package JDBCDriver;
import java.math.BigDecimal;

import java.sql.*;
import PREDATOR.*;

/**
 * <P>A ResultSet provides access to a table of data generated by
 * executing a Statement. The table rows are retrieved in
 * sequence. Within a row its column values can be accessed in any
 * order.
 * 
 * <P>A ResultSet maintains a cursor pointing to its current row of
 * data.  Initially the cursor is positioned before the first row.
 * The 'next' method moves the cursor to the next row.
 *
 * <P>The getXXX methods retrieve column values for the current
 * row.  You can retrieve values either using the index number of the
 * column, or by using the name of the column.  In general using the 
 * column index will be more efficient.  Columns are numbered from 1.
 *
 * <P>For maximum portability, ResultSet columns within each row should be
 * read in left-to-right order and each column should be read only once.
 *
 * <P>For the getXXX methods, the JDBC driver attempts to convert the
 * underlying data to the specified Java type and returns a suitable
 * Java value.  See the JDBC specification for allowable mappings
 * from SQL types to Java types with the ResultSet.getXXX methods.
 *
 * <P>Column names used as input to getXXX methods are case insensitive.
 * When performing a getXXX using a column name, if several columns have
 * the same name, then the value of the first matching column will be
 * returned.
 *
 * <P>A ResultSet is automatically closed by the Statement that
 * generated it when that Statement is closed, re-executed, or is used
 * to retrieve the next result from a sequence of multiple results.
 * 
 * <P>The number, types and properties of a ResultSet's columns are
 * provided by the ResulSetMetaData object returned by the getMetaData
 * method.
 *
 * @see Statement#executeQuery 
 * @see Statement#getResultSet 
 * @see ResultSetMetaData 
 */

public class ResultSetImpl implements java.sql.ResultSet {
	
	/**
	 * Statement and query result
	 */
	StatementImpl stmt;
	QueryResult qr;
	
	/**
	* index of the current row.
	*/
	private int current_row_index = 0;

	/** 
	* The meta data
	*/
	private ResultSetMetaDataImpl meta;

	/** 
	* A hashtable that maps column names to columns
	*/
	private java.util.Hashtable column_map = null;
	
	/**
	 * The stream of records in the result set
	 */
	private RecordStreamResult  cursor;
	
	/**
	 * The current row in the result set.
	 */
	private PREDATOR.ADT.DataObject record[];
	
	/**
	 * Last value red with a call to one of the getXXX function
	 */
	private PREDATOR.ADT.DataObject object;
	
	/**
	 * Constructor
	 */
	public ResultSetImpl(QueryResult qr, StatementImpl stmt)
		throws java.sql.SQLException
	{
		this.stmt = stmt;
		this.qr   = qr;
		
		if (qr == null) {
			// There was no result to the Query
			this.cursor = null;
		} else if (qr instanceof RecordStreamResult) {
			// Process the record stream
			this.cursor = (RecordStreamResult) qr;
			this.current_row_index = 0;
		} else {
			throw new java.sql.SQLException();
		}
		
	}
	
  /**
   * A ResultSet is initially positioned before its first row; the
   * first call to next makes the first row the current row; the
   * second call makes the second row the current row, etc. 
   *
   * <P>If an input stream from the previous row is open, it is
   * implicitly closed. The ResultSet's warning chain is cleared
   * when a new row is read.
   *
   * @return true if the new current row is valid; false if there
   * are no more rows 
   */
  public boolean next() throws java.sql.SQLException 
  {
		  if (cursor != null) 
		  {
			  try
			  {
				record = cursor.getNextRecord(null);
				if (record == null) 
					return false;
				else
				{
					current_row_index ++;
					return true;
				}
			  }
			  catch (PREDATOR.ADT.BadSubclassSelectorException e)
			  {
				  throw new java.sql.SQLException();
			  }
			  catch (InterruptedException ie)
			  {
				  throw new java.sql.SQLException();
			  }
		  } 
		  else 
			  return false;

  }

  /**
   * In some cases, it is desirable to immediately release a
   * ResultSet's database and JDBC resources instead of waiting for
   * this to happen when it is automatically closed; the close
   * method provides this immediate release.
   *
   * <P><B>Note:</B> A ResultSet is automatically closed by the
   * Statement that generated it when that Statement is closed,
   * re-executed, or is used to retrieve the next result from a
   * sequence of multiple results. A ResultSet is also automatically
   * closed when it is garbage collected.  
   */
  public void close() throws java.sql.SQLException {
  }

  /**
   * A column may have the value of SQL NULL; wasNull reports whether
   * the last column read had this special value.
   * Note that you must first call getXXX on a column to try to read
   * its value and then call wasNull() to find if the value was
   * the SQL NULL.
   *
   * @return true if last column read was SQL NULL
   */
  public boolean wasNull() throws java.sql.SQLException {
    return (object == null);
  }
    
  //======================================================================
  // Methods for accessing results by column index
  //======================================================================

  /**
   * Get the value of a column in the current row as a Java String.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return the column value; if the value is SQL NULL, the result is null
   */
  public String getString(int columnIndex) throws java.sql.SQLException {
		object = record[columnIndex];
		if (object == null) 
			return null;
		else
			return object.toString(cursor.getSchema().getAttribute(columnIndex).getMetaInfo());
  }

  /**
   * Get the value of a column in the current row as a Java boolean.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return the column value; if the value is SQL NULL, the result is false
   */
  public boolean getBoolean(int columnIndex) throws java.sql.SQLException {
	  try {
		  String str = getString(columnIndex);
		  if (str == null) return false;
		  if (str.equals("0")) return false;
		  return true;
	  } catch (Exception e) {
		  throw new SQLException(e + " ");
	  }
  }

  /**
   * Get the value of a column in the current row as a Java byte.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return the column value; if the value is SQL NULL, the result is 0
   */
  public byte getByte(int columnIndex) throws java.sql.SQLException {
		String str = getString(columnIndex);
		if (str == null) return 0;
		if (str.equals("")) {
			return 0;
		} else {
			return (byte) str.charAt(0);
		} 
  }

  /**
   * Get the value of a column in the current row as a Java short.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return the column value; if the value is SQL NULL, the result is 0
   */
  public short getShort(int columnIndex) throws java.sql.SQLException {
		String str = getString(columnIndex);
		if (str == null) return 0;
		
		if (str.equals("")) return 0;

		try {
			return (short) (Integer.valueOf(str).intValue());
		} catch ( NumberFormatException e ) {
			throw new SQLException(e + " ");
		}
  }

  /**
   * Get the value of a column in the current row as a Java int.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return the column value; if the value is SQL NULL, the result is 0
   */
  public int getInt(int columnIndex) throws java.sql.SQLException {
		String str = getString(columnIndex);

		if ( str == null) return 0;

		try {
			return Integer.valueOf(str).intValue();
		} catch (NumberFormatException e) {
			throw new SQLException(e + " ");
		}
  }

  /**
   * Get the value of a column in the current row as a Java long.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return the column value; if the value is SQL NULL, the result is 0
   */
  public long getLong(int columnIndex) throws java.sql.SQLException {
		String str = getString(columnIndex);

		if (str == null) return 0;

		try {
			return Long.valueOf(str).longValue();
		} catch (NumberFormatException e) {
			throw new SQLException(e + " ");
		}
  }

  /**
   * Get the value of a column in the current row as a Java float.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return the column value; if the value is SQL NULL, the result is 0
   */
  public float getFloat(int columnIndex) throws java.sql.SQLException {
		String str = getString(columnIndex);

		if (str == null) return 0;

		try {
			return Float.valueOf(str).floatValue();
		} catch (NumberFormatException e) {
			throw new SQLException(e + " ");
		}
  }

  /**
   * Get the value of a column in the current row as a Java double.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return the column value; if the value is SQL NULL, the result is 0
   */
  public double getDouble(int columnIndex) throws java.sql.SQLException {
		String str = getString(columnIndex);

		if (str == null) return 0;

		try {
			return Double.valueOf(str).doubleValue();
		} catch (NumberFormatException e) {
			throw new SQLException(e + " ");
		}
  }

  /**
   * Get the value of a column in the current row as a java.lang.BigDecimal object.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @param scale the number of digits to the right of the decimal
   * @return the column value; if the value is SQL NULL, the result is null
   */
  public BigDecimal getBigDecimal(int columnIndex, int scale)
  throws java.sql.SQLException {
    return null;
  }

  /**
   * Get the value of a column in the current row as a Java byte array.
   * The bytes represent the raw values returned by the driver.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return the column value; if the value is SQL NULL, the result is null
   */
  public byte[] getBytes(int columnIndex) throws java.sql.SQLException {
		String str = getString(columnIndex);

		byte b[];

		if (str == null ) return null;

		b = new byte[str.length() + 10];

		str.getBytes(0, str.length(), b,0);
		return b;
  }

  /**
   * Get the value of a column in the current row as a java.sql.Date object.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return the column value; if the value is SQL NULL, the result is null
   */
  public java.sql.Date getDate(int columnIndex) throws java.sql.SQLException {
		String str = getString(columnIndex);

		if (str == null) return null;

		try {
			java.util.Date d = new java.util.Date(str);
			return new java.sql.Date(d.getYear(), d.getMonth(), d.getDate());
		} catch (Exception e) {
			throw new SQLException("Date format error :" + e + " ");
		}

  }

  /**
   * Get the value of a column in the current row as a java.sql.Time object.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return the column value; if the value is SQL NULL, the result is null
   */
  public java.sql.Time getTime(int columnIndex) throws java.sql.SQLException {
    return null;
  }

  /**
   * Get the value of a column in the current row as a java.sql.Timestamp object.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return the column value; if the value is SQL NULL, the result is null
   */
  public java.sql.Timestamp getTimestamp(int columnIndex)
  throws java.sql.SQLException {
    return null;
  }

  /**
   * A column value can be retrieved as a stream of ASCII characters 
   * and then read in chunks from the stream.  This method is particularly
   * suitable for retrieving large LONGVARCHAR values.  The JDBC driver will
   * do any necessary conversion from the database format into ASCII.
   *
   * <P><B>Note:</B> All the data in the returned stream must be
   * read prior to getting the value of any other column. The next
   * call to a get method implicitly closes the stream. . Also, a
   * stream may return 0 for available() whether there is data
   * available or not.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return a Java input stream that delivers the database column value
   * as a stream of one byte ASCII characters.  If the value is SQL NULL
   * then the result is null.  
   */
  public java.io.InputStream getAsciiStream(int columnIndex)
  throws java.sql.SQLException {
    return null;
  }

  /**
   * A column value can be retrieved as a stream of Unicode characters 
   * and then read in chunks from the stream.  This method is particularly
   * suitable for retrieving large LONGVARCHAR values.  The JDBC driver will
   * do any necessary conversion from the database format into Unicode.
   *
   * <P><B>Note:</B> All the data in the returned stream must be
   * read prior to getting the value of any other column. The next
   * call to a get method implicitly closes the stream. . Also, a
   * stream may return 0 for available() whether there is data
   * available or not.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return a Java input stream that delivers the database column value
   * as a stream of two byte Unicode characters.  If the value is SQL NULL
   * then the result is null.  
   */
  public java.io.InputStream getUnicodeStream(int columnIndex)
  throws java.sql.SQLException {
    return null;
  }

  /**
   * A column value can be retrieved as a stream of uninterpreted bytes 
   * and then read in chunks from the stream.  This method is particularly
   * suitable for retrieving large LONGVARBINARY values.
   *
   * <P><B>Note:</B> All the data in the returned stream must be
   * read prior to getting the value of any other column. The next
   * call to a get method implicitly closes the stream. Also, a
   * stream may return 0 for available() whether there is data
   * available or not.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return a Java input stream that delivers the database column value
   * as a stream of uninterpreted bytes.  If the value is SQL NULL
   * then the result is null.  
   */
  public java.io.InputStream getBinaryStream(int columnIndex)
  throws java.sql.SQLException {
    return null;
  }


  //======================================================================
  // Methods for accessing results by column name
  //======================================================================

  /**
   * Get the value of a column in the current row as a Java String.
   *
   * @param columnName is the SQL name of the column
   * @return the column value; if the value is SQL NULL, the result is null
   */
  public String getString(String columnName) throws java.sql.SQLException {
		return getString(findColumn(columnName));
  }

  /**
   * Get the value of a column in the current row as a Java boolean.
   *
   * @param columnName is the SQL name of the column
   * @return the column value; if the value is SQL NULL, the result is false
   */
  public boolean getBoolean(String columnName) throws java.sql.SQLException {
		return getBoolean(findColumn(columnName));
  }

  /**
   * Get the value of a column in the current row as a Java byte.
   *
   * @param columnName is the SQL name of the column
   * @return the column value; if the value is SQL NULL, the result is 0
   */
  public byte getByte(String columnName) throws java.sql.SQLException {
		return getByte(findColumn(columnName));
  }

  /**
   * Get the value of a column in the current row as a Java short.
   *
   * @param columnName is the SQL name of the column
   * @return the column value; if the value is SQL NULL, the result is 0
   */
  public short getShort(String columnName) throws java.sql.SQLException {
		return getShort(findColumn(columnName));
  }

  /**
   * Get the value of a column in the current row as a Java int.
   *
   * @param columnName is the SQL name of the column
   * @return the column value; if the value is SQL NULL, the result is 0
   */
  public int getInt(String columnName) throws java.sql.SQLException {
		return getInt(findColumn(columnName));
  }

  /**
   * Get the value of a column in the current row as a Java long.
   *
   * @param columnName is the SQL name of the column
   * @return the column value; if the value is SQL NULL, the result is 0
   */
  public long getLong(String columnName) throws java.sql.SQLException {
		return getLong(findColumn(columnName));
  }

  /**
   * Get the value of a column in the current row as a Java float.
   *
   * @param columnName is the SQL name of the column
   * @return the column value; if the value is SQL NULL, the result is 0
   */
  public float getFloat(String columnName) throws java.sql.SQLException {
		return getFloat(findColumn(columnName));
  }

  /**
   * Get the value of a column in the current row as a Java double.
   *
   * @param columnName is the SQL name of the column
   * @return the column value; if the value is SQL NULL, the result is 0
   */
  public double getDouble(String columnName) throws java.sql.SQLException {
		return getDouble(findColumn(columnName));
  }

  /**
   * Get the value of a column in the current row as a java.lang.BigDecimal object.
   *
   * @param columnName is the SQL name of the column
   * @param scale the number of digits to the right of the decimal
   * @return the column value; if the value is SQL NULL, the result is null
   */
  public BigDecimal getBigDecimal(String columnName, int scale)
  throws java.sql.SQLException {
    return null;
  }

  /**
   * Get the value of a column in the current row as a Java byte array.
   * The bytes represent the raw values returned by the driver.
   *
   * @param columnName is the SQL name of the column
   * @return the column value; if the value is SQL NULL, the result is null
   */
  public byte[] getBytes(String columnName) throws java.sql.SQLException {
	return getBytes(findColumn(columnName)) ;
  }

  /**
   * Get the value of a column in the current row as a java.sql.Date object.
   *
   * @param columnName is the SQL name of the column
   * @return the column value; if the value is SQL NULL, the result is null
   */
  public java.sql.Date getDate(String columnName) throws java.sql.SQLException {
     return getDate(findColumn(columnName));
  }

  /**
   * Get the value of a column in the current row as a java.sql.Time object.
   *
   * @param columnName is the SQL name of the column
   * @return the column value; if the value is SQL NULL, the result is null
   */
  public java.sql.Time getTime(String columnName) throws java.sql.SQLException {
    return null;
  }

  /**
   * Get the value of a column in the current row as a java.sql.Timestamp object.
   *
   * @param columnName is the SQL name of the column
   * @return the column value; if the value is SQL NULL, the result is null
   */
  public java.sql.Timestamp getTimestamp(String columnName)
  throws java.sql.SQLException {
    return null;
  }

  /**
   * A column value can be retrieved as a stream of ASCII characters 
   * and then read in chunks from the stream.  This method is particularly
   * suitable for retrieving large LONGVARCHAR values.  The JDBC driver will
   * do any necessary conversion from the database format into ASCII.
   *
   * <P><B>Note:</B> All the data in the returned stream must
   * be read prior to getting the value of any other column. The
   * next call to a get method implicitly closes the stream.
   *
   * @param columnName is the SQL name of the column
   * @return a Java input stream that delivers the database column value
   * as a stream of one byte ASCII characters.  If the value is SQL NULL
   * then the result is null.
   */
  public java.io.InputStream getAsciiStream(String columnName)
  throws java.sql.SQLException {
    return null;
  }

  /**
   * A column value can be retrieved as a stream of Unicode characters 
   * and then read in chunks from the stream.  This method is particularly
   * suitable for retrieving large LONGVARCHAR values.  The JDBC driver will
   * do any necessary conversion from the database format into Unicode.
   *
   * <P><B>Note:</B> All the data in the returned stream must
   * be read prior to getting the value of any other column. The
   * next call to a get method implicitly closes the stream.
   *
   * @param columnName is the SQL name of the column
   * @return a Java input stream that delivers the database column value
   * as a stream of two byte Unicode characters.  If the value is SQL NULL
   * then the result is null.
   */
  public java.io.InputStream getUnicodeStream(String columnName)
  throws java.sql.SQLException {
    return null;
  }

  /**
   * A column value can be retrieved as a stream of uninterpreted bytes 
   * and then read in chunks from the stream.  This method is particularly
   * suitable for retrieving large LONGVARBINARY values.
   *
   * <P><B>Note:</B> All the data in the returned stream must
   * be read prior to getting the value of any other column. The
   * next call to a get method implicitly closes the stream.
   *
   * @param columnName is the SQL name of the column
   * @return a Java input stream that delivers the database column value
   * as a stream of uninterpreted bytes.  If the value is SQL NULL
   * then the result is null.
   */
  public java.io.InputStream getBinaryStream(String columnName)
  throws java.sql.SQLException {
    return null;
  }


  //=====================================================================
  // Advanced features:
  //=====================================================================

  /**
   * <p>The first warning reported by calls on this ResultSet is
   * returned. Subsequent ResultSet warnings will be chained to this
   * SQLWarning.
   *
   * <P>The warning chain is automatically cleared each time a new
   * row is read.
   *
   * <P><B>Note:</B> This warning chain only covers warnings caused
   * by ResultSet methods.  Any warning caused by statement methods
   * (such as reading OUT parameters) will be chained on the
   * Statement object. 
   *
   * @return the first SQLWarning or null 
   */
  public SQLWarning getWarnings() throws java.sql.SQLException {
    return null;
  }

  /**
   * After this call getWarnings returns null until a new warning is
   * reported for this ResultSet.  
   */
  public void clearWarnings() throws java.sql.SQLException {
  }

  /**
   * Get the name of the SQL cursor used by this ResultSet.
   *
   * <P>In SQL, a result table is retrieved through a cursor that is
   * named. The current row of a result can be updated or deleted
   * using a positioned update/delete statement that references the
   * cursor name. 
   * 
   * <P>JDBC supports this SQL feature by providing the name of the
   * SQL cursor used by a ResultSet. The current row of a ResultSet
   * is also the current row of this SQL cursor.
   *
   * <P><B>Note:</B> If positioned update is not supported a
   * java.sql.SQLException is thrown
   *
   * @return the ResultSet's SQL cursor name
   */
  public String getCursorName() throws java.sql.SQLException {
    return null;
  }

  /**
   * The number, types and properties of a ResultSet's columns
   * are provided by the getMetaData method.
   *
   * @return the description of a ResultSet's columns
   */
  public ResultSetMetaData getMetaData() throws java.sql.SQLException {
     return meta = new ResultSetMetaDataImpl(stmt, qr, this);
  }

  /**
   * <p>Get the value of a column in the current row as a Java object.
   *
   * <p>This method will return the value of the given column as a
   * Java object.  The type of the Java object will be the default
   * Java Object type corresponding to the column's SQL type,
   * following the mapping specified in the JDBC spec.
   *
   * <p>This method may also be used to read datatabase specific abstract
   * data types.
   *
   * @param columnIndex the first column is 1, the second is 2, ...
   * @return A java.lang.Object holding the column value.  
   */
  public Object getObject(int columnIndex) throws java.sql.SQLException {
    return (Object)record[columnIndex];
  }

  /**
   * <p>Get the value of a column in the current row as a Java object.
   *
   * <p>This method will return the value of the given column as a
   * Java object.  The type of the Java object will be the default
   * Java Object type corresponding to the column's SQL type,
   * following the mapping specified in the JDBC spec.
   *
   * <p>This method may also be used to read datatabase specific abstract
   * data types.
   *
   * @param columnName is the SQL name of the column
   * @return A java.lang.Object holding the column value.  
   */
  public Object getObject(String columnName) throws java.sql.SQLException {
		return (Object)record[findColumn(columnName)];
  }

  //----------------------------------------------------------------

  /**
   * Map a Resultset column name to a ResultSet column index.
   *
   * @param columnName the name of the column
   * @return the column index
   */
  public int findColumn(String columnName) throws java.sql.SQLException {
		meta = new ResultSetMetaDataImpl(stmt, qr, this);
		while (true) {
			try {
				int i = 0;
				if (meta.getColumnName(i).equals(columnName)) return i;
				i++;
			} catch (Exception e) {
				return -1;
			}
		}
  }

};

