package PREDATOR.ADT;

import java.io.*;
import PREDATOR.*;

/**
 * The base class for all E-ADTs.  All java ADT classes subclass this
 * class.  This class essentially tells the Java interface how to interact
 * with a data type.  It provides the following information to the client:
 * <OL>
 *   <LI> Whether this type is variable size in a record
 *   <LI> The required DataObject superclass
 * </OL>
 * An DataObject object is akin to an instantiation of a data type.  To
 * permit interface extensibility, before creating a new DataObject,
 * the ADT gives the interface the chance to subclass the DataObject
 * to provide additional functionality.  (This all occurs in the 
 * NewObject function.)  The ObjectClass function is used by the
 * interface to decide which subclass to offer the ADT.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 */
public abstract class ADT {
  /// The name of the ADT
  private String name;
  /// The server ID number of the ADT
  private int id;

  /**
   * Sets the name of this ADT.
   * @param Name The name of this ADT
   */
  public final void setName(String Name) { name = Name; }
  /**
   * Returns the (server) name of the ADT.
   * @return The name of this ADT
   */
  public final String getName() { return name; }
  /**
   * Sets the type ID of this ADT.
   * @param The type ID number of the ADT (server-side)
   */
  public final void setTypeId (int TypeId) { id = TypeId; }
  /**
   * Returns the (server) ID number of the ADT.
   * @return The ID number of this ADT
   */
  public final int getTypeId() { return id; }
  /**
   * Determines if the binary version of the ADT is variable size.  By
   * default, this method returns false.
   * @param MetaInfo The meta-info
   * @return true if the in-record version of the ADT is variable size,
   *         false otherwise
   */
  public boolean isVariableSize(ADTMetaInfo MetaInfo) { return false; }
  /**
   * Returns the class of the ADT's object data.  For instance, the
   * integer ADT returns IntegerObject.
   * @param MetaInfo The meta-info
   * @return The class of the ADT's object data
   */
  public abstract Class getObjectClass(ADTMetaInfo MetaInfo);
  /**
   * Returns the class of the ADT's meta info.  For instance, the
   * audio ADT returns AudioMetaInfo.
   * @return The class of the ADT's meta info
   */
  public Class getMetaInfoClass()
  {
    return null;
  }
  /**
   * Returns a new ADTMetaInfo suitable for use with this ADT.
   * @param bytes The binary version of the meta-info
   * @exception IllegalAccessException If the DataObject class is inaccessible
   * @exception InstantiationException If an error occurs while instantiating
   *                                   the DataObject class's object
   * @return A new ADTMetaInfo object suitable for use with this ADT
   */   
  public final ADTMetaInfo newMetaInfo(byte []bytes)
    throws IllegalAccessException, 
           InstantiationException
  {
    if (getMetaInfoClass() == null)
      return null;
    ADTMetaInfo MetaInfo = (ADTMetaInfo)getMetaInfoClass().newInstance();
    MetaInfo.setBytes(bytes);
    return MetaInfo;
  }
  /**
   * Returns a new DataObject suitable for use with this ADT.
   * It permits the GUI to specify a subclass of the original ObjectData
   * class to that further GUI-oriented functionality can be used.
   * @param bytes The binary version of the object
   * @param subclasser A object which will select a subclass of the
   *                   DataObject's class to extend functionality, or
   *                   null if no subclass is desired
   * @param MetaInfo The meta-info
   * @exception IllegalAccessException If the DataObject class is inaccessible
   * @exception InstantiationException If an error occurs while instantiating
   *                                   the DataObject class's object
   * @exception BadSubclassSelectorException If the subclasser returns a
   *                                         null class or one which is not
   *                                         a subclass of the parameter class
   * @return A new ObjectData object suitable for use with this ADT
   * @see SubclassSelector
   */   
  public final DataObject newObject(byte []bytes,
				    ADTMetaInfo MetaInfo,
				    SubclassSelector subclasser) 
    throws IllegalAccessException, 
           InstantiationException, 
           BadSubclassSelectorException
  {
    DataObject object;
    Class dataObjectClass;
    if (subclasser != null) {
		try {
      dataObjectClass = subclasser.getSubclass(getObjectClass(MetaInfo));
		} catch (ClassNotFoundException X) {
			throw new BadSubclassSelectorException(X.getMessage()); }
      if (dataObjectClass == null)
	throw new BadSubclassSelectorException("The returned class was null");
      if (!getObjectClass(MetaInfo).isAssignableFrom(dataObjectClass))
	throw new 
	  BadSubclassSelectorException("The returned class was not a direct " +
				       "subclass of the supplied class nor " +
				       "the class itself.");
    } else
      dataObjectClass = getObjectClass(MetaInfo);
    object = (DataObject)(dataObjectClass.newInstance());
    object.setBytes(bytes);
    return object;
  }
}
