package PREDATOR.ADT;

import java.io.*;

/**
 * Encapsulates all necessary object data for ADT display, querying,
 * etc.  For instance, an integer's object data will contain only its
 * value.  An image's object data will contain OID information, and
 * a handle on an in-memory version of the image.
 *
 * All data members of an ObjectData class should be declared either
 * public or protected so that subclasses can access the underlying
 * data.  In that sense, this class is a "client-side" copy of the
 * server's binary data.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see CachingObject
 * @see MIMETypeObject
 * @see DisplayObject
 * @see CallbackObject
 */
public class DataObject {
  /**
   * Set if this object is a null value
   */
  public final static byte NULL = 0x01;
  /**
   * Set if this object is a suspended value.  If this is set, the 
   * object's suspended computation id is valid; otherwise, it is not.
   */
  public final static byte SUSPENDED = 0x02;
  /**
   * Set if this object's entire value is in-field.
   */
  public final static byte IN_FIELD = 0x04;
  /**
   * Set if this object is in a compressed format.
   */
  public final static byte COMPRESSED = 0x08;
  /**
   * Set if this object is in little-endian format.
   */
  public final static byte LITTLE_ENDIAN = 0x10;
  /**
   * The binary format of this object.
   */
  protected byte[] bytes;
  /**
   * Checks to see if a flag is set in the status byte of the object.
   * The status byte is the first byte of every object.
   */
  public final boolean is(byte flag)
  {
    return ((bytes[0] & flag) != 0);
  }
  /**
   * Every data object is stored as a contiguous sequence of bytes at 
   * the server.  This function assigns those bytes to the object.
   * @param bytes The binary form of this object.
   */
  public void setBytes(byte[] bytes) 
  {
    this.bytes = bytes;
  }
  /**
   * Gets the binary form of this object.
   */
  public byte[] getBytes()
  {
    return bytes;
  }
  /**
   * Provides a textual version of this object suitable for user-display.
   * @param MetaInfo The meta-info for this object
   * @return A textual version of this object
   */
  public String toString(ADTMetaInfo MetaInfo)
  {
    if (is(NULL))
      return "Null";
    else if (is(SUSPENDED))
      return "Suspended";
    else return null;
  }
}
