package PREDATOR.ADT;

import java.awt.*;
import java.io.*;
import java.net.*;
import PREDATOR.*;

/**
 * Encapsulates all in-memory data regarding an polygon ADT object.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see ADT_polygon
 */
public class PolygonObject extends DataObject 
  implements CachingObject, DisplayObject 
{
  /**
   * The X points.
   */
  double []Xpoints;
  /**
   * The Y points.
   */
  double []Ypoints;
  /**
   * Prints the value of the polygon
   * @param MetaInfo The meta-info for this object
   * @return A string holding the value of the polygon
   */
  public String toString(ADTMetaInfo MetaInfo) {
    String tmp = super.toString(MetaInfo);
    if (tmp != null)
      return (tmp);
    try {
      DataInputStream dataStream = 
	new DataInputStream(new ByteArrayInputStream(getBytes(), 12,
						     getBytes().length - 12));
      int numVertices = (int)dataStream.readInt();
      double Xmin = dataStream.readDouble();
      double Ymin = dataStream.readDouble();
      double Xmax = dataStream.readDouble();
      double Ymax = dataStream.readDouble();
      return ((numVertices-1) + "-gon [(" +
	      Xmin + ", " + Ymin + "), (" + Xmax + ", " + Ymax + ")]");
    } catch (IOException X) {
      throw new PREDATOR.InternalError(X);
    }
  }
  /**
   * Callback objects have the option of supplying a byte string
   * containing additional information about the callback.  For
   * instance, an image object might specify one of many stored
   * resolutions.
   * @return A binary message
   */
  public byte [] getMessage()
  {
    return null;
  }
  /**
   * Performs a callback to the server, using the given URL.  The URL
   * may be opened as a stream, if desired.  The contents of the URL
   * will be determined by what the server-side ADT writes out in the
   * <TT>ServiceCallback()<TT> function (in <TT>types.h</TT>).
   * @param CallbackURL The URL of the callback object
   * @param MetaInfo The meta-info for this object
   */
  public void callback(URL CallbackURL, ADTMetaInfo MetaInfo)
  {
    try {
      DataInputStream inputStream = 
	new DataInputStream(CallbackURL.openStream());
      DataInputStream dataStream = 
        new DataInputStream(new ByteArrayInputStream(getBytes(), 4, 
                                                     getBytes().length - 4));
      int numVertices = (int)dataStream.readInt();
      Xpoints = new double[numVertices];
      Ypoints = new double[numVertices];
      for (int i = 0; i < numVertices; i++) {
	Xpoints[i] = inputStream.readDouble();
      }
      for (int i = 0; i < numVertices; i++)
	Ypoints[i] = inputStream.readDouble();
    } catch (Exception X) {
      throw new PREDATOR.InternalError(X);
    }
  }
  /**
   * Determines if this object has a handle on the "full value", or if
   * a callback or deserialization is necessary.
   * @param MetaInfo The meta-info for this object
   * @return true if this object does not have its "full value" available
   */
  public boolean requiresCallback(ADTMetaInfo MetaInfo)
  {
    return (Xpoints == null);
  }
  /**
   * Displays an object indepently.  Implementations of this function
   * should be non-blocking for best performance.  (For instance, an
   * image should be displayed in an independent frame and execution
   * should proceed while the frame is still open.)
   * @param MetaInfo The meta-info for this object
   * @exception RequiresCallbackException when a callback is required
   *                                      before a display can be made
   */
  public void display(ADTMetaInfo MetaInfo) 
    throws RequiresCallbackException 
  {
    if (requiresCallback(MetaInfo))
      throw new RequiresCallbackException();
    // Launch the thread
    try {
      DataInputStream dataStream = 
        new DataInputStream(new ByteArrayInputStream(getBytes(), 4, 
                                                     getBytes().length - 4));
      int numVertices = (int)dataStream.readInt();
      double Xmin = dataStream.readDouble();
      double Ymin = dataStream.readDouble();
      double Xmax = dataStream.readDouble();
      double Ymax = dataStream.readDouble();
      new Thread(new PolygonPlayer(Xmin, Ymin, Xmax, Ymax, 
				   Xpoints, Ypoints, toString())).run();    
    } catch (IOException X) {
      throw new PREDATOR.InternalError(X);
    }
  }
}

/**
 * The polygon ADT's independent method of display; a thread which displays 
 * it in a frame.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version May 28, 1997
 * @see ADT_polygon
 */
class PolygonPlayer extends Frame implements Runnable {
  private double Xmin;
  private double Ymin;
  private double Xmax;
  private double Ymax;
  private int numVertices;
  private double Xpoints[];
  private double Ypoints[];
  private final int MAX_EDGE = 500;  
  private final int RADIUS = 3;

  /**
   * Constructor.
   * @param values The integer values of the polygon
   * @param width the width of the polygon
   * @param height the height of the polygon
   * @param frameTitle The title given to the frame
   */
  public PolygonPlayer(double Xmin,
		       double Ymin,
		       double Xmax,
		       double Ymax,
		       double Xpoints[],
		       double Ypoints[],
		       String frameTitle) {
    super(frameTitle);
    this.Xmin = Xmin;
    this.Xmax = Xmax;
    this.Ymin = Ymin;
    this.Ymax = Ymax;
    this.Xpoints = Xpoints;
    this.Ypoints = Ypoints;
    setBackground(Color.black);
    resize(MAX_EDGE, MAX_EDGE + 50);
    show();
  }

  /**
   * The runnable's run() function;
   */
  public void run() {
    Draw(getGraphics());
    repaint();
  }

  private void Draw(Graphics g) {
    g.translate(50, 50);
    double Xrange = Xmax - Xmin;
    double Yrange = Ymax - Ymin;
    int X, Y;
    int height, width;
    width = height = (MAX_EDGE - 100);
    g.setColor(Color.yellow);
    for (int i = 0; i < Xpoints.length; i++) {
      X = (int)((double)width * ((Xpoints[i] - Xmin) / Xrange));
      Y = (int)((double)height * ((Ypoints[i] - Ymin) / Yrange));
      g.fillOval(X - RADIUS, Y - RADIUS, RADIUS * 2, RADIUS * 2);
      if (i != 0) {
	g.drawLine(X, Y,
		   (int)((double)width * ((Xpoints[i - 1] - Xmin) / Xrange)),
		   (int)((double)height * ((Ypoints[i - 1] - Ymin) /Yrange)));
      }
    }
  }
  /**
   * The frame's paint() function.  Repaints the polygon.
   * @param g The frame's graphics context.
   */
  public void paint(Graphics g) {
    Draw(g);
  }
  /**
   * Handles the destruction of the window.
   * @param event The event to handle
   * @return true if the event was handled, false otherwise
   */
  public boolean handleEvent(Event event) {
    if (event.id == Event.WINDOW_DESTROY) {
      dispose();	
      return true;
    }
    return false;
  }
}
