package PREDATOR.ADT;

import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.net.*;
import PREDATOR.*;

/**
 * Encapsulates all in-memory data regarding a raster ADT object.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see ADT_raster
 */
public class RasterObject 
  extends DataObject implements CachingObject, DisplayObject {
  /// The frequency of the raster
  public int frequency;
  /// The number of samples in the raster
  public int samples;
  /// The number of lines in the raster
  public int lines;
  /// The resolution of the raster
  public int resolution;
  /// The X coordinate of the min point
  public double Xmin;
  /// The Y coordinate of the min point
  public double Ymin;
  /// The X coordinate of the max point
  public double Xmax;
  /// The Y coordinate of the max point
  public double Ymax;
  /// The pixel data
  public int values[] = null;
  /**
   * Prints the value of the raster
   * @param MetaInfo The meta-info for this object
   * @return A string holding the value of the raster
   */
  public String toString(ADTMetaInfo MetaInfo) {
    String tmp = super.toString(MetaInfo);
    if (tmp != null)
      return (tmp);
    try {
      DataInputStream dataStream = 
	new DataInputStream(new ByteArrayInputStream(getBytes(), 4, 
						     getBytes().length - 4));
      frequency = dataStream.readInt();
      samples = dataStream.readInt();
      lines = dataStream.readInt();
      dataStream.close();
    } catch (IOException X) {
      throw new PREDATOR.InternalError(X);
    }
    return ("raster [" + samples + "x" + lines +"]");
  }
  /**
   * Callback objects have the option of supplying a byte string
   * containing additional information about the callback.  For
   * instance, an image object might specify one of many stored
   * resolutions.
   * @return A binary message
   */
  public byte [] getMessage()
  {
    return null;
  }
  /**
   * Performs a callback to the server, using the given URL.  The URL
   * may be opened as a stream, if desired.  The contents of the URL
   * will be determined by what the server-side ADT writes out in the
   * <TT>ServiceCallback()<TT> function (in <TT>types.h</TT>).
   * @param CallbackURL The URL of the callback object
   * @param MetaInfo The meta-info for this object
   */
  public void callback(URL CallbackURL, ADTMetaInfo MetaInfo)
  {
    try {
      DataInputStream inputStream = 
	new DataInputStream(CallbackURL.openStream());
      int numBytes = inputStream.readInt();
      values = new int[numBytes / 2];
      for (int i = 0; i < numBytes / 2; i++) {
	values[i] = inputStream.readShort();
      }
    } catch (Exception X) {
      throw new PREDATOR.InternalError(X);
    }
  }
  /**
   * Determines if this object has a handle on the "full value", or if
   * a callback or deserialization is necessary.
   * @param MetaInfo The meta-info for this object
   * @return true if this object does not have its "full value" available
   */
  public boolean requiresCallback(ADTMetaInfo MetaInfo)
  {
    return (values == null);
  }
  /**
   * Displays an object indepently.  Implementations of this function
   * should be non-blocking for best performance.  (For instance, an
   * image should be displayed in an independent frame and execution
   * should proceed while the frame is still open.)
   * @param MetaInfo The meta-info for this object
   * @exception RequiresCallbackException when a callback is required
   *                                      before a display can be made
   */
  public void display(ADTMetaInfo MetaInfo) 
    throws RequiresCallbackException 
  {
    if (requiresCallback(MetaInfo))
      throw new RequiresCallbackException();
    try {
      DataInputStream dataStream = 
        new DataInputStream(new ByteArrayInputStream(getBytes(), 4, 
                                                     getBytes().length - 4));
      int frequency = dataStream.readInt();
      int samples = dataStream.readInt();
      int lines = dataStream.readInt();
      /*
       * The RasterStruct in PREDATOR contains a padding of 4 bytes for the 
       * doubles to be aligned.
       */
      dataStream.readInt();
      double Xmin = dataStream.readDouble();
      double Ymin = dataStream.readDouble();
      double Xmax = dataStream.readDouble();
      double Ymax = dataStream.readDouble();
      int resolution = dataStream.readInt();
      // Launch the thread
      new Thread(new RasterPlayer(values, samples, lines, toString())).run();
    } catch (IOException X) {
      throw new PREDATOR.InternalError(X);
    }
  }
}

/**
 * The raster ADT's independent method of display; a thread which displays 
 * it in a frame.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see PREDATOR.ADT.ADT_raster
 */
class RasterPlayer extends Frame implements Runnable, ImageObserver {
  private final int MAX_EDGE = 500;
  /// The raster displayed in the frame
  private Image image;
  private int width;
  private int height;
  private int values[];
  private int maxEdge;

  /**
   * Constructor.
   * @param values The integer values of the raster
   * @param width the width of the raster
   * @param height the height of the raster
   * @param frameTitle The title given to the frame
   */
  public RasterPlayer(int values[], 
		      int width,
		      int height,
		      String frameTitle) {
    super(frameTitle);
    this.values = values;
    this.width = width;
    this.height = height;
    maxEdge = Math.min(MAX_EDGE, Math.max(width, height));
    setBackground(Color.black);
    resize(250,250);
    show();
  }

  /**
   * The runnable's run() function; fetches the raster via http and
   * displays it.
   */
  public void run() {
    if (values == null)
      return;
    int size = width * height;
    int pixels[] = new int[size];
    for (int i = 0; i < size; i++) {
      pixels[i] = (int) (255 << 24) | (values[i] << 16) | (150 << 8) | 150;
    }
    image = createImage(new MemoryImageSource(width, 
					      height, pixels, 0, width));
    if ((width > height) && (width > MAX_EDGE)) {
      height /= (double)width / (double)MAX_EDGE;
      width = MAX_EDGE;
    } else if ((width < height) && (height > MAX_EDGE)) {
      width /= (double)height / (double)MAX_EDGE;
      height = MAX_EDGE;
    }
    getGraphics().drawImage(image, 0, 0, width, height, this);
    repaint();
  }

  /**
   * The frame's paint() function.  Repaints the raster.
   * @param g The frame's graphics context.
   */
  public void paint(Graphics g) {
    if (image != null) 
      g.drawImage(image, 0, 0, width, height, this);
  }

  /**
   * Used by AWT to handle asynchronous transfer of images.
   * @param img The raster being observed
   * @param infoFlags The OR of some flags (see the interface's API)
   * @param x an x coordinate
   * @param y a y coordinate
   * @param width the width of the raster
   * @param height the height of the raster
   * @return true if further calls to the raster observer are necesary,
   *         false otherwise
   */
  public boolean imageUpdate(Image img, int infoflags, int x, int y,
			     int width, int height) {
    if ((infoflags & (WIDTH | HEIGHT)) != 0) {
      resize(width, height);
    }
    boolean done = ((infoflags & (ERROR | FRAMEBITS | ALLBITS)) != 0);
    if (done) {
      resize(width, height);
      repaint(0);
    } else {
      repaint(100);
    }
    return (!done);
  }

  /**
   * Handles the destruction of the window.
   * @param event The event to handle
   * @return true if the event was handled, false otherwise
   */
  public boolean handleEvent(Event event) {
    if (event.id == Event.WINDOW_DESTROY) {
      dispose();	
      return true;
    }
    return false;
  }
}
