package PREDATOR.ADT;

/**
 * The interface implemented by objects which wish to subclass the
 * object data objects used by the ADTs.  Interface designers may wish
 * to use an object which implements this interface so that they may
 * add some type of functionality to the data type objects.  For instance,
 * the designer of an interface to the Sequoia 2000 benchmark data may wish
 * to have all of the GIS (Geographic Information System) types to be
 * displayed on a map of California.  The GIS types (Point, Polygon, and 
 * Raster E-ADTs) currently do not have this capability.  To extend their
 * functionality, the interface designer would take the following steps:
 * <OL>
 *   <LI>Design an interface (in the Java sense) which defines methods
 *       to extend functionality.  In the Sequoia example, the interface
 *       would look much like this:
 *       <PRE>
 *       public interface CommonDisplayObject {
 *          public void Draw(CaliforniaMap Map);
 *       }
 *       </PRE>
 *   <LI>Create subclasses of the RasterObject, PolygonObject, and 
 *       PointObject classes which implement the new interface described
 *       above.
 *   <LI>Create an object which implements the <TT>SubclassSelector</TT>
 *       interface.  In the <TT>getSubclass(Class)</TT> function, it should
 *       return the class of the subclass defined in the previous step,
 *       given the class of the original <TT>DataObject</TT> class.  For
 *       those <TT>DataObject</TT> subclasses which are not GIS types,
 *       the <TT>getSubclass(Class)</TT> function must return the original
 *       class.  For example, if the subclasses were given the prefix 
 *       <TT>GIS</TT>, the <TT>getSubclass(Class)</TT> function would be 
 *       implemented as follows: 
 *       <PRE>
 *       public Class getSubclass(Class SuperClass)
 *       {
 *         // Form the name of the subclass, if it exists
 *         String BaseClassName = SuperClass.getName();
 *         String SubclassName = "Sequoia.GIS" +
 *           BaseClassName.substring("PREDATOR.ADT.".length(), 
 *                                   BaseClassName.length());
 *         // Try to load the class.  If it exists, return it.  Otherwise
 *         // return the original class.
 *         Class returnClass = null;
 *         try {
 *            returnClass = Class.forName(SubclassName);
 *         } catch (java.lang.ClassNotFoundException X) {
 *           return SuperClass;
 *         }
 *         return returnClass;
 *       }
 *       </PRE>
 *       Note how this function works: when the Raster E-ADT asks the 
 *       <TT>SubclassSelector</TT> for a subclass of <TT>RasterObject</TT>,
 *       the function above attempts to load the class 
 *       <TT>Sequoia.GISRasterObject</TT> and substitute it.  If it
 *       cannot find the class, it returns the original class.  That insures
 *       that <TT>DataObject</TT>s such as <TT>IntegerData</TT> are 
 *       unaffected by the <TT>SubclassSelector</TT>.
 *
 *       As long as the <TT>getSubclass(Class)</TT> is written in this
 *       extensible manner, it is possible for classes added later to also
 *       assume the new functionality.  For instance, if a PolyLine E-ADT
 *       were added to the system, the Sequoia interface designer need only
 *       define the <TT>Sequoia.GISPolyLineObject</TT> class to allow
 *       it to draw itself on a map of California.  If the function checked
 *       to see if the incoming class was <TT>RasterObject</TT>,
 *       <TT>PointObject</TT> or <TT>PolygonObject</TT>, the
 *       <TT>SubclassSelector</TT> would be much less useful.
 *   <LI>Pass in the <TT>SubclassSelector</TT> object to all calls to
 *       <TT>RecordStreamResult.getNextRecord(SubclassSelector)</TT>.
 * </OL>
 * If all of these steps are followed, the objects returned from calls to
 * <TT>RecordStreamResult.getNextRecord(SubclassSelector)</TT> will be of the
 * class specified by the <TT>SubclassSelector</TT>.  (In the Sequoia 
 * example, they will be capable of drawing themselves on a common map
 * of California.
 * 
 * If the <TT>SubclassSelector</TT> supplied to the 
 * <TT>RecordStreamResult.getNextRecord(SubclassSelector)</TT> function
 * returns null or a class which is not either the supplied class or a 
 * <I>direct</I> subclass of the supplied class,
 * a <TT>BadSubclassSelectorException</TT> will be thrown.
 * @see DataObject
 * @see PREDATOR.RecordStreamResult#getNextRecord(SubclassSelector)
 * @see BadSubclassSelectorException
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 */
public class SubclassSelector {

  /// used by getGenFuncClass
  final String GENFUNC_PREFIX = "PREDATOR.UDF.Wrapper_";

  /**
   * Permits an interface to add functionality to <TT>DataObject</TT>
   * objects by selecting a direct subclass of that class which implements
   * an interface describing the new functionality.
   * @param SuperClass the class of the <TT>DataObject</TT> which might
   *                   need to be subclassed for added functionality
   * @return SuperClass, a <I>direct</I> subclass of it.  If something is
   *         returned which does not satisfy this criteria (null, for
   *         instance), a <TT>BadSubclassSelectorException</TT> exception
   *         will be thrown from the call to 
   *         <TT>RecordStreamResult.getNextRecord(SubclassSelector)</TT>.
   */
  public Class getSubclass(Class SuperClass)
       throws ClassNotFoundException
  {
    return SuperClass; // default: no subclassing
  } 

   /**
   * Locates the class which implements the generic function named UDFName.
   * Users can specify where they keep their UDFs.
   * @param UDFName The name of the UDF.
   * @return A class that implements the desired static method.
   */
  public Class getGenFuncClass(String UDFName)
       throws ClassNotFoundException
  {
    return Class.forName(GENFUNC_PREFIX + UDFName);
  }
    
}














