package PREDATOR;

import java.io.*;
import java.net.*;
import java.util.*;
import PREDATOR.ADT.*;

/**
 * The main interface for a Java applet or application to maintain a client 
 * connection
 * with a <FONT COLOR="#CC0000"><FONT SIZE=+1>P</FONT>REDATOR</FONT> server.  
 * It provides methods to connect and disconnect, and to process queries and 
 * callbacks.  A Connection object is thread-safe in that many threads can 
 * use it simultanously with no possibility of race conditions or blocking 
 * threads.
 * <BR>
 * An object of this class must be maintained for <I>each</I> client session
 * with the 
 * <FONT COLOR="#CC0000"><FONT SIZE=+1>P</FONT>REDATOR</FONT> server.  The
 * standard way to use this class is:
 * <OL>
 *   <LI>Create this object.
 *   <LI>Create a <TT>ServerSite</TT> object, which will identify the location
 *       of the <FONT COLOR="#CC0000"><FONT SIZE=+1>P</FONT>REDATOR</FONT> 
 *       server.
 *   <LI>Call the <TT>open(ServerSite)</TT> method of this class.
 *   <LI>Run queries (see below).
 *   <LI>Close the connection via a call to the <TT>close()</TT> method of
 *       class.
 * </OL>
 * To run a query, first create a <TT>Query</TT> object.  Then call 
 * <TT>processQuery(Query)</TT> and get the <TT>ResultObject</TT> it returns.
 * Then, using the <TT>instanceof</TT> operator, determine if the result
 * is a <TT>UnformattedResult</TT> or a <TT>RecordStreamResult</TT> object.
 * If desired, make callbacks for "full values" of callback objects.  Then,
 * end the query (freeing up server resources) by calling <TT>endQuery()</TT>.
 * <P>
 * A database server error may arise if the queries sent are inappropriate.
 * To check if the server has noted an error, use the <TT>isError()</TT> 
 * function.  If an error has occurred, the actual error (which provides 
 * messages and stack information) can be obtained via a call to 
 * <TT>getError()</TT>.  Applications should check for errors at the 
 * following times:
 * <OL>
 *   <LI>After calls to <TT>Connection.open(ServerSite S)</TT>
 *   <LI>After calls to <TT>Connection.endQuery()</TT> (or 
 *       <TT>Connection.processQuery(Query Q, true)</TT>).
 *   <LI>After calls to <TT>Connection.close()</TT>
 * </OL>
 *
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see ServerSite
 * @see Query
 * @see QueryResult
 * @see RecordStreamResult
 * @see UnformattedResult
 */
public class Connection {
  /**
   * The string which tells the server to start a new client session
   */
  protected final static String InitString = "NEW SESSION\n";
  /**
   * The string the server sends over in response to the InitString
   */
  protected final static String SERVER_SETUP_STRING = "Setup?";
  /**
   * The primary socket connection; used for sending queries and
   * reading results
   */
  protected SocketConnection primary;
  /**
   * The secondary socket connection; used for reading asynchronous errors
   */
  protected SocketConnection secondary;
  /**
   * The PREDATOR site to which this client is connected
   */
  protected ServerSite site;
  /**
   * The client identification number
   */
  protected int clientId;
  /**
   * The ADT table
   */
  protected ADT ADTs[];
  /**
   * True if a query is in process
   */
  protected boolean queryInProcess = false;
  /**
   * The byte which prefixes all formatted query results
   */
  protected static final byte FORMATTED_RESULTS_SENTINEL = 27;
  /**
   * The binary transaction ID of the current query
   */
  protected byte xactId[];
  /**
   * The size of a SHORE transaction ID (in bytes)
   */
  protected final static int SHOREXactIdSize = 8;
  /**
   * The thread which asynchronously reads off the error stream
   */
  protected ErrorReader errorReader;
  /**
   * The buffer which holds server errors.  This object is shared between
   * the main thread and the error reader thread.
   */
  protected ServerErrorBuffer errorBuffer;

  /**
   * Determines if this connection is open.
   * @return true if the connection is open, false otherwise.
   */
  public boolean isOpen()
  {
    return (primary != null);
  }
  /**
   * Determines if a query is currently in process, i.e.,
   * <TT>processQuery()</TT> has been called, but not <TT>endQuery()</TT>.
   * @return true if a query is in process
   */
  public boolean queryInProcess()
  {
    return (queryInProcess);
  }
  /**
   * Opens a new connection to a 
   * <FONT COLOR="#CC0000"><FONT SIZE=+1>P</FONT>REDATOR</FONT> server.
   * @param Site The site to which the connection should be made
   * @exception CannotConnectException if the connection cannot be opened
   * @exception ConnectionAlreadyOpenException if the connection is already 
   *                                           open
   */
  public synchronized void open(ServerSite Site) 
    throws CannotConnectException, ConnectionAlreadyOpenException
  {
    if (isOpen())
      throw new ConnectionAlreadyOpenException();
    // Open the primary socket connection to the server
    primary = new SocketConnection();
    try { 
      primary.open(Site);
      // Send the string which informs the server to start a new session.
      DataOutputStream dStr = new DataOutputStream(primary.getOutputStream());
      dStr.writeBytes(InitString);
      dStr.flush();
      // Read back the server setup message
      String ServerSetUp;
      if (!(((ServerSetUp = primary.getInputStream().readLine()) != null) &&
	    (ServerSetUp.equals(SERVER_SETUP_STRING))))
	throw new CannotConnectException("Server setup failed.");
      // Send back client characteristics
      dStr.writeBytes("praveen password javaclient ");
      dStr.flush();
      // Read in the client ID
      clientId = primary.getInputStream().readInt();
      // Read in the ADT names and instantiate the ADT classes
      int numADTs = primary.getInputStream().readInt();
      ADTs = new ADT[numADTs];
      int nameSize, typeId;
      byte name[];
      for (int i = 0; i < numADTs; i++) {
	typeId = primary.getInputStream().readInt();
	nameSize = primary.getInputStream().readInt();
	name = new byte[nameSize];
	primary.getInputStream().readFully(name);
	try { 
	  String ADTName = new String(name, 0);
	  String ADTString = "PREDATOR.ADT.ADT_" + ADTName;
	  Class ADTClass = Class.forName(ADTString);
	  ADT ADTObject = (ADT)(ADTClass.newInstance());
	  ADTObject.setName(ADTName);
	  ADTObject.setTypeId(typeId);
	  ADTs[typeId] = ADTObject;
	} catch (ClassNotFoundException X) {
	  ADTs[typeId] = new NullADT();
	} catch (InstantiationException X) {
	  throw new PREDATOR.InternalError(X);
	} catch (IllegalAccessException X) {
	  throw new InternalError(X);
	}
      }
      // Read in the socket number of the secondary connection
      int secondaryPort = primary.getInputStream().readInt();
      // Open the secondary socket connection to the server
      secondary = new SocketConnection();
      secondary.open(new ServerSite(primary.getHostAddress(), secondaryPort));
      // Spawn the thread which asynchronously reads off the error stream
      errorBuffer = new ServerErrorBuffer();
      errorReader = new ErrorReader(secondary.getInputStream(), 
                                    secondary.getOutputStream(), 
				    errorBuffer);
      errorReader.setPriority(Thread.MAX_PRIORITY);
      errorReader.start();
    } catch (IOException X) {
      throw new CannotConnectException(X.getMessage());
    }
    site = Site;
  }
  /**
   * Closes the connection to the 
   * <FONT COLOR="#CC0000"><FONT SIZE=+1>P</FONT>REDATOR</FONT> server.
   * @exception NoOpenConnectionException if no connection is open
   * @exception UnterminatedQueryException if the previous query was not
   *                                       ended using <TT>endQuery()</TT>
   * @exception InterruptedException if this thread is interrupted while
   *                                 closing the connection
   * @see #endQuery()
   */
  public synchronized void close() 
    throws UnterminatedQueryException, NoOpenConnectionException, 
           InterruptedException
  {
    if (!isOpen())
      throw new NoOpenConnectionException();
    //if (queryInProcess())
    //  throw new UnterminatedQueryException();

    /* TODO
      A Netscape bug makes this hang -- lets see if the bug is fixed.
      No, this isnt just a Netscape bug but something deeper.
      It crashes in VJ++ too but in a very strange place  
	  Binquan : It's due to the bug in "Thread.join()", so I changed
	  "errorReader.join()" to "errorReader.join(500)".
	*/
    errorReader.interrupt();
    errorReader.join(5000);

    errorReader.stop();

	try {
      primary.getOutputStream().writeInt(0);
      primary.getOutputStream().flush();
      primary.getInputStream().readLine();
      primary.close();
      secondary.close();
    } catch (IOException X) {  }
    primary = null;
    site = null;
  }
  /**
   * Processes a query, returning its associated result object.  After this
   * function is called, any number of callbacks can be made for data 
   * objects (assuming the result was a <TT>RecordStreamResult</TT> object).
   * If the <TT>endQuery</TT> parameter is set to false, the query will not
   * be ended automatically after the results are read.  In this case,
   * the <TT>endQuery()</TT> function must be called before a new query can
   * be processed.  Note that while records are still being transferred
   * through a <TT>RecordStreamResult</TT>, the query has not ended.  Only
   * after the last record is read will the query be automatically ended.
   * @param Query The query to process
   * @param enwQuery If this parameter is false, the transaction will not
   *                 be automatically ended after the result is obtained;
   *                 otherwise, it will.
   * @return The result of the query; this can be either an UnformattedResult
   *         or a RecordStream result.
   * @exception NoOpenConnectionException if no connection is open
   * @exception UnterminatedQueryException if the previous query was not
   *                                       ended using <TT>endQuery()</TT>
   * @see #endQuery()
   * @see RecordStreamResult 
   * @see UnformattedResult
   */
  public synchronized QueryResult processQuery(Query Q, boolean endQuery) 
    throws NoOpenConnectionException, 
           UnterminatedQueryException
  {
    if (!isOpen())
      throw new NoOpenConnectionException();
    if (queryInProcess())
      throw new UnterminatedQueryException();
    queryInProcess = true;
    // Reset the error buffer
    errorBuffer.reset();
    try {
      // First send the number of bytes across
      primary.getOutputStream().writeInt(Q.toString().length());
      // Then send the actual data
      DataOutputStream dStr = new DataOutputStream(primary.getOutputStream());
      dStr.writeBytes(Q.toString());
      dStr.flush();
      // Now read the results
      ((FilterInputStream)primary.getInputStream()).mark(2);
      byte b;
      b = primary.getInputStream().readByte();
      if (b == FORMATTED_RESULTS_SENTINEL) {
	RecordStreamResult result = new RecordStreamResult(new Schema());
	// Read the transaction ID, size first
	xactId = new byte[primary.getInputStream().readInt()];
	primary.getInputStream().readFully(xactId);
	// Read the schema
	int schemaArity = primary.getInputStream().readInt();
	int attrTypeId, attrNameLen, attrOffset, metaInfoSize;
	ADT attrADT;
	byte attrName[], metaInfoBytes[];
	ADTMetaInfo MetaInfo = null;
	for (int i = 0; i < schemaArity; i++) {
	  // Read the type id
	  attrTypeId = primary.getInputStream().readInt();
	  attrADT = ADTs[attrTypeId];
	  // Read the attr name
	  attrNameLen = primary.getInputStream().readInt();
	  attrName = new byte[attrNameLen];
	  primary.getInputStream().readFully(attrName);
	  // Read the attr offset
	  attrOffset = primary.getInputStream().readInt();
	  // Read the ADT's meta-info (size first)
	  metaInfoSize = primary.getInputStream().readInt();
	  if (metaInfoSize > 0) {
	    metaInfoBytes = new byte[metaInfoSize];
	    primary.getInputStream().readFully(metaInfoBytes);
	    try {
	      MetaInfo = attrADT.newMetaInfo(metaInfoBytes);
	    } catch (Exception X) {
	      throw new InternalError(X);
	    }
	  } else {
	    MetaInfo = null;
	  }
	  Attribute attr = new Attribute(attrADT, MetaInfo, 
					 new String(attrName, 0), attrOffset);
	  result.getSchema().addAttribute(i, attr);
	}
	RecordReader reader = new RecordReader(this, result, endQuery);
	// The record reader will unlock the primary socket when it is finished
	reader.start();
	return result;
      } else {
	((FilterInputStream)primary.getInputStream()).reset();
	UnformattedResult result = readUnformattedResult();
	primary.unlock();
	if (endQuery) {
	  try {
	    endQuery();
	  } catch (Exception X) {
	    throw new InternalError(X);
	  }
	}
	return result;
      }
    } catch (IOException X) {
      throw new InternalError(X);
    }
  }
  /**
   * Ends the last query to be processed.  This function must be called
   * before a new query is processed.  It is highly advisable
   * to call this function as soon as possible, because it terminates the
   * query's transaction at the server, freeing up system resources for other
   * clients.  This function will block until the query can be properly
   * ended.
   * @exception NoOpenConnectionException if no connection is open
   * @exception NoQueryRunningException if no query has been processed or the
   *                                    last query was ended
   * @exception InterruptedException If the current thread is interrupted
   *                                 while blocking waiting to end the query
   */
  public synchronized void endQuery() 
    throws NoOpenConnectionException, 
           NoQueryRunningException,
           InterruptedException
  {
    if (!isOpen())
      throw new NoOpenConnectionException();
    // Wait for all the records to be read
    while (!primary.lock())
      wait();
    if (!queryInProcess())
      throw new NoQueryRunningException();
    // Get the obligatory error communication from the server
    errorBuffer.complete();
    queryInProcess = false;
    xactId = null;
    try {
      primary.getOutputStream().writeBytes("END");
      primary.getOutputStream().flush();
      byte [] reply = new byte[2];
      primary.getInputStream().readFully(reply);
      if (!(new String(reply, 0, 0, 2).equals("OK")))
	throw new InternalError("Server improperly terminated query");
    } catch (IOException X) {
      throw new InternalError(X);
    }
    primary.unlock();
  }
  /**
   * Reads unformatted results off the primary stream.
   */
  UnformattedResult readUnformattedResult() throws IOException
  {
    StringBuffer buf = new StringBuffer();
    char c;
    while ((c = (char)(primary.getInputStream().readByte())) != (char)0)
      buf.append(c);
    if (buf.toString().trim().length() > 0)
      return new UnformattedResult(buf.toString());
    else
      return null;
  }
  /**
   * Returns a URL to a callback object.
   *
   * @param Object The CallbackObject for which a URL is desired
   * @param TypeADT The ADT associated with the object.  This
   *                can be obtained by the <TT>getTypeADT()</TT> method
   *                of the attribute corresponding to this object.
   *                The attribute is available from the schema of the 
   *                record from which the object was obtained.
   * @param MetaInfo The meta-information associated with the object.  This
   *                 can be obtained by the <TT>getMetaInfo()</TT> method
   *                 of the attribute corresponding to this object.
   *                 The attribute is available from the schema of the 
   *                 record from which the object was obtained.
   * @exception NoOpenConnectionException if no connection is open
   * @see RecordStreamResult#getSchema()
   * @see Schema#getAttribute(int)
   * @see Attribute#getMetaInfo()
   * @see Attribute#getTypeADT()
   */
  public URL getCallbackObjectURL(DataObject Object,
				  ADT TypeADT,
				  ADTMetaInfo MetaInfo)
    throws NoOpenConnectionException
  {
    if (!(Object instanceof CallbackObject))
      throw 
	new IllegalArgumentException("The object is not a callback object");
    if (!isOpen())
      throw new NoOpenConnectionException();
    try {
      /*
       * Form the URL to communicate with the server; the URI provided will 
       * be a hexidecimal string representing the byte-message below:
       *   (a.) (b.)    (c.) (d.) (e.)  (f.) (g.)  (h.) (i.)
       *  |----|--...--|----|----|-...-|----|-...-|----|-...-|
       *    a. Client ID (integer)
       *    b. Transaction Id (a binary sequence of known length)
       *    c. Type ID (integer)
       *    d. Size of the object (integer)
       *    e. Object (sequence of (d.) bytes)
       *    f. Size of the object's meta-info (integer)
       *    g. Meta-info (sequence of (f.) bytes)
       *    h. Size of the additional message (integer)
       *    i. Additional message (sequence of (h.) bytes)
       */
      // First form the URI byte string
      byte ObjectBytes[] = Object.getBytes();
      byte MetaInfoBytes[] = (MetaInfo != null) ? MetaInfo.getBytes() : null;
      byte AdditionalMessage[] = ((CallbackObject)Object).getMessage();
      ByteArrayOutputStream bStr = new ByteArrayOutputStream();
      DataOutputStream dStr = new DataOutputStream(bStr);
      dStr.writeInt(clientId);
      if (xactId != null)
	dStr.write(xactId, 0, xactId.length);
      else {
	// Create a null transaction identifier
	for (int i = 0; i < SHOREXactIdSize; i++)
	  dStr.writeByte(0);
      }
      dStr.writeInt(TypeADT.getTypeId());
      if (ObjectBytes != null) {
	dStr.writeInt(ObjectBytes.length);
	dStr.write(ObjectBytes, 0, ObjectBytes.length);
      } else {
	dStr.writeInt(0);
      }
      if (MetaInfoBytes != null) {
	dStr.writeInt(MetaInfoBytes.length);
	dStr.write(MetaInfoBytes, 0, MetaInfoBytes.length);
      } else {
	dStr.writeInt(0);
      }
      if (AdditionalMessage != null) {
	dStr.writeInt(AdditionalMessage.length);
	dStr.write(AdditionalMessage, 0, AdditionalMessage.length);
      } else {
	dStr.writeInt(0);
      }
      dStr.flush();
      byte URIbytes[] = bStr.toByteArray();
      // Now convert it to a hexidecimal String
      StringBuffer URI = new StringBuffer();
      int b = 0;
      for (int i = 0; i < URIbytes.length; i++) {
	b = (URIbytes[i] & 0xf0) >> 4;
	if (b < 10)
	  URI.append((char)((int)'0' + b));
	else
	  URI.append((char)((int)'A' + b - 10));
	b = URIbytes[i] & 0x0f;
	if (b < 10)
	  URI.append((char)((int)'0' + b));
	else
	  URI.append((char)((int)'A' + b - 10));
      }
      // Now create the URL
      return new URL("http", site.getHostAddress().getHostName(),
		     site.getPortNumber(), "/" + URI.toString());
    } catch (Exception X) {
      throw new InternalError(X);
    }
  }
  /**
   * Checks to see if there is a pending server error.
   * @return true if the server has registered an error, false otherwise
   */
  public boolean isError()
  {
    errorBuffer.complete();
    return errorBuffer.isError();
  }  
  /**
   * If an error exists, returns the error and clears the error buffer.  If
   * none exists, returns null.
   * @return The server error
   */
  public ServerError getError()
  {
    errorBuffer.complete();
    ServerError E = errorBuffer.getError();
    errorBuffer.clear();
    return E;
  }
}
