package PREDATOR;

import java.io.*;

/**
 * Reads off the error stream asynchronously, preventing the server from
 * blocking on writes.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 */
class ErrorReader extends Thread {
  /**
   * The error buffer between the main thread and this thread.
   */
  protected ServerErrorBuffer ErrorBuffer;
  /**
   * The error stream which this thread continually monitors
   */
  protected DataInputStream ErrorStream;
  /**
   * The stream on which the thread signals complete receipt
   */
  protected DataOutputStream OutStream;
  /**
   * Default Constructor.
   * @param ErrorStream The stream which this thread reads
   * @param OutStream The stream on which this thread signals receipt
   * @param ErrorBuffer The buffer to hold error information
   */
  ErrorReader(DataInputStream ErrorStream, 
	      DataOutputStream OutStream,
	      ServerErrorBuffer ErrorBuffer) {
    this.ErrorStream = ErrorStream;
    this.OutStream = OutStream;
    this.ErrorBuffer = ErrorBuffer;
  }
  /**
   * The thread's run method.  Reads off the error stream, appending it
   * to the Connection's error buffer.
   */
  public void run() {
    int size;
    byte buf[];
    ServerError error = null;
    while (true) {
      if (isInterrupted())
	return;
      try {
	// Read the error message, length first
	size = ErrorStream.readInt();
	// If the size coming over is a zero, then there was no error.
	if (size == 0) {
	  ErrorBuffer.noError();
	  OutStream.writeBytes("END");
	  OutStream.flush();
	  continue;
	}
	try {
	  buf = new byte[size];
	} catch (OutOfMemoryError E) {
	  E.printStackTrace();
	  throw E;
	}
	ErrorStream.readFully(buf);
	error = new ServerError(new String(buf, 0));
	// Read the error stack
	while (true) {
	  if (isInterrupted())
	    return;
	  // Read the filename, length first
	  size = ErrorStream.readInt();
	  if (size == 0)
	    break;
	  buf = new byte[size];
	  ErrorStream.readFully(buf);
	  String filename = new String(buf, 0);
	  // Read the line number
	  int line = ErrorStream.readInt();
	  // Read in the error string, size first
	  size = ErrorStream.readInt();
	  String errorString = null;
	  if (size > 0) {
	    buf = new byte[size];
	    ErrorStream.readFully(buf);
	    errorString = new String(buf, 0);
	  }
	  // Read in the detail, size first
	  size = ErrorStream.readInt();
	  buf = new byte[size];
	  ErrorStream.readFully(buf);
	  String detail = new String(buf, 0);
	  // Push the new node
	  error.pushError(filename, line, errorString, detail);
	  OutStream.writeBytes("END");
	  OutStream.flush();
	}
	ErrorBuffer.noteError(error);
      } catch (InterruptedIOException X) {
	return;
      } catch (EOFException X) {
	return;
      } catch (IOException X) {
	return;
      } catch (ThreadDeath X) {
	throw new ThreadDeath();
      }
    }
  }
}
