package PREDATOR;

import java.util.*;
import java.io.*;
import java.net.*;
import PREDATOR.ADT.*;

/**
 * This is a simple textual interface for a PREDATOR client connection.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 */
public class JagClient {
  /// true if input lines should be echoed (defaults to false)
  protected boolean printEcho = false;
  /// true if the times should be printed out (defaults to true)
  protected boolean printTimer = true;
  /// true if the schema should be printed out (defaults to true)
  protected boolean printSchemas = true;
  /// true if the results should be printed out (defaults to true)
  protected boolean printResults = true;
  /// true if the serverErrors should be printed out (defaults to true)
  protected boolean printErrors = true;
  /// how detailed should the error printing be (0:low, 1:medium, 2:high)
  protected int errorLevel = 2;
  /// true if verbose messages should be printed out (defaults to true)
  protected boolean printVerbose = true;
  /// true if the prompt should be printed out (defaults to true)
  protected boolean printPrompt = true;
  /// The output stream (defaults to System.out)
  protected PrintStream outStr = System.out;
  /// The input file, if any
  protected String inputFile;
  /// The tokenizer for the input stream
  protected StreamTokenizer tokenizer;
  /// The default prompt
  protected static String DEFAULT_PROMPT = "Q> ";
  /// The command prompt
  protected String prompt = DEFAULT_PROMPT;
  /// All input lines are terminated by a semicolon
  protected final static char LINE_TERMINATOR = ';';
  /// Client commands start with a period
  protected char COMMAND_CHAR = '.';
  /// The command which invokes a callback
  protected final static String CALLBACK_COMMAND = "callback";
  /// The command which runs a system command
  protected final static String SYSTEM_COMMAND = "system";
  /// The command which terminates a running query
  protected final static String END_COMMAND = "end";
  /// The command which downloads a MIME type object
  protected final static String DOWNLOAD_COMMAND = "download";
  /// The command which invokes display of an object
  protected final static String DISPLAY_COMMAND = "display";
  /// The command which changes the default language
  protected final static String LANGUAGE_COMMAND = "language";
  /// The command which changes the prompt
  protected final static String PROMPT_COMMAND = "prompt";
  /// The command which quits the client
  protected final static String ECHO_COMMAND = "echo";
  /// The command which quits the client
  protected final static String HELP_COMMAND = "help";
  /// The command which quits the client
  protected final static String QUIT_COMMAND = "quit";
  /// The command which reads input from another file
  protected final static String READ_COMMAND = "read";
  /// The command which writes ouput to another file
  protected final static String WRITE_COMMAND = "write";
  /// The command which suppresses results
  protected final static String RESULTS_COMMAND = "results";
  /// The command which suppresses schemas
  protected final static String SCHEMAS_COMMAND = "schemas";
  /// The command which suppresses timer printout
  protected final static String TIMER_COMMAND = "timer";

  protected final static String SELECTIVITY_COMMAND = "selectivity";
  protected final static String RESULTSIZE_COMMAND = "resultsize";
  protected final static String DROPRECORDS_COMMAND = "droprecords";

  /// The server site information
  protected ServerSite serverSite;
  /// The PREDATOR connection
  protected JConnection connection;
  /// The download buffer
  byte downloadBuffer[] = new byte[4096];
  /// The default query language
  protected final static String DEFAULT_LANGUAGE = "SQL";
  /// The name of the language of this client, e.g., "SQL"
  protected String languageName = DEFAULT_LANGUAGE;
  /// A container for all the returned objects which a command can reference
  protected Vector complexObjects;

  public static int Selectivity = 10;
  public static int ResultSize = 100;
  public static boolean DropRecords = false;

  /**
   * The main program.  Starts a thread to run the JagClient class and
   * then exits.
   */
  public static void main(String[] args)
  {
    System.out.println("JagClient running ...");
    JagClient Client = new JagClient();
    // Parse the command line
    if (!Client.parseCommandLine(args)) {
      Client.printUsage();
      System.exit(1);
    }
    // Open the connection
    try {
      Client.makeConnection();
    } catch (CannotConnectException X) {
      System.err.println("Cannot make connection to " + Client.serverSite);
      X.printStackTrace();
      System.exit(1);
    } catch (ConnectionAlreadyOpenException X) {
      // This should never happen
      X.printStackTrace();
      System.exit(1);
    }
    // Print the welcome message
    Client.printWelcomeMessage();
    // Run queries
    Client.runQueries();
    // Close the connection
    try {
      Client.closeConnection();
      System.out.println("Connection closed.");
    } catch (NoOpenConnectionException X) {
      System.err.println("Connection was prematurely terminated.");
      System.err.println("No open connection.");
      X.printStackTrace();
      System.exit(1);
    } catch (InterruptedException X) {
      System.err.println("Interrupted while closing connection!");
      X.printStackTrace();
      System.exit(1);
    } catch (UnterminatedQueryException X) {
      System.err.println("A query was improperly terminated.");
      X.printStackTrace();
      System.exit(1);
    }
    System.exit(0);
  }
  /**
   * Protected constructor.
   */
  protected JagClient()
  {
    complexObjects = new Vector();
    initTokenizer(System.in);
  }
  /**
   * Constructor.
   */
  public JagClient(ServerSite Site)
  {
    serverSite = Site;
    complexObjects = new Vector();
    initTokenizer(System.in);
  }
  /**
   * Sets the printEcho flag.
   */
  public void setPrintEcho(boolean NewValue)
  {
    printEcho = NewValue;
  }
  /**
   * Sets the printTimer flag.
   */
  public void setPrintTimer(boolean NewValue)
  {
    printTimer = NewValue;
  }
  /**
   * Sets the printSchemas flag.
   */
  public void setPrintSchemas(boolean NewValue)
  {
    printSchemas = NewValue;
  }
  /**
   * Sets the printResults flag.
   */
  public void setPrintResults(boolean NewValue)
  {
    printResults = NewValue;
  }
  /**
   * Sets the printErrors flag.
   */
  public void setPrintErrors(boolean NewValue)
  {
    printErrors = NewValue;
  }
  /**
   * Sets the errorLebel flag.
   */
  public void setErrorLevel(int NewValue)
  {
    errorLevel = NewValue;
  }
  /**
   * Sets the printVerbose flag.
   */
  public void setPrintVerbose(boolean NewValue)
  {
    printVerbose = NewValue;
  }
  /**
   * Sets the printPrompt flag.
   */
  public void setPrintPrompt(boolean NewValue)
  {
    printPrompt = NewValue;
  }
  /**
   * Sets the current prompt.
   */
  public void setPrompt(String NewPrompt)
  {
    if ((prompt = NewPrompt) == null)
      prompt = DEFAULT_PROMPT;
  }

  /**
   * Initializes the input tokenizer with the input stream.
   */
  protected void initTokenizer(InputStream inStr)
  {
    tokenizer = new StreamTokenizer(inStr);
    tokenizer.resetSyntax();
    tokenizer.slashSlashComments(true);
    tokenizer.slashStarComments(true);
    tokenizer.lowerCaseMode(false);
    tokenizer.eolIsSignificant(false);    
    tokenizer.wordChars(32, ((int)';') - 1);
    tokenizer.wordChars(((int)';') + 1, 255);
    // tokenizer.quoteChar((int)'"');
    // tokenizer.quoteChar((int)'\'');
    tokenizer.whitespaceChars((int)' ', (int)' ');
    tokenizer.whitespaceChars((int)'\t', (int)'\t');
    tokenizer.whitespaceChars((int)'\n', (int)'\n');
  }
  /**
   * Opens the connection.
   */
  public void makeConnection()
    throws CannotConnectException, ConnectionAlreadyOpenException
  {
    connection = new JConnection();
    connection.open(serverSite);
  }
  /**
   * Processes the contents of the input file, directing output to the
   * supplied output file name.
   */
  public void processFile(String inputFilename, String outputFilename)
    throws IOException
  {
    // Redirect input
    try {
      initTokenizer(new FileInputStream(inputFilename));
    } catch (IOException X) {
      initTokenizer(System.in);
      throw X;
    }
    // Redirect output
    try {
      outStr = new PrintStream(new FileOutputStream(outputFilename));
    } catch (IOException X) {
      outStr = System.out;
      throw X;
    }
    // Process the input file
    runQueries();
    // Reset the streams
    outStr.close();
    outStr = System.out;
    initTokenizer(System.in);
  }
  /**
   * Closes the connection.
   */
  public void closeConnection() 
    throws NoOpenConnectionException, InterruptedException, 
    UnterminatedQueryException
  {
    terminateQuery();
    connection.close();
  }

  protected void terminateQuery()
  {
    if (!connection.queryInProcess())
      return;
    complexObjects.removeAllElements();
    try {
      connection.endQuery();
    } catch (Exception X) {
      outStr.println("Cannot end query.");
      X.printStackTrace();
    }
  }

  /**
   * Prints out the welcome message.
   */
  protected void printWelcomeMessage()
  {
    if (printVerbose)
      outStr.
	println("Welcome to PREDATOR.  Type \"" + COMMAND_CHAR + HELP_COMMAND +
		LINE_TERMINATOR + "\" for assistance.");
  }
  /**
   * Parses the command line.
   * @return false if there is an error in the command line parameters
   */
  protected boolean parseCommandLine(String[] args)
  {
    if (args.length < 2)
      return false;
    try {
      serverSite = new ServerSite(args[0], Integer.parseInt(args[1]));
    } catch (java.net.UnknownHostException X) {
      outStr.println("Host \"" + args[0] + "\" is unknown.");
      return false;
    } catch (NumberFormatException X) {
      outStr.println("Port number \"" + args[1] + 
			 "\" is not an integer.");
      return false;
    }
    // Now parse the rest of the flags
    int i = 2;
    while (i < args.length) {
      if (args[i].trim().charAt(0) == '-') {
	switch (args[i].trim().charAt(1)) {
	case ('h'):
	  i++;
	  return false;
	case ('v'):
	  setPrintVerbose(false);
	  i++;
	  break;
	case ('t'):
	  setPrintTimer(false);
	  i++;
	  break;
	case ('e'):
	  setPrintErrors(false);
	  i++;
	  break;
	case ('c'):
	  setPrintEcho(true);
	  i++;
	  break;
	case ('s'):
	  setPrintSchemas(false);
	  i++;
	  break;
	case ('r'):
	  setPrintResults(false);
	  i++;
	  break;
	case ('i'):
	  if (i + 1 < args.length) {
	    inputFile = args[i + 1];
	    try {
	      initTokenizer(new FileInputStream(inputFile));
	    } catch (Exception X) {
	      outStr.println("The file " + inputFile + 
			     " cannot be read: " + X);
	    }
	    i++; i++;
	  } else {
	    outStr.println("Input file flag (-i) requires a parameter.");
	    return false;
	  }
	  break;
	case ('p'):
	  if (i + 1 < args.length) {
	    setPrompt(args[i + 1]);
	    i++; i++;
	  } else {
	    outStr.println("Prompt flag (-p) requires a parameter.");
	    return false;
	  }
	  break;
	case ('l'):
	  if (i + 1 < args.length) {
	    languageName = args[i + 1];
	    i++; i++;
	  } else {
	    outStr.println("Language flag (-l) requires a parameter.");
	    return false;
	  }
	  break;
	default:
	  outStr.println("Flag -" + args[i].trim().charAt(1) + 
			     " unrecognized.");
	  return false;
	}
      } else {
	outStr.println("Options must start with a dash (-).");
	return false;
      }
    }
    return true;
  }
  /**
   * Prints the available options and correct syntax.
   */
  protected void printUsage()
  {
    outStr.println("Usage: java JagClient <host name> " + 
		       "<port number> [options]");
    outStr.println("where [options] include:");
    outStr.println("   -e             Suppress printing of server errors");
    outStr.println("   -c             Echoes each input line");
    outStr.println("   -h             Displays this message");
    outStr.println("   -i <file name> Reads from a file for input");
    outStr.println("   -l <language>  Query language (default \"" + 
		   DEFAULT_LANGUAGE + "\")");
    outStr.println("   -p <prompt>    Prompt (default \"" + 
		   DEFAULT_PROMPT + "\")");
    outStr.println("   -r             Suppress printing of results");
    outStr.println("   -s             Suppress printing of schemas");
    outStr.println("   -t             Suppress printing of timer");
    outStr.println("   -v             Suppress verbose messages");
  }
  /**
   * Runs queries until the user signals EOF.
   */
  protected void runQueries()
  {
    Query Q;
    QueryResult Result;
    String Input;
    while ((Input = getInput()) != null) {
      if (printEcho)
	outStr.println(prompt + Input + LINE_TERMINATOR);
      java.util.Date start = new java.util.Date();
      if (isCommand(Input)) {
	if (processCommand(Input.substring(1)))
	  break;
      } else {
	processQuery(Input);
      }
      if (connection.isError())
	processServerError(connection.getError());
      java.util.Date end = new java.util.Date();
      // Print out the time if requested
      if (printTimer)
	if (printVerbose) 
	  outStr.println("Processing time: " +
			 (double)(end.getTime()-start.getTime()) / 
			 1000.0 + " seconds");
	else
	  outStr.println(end.getTime()-start.getTime());
    }
  }
  /**
   * Determines if an input is a client command.
   */
  protected boolean isCommand(String Input)
  {
    return ((Input.length() > 0) && 
	    (Input.charAt(0) == COMMAND_CHAR));
  }
  /**
   * Runs a single query.
   */
  protected void processQuery(String Query)
  {
    terminateQuery();
    try {
      Query query = new Query(languageName, Query);
      QueryResult Result = null;
      try {
	Result = connection.processQuery(query, false);
      } catch (UnterminatedQueryException X) {
	outStr.println("The query was improperly terminated.");
	X.printStackTrace();
      }
      if (Result == null)
	processNoResponse();
      else if (Result instanceof UnformattedResult)
	processUnformatted((UnformattedResult)Result);
      else if (Result instanceof RecordStreamResult)
	processRecordStream((RecordStreamResult)Result);
      else
	outStr.println("Unrecognized result type.");
    } catch (BadQueryException X) {
      outStr.println("The query is improperly formatted");
      X.printStackTrace();
    } catch (NoOpenConnectionException X) {
      outStr.println("Cannot process query.");
      X.printStackTrace();
    }
  }
  /**
   * Processes a client command.
   * @return true if the command should initiate a shutdown.
   */
  protected boolean processCommand(String Command)
  {
    try {
      /*
       * JDK 1.1
       * Use a string reader instead of a StringBufferInputStream
       */
      StreamTokenizer tokenizer = 
	new StreamTokenizer(new StringBufferInputStream(Command));
      tokenizer.parseNumbers();
      tokenizer.wordChars(0, 255);
      tokenizer.quoteChar((int)'"');
      tokenizer.quoteChar((int)'\'');
      tokenizer.whitespaceChars((int)' ', (int)' ');
      tokenizer.whitespaceChars((int)'\t', (int)'\t');
      tokenizer.whitespaceChars((int)'\n', (int)'\n');
      tokenizer.nextToken();
      if (tokenizer.ttype != StreamTokenizer.TT_WORD) {
	displayHelp();
	return false;
      }
      String commandWord = tokenizer.sval;
      if (commandWord.toLowerCase().equals(CALLBACK_COMMAND)) {
	tokenizer.nextToken();
	if (tokenizer.ttype != StreamTokenizer.TT_NUMBER) {
	  outStr.println("Usage: " + COMMAND_CHAR + CALLBACK_COMMAND + 
			 " <object number>" + LINE_TERMINATOR);
	  return false;
	}
	performCallback((int)tokenizer.nval);
      } else if (commandWord.toLowerCase().equals(DOWNLOAD_COMMAND)) {
	tokenizer.nextToken();
	if (tokenizer.ttype != StreamTokenizer.TT_NUMBER) {
	  outStr.println("Usage: " + COMMAND_CHAR + DOWNLOAD_COMMAND + 
			 " <object number> [<file name>]" + LINE_TERMINATOR);
	  return false;
	}
	int objectId = (int)tokenizer.nval;
	tokenizer.nextToken();
	if (tokenizer.ttype == StreamTokenizer.TT_WORD) {
	  performDownload((int)tokenizer.nval, tokenizer.sval);
	} else {
	  performDownload((int)tokenizer.nval, null);
	}
      } else if (commandWord.toLowerCase().equals(END_COMMAND)) {
	terminateQuery();
      } else if (commandWord.toLowerCase().equals(SYSTEM_COMMAND)) {
	tokenizer.nextToken();
        if ((tokenizer.ttype != StreamTokenizer.TT_WORD) &&
            (tokenizer.ttype != (int)'\'') && (tokenizer.ttype != (int)'\"')) {
	  outStr.println("Usage: " + COMMAND_CHAR + SYSTEM_COMMAND + 
			 " '<system command>'" + LINE_TERMINATOR);
	  return false;
	}
	Runtime.getRuntime().exec(tokenizer.sval);
      } else if (commandWord.toLowerCase().equals(DISPLAY_COMMAND)) {
	tokenizer.nextToken();
	if (tokenizer.ttype != StreamTokenizer.TT_NUMBER) {
	  outStr.println("Usage: " + COMMAND_CHAR + DISPLAY_COMMAND + 
			 " <object number>" + LINE_TERMINATOR);
	  return false;
	}
	performDisplay((int)tokenizer.nval);
      } else if (commandWord.toLowerCase().equals(PROMPT_COMMAND)) {
        tokenizer.nextToken();
        if ((tokenizer.ttype != StreamTokenizer.TT_WORD) &&
            (tokenizer.ttype != (int)'\'') && (tokenizer.ttype != (int)'\"')) {
	  outStr.println("Usage: " + COMMAND_CHAR + PROMPT_COMMAND + 
			 " '<new prompt>'" + LINE_TERMINATOR);
	  return false;
	}
	prompt = tokenizer.sval;
      } else if (commandWord.toLowerCase().equals(RESULTS_COMMAND)) {
	tokenizer.nextToken();
	if ((tokenizer.ttype != StreamTokenizer.TT_WORD) ||
	    ((!tokenizer.sval.toLowerCase().equals("on")) &&
	     (!tokenizer.sval.toLowerCase().equals("off")))) {
	  outStr.println("Usage: " + COMMAND_CHAR + RESULTS_COMMAND + 
			 " [ on | off ]" + LINE_TERMINATOR);
	  return false;
	}
	printResults = tokenizer.sval.toLowerCase().equals("on");
      } else if (commandWord.toLowerCase().equals(SCHEMAS_COMMAND)) {
	tokenizer.nextToken();
	if ((tokenizer.ttype != StreamTokenizer.TT_WORD) ||
	    ((!tokenizer.sval.toLowerCase().equals("on")) &&
	     (!tokenizer.sval.toLowerCase().equals("off")))) {
	  outStr.println("Usage: " + COMMAND_CHAR + SCHEMAS_COMMAND + 
			 " [ on | off ]" + LINE_TERMINATOR);
	  return false;
	}
	printSchemas = tokenizer.sval.toLowerCase().equals("on");
      } else if (commandWord.toLowerCase().equals(TIMER_COMMAND)) {
	tokenizer.nextToken();
	if ((tokenizer.ttype != StreamTokenizer.TT_WORD) ||
	    ((!tokenizer.sval.toLowerCase().equals("on")) &&
	     (!tokenizer.sval.toLowerCase().equals("off")))) {
	  outStr.println("Usage: " + COMMAND_CHAR + TIMER_COMMAND + 
			 " [ on | off ]" + LINE_TERMINATOR);
	  return false;
	}
	printTimer = tokenizer.sval.toLowerCase().equals("on");
      } else if (commandWord.toLowerCase().equals(LANGUAGE_COMMAND)) {
	tokenizer.nextToken();
	if (tokenizer.ttype != StreamTokenizer.TT_WORD) {
	  outStr.println("Usage: " + COMMAND_CHAR + LANGUAGE_COMMAND + 
			 " <default language>" + LINE_TERMINATOR);
	  return false;
	}
	languageName = tokenizer.sval;
      } else if (commandWord.toLowerCase().equals(READ_COMMAND)) {
	tokenizer.nextToken();
	if (tokenizer.ttype != StreamTokenizer.TT_WORD) {
	  outStr.println("Usage: " + COMMAND_CHAR + READ_COMMAND + 
			 " <file name>" + LINE_TERMINATOR);
	  return false;
	}
	readFile(tokenizer.sval);
      } else if (commandWord.toLowerCase().equals(WRITE_COMMAND)) {
	tokenizer.nextToken();
	if (tokenizer.ttype != StreamTokenizer.TT_WORD) {
	  outStr.println("Usage: " + COMMAND_CHAR + WRITE_COMMAND + 
			 " <file name>" + LINE_TERMINATOR);
	  return false;
	}
	writeFile(tokenizer.sval);
      } else if (commandWord.toLowerCase().equals(ECHO_COMMAND)) {
	tokenizer.nextToken();
	if (tokenizer.ttype != StreamTokenizer.TT_WORD) {
	  outStr.println("Usage: " + COMMAND_CHAR + ECHO_COMMAND + 
			 " <word to echo>" + LINE_TERMINATOR);
	  return false;
	}
	outStr.println(tokenizer.sval);
      } else if (commandWord.toLowerCase().equals(HELP_COMMAND)) {
	displayHelp();
      } else if (commandWord.toLowerCase().equals(QUIT_COMMAND)) {
        return true;
      } else if (commandWord.toLowerCase().
		 equals(SELECTIVITY_COMMAND)) {
	tokenizer.nextToken();
	if (tokenizer.ttype != StreamTokenizer.TT_NUMBER) {
	  outStr.println("Usage: " + COMMAND_CHAR + SELECTIVITY_COMMAND + 
			 " <selectivity>" + LINE_TERMINATOR);
	  return false;
	}
	Selectivity = (int)tokenizer.nval;
      } else if (commandWord.toLowerCase().equals(RESULTSIZE_COMMAND)) {
	tokenizer.nextToken();
	if (tokenizer.ttype != StreamTokenizer.TT_NUMBER) {
	  outStr.println("Usage: " + COMMAND_CHAR + RESULTSIZE_COMMAND + 
			 " <result size>" + LINE_TERMINATOR);
	  return false;
	}
	ResultSize = (int)tokenizer.nval;
     } else if (commandWord.toLowerCase().equals(DROPRECORDS_COMMAND)) {
	tokenizer.nextToken();
	if (tokenizer.ttype != StreamTokenizer.TT_NUMBER) {
	  outStr.println("Usage: " + COMMAND_CHAR + RESULTSIZE_COMMAND + 
			 " <result size>" + LINE_TERMINATOR);
	  return false;
	}
	DropRecords = ((int)tokenizer.nval == 1);
      } else {
	outStr.println("Unrecognized command: " + commandWord);
	return false;
      }
    } catch (IOException X) {
      outStr.println("An error occurred while parsing the command.");
      X.printStackTrace();
      return false;
    }
    return false;
  }
  /**
   * Displays help.
   */
  protected void displayHelp()
  {
    outStr.println("All lines of input must be terminated with a '" +
		   LINE_TERMINATOR + "'.");
    outStr.println("All client commands must begin with a '" + 
		   COMMAND_CHAR + "'.");
    outStr.println("Query syntax: <query>" + LINE_TERMINATOR);
    outStr.println("Command syntax: " + COMMAND_CHAR + 
		       "<command> <parameters>" + LINE_TERMINATOR);
    outStr.println("  Available commands:");
    outStr.println("    \"" + COMMAND_CHAR + RESULTS_COMMAND +
		   " [ on | off ]" + LINE_TERMINATOR + 
		   "\" prints/suppresses printing of results");
    outStr.println("    \"" + COMMAND_CHAR + SCHEMAS_COMMAND +
		   " [ on | off ]" + LINE_TERMINATOR + 
		   "\" prints/suppresses printing of schemas");
    outStr.println("    \"" + COMMAND_CHAR + TIMER_COMMAND +
		   " [ on | off ]" + LINE_TERMINATOR + 
		   "\" prints/suppresses printing of timer measurements");
    outStr.println("    \"" + COMMAND_CHAR + PROMPT_COMMAND +
		   " '<new prompt>'" + LINE_TERMINATOR + 
		   "\" changes the query prompt");
    outStr.println("    \"" + COMMAND_CHAR + LANGUAGE_COMMAND +
		   " <language name>" + LINE_TERMINATOR + 
		   "\" changes the default language");
    outStr.println("    \"" + COMMAND_CHAR + READ_COMMAND +
		   " <file name>" + LINE_TERMINATOR + 
		   "\" reads input from the supplied file");
    outStr.println("    \"" + COMMAND_CHAR + WRITE_COMMAND +
		   " <file name>" + LINE_TERMINATOR + 
		   "\" writes output to the supplied file");
    outStr.println("    \"" + COMMAND_CHAR + DISPLAY_COMMAND + 
		   " <object #>" + LINE_TERMINATOR + 
		   "\"displays an object");
    outStr.println("    \"" + COMMAND_CHAR + CALLBACK_COMMAND + 
		   " <object #>" + LINE_TERMINATOR + 
		   "\" performs a callback");
    outStr.println("    \"" + COMMAND_CHAR + DOWNLOAD_COMMAND + 
		   " <object #> [<file name>]" + LINE_TERMINATOR + 
		   "\" downloads a MIME type object [to a local file]");
    outStr.println("    \"" + COMMAND_CHAR + SYSTEM_COMMAND + 
		   " '<command>'" + LINE_TERMINATOR + 
		   "\" runs the command in a separate process");
    outStr.println("    \"" + COMMAND_CHAR + HELP_COMMAND + 
		   LINE_TERMINATOR + "\" displays this message");
    outStr.println("    \"" + COMMAND_CHAR + QUIT_COMMAND + 
		   LINE_TERMINATOR + 
		   "\" ends the session (as does Ctrl-D)");
  }
  /**
   * Switches input to a file.
   */
  protected void readFile(String filename)
  {
    try {
      initTokenizer(new FileInputStream(filename));
    } catch (Exception X) {
      outStr.println("The file " + filename + " cannot be read: " + X);
      initTokenizer(System.in);
      return;
    }
    String oldPrompt = prompt;
    prompt = "";
    runQueries();
    prompt = oldPrompt;
    System.out.println("Finished reading " + filename + ".");
    initTokenizer(System.in);
  }
  /**
   * Switches output to a file.
   */
  protected void writeFile(String filename)
  {
    try {
      outStr = new PrintStream(new FileOutputStream(filename));
    } catch (Exception X) {
      outStr.println("Could not open " + filename + " for writing: " + X);
      outStr = System.out;
      return;
    }
    System.out.println("Output redirected to " + filename + ";");
    System.out.println("type \".quit;\" to restore output to screen.");
    runQueries();
    System.out.println("Finished writing to " + filename + ".");
    outStr.close();
    outStr = System.out;
  }
  /**
   * Performs a callback on the specified object.
   */
  protected void performCallback(int objectId)
  {
    if (objectId < complexObjects.size()) {
      JagMetaInfoObject MIObject = 
	(JagMetaInfoObject)(complexObjects.elementAt(objectId));
      ADT typeADT = MIObject.TypeADT;
      ADTMetaInfo metaInfo = MIObject.MetaInfo;
      DataObject object = MIObject.Object;
      if (!(object instanceof CachingObject)) {
	outStr.println("Object {" + objectId + 
			   "} does not require a callback.");
	return;
      }
      outStr.println("Performing callback...");
      try {
	URL callbackURL = 
	  connection.getCallbackObjectURL(object, typeADT, metaInfo);
	((CachingObject)object).callback(callbackURL, metaInfo);
      } catch (Exception X) {
	outStr.println("Callback failed.");
	X.printStackTrace();
      }
    } else {
      outStr.println("There is no object with id {" + objectId + "}");
    }
  }
  /**
   * Performs a download of the specified object.
   */
  protected void performDownload(int objectId, String filename)
  {
    if (objectId < complexObjects.size()) {
      JagMetaInfoObject MIObject = 
	(JagMetaInfoObject)(complexObjects.elementAt(objectId));
      ADT typeADT = MIObject.TypeADT;
      ADTMetaInfo metaInfo = MIObject.MetaInfo;
      DataObject object = MIObject.Object;
      if (!(object instanceof MIMETypeObject)) {
	outStr.println("Object {" + objectId + 
			   "} is not a MIME type object.");
	return;
      }
      if (!(object instanceof CallbackObject)) {
	outStr.println("Object {" + objectId + 
			   "} doesn't contact the server for its full value.");
	return;
      }
      FileOutputStream fStr = null;
      try { 
	if (filename != null) {
	  // Open a stream to the file
	  String downloadFile = filename + "." + 
	    ((MIMETypeObject)object).getFilenameExtension(metaInfo);
	  outStr.println("Downloading object " + objectId + 
			 " to " + downloadFile + "..." );
	  fStr = new FileOutputStream(downloadFile);
	} else {
	  outStr.println("Downloading object " + objectId + 
			 " to main memory..." );
	}
	// Get the URL of the object
	URL callbackURL = 
	  connection.getCallbackObjectURL(object, typeADT, metaInfo);
	// Open a stream to the URL
	URLConnection URLConn = callbackURL.openConnection();
	int Size;
	int TotalBytes = 0;
	while (true) {
	  if ((Size = URLConn.getInputStream().read(downloadBuffer)) == -1)
	    break;
	  TotalBytes += Size;
	  if (fStr != null)
	    fStr.write(downloadBuffer, 0, Size);
	}
	if (fStr != null)
	  fStr.close();
	URLConn.getInputStream().close();
	outStr.println("Download complete: " + TotalBytes + " bytes read.");
      } catch (Exception X) {
	outStr.println("The download failed:");
	X.printStackTrace(outStr);
      }
    } else {
      outStr.println("There is no object with id {" + objectId + "}");
    }
  }
  /**
   * Performs a display of the specified object.
   */
  protected void performDisplay(int objectId)
  {
    if (objectId < complexObjects.size()) {
      JagMetaInfoObject MIObject = 
	(JagMetaInfoObject)(complexObjects.elementAt(objectId));
      ADTMetaInfo metaInfo = MIObject.MetaInfo;
      DataObject object = MIObject.Object;
      if (!(object instanceof DisplayObject)) {
	outStr.println("Object {" + objectId + "} cannot display itself.");
	return;
      }
      if ((object instanceof CachingObject) &&
	  (((CachingObject)object).requiresCallback(metaInfo))) {
	performCallback(objectId);
      }
      try {
	outStr.println("Displaying object...");
	((DisplayObject)object).display(metaInfo);
      } catch (RequiresCallbackException X) {
	outStr.println("The object cannot be displayed " + 
		       "before a callback is made.");
	return;
      }
    } else {
      outStr.println("There is no object with id {" + objectId + "}");
    }
  }
  /**
   * Reads an input from the user.  Inputs are always terminated with a 
   * semicolon.  Returns null if the user signals EOF.
   */
  protected String getInput()
  {
    StringBuffer Line = new StringBuffer();
    if (printPrompt) {
      outStr.print(prompt);
      outStr.flush();
    }
    try {
      tokenizer.nextToken();
      while ((tokenizer.ttype != StreamTokenizer.TT_EOF) &&
	     (tokenizer.ttype != LINE_TERMINATOR)) {
	if (tokenizer.ttype != StreamTokenizer.TT_WORD) {
	  Line.append((char)(tokenizer.ttype));
	  tokenizer.nextToken();
          continue;
	}
	Line.append(tokenizer.sval);
	Line.append(' ');
	tokenizer.nextToken();
      }
    } catch (IOException X) {
      outStr.println("Received IO Exception.");
      X.printStackTrace();
      return null;
    }
    if (tokenizer.ttype == StreamTokenizer.TT_EOF) {
      return null;
    } else {
      return Line.toString().trim();
    }
  }
  /**
   * Processes a "no response" result from a query.
   */
  protected void processNoResponse()
  {
  }
  /**
   * Processes a server error.
   */
  protected void processServerError(ServerError Error)
  {
    if (printErrors) {
      //outStr.println("SERVER ERROR: " + Error.getMessage() + "\nStack Trace:");
      outStr.println("Error: " + Error.getMessage());
      Error.printStackTrace(outStr, 1 /* ErrorLevel */);
      outStr.flush();
    }
  }
  /**
   * Processes an unformatted result from a query.
   */
  protected void processUnformatted(UnformattedResult Unformatted)
  {
    if (printResults)
      outStr.println(Unformatted.toString());
  }
  /**
   * Processes a RecordStream result from a query.
   */
  protected synchronized void processRecordStream(RecordStreamResult 
						  RecordStream)
  {
    DataObject record[];
    if ((printSchemas) && (printResults))
      printSchema(RecordStream.getSchema());
    try {
      while (!RecordStream.isEmpty())
	try {
	  // Get the next record
	  record = RecordStream.getNextRecord(null);
	  // Print out the record
	  if (printResults) {
	    for (int i = 0; i < RecordStream.getSchema().getArity(); i++) {
	      if ((record[i] instanceof CallbackObject) ||
		  (record[i] instanceof DisplayObject)) {
		complexObjects.
		  addElement(new JagMetaInfoObject(record[i], 
						RecordStream.
						getSchema().getAttribute(i).
						getTypeADT(),
						RecordStream.
						getSchema().getAttribute(i).
						getMetaInfo()));
		outStr.print("{" + (complexObjects.size() - 1) + "}");
	      }
	      outStr.print(record[i].toString(RecordStream.
					      getSchema().getAttribute(i).
					      getMetaInfo()));
	      if (i < RecordStream.getSchema().getArity() - 1)
		outStr.print('\t');
	      else 
		outStr.print('\n');
	    }
	  }
        } catch (BadSubclassSelectorException X) {
	  // this should never happen
	  X.printStackTrace();
	  continue;
	}
    } catch (InterruptedException X) {
      // this should never happen
      return;
    }
  }
  /**
   * Prints a schema.
   */
  protected void printSchema(Schema schema) 
  {
    outStr.print("<");
    for (int i = 0; i < schema.getArity(); i++) {
      outStr.print(schema.getAttribute(i).toString());
      if (i < schema.getArity() - 1)
	outStr.print(", ");
    }
    outStr.println(">");
  }
}
