package PREDATOR;

import java.io.*;
import java.net.*;
import java.util.*;
import reflector.*;

/**
 * This class implements a reflector object.  A reflector routes socket
 * connections from one machine to another.  These are used to permit
 * applets to connect to sockets on machines different from the machine
 * from which they were downloaded.  This particular reflector is used
 * to reflect JCL connections to PREDATOR, which are more complicated, 
 * since they use two socket connections: one for standard communications
 * and one for error communications.  As a result, the reflector must snoop
 * on the protocol setup to read the error port and set up a new reflector
 * for that port.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 */
public class JCLReflector extends Reflector {
  /// true if this reflector is for a single error socket connection
  protected boolean ErrorConn;

  /**
   * Initializes a reflector to accept a connection on the supplied
   * service port and route it to the supplied service site.
   */
  public JCLReflector(int ServicePort, ServerSite Site, boolean ErrorConn)
  {
    super(ServicePort, Site);
    this.ErrorConn = ErrorConn;
  }

  /**
   * The thread's run function.
   */
  public void run()
  {
    ServerSocket InSock;
    Socket ConnectSock;
    int connectionNum = 0;

    try {
      InSock = new ServerSocket(ServicePort);
      System.out.println("Accepting incoming connections on port " + 
			 ServicePort + ".");
      while (true) {
	ConnectSock = InSock.accept();
	System.out.println("Incoming connection " + connectionNum + 
			   " accepted.");
	new RouterThread(ConnectSock, Site, connectionNum, ErrorConn).start();
	if (ErrorConn)
	  break;
	connectionNum++;
      }
    } catch (IOException X) {
      X.printStackTrace();
      return;
    }
  }

  /**
   * Starts a reflector.  The required arguments are:
   * <OL>
   *   <LI>The local port on which the reflector will accept connections
   *   <LI>The host machine to which the socket connections will be directed
   *   <LI>The port number of that host which will accept socket connections
   * </OL>
   */
  public static void main(String args[])
  {
    if (args.length != 3) {
      System.err.println("Required arguments: <local port> " + 
			 "<destination host> <destination port>");
      System.exit(1);
    }
    try {
      int localPort;
      ServerSite Site;
      localPort = Integer.parseInt(args[0]);
      Site = new ServerSite(args[1], Integer.parseInt(args[2]));
      new JCLReflector(localPort, Site, false).start();
    } catch (NumberFormatException X) {
      System.err.println("Required arguments: <local port> " + 
			 "<destination host> <destination port>");
      System.exit(1);
    } catch (UnknownHostException X) {
      System.err.println("Host " + args[1] + " is unknown.");
      System.exit(1);
    }
  }
}

class RouterThread extends Thread {
  /// The socket connection to the client
  Socket ClientSock;
  /// The server site to connect to
  ServerSite Site;

  boolean ErrorConn;
  int id;

  public RouterThread (Socket ClientSock, ServerSite Site, 
		       int id, boolean ErrorConn)
  {
    this.ClientSock = ClientSock;
    this.Site = Site;
    this.id = id;
    this.ErrorConn = ErrorConn;
  }

  public void run()
  {
    Socket ServerSock = null;
    StreamBinderThread SBT1, SBT2;
    try {
      ServerSock = new Socket(Site.getHostAddress(), Site.getPortNumber());
      System.out.println("Connection opened to " + Site + 
			 " for incoming connection " + id + ".");
      SBT1 = new StreamBinderThread(ServerSock.getInputStream(),
				    ClientSock.getOutputStream(),
				    "Server ==[" + id + "]==> Client",
				    !ErrorConn, Site);
      SBT2 = new StreamBinderThread(ClientSock.getInputStream(),
				    ServerSock.getOutputStream(),
				    "Client ==[" + id + "]==> Server",
				    false, Site);
      SBT1.setPriority(Thread.MIN_PRIORITY);
      System.out.println("Starting stream binder: " + SBT1.getName());
      SBT1.start();
      SBT2.setPriority(Thread.MIN_PRIORITY);
      System.out.println("Starting stream binder: " + SBT2.getName());
      SBT2.start();
      System.out.println("Connections have been bound for incoming connection "
			 + id + ".");
      SBT1.join();
      System.out.println("Stream binder " + SBT1.getName() + " finished.");
      SBT2.join();
      System.out.println("Stream binder " + SBT2.getName() + " finished.");
      ClientSock.close();
      ServerSock.close();
      System.out.println("Connections closed for incoming connection "
			 + id + ".");
    } catch (Exception X) {
      try {
	if (ServerSock != null)
	  ServerSock.close();
	ClientSock.close();
      } catch (IOException Y) { }
      X.printStackTrace();
      return;
    }
  }
}

class StreamBinderThread extends Thread {
  InputStream IStr;
  OutputStream OStr;
  boolean ServerToClient;
  ServerSite Site;

  public StreamBinderThread (InputStream IStr, OutputStream OStr, 
			     String name, boolean ServerToClient,
			     ServerSite Site)
  {
    super(name);
    this.IStr = IStr;
    this.OStr = OStr;
    this.ServerToClient = ServerToClient;
    this.Site = Site;
  }

  public void run()
  {
    byte buf[] = new byte[2048];
    int size = 0;
    String str = null;

    try {
      if (ServerToClient) {
	/* 
	 * Snoop on the protocol to spawn another reflector for the JCL
	 * error thread
	 */
	// Read the setup string
	size = IStr.read(buf);
	OStr.write(buf, 0, size);
	OStr.flush();
	str = new String(buf, 0, 0, size);
      }
      /* Check to see if this connection is an initial setup; if so, snoop
       * the communications. */
      if ((str != null) && str.startsWith(Connection.SERVER_SETUP_STRING)) {
	// Read the client ID
	size = IStr.read(buf, 0, 4);
	OStr.write(buf, 0, size);
	// Read the number of ADTs
	size = IStr.read(buf, 0, 4);
	OStr.write(buf, 0, size);
	int numADTs = 
	  new DataInputStream(new ByteArrayInputStream(buf)).readInt();
	for (int i = 0; i < numADTs; i++) {
	  // Read the type ID
	  size = IStr.read(buf, 0, 4);
	  OStr.write(buf, 0, size);
	  // Read the name length
	  size = IStr.read(buf, 0, 4);
	  OStr.write(buf, 0, size);
	  int nameLength = 
	    new DataInputStream(new ByteArrayInputStream(buf)).readInt();
	  // Read the name
	  size = IStr.read(buf, 0, nameLength);
	  OStr.write(buf, 0, size);
	}
	// Read the error stream's port number
	size = IStr.read(buf, 0, 4);
	int errorPort = 
	  new DataInputStream(new ByteArrayInputStream(buf)).readInt();
	ServerSite NewSite = new ServerSite(Site.getHostAddress(), errorPort);
	new JCLReflector(errorPort, NewSite, true).start();
	OStr.write(buf, 0, size);
	OStr.flush();
      }
      while (size != -1) {
	size = IStr.read(buf);
	if (size == -1)
	  break;
	OStr.write(buf, 0, size);
	OStr.flush();
	System.out.println(getName() + " transferred " + size + " bytes.");
      }
      IStr.close();
      OStr.close();
    } catch (IOException X) {
      try {
	IStr.close();
	OStr.close();
      } catch (IOException Y) { }
      return;
    }
  }
}
