package PREDATOR;

/**
 * This class embodies a "query", which can also be a non-query
 * statement, such as an update statement or commands like <TT>TRANS begin</TT>
 * or <TT>list rels</TT>.  All 
 * <FONT COLOR="#CC0000"><FONT SIZE=+1>P</FONT>REDATOR</FONT> queries are 
 * phrased in a <I>language</I>, which identifies the database engine 
 * responsible for processing the query.  The relational engine has the 
 * language "SQL", and the default engine has the language "XXX".  
 * Query objects can only contain one logical query.  For example:
 * <PRE>
 * Query Q = new Query("SELECT I.picture FROM Images I");
 * </PRE>
 * is acceptable, whereas
 * <PRE>
 * Query Q = new Query("TRANS begin; SELECT I.picture FROM Images I; TRANS commit;");
 * </PRE>
 * is not.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see Connection#processQuery(Query)
 */
public class Query {
  /// The name of the language in which the query is stated
  String languageName;
  /// The query text
  String queryText;
  /// The character which terminates a query (semicolon)
  final char QUERY_TERMINATOR = ';';
  /// The character which separates the language from the query (colon)
  final char QUERY_SEPARATOR = ':';
  /// The SQL language name
  final String SQL_LANGUAGE = "SQL";

  /**
   * Constructor.
   * @param LanguageName The name of the language the query is phrased in.
   *                     SQL queries should use the language "SQL".
   * @param QueryText The actual query.
   * @exception BadQueryException If the query is improperly formatted
   */
  public Query (String LanguageName, String QueryText) throws BadQueryException
  {
    languageName = LanguageName;
    queryText = QueryText;
    if (queryText.indexOf((int)QUERY_TERMINATOR) != -1) {
      throw new BadQueryException("the query contains a '" + 
				  QUERY_TERMINATOR + 
				  "', which is disallowed.");
    }
  }
  /**
   * This constructor is short-hand for queries phrased in SQL.
   * @param QueryText The actual query.
   * @exception BadQueryException If the query is improperly formatted
   */
  public Query (String QueryText) throws BadQueryException
  {
    languageName = SQL_LANGUAGE;
    queryText = QueryText;
    if (queryText.indexOf((int)QUERY_TERMINATOR) != -1) {
      throw new BadQueryException("the query contains a '" + 
				  QUERY_TERMINATOR + 
				  "', which is disallowed.");
    }
  }
  /**
   * Produces a text-version of this query suitable for parsing by the
   * server.
   */
  public String toString()
  {
    return (languageName + QUERY_SEPARATOR + queryText + QUERY_TERMINATOR);
  }
}
