package PREDATOR;

import java.util.*;
import PREDATOR.ADT.*;

/**
 * The result returned by a call to <TT>Connection.processQuery()</TT> when 
 * the results are records.  This class provides an iteration interface over 
 * the resultant records.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see Connection#processQuery(Query)
 */
public class RecordStreamResult implements QueryResult {
  /// The schema of the answers
  Schema schema;
  /// The buffer of records
  DataRecord[] buffer;
  /// The size of the buffer
  final static int BUFFER_SIZE = 10;
  /// The index of the next slot to fill
  int nextSlotToFill;
  /// The index of the next slot to return
  int nextSlotToReturn;
  /**
   * A flag which indicates that no more records will be placed in this
   * result variable.  It is set by the RecordReader thread when the read
   * is complete.
   */
  boolean noMoreInput;

  /**
   * Constructor.
   */
  RecordStreamResult(Schema schema)
  {
    this.schema = schema;
    noMoreInput = false;
    buffer = new DataRecord[BUFFER_SIZE];
    for (int i = 0; i < BUFFER_SIZE; i++)
      buffer[i] = null;
    nextSlotToFill = nextSlotToReturn = 0;
  }
  /**
   * Puts a record into this result object.
   * @exception InterruptedException if the thread is interrupted while
   *                                 blocking, waiting to add the record
   */
  synchronized void putRecord(DataRecord record) throws InterruptedException
  {
    while (buffer[nextSlotToFill] != null)
      wait();
    buffer[nextSlotToFill] = record;
    nextSlotToFill = (nextSlotToFill + 1) % BUFFER_SIZE;
    notifyAll();
  }
  /**
   * Returns true if the results have been read completely.  This function 
   * will block until it is known if the stream is empty or not.
   * @return true if all records have been read.
   * @exception InterruptedException If the current thread is interrupted
   *                                 while blocking waiting for the status
   */
  public synchronized boolean isEmpty() throws InterruptedException
  {
    while (!isReady())
      wait();
    return (noMoreInput && (buffer[nextSlotToReturn] == null));
  }
  /**
   * Called by the RecordReader thread when all records have been placed
   * in this object.
   */
  synchronized void setNoMoreInput()
  {
    noMoreInput = true;
    notifyAll();
  }
  /**
   * Returns true if the stream is ready to correctly report if it is empty,
   * and if not, return the next record.
   * @return true if this result object is ready
   */
  synchronized boolean isReady()
  {
    return (noMoreInput || (buffer[nextSlotToReturn] != null));
  }
  /**
   * Gets the schema for this stream of records.
   * @return The schema associated with these records.
   */
  public synchronized Schema getSchema()
  {
    return (Schema)schema;
  }
  /**
   * Gets an array of the data objects of the attributes of the next 
   * record from this result object.  This function will block until
   * the next record is ready to be returned.  If the stream is empty, 
   * it will return a null reference.
   * @param Subclasser An object which determines a subclass to use of 
   *                   DataObjects, or null if no subclass should be
   *                   selected.
   * @return An array of the data objects of the attributes of the next 
   *         record from this result object, or null if the stream is empty.
   * @exception BadSubclassSelectorException If the subclasser returns a
   *                                         null class or one which is not
   *                                         a subclass of the parameter class
   * @exception InterruptedException If the current thread is interrupted
   *                                 while blocking waiting for a record
   * @see PREDATOR.ADT.DataObject
   */
  public synchronized DataObject[] getNextRecord(SubclassSelector Subclasser) 
    throws BadSubclassSelectorException,
           InterruptedException
  {
    while (!isReady()) {
      wait();
    }
    if (isEmpty())
      return null;
    DataRecord record = buffer[nextSlotToReturn];
    buffer[nextSlotToReturn] = null;
    nextSlotToReturn = (nextSlotToReturn + 1) % BUFFER_SIZE;
    notifyAll();
    // Get an array of the objects
    DataObject objects[] = record.getObjects(schema, Subclasser);
    return objects;
  }
}
