package PREDATOR;

/**
 * The ServerErrorBuffer is the error buffer between the ErrorReader thread
 * and the main thread operating the Connection object.  It provides the
 * means for the ErrorReader thread to read in the (essentially) asynchronous
 * server errors and allow the Connection object to present them synchronously.
 * @author Mark A. Paskin
 * @version 0.9
 * @see ErrorReader
 * @see Connection
 */
public class ServerErrorBuffer {
  /**
   * The error buffer
   */
  protected ServerError error;
  /**
   * A status code, indicating if an error has been received or not.
   */
  protected int status;
  /**
   * If the status variable has this value, then the server indicated that
   * no error occurred during the query.
   */
  protected final static int EMPTY = 0;
  /**
   * If the status variable has this value, then the server indicated that
   * an error occurred during the query.
   */
  protected final static int FULL = 1;
  /**
   * If the status variable has this value, then the the server has not yet
   * communicated the error status.
   */
  protected final static int WAITING = 2;
  /**
   * Resets this buffer to be empty, without any error information, and
   * waiting for an error (or an A-OK) signal to be sent.
   */
  public synchronized void reset() 
  {
    error = null;
    status = WAITING;
  }
  /**
   * Resets this buffer to be empty, without any error information.
   */
  public synchronized void clear() 
  {
    error = null;
    status = EMPTY;
  }
  /**
   * Waits until this error buffer is filled (with an error or an A-OK signal).
   */
  public synchronized void complete()
  {
    try {
      while (status == WAITING)
	wait();
    } catch (InterruptedException X) {
      throw new PREDATOR.InternalError(X);
    }
  }

  /**
   * Checks to see if there is a pending server error.
   * @return true if the server has registered an error, false otherwise
   */
  public synchronized boolean isError()
  {
    return (status == FULL);
  }  
  /**
   * If an error exists, returns the error and clears the error buffer.  If
   * none exists, returns null.
   * @return The server error
   */
  public synchronized ServerError getError()
  {
    if (status == FULL) {
      ServerError E = this.error;
      error = null;
      status = EMPTY;
      return E;
    } else
      return null;
  }
  /**
   * Places an error in the buffer, replacing any error already buffered.
   */
  public synchronized void noteError(ServerError E)
  {
    this.error = E;
    status = FULL;
    notifyAll();
  }
  /**
   * Signals that no error occurred during a query.
   */
  public synchronized void noError()
  {
    status = EMPTY;
    notifyAll();
  }
}
