#ifndef XXX_ALIAS_H
#define XXX_ALIAS_H

#include "prd-types.h"

/** 
* XxxAliasADT is a data type which is an alias to another data type
* specified during construction. All the virtual functions of
* the ancestor class XxxADTClass are overridden to call the
* corresponding functions of the alias-to ADTClass, except Print()
* @author BinQuan Gu, Jyhtar Chong
* @version January 28, 1997
* @see
*/

class XxxAliasADT : public XxxADTClass {
	
protected:
  XxxADTMetaInfo* GetMIInfo(const XxxADTMetaInfo* inMI) const;
	
public:
	
  /**
   *  Check the existence of data type with "Name" in the 
   *  "typelist" file of defdisk, if not, return XXX_OK. 
   */
  static XxxErrCode CheckName(int typeId, const char *Name);
	
  /**
   *  When a new alias is created(in sql.y), first "CheckName" 
   *  is called to check the existence of the name in the system 
   *  type list. If the name is not in the system yet, "Register" 
   *  method is called to insert the record about this alias into
   *  the file "typeFileId", and the whole process is synchronized 
   *  by mutex.
   */
  XxxErrCode Register();
	
  /**
   *  Drop an alias with name "typeName", doesn't work yet. TODO
   */
  //static XxxErrCode Drop(const char* typeName);
	
  /**
   *  Parse the data stored in typeList file.
   *  @param data: contains all the information to recreate the alias ADT.
   */
  static XxxErrCode UnSerialize(const char* data, XxxBasicType& Id, char *&aliasToName, char *&metaString);
	
  /// the ID of the type that this alias is associated to (to be known as "original type") 
  int AliasToID;
  // Meta Handle
  XxxADTClass::MetaInfoHandle MIHndl;
  // Meta string with the alias ADT
  char *MetaString;
	
  ///the pointer to the instance of the original type 
  XxxADTClass *AliasTo;  
	
  XxxAliasADT(XxxBasicType Id, const char *Name, const char *aliasToName, 
	      const char *metaString, XxxErrCode &status);
	
  virtual ~XxxAliasADT();
	
  /**
   * Check whether this type maintains metainformation. Used by 
   * RelShoreDataSrc::RelShoreCatalog::getAttrInfo to determine whether it
   * should retrieve metainfo for this ADT.
   * Alias ADT will overwrite it, and return the maintainsMetaInfo field
   * of the ADT it aliases to.
   */
  virtual XxxBool MaintainsMetaInfo()
  { return AliasTo->MaintainsMetaInfo();}
	
  /**
   * Check whether the meta information has been stored for this ADT. 
   */
  virtual XxxBool SpecifiedMetaInfo()
  {	
    if (AliasTo->MaintainsMetaInfo() && MetaString == NULL)
      return XXX_TRUE;
    return XXX_FALSE;
  }
	
  XxxBool IsAlias(){ return XXX_TRUE;}
	
  // Get the methodinfo for a method by its index in meth array
  virtual FuncMethodInfo* GetMethByIndex(int index) const
  { 
    return AliasTo->GetMethByIndex(index);
  }
	
  // Get aggrmethinfo for an aggregate method by its index in
  // aggr meth array
  virtual AggrMethodInfo* GetAggrMethByIndex(int index) const
  { 
    return AliasTo->GetAggrMethByIndex(index);
  }

  virtual int MaxObjectSize(const XxxADTMetaInfo* MI) const 
  { return AliasTo->MaxObjectSize(GetMIInfo(MI)); }
	
  virtual XxxBool IsSmall(const XxxADTMetaInfo* MI) const 
  { return AliasTo->IsSmall(GetMIInfo(MI)); }
	
  virtual void PrintStats(ostream& OStr) const 
  { AliasTo->PrintStats(OStr); }
	
  virtual void ResetStats() 
  { AliasTo->ResetStats(); }
	
  virtual int TypeSize(const XxxADTValue* c/* Object */, 
		       const XxxADTMetaInfo* MetaInfo) const
  { return AliasTo->TypeSize(c, GetMIInfo(MetaInfo)); }
	
  virtual int TypeCopy(const XxxADTValue* ObjectRef, XxxADTValue *ResultObject,
		       const XxxADTMetaInfo* MetaInfo) const 
  { return AliasTo->TypeCopy(ObjectRef, ResultObject, GetMIInfo(MetaInfo)); }
	
  virtual XxxBool  Equals(const XxxADTValue* Obj1, const XxxADTMetaInfo* MI1, 
			  const XxxADTValue* Obj2, const XxxADTMetaInfo* MI2)  const 
  { return AliasTo->Equals(Obj1, MI1, Obj2, MI2);}
	
  virtual XxxBool OrderedType(const XxxADTMetaInfo* MI) const 
  { return AliasTo->OrderedType(GetMIInfo(MI));}
	
  virtual XxxBool  LessThan(const XxxADTValue* Val1, 
			    const XxxADTMetaInfo* MI1,
			    const XxxADTValue* Val2,
			    const XxxADTMetaInfo* MI2) const 
  { return AliasTo->LessThan(Val1, MI1, Val2, MI2);}
	
  virtual XxxBool  GreaterThan(const XxxADTValue* Val1, 
			       const XxxADTMetaInfo* MI1,
			       const XxxADTValue* Val2,
			       const XxxADTMetaInfo* MI2) const
  { return AliasTo->GreaterThan(Val1, MI1, Val2, MI2);}
	
  virtual int TypeStoreCopy(const XxxADTValue* ObjectRef, XxxADTValue *ResultObject,
			    const XxxADTMetaInfo* MetaInfo) const 
  { return AliasTo->TypeStoreCopy(ObjectRef, ResultObject, GetMIInfo(MetaInfo)); }
	
  virtual XxxErrCode CopyHToN(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const
  { return AliasTo->CopyHToN(SrcObject, ResultObject, GetMIInfo(MetaInfo)); }
	
  virtual XxxErrCode CopyNToH(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const
  { return AliasTo->CopyNToH(SrcObject, ResultObject, GetMIInfo(MetaInfo)); }
	
  virtual char* TypeCreate(const XxxADTValue* ObjectRef, 
			   const XxxADTMetaInfo* MetaInfo) const 
  { return AliasTo->TypeCreate(ObjectRef, GetMIInfo(MetaInfo));}
	
  virtual void TypeDestroy(XxxADTValue *Object, 
			   const XxxADTMetaInfo* MetaInfo) const 
  { AliasTo->TypeDestroy(Object, GetMIInfo(MetaInfo)); }
	
  virtual XxxErrCode SerializeCopySize(const XxxADTMetaInfo* MetaInfo,
				       unsigned int& Size) const
  { return AliasTo->SerializeCopySize(GetMIInfo(MetaInfo), Size); }
	
  virtual XxxErrCode SerializeCopy(const XxxADTValue* Object,
				   const XxxADTMetaInfo* MetaInfo,
				   char* MemLocation) const
  { return AliasTo->SerializeCopy(Object, GetMIInfo(MetaInfo), MemLocation); }
	
  virtual XxxErrCode UnSerializeCopy(const char* MemLocation,
				     const XxxADTMetaInfo* MetaInfo,
				     XxxADTValue* Object) const
  { return AliasTo->UnSerializeCopy(MemLocation, GetMIInfo(MetaInfo), Object); }
	
  virtual XxxErrCode DefaultVal(XxxADTValue* ObjectRef, 
				const XxxADTMetaInfo* MetaInfo) const 
  { return AliasTo->DefaultVal(ObjectRef, GetMIInfo(MetaInfo)); }
	
  virtual XxxErrCode WriteText(ostream& OutStream, const XxxADTValue* Object,
			       const XxxADTMetaInfo* MetaInfo) const
  { return AliasTo->WriteText(OutStream, Object, GetMIInfo(MetaInfo)); }
	
  virtual XxxErrCode ReadText(istream& InStream, XxxADTValue* ObjectRef,
			      const XxxADTMetaInfo* MetaInfo) const
  { return AliasTo->ReadText(InStream, ObjectRef, GetMIInfo(MetaInfo));}
	
  virtual XxxErrCode AddMetaInfo(const char* Name, 
				 const XxxBool Named,
				 const char* MetaInfoString,
				 MetaInfoHandle& MIHandle)
  { return AliasTo->AddMetaInfo(Name, Named, MetaInfoString, MIHandle); }
	
//Bjarke, aug. 8: const added to char* parameter
  virtual XxxErrCode GetMetaInfo(const char* ObjName,
				 const MetaInfoHandle& MIHandle,
				 XxxADTMetaInfo* &MetaInfo,
				 XxxBool& Found) const
  { return AliasTo->GetMetaInfo(ObjName, MIHandle, MetaInfo, Found);}
	
  virtual XxxBool SupportsCallback() const
  { return AliasTo->SupportsCallback(); }
	
  virtual const char * GetContentType(const XxxADTValue *ResultHandle,
				      const XxxADTMetaInfo *MetaInfo,
				      const char *Message,
				      const int MessageSize) const
  { return AliasTo->GetContentType(ResultHandle, MetaInfo, Message, MessageSize); }
	
  virtual XxxErrCode GetResultSize(const XxxADTValue *ResultHandle,
				   const XxxADTMetaInfo *MetaInfo,
				   const char *Message,
				   const int MessageSize,
				   int &ResultSize) const
  { return AliasTo->GetResultSize(ResultHandle, MetaInfo, Message, MessageSize, ResultSize); }
	
  virtual XxxErrCode ServiceCallback(const XxxADTValue *ResultHandle,
				     const XxxADTMetaInfo *MetaInfo,
				     const char *Message,
				     const int MessageSize,
				     ostream &OStr) const 
  { return AliasTo->ServiceCallback(ResultHandle, MetaInfo, Message, MessageSize, OStr); }
	
  virtual XxxErrCode CreateStatsInfo(const XxxADTMetaInfo *MetaInfo,
				     TypeStats*& StatsInfo) const
  { return AliasTo->CreateStatsInfo(GetMIInfo(MetaInfo), StatsInfo);}
	
  virtual XxxErrCode StoreStatsInfo(const XxxADTMetaInfo *MetaInfo,
				    const TypeStats* StatsInfo,
				    MetaInfoHandle& MIHandle)
  { return AliasTo->StoreStatsInfo(GetMIInfo(MetaInfo), StatsInfo, MIHandle); }
	
  virtual XxxErrCode RetrieveStatsInfo(const XxxADTMetaInfo *MetaInfo,
				       const MetaInfoHandle& MIHandle,
				       TypeStats*& StatsInfo)
  { return AliasTo->RetrieveStatsInfo(GetMIInfo(MetaInfo), MIHandle, StatsInfo); }
	
  virtual XxxErrCode FuncTypeCheck(const char*FuncName,
				   const XxxValueExpression* Owner,
				   const XxxValueExprList *ArgList,
				   XxxBool IsAggr,
				   XxxBasicType& ReturnType,
				   XxxFuncParseInfo*& FuncParse) const;
	
  virtual XxxErrCode FuncReOrganize(XxxValueExpression*& FuncExpression) const
  { return AliasTo->FuncReOrganize(FuncExpression); }
	
  virtual XxxErrCode FuncOptimize(XxxFuncParseInfo* FuncParse,
				  XxxValueExprPlan* OwnerPlan,
				  XxxValExprPlanList* ArgPlans,
				  XxxFuncPlanInfo*& FuncPlan) const
  { return AliasTo->FuncOptimize(FuncParse, OwnerPlan, ArgPlans, FuncPlan); }
	
  /**
   * Determines if this ADT can cast an object of another type into
   * its own type.
   * @param SourceType The type to cast from
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   */
  virtual XxxBool CanCastFrom(XxxBasicType SourceType, 
			      const XxxADTMetaInfo* SourceMetaInfo,
			      const XxxADTMetaInfo* TargetMetaInfo)
  { return AliasTo->CanCastFrom(SourceType, SourceMetaInfo, TargetMetaInfo); }
  /**
   * Executes a cast from another type into this type.
   * @param CastType The type id of the old object
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   * @param OldObject The object to be casted
   * @param NewObject The new object to be created
   */
  virtual XxxErrCode CastFrom(XxxBasicType SourceType, 
			      const XxxADTMetaInfo* SourceMetaInfo,
			      const XxxADTMetaInfo* TargetMetaInfo,
			      XxxADTValue* OldObject, 
			      XxxADTValue* NewObject)
  { return AliasTo->CastFrom(SourceType, SourceMetaInfo, TargetMetaInfo, OldObject, NewObject); }
  /**
   * Determines if this ADT can cast an object of its type into
   * another type.
   * @param TargetType The type to cast to
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   */
  virtual XxxBool CanCastTo(XxxBasicType TargetType, 
			    const XxxADTMetaInfo* SourceMetaInfo,
			    const XxxADTMetaInfo* TargetMetaInfo)
  { return AliasTo->CanCastTo(TargetType, SourceMetaInfo, TargetMetaInfo); }
  /**
   * Executes a cast from this type into another type.
   * @param CastType The type id of the new object
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   * @param OldObject The object to be casted
   * @param NewObject The new object to be created 
   */
  virtual XxxErrCode CastTo(XxxBasicType TargetType, 
			    const XxxADTMetaInfo* SourceMetaInfo,
			    const XxxADTMetaInfo* TargetMetaInfo,
			    XxxADTValue* OldObject, 
			    XxxADTValue* NewObject)
  {return AliasTo->CastTo(TargetType, SourceMetaInfo, TargetMetaInfo, OldObject, NewObject); }
	
  virtual XxxErrCode CreateEnv( XxxADTEnv* &Env ) const
  { return AliasTo->CreateEnv(Env); }
	
  virtual XxxErrCode DeleteEnv( XxxADTEnv* Env ) const
  { return AliasTo->DeleteEnv(Env); }
	
  virtual XxxBool GetBoundingBox(const XxxADTValue* &Object, XxxBoundingBox& Box) 
  { return AliasTo->GetBoundingBox(Object, Box); }
	
  virtual XxxErrCode GetCompOp(const char* FuncName, 
			       XxxCompareOp& CompOp, XxxBool& Success) const
  { return AliasTo->GetCompOp(FuncName, CompOp, Success);}
	
};
 
#endif
