#ifndef XXX_COUNTER_H
#define XXX_COUNTER_H
#include "genadt.h"
#include "shoresm.h"

/**
 * This ADT is a counter which can be used in a relation to provide a
 * sequence of unique ordered integer values.  This is useful for adding a
 * record ID to a record when inserting it into a relation.
 *
 * The meta-info stores the most recently-assigned value, so it is always 
 * accessible as a default value for insertion.  This data type guarantees
 * that all objects bound to a single meta-info object will have unique
 * values.  It is not necessarily guaranteed that all values are used.
 *
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version June 26, 1997
 * @see
 */
class XxxCounterADT : public XxxSmallFixedADT {
protected:
  class CounterStruct :public XxxADTValue {
   public:
    int value;
   };

  /**
   * Meta-info for counters.  This structure holds information on the
   * domain of the counter as well as the most-recently assigned value.
   */
  class CounterMetaInfo : public XxxADTMetaInfo { 
    friend class XxxCounterADT;
  protected:
    /// The next value to be assigned
    int nextValue;
    /// Strangely enough, the OID of this object in SHORE
    XxxShoreObjectId myId;
  public:
    /// Constructor
    CounterMetaInfo() : nextValue(0) {}
    /// Destructor
    virtual ~CounterMetaInfo() {}
    /**
     * Makes a copy of this meta-info object
     */
    virtual XxxErrCode Replicate(XxxADTMetaInfo*& NewMI) const;
    /**
     * Prints this meta-info object to the output stream
     */
    virtual void Print(ostream& OStr) const;
    /**
     * "Constant" values for counters should not define their own 
     *  meta-infos.
     */
    virtual XxxBool IsCompatible(const XxxADTMetaInfo* OtherMI) const {
      return XXX_FALSE;
     }
    /**
     * Returns the number of bytes necessary to hold the serialized version
     * of this meta-info object.
     * @param Size (out) The required number of bytes
     * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
     */
    virtual XxxErrCode SerializeCopySize(int &Size) const {
      // Size = sizeof(int) + sizeof(XxxShoreObjectId); return XXX_OK;
      // do not print out the shore OID
      Size = sizeof(int); return XXX_OK;
    }
    /**
     * Creates a binary version of this meta-info object.
     * @param Data An array of bytes of at least the size specified by the
     *             SerializeCopySize() function.
     * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
     */
    virtual XxxErrCode SerializeCopy(char *Data) const {
      int TmpNextVal;
      XxxNetworkInt(&TmpNextVal, nextValue);
      memcpy(Data, (char *)(&TmpNextVal), sizeof(int));
      return XXX_OK;
    }
  };

public:
  /// Type ID accessible without object
  static XxxBasicType CounterTypeId;

  /**
   * Constructor.
   */
  XxxCounterADT(XxxBasicType Id);

  /**
   * Destructor.
   */
  virtual ~XxxCounterADT();

  virtual XxxBool OrderedType() const 
   { return XXX_TRUE;}

  /**
   * Prints the name of the ADT.
   */
  virtual void Print(ostream& OStr, const XxxADTMetaInfo* MetaInfo) const;

  /**
   * A method to determine if two objects of this type are identical.
   */
  virtual XxxBool  Equals(const XxxADTValue* Obj1, const XxxADTMetaInfo*, 
			  const XxxADTValue* Obj2, const XxxADTMetaInfo*)  const 
  { return (((CounterStruct *)Obj1)->value == 
	    ((CounterStruct *)Obj2)->value); }

  /**
   * Assigns a "default" value--the next value to be assigned.
   */
  virtual XxxErrCode DefaultVal(XxxADTValue* ObjectRef, 
				const XxxADTMetaInfo* MetaInfo) const;

  /**
   * Makes a readable text string version of the object from the tuple
   * representation.
   */
  virtual XxxErrCode WriteText(ostream& OutStream, 
			       const XxxADTValue* Object,
			       const XxxADTMetaInfo* MetaInfo) const;

  /**
   * Reads a text version of the object into the tuple representation of it, 
   * after memory has been allocated for the tuple representation.
   */
  virtual XxxErrCode ReadText(istream& InStream, XxxADTValue* ObjectRef,
			      const XxxADTMetaInfo* MetaInfo) const;

  /**
   * Compares two objects of this type.
   * @return XXX_TRUE if the first argument is "less than" the second
   */
  virtual XxxBool  LessThan(const XxxADTValue* Val1, 
			    const XxxADTMetaInfo*,
			    const XxxADTValue* Val2,
			    const XxxADTMetaInfo*) const 
  { return (((CounterStruct *)Val1)->value < 
	    ((CounterStruct *)Val2)->value); }

  /**
   * Compares two objects of this type.
   * @return XXX_TRUE if the first argument is "greater than" the second
   */
  virtual XxxBool  GreaterThan(const XxxADTValue* Val1, 
			       const XxxADTMetaInfo*,
			       const XxxADTValue* Val2,
			       const XxxADTMetaInfo*) const
  { return (((CounterStruct *)Val1)->value > 
	    ((CounterStruct *)Val2)->value); }

  /**
   * Determines if it is possible to cast a counter into another type .
   */
  virtual XxxBool CanCastTo(XxxBasicType TargetType, 
			    const XxxADTMetaInfo* SourceMetaInfo,
			    const XxxADTMetaInfo* TargetMetaInfo);
  /**
   * Executes a cast from counter into another type
   */
  virtual XxxErrCode CastTo(XxxBasicType TargetType, 
			    const XxxADTMetaInfo* SourceMetaInfo,
			    const XxxADTMetaInfo* TargetMetaInfo,
			    XxxADTValue* OldObject, 
			    XxxADTValue* NewObject);

  virtual XxxErrCode CopyHToN(const XxxADTValue* ObjectRef, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxErrCode CopyNToH(const XxxADTValue* ObjectRef, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
 protected:
  /**
   * parse a new meta-info object for this class.
   * @param MetaInfoString A string containing information on the contents
   *                       of the meta-info.  In the counter's case, empty.
   * @param MIHandle (out) A handle on the newly-created meta info
   */
  virtual XxxErrCode parseMetaInfo(const char* MetaInfoString,
				   MetaInfoHandle& MIHandle) const;

  /**
   * Make the meta-info object associated with its handle and pins
   * the counter.
   * @param ObjName The name given to the object if one is provided
   * @param MIHandle (in) A handle on the meta info
   * @param MetaInfo (out) A pointer to a newly-created MI object
   * @param Found (out) XXX_FALSE if the MI was not found
   */
  virtual XxxErrCode makeMetaInfo(const MetaInfoHandle& MIHandle,
				  XxxADTMetaInfo* &MetaInfo) const;
};

#endif /* XXX_COUNTER_H */
