#ifndef DATE_H
#define DATE_H

#include "types.h"
#include "basictypes.h"
#include "genadt.h"
#include <time.h>
#include "adtfunc.h"

/**
 * JulianCalADT (base class for all date ADTs ... never directly used)
 **/
class JulianCalADT: public XxxSmallFixedADT {
protected:
  // string used by ReadText to specify a get of the current date
  static const char NowString[];

  static const int SEQ_WEEK;
  static const int SEQ_MONTH;
  static const int SEQ_YEAR;
  static const int SEQ_CENTURY;
  static const int StartDay;
  static const int StartWeek;
  static const int StartMonth;
  static const int StartYear;
  static const int MonthDays[13];
  // running sum of the above
  static const int AccumMonthDays[13];
  static const int LeapYears[100];
  // running sum of the above
  static const int AccumLeapYears[100];

  static const int YearsPerCent;
  // assuming that it is not a century divisible by 4!
  static const int DaysPerCent;
  
  // some constants useful for approximate calculations
  static const float DaysPerMonth;
  static const float DaysPerYear;

  // const values of interest for hour/minute/sec calcs
  static const int SEQ_SECOND;
  static const int SEQ_MINUTE;
  static const int SEQ_HOUR;

  static const int StartSecond;
  static const int EndSecond;
  static const int StartMinute;
  static const int EndMinute;
  static const int StartHour;
  static const int EndHour;


  static int endYear(int) { return 99;}
  static int endMonth(int) { return 12;}
  static int endWeek(int) { return 53;}

 public:
  inline static XxxBool isLeapYear(int Year, int Century); 
  inline static int endDay(int Month, int Year, int Century); 
  inline static void nextYear(int& Year, int& Cent);
  inline static void nextMonth(int& Month, int& Year, int& Cent); 
  inline static void nextDay(int& Day, int& Month, int& Year, int& Cent);
  inline static void prevYear(int& Year, int& Cent);
  inline static void prevMonth(int& Month, int& Year, int& Cent);
  inline static void prevDay(int& Day, int& Month, int& Year, int& Cent);
  inline static int daysPerYear(int Year, int Century);
  inline static int daysPerCent(int Century);


  // cumulative sum of the days in the year upto and including EndYear
  inline static int daysCentTotal(int EndYear, int Century);

  // approximate only
  inline static int numDays(int Day1, int Month1, int Year1, int Cent1,
			    int Day2, int Month2, int Year2, int Cent2);
  inline static int numMonths(int Month1, int Year1, int Cent1,
			      int Month2, int Year2, int Cent2);

  // routines for converting to/from the stored representation
  inline static int makeIntPos(int Day, int Month, int Yr, int Cent);
  inline static void splitIntPos(int IntPos, int& Day, int& Month, 
				 int& Yr, int& Cent);
 protected:

  // note: methOptimize *must* be overrided by final ADT

  /**
    * MetaInfo needed to determine input/output formats
    * class useful for representing generic date formats (also needed for MetaInfo)
    */
  class DateFormat { 
   public:
    // The date format can specify the following
    //   1) Century : C        19
    //   2) Year    : Y        00
    //   3) Month   : M        01
    //   4) Day     : D        01
    //   5) Hour    : h        00
    //   6) Minute  : m        00
    //   7) Second  : s        00
    //   8) Defaults for all of the above
    //   9) Allowed Separators  :  whitespace(w), /, <tab> (t), :, -, etc
    //  10) End Of Format String : e
    //   
    // The defaults used for these are listed in the second column
    // They should be settable by special commands from the prompt.
    // The format is expressed as a string. Here are some examples:
    //  05/22/1970             .....    M/D/CYe
    //  1970-05-22 23:54:13    .....    CY-M-Dwh:m:se
    //
    // The format can have at most 7 characters + 12 separators. Total size 19.
    // The format cannot start with the same character as special strings such as
    //     NowString
    char   Spec[20];
    static const char DefaultSpec[20];
    static unsigned int SpecSize() { return 20;}
    void ResetDefault()
     { strcpy(Spec, DefaultSpec); }

    DateFormat()
     { ResetDefault(); }
    DateFormat(const DateFormat& SrcFormat)
     { strcpy(Spec, SrcFormat.Spec); }

    // returns whether Century is specified as part of format
    XxxBool CentSpecified();
  };
  ////
  class DateMetaInfo : public XxxADTMetaInfo { 
   public:
    DateFormat       Format;
    static const int DefaultCent;
    static const int DefaultYear;
    static const int DefaultMonth;
    static const int DefaultDay;
    static const int DefaultHr;
    static const int DefaultMin;
    static const int DefaultSec;

    DateMetaInfo() : Format() { }
    DateMetaInfo(const DateFormat& FormatSpec) : Format(FormatSpec) { }
    virtual XxxErrCode Replicate(XxxADTMetaInfo*& NewMI) const {
      NewMI = new DateMetaInfo( this->Format );
      return XXX_OK;
    }
    virtual void Print(ostream& OStr) const {
      OStr << "\"" << Format.Spec << "\"";
    }
    // it doesnt matter what format is used: the date is represented
    // the same way in binary form: so any metainfo is compatible
    virtual XxxBool IsCompatible(const XxxADTMetaInfo* OtherMI) const {
      return XXX_TRUE;
     }
    /**
     * Returns the number of bytes necessary to hold the serialized version
     * of this meta-info object.
     * @param Size (out) The required number of bytes
     * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
     */
    virtual XxxErrCode SerializeCopySize(int &Size) const {
      Size = Format.SpecSize();  return XXX_OK;
    }
    /**
     * Creates a binary version of this meta-info object.
     * @param Data An array of bytes of at least the size specified by the
     *             SerializeCopySize() function.
     * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
     */
    virtual XxxErrCode SerializeCopy(char *Data) const {
      memcpy(Data, Format.Spec, Format.SpecSize());  return XXX_OK;
    }

    static XxxBool ValidDateFormat(const char *FormatString);
  };
  friend class DateMetaInfo;

  static XxxADTMetaInfo* OwnerMIGen(XxxValueExpression* Owner,
      XxxValueExprList* ArgList) {
    DateMetaInfo* OwnerMetaInfo = (DateMetaInfo*) (Owner->GenerateCatInfo());
    if (OwnerMetaInfo) {
      DateMetaInfo* ReturnMetaInfo = new DateMetaInfo(*OwnerMetaInfo);
      return ReturnMetaInfo;
    }
    else return NULL;
  }

  // necessary overrides for virtuals in XxxADTClass
  virtual XxxErrCode parseMetaInfo(const char* MetaInfoString,
				   MetaInfoHandle& MIHandle) const;

  virtual XxxErrCode makeMetaInfo(const MetaInfoHandle& MIHandle,
				  XxxADTMetaInfo*& MI) const;

public:
  typedef XxxCoreADTClass::IntValue DateValue;
  //// Function Plans for date data type
  class DatePlanInfo : public XxxFuncPlanInfo {
  public:
    virtual void NextPos(const XxxCoreADTClass::IntValue* Pos, 
			 int HowMany, 
			 const XxxADTMetaInfo* MetaInfo, 
			 XxxCoreADTClass::IntValue* NextPos) const;
    virtual void PrevPos(const XxxCoreADTClass::IntValue* Pos, 
			 int HowMany, 
			 const XxxADTMetaInfo* MetaInfo, 
			 XxxCoreADTClass::IntValue* PrevPos) const;

    virtual void NextPos(const XxxCoreADTClass::IntValue* Pos, 
			 const XxxADTMetaInfo* MetaInfo, 
			 XxxCoreADTClass::IntValue* NextPos) const = 0;
    virtual void PrevPos(const XxxCoreADTClass::IntValue* Pos, 
			 const XxxADTMetaInfo* MetaInfo, 
			 XxxCoreADTClass::IntValue* PrevPos) const = 0;
    virtual int  NumPositions(const XxxCoreADTClass::IntValue* Pos1, 
			      const XxxCoreADTClass::IntValue* Pos2) const = 0;
  public:
    DatePlanInfo(const XxxFuncParseInfo* FuncParse, int argCount) :
      XxxFuncPlanInfo(FuncParse, argCount) {}
  };

  JulianCalADT(XxxBasicType Id, const char* Name, int DataSize) 
    : XxxSmallFixedADT(Id, Name, DataSize, XXX_TRUE) {}

  virtual XxxBool OrderedType(const XxxADTMetaInfo*) const
    { return XXX_TRUE; }

  virtual XxxBool  Equals(const XxxADTValue* Pos1, const XxxADTMetaInfo*, 
			  const XxxADTValue* Pos2, const XxxADTMetaInfo*)  const 
    { return (((XxxCoreADTClass::IntValue*)Pos1)->Value == 
	      ((XxxCoreADTClass::IntValue*)Pos2)->Value); }
  virtual XxxBool  LessThan(const XxxADTValue* Pos1, 
			    const XxxADTMetaInfo*,
			    const XxxADTValue* Pos2,
			    const XxxADTMetaInfo*) const 
    { return (((XxxCoreADTClass::IntValue*)Pos1)->Value < 
	      ((XxxCoreADTClass::IntValue*)Pos2)->Value); }
  virtual XxxBool  GreaterThan(const XxxADTValue* Pos1, 
			       const XxxADTMetaInfo*,
			       const XxxADTValue* Pos2,
			       const XxxADTMetaInfo*) const
    { return (((XxxCoreADTClass::IntValue*)Pos1)->Value > 
	      ((XxxCoreADTClass::IntValue*)Pos2)->Value); }

};  // JulianCalADT

/**
 * JulianDayCalADT
 * Note: not yet a full ADT
 */
class JulianDayCalADT: public JulianCalADT {
protected:
  // actually, these should be virtual functions here that are
  // defined in the subclasses. Instead, since i know that there
  // will only be one instance of each class created, i use 
  // variables maxPos and minPos that will be initialized in the
  // subclas constructors.
  int maxPos, minPos;
public:
  JulianDayCalADT(XxxBasicType Id, const char* Name, 
		  int DataSize=sizeof(DateValue))
    : JulianCalADT(Id, Name, DataSize) {}
  virtual void MaxPos(char*& Pos) const { *((int*)Pos) = maxPos;}
  virtual void MinPos(char*& Pos) const { *((int*)Pos) = minPos;}
};

/**
 * JulianMDYCalADT            (MONTH, DAY, YEAR) support
 */
class JulianMDYCalADT: public JulianDayCalADT {
public:
  ////
  class MDYPlanInfo : public JulianCalADT::DatePlanInfo {
  public:
    XxxErrCode execMonth(XxxValueEnv* Env, XxxADTValue* ReturnValue) const;
    XxxErrCode execDay(XxxValueEnv* Env, XxxADTValue* ReturnValue) const;
    XxxErrCode execYear(XxxValueEnv* Env, XxxADTValue* ReturnValue) const;
    XxxErrCode execCentYear(XxxValueEnv* Env, XxxADTValue* ReturnValue) const;
    XxxErrCode execCent(XxxValueEnv* Env, XxxADTValue* ReturnValue) const;
    XxxErrCode execNextYear(XxxValueEnv* Env, XxxADTValue* ReturnValue) const;
    XxxErrCode execPrevYear(XxxValueEnv* Env, XxxADTValue* ReturnValue) const;
    XxxErrCode execNextMonth(XxxValueEnv* Env, XxxADTValue* ReturnValue) const;
    XxxErrCode execPrevMonth(XxxValueEnv* Env, XxxADTValue* ReturnValue) const;
    XxxErrCode execNextPos(XxxValueEnv* Env, XxxADTValue* ReturnValue) const;
    XxxErrCode execPrevPos(XxxValueEnv* Env, XxxADTValue* ReturnValue) const;

    virtual void NextYear(const JulianCalADT::DateValue* Pos, 
			 int HowMany,
			 const XxxADTMetaInfo* MetaInfo, 
			 JulianCalADT::DateValue* NextPos) const;
    virtual void PrevYear(const JulianCalADT::DateValue* Pos, 
			 int HowMany,
			 const XxxADTMetaInfo* MetaInfo, 
			 JulianCalADT::DateValue* PrevPos) const;
    virtual void NextMonth(const JulianCalADT::DateValue* Pos, 
			 int HowMany,
			 const XxxADTMetaInfo* MetaInfo, 
			 JulianCalADT::DateValue* NextPos) const;
    virtual void PrevMonth(const JulianCalADT::DateValue* Pos, 
			 int HowMany,
			 const XxxADTMetaInfo* MetaInfo, 
			 JulianCalADT::DateValue* PrevPos) const;
    virtual void NextPos(const JulianCalADT::DateValue* Pos, 
			 int HowMany,
			 const XxxADTMetaInfo* MetaInfo, 
			 JulianCalADT::DateValue* NextPos) const;
    virtual void PrevPos(const JulianCalADT::DateValue* Pos, 
			 int HowMany,
			 const XxxADTMetaInfo* MetaInfo, 
			 JulianCalADT::DateValue* PrevPos) const;

    virtual void NextPos(const JulianCalADT::DateValue* Pos, 
			 const XxxADTMetaInfo* MetaInfo, 
			 JulianCalADT::DateValue* NextPos) const;
    virtual void PrevPos(const JulianCalADT::DateValue* Pos, 
			 const XxxADTMetaInfo* MetaInfo, 
			 JulianCalADT::DateValue* PrevPos) const;
    virtual int  NumPositions(const JulianCalADT::DateValue* Pos1, 
			      const JulianCalADT::DateValue* Pos2) const;
  public:
    MDYPlanInfo(const XxxFuncParseInfo* FuncParse, int argCount) :
      JulianCalADT::DatePlanInfo(FuncParse, argCount) {}
  };
  friend class MDYPlanInfo;

  // XxxADTClass::methOptimize not good enough
  virtual XxxErrCode methOptimize(XxxFuncParseInfo* FuncParse,
				  XxxValueExprPlan* OwnerPlan,
				  XxxValExprPlanList* ArgPlans,
				  XxxFuncPlanInfo*& FPlan) const {
    FPlan = new MDYPlanInfo(FuncParse, (ArgPlans?ArgPlans->Count():0));
    return XXX_OK;
  }

public:
  static int MDYOrdDomId; // offset of this type in global type table

  JulianMDYCalADT(XxxBasicType Id);

  virtual XxxErrCode WriteText(ostream& OutStream, 
			       const XxxADTValue* Pos,
			       const XxxADTMetaInfo* MInfo) const;
  virtual XxxErrCode ReadText(istream& InStream, 
			      XxxADTValue* PosData,
			      const XxxADTMetaInfo* MInfo) const;

  virtual XxxErrCode CopyHToN(const XxxADTValue* ObjectRef, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxErrCode CopyNToH(const XxxADTValue* ObjectRef, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
};

/********************************************************************************
 * Now the definitions of the VIRTUAL methods
 *******************************************************************************/

inline XxxBool JulianCalADT::isLeapYear(int Year, int Century) {
  if (!(Year%4) && (Year%100 || !(Century%4)))
    return XXX_TRUE;
  else
    return XXX_FALSE;
}
//------------------------------------------------------------------
inline int JulianCalADT::endDay(int Month, int Year, int Century) {
  int End = MonthDays[Month];
  if ((Month == 2) && isLeapYear(Year, Century)) {
    End++;
  }
  return End;
}
//------------------------------------------------------------------
inline void JulianCalADT::nextYear(int& Year, int& Cent) {
  if (Year >= endYear(Cent)) {
    Year = StartYear;
    Cent++;
  }
  else Year++;
}
//------------------------------------------------------------------
inline void JulianCalADT::nextMonth(int& Month, int& Year, int& Cent) {
  if (Month >= endMonth(Year)) {
    Month = StartMonth;
    nextYear(Year, Cent);
  }
  else Month++;
}
//------------------------------------------------------------------
inline void JulianCalADT::nextDay(int& Day, int& Month, int& Year, int& Cent) {
  if (Day >= endDay(Month, Year, Cent)) {
    Day = StartDay;
    nextMonth(Month, Year, Cent);
  }
  else Day++;
}
//------------------------------------------------------------------
inline void JulianCalADT::prevYear(int& Year, int& Cent) {
  if (Cent >= 0 && Year > 0) {
    if (Year <= StartYear) {
      Cent--;
      Year = endYear(Cent);
    }
    else Year--;
  }
}
//------------------------------------------------------------------
inline void JulianCalADT::prevMonth(int& Month, int& Year, int& Cent) {
  if (Month <= StartMonth) {
    prevYear(Year, Cent);
    Month = endMonth(Year);
  }
  else Month--;
}
//------------------------------------------------------------------
inline void JulianCalADT::prevDay(int& Day, int& Month, int& Year, int& Cent) {
  if (Day <= StartDay) {
    prevMonth(Month, Year, Cent);
    Day = endDay(Month, Year, Cent);
  }
  else Day--;
}
//------------------------------------------------------------------
inline int JulianCalADT::daysPerYear(int Year, int Century) {
  return (isLeapYear(Year, Century)?366:365);
}
//------------------------------------------------------------------
inline int JulianCalADT::daysPerCent(int Century) {
  return (isLeapYear(0, Century)?DaysPerCent:DaysPerCent+1);
}
//------------------------------------------------------------------
inline int JulianCalADT::daysCentTotal(int EndYear, int Century) {
  int LeapYears = AccumLeapYears[EndYear];
  if (!(Century%4)) LeapYears++;
  int NonLeapYears = EndYear + 1 - LeapYears;
  return (LeapYears*366)+(NonLeapYears*365);
}
//------------------------------------------------------------------
inline int JulianCalADT::numDays(int Day1, int Month1, 
    int Year1, int Cent1,
    int Day2, int Month2, 
    int Year2, int Cent2) 
{
  // int PosCount = 1;
  int PosCount = 0; 

  PosCount += (int)((Cent2 - Cent1)*DaysPerCent);
  if (Year2 > Year1)
    PosCount += (int)((Year2 - Year1)*DaysPerYear);
  else
    PosCount -= (int)((Year1 - Year2)*DaysPerYear);

  if (Month2 > Month1) 
    PosCount += (int)((Month2 - Month1)*DaysPerMonth);
  else 
    PosCount += (int)((Month1 - Month2)*DaysPerMonth);
  PosCount += (Day2 - Day1);
  return PosCount;
}
//------------------------------------------------------------------
inline int JulianCalADT::numMonths(int Month1, int Year1, 
    int Cent1,  int Month2, 
    int Year2, int Cent2) 
{
  // int PosCount = 1;
  int PosCount = numDays(1, Month1, Year1, Cent1,
      1, Month2, Year2, Cent2);
  return (int)(PosCount/DaysPerMonth);
}

//------------------------------------------------------------------
inline int JulianCalADT::makeIntPos(int Day, int Month, 
    int Yr, int Cent) {
  return Cent*SEQ_CENTURY+Yr*SEQ_YEAR+Month*SEQ_MONTH+Day;
  // return Cent*SEQ_CENTURY+Yr*SEQ_YEAR+Month; if only MYC storage
}
//------------------------------------------------------------------

inline void JulianCalADT::splitIntPos(int IntPos, int& Day, 
    int& Month, int& Yr, 
    int& Cent) {
  Cent =  (IntPos/SEQ_CENTURY);
  Yr = ((IntPos%SEQ_CENTURY)/SEQ_YEAR);
  Month = (((IntPos%SEQ_CENTURY)%SEQ_YEAR)/SEQ_MONTH);
  Day = (((IntPos%SEQ_CENTURY)%SEQ_YEAR)%SEQ_MONTH);
  // omit last statement if only MYC storage
}

#endif
