/*
 *-------------------------------------------------------------
 * geometry.C --  
 *
 * -- Dave Koster, 11/18/96
 *
 * Contains all the point, box, circle, line segment, polygon,
 * line, and path operations code.
 *-------------------------------------------------------------
 */

#include "geometry.h"
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <memory.h>

/*

  Taken from s2k site at UW-Madison:

  Modified by Paul Brown, February, 1994, to get the point_in_poly
  better. Check the box boundary to exclude all by the likely cases,
  and then perform the complete point_in_bound_path algorhythm.

  Points on the perimiter are included.

  For a detailed explaination of this algorhythm, refer to    

  "Point in Polygon Strategies", Eric Haines,
  Graphics Gems IV, ed. Paul Heckbert, Academic Press, p. 23-45.

*/
  


/*
 *-------------------------------------------------------------
 *
 * Circle Operations:
 *
 * circle_make
 * circle_MBR
 * circle_equals
 *
 *-------------------------------------------------------------
 */

/*
 * circle_make --
 *
 * Returns a new XXX_CIRCLE given a XXX_POINT and radius.
 */

XXX_CIRCLE *circle_make(XXX_POINT *loc, double radius) {
  XXX_CIRCLE *circle = new XXX_CIRCLE();
  
  circle->center = *loc;
  circle->radius = radius;
  return (circle);
}


/*
 * circle_MBR --
 *
 * Returns a XXX_BOX that is the MBR of the given XXX_CIRCLE.
 */

XXX_BOX *circle_MBR(XXX_CIRCLE *c) {
  XXX_BOX *box = new XXX_BOX();
  
  box->xl = c->center.x - c->radius;
  box->xh = c->center.x + c->radius;
  box->yl = c->center.y - c->radius;
  box->yh = c->center.y + c->radius;
  return box;
}

/*
 * circle_equals --
 *
 * Return 1 if the two circles are equal; 0 otherwise.
 */

int circle_equals(XXX_CIRCLE *c1, XXX_CIRCLE *c2) {
  return ((point_equals(&c1->center, &c2->center))&&isequal(c1->radius, c2->radius));
}



/*
 *-------------------------------------------------------------
 *
 * Box Operations:
 *
 * box_make
 * box_in_box
 * box_iin_box
 * box_equals
 * box_2_poly
 * box_overlap
 * box_ooverlap
 *
 *-------------------------------------------------------------
 */

/*
 * box_make --
 *
 * Given the (xl, yl), (xh, yh) pair, create and return
 * a point to a BOX.
 */

XXX_BOX *box_make(double xl, double yl, double xh, double yh) {
  XXX_BOX *b = new XXX_BOX();
  b->xl = xl;
  b->yl = yl;
  b->xh = xh;
  b->yh = yh;
  return b;
}

/* Overloaded function: */
XXX_BOX *box_make(XXX_POINT *ll, XXX_POINT *ur) {
  XXX_BOX *b = new XXX_BOX();
  b->xl = ll->x;
  b->yl = ll->y;
  b->xh = ur->x;
  b->yh = ur->y;
  return b;
}


/*
 * box_in_box --
 *
 * Return 1 if XXX_BOX b1 lies within XXX_BOX b2; 0 otherwise.
 */

int box_in_box(XXX_BOX *b1, XXX_BOX *b2) {
  return ( isge(b1->xl, b2->xl) && isge(b1->yl, b2->yl) && isle(b1->xh, b2->xh) && isle(b1->yh, b2->yh));
}


/*
 * box_iin_box --
 *
 * Return 1 if XXX_BOX b1 lies strictly within XXX_BOX b2; 0 otherwise.
 */

int box_iin_box(XXX_BOX *b1, XXX_BOX *b2) {
  return ( isgt(b1->xl, b2->xl) && isgt(b1->yl, b2->yl) && islt(b1->xh, b2->xh) && islt(b1->yh, b2->yh));
}


/*
 * box_equals --
 *
 * Return 1 if the two XXX_BOX's are equal; 0 otherwise.
 */

int box_equals(XXX_BOX *b1, XXX_BOX *b2) {
  return ( isequal(b1->xl, b2->xl) && isequal(b1->xh, b2->xh) && isequal(b1->yl, b2->yl) &&
	   isequal(b1->yh, b2->yh));
}


/*
 * box_2_poly --
 *
 * Converts the given XXX_BOX to a XXX_POLYGON and returns this
 * new polygon.
 */

XXX_POLYGON *box_2_poly(XXX_BOX *box) {
  XXX_POLYGON *p = new XXX_POLYGON();
  double *pts;
  
  p->npts = 4;
  p->pts = new double[(sizeof(double) * 8)];
  p->mbr = *box;
  pts = p->pts;
  *pts++ = box->xl;
  *pts++ = box->xh;
  *pts++ = box->xh;
  *pts++ = box->xl;
  *pts++ = box->yl;
  *pts++ = box->yl;
  *pts++ = box->yh;
  *pts   = box->yh;
  return p;
}


/*
 * box_overlap --
 *
 * Return 1 if the two boxes overlap; 0 otherwise.  This
 * is a relaxed check.  Two boxes need only share boundary
 * points.
 */

int box_overlap(XXX_BOX *b1, XXX_BOX *b2) {
  return ( (!(islt(b1->xh,b2->xl))) &&
	   (!(islt(b1->yh,b2->yl))) &&
	   (!(isgt(b1->xl,b2->xh))) &&
	   (!(isgt(b1->yl,b2->yh))));
}


/*
 * box_ooverlap --
 *
 * Return 1 if the two boxes strictly overlap; 0 otherwise.
 * This is a strict check.  Two boxes cannnot only share boundary
 * points.
 */

int box_ooverlap(XXX_BOX *b1, XXX_BOX *b2) {
  return ( (!(isle(b1->xh,b2->xl))) &&
	   (!(isle(b1->yh,b2->yl))) &&
	   (!(isge(b1->xl,b2->xh))) &&
	   (!(isge(b1->yl,b2->yh))));
}



/*
 *-------------------------------------------------------------
 *
 * Point Operations:
 *
 * point_make
 * point_equals
 * point_2_box
 * point_dist
 * point_in_box
 * point_iin_box
 * point_on_box
 * point_2_poly
 * point_in_poly
 * point_iin_poly
 * point_on_poly
 *
 *-------------------------------------------------------------
 */

/*
 * point_make --
 *
 * Create a new XXX_POINT with the given (x,y) coordinates
 * and return a pointer to it.
 */

XXX_POINT *point_make(double x, double y) {
  XXX_POINT *p = new XXX_POINT();
  p->x = x;
  p->y = y;
  return p;
}


/*
 * point_2_box --
 *
 * Converts a point into a XXX_BOX and returns a pointer to this
 * new BOX.
 */

XXX_BOX *point_2_box(XXX_POINT *p) {
  XXX_BOX *b = new XXX_BOX();
  b->xl = b->xh = p->x;
  b->yl = b->yh = p->y;
  return b;
}


/*
 * point_dist --
 *
 * Computes and returns the distance between two
 * points.
 */

double point_dist(XXX_POINT *p1, XXX_POINT *p2) {
  double dx, dy, dist;

  dx = (p1->x - p2->x);
  dy = (p1->y - p2->y);
  dist = dx*dx + dy*dy;
  return (sqrt(dist));
}


/*
 * point_in_box --
 *
 * Return 1 if the XXX_POINT lies within or on the boundary
 * of the given XXX_BOX; 0 otherwise.
 */

int point_in_box(XXX_POINT *p, XXX_BOX *b) {
  return (isge(p->x, b->xl) && isle(p->x, b->xh) && isge(p->y, b->yl) && isle(p->y, b->yh));
}


/*
 * Return 1 if the XXX_POINT lies strictly within the given XXX_BOX;
 * 0 otherwise.
 */

int point_iin_box(XXX_POINT *p, XXX_BOX *b) {
  return (isgt(p->x, b->xl) && islt(p->x, b->xh) && isgt(p->y, b->yl) && islt(p->y, b->yh));
}


/*
 * point_on_box --
 *
 * Return 1 if the XXX_POINT lies on the boundary of the XXX_BOX;
 * 0 otherwise.
 */

int point_on_box(XXX_POINT *p, XXX_BOX *b) {
  /*  This isn't obvious:
  return ( (isequal(p->x, b->xl) && isge(p->y, b->yl) && isle(p->y, b->yh)) ||
	   (isequal(p->x, b->xh) && isge(p->y, b->yl) && isle(p->y, b->yh)) ||
	   (isequal(p->y, b->yh) && isge(p->x, b->xl) && isle(p->x, b->xh)) ||
	   (isequal(p->y, b->yh) && isge(p->x, b->xl) && isle(p->x, b->xh)));
	   */
  /* Try this instead: */
  return ((point_in_box(p, b))&&(!(point_iin_box(p, b))));
}


/*
 * point_2_poly --
 *
 * Converts a XXX_POINT to a XXX_POLYGON and returns a
 * pointer to the new XXX_POLYGON.
 */

XXX_POLYGON *point_2_poly(XXX_POINT *pt) {
  XXX_POLYGON *p = new XXX_POLYGON();
  double *pts;
  
  p->npts = 1;
  p->pts = new double[(sizeof(double) * 2)];
  p->mbr.xh = p->mbr.xl = pt->x;
  p->mbr.yh = p->mbr.yl = pt->y;
  pts = (double *)p->pts;
  *pts++ = pt->x;
  *pts = pt->y;
  return p;
}



/*
 * point_on_poly --
 *
 * Return 1 if the point lies on the boundary of the given
 * polygon; 0 otherwise.
 */

int point_on_poly(XXX_POINT *pt, XXX_POLYGON *poly) {
  return ( (point_in_poly(pt, poly))&&(!(point_iin_poly(pt, poly))));
}


/*
 *-------------------------------------------------------------
 *
 * Line Segment Operations:
 *
 * lseg_make
 * lseg_smake
 * lseg_parallel
 * lseg_intersect
 * lseg_intercept
 *
 *-------------------------------------------------------------
 */

/*
 * lseg_smake --
 *
 * Fill in the given XXX_LSEG with the 2 XXX_POINTS.
 */

void lseg_smake(XXX_LSEG *lseg, XXX_POINT *p1, XXX_POINT *p2) {
  lseg->p[0] = *p1;
  lseg->p[1] = *p2;
  lseg->m = (p2->y - p1->y)/(p2->x - p1->x);
}

/*
 * lseg_make --
 *
 * Create a new XXX_LSEG, fill it in with the 2 XXX_POINTS,
 * and return it.
 */

XXX_LSEG *lseg_make(XXX_POINT *p1, XXX_POINT *p2) {
  XXX_LSEG *lseg = new XXX_LSEG();

  if (lseg) {
    lseg->p[0] = *p1;
    lseg->p[1] = *p2;
    lseg->m = (p2->y - p1->y)/(p2->x - p1->x);
  }
  return lseg;
}

/*
 * lseg_intersect --
 *
 * Return 1 if the two given XXX_LSEG's intersect,
 * 0 otherwise.
 */

int lseg_intersect(XXX_LSEG *l1, XXX_LSEG *l2) {
  double xp, yp;

  /*
   * If the segments are parallel:
   */
  if (isequal(l1->m, l2->m)) {
    /*
     * If they lie along the same lines:
     */
    if (iszero( (double)( l2->m * (l1->p[0].x - l2->p[0].x) - (l1->p[0].y - l2->p[0].y)))) {
      /*
       * If either endpoint of one segment lies within the segment of the other:
       * then they intersect; otherwise, no.
       */
      return ( (isge(l1->p[0].x, l2->p[0].x) && isle(l1->p[0].x, l2->p[1].x)) ||
	       (isge(l1->p[0].x, l2->p[1].x) && isle(l1->p[0].x, l2->p[0].x)) ||
	       (isge(l1->p[1].x, l2->p[0].x) && isle(l1->p[1].x, l2->p[1].x)) ||
	       (isge(l1->p[1].x, l2->p[1].x) && isle(l1->p[1].x, l2->p[0].x)));
    }
    else return 0;
  }
  /*
   * If the segments aren't parallel:
   */
  else {
    /*
     * Find the point of intersection between the two lines
     * on which the two segments lie:
     */
    xp = (((l1->m*l1->p[0].x)-(l2->m*l2->p[0].x)+l2->p[0].y-l1->p[0].y)/(l1->m-l2->m));
    yp = ((l1->m*(xp-l1->p[0].x))+l1->p[0].y);
    /*
     * If the intercept lies on both segments then the two segments
     * intersect; otherwise, no:
     */
    return ( ((isge(xp, l1->p[0].x) && isle(xp, l1->p[1].x)) ||
	      (isge(xp, l1->p[1].x) && isle(xp, l1->p[0].x))) &&
	     ((isge(xp, l2->p[0].x) && isle(xp, l2->p[1].x)) ||
	      (isge(xp, l2->p[1].x) && isle(xp, l2->p[0].x))));
  } 
}


/*
 * lseg_parallel --
 *
 * Return 1 if the two XXX_LSEG's are parallel
 * (i.e. if they equal slopes); otherwise
 * return 0.
 */

int lseg_parallel(XXX_LSEG *l1, XXX_LSEG *l2) {
  return (isequal(l1->m, l2->m));
}


/*
 * lseg_intercept --
 *
 * Assume the two XXX_LSEG's already intersect and are
 * not parallel.  Return the point intersection.
 */

XXX_POINT *lseg_intercept(XXX_LSEG *l1, XXX_LSEG *l2) {
  double xp, yp;
  XXX_POINT *p = new XXX_POINT();

  /*
   * Assume the two segments intersect and are not
   * parallel.  If they are parallel, we can not
   * specify one particular point unless the segments
   * touch only at one end.
   */

  xp = (((l1->m*l1->p[0].x)-(l2->m*l2->p[0].x)+l2->p[0].y-l1->p[0].y)/(l1->m-l2->m));
  yp = ((l1->m*(xp-l1->p[0].x))+l1->p[0].y);
  p->x = xp;
  p->y = yp;
  return p;
}

/*
 *-------------------------------------------------------------
 *
 * Polygon Operations:
 *
 * poly_make
 * poly_MBR
 * poly_2_chars
 * chars_2_poly
 * poly_free
 * poly_equals
 * poly_in_circle
 * poly_overlap
 * poly_ooverlap
 * poly_in_poly
 * poly_iin_poly
 *
 *-------------------------------------------------------------
 */

/*
 * poly_make --
 *
 * Given an x and y arrays and the number of points, create
 * a XXX_POLYGON and return a pointer to it.
 */

XXX_POLYGON *poly_make(long npts, double *xpts, double *ypts) {
  XXX_POLYGON *p = new XXX_POLYGON();

  p->npts = npts;
  p->pts = new double[(sizeof(double)*2*npts)];
  memcpy((char *)(p->pts), (char *)xpts, (sizeof(double)*npts));
  memcpy((char *)(p->pts + (sizeof(double)*npts)), (char *)xpts, (sizeof(double)*npts));
  poly_MBR(p);
  return p;
}


/*
 * poly_MBR --
 *
 * Determines the MBR of the given XXX_POLYGON and fills in its
 * XXX_BOX field.
 */

int poly_MBR(XXX_POLYGON *p) {
  int i;
  
  p->mbr.xl = p->mbr.xh = p->pts[0];
  p->mbr.yl = p->mbr.yh = p->pts[p->npts];
  for (i = 1; i < p->npts; i++) {
    if (p->pts[i] < p->mbr.xl) p->mbr.xl = p->pts[i];
    if (p->pts[i] > p->mbr.xh) p->mbr.xh = p->pts[i];
    if (p->pts[i+p->npts] < p->mbr.yl) p->mbr.yl = p->pts[i+p->npts];
    if (p->pts[i+p->npts] > p->mbr.yh) p->mbr.yh = p->pts[i+p->npts];
  }
  return GEOM_OK;
}

/* Overloaded function: */
int poly_MBR(XXX_POLYGON *p, XXX_BOX *b) {
  
  *b = p->mbr;
  return GEOM_OK;
}


/*
 * poly_2_chars --
 *
 * Reads a XXX_POLYGON into a character string.
 */

int poly_2_chars(XXX_POLYGON *p, char *s) {
  memcpy(s, (char *)p, sizeof(*p));
  memcpy(((char *)s + sizeof(*p)), (char *)p->pts, (sizeof(double) * 2 * p->npts));
  return GEOM_OK;
}

/* Overloaded function: */
int poly_2_chars(XXX_POLYGON *p, const char *s) {
  memcpy((char *)s, (char *)p, sizeof(*p));
  memcpy(((char *)s + sizeof(*p)), (char *)p->pts, (sizeof(double) * 2 * p->npts));
  return GEOM_OK;
}



/*
 * chars_2_poly --
 *
 * Reads a XXX_POLYGON from a character string.
 */

int chars_2_poly(XXX_POLYGON *p, char *s) {
  memcpy((char *) p, s, sizeof(XXX_POLYGON));
  p->pts = new double[(sizeof(double)*2*p->npts)];
  memcpy((char *) p->pts, (char *) (s + sizeof(XXX_POLYGON)), (sizeof(double) * 2 * p->npts));
  return GEOM_OK;
}

/* Overloaded function: */
int chars_2_poly(XXX_POLYGON *p, const char *s) {
  memcpy((char *) p, s, sizeof(XXX_POLYGON));
  p->pts = new double[(sizeof(double)*2*p->npts)];
  memcpy((char *) p->pts, (char *) (s + sizeof(XXX_POLYGON)), (sizeof(double) * 2 * p->npts));
  return GEOM_OK;
}


/*
 * poly_free --
 *
 * Frees a dynamically allocated XXX_POLYGON.
 */

int poly_free(XXX_POLYGON *p) {
  if (p) {
    if (p->pts) delete [] p->pts;
    delete p;
    p = NULL;
  }
  return GEOM_OK;
}


/*
 * poly_equals --
 *
 * Return 1 if the two XXX_POLYGON's are equal; 0 otherwise.
 */

int poly_equals(XXX_POLYGON *p1, XXX_POLYGON *p2) {
  long i, j;

  // Must have same # of points

  if (p1->npts != p2->npts) return 0;

  // loop through pts until 1st match found, then compare
  // all until then

  for (i = 0; i < p1->npts; i++) {
    if (isequal(p1->pts[i], p2->pts[0]) && isequal(p1->pts[i+p1->npts], p2->pts[p2->npts])) break;
  }
  if (i >= p1->npts) return 0;
  for (j = 0; j < p1->npts; j++) {
    if (!(isequal(p1->pts[i], p2->pts[j]) && isequal(p1->pts[i+p1->npts], p2->pts[j+p2->npts]))) return 0;
    i++;
    if (i >= p1->npts) i = 0;
  }
  return 1;
}


/*
 * poly_in_circle --
 *
 * Return 1 if the XXX_POLYGON p lies within XXX_CIRCLE c;
 * 0 otherwise.
 */

int poly_in_circle(XXX_POLYGON *p, XXX_CIRCLE *c) {
    double *xpts, *ypts;
    XXX_POINT pt;
    int i;

    xpts = (double *)p->pts;
    ypts = (double *)(p->pts + p->npts);
    for (i=0; i<p->npts; i++) {
        pt.x = *xpts++;
        pt.y = *ypts++;
        if (isgt((point_dist(&pt, &c->center)), c->radius))
           return 0;
    }
    return 1;
}


/*
 * poly_iin_circle --
 *
 * Return 1 if the XXX_POLYGON p lies strictly within XXX_CIRCLE c;
 * 0 otherwise.
 */

int poly_iin_circle(XXX_POLYGON *p, XXX_CIRCLE *c) {
    double *xpts, *ypts;
    XXX_POINT pt;
    int i;

    xpts = (double *)p->pts;
    ypts = (double *)(p->pts + p->npts);
    for (i=0; i<p->npts; i++) {
        pt.x = *xpts++;
        pt.y = *ypts++;
        if (isge((point_dist(&pt, &c->center)), c->radius))
           return 0;
    }
    return 1;
}


/*
 * poly_overlap --
 *
 * Return 1 if the two XXX_POLYGON's overlap; 0 otherwise.
 * This is a relaxed overlap check;  i.e. the two
 * XXX_POLYGON's need only touch or share a boundary point.
 */

int poly_overlap(XXX_POLYGON *p1, XXX_POLYGON *p2) 
{
  /*
    XXX_POLYGON A is deemed to CLIP XXX_POLYGON B iff
    
    - They share a rectangle extension.
    
    - One point from either is inside the other or
    
    - Line segments intersect.  This is "loose" overlapping
    b/c it allows for polygons to just "touch" instead of
    actually overlap.
    
    */
  int i, j, a, b;
  double *ixp, *iyp, *oxp, *oyp;
  XXX_POLYGON * pOuter, * pInner;
  XXX_POINT *pIntercept;
  XXX_LSEG seg1, seg2;
  XXX_POINT pCons1, pCons2;
  
  if (box_overlap(&p1->mbr, &p2->mbr)) {     
    //If candidate (exclude all 
    // obvious ones.)          
    if (p1->npts > p2->npts) {  
      // Do whichever is the simpler
      // as the inner loop.         
      pOuter=p1;
      pInner=p2;
    } 
    else {
      pOuter=p2;
      pInner=p1;
    }
    // Standard way of getting the pairs of points given how we have stored
    // the polygons in structures. No 'module' exists, of functions to get the
    // 'n'th point_in_polygon - which would provide a standard interface.

    oxp = (double *) pOuter->pts;          
    oyp = (double *) (pOuter->pts + pOuter->npts);
    ixp = (double *) pInner->pts;
    iyp = (double *) (pInner->pts + pInner->npts);
    
    // first check to see if points of inner are in outer polygon
    
    for (i=0;i <= pInner->npts -1; i++) {
      pCons1.x=ixp[i];
      pCons1.y=iyp[i];
      
      if (point_in_poly(&pCons1, pOuter)) {
	return 1;
      }
    }
    
    // Then check if points of outer are in inner polygon
    
    for (i=0;i<=pOuter->npts-1;i++) {
      pCons1.x=oxp[i];  
      pCons1.y=oyp[i];
      
      if (point_in_poly(&pCons1, pInner)) {
	return 1;
      }
    }
    
    // Then check to see if any of the sides intersect.
    
    for (i=0, j=pOuter->npts - 1; i <= pOuter->npts - 1; j=i++) {
      // Make the points
      pCons2.x=oxp[i];  
      pCons2.y=oyp[i];
      
      pCons1.x=oxp[j];
      pCons1.y=oyp[j];
      
      // Make the line segment from the pair of points.
     
      lseg_smake(&seg1, &pCons1, &pCons2);
      
      for (a=0, b=pInner->npts - 1; a <= pInner->npts - 1; b=a++) {
        pCons2.x=ixp[a];
        pCons2.y=iyp[a];
	
        pCons1.x=ixp[b];
        pCons1.y=iyp[b];
	
	// Make the line segment 
        lseg_smake(&seg2, &pCons2, &pCons1);
	
	// If the inner segment and the outer intersect and
	// aren't parallel.

	if ((!(lseg_parallel(&seg1, &seg2))) && 
	    (lseg_intersect(&seg1, &seg2))) {
          pIntercept=lseg_intercept(&seg1, &seg2);
	  
	  // If intercept is at the extremes of both lsegs then the
	  // intercept is at a point.
	  
          if (!((point_equals(pIntercept,seg1.p[0])) || 
		(point_equals(pIntercept,seg1.p[1]))) && 
              ((point_equals(pIntercept,seg2.p[0])) ||  
               (point_equals(pIntercept,seg2.p[1])))) {
	    if(pIntercept!=NULL) delete pIntercept;
            return 1;
          }
          if(pIntercept!=NULL) delete pIntercept;
        }
      }
    }
  }
  return 0;
}


/*
 * poly_ooverlap --
 *
 * Return 1 if the two XXX_POLYGON's overlap; 0 otherwise.
 * This is a strict overlap check;  i.e. the two
 * XXX_POLYGON's must intersect at more than just boundary
 * points.
 */

int poly_ooverlap(XXX_POLYGON *p1, XXX_POLYGON *p2) 
{
  /*
    XXX_POLYGON A is deemed to CLIP XXX_POLYGON B iff

    - They share a rectangle extension.
    
    - One point from either is inside the other or
    
    - two non-parallel line segments intersect at some
    point which isn't a boundary point for both
    polygons.
    */
  int i, j, a, b;
  double *ixp, *iyp, *oxp, *oyp;
  XXX_POLYGON * pOuter, * pInner;
  XXX_POINT *pIntercept;
  XXX_LSEG seg1, seg2;
  XXX_POINT pCons1, pCons2;
  
  if (box_overlap(&p1->mbr, &p2->mbr)) {
    // If candidate (exclude all  
    // obvious ones.)             
    if (p1->npts > p2->npts) {                    
      // Do whichever is the simpler
      // as the inner loop.         
      pOuter=p1;
      pInner=p2;
    } 
    else {
      pOuter=p2;
      pInner=p1;
    }

    // Standard way of getting the pairs of points given how we have stored 
    // the polygons in structures. No 'module' exists, of functions to get the
    // 'n'th point_in_polygon - which would provide a standard interface.

    oxp = (double *) pOuter->pts;          
    oyp = (double *) (pOuter->pts + pOuter->npts);
    ixp = (double *) pInner->pts;
    iyp = (double *) (pInner->pts + pInner->npts);
    
    // first check to see if points of inner are in outer polygon
    
    for (i=0;i <= pInner->npts -1; i++) {
      pCons1.x=ixp[i];
      pCons1.y=iyp[i];
      
      if (point_iin_poly(&pCons1, pOuter)) {
	return 1;
      }
    }
    
    // Then check if points of outer are in inner polygon
    
    for (i=0;i<=pOuter->npts-1;i++) {
      pCons1.x=oxp[i];  
      pCons1.y=oyp[i];
      
      if (point_iin_poly(&pCons1, pInner)) {
	return 1;
      }
    }
    
    // Then check to see if any of the sides intersect.


    for (i=0, j=pOuter->npts - 1; i <= pOuter->npts - 1; j=i++) {
      // Make the points
      pCons2.x=oxp[i]; 
      pCons2.y=oyp[i];
      
      pCons1.x=oxp[j];
      pCons1.y=oyp[j];
      
      // Make the line segment from the pair of points.
      
      lseg_smake(&seg1, &pCons1, &pCons2);
      
      for (a=0, b=pInner->npts - 1; a <= pInner->npts - 1; b=a++) {
        pCons2.x=ixp[a];
        pCons2.y=iyp[a];
	
        pCons1.x=ixp[b];
        pCons1.y=iyp[b];
	
	/* Make the line segment */
        lseg_smake(&seg2, &pCons2, &pCons1);
	
	/* 
	 * If the inner segment and
	 * the outer intersect and
	 * aren't parallel.
	 */
	if ((!(lseg_parallel(&seg1, &seg2))) && 
	    (lseg_intersect(&seg1, &seg2))) {
          pIntercept=lseg_intercept(&seg1, &seg2);
	  
	  /*
	   * If intercept is at the extremes of both lsegs then the
	   * intercept is at a point.
	   */
          if (!((point_equals(pIntercept,seg1.p[0])) || 
		(point_equals(pIntercept,seg1.p[1]))) && 
              ((point_equals(pIntercept,seg2.p[0])) ||  
               (point_equals(pIntercept,seg2.p[1])))) {
	    if(pIntercept!=NULL) delete pIntercept;
            return 1;
          }
          if(pIntercept!=NULL) delete pIntercept;
        }
     }
    }
  }
  return 0;
}


/*
 * poly_in_poly --
 *
 * Return 1 if the if XXX_POLYGON p1 lies within
 * XXX_POLYGON p2; otherwise return 0.  This is a relaxed
 * check.  p1 doesn't need to be strictly within p2
 * (i.e. it can share boundary points)
 */

int poly_in_poly(XXX_POLYGON *p1, XXX_POLYGON *p2) {
  XXX_POINT p;
  int k;

  for (k = 0; k < p1->npts; k++) {
    p.x = p1->pts[k];
    p.y = p1->pts[k+p1->npts];
    if ((point_in_poly(&p, p2))) {
      continue;
    }
    else {
      return 0;
    }
  }
  return 1;
}


/*
 * poly_iin_poly --
 *
 * Return 1 if the if XXX_POLYGON p1 lies strictly within
 * XXX_POLYGON p2; otherwise return 0.  This is a strict
 * check.  p1 can share no boundary points with p2.
 */

int poly_iin_poly(XXX_POLYGON *p1, XXX_POLYGON *p2) {
  XXX_POINT p;
  int k;

  for (k = 0; k < p1->npts; k++) {
    p.x = p1->pts[k];
    p.y = p1->pts[k+p1->npts];
    if ((point_iin_poly(&p, p2))) {
      continue;
    }
    else {
      return 0;
    }
  }
  return 1;
}

// poly_size returns the approximate area of a polygon by returning
// the area of the XXX_POLYGON's bounding box.

double poly_size(XXX_POLYGON *p)
{
  XXX_BOX b = p->mbr;

  return ((b.xh-b.xl) * (b.yh-b.yl));
}

  
