/*
 *-------------------------------------------------------------------------
 *
 * polygon.h
 *
 * Polygon ADT
 *-------------------------------------------------------------------------
 */

#ifndef XXX_POLYGON_H
#define XXX_POLYGON_H

#include "geometry.h"
#include "genadt.h"
#include "shoresm.h"
#include "adtfunc.h"

XXX_EXTERN_CLASS_DECL(XxxBoundingBox);

/**
 * XxxPolygonADT:  
 * Basic ADT for handling any generic polygon data.  Polygons are
 * stored as fixed size objects within a record.  Consequently they
 * hold an object ID of the actual polygon data stored in SHORE.  
 * The polygon data is a file containing a binary stream of doubles,
 * first the x-coords, then the y-coords.  The polygon ADT currently
 * supports the following methods: disjoint(), meets(), covers(),
 * contains(), inside(), covered_by(), overlaps(), equals(), size(),
 * and bounding_box().  The first 6 of these methods come from the
 * 9-point intersection model of spatial objects.
 *
 * @see geometry.h box.h point.h
 * @author Dave Koster dkoster@cs.cornell.edu
 * @version 6/17/97
 */
class XxxPolygonADT : public XxxLargeFixedADT {
public:
  // friends, so they can call some of the polygon methods
  friend class XxxPointADT;
  friend class XxxBoxADT;
  friend class XxxRasterADT;
  friend class XxxGISModule;

  /**
   * PolygonMetaInfo:
   * Meta Info for polygons.  Currently it only contains info on 
   * the average polygon size.
   *
   * @see genadt.h types.
   */
  class PolygonMetaInfo : public XxxADTMetaInfo {
  public:
    /// Average polygon size
    int AvgPolygonSize;
    /**
     * Constructor for PolygonMetaInfo.
     */
    PolygonMetaInfo() {
      AvgPolygonSize = 16000;
    }
    /**
     * Replication function for this polygon meta info
     * @param NewMI Reference to the newly replicated meta info.
     * @return The status code.
     */
    virtual XxxErrCode Replicate(XxxADTMetaInfo*& NewMI) const {
      PolygonMetaInfo* NewMetaInfo = new PolygonMetaInfo(*this);
      NewMI = NewMetaInfo;
      return XXX_OK;
    }
    /**
     * Destructor
     */
    virtual ~PolygonMetaInfo() {}
    /**
     * Print
     * @param OStr Output stream.
     * @return Nothing.
     */
    virtual void Print(ostream& OStr) const
      { OStr << "(" << AvgPolygonSize << ")";}
    virtual XxxBool IsCompatible(const XxxADTMetaInfo* OtherMI) const 
      { return XXX_TRUE; }
    /**
     * Returns the number of bytes necessary to hold the serialized version
     * of this meta-info object.
     * @param Size (out) The required number of bytes
     * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
     */
    virtual XxxErrCode SerializeCopySize(int &Size) const {
      Size = sizeof(AvgPolygonSize); return XXX_OK;
    }
    /**
     * Creates a binary version of this meta-info object.
     * @param Data An array of bytes of at least the size specified by the
     *             SerializeCopySize() function.
     * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
     */
    virtual XxxErrCode SerializeCopy(char *Data) const {
      memcpy(Data, (char *)(&AvgPolygonSize), sizeof(AvgPolygonSize));
      return XXX_OK;
    }
  }; // PolygonMetaInfo
  friend class PolygonMetaInfo;

  /**
   * InMemPolygon:
   * In Memory polygon information.  Used by XxxPolygonADT whenever
   * it needs to retrieve a polygon's data from SHORE and read it
   * into a POLYGON.
   *
   * @see geometry.h genadt.h
   * @author Dave Koster dkoster@cs.cornell.edu
   * @version 6/17/97
   */
  class InMemPolygon {
  public:
    /**
     * The actual polygon represention, containing the array
     * points, number of points, and bounding box.
     */
    XXX_POLYGON polygon;
    /**
     * Constructor for InMemPolygon.
     * @param PS Polygon info (number of points, etc.).
     */
    InMemPolygon(XXX_POLYGON PS) : polygon(PS) {
      polygon.pts = new double[polygon.npts * 2];
    }
    /**
     * Destructor for InMemPolygon.
     * Deletes the array of points in its XXX_POLYGON.
     */
    ~InMemPolygon() {
      delete [] polygon.pts;
      polygon.pts = NULL;
    }
  }; // InMemPolygon

  /**
   * PolygonStruct:
   * Contains the ADT information on the polygon.  This includes more
   * than just the number of points and bounding box; namely, the
   * Shore ID of the polygon's data and also a pointer to the polygon's
   * in-memory location, if it exists.
   *
   * @see genadt.h
   * @author Dave Koster dkoster@cs.cornell.edu
   * @version 6/17/97
   */
  class PolygonStruct : public XxxADTValue {
  public:
    /// Basic polygon info.
    XXX_POLYGON polygon;
    /// Shore ID of polygon data.
    XxxShoreLogicalOID PolygonId;
    /// In-memory polygon.
    InMemPolygon *MemPolygon;

    XxxErrCode CopyHToN(const PolygonStruct* AS,
			const PolygonMetaInfo* MetaInfo);
    XxxErrCode CopyNToH(const PolygonStruct* AS,
			const PolygonMetaInfo* MetaInfo);
  };
  
  /**
   * PlanInfo:
   * Plan structure for methods on polygons.  Contains
   * all the different geographic intersection methods as well.
   *
   * @see genadt.h
   * @author Dave Koster dkoster@cs.cornell.edu
   * @version 6/17/97
   */
  class PlanInfo : public XxxFuncPlanInfo {
  protected:
    /// friends that need to use the getMemPolygon method.
    friend class XxxPolygonADT;
    friend class XxxPointADT;
    friend class XxxBoxADT;
    
    /**
     * Disjoint method between this polygon and other objects.  These other
     * objects can be either polygons or boxes.
     *
     * @param Env Function environment.
     * @param ReturnValue Where to write the answer.
     * @return The status code.
     */
    XxxErrCode execDisjoint(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;

    /**
     * Meets method between this polygon and other objects.  These other
     * objects can be either polygons or boxes.  Meets simply checks
     * if any of the two object's edges lie tangent to each other.
     *
     * @param Env Function environment.
     * @param ReturnValue Where to write the answer.
     * @return The status code.
     */
    XxxErrCode execMeets(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;

    /**
     * Covers method between this polygon and other objects -- polygons,
     * boxes, or points.  This simply checks if this polygon "covers"
     * another object.
     *
     * @param Env Function environment.
     * @param ReturnValue Where to write the answer.
     * @return The status code.
     */
    XxxErrCode execCovers(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;

    /**
     * Contains method between this polygon and other objects -- polygons,
     * boxes, or points.  This is the strict version of Covers -- i.e.
     * the inner object must lie completely within this polygon.
     *
     * @param Env Function environment.
     * @param ReturnValue Where to write the answer.
     * @return The status code.
     */
    XxxErrCode execContains(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;

    /**
     * Overlaps method between this polygon and other objects -- polygons
     * or boxes.  This simply checks if this polygon overlaps any part of
     * the argument object.
     *
     * @param Env Function environment.
     * @param ReturnValue Where to write the answer.
     * @return The status code.
     */
    XxxErrCode execOverlaps(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;

    /**
     * Inside method between this polygon and other objects -- polygons,
     * circles, or boxes.  This checks if this polygon lies completely
     * inside the outer object.  This is the stricter form of Covered-by.
     *
     * @param Env Function environment.
     * @param ReturnValue Where to write the answer.
     * @return The status code.
     */
    XxxErrCode execInside(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;

    /**
     * Covered-by method between this polygon and other objects -- polygons,
     * circles, or boxes.  This checks if this polygon lies inside the 
     * outer object.
     *
     * @param Env Function environment.
     * @param ReturnValue Where to write the answer.
     * @return The status code.
     */
    XxxErrCode execCoveredBy(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;

    /**
     * Equals method between this polygon and other objects -- obviously
     * another polygon or box.
     *
     * @param Env Function environment.
     * @param ReturnValue Where to write the answer.
     * @return The status code.
     */
    XxxErrCode execEquals(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;

    /**
     * Bounding-box method of this polygon.  Simply writes the bounding
     * box of this polygon into the ReturnValue.
     *
     * @param Env Function environment.
     * @param ReturnValue Where to write the answer.
     * @return The status code.
     */
    XxxErrCode execBoundingBox(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;

    /**
     * Size method of this polygon.  Currently, this incorrectly returns
     * the area of the bounding box.
     *
     * @param Env Function environment.
     * @param ReturnValue Where to write the answer.
     * @return The status code.
     */
    XxxErrCode execSize(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;

    /**
     * Fetches the polygon data into memory.
     */
    static XxxErrCode getMemPolygon(PolygonStruct *PS,
			     InMemPolygon*& Polygon);
    /**
     * Writes the in-memory polygon data into SHORE.
     */
    static XxxErrCode putMemPolygon(InMemPolygon* Polygon,
				    PolygonStruct *PS,
				    XxxShoreVolumeId &VolId,
				    XxxShoreObjectId &FileId);
    /**
     * Reclaims an in-memory polygon and its data.
     */
    static XxxErrCode reclaimMemPolygon(PolygonStruct *PS);
    
  public:

    /**
     * Constructor for PlanInfo.
     * @param FuncParse Parse info for this func expression.
     * @param ArgCount Number of arguments to this function.
     */
    PlanInfo(const XxxFuncParseInfo* FuncParse, int ArgCount)
    : XxxFuncPlanInfo(FuncParse, ArgCount)  {}
    /**
     * Destructor for PlanInfo.
     */
    virtual ~PlanInfo() {}
  }; // PlanInfo
  friend class PlanInfo;
  
protected:
  // local methods

  /**
   * Parses the meta info.  Currently this can only consist
   * of an average polygon size.
   */
  XxxErrCode parseMetaInfo(const char* MetaInfoString,
                           MetaInfoHandle& MIHandle) const;
  
  /**
   * Reads in the meta info for polygons.
   *
   * @param MIHandle Handle for storing the meta info.
   * @param MetaInfo Reference to the meta info.
   * @return The status code.
   */
  virtual XxxErrCode makeMetaInfo(const MetaInfoHandle& MIHandle,
                                  XxxADTMetaInfo*& MI) const;
  
  //// Polygon temp file name.
  static const char* const POLYGONFILENAME;
  
  /**
   * Retrieves a SHORE logical object ID from a char buffer.
   *
   * @param Object Char buffer.
   * @param VolId Reference to volume ID.
   * @param LocalId Reference to object ID.
   * @return The status code.
   */
  virtual XxxErrCode getOid(const XxxADTValue *Object,
			    XxxShoreVolumeId& VolId,
			    XxxShoreObjectId& LocalId) const {
    XxxShoreLogicalOID* PolygonId = &(((PolygonStruct*)Object)->PolygonId);
    VolId = PolygonId->VolId;
    LocalId = PolygonId->LocalId;
    return XXX_OK;
  }

  /**
   * Does nothing for now.  The actual display goes on in the
   * WriteText method.
   */
  virtual XxxErrCode showDataFile(const char* FileName,
                                  ostream& OutStream,
                                  const XxxADTValue* Object,
                                  const XxxADTMetaInfo* MetaInfo) const {
    // this does not get called, because Polygon overrides WriteText()
    return XXX_NOT_OK;
  }

  // dummy wrappers
  static XxxErrCode getMemPolygon(PolygonStruct *PS,
    InMemPolygon*& Polygon) 
  { return PlanInfo::getMemPolygon(PS, Polygon);}
  static XxxErrCode putMemPolygon(InMemPolygon* Polygon,
    PolygonStruct *PS,
    XxxShoreVolumeId &VolId,
    XxxShoreObjectId &FileId) { 
    return PlanInfo::putMemPolygon(Polygon, PS,
                                   VolId, FileId);
  }
  static XxxErrCode reclaimMemPolygon(PolygonStruct *PS) 
  { return PlanInfo::reclaimMemPolygon(PS); }
  
public:
  
  /// Polygon's type ID.
  static XxxBasicType PolygonTypeId;
  
  /**
   * Constructor for XxxPolygonADT.
   *
   * @param Id Type ID for polygon.
   */
  XxxPolygonADT(XxxBasicType Id);
  
  /**
   * Destructor for XxxPolygonADT.
   */
  virtual ~XxxPolygonADT() {};
  
  /**
   * Display function for the txtclient protocol.
   *
   * @param OutStream Output stream to write to.
   * @param Object Buffer containing the PolygonStruct.
   * @param MetaInfo Meta info for polygons.
   * @return The status code.
   */
  virtual XxxErrCode WriteText(ostream& OutStream, const XxxADTValue* Object, 
			       const XxxADTMetaInfo *MetaInfo) const;
  
  /**
   * Input function for loading polygons into tables.
   *
   * @param is Input stream from which to read the PolygonStruct.
   * @param ObjectRef Buffer in which to store the PolygonStruct.
   * @param MetaInfo Meta info for polygons.
   * @return The status code.
   */
  virtual XxxErrCode ReadText(istream& is, XxxADTValue* ObjectRef, 
			      const XxxADTMetaInfo* MetaInfo) const;
  
  // Methods related to using indexes on box types.

  /**
   * Retrieves the bounding box of this spatial data type.  Used by
   * the indexing methods.
   */
  virtual XxxBool GetBoundingBox(const  XxxADTValue *&Object,
				 XxxBoundingBox& Box);

  /**
   * Retrieves the compare method given a function name
   * and a array of arguments for that function.
   * This is used in index scans and is index independent.
   */
  virtual XxxErrCode GetCompOp(const char* FuncName,
			       XxxCompareOp& CompOp,
			       XxxBool& Success) const;

  virtual XxxErrCode CopyHToN(const XxxADTValue* ObjectRef, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxErrCode CopyNToH(const XxxADTValue* ObjectRef, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
}; // XxxPolygonADT




#endif /* XXX_POLYGON]_H */

