#include "gisadt.h"
#include "optrules.h"
#include "memmap.h"

const char* const XxxRasterADT::AVHRR = "avhrr";
const char* const XxxRasterADT::JPEG = "jpeg";

/*****
Here's the info on Raster EADT.

 Each raster schema has :
    *) a format (avhrr, jpeg, etc)
    *) compression information 

 The raster meta-info includes:
    *) avg. raster size
    *) format (just avhrr)
    
 Raster input format: <file-name>
 Raster schema format: raster {<format>, compression info, avg size}

****/

// define the static variables
XxxBasicType  XxxRasterADT::RasterTypeId = XXX_UNKNOWN;

//
// Raster Type
//

XxxRasterADT::XxxRasterADT(XxxBasicType TypeId)
: XxxLargeFixedADT(TypeId,          // id of type
		   "raster",         // type name
		   sizeof(RasterStruct))
{
 RasterTypeId = TypeId;

 NumMethods = 2;

 MethArray = new FuncMethodInfo*[NumMethods];
  
 MethArray[0] = new FuncMethodInfo("clip", TypeId, 0, 0,
				      (ExecuteFunc) &(PlanInfo::execClip),
				      1, XxxBoxADT::BoxTypeId);
 MethArray[1] = new FuncMethodInfo("deres", TypeId, 0, 0,
				      (ExecuteFunc) &(PlanInfo::execDeres),
				      1, XXX_INT);

 XxxFuncMethodInfo::ArgInfo EmptySignature(0);
 NumAggrMethods = 2;
 AggrMethArray = new AggrMethodInfo*[NumAggrMethods];
 AggrMethArray[0] = new AggrMethodInfo("ravg", TypeId, EmptySignature,
				       0, 0, 
				       (AggrInitFunc)(PlanInfo::ravgInit),
				       (AggrAccumFunc)(PlanInfo::ravgAccum),
				       (AggrTermFunc)(PlanInfo::ravgTerm), 0);
 AggrMethArray[1] = new AggrMethodInfo("rany", TypeId, EmptySignature,
                                       0, 0, 
				       (AggrInitFunc)(PlanInfo::ranyInit),
				       (AggrAccumFunc)(PlanInfo::ranyAccum),
				       (AggrTermFunc)(PlanInfo::ranyTerm), 0);
 
 RuleEngine = new XxxOptRuleEngine;

 // the deres and clip can both be moved into the aggregate ravg
 char Funcs[][XXX_FUNCNAME_MAX] = {"clip", "deres"};
 char Aggrs[][XXX_FUNCNAME_MAX] = {"ravg"};
 XxxRule *AggrFuncRule = new XxxMoveFuncIntoAggrRule(RasterTypeId, Aggrs, 1,
						     Funcs, 2);
 RuleEngine->AddRule(AggrFuncRule);

 // the deres and clip can both be moved out of the aggregate any
 char Aggrs2[][XXX_FUNCNAME_MAX] = {"rany"};
 XxxRule *FuncAggrRule = new XxxMoveFuncOutOfAggrRule(RasterTypeId, Aggrs2, 1,
						      Funcs, 2);
 RuleEngine->AddRule(FuncAggrRule);



}

//--------------------------------------------------------------------------------
XxxRasterADT::~XxxRasterADT()
{
}

//--------------------------------------------------------------------------------
XxxErrCode 
XxxRasterADT::parseMetaInfo(const char* MetaInfoString,
			    MetaInfoHandle& MIHandle) const
{
 MIHandle.Size = sizeof(RasterMetaInfo);
 ASSERT((MIHandle.Size <= XXX_MIHNDL_MAX));

 istrstream InStr((char *)MetaInfoString);
 char Format[10], Punct;
 int AvgSizeKB;
 InStr >> Format >> Punct >> AvgSizeKB;
 if (InStr.bad()) {
   XXX_NOTE_ERROR(XXX_BADADT, "Bad raster meta info");
   return XXX_NOT_OK;
  }
 RasterMetaInfo* RMI = (RasterMetaInfo*) (MIHandle.Data);
 RMI->AvgRasterSize = AvgSizeKB;
 if (!strcasecmp(Format, AVHRR))
   RMI->Format = RASTER_AVHRR;
 else  if (!strcasecmp(Format, JPEG))
   RMI->Format = RASTER_JPEG;
 else {
   char buf[100];
   sprintf(buf, "Unknown raster format: %s", Format);
   XXX_NOTE_ERROR(XXX_BADADT, buf);
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

//--------------------------------------------------------------------------------
XxxErrCode XxxRasterADT::makeMetaInfo(const MetaInfoHandle& MIHandle,
				      XxxADTMetaInfo*& MI) const
{
 RasterMetaInfo* HandleInfo = (RasterMetaInfo*)(MIHandle.Data);
 MI = new RasterMetaInfo(*HandleInfo);
 return XXX_OK;
}

//--------------------------------------------------------------------------------
// called by WriteText
XxxErrCode XxxRasterADT::showDataFile(const char* FileName,
				      ostream& OutStream, 
				      const XxxADTValue* Object,
				      const XxxADTMetaInfo* MetaInfo) const
{
#ifdef XXX_WIN32
  XXX_NOTE_ERROR(XXX_BADADT, "Cannot display jpeg raster");
  return XXX_NOT_OK;
#else
 RasterStruct *RS = (RasterStruct *)Object;
 // and xv it -- I need it in jpeg before I can do this
 char Command[5 + XXX_DIRNAME_MAX];
 char *TmpJpeg = tempnam(XXX_THISENV().TmpDataDir, "jpeg");
 
 if(avhrrToJpeg(FileName, RS->freq, TmpJpeg, RS->samples, RS->lines)){
   XXX_NOTE_ERROR(XXX_BADADT, "Error converting raster");
   return XXX_NOT_OK;
  }

 sprintf(Command, "xv %s &", TmpJpeg);
 if (system(Command)) {
   XXX_NOTE_ERROR(XXX_BADADT, " Error displaying raster");
   return XXX_NOT_OK;
  }
 return XXX_OK;
#endif
}

//--------------------------------------------------------------------------------
XxxErrCode 
XxxRasterADT::ReadText(istream& InStream, XxxADTValue* ObjectRef,
		       const XxxADTMetaInfo* MetaInfo) const
{
 XXX_SM_CHECK;
 XxxErrCode rc = XXX_OK;

 char BigRasterName[XXX_DIRNAME_MAX] = "\0";
 char WhichFreq[2] = "\0";
 char NumSamples[10] = "\0";
 char NumLines[10] = "\0";
 char xl[10] = "\0";
 char yl[10] = "\0";
 char xh[10] = "\0";
 char yh[10] = "\0";

 InStream >> WhichFreq >> NumLines >> NumSamples >> xl >> yl >> xh >> yh >> BigRasterName;

 if (!InStream.good()) {
   XXX_NOTE_ERROR(XXX_BADADT, "Cannot read raster data");
   return XXX_NOT_OK;
 }
 RasterStruct* RS = (RasterStruct*)ObjectRef;
 RS->freq = atoi(WhichFreq);
 RS->samples = atoi(NumSamples);
 RS->lines = atoi(NumLines);
 RS->geoArea.xl = atoi(xl);
 RS->geoArea.yl = atoi(yl);
 RS->geoArea.xh = atoi(xh);
 RS->geoArea.yh = atoi(yh);
 RS->resolution = 1; // All rasters are read in at full resolution
 RS->MemRaster = NULL;

 XxxShoreObjectId RasterOID;
 XxxLargeADTEnv *RasterEnv = (XxxLargeADTEnv*) XXX_THISADTENV(TypeId);
 // note: the constructor mmaps the file, while
 // the destructor unmaps the file
 XxxMemoryMap RasterMap(BigRasterName);
 void *MMapAddress = NULL;
 XXX_CALL_RET(RasterMap.GetHandle(MMapAddress),
	      XXX_BADFILE, "Error in mmap");

#ifdef XXX_LITTLEENDIAN
 int NumReadings = RS->lines * RS->samples;
 unsigned short *NewData = new unsigned short[NumReadings];
 unsigned short *OldData = (unsigned short *)MMapAddress;
 for (int i = 0; i < NumReadings; i++) {
   short TempShort;
   XxxHostShort(&TempShort, OldData[i]);
   *(NewData+i) = TempShort;
 }
  MMapAddress = (void *)NewData;
#endif
 
 // insert a "record" of this object
 XXX_CALL_RET (XxxShoreSM->InsertRecord(RasterEnv->StorageVolume, 
					RasterEnv->StorageFile,
					MMapAddress,
					RasterMap.Size(),
					RasterOID),
	       XXX_BADFILE, "Unable to store in SHORE");

 RS->RasterId.VolId = RasterEnv->StorageVolume;
 RS->RasterId.LocalId = RasterOID;
 
 RasterMetaInfo* RMI = (RasterMetaInfo*) (MetaInfo);

#ifdef XXX_LITTLEENDIAN
 if (NewData) delete [] NewData;
#endif

 return XXX_OK;
}

XxxErrCode XxxRasterADT::CopyHToN(const XxxADTValue* Object, 
				  XxxADTValue *ResultObject,
				  const XxxADTMetaInfo* MetaInfo) const
{
  RasterStruct* RS = (RasterStruct*)Object;
  RasterStruct* RRS = (RasterStruct*)ResultObject;
  ASSERT ((RRS && RS));
  return RRS->CopyHToN(RS, (RasterMetaInfo*)MetaInfo);
}

XxxErrCode XxxRasterADT::CopyNToH(const XxxADTValue* Object, 
				  XxxADTValue *ResultObject,
				  const XxxADTMetaInfo* MetaInfo) const
{
  RasterStruct* RS = (RasterStruct*)Object;
  RasterStruct* RRS = (RasterStruct*)ResultObject;
  ASSERT ((RRS && RS));
  return RRS->CopyNToH(RS, (RasterMetaInfo*)MetaInfo);
}

XxxErrCode XxxRasterADT::RasterStruct::
CopyHToN(const RasterStruct* RS,
	 const RasterMetaInfo* MetaInfo) 
{
  XxxNetworkInt(&(freq), RS->freq);
  XxxNetworkInt(&(samples), RS->samples);
  XxxNetworkInt(&(lines), RS->lines);
  XxxNetworkValue(&(geoArea), RS->geoArea);
  memcpy((char *)&(RasterId), (char *)&(RS->RasterId),
	 sizeof(XxxShoreLogicalOID));
  MemRaster = 0;
  return XXX_OK;
}
XxxErrCode XxxRasterADT::RasterStruct::
CopyNToH(const RasterStruct* RS,
	 const RasterMetaInfo* MetaInfo) 
{
  XxxHostInt(&(freq), RS->freq);
  XxxHostInt(&(samples), RS->samples);
  XxxHostInt(&(lines), RS->lines);
  XxxHostValue(&(geoArea), RS->geoArea);
  memcpy((char *)&(RasterId), (char *)&(RS->RasterId),
	 sizeof(XxxShoreLogicalOID));
  MemRaster = 0;
  return XXX_OK;
}
