/*
 *-------------------------------------------------------------------------
 *
 * raster.h
 * Raster ADT
 *-------------------------------------------------------------------------
 */

#ifndef XXX_RASTER_H
#define XXX_RASTER_H
#include "genadt.h"
#include "shoresm.h"
#include "geometry.h"
#include "adtfunc.h"
#include "aggrexpr.h"

#define RASTER_DUMP_DFLT "RASTER_DUMP"
#define RASTER_DUMP_DIR "RASTER_DIR"

/**
 * Raster Type
 */
class XxxRasterADT : public XxxLargeFixedADT {
  // I can put the avhrr2jpeg stuff in here later
  //  friend class JpegHandler;
 protected:
  // local typedefs
  enum RasterType {RASTER_AVHRR, RASTER_JPEG }; 

  //// Meta-information for raster ADT
  class RasterMetaInfo : public XxxADTMetaInfo {
   public:
    RasterType Format;
    int AvgRasterSize; // size of compressed image, in bytes
    RasterMetaInfo ();
    virtual XxxErrCode Replicate(XxxADTMetaInfo*& NewMI) const {
      RasterMetaInfo* NewMetaInfo = new RasterMetaInfo(*this);
      NewMI = NewMetaInfo;
      return XXX_OK;
     }
    virtual void Print(ostream& OStr) const 
      { OStr << "(" << Format << ", " << AvgRasterSize << ")"; }
    virtual XxxBool IsCompatible(const XxxADTMetaInfo* OtherMI) const {
      if (Format == ((RasterMetaInfo*)OtherMI)->Format)
	return XXX_TRUE;
      else return XXX_FALSE;
     }
    /**
     * Returns the number of bytes necessary to hold the serialized version
     * of this meta-info object.
     * @param Size (out) The required number of bytes
     * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
     */
    virtual XxxErrCode SerializeCopySize(int &Size) const {
      Size = sizeof(int) + sizeof(AvgRasterSize); return XXX_OK;
    }
    /**
     * Creates a binary version of this meta-info object.
     * @param Data An array of bytes of at least the size specified by the
     *             SerializeCopySize() function.
     * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
     */
    virtual XxxErrCode SerializeCopy(char *Data) const {
      int intFormat = (int)Format;
      memcpy(Data, (char *)(&intFormat), sizeof(intFormat));
      memcpy((char *)(Data + sizeof(intFormat)), (char *)(&AvgRasterSize), 
	     sizeof(AvgRasterSize));
      return XXX_OK;
    }
  };
  friend class RasterMetaInfo;

  //// In-memory representation of a raster
  class InMemRaster {
    public :
    unsigned short* RasData;
    unsigned int Lines;
    unsigned int Samples;
    int Freq;          // frequency of this raster
    XXX_BOX GeoArea; 
    int resolution;
    InMemRaster(unsigned int NumLines, unsigned int NumSamples,
		int Frequency, const XXX_BOX& GeoBox, int res=1) :
    Lines(NumLines), Samples(NumSamples), Freq(Frequency), GeoArea(GeoBox),
    resolution(res) {
      RasData = new unsigned short[Lines*Samples];
     }
    ~InMemRaster() {
      delete [] RasData;
     }
    unsigned short GetVal(unsigned int i, unsigned int j) {
      return *(RasData+(i*Samples)+j);
     }
    void SetVal(unsigned int i, unsigned int j, unsigned short NewVal) {
      *(RasData+(i*Samples)+j) = NewVal;
     }
   };

  // structure used to represent a raster within a record
  class RasterStruct : public XxxADTValue { 
    public:
    int freq;          // frequency of this raster
    int samples;       // samples / line
    int lines;         // lines in raster
    XXX_BOX geoArea; 
    int resolution;    // resolution of image stored. 1 means full
    XxxShoreLogicalOID  RasterId;
    InMemRaster* MemRaster;

    XxxErrCode CopyHToN(const RasterStruct* RS,
			const RasterMetaInfo* MetaInfo);
    XxxErrCode CopyNToH(const RasterStruct* RS,
			const RasterMetaInfo* MetaInfo);
  } ;

  static const char* const AVHRR;
  static const char* const JPEG;

  //// Plan structure for functions over rasters
  class PlanInfo : public XxxFuncPlanInfo {
    friend class XxxRasterADT;
    friend class AggrOpSpec;

    XxxErrCode execClip(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;
    XxxErrCode execDeres(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;
      
    XxxErrCode getMemRaster(RasterStruct* RS,
			    XxxBool ToDecompress,
			    InMemRaster*& Raster,
			    const XXX_BOX* DesiredBox = NULL) const;
    XxxErrCode putMemRaster(InMemRaster* Raster, 
			    XxxBool ToCompress,
			    RasterStruct* RS,
			    XxxShoreVolumeId &VolId,
			    XxxShoreObjectId &FileId) const;
    XxxErrCode reclaimMemRaster(RasterStruct* RS,
				XxxBool ToDecompress) const;

    XxxErrCode ravgInit(XxxAggrOpHandle& Hndl) const;
    XxxErrCode ravgAccum(XxxValueEnv* Env, XxxAggrOpHandle& Hndl) const;
    XxxErrCode ravgTerm(XxxAggrOpHandle& Hndl, XxxADTValue *FinalVal) const;

    XxxErrCode ranyInit(XxxAggrOpHandle& Hndl) const;
    XxxErrCode ranyAccum(XxxValueEnv* Env, XxxAggrOpHandle& Hndl) const;
    XxxErrCode ranyTerm(XxxAggrOpHandle& Hndl, XxxADTValue *FinalVal) const;

   public:
    PlanInfo(const XxxFuncParseInfo* FuncParse, int ArgCount) 
    : XxxFuncPlanInfo(FuncParse, ArgCount)  {}
    virtual ~PlanInfo() {}
  };
  friend class PlanInfo;

  //// Encapsulates all informaton about moving raster data to/from SHORE
  class RasterHandler { 
   public:    
    static XxxErrCode Put_Source (RasterStruct *AS1, 
				  XxxShoreObjectId RasterFileId,
				  RasterStruct *AS, char *mem_buffer);
    static XxxErrCode Get_Source (RasterStruct *AS, char *mem_buffer);
    static XxxErrCode Get_Source1 (RasterStruct *AS, char *mem_buffer, 
				   double t1, double t2);
   }; // class RasterHandler

  // local method defs

  XxxErrCode parseMetaInfo(const char* MetaInfoString,
			   MetaInfoHandle& MIHandle) const;
  virtual XxxErrCode makeMetaInfo(const MetaInfoHandle& MIHandle,
				  XxxADTMetaInfo*& MI) const;
  
 public:
  static XxxBasicType RasterTypeId;  // type id accessible without object

 protected:
  virtual XxxErrCode getOid(const XxxADTValue* Object, 
			    XxxShoreVolumeId& VolId, 
			    XxxShoreObjectId& LocalId) const {
    XxxShoreLogicalOID* RasterId = &(((RasterStruct*)Object)->RasterId);
    VolId = RasterId->VolId;
    LocalId = RasterId->LocalId;
    return XXX_OK;
   }
 private:
  XxxErrCode avhrrToJpeg(const char* FileName, int freq, 
			 const char* JpegName, int pixels, int scans) const;

protected:
  virtual XxxErrCode showDataFile(const char* FileName,
				  ostream& OutStream, 
				  const XxxADTValue* Object,
				  const XxxADTMetaInfo* MetaInfo) const;

  // for now I'm just making the AggrOpSpec class a friend of XxxRasterADT
  // should be easier to work this into the new framework -- Anil
  friend class AggrOpSpec;
  //// 
  class RasterAvgInfo : public XxxAggrOpHandle::AggrState { 
   public:
    InMemRaster* Raster; 
    RasterAvgInfo() { Raster = NULL;}
    virtual ~RasterAvgInfo() { if (Raster) delete Raster;}
  };

 public:

  XxxRasterADT(XxxBasicType Id);
  ~XxxRasterADT();

  virtual XxxErrCode ReadText(istream& InStream, XxxADTValue* ObjectRef,
			      const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxErrCode CopyHToN(const XxxADTValue* ObjectRef, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxErrCode CopyNToH(const XxxADTValue* ObjectRef, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
};

inline XxxRasterADT::RasterMetaInfo::RasterMetaInfo () { 
  // default constructor
  Format = RASTER_AVHRR;
  AvgRasterSize = 8192000;
}

#endif /* XXX_RASTER_H */
