#ifndef XXX_STRUCT_H
#define XXX_STRUCT_H

#include "prd-types.h"
#include "genadt.h"
#include "reldata.h"
#include "adtfunc.h"

XXX_EXTERN_CLASS_DECL(XxxFuncPlanInfo);

/* For future use in sql.y
struct StructType {
	int Type;
	XxxGenDataSrc::AttrTypeList *AttrsList;
};
*/

class StructPlanInfo : public XxxFuncPlanInfo {
public:
	StructPlanInfo(const StructPlanInfo &FPI):
	  XxxFuncPlanInfo(FPI){}
	  
	  StructPlanInfo(const char* Name, XxxBasicType Type,
		  int EngIndex, XxxGenericPlan* Exp):
	  XxxFuncPlanInfo(Name, Type, EngIndex, Exp){}
	  
	  StructPlanInfo(const XxxFuncParseInfo* ParseInfo,
		  int ArgCount, XxxGenericPlan* Plan = NULL):
	  XxxFuncPlanInfo(ParseInfo, ArgCount, Plan){}
	  
	  /// Evaluate this expression
	  virtual XxxErrCode Evaluate(XxxValueEnv* Env,
		  XxxFuncMethodInfo* ThisMethInfo,
		  XxxADTValue *ReturnValue) const;
};

/** 
 * XxxStructADT is a composite data type which allows the user
 * to create its own struct type based on other basic data types 
 * or other struct types. (Nesting of struct type is supported)
 * 
 * The names of the methods which can be applied on a struct ADT
 * are this struct's member field names. The return value is the
 * value of that member in the struct. The typechecking and optimization
 * for struct ADT is different from other types: it only has dummmy
 * metharray and aggrmetharray.
 *
 * @author BinQuan Gu, Jythar Chong
 */
class XxxStructADT : public XxxSmallVarADT {
public:
	/**
	 *	This struct stores the value of this struct object.
	 *  Memory layout Format:    
	 *  ||int(size) |   char array, with length size(data)||
	 *  size: the size of char array
	 *  data: the serialized data for a record, which contains
	 *		  value for this struct object
	 */
	class StructStruct : public XxxADTValue {
	public:
		char Val[1];
	};
	
	static XxxBasicType StructTypeId;

	/// meta-info for struct
	class StructMetaInfo : public XxxADTMetaInfo { 
	public:
		// Number of field types contained in this struct
		int numAttrs;
		// The schema for all the field types
		XxxRecordSchema *StructSchema;
		/**
		 *  Create the Struct MetaInfo, MetaHandle from the input
		 *  meta attribute list
		 *	@param Attr, the attribute list for the meta string
		 *  @param MIHandle, Metahandle of the struct ADT
		 *  @param status, the status of constuctor
		 */
		StructMetaInfo(XxxGenDataSrc::AttrTypeList *Attr, 
			           XxxADTClass::MetaInfoHandle &MIHandle, 
					   XxxErrCode& status);

		StructMetaInfo(int numAttrs, XxxRecordSchema *StructSchema) {
			this->numAttrs = numAttrs;
			this->StructSchema = new XxxRecordSchema(*StructSchema);
		}
		//Bjarke, sep. 13 2001: explicitly initialize base class XxxAdtMetaInfo
		StructMetaInfo(const StructMetaInfo &structMetaInfo) :
		  XxxADTMetaInfo()
		  {
		    numAttrs = structMetaInfo.numAttrs;
		    StructSchema = new XxxRecordSchema(*structMetaInfo.StructSchema);
		  }

		virtual ~StructMetaInfo() ;
		/**
		 *	Make a replicate of another struct metainfo
		 */
		virtual XxxErrCode Replicate(XxxADTMetaInfo*& NewMI) const {
			StructMetaInfo* NewMetaInfo = new StructMetaInfo(*this);
			NewMI = NewMetaInfo;
			return XXX_OK;
		}
		/**
		 *	Print this metainfo
		 */
		virtual void Print(ostream& OStr) const;
		/**
		 *	Check whether this struct metainfo is compatible with
		 *  "OtherMI"
		 */
		virtual XxxBool IsCompatible(const XxxADTMetaInfo* OtherMI) const ;
		/**
		* Returns the number of bytes necessary to hold the serialized version
		* of this meta-info object.
		* @param Size (out) The required number of bytes
		* @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
		*/
		virtual XxxErrCode SerializeCopySize(int &Size) const ;
		/**
		* Creates a binary version of this meta-info object.
		* @param Data An array of bytes of at least the size specified by the
		*             SerializeCopySize() function.
		* @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
		*/
		virtual XxxErrCode SerializeCopy(char *Data) const ;
	};  
	friend class StructMetaInfo;
	

	XxxStructADT(XxxBasicType Id);
	
	~XxxStructADT(){}
	
	// Get the methodinfo for a method by its index in meth array
	virtual FuncMethodInfo* GetMethByIndex(int index) const
	{ 
		// Return a dummy meth info
		return MethArray[0];
	}
	
	/**
	 *	Print the MetaInfo to Out stream "OStr"
	 */
	virtual void Print(ostream& OStr, const XxxADTMetaInfo* MetaInfo) const;
	
	/**
	 *	Write the value of this struct ADT to outstream
	 */
	virtual XxxErrCode WriteText(ostream& OutStream, const XxxADTValue* Object,
		const XxxADTMetaInfo* MetaInfo) const;
	
	/**
	 *	Read data of this struct ADT into ObjectRef, used by LoadRelation
	 */
	virtual XxxErrCode ReadText(istream& InStream, XxxADTValue* ObjectRef,
		const XxxADTMetaInfo* MetaInfo) const;

	/**
	 *	Used to parse Meta string into Attribute type list
	 */
	virtual XxxErrCode parseMetaString(const char* MetaInfoString,
				  XxxGenDataSrc::AttrTypeList*& structAttr) const;
	
	/**
	 *  Parse MetaInfoString, and Make MetaInfoHandle
	 */
    virtual XxxErrCode parseMetaInfo(const char* MetaInfoString,
				  MetaInfoHandle& MIHandle) const;

	/**
	 *	Make MetaInfo from MetaInfoHandle
	 */
    virtual XxxErrCode makeMetaInfo(const MetaInfoHandle& MIHandle,
				  XxxADTMetaInfo*& MI) const;


	/**
	 *	Get the Maximum size of this object
	 */
	virtual int MaxObjectSize(const XxxADTMetaInfo*) const;
	
	/**
	 *	Type check the method invoked on a struct object
	 */
	virtual XxxErrCode FuncTypeCheck(const char*FuncName,
							const XxxValueExpression* Owner,
							const XxxValueExprList *ArgList,
							XxxBool IsAggr,
							XxxBasicType& ReturnType,
							XxxFuncParseInfo*& FuncParse) const;
	
	/**
	 *	Optimize the execution of the method
	 */
	virtual XxxErrCode FuncOptimize(XxxFuncParseInfo* FParse,
							XxxValueExprPlan* OwnerPlan,
							XxxValExprPlanList* ArgPlans,
							XxxFuncPlanInfo*& FuncPlan) const;
	
	/**
	 *	Create Plan info
	 */
	virtual XxxErrCode methOptimize(XxxFuncParseInfo* FuncParse,
							XxxValueExprPlan* OwnerPlan,
							XxxValExprPlanList* ArgPlans,
							XxxFuncPlanInfo*& FPlan) const 
	{ 
		FPlan = new StructPlanInfo(FuncParse, (ArgPlans?ArgPlans->Count():0)); 
		return XXX_OK; 
	}
	
	/**
	 *	Host to Network data fromat transformation
	 */
	virtual XxxErrCode CopyHToN(const XxxADTValue* Object, 
		XxxADTValue *ResultObject,
		const XxxADTMetaInfo* MetaInfo) const;
	
	/**
	 *	Network to Host data format transformation
	 */
	virtual XxxErrCode CopyNToH(const XxxADTValue* Object, 
		XxxADTValue *ResultObject,
		const XxxADTMetaInfo* MetaInfo) const;
	
};

#endif
