#include "relexpr.h"
#include "relenv.h"

//
// Methods of RelUnarySetExpression
//
RelUnarySetExpression::RelUnarySetExpression(const XxxRelationEngine* REng,
					     RelViewClass* SubQuery,
					     RelUnarySetOpType OpType)
  : XxxRelBooleanExpression(XXX_OTHER_OP) 
{
  setOpType = OpType;
  queryRootNode = NULL;
  relEng = REng;
  if (relEng->QueryEng->ViewToOpDag(SubQuery, queryRootNode) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error in exists subquery");
    queryRootNode = NULL;
  }
  // delete SubQuery; ????? can i do this
}

XxxErrCode RelUnarySetExpression::TypeCheck()
{
  // simply type-check the nested query block
  if (!queryRootNode) {
    REL_NOTE_ERROR(REL_BADQUERY, "No nested query");
    return XXX_NOT_OK;
  }
  // determine the current XxxArgEnv
  XxxArgEnv* CurrentArgEnv = XXX_THISRELENV()->NestedState.TopStack();
  if (!CurrentArgEnv) {
    REL_NOTE_ERROR(REL_BADQUERY, "No current arg env");
    return XXX_NOT_OK;
  }

  if ((relEng->QueryEng->CheckOpDag(queryRootNode, CurrentArgEnv, 
				    queryObject) != XXX_OK)
      || (!queryObject)) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error type-checking nested query");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode 
RelUnarySetExpression::GetDependencies(XxxBitMap& Dependency,
				       int CorrelationHt) const
{
  XXX_CALL_RET(queryObject->GetDependencies(Dependency, CorrelationHt+1),
	       XXX_BADEXPR, "Error getting subquery dependencies");
  return XXX_OK;
}

XxxErrCode RelUnarySetExpression::Optimize(XxxExprPlan*& ExecPlan)
{
  // optimize the nested query block
  if (!queryObject) {
    REL_NOTE_ERROR(REL_BADQUERY, "No nested query");
    return XXX_NOT_OK;
  }
  RelSpecParameters Params(XXX_TRUE);
  RelQueryPlan* PlanQP = NULL;

  if ((relEng->Optimizer->OptimizeQuery(queryObject, &Params, 
					PlanQP) != XXX_OK)
      || (!PlanQP)) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error optimizing nested query");
    return XXX_NOT_OK;
  }

  RelUnarySetPlan* RPlan = makePlan(PlanQP);
  ExecPlan = RPlan;
 
  return XXX_OK;
}

RelUnarySetExpression::
RelUnarySetPlan::RelUnarySetPlan(const XxxRelationEngine* REng,
				 RelQueryPlan* PlanQP) :
  XxxBooleanExpressionPlan(XXX_REL_BOOL_EXPR, 0, XXX_FALSE, 0)
{
  relEng = REng;
  queryPlan = PlanQP;
}

RelUnarySetExpression::
RelUnarySetPlan::~RelUnarySetPlan()
{
  if (queryPlan) delete queryPlan;
}

XxxErrCode RelUnarySetExpression::
RelUnarySetPlan::RedirectDependencies(const VarRemapping& Remap,
				      int CorrelationHt)
{
  // need to redirect any correlation dependencies
  if (queryPlan) queryPlan->RedirectDependencies(Remap, CorrelationHt+1);
  return XXX_OK;
}

//
// Define four methods for RelExistsExpression
//   1) Print
//   2) Match
//   3) Selectivity
//   4) Plan::Evaluate()
//
XxxBool RelExistsExpression::Match(const XxxExpression* Expr, 
				   char *Env) const
{
  return XXX_FALSE;
}

double RelExistsExpression::Selectivity(SelectivityStatsInfo* StatsArray, 
					int ArraySize) const
{
  return 0.5;
}

void RelExistsExpression::Print(ostream& OStr) const
{
  OStr << " EXISTS (";
  if (queryRootNode) queryRootNode->PrintThis(OStr);
  OStr << ")";
}

XxxErrCode RelExistsExpression::
RelExistsPlan::Evaluate(const XxxRecord* Rec[],
			const XxxRecordSchema* Schema[],
			XxxADTValue *Result,
			const XxxValueEnv* Env) const
{
  // put the current "environment" on the stack
  XxxValueEnv CurrentValueEnv(0,0,0);
  // Note: "current" env is stored as a field of the new env.
  // this effectively creates a stack of envs refelcting the
  // nested query block invocations.
  CurrentValueEnv.NEEnv.Init(Rec, Schema, Env);

  // specify correlation parameters in ExecParams
  RelExecParameters ExecParams(&CurrentValueEnv);
  ExecParams.FirstTuple = XXX_TRUE;

  // interpret result as a ptr to a XxxBool
  XxxCoreADTClass::BoolValue* Answer = (XxxCoreADTClass::BoolValue*)Result;

  RelEvalCursor* EvalCursor = NULL;
  if (relEng->ExecEng->GetEvalCursor(queryPlan, &ExecParams, 
				     EvalCursor) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error executing nested query");
    return XXX_NOT_OK;
  }
  XxxBool ScanValid = XXX_FALSE;
  XxxRecord* RelRecord;
  if (EvalCursor->GetNext(RelRecord, ScanValid) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error executing nested query");
    return XXX_NOT_OK;
  }
  if (ScanValid)
    Answer->Value = XXX_TRUE;
  else
    Answer->Value = XXX_FALSE;

  delete EvalCursor;
  return XXX_OK;
}


//
// Methods of RelUniqueExpression
//
XxxBool RelUniqueExpression::Match(const XxxExpression* Expr, 
				   char *Env) const
{
  return XXX_FALSE;
}

double RelUniqueExpression::Selectivity(SelectivityStatsInfo* StatsArray, 
					int ArraySize) const
{
  return 0.5;
}

void RelUniqueExpression::Print(ostream& OStr) const
{
  OStr << " UNIQUE (";
  if (queryRootNode) queryRootNode->PrintThis(OStr);
  OStr << ")";
}

XxxErrCode RelUniqueExpression::
RelUniquePlan::Evaluate(const XxxRecord* Rec[],
			const XxxRecordSchema* Schema[],
			XxxADTValue *Result,
			const XxxValueEnv* Env) const
{
  // put the current "environment" on the stack
  XxxValueEnv CurrentValueEnv(0,0,0);
  CurrentValueEnv.NEEnv.Init(Rec, Schema, Env);

  // specify correlation parameters in ExecParams
  RelExecParameters ExecParams(&CurrentValueEnv);
  ExecParams.FirstTuple = XXX_TRUE;

  // interpret result as a ptr to a XxxBool
  XxxCoreADTClass::BoolValue* Answer = (XxxCoreADTClass::BoolValue*)Result;
  Answer->Value = XXX_FALSE;

  RelEvalCursor* EvalCursor = NULL;
  if (relEng->ExecEng->GetEvalCursor(queryPlan, &ExecParams, 
				     EvalCursor) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error executing nested query");
    return XXX_NOT_OK;
  }
  XxxBool ScanValid = XXX_FALSE;
  XxxRecord* RelRecord;
  if (EvalCursor->GetNext(RelRecord, ScanValid) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error executing nested query");
    return XXX_NOT_OK;
  }
  if (ScanValid) {
    if (EvalCursor->GetNext(RelRecord, ScanValid) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADQUERY, "Error executing nested query");
      return XXX_NOT_OK;
    }
    if (!ScanValid) { // failed on the second try: so it is unique
      Answer->Value = XXX_TRUE;
    }
  } // if ScanValid

  delete EvalCursor;
  return XXX_OK;
}



//
// Methods of RelBinarySetExpression
//
void RelBinarySetExpression::init(const XxxRelationEngine* REng,
				  RelViewClass* SubQuery,
				  RelBinarySetOpType Op1,
				  XxxRelOpType Op2,
				  XxxValueExpression* LHSValExpr)
{
  queryRootNode = NULL;
  relEng = REng;
  if (relEng->QueryEng->ViewToOpDag(SubQuery, queryRootNode) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error in exists subquery");
    queryRootNode = NULL;
  }
  // delete SubQuery; ????? can i do this
  binOpType.SetOp = Op1; 
  binOpType.RelOp = Op2; 
}

XxxErrCode RelBinarySetExpression::TypeCheck()
{
  // type-check the nested query block and the valexpr
  if (!queryRootNode) {
    REL_NOTE_ERROR(REL_BADQUERY, "No nested query");
    return XXX_NOT_OK;
  }
  // determine the current XxxArgEnv
  XxxArgEnv* CurrentArgEnv = XXX_THISRELENV()->NestedState.TopStack();
  if (!CurrentArgEnv) {
    REL_NOTE_ERROR(REL_BADQUERY, "No current arg env");
    return XXX_NOT_OK;
  }

  if ((relEng->QueryEng->CheckOpDag(queryRootNode, CurrentArgEnv, 
				    queryObject) != XXX_OK)
      || (!queryObject)) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error type-checking nested query");
    return XXX_NOT_OK;
  }

  if (XxxExpression::TypeCheck() != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error type-checking val expr in set subquery");
    return XXX_NOT_OK;
  }

  // make sure that types are compatible
  XxxRecordSchema* NestedSchema = ((RelCatalogInfo*)(queryObject->GenerateCatInfo()))->RecordSchema();
  if (NestedSchema->Arity() != 1) {
    REL_NOTE_ERROR(REL_BADQUERY, "IN Expr must have sinlge attribute");
    return XXX_NOT_OK;
  }

  if (NestedSchema->GetAttr(0)->Type != ChildExpr(0)->BasicType()) {
    REL_NOTE_ERROR(REL_BADQUERY, "Type mismatch in IN Expr");
    return XXX_NOT_OK;
  }

  return XXX_OK;
}

XxxErrCode 
RelBinarySetExpression::GetDependencies(XxxBitMap& Dependency,
				       int CorrelationHt) const
{
  XXX_CALL_RET(XxxRelBooleanExpression::GetDependencies(Dependency,
							CorrelationHt),
	       XXX_BADEXPR, "Error getting dependencies");
  XXX_CALL_RET(queryObject->GetDependencies(Dependency, CorrelationHt+1),
	       XXX_BADEXPR, "Error getting subquery dependencies");
  return XXX_OK;
}

XxxErrCode RelBinarySetExpression::Optimize(XxxExprPlan*& ExecPlan)
{
  // optimize the nested query block
  if (!queryObject) {
    REL_NOTE_ERROR(REL_BADQUERY, "No nested query");
    return XXX_NOT_OK;
  }
  RelSpecParameters Params(XXX_TRUE);
  RelQueryPlan* PlanQP = NULL;

  if ((relEng->Optimizer->OptimizeQuery(queryObject, &Params, 
					PlanQP) != XXX_OK)
      || (!PlanQP)) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error optimizing nested query");
    return XXX_NOT_OK;
  }

  // optimize the value expression
  XxxExprPlan* lValPlan;
  if (ChildExpr(0)->Optimize(lValPlan) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error optimizing LHS of nested query");
    return XXX_NOT_OK;
  }

  RelBinarySetPlan* RPlan = makePlan(PlanQP, (XxxValueExprPlan*)lValPlan);
  ExecPlan = RPlan;

  return XXX_OK;
}

RelBinarySetExpression::
RelBinarySetPlan::RelBinarySetPlan(const XxxRelationEngine* REng,
				   RelQueryPlan* PlanQP,
				   XxxValueExprPlan* ValPlan) :
  XxxBooleanExpressionPlan(XXX_REL_BOOL_EXPR, 0, XXX_FALSE, 0)
{
  relEng = REng;
  queryPlan = PlanQP;
  exprPlan = ValPlan;

  TypeL = ValPlan->BasicType();
  MetaInfoL = ValPlan->GenerateCatInfo();
  ArgL = (XxxADTValue *)(new char[XxxADTArray[TypeL]->MaxObjectSize(MetaInfoL)]);
  ArgL->Init();

  /***
  XxxRecordSchema* NestedSchema = 
    PlanQP->TopBlockPlan->CatalogInfo->RecordSchema();
  TypeR = NestedSchema->GetAttr(0)->Type;
  MetaInfoRL = NestedSchema->GetAttr(0)->TypeCatInfo();
  ArgR = (XxxADTValue *)(new char[XxxADTArray[TypeR]->MaxObjectSize(MetaInfoR)]);
  ArgR->Init();
  ***/
}

RelBinarySetExpression::
RelBinarySetPlan::~RelBinarySetPlan()
{
  if (queryPlan) delete queryPlan;
  if (ArgL) delete [] ((char *)ArgL);
}

XxxErrCode RelBinarySetExpression::
RelBinarySetPlan::RedirectDependencies(const VarRemapping& Remap,
				       int CorrelationHt)
{
  // need to redirect any correlation dependencies
  if (queryPlan) queryPlan->RedirectDependencies(Remap, CorrelationHt+1);
  // also redirect the simple expression
  if (exprPlan) exprPlan->RedirectDependencies(Remap, CorrelationHt);
  return XXX_OK;
}

XxxErrCode RelBinarySetExpression::
RelBinarySetPlan::Evaluate(const XxxRecord* Rec[],
			   const XxxRecordSchema* Schema[],
			   XxxADTValue *Result,
			   const XxxValueEnv* Env) const
{
  // put the current "environment" on the stack
  XxxValueEnv CurrentValueEnv(0,0,0);
  CurrentValueEnv.NEEnv.Init(Rec, Schema, Env);

  // specify correlation parameters in ExecParams
  RelExecParameters ExecParams(&CurrentValueEnv);
  ExecParams.FirstTuple = XXX_TRUE;

  // interpret result as a ptr to a XxxBool
  XxxCoreADTClass::BoolValue* Answer = (XxxCoreADTClass::BoolValue*)Result;
  Answer->Value = XXX_FALSE;

  // first evaluate the LHS expression
  if (exprPlan->Evaluate(Rec, Schema, ArgL, Env) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error executing LHS of IN Expr");
    return XXX_NOT_OK;
  }

  RelEvalCursor* EvalCursor = NULL;
  if (relEng->ExecEng->GetEvalCursor(queryPlan, &ExecParams, 
				     EvalCursor) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error executing nested query");
    return XXX_NOT_OK;
  }
  XxxBool ScanValid = XXX_FALSE;
  XxxRecord* RelRecord;
  while (1) {
    if (EvalCursor->GetNext(RelRecord, ScanValid) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADQUERY, "Error executing nested query");
      return XXX_NOT_OK;
    }
    if (!ScanValid)
      break;

    // ok, we have the record, lets evaluate the expression:
    REL_CALL_FN(getAnswer(RelRecord, EvalCursor->AnswerSchema, Answer),
		REL_BADQUERY, "Error getting answer from record");

    // if it is true, break out of loop
    if (Answer->Value) break;

  } // while (1)

  delete EvalCursor;
  return XXX_OK;
}

//
// Define four methods for RelAnyExpression
//   1) Print
//   2) Match
//   3) Selectivity
//
XxxBool RelAnyExpression::Match(const XxxExpression* Expr, 
				char *Env) const
{
  return XXX_FALSE;
}

double RelAnyExpression::Selectivity(SelectivityStatsInfo* StatsArray, 
				     int ArraySize) const
{
  return 0.8;
}

void RelAnyExpression::Print(ostream& OStr) const
{
  ChildExpr(0)->Print(OStr);
  switch (binOpType.RelOp) {
  case XXX_GT_OP: OStr << " > "; break;
  case XXX_LT_OP: OStr << " < "; break;
  case XXX_EQ_OP: OStr << " = "; break;
  case XXX_LEQ_OP: OStr << " <= "; break;
  case XXX_GEQ_OP: OStr << " >= "; break;
  case XXX_NEQ_OP: OStr << " <> "; break;
  default: OStr << " ?? "; break;
  }
  OStr << " ANY (";
  if (queryRootNode) queryRootNode->PrintThis(OStr);
  OStr << ")";
}

XxxErrCode RelAnyExpression::
RelAnyPlan::getAnswer(const XxxRecord* Record, 
		      const XxxRecordSchema* Schema,
		      XxxCoreADTClass::BoolValue* Answer) const
{
  const char* FieldPtr;
  Record->GetField(0, Schema, FieldPtr);
  XxxAttribute* Attr = Schema->GetAttr(0);
  
 XXX_CALL_RET(XxxRelBooleanPlan::
	      ExprEval(ArgL, TypeL, MetaInfoL, 
		       (XxxADTValue*)FieldPtr, Attr->Type, Attr->TypeCatInfo,
		       relOp, inverted, Answer->Value),
	      XXX_BADEXPR, "Error evaluating relational expression");
  return XXX_OK;
}

//
// Define four methods for RelInExpression
//   1) Print
//   2) Match
//   3) Selectivity
//
XxxBool RelInExpression::Match(const XxxExpression* Expr, 
				char *Env) const
{
  return XXX_FALSE;
}

double RelInExpression::Selectivity(SelectivityStatsInfo* StatsArray, 
				    int ArraySize) const
{
  return 0.8;
}

void RelInExpression::Print(ostream& OStr) const
{
  ChildExpr(0)->Print(OStr);
  OStr << " IN (";
  if (queryRootNode) queryRootNode->PrintThis(OStr);
  OStr << ")";
}

XxxErrCode RelInExpression::
RelInPlan::getAnswer(const XxxRecord* Record, 
		      const XxxRecordSchema* Schema,
		      XxxCoreADTClass::BoolValue* Answer) const
{
  const char* FieldPtr;
  Record->GetField(0, Schema, FieldPtr);
  if (XxxCoreADTClass::IsCoreADT(TypeL)) {
    Answer->Value = XxxCoreADTClass::
	SEquals(TypeL, ArgL, MetaInfoL, (XxxADTValue*)FieldPtr, MetaInfoL);
  }
  else {
    Answer->Value = XxxADTArray[TypeL]->
      Equals(ArgL, MetaInfoL, (XxxADTValue*)FieldPtr, MetaInfoL);
  }
  return XXX_OK;
}


