#ifndef REL_EXPR_H
#define REL_EXPR_H

#include "expr.h"
#include "releng.h"
#include "relquery.h"
#include "relopt.h"
#include "relexec.h"

typedef enum {
  REL_EXISTS_OP,
  REL_UNIQUE_OP
 } RelUnarySetOpType;

typedef enum {
  REL_ANY_OP,
  REL_ALL_OP,
  REL_1TUPLE_OP,
  REL_IN_OP,
  REL_NOTIN_OP
 } RelBinarySetOpType;

typedef struct RelBinaryOp {
  RelBinarySetOpType SetOp;
  XxxRelOpType       RelOp;
} RelBinaryOp;

/**
 * This is the parent class for EXISTS and UNIQUE subquery expressions
 */
class RelUnarySetExpression : public XxxRelBooleanExpression {
public:
  class RelUnarySetPlan : public XxxBooleanExpressionPlan { 
    friend class RelUnarySetExpression;
  protected:
    virtual PlanUsage& getPlanUsage() { return planUsage;}
    RelQueryPlan* queryPlan;
    const XxxRelationEngine* relEng;
  public:
    RelUnarySetPlan(const XxxRelationEngine* REng,
		  RelQueryPlan* PlanQP);
    RelUnarySetPlan(const RelUnarySetPlan& Plan)
      : XxxBooleanExpressionPlan(Plan) {
	queryPlan = Plan.queryPlan;
	relEng = Plan.relEng;
       }
    virtual ~RelUnarySetPlan();
    virtual XxxErrCode RedirectDependencies(const VarRemapping& Remap,
					    int CorrelationHt);
  };
  friend class RelUnarySetPlan;

protected:
  // method defined by each subclass
  virtual RelUnarySetPlan* makePlan(RelQueryPlan* PlanQP) = 0;

  RelQueryNode* queryRootNode;
  const XxxRelationEngine* relEng;
  RelQueryClass* queryObject;
  RelUnarySetOpType setOpType;
 public:
  RelUnarySetExpression(const XxxRelationEngine* RelEng,
			RelViewClass* SubQuery,
			RelUnarySetOpType OpType);
  /**
   * Copy constructor.  Makes a complete copy (recursing if necessary) of
   * the supplied expression.
   */
  RelUnarySetExpression(const RelUnarySetExpression& src) : 
    XxxRelBooleanExpression(src)  { 
      // TODO: dont have this Replicate() method implemented yet
      // queryRootNode = src.queryRootNode->Replicate(); 
      queryRootNode = src.queryRootNode;
      relEng = src.relEng;
      queryObject = src.queryObject;
      setOpType = src.setOpType;
  }
  virtual ~RelUnarySetExpression() {
    // should delete queryRootNode, but cant do it yet ... TODO
  }
  virtual XxxErrCode Optimize(XxxExprPlan*& ExecPlan);
  virtual XxxErrCode TypeCheck();

  //// Gets input stream dependencies 
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependendency,
				     int CorrelationHt) const;
 };

/**
 * This classis used to model EXISTS suqueries.
 */
class RelExistsExpression : public RelUnarySetExpression {
protected:
  /// Plan for exists subqueries
  class RelExistsPlan : public RelUnarySetExpression::RelUnarySetPlan { 
    friend class RelExistsExpression;
  protected:
  public:
    RelExistsPlan(const XxxRelationEngine* REng,
		  RelQueryPlan* PlanQP) 
      : RelUnarySetExpression::RelUnarySetPlan(REng, PlanQP) {}
    RelExistsPlan(const RelExistsPlan& Plan) 
      : RelUnarySetExpression::RelUnarySetPlan(Plan) {}
    virtual XxxGenericPlan * Replicate() const {
      return new RelExistsPlan(*this);
     }
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env = NULL) const;
  };
  friend class RelExistsPlan;
  virtual RelUnarySetPlan* makePlan(RelQueryPlan* PlanQP) {
    return new RelExistsPlan(relEng, PlanQP);
  }

public:
  RelExistsExpression(const XxxRelationEngine* RelEng, RelViewClass* SubQuery) 
    : RelUnarySetExpression(RelEng, SubQuery, REL_EXISTS_OP){}
  /**
   * Copy constructor.  Makes a complete copy (recursing if necessary) of
   * the supplied expression.
   */
  RelExistsExpression(const RelExistsExpression& src) : 
    RelUnarySetExpression(src)  { }
  virtual ~RelExistsExpression() {}
  /// Provides a convenient way to copy arbitrary expressions
  virtual XxxExpression * Replicate() const
    { return new RelExistsExpression(*this); }
  virtual void Print(ostream&) const;
  virtual XxxBool Match(const XxxExpression* Expr, 
			char *Env = NULL) const;
  virtual double Selectivity(SelectivityStatsInfo* StatsArray, 
			     int ArraySize) const;
};

/**
 * This classis used to model UNIQUE suqueries.
 */
class RelUniqueExpression : public RelUnarySetExpression{
protected:
  /// Plan for Unique subqueries
  class RelUniquePlan : public RelUnarySetExpression::RelUnarySetPlan { 
    friend class RelUniqueExpression;
  public:
    RelUniquePlan(const XxxRelationEngine* REng,
		  RelQueryPlan* PlanQP) 
      : RelUnarySetExpression::RelUnarySetPlan(REng, PlanQP) {}
    RelUniquePlan(const RelUniquePlan& Plan) 
      : RelUnarySetExpression::RelUnarySetPlan(Plan) {}
    virtual XxxGenericPlan * Replicate() const {
      return new RelUniquePlan(*this);
     }
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env = NULL) const;
  };
  friend class RelUniquePlan;
  virtual RelUnarySetPlan* makePlan(RelQueryPlan* PlanQP) {
    return new RelUniquePlan(relEng, PlanQP);
  }

public:
  RelUniqueExpression(const XxxRelationEngine* RelEng, RelViewClass* SubQuery) 
   : RelUnarySetExpression(RelEng, SubQuery, REL_UNIQUE_OP) {}
  /**
   * Copy constructor.  Makes a complete copy (recursing if necessary) of
   * the supplied expression.
   */
  RelUniqueExpression(const RelUniqueExpression& src) : 
    RelUnarySetExpression(src)  { 
  }
  virtual ~RelUniqueExpression() {}
  /// Provides a convenient way to copy arbitrary expressions
  virtual XxxExpression * Replicate() const
    { return new RelUniqueExpression(*this); }
  virtual void Print(ostream&) const;
  virtual XxxBool Match(const XxxExpression* Expr, 
			char *Env = NULL) const;
  virtual double Selectivity(SelectivityStatsInfo* StatsArray, 
			     int ArraySize) const;
};

/**
 * This is the parent class for ANY, ALL, IN and NOT IN subquery expressions
 */
class RelBinarySetExpression : public XxxRelBooleanExpression {
public:
  class RelBinarySetPlan : public XxxBooleanExpressionPlan { 
    friend class RelBinarySetExpression;
  protected:
    virtual PlanUsage& getPlanUsage() { return planUsage;}
    RelQueryPlan* queryPlan;
    const XxxRelationEngine* relEng;
    XxxValueExprPlan* exprPlan;
    XxxADTValue *ArgL;
    XxxBasicType TypeL;
    const XxxADTMetaInfo *MetaInfoL;

    virtual XxxErrCode getAnswer(const XxxRecord* Record, 
				 const XxxRecordSchema* Schema,
				 XxxCoreADTClass::BoolValue* Answer) const = 0;

  public:
    RelBinarySetPlan(const XxxRelationEngine* REng,
		     RelQueryPlan* PlanQP,
		     XxxValueExprPlan* ValPlan);
    virtual ~RelBinarySetPlan();
    virtual XxxErrCode RedirectDependencies(const VarRemapping& Remap,
					    int CorrelationHt);
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env = NULL) const;
  };
  friend class RelBinarySetPlan;

protected:
  // method defined by each subclass
  virtual RelBinarySetPlan* makePlan(RelQueryPlan* PlanQP,
				     XxxValueExprPlan* ValPlan) = 0;

  RelQueryNode* queryRootNode;
  const XxxRelationEngine* relEng;
  RelQueryClass* queryObject;
  RelBinaryOp binOpType;
  void init(const XxxRelationEngine* RelEng,
	    RelViewClass* SubQuery,
	    RelBinarySetOpType Op1,
	    XxxRelOpType Op2,
	    XxxValueExpression* LHSValExpr);
 public:
  RelBinarySetExpression(const XxxRelationEngine* RelEng,
			 RelViewClass* SubQuery,
			 RelBinaryOp OpType,
			 XxxValueExpression* LHSValExpr)
  : XxxRelBooleanExpression(XXX_OTHER_OP, 1, (XxxExpression**)(&LHSValExpr))
    // this makes sure that the LHSValExpr is type-checked normally 
  { init(RelEng, SubQuery, OpType.SetOp, OpType.RelOp, LHSValExpr); }
  RelBinarySetExpression(const XxxRelationEngine* RelEng,
			 RelViewClass* SubQuery,
			 RelBinarySetOpType Op1,
			 XxxRelOpType Op2,
			 XxxValueExpression* LHSValExpr)
  : XxxRelBooleanExpression(XXX_OTHER_OP, 1, (XxxExpression**)(&LHSValExpr))
    // this makes sure that the LHSValExpr is type-checked normally
    { init(RelEng, SubQuery, Op1, Op2, LHSValExpr); }
  /**
   * Copy constructor.  Makes a complete copy (recursing if necessary) of
   * the supplied expression.
   */
  RelBinarySetExpression(const RelBinarySetExpression& src) : 
    XxxRelBooleanExpression(src), binOpType(src.binOpType)  { 
      // TODO: dont have this Replicate() method implemented yet
      // queryRootNode = src.queryRootNode->Replicate(); 
      queryRootNode = src.queryRootNode;
      relEng = src.relEng;
      queryObject = src.queryObject;
  }
  virtual ~RelBinarySetExpression() {
    // should delete queryRootNode, but cant do it yet ... TODO
  }
  virtual XxxErrCode Optimize(XxxExprPlan*& ExecPlan);
  virtual XxxErrCode TypeCheck();
  //// Gets input stream dependencies 
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependendency,
				     int CorrelationHt) const;
 };

/**
 * This class is used to model ANY subqueries
 */
class RelAnyExpression : public RelBinarySetExpression {
protected:
  /// Plan for Any subqueries
  class RelAnyPlan : public RelBinarySetExpression::RelBinarySetPlan { 
    friend class RelAnyExpression;
    XxxRelOpType relOp;
    XxxBool inverted;
  protected:
    virtual XxxErrCode getAnswer(const XxxRecord* Record, 
				 const XxxRecordSchema* Schema,
				 XxxCoreADTClass::BoolValue* Answer) const;
  public:
    RelAnyPlan(const XxxRelationEngine* REng,
	       RelQueryPlan* PlanQP,
	       XxxValueExprPlan* ValPlan,
	       XxxRelOpType RelOp) 
      : RelBinarySetExpression::RelBinarySetPlan(REng, PlanQP, ValPlan) 
    { XxxRelBooleanPlan::InvertOp(RelOp, relOp, inverted); }
    RelAnyPlan(const RelAnyPlan& Plan) 
      : RelBinarySetExpression::RelBinarySetPlan(Plan) {}
    virtual XxxGenericPlan * Replicate() const {
      return new RelAnyPlan(*this);
     }
  };
  friend class RelAnyPlan;
  virtual RelBinarySetPlan* makePlan(RelQueryPlan* PlanQP,
				     XxxValueExprPlan* ValPlan) {
    return new RelAnyPlan(relEng, PlanQP, ValPlan, binOpType.RelOp);
  }

public:
  XxxRelOpType RelOpType;
  RelAnyExpression(const XxxRelationEngine* RelEng, 
		   RelViewClass* SubQuery,
		   XxxRelOpType OpType,
		   XxxValueExpression* LHSValExpr) 
    : RelBinarySetExpression(RelEng, SubQuery, 
			     REL_ANY_OP, OpType, LHSValExpr) {}
  /**
   * Copy constructor.  Makes a complete copy (recursing if necessary) of
   * the supplied expression.
   */
  RelAnyExpression(const RelAnyExpression& src) : 
    RelBinarySetExpression(src)  {}
  virtual ~RelAnyExpression() {}
  /// Provides a convenient way to copy arbitrary expressions
  virtual XxxExpression * Replicate() const
    { return new RelAnyExpression(*this); }
  virtual void Print(ostream&) const;
  virtual XxxBool Match(const XxxExpression* Expr, 
			char *Env = NULL) const;
  virtual double Selectivity(SelectivityStatsInfo* StatsArray, 
			     int ArraySize) const;
};

/**
 * This class is used to model IN subqueries
 */
class RelInExpression : public RelBinarySetExpression {
protected:
  /// Plan for Any subqueries
  class RelInPlan : public RelBinarySetExpression::RelBinarySetPlan { 
    friend class RelInExpression;
  protected:
    virtual XxxErrCode getAnswer(const XxxRecord* Record, 
				 const XxxRecordSchema* Schema,
				 XxxCoreADTClass::BoolValue* Answer) const;
  public:
    RelInPlan(const XxxRelationEngine* REng,
	      RelQueryPlan* PlanQP,
	      XxxValueExprPlan* ValPlan) 
      : RelBinarySetExpression::RelBinarySetPlan(REng, PlanQP, ValPlan) {}
    RelInPlan(const RelInPlan& Plan) 
      : RelBinarySetExpression::RelBinarySetPlan(Plan) {}
    virtual XxxGenericPlan * Replicate() const {
      return new RelInPlan(*this);
     }
  };
  friend class RelInPlan;
  virtual RelBinarySetPlan* makePlan(RelQueryPlan* PlanQP,
				     XxxValueExprPlan* ValPlan) {
    return new RelInPlan(relEng, PlanQP, ValPlan);
  }

public:
  RelInExpression(const XxxRelationEngine* RelEng, 
		  RelViewClass* SubQuery,
		  XxxValueExpression* LHSValExpr) 
    : RelBinarySetExpression(RelEng, SubQuery, 
			     REL_IN_OP, XXX_OTHER_OP,
			     LHSValExpr) {}
  /**
   * Copy constructor.  Makes a complete copy (recursing if necessary) of
   * the supplied expression.
   */
  RelInExpression(const RelAnyExpression& src) : 
    RelBinarySetExpression(src)  { }
  virtual ~RelInExpression() {}
  /// Provides a convenient way to copy arbitrary expressions
  virtual XxxExpression * Replicate() const
    { return new RelInExpression(*this); }
  virtual void Print(ostream&) const;
  virtual XxxBool Match(const XxxExpression* Expr, 
			char *Env = NULL) const;
  virtual double Selectivity(SelectivityStatsInfo* StatsArray, 
			     int ArraySize) const;
};

inline XxxRelOpType RelOpNeg(XxxRelOpType Op) {
  switch (Op) {
  case XXX_GT_OP: return XXX_LEQ_OP;
  case XXX_LT_OP: return XXX_GEQ_OP;
  case XXX_EQ_OP: return XXX_NEQ_OP;
  case XXX_LEQ_OP: return XXX_GT_OP;
  case XXX_GEQ_OP: return XXX_LT_OP;
  case XXX_NEQ_OP: return XXX_EQ_OP;
  default: {
    REL_NOTE_ERROR(REL_BADEXEC, "Bad RelOp type passed to RelNegOp");
    return XXX_EQ_OP;
  }
  } // switch
};

#endif
