#ifndef REL_QUERY_H
#define REL_QUERY_H
#include "relconfig.h"
#include "templates.h"
#include "absquery.h"
#include "reladt.h"
#include "expr.h"
#include "values.h"
#include "relops.h"
#include "relcat.h"

XXX_EXTERN_CLASS_DECL(RelQueryEngine);
/**
 * Defines a relational query node for construction of a query graph.
 */
class RelQueryNode : public XxxQueryNode {
public:
  /// A handle on the operator associated with this node
  RelOperator* Op;
  /// A handle on the catalog information associated with this node
  RelCatalogInfo* CatalogInfo;

  /**
   * Constructor.
   * @param Op The operator associated with this node
   * @param ChildArity The number of children of this node
   * @param Children An array of pointers to this node's children
   * @param NumNames The number of names supplied for children (optional)
   * @param NodeNames The names of the children operators (optional)
   */
  RelQueryNode(RelOperator* Op, 
	       int ChildArity, 
	       RelQueryNode **Children,
	       int NumNames = 0, 
	       char **NodeNames = NULL);

  /**
   * Adds children to this node.  Added to facilitate query rewrite.
   * @param NodeList An array of the children to add to this node
   * @param Quantity The number of children to add
   */
  XxxErrCode AddChildren(RelQueryNode **NodeList, 
			 int Quantity); 

  /**
   * Destructor.
   */
  virtual ~RelQueryNode();

  /**
   * Prints a textual representation of this node to the output stream.
   * @param OStr The stream to print to
   */
  virtual void PrintThis(ostream& OStr) const;

  /**
   * Prints a textual representation of this node, and its children to the output 
   * stream.  The output is formatted such that it can be read by the Java client's 
   * QueryProcDisplayer frame and be displayed graphically.
   * @param OStr The stream to print to
   * @param CurBlockId The node id to be given to this node
   * @param Level The level of this node in the tree (0 is root)
   */
  virtual void TreePrint(ostream& OStr, 
			 int &CurBlockId, 
			 int Level = 0) const;

  /**
   * Prints a textual representation of this node to the output stream.  For
   * use with TreePrint.
   * @param OStr The stream to print to
   * @param CurBlockId The node id to be given to this node
   * @param Level The level of this node in the tree (0 is root)
   * @param NumChildren The number of children of this node
   * @param ChildrenIds The id numbers of this node's children
   */
  virtual void NodePrint(ostream &OStr, int CurBlockId, int Level, 
			 int NumChildren, int *ChildrenIds) const;

  /**
   * Propagates meta-info, such as schema and statistical information, 
   * up the query tree.
   * @param Env The environment information
   * @return XXX_NOT_OK if an error occurs.
   */
  virtual XxxErrCode PropagateMetaInfoUp(const XxxArgEnv* Env = NULL);

  /**
   * There is currently no information which must be propagated down
   * the query graph tree.  This function is empty.
   * @return XXX_NOT_OK if an error occurs.
   */
  virtual XxxErrCode PropagateMetaInfoDown();

  /**
   * Returns the schema associated with the output of this node.
   * @return The schema associated with this node's output, or NULL
   *         if no catalog information is available.
   */
  virtual XxxRecordSchema* Schema() const { 
    if (!CatalogInfo)
      return NULL;
    else
      return CatalogInfo->RecordSchema();
  }
  //// Used to get correlation dependencies
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependency,
				     int CorrelationHt) const {
    if (Op) Op->GetDependencies(Dependency, CorrelationHt);
    return XXX_OK;
  }
};

/**
 * Encapsulates a relational query.  This class holds the query tree
 * associated with a relational query.
 */
class RelQueryClass : public XxxQueryDag {
public:
  /// The root (final) node of the query tree
  RelQueryNode* Root;

  //// Constructor.
  RelQueryClass(RelQueryNode* RootNode) { Root = RootNode;}

  //// Return a copy of this object, sharing the same underlying query tree
  virtual XxxGenericParse* Replicate() const;

  //// Destructor.  Reclaims the underlying query tree.
  virtual ~RelQueryClass() { if (Root) delete Root;}

  //// Prints a textual version of this query.
  virtual void Print(ostream& OStr) const;

  /**
   * Generates catalog information for the query.  This is calculated
   * by returning the root's output catalog information.
   */
  virtual XxxADTMetaInfo* GenerateCatInfo() const 
   { return Root->CatalogInfo; }

  //// Returns the type of the query.  (Always RelationTypeId.)
  virtual XxxBasicType GenerateType() const { return RelationTypeId;}

  //// Used to get correlation dependencies
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependency,
				     int CorrelationHt) const {
    if (Root) return Root->GetDependencies(Dependency, CorrelationHt);
    return XXX_OK;
  }
};

XXX_EXTERN_CLASS_DECL(RelViewClass);

/**
 * Contains information about a relation necesary to process a query.
 * This is an abstract base class.
 */
class RelInfo { public:
  /// The relation's alias; if none is supplied, it is the relation's name
  char AliasName[XXX_NODE_NAME_MAX];
  /// XXX_TRUE if the relation is nested
  XxxBool Nested;
  /**
   * Constructor.
   * @param AName The alias of the relation.
   */
  RelInfo(char* AName) { 
    strncpy(AliasName, AName, XXX_NODE_NAME_MAX); 
    AliasName[XXX_NODE_NAME_MAX-1] = 0;
    Nested  = XXX_FALSE;
  }
  /**
   * Destructor.
   */
  virtual ~RelInfo() {};
  /**
   * Prints a textual version of this relation's information.
   * @param OStr The stream to print to
   */
  virtual void Print(ostream& OStr) const = 0;
};

/**
 * Contains information about a relation necesary to process a query.
 */
class SimpleRelInfo : public RelInfo { public:
  /// The name of the relation
  char RelName[XXX_NODE_NAME_MAX];
  /**
   * Constructor.
   * @param Name The name of the relation
   * @param AName The alias of the relation
   */
  SimpleRelInfo(char* Name, char* AName) : RelInfo(AName) { 
    strncpy(RelName, Name, XXX_NODE_NAME_MAX);
    RelName[XXX_NODE_NAME_MAX-1] = 0;
  }
  /**
   * Prints a textual version of this relation's information.
   * @param OStr The stream to print to
   */
  virtual void Print(ostream& OStr) const;
};

/**
 * Contains information about a nested relation, e.g., a view,
 * necesary to process a query.
 */
class NestedRelInfo : public RelInfo { public:
  /// A pointer to the view associated with this relation
  RelViewClass* View;
  /**
   * Constructor.
   * @param Name The name of the relation
   * @param AName The alias of the relation
   */
  NestedRelInfo(RelViewClass* NestedView, char* AName) : RelInfo(AName)
  { View = NestedView; Nested = XXX_TRUE;}
  /**
   * Prints a textual version of this relation's information.
   * @param OStr The stream to print to
   */
  virtual void Print(ostream& OStr) const;
};

/**
 *
 */
class RelViewClass : public XxxViewClass {
friend class RelOpDagBuilder;
friend class XxxRelationEngine;
friend class RelQueryEngine;
public:
  //// 
  class FromInfo {
  public:
    int Arity;         // number of relations in the FROM clause
    RelInfo** RelList; // an array of size Arity
	//ihab begin
	XxxBool KRJOp;
	//ihab end
    FromInfo(RelInfo* First) {
      Arity = 1;
      RelList = new RelInfo*[1];
      RelList[0] = First;
    }
    virtual ~FromInfo() {
      delete [] RelList;
    }
    RelInfo* operator [] (int index) const
      { return RelList[index]; }
    void AddRelInfo(RelInfo* Next);
    virtual void Print(ostream& OStr) const;
  };
  //// 
  class ProjInfo {
  public:
    XxxBool Distinct;    
    XxxValueExprList* TargetList;
    ProjInfo() { Distinct = XXX_FALSE; TargetList = NULL;}
    virtual void Print(ostream& OStr) const;
  };
  //// 
  class OrderInfo {
  public:
    // NOTE: These arrays store info in decreasing order of sort priority; 
    //       the first item is the main sort column.
    XxxBool *Ascending;
    XxxValueExprList *TargetList;
    int ColArity;
    OrderInfo () : Ascending (NULL), 
    TargetList (NULL), ColArity (0) { }
    virtual void Print(ostream& OStr) const;
  };

  ////
//ihab:stop after information
 class StopInfo {
  public:
    int K;
    StopInfo () : K (0) { }
    StopInfo (int k) : K (k) { }
    virtual void Print(ostream& OStr) const;
  };
 /////

  //Begin Moustafa: Window predicate
  class WindowInfo {
  public:
    // NOTE: Stores input interval for later parsing
    unsigned int 
      IntervalHours,
      IntervalMinutes, 
      IntervalSeconds,
      IntervalMicroSecs;// HHMMSSZZZZZZ
    
    WindowInfo (){}
    /*
      WindowInfo (char* interval){ 
      if(interval){
      Interval = (char*)malloc(22*sizeof(char));
      strcpy(Interval, interval);
      }
      
      }
      WindowInfo (float interval){ 
      Interval = interval;
      }
    */
    void GetWindowValue(unsigned int *sec, unsigned int *microsec){
      *sec= IntervalSeconds;
      *microsec = IntervalMicroSecs;
    } // return the window value (TODO multiple windows)
    XxxBool IsValid(){
      // check that the interval has vaild values
      // hh mm ss zzzzzz
      XxxBool IsV = XXX_FALSE;
      if(IntervalHours>23) return  XXX_FALSE;
      if(IntervalMinutes>59) return  XXX_FALSE;
      if(IntervalSeconds>59) return  XXX_FALSE;
      //      if(IntervalMicroSecs<0) return  XXX_FALSE;
      IntervalSeconds = (IntervalHours*60+IntervalMinutes)*60+IntervalSeconds;
      return XXX_TRUE;
    } // is the window valid
    virtual void Print(ostream& OStr) const;
  };
  //End Moustafa:


  //// 
  class GroupByInfo {
  public:
    XxxValueExprList *GroupList;
    int ColArity;
    GroupByInfo () : GroupList (NULL), ColArity (0) { }
    virtual void Print(ostream& OStr) const;
  };

  void AddFrom(FromInfo *);
  void AddSelection(XxxBooleanExpression *);
  void AddProjection(ProjInfo *);
  void AddGroupBy(GroupByInfo *);
  void AddHaving(XxxBooleanExpression*);
  void AddOrderBy(OrderInfo *);
  void AddStopAfter(StopInfo *); //ihab:stop after
  //Begin Moustafa: Window predicate
  void AddWindow(WindowInfo *);
  //End Moustafa: Window predicate
protected:
  FromInfo* FromList;
  ProjInfo* TargetList;
  XxxBooleanExpression* WhereClause;
  GroupByInfo* GroupByList;
  XxxBooleanExpression* HavingClause;
  OrderInfo* OrderByList;
  StopInfo * StopValue;  // ihab

  //Begin Moustafa: Window predicate
  WindowInfo * WindowValue;
  //End Moustafa: Window predicate
public:
  RelViewClass(const char* HeadName);
  virtual ~RelViewClass();
  virtual void Print(ostream& OStr) const;
  virtual void PrintWithoutName(ostream& OStr) const;
};

/** 
 * The RelOpDagBuilder class encapsulates the process of checking query 
 * semantics and structuring a query directed acyclic graph (DAG).  It 
 * generates extra info which it uses to construct a DAG.  Note that the 
 * redundant RelQueryEngine::ViewToOpDag function is needed since a 
 * RelOpDagBuilder is created for every view transformed--a process which 
 * can be recursively necessary if a View is a FROM argument.
 *
 * The query graph is structured in the following way: an initial SPJ
 * (select-project-join) node is created which takes all the query inputs
 * as its children.  Its projection list is a superset of the actual
 * projection list, including also those attributes necessary for
 * intermediate operations, such as ordering on a non-projected field.
 *
 * If aggregates (with or without grouping) are specified, a Group-By
 * operator is placed above.  If ordering is supplied, an Order-by
 * operator is placed above.  Finally, if items were projected in which
 * were used only for intermediate operators, a final SPJ operator is
 * placed above which projects these fields out.
 *
 */
class RelOpDagBuilder {
protected:
  /// A handle on this relational query engine
  const RelQueryEngine *Engine;
  /// The projection list specified by the query
  XxxValueExprList *UserProjection;
  /// The list of all items that must be projected from the query inputs
  XxxValueExprList *InitProjection;  
  /// The group-by projection list
  XxxValueExprList *GroupByProjection; 
  /// All group-by expressions
  XxxValueExprList *Groupings; 
  /// All expressions referenced which contain an aggregate
  XxxValueExprList *Aggregates;       
  /// XXX_TRUE if a final SPJ is needed to project out intermediate fields
  XxxBool FinalSPJNeeded;
  /// The projection which must be applied at the final SPJ
  XxxValueExprList *FinalProjection;
  /// XXX_TRUE if the selection is distinct
  XxxBool DistinctSelection;
  /// The selection condition
  XxxBooleanExpression *Selection;
  /// All value expressions referenced in the WHERE clause
  XxxValueExprList *SelectionExprsRef;   
  /// The source relations specified in the FROM clause
  RelInfo **SourceRelations;
  /// The number of source relations
  int Arity;
  /// The having condition
  XxxBooleanExpression *GroupSelection;
  /// All value expressions referenced in the HAVING clause
  XxxValueExprList *GrpSelExprsRef;   
  /// The ordering expressions
  XxxValueExprList *Ordering;

  //ihab begin
  XxxBool KRJOp; // is this rank spj block or not. 
  int StopValue;  // the stop after parameter
  //ihab end

  //Begin Moustafa
  RelViewClass::WindowInfo* WindowValue;
  //End Moustafa
  /**
   * An array of booleans corresponding to the ordering expressions; 
   * XXX_TRUE if ascending
   */
  XxxBool *Ascending;
  /**
   * Takes the query view information and organizes it
   * @param View The view to process
   * @return XXX_OK if extraction worked.
   */
  XxxErrCode GetViewInfo(const RelViewClass *View);
  /**
   * Reorganizes the information, changing the "implied expressions" to 
   * explicit ones; e.g., performs intermediate projections and gathers 
   * aggregates.
   * @return XXX_OK if necessities were determined successfully.
   */
  XxxErrCode ReorganizeInfo();
  /**
   * Checks the semantics of the information extracted.  This is where all 
   * semantic SQL rules, such as "you cannot group by if no aggregates are 
   * present", are enforced.
   * @return XXX_OK if semantics were correct
   */
  XxxErrCode CheckSemantics();
  /**
   * The query may reference expressions that are not explicitly projected;
   * In this case, these expressions must be added to the initial projection
   * and a final projection must be formed to remove them later.
   */
  XxxErrCode AddImpliedProjections();
  const char AggContains(XxxValueExpression *Expr);
  /**
   * Prints the Query DAG to the output stream.
   */
  void PrintQueryDag(ostream &OStr) const;
  /**
   * Creates query nodes for each of the relational inputs.
   * @param FromNodes (output) An array of (arity) query nodes based on the 
   *                  input relations
   * @return XXX_OK if all from nodes could be constructed
   */
  XxxErrCode BuildFromNodes(RelQueryNode **&FromNodes);
  /**
   * Creates the initial SPJ node with the from nodes as inputs
   * @param FromNodes (input) An array of (arity) query nodes based on the 
   *                  input relations
   * @param SPJNode (output) The initial SPJ node
   * @return XXX_OK if no errors occurred
   */
  XxxErrCode AddInitialSPJNode(RelQueryNode *&SPJNode,
			       RelQueryNode **FromNodes);
  /**
   * Creates the group-by node over the current root, which is passed in.
   * @param Root (input) the current root (output) the new root
   * @return XXX_OK if no errors occurred
   */
  XxxErrCode AddGroupByNode(RelQueryNode *&Root);
  /**
   * Creates the order-by node over the current root, which is passed in.
   * @param Root (input) the current root (output) the new root
   * @return XXX_OK if no errors occurred
   */
  XxxErrCode AddOrderByNode(RelQueryNode *&Root);
  /**
   * Creates the final SPJ node over the current root, which is passed in.
   * @param Root (input) the current root (output) the new root
   * @return XXX_OK if no errors occurred
   */
  XxxErrCode AddFinalSPJNode(RelQueryNode *&Root);
  /**
   * Constructor.
   * @param Engine A handle on the relational engine
   */
 //ihab begin 
  // creating the Rankjoin Operators and the stop after operators nodes
  XxxErrCode AddInitialKRJNode(RelQueryNode *&KRJNode,RelQueryNode **FromNodes);
  XxxErrCode AddFinalKRJNode(RelQueryNode *&Root); 
  XxxErrCode AddStopAfterNode(RelQueryNode *&Root);
  //ihab end

 
public:
  RelOpDagBuilder (const RelQueryEngine *Engine);

  /**
   * Destructor.
   */
  ~RelOpDagBuilder ();

  /**
   * Builds a query DAG, given a query view.  This function calls all of the other
   * functions.
   * @param View The view to be transformed into a tree
   * @param Root The root of the newly contructed tree
   * @return XXX_OK if no errors occurred
   */
  XxxErrCode BuildQueryDag (RelViewClass *View,
			    RelQueryNode *&Root);
};

/**
 * The query engine is a repository for views, and also performs semantic checking
 * for queries before optimization.  It talks directly to the parser, which gives it 
 * views as its input, and produces adorned query structures as outputs which can be 
 * fed into the optimizer.
 */
class RelQueryEngine {
  friend class RelOpDagBuilder;
  /// a handle on the relation engine
  XxxRelationEngine* RelEng;

public:  
  /**
   * Constructor.
   * @param SrcEng The Relation engine to which this engine belongs
   */
  RelQueryEngine(XxxRelationEngine* SrcEng);
  /**
   * Destructor.
   */
  ~RelQueryEngine();
  /**
   * Stores a view in this query engine.
   * @param ViewName The name of the view to store
   * @param View The View data structure to store
   * @return XXX_OK if no errors occured
   */
  XxxErrCode StoreView(const char* ViewName, RelViewClass* View);
  /**
   * Deletes a view in this query engine.
   * @param ViewName The name of the view to store
   * @return XXX_OK if no errors occured
   */
  XxxErrCode DropView(const char* ViewName);
  /**
   * Retrieves a view from this query engine.
   * @param ViewName The name of the view to retrieve
   * @param View The View data structure retrieved
   * @return XXX_OK if no errors occured
   */
  XxxErrCode RetrieveView(const char* ViewName, 
			  RelViewClass*& View, XxxBool& Found) const;
  /**
   * Takes a view and produces a query from it.  This function transforms the
   * view into a DAG, performs semantic checking, propagates all meta info,
   * performed heuristic rewrite, etc.
   * @param View The view to transform
   * @param Query (output) the query generated
   * @param Env The query environment (optional)
   * @return XXX_OK if no errors occured
   */
  XxxErrCode CheckQuery(RelViewClass* View, RelQueryClass*& Query,
			const XxxArgEnv* Env = NULL) const;
  /**
   * Creates an operator DAG for a query view.  
   * Uses a RelOpDagBuilder which actually performs the translation.
   * @param View The view to be transformed into a tree
   * @param Root The root of the newly contructed tree
   */
  XxxErrCode ViewToOpDag(RelViewClass *View, RelQueryNode *&Root) const;
  /**
   * Creates an operator DAG for a query view.  
   * Uses a RelOpDagBuilder which actually performs the translation.
   * @param View The view to be transformed into a tree
   * @param Root The root of the newly contructed tree
   */
  XxxErrCode CheckOpDag(RelQueryNode *Root, const XxxArgEnv* Env,
			RelQueryClass*& Query) const;
  /**
   * Prints all views stored in this query engine
   * @param OStr The stream to which the views should be printed
   */
  XxxErrCode PrintViews(ostream& OStr) const;
};

#endif
