#include "relrewrite.h"
#include "relops.h"
#include "relenv.h"
#include "expr.h"
#include "values.h"
#include "funcexpr.h"

// default constructor for the RelRewriteEngine
RelRewriteEngine::RelRewriteEngine(int MaxRules) :MAXRULES(MaxRules)
{
 numUpRules = 0;
 numDownRules = 0;
 numRulesFired = 0;
 upPhaseRules = new RelRewriteRule*[MAXRULES];
 downPhaseRules = new RelRewriteRule*[MAXRULES];

 // up phase rules
 upPhaseRules[numUpRules++] = new RelSelectMerge;

 // down phase rules
 downPhaseRules[numDownRules++] = new RelDistinctPushDown;
 // downPhaseRules[numDownRules++] = new RelAggrFuncPushDown;
 downPhaseRules[numDownRules++] = new RelElimUnrefTargets;

 //  const int NumUp=2, NumDown=3;
 //Rule *UpRules = new Rule[NumUp];
 //UpRules[0] = GRPBYPULLUP;
 //UpRules[1] = SELMERGE;

 //Rule *DownRules = new Rule[NumDown];
 //DownRules[0] = DISTPDFR;
 //DownRules[1] = DISTPDTO;
 //DownRules[2] = E2F;
}

// PURPOSE: destructor for the RelRewriteEngine
RelRewriteEngine::~RelRewriteEngine() 
{
 // cleanup up phase rules
 if (numUpRules > 0) {
   for (int i=0; i<numUpRules; i++)
     delete upPhaseRules[i];
   delete [] upPhaseRules;
  }
 // cleanup down phase rules
 if (numDownRules > 0) {
   for (int i=0; i<numDownRules; i++)
     delete downPhaseRules[i];
   delete [] downPhaseRules;
  }

  if (XXX_THISRELENV()->DebugQrw) {
    cerr << "Total Number of Rewrite Rules Fired = " << numRulesFired << endl;
  }
}

// apply upPhase or downPhase Rules on a particular RelQueryNode
XxxErrCode RelRewriteEngine::executeRules(RelRewriteRule *Rules[], int NumRules, 
					  RelQueryNode *InNode, 
					  RelQueryNode *&OutNode) 
{
 XxxBool Success;
 OutNode = InNode;
 for (int i=0; i < NumRules; i++) {
   if (Rules[i]->ApplyRule(InNode, Success, OutNode) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADOPT, "Error applying rewrite rule.");
     return XXX_NOT_OK;
    }
   if (Success) {
     if (XXX_THISRELENV()->DebugQrw) {
       cerr << "Rule successfully fired!" << endl;
     }
     numRulesFired++;
     InNode = OutNode;
    } 
   else {
     if (XXX_THISRELENV()->DebugQrw) {
       cerr << "Rule unsuccessful" << endl;
     }
    }
  } // for each rule
 return XXX_OK;
}

// The main rewrite function.  Clients can only call this.
// It performs exactly one downward and one upward pass.
XxxErrCode RelRewriteEngine::Rewrite(RelQueryNode *InNode,
				     RelQueryNode *&OutNode) 
{
 if (InNode == NULL) {
   REL_NOTE_ERROR(REL_BADQUERY, "Rewrite: Received NULL argument");
   return XXX_NOT_OK;
  }

 OutNode = InNode;
 
 if (InNode->Op == NULL) {
   // a base table.
   // we don't have any rules for these
   return XXX_OK;
  }
 
 // first apply the down phase rules
 if (executeRules(downPhaseRules, numDownRules, InNode, OutNode) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADSERVER, "Error executing down phase rules.");
   return XXX_NOT_OK;
  }
 // Now recursively apply Rewrite to the children
 for (int i=0; i<OutNode->Arity(); i++) {
   RelQueryNode *InChild = (RelQueryNode *)OutNode->GetChild(i);
   RelQueryNode *OutChild;
   if (Rewrite(InChild, OutChild) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADSERVER, "Error applying rewrite rules on children");
     return XXX_NOT_OK;
    }
   if (InChild != OutChild) {
     XxxQueryNode *OldChild;
     OutNode->ReplaceChild(i, OutChild, OldChild);
    }
  }

 InNode = OutNode;

 // and now apply the up phase Rules
 if (executeRules(upPhaseRules, numUpRules, InNode, OutNode) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADSERVER, "Error executing up phase rules.");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

// 
// The Select Merge Rule
// Our main merging rule.  It takes a node and tries to merge each 
// of its children into itself.
// 
XxxErrCode RelSelectMerge::ApplyRule(RelQueryNode *InNode, XxxBool &Success,
				     RelQueryNode *&OutNode) 
{
 Success = XXX_FALSE;

  if (XXX_THISRELENV()->DebugQrw) {
    cerr << "Trying to apply the Select Merge Rule" << endl;
  }

 // check the basics: Node must be a RelSPJ node and have children.
 if ((InNode == NULL) || (InNode->Arity() == 0) || (InNode->Op == NULL) || 
     (InNode->Op->Type() != REL_SPJ))
   return XXX_OK;

 RelSPJOp *ParentOp = (RelSPJOp *)InNode->Op;

 for (int i=0; i<InNode->Arity(); i++) {
   RelQueryNode *Child=(RelQueryNode*)InNode->GetChild(i);
   // if Child is not a SPJ node, we can't do anything with it, so go on
   // to the next child
   if ((Child->Op == NULL) || (Child->Op->Type() != REL_SPJ)) {
     continue;
    }

   RelSPJOp *ChildOp = (RelSPJOp *)Child->Op;
   // see if this pair of nodes matches the criteria
   if ((ParentOp->Distinct == XXX_TRUE) || 
       (ParentOp->DistinctCond == QR_PERMIT) ||
       (ChildOp->DistinctCond != QR_ENFORCE)) {
     // ok, go ahead and merge
     if ((ChildOp->DistinctCond == QR_ENFORCE) && 
	 (ParentOp->DistinctCond != QR_PERMIT))
       ParentOp->DistinctCond = QR_ENFORCE;

     if (merge(InNode, Child) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error merging query blocks");
       return XXX_NOT_OK; 
      }
     Success = XXX_TRUE;
     OutNode = InNode;
    } // if okay to merge
  } // for each child
 return XXX_OK;
}

// Merges a parent and a child
XxxErrCode RelSelectMerge::merge(RelQueryNode *Parent, RelQueryNode *Child) const 
{
 // First, walk through select clause of subquery and update the SrcChild
 // fields of any UnknownValue expressions.
 RelSPJOp *ChildOp=(RelSPJOp*)Child->Op;
 XxxValueExprList* iter=ChildOp->TargetList;
 
 while (iter!=NULL) {
   if (updateSrcChildren(iter->Expr, Child->WhichChildAmI(), 
			 Child->Arity(), Parent->Arity()) != XXX_OK)
     return XXX_NOT_OK;
   iter=iter->Next;
  }

 // walk through Child's where clause, and update the SrcChild fields of
 // any UnknownValue expressions.
 if ((ChildOp->Condition != NULL) &&
     (updateSrcChildren(ChildOp->Condition,Child->WhichChildAmI(),
			Child->Arity(), Parent->Arity()) != XXX_OK)) {
   REL_NOTE_ERROR(REL_BADQUERY, "SelectMerge::merge: WHERE clause of subquery can't be updated for an unknown reason.");
   return XXX_NOT_OK;
  }

 // walk through the select clause of the parent and replace any aliases
 // with real name references (looking at the child query).
 RelSPJOp *ParOp = (RelSPJOp *)Parent->Op;
 iter = ParOp->TargetList;
 while (iter!=NULL) {
   XxxValueExpression *ValExpr = iter->Expr;
   if ((iter->Expr = 
	(XxxValueExpression *)searchAndReplace(ValExpr, Child)) == NULL)
     return XXX_NOT_OK;
   iter = iter->Next;
  }
 
 // walk through the where clause of the parent and replace any aliasees
 // with real name references (looking at the child query).
 XxxBooleanExpression *ParWhereClause = ParOp->Condition;
 if ((ParWhereClause != NULL) && 
     ((ParOp->Condition = 
       (XxxBooleanExpression *)searchAndReplace(ParWhereClause, 
						Child)) == NULL))
   return XXX_NOT_OK;
 
 // bring Child's from clause into parent's from clause.
 // The first element of the child's from clause will take the place of
 // the nested query in parent.  The other elements will be added to the end
 // of the Parent->children array.  Don't replace the nested query until
 // the very end (otherwise, we won't be able to access the nested query
 // again).
 
 if (Child->Arity()>1) {
   RelQueryNode **AddOns = new RelQueryNode*[Child->Arity()-1];

   for (int i=1; i<Child->Arity(); i++) {
     AddOns[i-1] = (RelQueryNode*)Child->GetChild(i);
     AddOns[i-1]->IncrRef();
    }
   Parent->AddChildren(AddOns, Child->Arity()-1);
  }
 // Bring the child WHERE clause up.
 // Make a copy of the child where clause.
 if (ChildOp->Condition != NULL) {
   XxxBooleanExpression *ChildWhereCopy=
     (XxxBooleanExpression*)ChildOp->Condition->Replicate();
   if (ParOp->Condition != NULL) {
     XxxExpression **WhereChildren = new XxxExpression*[2];
     WhereChildren[0] = ParOp->Condition;
     WhereChildren[1] = ChildWhereCopy;
     XxxBoolBooleanExpression *NewWhere =
       new XxxBoolBooleanExpression(XXX_AND_OP, 2, WhereChildren);
     ParOp->Condition = NewWhere;
    } else {
      ParOp->Condition = ChildWhereCopy;
     }
  }
 RelQueryNode *FirstGrandChild = (RelQueryNode*)Child->GetChild(0);
 FirstGrandChild->SetParent(Parent);
 FirstGrandChild->IncrRef();
 Parent->SetChild(Child->WhichChildAmI(), (RelQueryNode*)Child->GetChild(0));

 // now, we have the one added problem that each expression keeps track
 // of a "dependsArity" that is a count of the number of children in
 // the query block (this is a nasty part of the code design, and we
 // should eventually take it out)// TODO
 // So, we need to make sure that all of these expressions get the
 // correct arity, now that we have merged the blocks
 if (Parent->Op) {
   RelQueryNode **Children = new RelQueryNode*[Parent->Arity()];
   for (int i=0; i<Parent->Arity(); i++) {
     Children[i] = (RelQueryNode*)(Parent->GetChild(i));
    }
   if (Parent->Op->ResolveReferences(Parent->Arity(), 
				     Children, 0) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADQUERY, "Error resolving references after merge");
     return XXX_NOT_OK;     
   }
   delete [] Children;
 } // if (Parent->Op)
 return XXX_OK;
}

// Modifies the SrcChild fields of the Child's 
// XxxUnknownValue Expressions (TargetList or WhereClause)
// NOTES: Used by SelectMerge::merge
//        Some BackGround: The first child of the child node (which is merged) 
//        is put in place of that child, i.e. the first grandchild takes it 
//        parent's place in the grandparent's from list.  All the other 
//        grandchildren go at the end of the grandparent FROM list.
XxxErrCode RelSelectMerge::updateSrcChildren(XxxExpression *Exp,
					     int ChildRelPos, 
					     int ChildArity,
					     int ParentArity) const 
{
 // If this expression is a leaf, check if it's an XxxUnknownValue and if so,
 // update it's SrcChild field.
 if (Exp->Arity() == 0) {
   XxxValueExpression *ValExpr = (XxxValueExpression *)Exp;
   if (ValExpr->ValType() != XxxValueExpression::COLUMN_VAR) {
     // maybe its a func-expression
     if (ValExpr->ValType() == XxxValueExpression::FUNC_EXPR) {
       XxxFuncExpression *FuncExpr = (XxxFuncExpression *)ValExpr;
       if (FuncExpr->Owner && updateSrcChildren(FuncExpr->Owner, ChildRelPos, ChildArity, ParentArity) 
	   != XXX_OK)
	 return XXX_NOT_OK;
       // now do it for the arguments
       XxxValueExprList *ArgList = FuncExpr->ArgList;
       if (ArgList) {
	 int Length = ArgList->Count();
	 for (int i=0; i<Length; i++) {
	   if (updateSrcChildren(ArgList->Expr, ChildRelPos, ChildArity, ParentArity) 
	       != XXX_OK)
	     return XXX_NOT_OK;
	 }
       }
       return XXX_OK;
     } 
     else {
       // not a func expression
       return XXX_OK;
     }
   }
   // this is an UnknownValue, so update its SrcChild.
   XxxUnknownValue *UnknownExp=(XxxUnknownValue*)Exp;
   if (UnknownExp->SrcChild==0)
     UnknownExp->SrcChild=ChildRelPos;
   else 
     UnknownExp->SrcChild+=(ParentArity-1);
   return XXX_OK;
  } // if expression with arity 0
 for (int i=0; i<Exp->Arity(); i++) {
   updateSrcChildren(Exp->ChildExpr(i), ChildRelPos, ChildArity, ParentArity);
  }
 return XXX_OK;
}

// go through the parent expressions, and modify references to the
// child's expressions
// Used by SelectMerge::merge
XxxExpression *RelSelectMerge::searchAndReplace(XxxExpression *OuterExpr,
						RelQueryNode *Node) const 
{
 if (OuterExpr->Arity() > 0) {
   for (int i=0; i<OuterExpr->Arity() ; i++) {
     XxxExpression *NewChildExpr;
     if ((NewChildExpr = searchAndReplace(OuterExpr->ChildExpr(i), Node)) 
	 == NULL)
       return NULL;
     // only change child if it has actually changed
     if (NewChildExpr != OuterExpr->ChildExpr(i))
       OuterExpr->SetChild(i, NewChildExpr);
    } // for
   return OuterExpr;
  } // if OuterEpr->Arity() > 0
 else {
   // we've hit an expression with no children.  Check if this is a leaf and
   // is an UnknownValueExpression whose SrcChild field is the same as the
   // "Node"'s "childnumber" field.
   
   XxxValueExpression *OuterValExpr = (XxxValueExpression *)OuterExpr;
   if (OuterValExpr->ValType() != XxxValueExpression::COLUMN_VAR) {
     // maybe its a func-expression
     if (OuterValExpr->ValType() == XxxValueExpression::FUNC_EXPR) {
       XxxFuncExpression *OuterFuncExpr = (XxxFuncExpression *)OuterValExpr;
       XxxValueExpression *NewOwner = 
	 (XxxValueExpression *)searchAndReplace(OuterFuncExpr->Owner, 
						Node);
       if(NewOwner == NULL)
	 return NULL;
       if(NewOwner != OuterFuncExpr->Owner) {
	 delete OuterFuncExpr->Owner;
	 OuterFuncExpr->Owner = NewOwner;
	}
       // now do it for the arguments
       XxxValueExprList *ArgList = OuterFuncExpr->ArgList;
       if (ArgList) {
	 int Length = ArgList->Count();
	 for (int j=0; j<Length; j++) {
	   XxxValueExpression *NewArgExpr = 
	     (XxxValueExpression *)searchAndReplace(ArgList->Expr, 
						    Node);
	   if (NewArgExpr == NULL)
	     return NULL;
	   if (NewArgExpr != ArgList->Expr) {
	     delete ArgList->Expr;
	     ArgList->Expr = NewArgExpr;
	    }	    
	  }
	}
       return OuterFuncExpr;
      } 
     else {
       // not a column var or a func expression
       return OuterValExpr;
      }
    } // if not a column var
   
   // this expr is definitely a column variable leaf.
   XxxUnknownValue *OuterUnknown = (XxxUnknownValue *)OuterValExpr;
   if (Node->WhichChildAmI() != OuterUnknown->SrcChild) {
     // doesn't match, so we don't change anything and go to the next
     // ChildExpr.
     return OuterUnknown;
    }
   // we are now at a point where we should go to the node's select clause's
   // SrcAttrIndex'th expression that will replace this child.
   XxxValueExprList *InnerSelList=((RelSPJOp*)Node->Op)->TargetList;
   for (int j=0; j<OuterUnknown->SrcAttrIndex; j++)
     InnerSelList = InnerSelList->Next;
   XxxExpression *Copy = InnerSelList->Expr->Replicate();
   return Copy;
  }
 // we should never be here
 cerr << "SelectMerge::searchAndReplace:  we should never be here" << endl;
 return NULL;
}

//
// The Aggregate Function PushDown Rule
//

// application of the rule to a candidate node
XxxErrCode RelAggrFuncPushDown::ApplyRule(RelQueryNode *InNode, 
					  XxxBool &Success,
					  RelQueryNode *&OutNode) 
{
  Success = XXX_FALSE; // was the rule applicable?

  if (XXX_THISRELENV()->DebugQrw) {
    cerr << "Trying to apply the AggrFuncPushDown Rule" << endl;
  }

  // check the basics: Node must be a RelSPJ node and have children.
  if ((InNode == NULL) || (InNode->Arity() == 0) || (InNode->Op == NULL) || 
      (InNode->Op->Type() != REL_SPJ))
    return XXX_OK; // with Success == XXX_FALSE
  
  // loop over children, ignoring nodes that are not GBY operators
  for (int i=0; i<InNode->Arity(); i++) {
    RelQueryNode *Child=(RelQueryNode*)InNode->GetChild(i);
    if ((Child->Op == NULL) || (Child->Op->Type() != REL_GBY)) {
      continue;
    }
   // check if there are subexpressions of targets of InNode, that
   // can be pushed down to this Child.
   XxxBool ChildSuccess = XXX_FALSE;
   if( pushDownAggrMethods(InNode, Child, ChildSuccess) != XXX_OK ) {
     REL_NOTE_ERROR(REL_BADOPT, "Error merging query blocks");
     return XXX_NOT_OK; 
   }
   // if this was successful, the whole rewrite was. 
   // This child's and the parent's CatalogInfo, have to be recomputed.
   if(ChildSuccess == XXX_TRUE) {
     Success = XXX_TRUE; 
     Child->CatalogInfo = NULL;
   }
 }
 // Recursively update the CatalogInfo of all modified children 
 // and the parent.
 if(Success == XXX_TRUE) {
   InNode->CatalogInfo = NULL;
   InNode->PropagateMetaInfoUp(NULL); // recurses to children with NULL info
 }   
 return XXX_OK;
}


// initialize the generator for unique id strings (for column references)
int RelAggrFuncPushDown::GeneratedIDString::number = 1;


/* pushes down all applications of methods on aggregates from an SPJ down
 * to a child GrpBy node. 
 * IMPORTANT: 
 * Parent is assumed to have a SPJ operator and child a GBY operator
 */
XxxErrCode 
RelAggrFuncPushDown::pushDownAggrMethods(RelQueryNode *Parent, 
					 RelQueryNode *Child, 
					 XxxBool &Success) const
{
  Success = XXX_FALSE;
  RelGrpByOp* ChildOp = (RelGrpByOp*)Child->Op;

  // list of subexpressions that we want to push down
  ValExprRefList* targets = NULL;
  ValExprRefList** nextTargetPtr = &targets;

  // for each target of the parent, append its interesting subexpressions
  // to targets
  XxxValueExprList* Tmp = ((RelSPJOp*)Parent->Op)->TargetList;
  while(Tmp) {
    (*nextTargetPtr) = 
      findAggrMethodPushDown(&Tmp->Expr, NULL, Child->WhichChildAmI());
    while(*nextTargetPtr) 
      nextTargetPtr = &(*nextTargetPtr)->Next;
    Tmp = Tmp->Next;
  }

  // generate the necessary new targets, aggregates, and grouping attributes 
  // for the child
  int ChildAttrNumber = ChildOp->TargetList->Count();
  // identify the ends of the existing lists
  XxxValueExprList** TargetListEnd = &ChildOp->TargetList;
  XxxValueExprList** AggrListEnd = &ChildOp->AggrList;
  XxxValueExprList** GroupListEnd = &ChildOp->GroupList;
  while(*TargetListEnd)
    TargetListEnd = &(*TargetListEnd)->Next;
  while(*AggrListEnd)
    AggrListEnd = &(*AggrListEnd)->Next;
  while(*GroupListEnd)
    GroupListEnd = &(*GroupListEnd)->Next;

  // walk through the list of new targets
  ValExprRefList* currentTarget = targets;
  while(currentTarget) {
    XxxValueExpression* exp = *currentTarget->Ptr;
    // prepare the parent target subexpression for being moved to the
    // childs targets, by substituting all column references
    if(substituteColumnValues(exp, ChildOp->TargetList) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_INTERNAL, "Error substituting column variables");
      return XXX_NOT_OK; 
    }
    // add the new target
    (*TargetListEnd) = new XxxValueExprList(exp);
    TargetListEnd = &(*TargetListEnd)->Next;
    if(exp->ReferencesAggregate() == XXX_TRUE) {
      (*AggrListEnd) = new XxxValueExprList(exp->Replicate());
      AggrListEnd = &(*AggrListEnd)->Next;
      /*   } else {
      (*GroupListEnd) = new XxxValueExprList(exp->Replicate());
      GroupListEnd = &(*GroupListEnd)->Next;*/
    }
    // now substitute a reference to the new child target for the old
    // parent target
    char *ColumnName = GeneratedIDString::NewIDString();
    XxxUnknownValue* newUV = 
      new XxxUnknownValue(ColumnName, Child->BasicName(),
			  (*currentTarget->Ptr)->BaseType,
			  ChildAttrNumber,
			  Child->WhichChildAmI());
    delete[] ColumnName;
    *currentTarget->Ptr = newUV;

    currentTarget = currentTarget->Next;
    ChildAttrNumber++;
  }
  // if any targets where pushed down, the rule was applicable
  if(targets)
    Success = XXX_TRUE;
  return XXX_OK;
}


// exp is a subexpression of one of the parents targets, that is added to
// the child's TargetList. Its 'unknown' references to that child have to
// be substituted by the referenced target expressions of the child.
// ALL unknown references are assumed to refer to the owner of the 
// targetlist.
XxxErrCode 
RelAggrFuncPushDown::substituteColumnValues(XxxValueExpression* &Exp,
					    XxxValueExprList* TargetList) const
{
  // Unfortunately Replace does not work on Unknowns themselves,
  // we have to do that by hand:
  if(Exp->ValType() == XxxValueExpression::COLUMN_VAR) {
    if(TargetList-> // get the corresponding target expression
       GetElement(((XxxUnknownValue*)Exp)->SrcAttrIndex, Exp)
       != XXX_OK)
      return XXX_NOT_OK;
    return XXX_OK;
  }	
 // Exp is itself not a column reference
 XxxValueExprList *Subexpressions = Exp->ValExprsReferenced();
  XxxValueExprList *Unknowns =  // all the subexpressions that are column refs
    Subexpressions->TypeFilter(XxxValueExpression::COLUMN_VAR);
  while(Unknowns) { // for each one substitute the corresponding target
    XxxValueExpression* Target;
    if(TargetList->
       GetElement(((XxxUnknownValue*)Unknowns->Expr)->SrcAttrIndex, Target)
       != XXX_OK) // returns the original, not a replication!
      return XXX_NOT_OK;
    Exp->Replace((XxxUnknownValue*)Unknowns->Expr, Target); // replicates Target
    XxxValueExprList *Deathbound = Unknowns; // Unknowns are just replica
    Unknowns = Unknowns->Next;
    Deathbound->Next = NULL;
    Deathbound->Destroy(); // destroys the replica of the substituted unknown
  }
  return XXX_OK;
}
  
// identify all subexpressions of exp that can be pushed down to childNumber.
// Recursive in exp. ancestor is the highest parent of exp int the expression
// tree that could be in such a subexpression. 
// The function returns the ancestor (as a list) if it can be pushed down 
// completely, otherwise it returns a list of its own pushable subexpressions.
RelAggrFuncPushDown::ValExprRefList* 
RelAggrFuncPushDown::findAggrMethodPushDown(XxxValueExpression **Exp,
					    XxxValueExpression **ancestor,
					    int childNumber) const
{
  if((*Exp)->ValType() == XxxValueExpression::COLUMN_VAR) {
    if(((XxxUnknownValue*)*Exp)->SrcChild == childNumber)
      if (ancestor)
	return new ValExprRefList(ancestor);
      else
	// empty list, we do not want to push a single unknown
	return NULL;
    else 
      // return empty list
      return NULL;
  }
    
  XxxValueExprList* DirectSubexpressions = (*Exp)->DirectSubexpressions();
  // do all subexpressions allow this expression to be pushed down?
  XxxBool allAlike = XXX_TRUE;
  if (!ancestor && (*Exp)->ValType() == XxxValueExpression::FUNC_EXPR)
    ancestor = Exp; // this exp does it as a ancestor for its subexpressions
  else if ((*Exp)->ValType() == XxxValueExpression::CAST_EXPR ||
	   (*Exp)->ValType() == XxxValueExpression::BOOL_EXPR ||
	   (*Exp)->ValType() == XxxValueExpression::ARITH_EXPR) {
    ancestor = NULL;
    allAlike = XXX_FALSE;
  }
  int NumDSubexp = (DirectSubexpressions?DirectSubexpressions->Count():0);
  // collect the results of the children in an array
  ValExprRefList** ResultLists = new ValExprRefList*[NumDSubexp];
  XxxValueExprList* tmp = DirectSubexpressions;
  for(int i=0; i<NumDSubexp; i++) {
    ResultLists[i] = 
      findAggrMethodPushDown(&tmp->Expr, ancestor, childNumber);
    // check if all subexpressions so far returned the ancestor
    if(!ResultLists[i] || ResultLists[i]->Ptr != ancestor) 
      allAlike = XXX_FALSE;
    tmp = tmp->Next;
  }
  // is the ancestor pushable?
  if ( allAlike == XXX_TRUE ) {
    for(int i=0; i<NumDSubexp; i++)
      delete ResultLists[i];
    delete[] ResultLists;
    return new ValExprRefList(ancestor); // here: ancestor != NULL 
  } else {
    // otherwise concatenate the results of the subexpressions
    ValExprRefList* ResultList = NULL;
    ValExprRefList** EndOfResultList = &ResultList;
    tmp = ((XxxFuncExpression*)Exp)->ArgList;
    for(int i=0; i < NumDSubexp; i++) {
      if(ResultLists[i] != NULL) {
	if(ResultLists[i]->Ptr == ancestor) {
	  // now the ancestor can't be pushed, so take the subexpression
	  *EndOfResultList = new ValExprRefList(&tmp->Expr);
	  delete ResultLists[i];
	} else {
	  // ..., or its subexpressions
	  *EndOfResultList = ResultLists[i];
	}
	while(*EndOfResultList) 
	  EndOfResultList = &(*EndOfResultList)->Next;
      }	  
      tmp=tmp->Next;
    }
    delete[] ResultLists;
    return ResultList;
  }
}


// 
// The Eliminate Unreferenced Targets Rule
// 
// PURPOSE: used to eliminate targets and aggregates
//          that are not referenced in SPJ nodes above, a top-down rule
//          This rule basically cleans up after the AggrFuncPushDown rule

XxxErrCode RelElimUnrefTargets::ApplyRule(RelQueryNode *InNode, 
					  XxxBool &Success, 
					  RelQueryNode *&OutNode) 
{
 Success = XXX_FALSE;
 if (XXX_THISRELENV()->DebugQrw) {
   cerr << "Trying to apply the Eliminate Unreferenced Targets Rule" << endl;
 } 
 if ((InNode == NULL) || (InNode->Op->Type() != REL_SPJ))
   return XXX_OK;
 RelSPJOp *ParentSPJOp = (RelSPJOp *)InNode->Op;
 for(int i=0; i < InNode->Arity(); i++) { // loop over the children
   RelQueryNode *Child = (RelQueryNode *)InNode->GetChild(i);
   XxxBool ChildSuccess = XXX_FALSE; // needed to clean the child's CatalogInfo
   if(Child->Op &&  
      (Child->Op->Type() == REL_SPJ || Child->Op->Type() == REL_GBY)) {
     // not for REL_OBY, because order might depend on unreferenced targets
     int OldTargetNumber = 0;  // as we walk through the original targetlist
     int NewTargetNumber = 0;  // as above, but not counting eliminated targets
     XxxValueExprList** TargetListPtr; // points to the child's 'TargetList'
     if(Child->Op->Type() == REL_SPJ)
       TargetListPtr = &((RelSPJOp*)Child->Op)->TargetList;
     else // Child->Op == REL_GBY
       TargetListPtr = &((RelGrpByOp*)Child->Op)->TargetList;
     XxxValueExprList* current = *TargetListPtr;
     XxxValueExprList* last = NULL;
     while(current) { // for each of the child's targets
       XxxBool appearsInParent = XXX_FALSE;
       if(ParentSPJOp->Condition &&
	  ParentSPJOp->Condition->UpdateUnknowns(i, // # of the child
						 OldTargetNumber,
						 NewTargetNumber, 
						 appearsInParent) != XXX_OK) {
	 XXX_NOTE_ERROR(XXX_INTERNAL, "Error updating the parent's condition");
	 return XXX_NOT_OK;
       }
       if(ParentSPJOp->TargetList &&
	  ParentSPJOp->TargetList->UpdateUnknowns(i,
						 OldTargetNumber,
						 NewTargetNumber, 
						 appearsInParent) != XXX_OK) {
	 XXX_NOTE_ERROR(XXX_INTERNAL, "Error updating the parent's targets");
	 return XXX_NOT_OK;
       }
       if(appearsInParent == XXX_TRUE) { // we will preserve this target
	 OldTargetNumber++;
	 NewTargetNumber++; 
	 last = current;
	 current = current->Next;
       } else { // eliminate this target
	 OldTargetNumber++; // NewTargetNumber is not increased in this case
	 Success = ChildSuccess = XXX_TRUE;
	 // take care of corresponding, now redundant, aggregates
	 if(Child->Op->Type() == REL_GBY && 
	    (((RelGrpByOp*)Child->Op)->AggrList ?
	     ((RelGrpByOp*)Child->Op)->AggrList->Count() : 0) > 1) {
	   // >0 : A GBY node must have at least one aggregate reference!
	   // GroupList members do not become redundant!
	   deleteFirstExprFromList(&(((RelGrpByOp*)Child->Op)->AggrList),
				   current->Expr);
	 } // now take care of the target itself 
	 if(last) {
	   last->Next = current->Next;
	   delete current->Expr;
	   delete current;
	   current = last->Next;
	 } else { 
	   *TargetListPtr = current->Next;
	   delete current->Expr;
	   delete current;
	   current = (*TargetListPtr);
	 }
       }
     }
     if(ChildSuccess == XXX_TRUE) 
       Child->CatalogInfo = NULL; // will be regenerated in the end
   }
 }
 // Recursively update the CatalogInfo of all modified children 
 // and the parent.
 if(Success == XXX_TRUE) {
   InNode->CatalogInfo = NULL;
   InNode->PropagateMetaInfoUp(NULL); // recurses to children with NULL info
 }
 return XXX_OK;
}


XxxErrCode RelElimUnrefTargets::
deleteFirstExprFromList(XxxValueExprList** ExprListPtr, 
			XxxValueExpression* Expr)
{
  // we delete at most one matching expression from each list
  XxxValueExprList *ExprList = *ExprListPtr;
  XxxValueExprList *Last = NULL;
  while(ExprList && ExprList->Expr->Match(Expr) != XXX_TRUE) {
    Last = ExprList;
    ExprList = ExprList->Next;
  }
  if(ExprList) {
    if(Last) {
      delete ExprList->Expr;
      Last->Next = ExprList->Next;
      delete ExprList; // only this element
    } else { // delete the first element
      delete (*ExprListPtr)->Expr;
      *ExprListPtr = ExprList->Next;
      delete ExprList;
    }
  }
  return XXX_OK; // always (?)
}

// 
// The Distinct Pushdown Rule
// 
// PURPOSE: used to push the distinct attribute down to the children 
XxxErrCode RelDistinctPushDown::ApplyRule(RelQueryNode *InNode, XxxBool &Success, 
					  RelQueryNode *&OutNode) 
{
 Success = XXX_FALSE;
 
  if (XXX_THISRELENV()->DebugQrw) {
    cerr << "Trying to apply the Distinct Push Down Rule" << endl;
  }
 
 if ((InNode == NULL) || (InNode->Op == NULL))
   return XXX_OK;
 // currently we only deal with select boxes, but this rule can just
 // as well be applied to union, intersect, except
 if (InNode->Op->Type() == REL_SPJ) {
   RelSPJOp *ParentSPJOp = (RelSPJOp *)InNode->Op;
   RewriteDupCond ParentDistCond= ParentSPJOp->DistinctCond;
   
   // see if this node matches our criteria
   if((ParentDistCond == QR_PERMIT) || (ParentDistCond == QR_ENFORCE)) {
     int NumChildren = InNode->Arity();
     
     for(int i=0; i < NumChildren; i++) {
       RelQueryNode *Child = (RelQueryNode *)InNode->GetChild(i);
       if((Child->Op == NULL) || (Child->Op->Type() != REL_SPJ)) {
	 continue;
	}
       RelSPJOp *ChildOp = (RelSPJOp *)Child->Op;
       
       // we want to know if the rule actually made any changes
       if(ChildOp->DistinctCond != QR_PERMIT) {
	 ChildOp->DistinctCond = QR_PERMIT;
	 Success = XXX_TRUE;
	}
      }
    }
  }
 return XXX_OK;
}

/******************* The rest may be added in later releases...
  
// the groupby pullup rule
// Node is the SPJ node above the groupby node.
// We first see if a groupby pullup is valid, if so we pull it above the 
// SPJ node
XxxBool RelRewriteEngine::grpByPullUp(RelQueryNode *Node) const {

  // do some initial checking
  if(Node == NULL) {
    char buf[100];
    sprintf(buf, 
            "RelRewriteEngine::grpByPullUp:  Received bad argument.");
    REL_NOTE_ERROR(REL_BADQUERY, buf);
    return XXX_FALSE;
  }

  if(Node->Op == NULL) {
    // a base table; we can't do anything with it
    return XXX_FALSE;
  }
   
  // we only apply this rule on a parent SPJ node
  if(Node->Op->Type() == REL_SPJ) {

    RelSPJOp *ParentSPJOp = (RelSPJOp *)Node->Op;

    // try to apply this rule for any children that are group-by's
    for(int i=0; i<Node->Arity(); i++) {
      RelQueryNode *Child = (RelQueryNode *)Node->GetChild(i);
      if(Child->Op) {  // is this a base table or an operator node
	if(Child->Op->Type() == REL_GBY) {
	  // found a child which is a group-by
	  RelGrpByOp *ChildGrpByOp = (RelGrpByOp *)Child->Op;
	  
	  // see if the groupby pullup rule is applicable
	  XxxBool Valid = isGrpByPullUpApplicable(ParentSPJOp, 
						  ChildGrpByOp, i);

	  if(Valid == XXX_FALSE) {
	    cout << "GrpByPullUp Not Applicable" << endl;
	    return XXX_FALSE;
	  }
	  
	  // perform groupby pullup

	  cout << "GroupBy PullUp can be applied, but we just ain't got the code for it." << endl;
	  return XXX_TRUE;

	  // first traverse the parent's target list updating the SrcChild,
	  // SrcAttrIndex fields.  We know that these are just output

	  XxxValueExprList *ParentOutputList = ParentSPJOp->TargetList;	  
	  int ParentOutputSize = ParentOutputList->Count();

	  int SrcChild = i;
	  
	  for(int j=0; j<ParentOutputSize; j++, 
		ParentOutputList = ParentOutputList->Next) {
	    XxxValueExpression *ParentExpr = ParentOutputList->Expr;
	    if(ParentExpr->ValType() == XxxValueExpression::COLUMN_VAR) {
	      // a column variable
	      XxxUnknownValue *ParentUnknownExpr = 
		(XxxUnknownValue *)ParentExpr;
	      int ParentSrcChild = ParentUnknownExpr->SrcChild;
	      int ParentSrcAttrIndex = ParentUnknownExpr->SrcAttrIndex;

	      if(SrcChild == ParentSrcChild) {
		// we have a match
		// go through the Child's TargetList and retrieve the entry
		XxxValueExprList *ChildTargetList = ChildGrpByOp->TargetList;
		for(int k=0; k<ParentSrcAttrIndex; k++)
		  ChildTargetList = ChildTargetList->Next;

		XxxValueExpression *ChildExpr = ChildTargetList->Expr;
		if(ChildExpr->ValType() == XxxValueExpression::COLUMN_VAR) {
		  // a groupby column
		  XxxUnknownValue *ChildUnknownExpr = 
		    (XxxUnknownValue *)ChildExpr;
		  ParentUnknownExpr->SrcChild = ChildUnknownExpr->SrcChild;
		  ParentUnknownExpr->SrcAttrIndex = 
		    ChildUnknownExpr->SrcAttrIndex;
		} else {
		  // an aggregate expression

		  // make sure to handle all of this
		}
	      }
	    }
	  }
	} // if groupby
      }
    }
  }
  // rule not applied
  return XXX_FALSE;
}

// this method determines whether a group-by pullup rule is applicable.
// the ParentOp is the parent SPJ node, and the ChildOp is the child 
// GroupBy Operator Node.
// It pretty much determines whether the group-by columns
// of the child are keys of the parent, i.e. in the parent's target list,
// they must only be output by the parent.
// In addition, the aggregate columns must only be projected in the parent,
// and not operated on in any way
XxxBool RelRewriteEngine::isGrpByPullUpApplicable(const RelSPJOp *ParentOp, 
						  const RelGrpByOp *ChildOp,
						  const int SrcChild) const {
  
  XxxValueExprList *ChildTargetList = ChildOp->TargetList;
  int ChildTargetListLength = ChildTargetList->Count();
  XxxValueExprList *ParentTargetList = ParentOp->TargetList;

  XxxBool *Output = new XxxBool[ChildTargetListLength];
  // Output[i] will hold whether column i in the child is output by the
  // parent
  // initialize Output to False
  for(int i=0; i<ChildTargetListLength; i++)
    Output[i] = XXX_FALSE;

  XxxBool Valid;
  // Now traverse the Parent's TargetList once, setting Output Values
  // if a column from this child is involved in a complex expression,
  // return False;
  Valid = areExpProjectedOnly(ParentTargetList, SrcChild, Output);
  
  if(Valid == XXX_FALSE)
    return XXX_FALSE;

  // make sure all Output Values are TRUE
  for(int i=0; i<ChildTargetListLength; i++)
    if(Output[i] == XXX_FALSE)
      return XXX_FALSE;

  delete[] Output;

  // at this point we know that this child's columns are only output by  
  // the parent in its target list, and are not involved in any complex 
  // expressions in the parent's target list
  // Now, let's ascertain whether the aggragte columns get operated on in 
  // the parent's where clause.
  // Remember: the groupby columns can have operations done on them, but 
  // the aggregate columns can't
  XxxBooleanExpression *Condition = ParentOp->Condition;
  if(Condition)
    return (!aggsUsed((XxxExpression *)Condition, ChildOp, SrcChild));
  
  return XXX_TRUE;
}

// this method assumes SrcChild to represent a child groupby Op and 
// Condition to be a Parent WhereClause
// it returns TRUE if the aggregation columns are involved in any operations 
//                 else it returns FALSE
XxxBool RelRewriteEngine::aggsUsed(XxxExpression *Condition,
				   const RelGrpByOp *ChildOp,
				   const int SrcChild) const {
  if(Condition->Arity() == 0) {
    XxxValueExpression *ValExp = (XxxValueExpression *)Condition;
    if(ValExp->ValType() == XxxValueExpression::COLUMN_VAR) {
      XxxUnknownValue *UnknownExp = (XxxUnknownValue *)ValExp;
      int UnknownSrcChild = UnknownExp->SrcChild;
      int UnknownSrcAttrIndex = UnknownExp->SrcAttrIndex;
      
      if(SrcChild == UnknownSrcChild) {
	// this column belongs to our child
	XxxValueExprList *ChildTargetList = ChildOp->TargetList;
	for(int i=0; i<UnknownSrcAttrIndex; i++)
	  ChildTargetList = ChildTargetList->Next;
	// reached our expression; see if its an aggregate
	if ((ChildTargetList->Expr->ValType() == XxxValueExpression::FUNC_EXPR) ||
	((XxxFuncExpression*)ChildTargetList->Expr->IsAggr()))
	  return XXX_TRUE;
      }
    }
    return XXX_FALSE;
  }
  
  XxxBool Used = XXX_FALSE;
  for(int i=0; i<Condition->Arity(); i++) {
    Used = Used || aggsUsed(Condition->ChildExpr(i), ChildOp, SrcChild);
    if(Used)
      return XXX_TRUE;
  }
  return XXX_FALSE;
}

// this method determines which columns for a particular SrcChild are output
// by a RelSPJNode parent.
// if the child's columns are involved in a complex expression, it returns
// FALSE
XxxBool 
RelRewriteEngine::areExpProjectedOnly(XxxValueExprList *ParentTargetList, 
					const int SrcChild, 
					XxxBool *&Output) const {

  XxxValueExprList *TList = ParentTargetList;
  int Length = TList->Count();

  // go through the Parent Target List setting Output values
  for(int i=0; i<Length; i++, TList=TList->Next) {
    XxxValueExpression *ParentValExpr = TList->Expr;
    XxxValueExpression::ValExprType ParentExprType = 
      ParentValExpr->ValType();

    if(ParentExprType == XxxValueExpression::COLUMN_VAR) {
      // this column is simply being output by the parent
      XxxUnknownValue *ParentUnknown = (XxxUnknownValue *)ParentValExpr;

      if(ParentUnknown->SrcChild == SrcChild) {
	// one of the guys we are looking for
	Output[ParentUnknown->SrcAttrIndex] = XXX_TRUE;
      }
    } else {
      // a complex expression
      XxxBool Occurs = inExpr((XxxExpression *)ParentValExpr, SrcChild);
      if(Occurs) {
	// uh..oh... the groupby's output is involved in a complex expression
	return XXX_FALSE;
      }
    }
  }
}

// this method determines if any column (SrcAttrIndex) from the child
// represented by SrcChild, SrcAttrIndex in Exp, where Exp is in the 
// parent Operator Node.
XxxBool 
RelRewriteEngine::inExpr(XxxExpression *Exp, const int SrcChild) const {

  if(Exp->Arity() == 0) {
      XxxValueExpression *ValExp = (XxxValueExpression *)Exp;
      if(ValExp->ValType() == XxxValueExpression::COLUMN_VAR) {
	// a column variable
	XxxUnknownValue *UnknownExp = (XxxUnknownValue *)ValExp;
	if(UnknownExp->SrcChild == SrcChild) {
	  // bingo, we have found a match
	  return XXX_TRUE;
	}
      }
  }

  // a complex expression; let's check the children
  XxxBool Occurs = XXX_FALSE;
  for(int i=0; i<Exp->Arity(); i++) {
    Occurs = Occurs || inExpr(Exp->ChildExpr(i), SrcChild);
    if(Occurs)
      break;
  }
  return Occurs;
}

XxxBool RelRewriteEngine::eToF(RelQueryNode *Node) const {
  // start with the basics, as usual
  XxxBool DidSomething=XXX_FALSE; // this variable will be set to XXX_TRUE
  // if any where clause members are moved up.
  
  // check the basics: Node must be a RelSPJ node and have children.
  if (Node==NULL) {
    char buf[100];
    sprintf(buf,"RelRewriteEngine::eToF: Received NULL argument");
    REL_NOTE_ERROR(REL_BADQUERY,buf);
    return XXX_FALSE;
  }
  
  if (Node->Op==NULL || Node->Op->Type()!=REL_SPJ)
    return XXX_FALSE;
  
  RelSPJOp* Op=(RelSPJOp*)Node->Op;
  
  // Make sure there is a where clause.
  if (Op->Condition==NULL)
    return XXX_FALSE;
  
  // check if some easy to check necessary conditions hold.
  if (Node->GetHeadDistinct()==XXX_FALSE && Node->GetBodyDistinct()!=PERMIT &&
      !oneTuple(Node))
    return XXX_FALSE;
  
  // find all the EXIST quantifiers and store them in a list.
  Quantifier **ExistList=new Quantifier*[Node->QList->GetNumUsed()];
  int ListLen=0;
  for (int i=0;i<Node->QList->GetNumUsed();i++) {
    Quantifier *Q=Node->QList->GetQuantifier(i);
    if (Q->GetQType()==Q_EXIST) {
      // add this quantifier to the list;  but AddQuantifier has a side-effect
      // that it automatically resets the contained node's whichQuantifierAmI
      // field (which is nice in most cases, but bad here).  So, we pass it
      // additional XXX_FALSE argument to prevent it from doing this.
      ExistList[ListLen++]=Q;
    }
  }
  
  // Check if there are any EXIST quantifiers.  If there aren't there's nothing
  // else for us to do.
  if (ListLen==0) {
    delete[] ExistList;
    return XXX_FALSE;
  }
  
  // In this early version, we can only deal with queries that have the form
  // where exists (....); we can't handle anything complex like "where exists
  // (...) and E.id>1."  Thus, if the WHERE clause is complex, don't do
  // anything.  The rest of this function, however, supports multiple exist
  // quantifiers in the node (which can't happen unless there's a compound
  // where clause.)
  if (Op->Condition->BoolType!=REL_QUERY_BOOL_EXPR) {
    delete[] ExistList;
    return XXX_FALSE;
  }
  
  // Make the RelQueryNode's in the EXIST quantifiers become children of Node.
  // While we're at it, we should reset the QuantifierType of each Quantifier
  // to Q_FROM. Increment the Node refcounts while we're here.
  RelQueryNode **NodeList=new RelQueryNode*[ListLen];
  for (int i=0;i<ListLen;i++) {
    NodeList[i]=ExistList[i]->GetNode();
    NodeList[i]->IncrRef();
    ExistList[i]->SetQType(Q_FROM);
  }
  Node->AddChildren(NodeList, ListLen);
  
  // Remove EXISTS from WhereClause.  Essentially, erase Op->Condition
  // For now, we don't delete it because it damages the RelQueryNode inside
  //  delete Op->Condition;
  Op->Condition=NULL;
  delete[] NodeList;
  delete[] ExistList;
  return XXX_TRUE;
}

************/

