#ifndef _RELREWRITE_H_
#define _RELREWRITE_H_

/******
 This is a very initial version of a query rewrite engine along the lines
 of that in Starburst/DB2. A weak imitation, but we will improve on
 this as time goes.
 In the next release, the plan is to convert these into precondition/action
 rules, and in fact merge this with the other rule engine in optrules.[Ch]
*******/
#include "relquery.h"

/**
 * A rewrite rule that is used to transform a relational query 
 * with multiple query blocks
 */
class RelRewriteRule {
 public:
  RelRewriteRule() {}
  virtual ~RelRewriteRule() {}
  virtual XxxErrCode ApplyRule(RelQueryNode *InNode, XxxBool &Success,
			       RelQueryNode *&OutNode) = 0;
};

/**
 * This is a simple query rewrite engine for query block transformations.
 */ 
class RelRewriteEngine {
 public:
  RelRewriteEngine(int MaxRules = 20);
  virtual ~RelRewriteEngine();
  //// the main method --- rewrite the input query
  virtual XxxErrCode Rewrite(RelQueryNode *InNode, RelQueryNode *&OutNode);
private:
  const int MAXRULES;
  int numUpRules, numDownRules;
  RelRewriteRule **upPhaseRules;
  RelRewriteRule **downPhaseRules;
  int numRulesFired;

  XxxErrCode executeRules(RelRewriteRule *Rules[], int NumRules, 
			  RelQueryNode *InNode, RelQueryNode *&OutNode);
};

// and now the various rules

class RelSelectMerge : public RelRewriteRule {
public:
  virtual XxxErrCode ApplyRule(RelQueryNode *InNode, XxxBool &Success, 
			       RelQueryNode *&OutNode);
private:
  // helper functions
  XxxErrCode merge(RelQueryNode *Parent, RelQueryNode *Child) const;
  XxxErrCode updateSrcChildren(XxxExpression *Exp, 
			       int ChildRelPos,
			       int ChildArity,
			       int ParentArity) const;
  XxxExpression* searchAndReplace(XxxExpression *OuterExpr, 
				  RelQueryNode *Node) const;

};

/**
 * Rewrite Rule that pushes method applications on an aggregate down from
 * a SPJ to a GBY node.
 * select x.a.imag() from 
 *  ( select csum(y.f1) as a from C2 y group by y.f2) x;
 * becomes:
 * select x.XXX_GEN_ID_1 from
 *  ( select csum(y.f1) as a, csum(y.f1).imag() as XXX_GEN_ID_1 from .... )
 * For more examples, see the rewrite test suites.
 * @see suites/rewrite/aggregate_push_down.run
 * @author Tobias Mayr (mayr@cs.cornell.edu)
 * @version April '98
 */
class RelAggrFuncPushDown : public RelRewriteRule {
public:
  /// applying this rule to a RelQueryNode 
  virtual XxxErrCode ApplyRule(RelQueryNode *InNode, XxxBool &Success, 
			       RelQueryNode *&OutNode);
private:
  /** 
   * Used to extract the positions of XxxValueExpressions that will be
   * pushed down. Positions of XxxValueExpression* are XxxValueExpression**
   */
  class ValExprRefList 
  {
  public:
    XxxValueExpression **Ptr;
    ValExprRefList* Next;
    ValExprRefList(XxxValueExpression **ptr) : Ptr(ptr), Next(NULL) {}
  };

  /**
   * Used to generate fresh id strings for the introduced column references
   * in the parent SPJ node.
   */
  class GeneratedIDString
  {
  private:
    static int number;
  public:
    static char* NewIDString() {
      char* id = new char[25];
      sprintf(id, "XXX_GEN_ID_%d", number++);
      return id; }
  };

  /**
   * Application of the RelAggrFuncPushDown rewrite to a given SPJ operator
   * and a given GBY operator child.
   * For a given Parent (SPJ operator) and its child (GBY operator), this
   * identifies all subexpressions in the targetlist that are method 
   * applications to an aggregate. It substitutes new Unknowns for them and
   * adds them as new targets to the child node.
   */
  XxxErrCode pushDownAggrMethods(RelQueryNode *Parent, 
				 RelQueryNode *Child, 
				 XxxBool &Success) const;

  /**
   * Substitutes all column references of an expression that is pushed 
   * down to the child. Recursive in exp. Column references are assumed
   * to refer to the child's target list, and are simply replaced by a
   * replication of the corresponding target expression.
   */
  XxxErrCode substituteColumnValues(XxxValueExpression* &exp, 
				    XxxValueExprList* TargetList) const;
    
  /**
   * Identifies all subexpressions of exp, that are method applications
   * to constants or unknowns referencing childNumber.
   * Recursive in exp. ancestor denotes the greatest superexpression of
   * exp that consists of method applications to constants or unknowns 
   * referencing childNumber. Initially, ancestor should be NULL.
   */
  ValExprRefList* findAggrMethodPushDown(XxxValueExpression **exp, 
					 XxxValueExpression **ancestor,
					 int childNumber) const;

};

/**
 * Rewrite rule that eliminates superfluous targets in SPJ and GBY nodes, that are
 * not referenced in the parent node.
 */ 
class RelElimUnrefTargets : public RelRewriteRule 
{
public:
  /// applying this rule to a RelQueryNode 
  virtual XxxErrCode ApplyRule(RelQueryNode *InNode, XxxBool &Success, 
			       RelQueryNode *&OutNode);
private:
  // simply deletes the first occurence of Expr from the List
  XxxErrCode deleteFirstExprFromList(XxxValueExprList** ExprListPtr, 
				     XxxValueExpression* Expr);
};

class RelDistinctPushDown : public RelRewriteRule {
  virtual XxxErrCode ApplyRule(RelQueryNode *InNode, XxxBool &Success, 
			       RelQueryNode *&OutNode);
};


/***********
  TODO:L3 A number of other rewrite rules should be rewritten,
          and will be added slowly.
	  Read Pirahesh, Hellerstein, Hasan [SIGMOD92] for common rules.

  All the other rules including GrpByPullUp
XxxBool isGrpByPullUpApplicable(const RelSPJOp *ParentOp, 
				const RelGrpByOp *ChildOp, 
				const int SrcChild) const;

XxxErrCode executeRule(const Rule *Rule, 
		       RelQueryNode *Node, 
		       XxxBool &Success) const;
XxxBool aggsUsed(XxxExpression *Condition, const RelGrpByOp *ChildOp,
		 const int SrcChild) const;
XxxBool inExpr(XxxExpression *Exp, const int SrcChild) const;
XxxBool areExpProjectedOnly(XxxValueExprList *ParentTargetList, 
			    const int SrcChild, XxxBool *&Output) const;
// oneTuple checks if the one tuple condition holds.
XxxBool oneTuple(RelQueryNode *Node) const {return XXX_FALSE;}

// rule functions

XxxBool selMerge(RelQueryNode *Node) const;
XxxBool grpByPullUp(RelQueryNode *Node) const;
XxxBool distPdFr(RelQueryNode *Node) const;
XxxBool distPdTo(RelQueryNode *Node) const;
XxxBool eToF(RelQueryNode *Node) const;
};
****************/
#endif
