#ifdef __GNUG__
#pragma implementation
#endif

#include <assert.h>

#include "gbyplanop.h"
#include "relops.h"
#include "reldata.h"
#include "releng.h"
#include "funcexpr.h"
#include "aggrexpr.h"

//
// Methods of RelGrpByPlanOp
//

RelGrpByPlanOp::RelGrpByPlanOp(RelPlanOp* Child,
			       XxxValExprPlanList *TargetList,
			       XxxValExprPlanList *AggrList,
			       XxxValExprPlanList *GroupList,
			       RelCatalogInfo* CatInfo) 
: RelDerivedPlanOp(1, &Child, CatInfo), TargetExprPlanList(TargetList),
  AggrExprPlanList (AggrList), GroupExprPlanList(GroupList)
{
}

RelGrpByPlanOp::~RelGrpByPlanOp() 
{ 
 // do nothing: lists reclaimed by RelGrpByBlockPlan
}

void RelGrpByPlanOp::NodePrint (ostream &OStr, int curBlockId, int Level, 
				int numChildren, int *childrenIds) const {
  OStr << "NODE Grouping / Aggregation" << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0x00CCCC << endl;
  OStr << "INFO Projection: ";
  if (TargetExprPlanList)
    TargetExprPlanList->Print(OStr);
  else
    OStr << "*";
  OStr << endl;
  OStr << "INFO Groupings: ";
  if (GroupExprPlanList)
    GroupExprPlanList->Print(OStr);
  else
    OStr << "none";
  OStr << endl;
  OStr << "INFO Aggregations: ";
  if (AggrExprPlanList)
    AggrExprPlanList->Print(OStr);
  else
    OStr << "none";
  OStr << endl;
  OStr << "ENDNODE" << endl;
}

XxxFuncExpression::AggrPlan *
RelGrpByPlanOp::getAggrPlan(XxxValueExprPlan *Plan)
{
  if (Plan->ValType() != XxxValueExpression::FUNC_EXPR)
    return NULL;

  if (!(((XxxFuncExpression::FuncPlan*)Plan)->IsAggr()))
    return getAggrPlan(((XxxFuncExpression::FuncPlan *)Plan)->OwnerPlan);
  else 
    return (XxxFuncExpression::AggrPlan *)Plan;
}

RelGrpByPlanOp::
ScanInfo::ScanInfo(XxxValExprPlanList* AggrList, XxxErrCode& rc)
{
  rc = XXX_OK;
  // Create the operator handles
  NumAggrExprs = (AggrList ? AggrList->Count() : 0);
  ASSERT(NumAggrExprs > 0);

  OpHandles = new XxxAggrOpHandle*[NumAggrExprs];
  XxxValExprPlanList* AList = AggrList;
  int AggrIndex = 0;
  while (AList) {
    XxxFuncExpression::AggrPlan* AggrPlan = 
      RelGrpByPlanOp::getAggrPlan(AList->Plan);
    ASSERT ((AggrPlan));
    REL_CALL_RCFN(AggrPlan->MakeOpHandle(OpHandles[AggrIndex++]), rc,
		REL_BADCURSOR, "Can't create OpHandle for an aggregate");
    AList = AList->Next;
  }
  LastRec = 0;
}

RelGrpByPlanOp::ScanInfo::~ScanInfo()
{
  if (LastRec) delete LastRec;

  if (OpHandles) {
    for (int i = 0; i < NumAggrExprs; i++) {
      if (OpHandles[i])	delete OpHandles[i];
    } // for
    delete [] OpHandles;
  }
}

XxxErrCode RelGrpByPlanOp::
InitHandle(XxxDerivedRelation::CursorInfo* Handle,
	   XxxRelation* *RelationArray) const
{
  // If there is a GROUP BY clause, the input is a SortPlanOp
  // which is assumed to order the data by the grouping expressions
  if (Handle->ArrayLen != 1) {
    REL_NOTE_ERROR(REL_BADOP, "Need one input to aggregate");
    return XXX_NOT_OK;
  }

  // open a cursor on the input relation
  REL_CALL_FN(RelationArray[0]->InitCursor1(Handle->ChildCursorArray[0],
					    NULL, NULL, NULL,
					    Handle->Params),
	      REL_BADCURSOR, "Unable to open input relation cursor");

  // Create the operator handles
  XxxErrCode rc = XXX_OK;
  ScanInfo *SInfo = new ScanInfo(AggrExprPlanList, rc);
  Handle->OperatorHandle = (void *)SInfo;
  return rc;
}

XxxErrCode 
RelGrpByPlanOp::CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
			    XxxRelation* *RelationArray) const
{
  if (Handle->ChildCursorArray[0].State &&
      RelationArray[0]->CloseCursor(Handle->ChildCursorArray[0]) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADCURSOR, "Unable to close cursor");
    return XXX_NOT_OK;
  }
  ScanInfo *SInfo = (ScanInfo *)(Handle->OperatorHandle);
  if (SInfo) delete SInfo;

  return XXX_OK;
}

XxxErrCode RelGrpByPlanOp::GetNextRecord(XxxSetCursor& SCsr,
					 XxxRelation* *RelationArray,
					 const XxxRecordSchema* Schema,
					 XxxRecord*& ReturnRec) const
{
  XxxDerivedRelation::CursorInfo* Handle = 
    (XxxDerivedRelation::CursorInfo*)(SCsr.State);
  ASSERT((Handle));
  ASSERT((Schema->Arity() == TargetExprPlanList->Count())); 

  // Get the information passed in through the operator handle
  ScanInfo *SInfo = (ScanInfo *)(Handle->OperatorHandle);
  XxxRecord *&LastRec = SInfo->LastRec;
  XxxAggrOpHandle** OpHandles = SInfo->OpHandles;

  // Check to see if input cursor is valid
  if (Handle->ChildCursorArray[0].IsValid() == XXX_FALSE) {
    // no more records to process
    SCsr.AtEnd = XXX_TRUE;         
    ReturnRec = 0;
    if (LastRec) {
      delete LastRec;
      // set LastRec to NULL after deletion.
      // otherwise we delete LastRec twice (2nd time in CloseHandle)
      LastRec = NULL;
    }
    return XXX_OK;
  }

  REL_CALL_FN(initAggrExprs(OpHandles),
	      REL_BADEXEC, "Error initializing aggregates");

  // if LastRec is not NULL, it is the first record of the next group;
  // aggregate it in
  if (LastRec != 0) {
    REL_CALL_FN(aggregateRecord(LastRec, AggrExprPlanList, 
				Handle, OpHandles),
		REL_BADEXEC, "Cannot aggregate last record");
  }

  // Get first record of next/first group
  REL_CALL_FN(RelationArray[0]->NextItem(Handle->ChildCursorArray[0],
					 Handle->ChildRecs[0]),
	      REL_BADEXEC, "Error getting NextItem from relation");

  // aggregate all incoming records which are in the same group
  while ((Handle->ChildCursorArray[0].IsValid() == XXX_TRUE) &&
	 (sameGroup(LastRec, Handle->ChildRecs[0], Handle->ChildSchema[0], 
		    Handle->Params ? Handle->Params->Params : 0, 
		    GroupExprPlanList) == XXX_TRUE)) {
    // accumulate the record values into the aggregates
    REL_CALL_FN(aggregateRecord(Handle->ChildRecs[0], AggrExprPlanList, 
				Handle, OpHandles),
		REL_BADEXEC, "Cannot aggregate record");
    // store the processed record as the last one and destroy the previous one
    if (LastRec) {
      delete LastRec;
    }
    LastRec = Handle->ChildRecs[0];
    Handle->ChildRecs[0] = 0;
    
    REL_CALL_FN(RelationArray[0]->NextItem(Handle->ChildCursorArray[0],
					   Handle->ChildRecs[0]),
		REL_BADEXEC, "Error getting NextItem from relation");
  } // while ChildScanIsValid() && (sameGroup(...))

  if (!LastRec &&
      (TargetExprPlanList->Count() != AggrExprPlanList->Count())) {
    // No more records to process and this is not a scalar aggregate
    // query (so even if the input table is empty, there is some o/p)
    SCsr.AtEnd = XXX_TRUE;         
    ReturnRec = 0;
    return XXX_OK;
  }

  // Otherwise, there exists an aggregate record to be returned.
  // Terminate aggregate accumulations and form the RecValuesArray
  REL_CALL_FN(generateRecValues(OpHandles, Schema,
				LastRec, Handle->ChildSchema[0],
				Handle->RecValuesArray), 
	      REL_BADEXEC, "Error generating aggregate record");
    
  // maintain the lookahead record
  if (LastRec) delete LastRec;
  LastRec = Handle->ChildRecs[0];
  Handle->ChildRecs[0] = 0;

  // now make a new record out of the recValArray
  ReturnRec = new XxxRecord(Handle->RecValuesArray, Schema);
  return XXX_OK;
}

XxxErrCode RelGrpByPlanOp::
generateRecValues(XxxAggrOpHandle** OpHandles,
		  const XxxRecordSchema* Schema, 
		  XxxRecord *LastRec,
		  const XxxRecordSchema* LastRecSchema, 
		  XxxValuesArray& ValArray) const
{
  XxxValExprPlanList *CurTarget = TargetExprPlanList;
  int CurAggrIndex = 0;
  int CurColIndex = 0;
  while (CurTarget) {  // for every column
    // CurTarget->Plan may now be a group by column, an aggregate 
    // expression, or a function on an aggregate expression
    XxxFuncExpression::AggrPlan *CurAggrPlan = getAggrPlan(CurTarget->Plan);
    if (CurAggrPlan) {
      REL_CALL_FN(generateAggrValue(CurTarget->Plan, CurAggrPlan,
				    OpHandles[CurAggrIndex], Schema,
				    LastRec, LastRecSchema, 
				    ValArray.Values[CurColIndex]),
		  REL_BADEXEC, "Error generating aggr value");
      CurAggrIndex++;
    } // if an aggr expression is involved
    else {
      // the column is not an aggr expression--get its value from LastRec
      REL_CALL_FN(generateScalarValue(CurTarget->Plan, CurColIndex, 
				      Schema,
				      LastRec, LastRecSchema, 
				      ValArray.Values[CurColIndex]),
		  REL_BADEXEC, "Error generating aggr value");
    } 
    CurTarget = CurTarget->Next;
    CurColIndex++;
  } // for each attribute

  return XXX_OK;
}

XxxErrCode RelGrpByPlanOp::
generateAggrValue(XxxValueExprPlan* CurPlan, 
		  XxxFuncExpression::AggrPlan *CurAggrPlan,
		  XxxAggrOpHandle* OpHandle, 
		  const XxxRecordSchema* Schema, 
		  XxxRecord *LastRec,
		  const XxxRecordSchema* LastRecSchema, 
		  XxxADTValue* Result) const
{
  // allocate some space to store the computed aggregate value
  XxxAggrOpHandle *CurrOpHandle = OpHandle;
  XxxBasicType AggrOutType = CurAggrPlan->BasicType();
  const XxxADTMetaInfo *AggrOutMetaInfo = CurAggrPlan->GenerateCatInfo();
  int AggrSize = XxxADTArray[AggrOutType]->MaxObjectSize(AggrOutMetaInfo);
  XxxADTValue *AggrValue = (XxxADTValue *)new char[AggrSize];
  AggrValue->Init();
  
  REL_CALL_FN(CurAggrPlan->Terminate(OpHandle, AggrValue),
	      REL_BADEXEC, "Error ending aggregate computation");
  
  REL_CALL_FN(aggregatePostProcess((XxxFuncExpression::FuncPlan*)CurPlan, 
				   (const XxxRecord **)&LastRec,
				   (const XxxRecordSchema **)&LastRecSchema,
				   AggrValue, AggrSize, Result),
	      REL_BADEXEC, "problem in aggregate post-processing");
  
  // clean up, we don't need the computed aggregate value any more
  delete [] (char *)AggrValue;
  return XXX_OK;
}

XxxErrCode RelGrpByPlanOp::
generateScalarValue(XxxValueExprPlan* CurPlan, 
		    int CurColIndex, 
		    const XxxRecordSchema* Schema, 
		    XxxRecord *LastRec,
		    const XxxRecordSchema* LastRecSchema, 
		    XxxADTValue* Result) const
{
  // i'm not sure this is necessary
  ASSERT(CurPlan->ValType() == XxxValueExpression::COLUMN_VAR);

  //Bjarke, sep. 13 2001: added const qualifier to variable
  const char *FldPtr;
  LastRec->GetField(((XxxUnknownValue::UnknownValuePlan *)
		     CurPlan)->SrcAttrIndex, LastRecSchema, FldPtr);
  XxxADTValue* Data = (XxxADTValue*)FldPtr;
  ASSERT (Data);
  int DataSize = 0;

  XxxBasicType TypeId = (*Schema)[CurColIndex].Type;
  if (XxxCoreADTClass::IsCoreADT(TypeId)) {
    XxxCoreADTClass::STypeCopy(TypeId, Data, Result,
			       (*Schema)[CurColIndex].TypeCatInfo);
  }
  else {
    XxxADTArray[TypeId]->
      TypeCopy(Data, Result, (*Schema)[CurColIndex].TypeCatInfo);
  }
  return XXX_OK;
}

XxxErrCode RelGrpByPlanOp::
initAggrExprs(XxxAggrOpHandle** OpHandles) const
{
  // Init the aggr plans' AggrOpSpecs
  XxxValExprPlanList* CurAggr = AggrExprPlanList;
  int CurAggrIndex = 0;
  while (CurAggr) {
    // since we may have something as complex as a function on an aggregate,
    // we must retrieve the Aggregate plan from the current expression's plan
    // tree
    XxxFuncExpression::AggrPlan *CurAggrPlan = getAggrPlan(CurAggr->Plan);
    REL_CALL_FN(CurAggrPlan->Init(OpHandles[CurAggrIndex++]),
		REL_BADEXEC, "Error initializing aggregate expression");
    CurAggr = CurAggr->Next;
  }
  return XXX_OK;
}

XxxBool 
RelGrpByPlanOp::sameGroup(const XxxRecord *LastRec, 
			  const XxxRecord *NextRec,
			  const XxxRecordSchema *Schema,
			  const XxxValueEnv *Env,
			  XxxValExprPlanList *GroupExprPlanList) const
{
  if (!(LastRec && NextRec)) return XXX_TRUE;
  // Every expression in the Target list that is not an aggregate is a
  // grouping expression.
  XxxValExprPlanList *GList = GroupExprPlanList;
  int AttrIndex;
  while (GList) {
    AttrIndex = ((XxxUnknownValue::UnknownValuePlan *)
		 (GList->Plan))->SrcAttrIndex;
    XxxBasicType TypeId = (*Schema)[AttrIndex].Type;
    XxxADTMetaInfo* MInfo = (*Schema)[AttrIndex].TypeCatInfo;
    //Bjarke, sep. 13 2001: added const qualifier to variable
    const char *Fld1, *Fld2;
    LastRec->GetField(AttrIndex, Schema, Fld1);
    NextRec->GetField(AttrIndex, Schema, Fld2);
    if (XxxCoreADTClass::IsCoreADT(TypeId)) {
      if (!(XxxCoreADTClass::
        SEquals(TypeId,
	        (XxxADTValue*)Fld1, MInfo, (XxxADTValue*)Fld2, MInfo)))
        return XXX_FALSE;
    }
    else {
      if (!(XxxADTArray[(*Schema)[AttrIndex].Type]->
	    Equals((XxxADTValue*)Fld1, MInfo, (XxxADTValue*)Fld2, MInfo)))
        return XXX_FALSE;
    }
    GList = GList->Next;
  }
  return XXX_TRUE;
}

XxxErrCode 
RelGrpByPlanOp::aggregateRecord (XxxRecord *Record,
				 XxxValExprPlanList *AggrExprPlanList,
				 XxxDerivedRelation::CursorInfo* Handle,
				 XxxAggrOpHandle** OpHandles) const 
{
  XxxValExprPlanList *CurAggr = AggrExprPlanList;
  int CurAggrIndex = 0;
  while (CurAggr) {
    XxxFuncExpression::AggrPlan *CurAggrPlan = getAggrPlan(CurAggr->Plan);
    REL_CALL_FN(CurAggrPlan->Accumulate((const XxxRecord **)&Record, 
					(const XxxRecordSchema **)Handle->ChildSchema, 
					OpHandles[CurAggrIndex],
					Handle->Params ? Handle->Params->Params : 0),
		REL_BADEXEC, "Error accumulating record arguments");

    CurAggr = CurAggr->Next;
    CurAggrIndex++;
  }
  return XXX_OK;
}

// TODO: revamp this: also, get rid of the use of cerr
XxxErrCode 
RelGrpByPlanOp::aggregatePostProcess(const XxxFuncExpression::FuncPlan* CurPlan, 
				     const XxxRecord **Rec,
				     const XxxRecordSchema **RecSchema,
				     XxxADTValue *AggrValue, int AggrSize,
				     XxxADTValue *ReturnValue) const 
{
  if (CurPlan->IsAggr()) {
    // this case only happens when there is no post-processing necessary
    memcpy((char*)ReturnValue, (char*)AggrValue, AggrSize);
    return XXX_OK;
  } 
    
  // a function expression
  XxxValueExprPlan *OwnerPlan = CurPlan->OwnerPlan;

  XxxADTValue *OwnerValue;   // will be used as a temporary if the 
                             // owner is a func-expression also
  XxxValueExpression::ValExprType ValType;
  if ((ValType = OwnerPlan->ValType()) != XxxValueExpression::FUNC_EXPR) {
    // this case should never happen : TODO: i dont understand why
    assert(0);
  }
  // the owner is a func-expression also
  XxxFuncExpression::FuncPlan *OwnerFuncPlan = 
    (XxxFuncExpression::FuncPlan *)OwnerPlan;
  if (OwnerFuncPlan->IsAggr()) {
    // the owner is an aggregate expression, termination case in 
    // our recursion
    return CurPlan->Evaluate1(Rec, RecSchema, AggrValue, ReturnValue);
  }
  
  XxxBasicType OwnerReturnType = OwnerFuncPlan->ReturnType(); 
  XxxADTMetaInfo *OwnerCatInfo = OwnerFuncPlan->ReturnCatInfo();
  int OwnerSize = 
    XxxADTArray[OwnerReturnType]->MaxObjectSize(OwnerCatInfo);
  OwnerValue = (XxxADTValue *)new char[OwnerSize];
  OwnerValue->Init();
  
  XXX_CALL_RET(aggregatePostProcess(OwnerFuncPlan, Rec, RecSchema, 
				    AggrValue, AggrSize, OwnerValue),
	       XXX_BADEXEC, "Error in aggregatePostProcess (recursing in).");
    
  // we have come out of the recursion at this point     
  // evaluate the function expression.
  XXX_CALL_RET(CurPlan->Evaluate1(Rec, RecSchema, OwnerValue, ReturnValue),
	       XXX_BADEXEC, "Error in aggregatePostProcess (recursing out)");
  delete [] (char *)OwnerValue;
  return XXX_OK;
}

void RelGrpByPlanOp::PrintThis(ostream& OStr) const
{
  OStr << PrintString();
  if (TargetExprPlanList) {
    OStr << "[";
    TargetExprPlanList->Print(OStr);
    OStr << "]";
  } 
  if (AggrExprPlanList) {
    OStr << "[";
    AggrExprPlanList->Print(OStr);
    OStr << "]";
  } 
  if (GroupExprPlanList) {
    OStr << "[";
    GroupExprPlanList->Print(OStr);
    OStr << "]";
  } 
}
