//SMostafa
//
// FEW-join (Nested Loop Impl)
//

#ifndef REL_NLFEWPLANOP_H
#define REL_NLFEWPLANOP_H

#include <sys/time.h>
#include "relenv.h"
#include "relplanop.h"

XXX_EXTERN_CLASS_DECL(RelOperator);

#define MAX_HASH_SIZE  2097152 // 2M, power of 2 
//#define MAX_HASH_SIZE  2024  // 2M, power of 2 
#define MAX_BUCKET_CNT 100 // 
#define SWAP_RATIO 0.5

#define MAX_JOINED_STREAMS  100 // Max. number of joined streams in a multi-way fashion

#define ACTIVE  0
#define ZOMBIE  1
#define EXPIRED 2




//----------------------------------------------------------------
/**
 * The Merge-Window Join plan operator.  
 * Only selected as one of the inputs contains a stream field?
 */
class RelNLFEWPlanOp; 
class RelNLFEWPlanOp : public RelSPJPlanOp {
 protected:

  /*
    Encapsulates the concept of a record group; maintains a linked-
    list of records and information on which attribute the join
    is based upon.
  */
  class Group {
  private:
    // The linked-list node structure
    struct Member {
      XxxRecord *Rec;
      int value;   // the value in which hashing was based
      struct timeval HashTimeStamp; // time stamp at which the tuple is added to the hash table, used to eliminate duplicates
      Member *Next;
    };
    // The head of the linked list
    Member *members;
    // The cursor in scans
    Member *currentMember;
    // The attribute function plan of the records in the equality join condition
    XxxValueExprPlan* AttrFuncPlan;
    // Whether or not the current scan is valid
    XxxBool scanValid;
  public:
    /*
      Constructor.
    */
    Group(XxxValueExprPlan* AFP) : members(NULL), currentMember(NULL),
      AttrFuncPlan(AFP), scanValid(XXX_FALSE) { }
    Group() : members(NULL), currentMember(NULL),
      AttrFuncPlan(NULL), scanValid(XXX_FALSE) { }
    /*
      Destructor.
    */
    ~Group() {
      FlushAllRecords(XXX_FALSE);
    }
    /*
      Returns the attribute function plan of the critical attribute of the contained records.
    */
    XxxValueExprPlan* AttrFunc() { return AttrFuncPlan; }
    /*
      Returns XXX_TRUE if this group has no records in it.
    */
    XxxBool IsEmpty() { return members == NULL; }
    /*
      Adds a record to this group.
    */
    void AddRecord(XxxRecord *Rec) {
      ASSERT(Rec);
      Member *NewMember = new Member;
      NewMember->Rec = Rec;
      NewMember->Next = members;
      members = NewMember;
    }
    /*
      Adds a record to this group.
    */
    void AddRecord(XxxRecord *Rec, int v) {
      ASSERT(Rec);
      Member *NewMember = new Member;
      NewMember->Rec = Rec;
      NewMember->value =v;

      // append time stamp for each record in the hash table
      struct timezone Temp; 
      gettimeofday(&(NewMember->HashTimeStamp), &Temp); 
      NewMember->Next = members;
      members = NewMember;
    }
    /*
      Removes all records from this group.
      @param DestroyRecs If XXX_TRUE, the records within will be explicitly
      destroyed; otherwise, only the linked-list structure
      will be destroyed.
    */
    void FlushAllRecords(XxxBool DestroyRecs) {
      Member *cur = members, *tmp;
      while (cur) {
	if (DestroyRecs)
	  delete cur->Rec;
	tmp = cur->Next;
	delete cur;
	cur = tmp;
      }
      scanValid = XXX_FALSE;
      members = NULL;
    }

    void FlushAllRecords(XxxBool DestroyRecs, const XxxRecordSchema* RecSchema) {
      Member *cur = members, *tmp;
      while (cur) {
	tmp = cur->Next;
	if (DestroyRecs)
	  cur->Rec->Reclaim(RecSchema);
	delete cur;
	cur = tmp;
      }
      scanValid = XXX_FALSE;
      members = NULL;
    }

    /*
      Initializes a scan over the group.
    */
    void InitScan() {
      currentMember = members; 
      scanValid = XXX_TRUE;
    }
    /*
      Returns XXX_TRUE if the current scan is valid.
    */
    XxxBool ScanIsValid() { return (scanValid); }
    /*
      Gets the next record from the scan.  Returns NULL if the scan is
      now invalid.
    */
    XxxRecord *GetNext() {
      if (currentMember == NULL) {
	scanValid = XXX_FALSE;
	return NULL;
      } else {
	XxxRecord *Rec = currentMember->Rec;
	currentMember = currentMember->Next;
	scanValid = XXX_TRUE;
	return Rec;
      }
    }
    /*
      Gets the next record from the scan.  Returns NULL if the scan is
      now invalid. Also, returns the value of the hashed field
    */
    XxxRecord *GetNext(int* v) {
      if (currentMember == NULL) {
	scanValid = XXX_FALSE;
	return NULL;
      } else {
	XxxRecord *Rec = currentMember->Rec;
	*v = currentMember->value;
	currentMember = currentMember->Next;
	scanValid = XXX_TRUE;
	return Rec;
      }
    }

    /*
      Closes the current scan.
    */
    void CloseScan() { currentMember = NULL; scanValid = XXX_FALSE; }
  };


  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  // A doubly linked list to represent the records in the bucket
  // The structure for an entry in the chain
  class Entry{
  public:
    XxxRecord *Rec;
    int value;          // the value in which hashing was based
    char swapped;       // set to true if this record was swapped to disk
    // must have 3 time stamps, to keep track of duplicates
    struct timeval HashTimeStamp;//time stamp at which the tuple is added to the hash table
    struct timeval SwapTimeStamp;//time stamp at which the tuple is swapped
    struct timeval ReHashTimeStamp;//time stamp at which the tuple is re-added

    Entry(XxxRecord *R, int v, struct timeval ts){
      Rec = R;
      value = v;
      //      swapped = FALSE;
      HashTimeStamp.tv_sec  = ts.tv_sec;
      HashTimeStamp.tv_usec = ts.tv_usec;
      SwapTimeStamp.tv_sec = 0;
      SwapTimeStamp.tv_usec = 0;
      ReHashTimeStamp.tv_sec = 0;
      ReHashTimeStamp.tv_usec = 0;
    };
    
    Entry(char* data, int size){
      // convert raw data to entry contants
      
      int SizeOfTimeStamp = sizeof(struct timeval);
      int SizeOfChar = sizeof(char);
      int SizeOfInteger = sizeof(int);      

      // extract record size field
      int recSize=0;
      memcpy((char *)&recSize, data, SizeOfInteger);
      // extract the record field 
      //      Rec = new XxxRecord(size-SizeOfInteger-SizeOfChar-2*SizeOfTimeStamp);
      Rec = new XxxRecord(recSize);
      ASSERT(Rec!=NULL);
      memcpy(Rec->DataArea, data+SizeOfInteger, Rec->Size());

      // extract the value field 
      char *ptrV = (char *)&value;
      memcpy(ptrV, data+SizeOfInteger+Rec->Size(), SizeOfInteger);

      // extract the swapped field

      memcpy(&swapped, data+SizeOfInteger+Rec->Size()+SizeOfInteger, SizeOfChar);
      ASSERT(swapped==1);

      // extract the size of the time stamps vector
      int VecSize=0;
      memcpy(&VecSize, data+SizeOfInteger+Rec->Size()+SizeOfInteger+SizeOfChar, SizeOfInteger);
      Rec->TSVec = new TheTimeStampVector(VecSize);
      // extract the satrt time stamps 
      int tmpsize = SizeOfInteger+Rec->Size()+ SizeOfInteger+SizeOfChar+SizeOfInteger;

      for(int i=0;i<VecSize;i++)
	memcpy((char *)(&Rec->TSVec->TheVec[i]),
	       data + tmpsize + i*sizeof(struct timeval),
	       sizeof(struct timeval));
    }
    // to Convert the non records information in the Entry 
    // to Raw Data (sequence of bytes)
    // caller must deallocate the data
    // value + swapped + size of vector + sequence of Time Stamps
    ConvertEntryToRawData(char *&ExtraData,int *SizeOfExtraData){
      *SizeOfExtraData = sizeof(int)+sizeof(char)+ sizeof(int) + 
	Rec->TSVec->VecSize*sizeof(struct timeval);
      ExtraData = new char[*SizeOfExtraData];
      ASSERT(ExtraData!=NULL);
      memcpy(ExtraData, (char *)&value,  sizeof(int));
      memcpy(ExtraData+sizeof(int), (char *)&swapped,  sizeof(char));
      memcpy(ExtraData+sizeof(int)+sizeof(char), 
	     (char *)&Rec->TSVec->VecSize, sizeof(int));
      int tmpsize = sizeof(int) + sizeof(char) + sizeof(int);
      for(int i=0;i<Rec->TSVec->VecSize;i++)
	memcpy(ExtraData+tmpsize+i*sizeof(struct timeval), 
	       (char *)(&Rec->TSVec->TheVec[i]), sizeof(struct timeval));
    };
  };// class Entry

  //-----------------------------------------------------------------------
  class BucketChain{
  private:
    class Member {
    public:
      Entry  *ChainEntry; 
      Member *NextMember;
      Member *PrevMember;
      Member (Entry  *E,
	      Member *Next,
	      Member *Prev):
	ChainEntry(E), NextMember(Next), PrevMember(Prev){
	
	if (NextMember != NULL)
	  NextMember->PrevMember = this;
	if (PrevMember != NULL)
	  PrevMember->NextMember = this;
      };
      ~Member (){};
    };
    // The head of the linked list
    Member *Head;
    // The tail of the linked list
    Member *Tail;
    // The cursor in scans
    Member *currentMember;
    // The attribute function plan of the records in the equality join condition
    XxxValueExprPlan* AttrFuncPlan;
    // Whether or not the current scan is valid
    XxxBool scanValid;
  public:
    
    BucketChain(XxxValueExprPlan* AFP) : Head(NULL), Tail(NULL), currentMember(NULL),
      AttrFuncPlan(AFP), scanValid(XXX_FALSE) { }
    BucketChain() : Head(NULL), Tail(NULL), currentMember(NULL),
      AttrFuncPlan(NULL), scanValid(XXX_FALSE) { }
    
    ~BucketChain() {
      FlushAllRecords(XXX_FALSE);
    }
    
    XxxValueExprPlan* AttrFunc() { return AttrFuncPlan; }
    
    XxxBool IsEmpty() { return Head == NULL; }
    
    void AddEntry(Entry  *E) {
      ASSERT(E);
      if (Tail==NULL){
	// first entry
	Tail = new Member(E, NULL, NULL);
	Head = Tail;
	return;
      };
      // table has entries
      Tail =  new Member(E, NULL, Tail);
      Tail->PrevMember->NextMember = Tail;
      
    }

    //TODO: correct deletion of the records
    void FlushAllRecords(XxxBool DestroyRecs) {
      Member *cur = Head, *tmp;
      while (cur) {
	if (DestroyRecs){
	  delete cur->ChainEntry->Rec;
	  delete cur->ChainEntry;
	}
	tmp = cur->NextMember;
	delete cur;
	cur = tmp;
      }
      scanValid = XXX_FALSE;
      Head = NULL;
      Tail = NULL;
    }


    //TODO: correct deletion of the records
    void FlushAllRecords(XxxBool DestroyRecs, const XxxRecordSchema* RecSchema) {
      Member *cur = Head, *tmp;
      while (cur) {
	if (DestroyRecs){
	  cur->ChainEntry->Rec->Reclaim(RecSchema);
	  delete cur->ChainEntry;
	}
	tmp = cur->NextMember;
	delete cur;
	cur = tmp;
      }
      scanValid = XXX_FALSE;
      Head = NULL;
      Tail = NULL;
    }


    
    void InitScan() {
      currentMember = Head; 
      scanValid = XXX_TRUE;
    }

    void InitBackwardScan() {
      currentMember = Tail; 
      scanValid = XXX_TRUE;
    }
    
    XxxBool ScanIsValid() { return (scanValid); }

    Entry  *GetNext() {
      if (currentMember == NULL) {
	scanValid = XXX_FALSE;
	return NULL;
      } else {
	Entry  *E = currentMember->ChainEntry;
	currentMember = currentMember->NextMember;
	scanValid = XXX_TRUE;
	return E;
      }
    }

    Entry  *GetPrevious() {
      if (currentMember == NULL) {
	scanValid = XXX_FALSE;
	return NULL;
      } else {
	Entry  *E = currentMember->ChainEntry;
	currentMember = currentMember->PrevMember;
	scanValid = XXX_TRUE;
	return E;
      }
    }
    
    void CloseScan() { currentMember = NULL; scanValid = XXX_FALSE; }

    // remove an entry where the scan has just passed
    void RemoveEntry(Entry* E){
      
      Member *ThisMember = currentMember;
      if (currentMember==NULL)
	ThisMember = Tail;
      else
	ThisMember = currentMember->PrevMember;
      ASSERT(ThisMember!=NULL);
      if(E==NULL)
	return;
      else{
	if(ThisMember->NextMember == NULL)
	  // delele tail
	  Tail = ThisMember->PrevMember;

	if(ThisMember->PrevMember == NULL)
	  // delete head
	  Head = ThisMember->NextMember;
	if(ThisMember->NextMember != NULL)
	  ThisMember->NextMember->PrevMember = ThisMember->PrevMember;
	if(ThisMember->PrevMember != NULL)
	  ThisMember->PrevMember->NextMember = ThisMember->NextMember;
	
      }
        
    }


    //GetTop
    //1) Get tuple at top of bucket
    Entry* GetTop(){
      
      Member *currentMember = Head; 
      
      if(currentMember==NULL)
	return NULL;
      else
	return currentMember->ChainEntry;

    }


    //DeleteTop
    //1) Remove and return tuple at top of bucket 
    Entry* DeleteTop(){
      Member *currentMember = Head; 
      
      if(currentMember==NULL)
	return NULL;
      else{
	if(Tail == Head){
	  // only one entry
	  Tail = NULL;
	  Head = NULL;
	}
	else{
	  Head = currentMember->NextMember;
	  Head->PrevMember = NULL;
	}
	return currentMember->ChainEntry;
      }
    }

    // Remove expired tuples from this bucket
    // expired when (currentTS - ts) > w
    void RemoveExpired(struct timeval ts, Group *eGroup, int *cnt, int w){
      Entry* E;
      struct timeval EntryTS;
      double interval;

      while(1){
	E = GetTop();
	if(E==NULL) break;
	EntryTS = E->HashTimeStamp;
	interval = TimeStampsDiff(ts, EntryTS);
	if(interval < w) break;
	E = DeleteTop();
	eGroup->AddRecord(E->Rec);
	*cnt = *cnt+1;
      }
    }

    //1) Returns the tuple at bottom of the bucket 
    Entry* GetBottom(){
      
      Member *currentMember = Tail; 
      
      if(currentMember==NULL)
	return NULL;
      else
	return currentMember->ChainEntry;

    }
    //DeleteBottom
    //1) Remove and return tuple at bottom of bucket 
    Entry* DeleteBottom(){
      Member *currentMember = Tail; 

      if(currentMember==NULL)
	return NULL;
      else{
	if(Tail == Head){
	  // only one entry
	  Tail = NULL;
	  Head = NULL;
	}
	else{
	  Tail = currentMember->PrevMember;
	  Tail->NextMember = NULL;
	}
	//	ASSERT(currentMember->ChainEntry->Rec->Size()==72);
	return currentMember->ChainEntry;
      }
    }

    //---------------

    // TS1>TS2?
    XxxBool TimeStampsGreater(struct timeval TS1, struct timeval TS2){
      double diff = ((TS1.tv_sec-TS2.tv_sec)+
		     (TS1.tv_usec-TS2.tv_usec)/1000000.0);
      return ((diff>0?XXX_TRUE:XXX_FALSE));
    };
    // TS1-TS2
    double TimeStampsDiff(struct timeval TS1, struct timeval TS2){
      return ((TS1.tv_sec-TS2.tv_sec)+
	      (TS1.tv_usec-TS2.tv_usec)/1000000.0);
    };


  };//BucketChain
  
  //-------------------------------------------------------------------------
  class TheSwapFile{
  public:
    FILE *SFile; 
    char FileName[100]; 
    int  RecordSize;
    int FileOffset;
    
    TheSwapFile(){
      char *TmpName; 
      
      strcpy(FileName, "/u/u15/mhammad");					 
      TmpName = (FileName+strlen(FileName)); 
      tmpnam(TmpName);
      
      //	RecordSize = size;
      FileOffset = 0;
      //    write(fileno(myfile), origVideo->Frame, origVideo->FrameSize);
    };
    
    InsertRecord(char* data, int size){ // append to end of file
      SFile = fopen(FileName, "ab+"); 
      write(fileno(SFile), data, size);
      fclose(SFile); 
      RecordSize = size;
    };
    
    InitScan(){
      SFile = fopen(FileName, "ab+"); 
      fseek(SFile, FileOffset, SEEK_SET);
    };
    
    int GetNext(char* data){
      // read a record
      int retVal = 0;
      if(!feof(SFile)){
	retVal = fread(data, sizeof(char), RecordSize, SFile);
	// if valid return the record and set the offset to next
	// record
	if(retVal==0) return 0;
	ASSERT(retVal == RecordSize);
	//	  FileOffset = FileOffset + RecordSize;
      };
      
      return retVal;
    };
    
    CloseScan(){
      fclose(SFile); 
    };
    
    int DeleteTopRecord(char* data){// delete from top of file
      // only set a position of next record
      // read a record
      int retVal = 0;
      if(!feof(SFile)){
	retVal = fread(data, sizeof(char), RecordSize, SFile);
	// if valid return the record and set the offset to next
	// record
	if(retVal==0) return 0;
	ASSERT(retVal == RecordSize);
	FileOffset = FileOffset + RecordSize;
      };
      
      return retVal;
    };
    
    ~TheSwapFile(){
      // delete this swap file
      fclose(SFile);
    };
    
  }; //TheSwapFile
  //-------------------------------------------------------------------------------

 public:
  //The Largebucket structure (Same as bucket chain but canbe swapped to disk)
  class LargeBucket{
  public:
    int size;          // number of elements in the bucket (a list of records), < MAX_BUCKET_SIZE
    int OverallMaxSize;
    BucketChain* chain;    // a list of records in this bucket
    TheSwapFile *SwapFile; // swap file for the bucket
    //constructor
    
    LargeBucket(){
      size = 0;
      OverallMaxSize = 0;
      chain = NULL;
      SwapFile = NULL;
    }
    
    bool BucketOverflow(){
      return (size>=MAX_HASH_SIZE);
    }
    
    void AdvanceBucket(Group *fGroup, int* cnt, int w);
    XxxErrCode DeleteSwapFile(const XxxRecordSchema* RecSchema);
    XxxErrCode FloatTuples(int cnt, BucketChain *fGroup);

    XxxErrCode insertRec(XxxRecord *Rec);

    virtual XxxRecord* GetTopRec(){return NULL;};

    void DeleteTopRec(XxxBool DestroyRec, const XxxRecordSchema* RecSchema){ 
      if(chain!=NULL){
	Entry  *TempEntry = chain->DeleteTop();
	if(TempEntry!=NULL){
	  size--;
	  if(DestroyRec){
	    TempEntry->Rec->Reclaim(RecSchema);
	    //	    delete TempEntry->Rec;
	  }
	  delete TempEntry;
	}
      }
      
    };

    void InitScan(){
      chain->InitScan();
    };
    
    void InitBackwardScan(){
      chain->InitBackwardScan();
    };

    XxxRecord  *GetNext() {
      Entry  *TempEntry = chain->GetNext();
      if (TempEntry == NULL)
	return NULL;
      else 
	return TempEntry->Rec;
    };

    XxxRecord  *GetPrevious() {
      Entry  *TempEntry = chain->GetPrevious();
      if (TempEntry == NULL)
	return NULL;
      else 
	return TempEntry->Rec;
    };



    void CloseScan() {  chain->CloseScan(); };

    XxxBool ScanIsValid() { return chain->ScanIsValid(); };
  };// class LargeBucket


  class OutWindowLargeBucket : public LargeBucket{
  public:
    OutWindowLargeBucket():LargeBucket(){};
    virtual XxxRecord* GetTopRec(XxxRelation *TheRelation,
				 XxxDerivedRelation::CursorInfo* Handle,
				 int Index, XxxBool& Valid);
  };// OutWindowLargeBucket
  
  class InWindowLargeBucket : public LargeBucket{
  public:
    InWindowLargeBucket():LargeBucket(){};
    virtual XxxRecord* GetTopRec(OutWindowLargeBucket *source, 
				 XxxRelation *TheRelation,
				 XxxDerivedRelation::CursorInfo* Handle,
				 int Index, XxxBool& Valid, 
				 unsigned char mode1,
				 unsigned char& NewTupleAddedToInBuffers,
				 unsigned char& IsNewRecord){

      XxxRecord* TmpRec = NULL;

      if((chain==NULL)||(chain->IsEmpty())){
	TmpRec = source->GetTopRec(TheRelation, Handle, Index, Valid);
	if(TmpRec!=NULL){
	  insertRec(TmpRec);
	  source->DeleteTopRec(XXX_FALSE, NULL);
	  if (mode1)
	    NewTupleAddedToInBuffers = 1;
	  IsNewRecord = 1;
	}
	return TmpRec;
      }
      else{
	Entry* E = chain->GetTop();
	IsNewRecord = 0;
	return E->Rec;
      }
    };
  };// InWindowLargeBucket

  

  //----------------------------------------------------------------------------
  /*
    Contains the relevant information to store in the operator handle
    between calls to GetNextRecord.
  */
  class NLFEWJoinInfo {
  public:
    // Is the scan valid?
    XxxBool Valid;
    // Is the left scan is valid
    XxxBool LeftScanValid;
    // Is the left scan is valid
    XxxBool RightScanValid;
    // The indices of the left and right relations in the RelationArray
    int LeftRelIndex, RightRelIndex;
    int LeftAttrIndex, RightAttrIndex;

    /// The left, right, and output group structures
    // I will keep he left and right groups, even they may have only 
    // one record
    Group OutputGroup,
      ExpiredGroup;


    BucketChain FloatGroup;
    // whose turn is next
    XxxBool TurnOnLeftStream;

    unsigned char ChildScannedCnt;     // Cnt of children Scanned by the iteration
    unsigned int TheWindow; // the window between the two streams of the join

    // Merge window join global variables
    int StreamsCnt,
      CurrentStream,
      StartingStream,
      Mode2_CurrentStream,
      inner,
      SkippedStream,
      CurrentlyIn;

    unsigned char IsMode1,
      NewTupleAddedToInBuffers, 
      EndOfMode2Flag;


    XxxBool StreamScanValid[MAX_JOINED_STREAMS];
    XxxRecord* StartRec[MAX_JOINED_STREAMS];
    XxxRecord* LastRec[MAX_JOINED_STREAMS];
    XxxRecord* CandidateForJoin[MAX_JOINED_STREAMS];

    struct{	
      struct timeval SPeriod; 
      struct timeval EPeriod; 
    } CurrentPeriod;


    InWindowLargeBucket** InWindowBuffer;
    OutWindowLargeBucket** OutWindowBuffer;
	
    XxxRelation* *RelationArray;
    XxxRecordSchema* Schema;

  
    /*
      Constructor.
    */
    NLFEWJoinInfo(int RightRelIndex, int LeftRelIndex, int RightAttrIndex, int LeftAttrIndex,
	       XxxValueExprPlan* RightAttrFuncPlan, XxxValueExprPlan* LeftAttrFuncPlan,
	       unsigned int ws, unsigned int wmicros) : 
      Valid(XXX_TRUE), LeftRelIndex(LeftRelIndex), RightRelIndex(RightRelIndex),
      LeftAttrIndex(LeftAttrIndex), RightAttrIndex(RightAttrIndex), OutputGroup(0), 
      FloatGroup(0), ExpiredGroup(0), TurnOnLeftStream(XXX_TRUE), LeftScanValid(XXX_TRUE), RightScanValid(XXX_TRUE){

      TheWindow = ws;
      /*
	TheWindow = (XXX_THISRELENV()->QueryWindowHourValue*60+
	XXX_THISRELENV()->QueryWindowMinValue)*60+
	XXX_THISRELENV()->QueryWindowSecValue;
      */
      ChildScannedCnt = 0;
    };
  };

  /*
    Creates the join info object for the NLFEW run.
  */
  virtual XxxErrCode createJoinInfo(XxxDerivedRelation::CursorInfo* Handle,
				    NLFEWJoinInfo *&JoinInfo) const;
  /*
    Gets the next record from the relation in the RelationArray 
    indexed by the parameter.  Returns Valid == XXX_TRUE if the next
    record was returned.
  */
  virtual XxxErrCode getNextItem(XxxRelation* RelationArray,
				 XxxDerivedRelation::CursorInfo* Handle,
				 int Index, XxxBool& Valid) const;
  /*?????
    Joins the left and right groups by a cross-product (assuming the groups
    ought to be joined) and populates the output group.  In the process, it
    empties the input groups.
  */
  virtual XxxErrCode 
    joinGroups(XxxDerivedRelation::CursorInfo* Handle,
	       XxxRelation* *RelationArray, const XxxRecordSchema* Schema,
	       NLFEWJoinInfo *JoinInfo) const;

  virtual XxxErrCode 
    NLJoinTuples(XxxDerivedRelation::CursorInfo* Handle,
		 XxxRelation* *RelationArray,
		 const XxxRecordSchema* Schema,
		 NLFEWJoinInfo *JoinInfo,
		 XxxRecord *LeftRec,
		 XxxRecord *RightRec) const;
  /*
    Parses the join plan, finding the location of the left and right
    relations and finding the indices of the critical attributes in
    each's records.
  */
  virtual 
    XxxErrCode parseJoinPlan(XxxDerivedRelation::CursorInfo* Handle,
			     XxxValueExprPlan* &LeftAttrFuncP, 
			     int &LeftAttrIndex,
			     int &LeftRelIndex,
			     XxxValueExprPlan* &RightAttrFuncP, 
			     int &RightAttrIndex,
			     int &RightRelIndex) const;

 public:

  double LeftWindowSize,  // The window size for left buffer
    RightWindowSize; // The window size for right buffer
  /*
    Returns a string representation of this plan operator.
  */
  virtual const char *PrintString()  const { return "NLFEW SPJ" ;}
  /*
    Gets the next record from this plan operator.
  */
  virtual XxxErrCode GetNextRecord(XxxSetCursor& SCsr,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
				   XxxRecord*& ReturnRec) const;
  /*
    Initializes this plan operator.
  */
  virtual XxxErrCode InitHandle(XxxDerivedRelation::CursorInfo* Handle,
				XxxRelation* *RelationArray) const;
  /*
    Closes the scan this plan operator offers.
  */
  virtual XxxErrCode CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				 XxxRelation* *RelationArray) const;

  virtual XxxErrCode checkFinal(XxxDerivedRelation::CursorInfo* Handle,
				XxxCoreADTClass::BoolValue& Satisfied) const;
  /*
    Constructor.
  */
  RelNLFEWPlanOp(int NumChildren, 
	      RelPlanOp* Children[],
	      RelCatalogInfo* CatInfo,
	      XxxValExprPlanList* ExprPlanList,
	      XxxBooleanExpressionPlan** SelPredPlans,
	      XxxBooleanExpressionPlan** JoinPredPlans,
	      XxxBooleanExpressionPlan** ResidPredPlans,
	      XxxBooleanExpressionPlan* InitPlan,
	      double LWSize,
	      double RWSize);
  /*
    Destructor.
  */
  virtual ~RelNLFEWPlanOp() {}
  
  virtual void PrintNodeInfo(ostream &OStr) const;

  void RecursiveNestedLoop(XxxDerivedRelation::CursorInfo* Handle, int CurrStrm) const;

  void JoinInWindowBuffers(XxxDerivedRelation::CursorInfo* Handle, 
			   XxxRecord *TmpRec) const;

  void JoinInWindowBuffers(XxxDerivedRelation::CursorInfo* Handle) const;

   
  // CurrentTimeStamp-TS2
  double TimeStampsDiff(struct timeval TS2) const{
    struct timezone Temp; 
    struct timeval TS;
    gettimeofday(&TS, &Temp);
    return ((TS.tv_sec-TS2.tv_sec)+
	    (TS.tv_usec-TS2.tv_usec)/1000000.0);
  };

  // TS1>TS2?
  XxxBool TimeStampsGreater(struct timeval TS1, struct timeval TS2) const{
    double diff = ((TS1.tv_sec-TS2.tv_sec)+
		   (TS1.tv_usec-TS2.tv_usec)/1000000.0);
    return ((diff>0?XXX_TRUE:XXX_FALSE));
  };
  
};




#endif /* REL_NLFEWPLANOP_H */
//EMostafa
