#ifndef XXX_RELATION_H
#define XXX_RELATION_H

#include "record.h"
#include "expr.h"
#include "exprplan.h"
#include "relcat.h"


XXX_EXTERN_CLASS_DECL(RelDerivedPlanOp);
XXX_EXTERN_CLASS_DECL(RelBasePlanOp);
XXX_EXTERN_CLASS_DECL(RelExecParameters);
XXX_EXTERN_CLASS_DECL(RelAccessPlan);

/**
 * Root class for all relations. This is essentially a main-memory
 * "handle" onto a relation, whether it is a stored table on disk,
 * or a derived table which is the result of some operation
 */
class XxxRelation : public XxxRecordSet {
 protected:
 public:
  //// Cursor over relation
  class CommonCursorInfo : public XxxSetCursor::CursorState { 
  public:
    XxxRecordSchema* RecordSchema;
    XxxBooleanExpressionPlan* Filter;
    const RelExecParameters *Params; 
    
    // these should all be const: put it in sometime
    // parameters (runtime args) for this rel
    // the "join environment" in which this cursor is created
    const XxxRecord  **SiblingRecs;         // the input records from siblings 
    const XxxRecordSchema** SiblingSchema;  // schemas of siblings
    int CurrJoinPos;                  // position of this relation
    XxxBool ExactMatch; // does this cursor return exact matches
    
    CommonCursorInfo(XxxRecordSchema* Schema,
		     XxxBooleanExpressionPlan* Expr,
		     XxxBool Exact,
		     const RelExecParameters* Parameters = NULL,
		     const XxxRecord** SibRecs = NULL,
		     const XxxRecordSchema** SibSchema = NULL,
		     int CurrentJoinPos = 0) {
      RecordSchema = Schema;
      Filter = Expr;
      Params = Parameters;
      SiblingRecs = SibRecs;
      SiblingSchema = SibSchema;
      CurrJoinPos = CurrentJoinPos;
      ExactMatch = Exact;
     }
    virtual ~CommonCursorInfo() {}
   };
  
  const RelCatalogInfo*            CatalogInfo;
  XxxRelation(const RelCatalogInfo* Cat) { CatalogInfo = Cat; }
  virtual ~XxxRelation() {}

  // scan with filter predicate, and assuming a join environment in
  // which a partial join result has been computed. 
  virtual XxxErrCode InitCursor1(XxxSetCursor&, 
				 XxxBooleanExpressionPlan*,
				 const XxxRecord* Rec[] = NULL,
				 const XxxRecordSchema* Schema[] = NULL,	
				 const RelExecParameters* Parameters = NULL,
				 int SibArity = 1, 
				 int CurrJoinPos = 0) 
    const = 0;
  virtual XxxErrCode InitCursor(XxxSetCursor& Csr) const {
    return InitCursor1(Csr, NULL, NULL);
   }
  virtual XxxErrCode NextItem(XxxSetCursor&, XxxRecord*& ) const = 0;
  virtual XxxErrCode CloseCursor(XxxSetCursor&) const = 0;

  virtual XxxErrCode AddIndex(const char *IndxName, 
			      XxxValueExpression  *Expr,
			      XxxIndexType IndxType,
			      XxxBool Unique, 
			      const void *IndxColMetaInfo)
  {
    XXX_NOTE_ERROR (XXX_INCOMPL, "Adding index incompl");
    return XXX_NOT_OK;
  }
  
  virtual XxxErrCode DeleteRecords(XxxBooleanExpressionPlan *FilterPlan,
				   int &NumDeleted) {
    // This is a default implementation; it is implemented for shore relations
    XXX_NOTE_ERROR(XXX_INCOMPL, "Deleting records from relations incomplete.");
    return XXX_NOT_OK;
  }
  virtual XxxErrCode UpdateRecords(XxxBooleanExpressionPlan *FilterPlan,
				   XxxValExprPlanList* TargetPlanList,
				   int &NumUpdated) {
    // This is a default implementation; it is implemented for shore relations
    XXX_NOTE_ERROR(XXX_INCOMPL, "Updating records from relations incomplete.");
    return XXX_NOT_OK;
  }

  // this is used to change logging properties, set up bulk loads, etc
  virtual XxxErrCode PrepareForLoad() = 0;
 };

/**
 * Base class for all stored relations
 */
class XxxStoredRelation : public XxxRelation {
 public:
  enum AccessMode { UNDECIDED, FILESCAN, INDEXED /*Begin: Moustafa*/, SSCAN /*End: Moustafa*/};
  //// different access modes
  typedef enum AccessMode AccessMode;

  AccessMode AccMode;
  RelBasePlanOp* AccInfo; // any relation created from a query has this

  XxxStoredRelation(RelCatalogInfo* Cat) : XxxRelation(Cat) {
    AccMode = UNDECIDED; // default
    AccInfo = NULL;
   }
  virtual ~XxxStoredRelation() {
    // if (AccInfo) delete AccInfo;
   }
  // this is used to change logging properties, set up bulk loads, etc
  virtual XxxErrCode PrepareForLoad() {
    return XXX_OK;
   }
 };

/**
 * Base class for all derived relations
 */
class XxxDerivedRelation : public XxxRelation {
 protected:
  RelDerivedPlanOp* planOp;
  XxxRelation* *relInputs;
  int inputArity;
  
  virtual XxxErrCode addItem(XxxRecord* const& AddRecord) {
    return XXX_NOT_OK;
   }
 public:
  //// Cursor over derived relation
  class CursorInfo : public XxxRelation::CommonCursorInfo { 
   public:
    XxxValuesArray  RecValuesArray; // for operators to create records
    int ArrayLen;                   // input arity

    XxxSetCursor* ChildCursorArray; // cursors over the children
    XxxRecord  **ChildRecs;         // the input records from the children
    XxxRecordSchema** ChildSchema;  // schemas of children

    void *OperatorHandle;           // dummy pointer for op cursor info

    CursorInfo(XxxRecordSchema* Schema,
	       XxxBooleanExpressionPlan* Expr,
	       XxxBool Exact,
	       int InputArity,
	       const RelExecParameters* Parameters = NULL,
	       const XxxRecord** SibRecs = NULL,
	       const XxxRecordSchema** SiblingSchema = NULL,
	       int CurrentJoinPos = 0);
    virtual ~CursorInfo();
   };

  XxxDerivedRelation(RelDerivedPlanOp* Oper, RelCatalogInfo* Cat,
		     const XxxValueEnv* Params);
  virtual ~XxxDerivedRelation();

  // scan with filter predicate
  virtual XxxErrCode InitCursor1(XxxSetCursor&, 
				 XxxBooleanExpressionPlan*,
				 const XxxRecord* Rec[] = NULL,
				 const XxxRecordSchema* Schema[] = NULL,
				 const RelExecParameters* Parameters = NULL,
				 int SibArity = 1, 
				 int   CurrJoinPos = 0) const;
  virtual XxxErrCode InitCursor(XxxSetCursor& Csr) const {
    return InitCursor1(Csr, NULL);
   }
  virtual XxxErrCode NextItem(XxxSetCursor&, XxxRecord*& ) const;
  virtual XxxErrCode CloseCursor(XxxSetCursor&) const;

  virtual XxxErrCode RemoveItem(XxxRecord* const& DelRecord) {
    return XXX_NOT_OK;
   }
  // this is used to change logging properties, set up bulk loads, etc
  virtual XxxErrCode PrepareForLoad() {
    return XXX_NOT_OK;
   }
 };

#endif
