//
// A RelPlanOp is a run-time structure used to execute a relational plan.
//

#ifndef REL_PLANOP_H
#define REL_PLANOP_H

#include <string.h>
#include "reladt.h"
#include "relconfig.h"
#include "relerror.h"
#include "expr.h"
#include "exprplan.h"
#include "values.h"
#include "relcat.h"
#include "relexec.h"
#include "relation.h"
#include "prd-sort.h"
#include "indexmatch.h"
#include "xxx_thread.h" 


#define XXX_MAX_SHAREDBUFFER_RECORDS 1000

XXX_EXTERN_CLASS_DECL(RelOperator);

/**
 *  Class representing a run-time plan component
 */
class RelPlanOp {
 protected:
  RelPlanOp* *childPlanArray;
  int childPlanArrayLen;
 public:
  RelCatalogInfo* CatalogInfo;

  RelPlanOp* GetChild(int i) const { 
    if ( i >= Arity()) return NULL;
    else return childPlanArray[i]; }
  virtual int Arity() const { return childPlanArrayLen;}

  RelPlanOp(int NumChildren, RelPlanOp* Children[],
	    RelCatalogInfo* CatInfo);
  virtual ~RelPlanOp();

  // The following two functions are used to print a planOp tree
  virtual void TreePrint (ofstream &OStr, int &curPlanId, int level) 
    const;
  virtual void NodePrint (ostream &OStr, int curBlockId, int Level, 
			  int numChildren, int *childrenIds) const = 0;
  // create a relation object out of this operator
  virtual XxxErrCode MakeRelation(const XxxValueEnv* Params,
				  XxxRelation*& ReturnRel) = 0;

  virtual void PrintThis(ostream& OStr)  const 
   { OStr << "REL PLAN OP"; }
  virtual void Print(ostream& OStr)  const;
  virtual const char *PrintString()  const = 0;
  virtual XxxBool IsFullFileScan() const { return XXX_FALSE;}
  //Begin: Moustafa
  virtual XxxBool IsStreamScan() const 
    { return XXX_FALSE;} // will help identify the type of scan plan
  //End: Moustafa
 };

/**
 // Base Relation plan operator 
 */
class RelBasePlanOp: public RelPlanOp {
 protected:
  char relName[REL_RELNAME_MAX];
 public:
  RelBasePlanOp(const char* Name,
		RelCatalogInfo* CatInfo)
    : RelPlanOp(0, NULL, CatInfo) { 
      strncpy(relName, Name, REL_RELNAME_MAX); 
      relName[REL_RELNAME_MAX-1] = 0;
     }
    
  virtual ~RelBasePlanOp() {}
  const char*RelName() const { return relName;}
  virtual void PrintThis(ostream& OStr)  const
   { OStr << relName; }
  virtual const char *PrintString()  const { return relName ;}
  // create a relation object out of this operator
  virtual XxxErrCode MakeRelation(const XxxValueEnv* Params,
				  XxxRelation*& ReturnRel);
  virtual void NodePrint (ostream &OStr, int curBlockId, int Level, 
			  int numChildren, int *childrenIds) const;
 };

/**
 */
class RelFScanPlanOp: public RelBasePlanOp {
 public:
  XxxBooleanExpressionPlan* Selection;
  RelFScanPlanOp(const char* Name,
		 RelCatalogInfo* CatInfo,
		 XxxBooleanExpressionPlan* Filter)
    : RelBasePlanOp(Name, CatInfo) 
     { Selection = Filter; }
  virtual ~RelFScanPlanOp() {}
  virtual void PrintThis(ostream& OStr)  const
   {  OStr << "FSCAN:" << relName; }
  // create a relation object out of this operator
  virtual XxxErrCode MakeRelation(const XxxValueEnv* Params,
				  XxxRelation*& ReturnRel);
  virtual void NodePrint (ostream &OStr, int curBlockId, int Level, 
			  int numChildren, int *childrenIds) const;
  virtual XxxBool IsFullFileScan() const;
 };

//SMostafa: Stream Scan operator(RelSScanPlanOp)
 class RelSScanPlanOp: public RelBasePlanOp {
 public:
  XxxBooleanExpressionPlan* Selection;
  RelSScanPlanOp(const char* Name, 
		 RelCatalogInfo* CatInfo, 
		 XxxBooleanExpressionPlan* Filter)
    : RelBasePlanOp(Name, CatInfo) 
    { Selection = Filter; }
  virtual ~RelSScanPlanOp() {}
  virtual void PrintThis(ostream& OStr)  const
    {  OStr << "SSCAN:" << relName; }
  // create a relation object out of this operator
  virtual XxxErrCode MakeRelation(const XxxValueEnv* Params,
				  XxxRelation*& ReturnRel);
  virtual void NodePrint (ostream &OStr, int curBlockId, int Level, 
			  int numChildren, int *childrenIds) const;
  virtual XxxBool IsStreamScan() const {return XXX_TRUE;} ;// will help identify the type of scan plan
};
//EMostafa

/**
  */
class RelIScanPlanOp: public RelBasePlanOp {
 public:
  XxxBooleanExpressionPlan* IndxPlan;
  XxxBooleanExpressionPlan* ResidPlan;
  const XxxMatchInfo* MatchedInfo;
  char IndxName[REL_RELNAME_MAX];
  RelIScanPlanOp(const char* Name,
		 RelCatalogInfo* CatInfo,
		 XxxBooleanExpressionPlan* IndxPreds,
		 XxxBooleanExpressionPlan* ResidPreds,
		 const XxxMatchInfo* MInfo,
		 const char* IndexName)
    : RelBasePlanOp(Name, CatInfo) { 
      IndxPlan = IndxPreds;
      ResidPlan = ResidPreds;
      MatchedInfo =  MInfo;
      strncpy(IndxName, IndexName, REL_RELNAME_MAX);
     }
  virtual ~RelIScanPlanOp() {}
  virtual void PrintThis(ostream& OStr)  const
   {  OStr << "ISCAN:" << relName; }
  // create a relation object out of this operator
  virtual XxxErrCode MakeRelation(const XxxValueEnv* Params,
				  XxxRelation*& ReturnRel);
  virtual void NodePrint (ostream &OStr, int curBlockId, int Level, 
			  int numChildren, int *childrenIds) const;
 };

// TODO:L2 flesh out the Tmp plan.
// It should keep around the materialized temp relation
// after the first time it is materialized, and reuse
// it on subsequent accesses.
/**
 * Makes a temporary relation (breaks the pipeline)
 */
class RelTmpScanPlanOp: public RelPlanOp {
 public:
  XxxBooleanExpressionPlan* Selection;
  RelTmpScanPlanOp(RelCatalogInfo* CatInfo,
		   RelPlanOp* ChildPlan,
		   XxxBooleanExpressionPlan* Filter)
    : RelPlanOp(1, &ChildPlan, CatInfo)
     { Selection = Filter;}
  virtual ~RelTmpScanPlanOp() {}
  virtual const char *PrintString()  const { return "TMP";}
  virtual void PrintThis(ostream& OStr)  const
   {  OStr << "TMPFSCAN:"; }
  // create a relation object out of this operator
  virtual XxxErrCode MakeRelation(const XxxValueEnv* Params,
				  XxxRelation*& ReturnRel);
  virtual void NodePrint (ostream &OStr, int curBlockId, int Level, 
			  int numChildren, int *childrenIds) const;
 };

/**
 * Root class for all derived plan operators.
 */
class RelDerivedPlanOp : public RelPlanOp {
 protected:
 public:
  RelDerivedPlanOp(int NumChildren, RelPlanOp* Children[],
		   RelCatalogInfo* CatInfo) 
   : RelPlanOp(NumChildren, Children, CatInfo) {
     // Set the flag which identifies the relation as derived
     CatalogInfo->ImplInfo()->RelType = REL_DERIVED;
    }
  virtual ~RelDerivedPlanOp() {}

  // create a relation object out of this operator
  virtual XxxErrCode MakeRelation(const XxxValueEnv* Params,
				  XxxRelation*& ReturnRel);
  // These methods are called from InitCursor and CloseCursor
  virtual XxxErrCode InitHandle(XxxDerivedRelation::CursorInfo* Handle,
				XxxRelation* *RelationArray)  const = 0;
  virtual XxxErrCode CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				 XxxRelation* *RelationArray) const = 0;
  virtual XxxErrCode GetNextRecord(XxxSetCursor& SCsr,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
				   XxxRecord*& ReturnRec) const = 0;
  virtual void NodePrint (ostream &OStr, int curBlockId, int Level, 
			  int numChildren, int *childrenIds) const;
 };

/**
  // Class of join algorithms
  */
class RelSPJPlanOp : public RelDerivedPlanOp {
 protected:
  // Target list for output projection
  XxxValExprPlanList* ExprPlanList;
  XxxBooleanExpressionPlan** SelPredPlans;
  XxxBooleanExpressionPlan** JoinPredPlans;
  XxxBooleanExpressionPlan** ResidPredPlans;
  XxxBooleanExpressionPlan* InitPlan;
  // create a record once all the info is in the handle
  virtual XxxErrCode createRecord(XxxDerivedRelation::CursorInfo* Handle,
				  XxxRelation* *RelationArray,
				  const XxxRecordSchema* Schema,
				  XxxRecord*& ReturnRec) const;
 public:
  virtual void PrintThis(ostream& OStr)  const;
  RelSPJPlanOp(int NumChildren, 
	       RelPlanOp* Children[],
	       RelCatalogInfo* CatInfo,
	       XxxValExprPlanList* ExprPlanList,
	       XxxBooleanExpressionPlan** SelPredPlans,
	       XxxBooleanExpressionPlan** JoinPredPlans,
	       XxxBooleanExpressionPlan** ResidPredPlans,
	       XxxBooleanExpressionPlan* InitPlan);

  virtual ~RelSPJPlanOp();
  virtual void NodePrint (ostream &OStr, int curBlockId, int Level, 
			  int numChildren, int *childrenIds) const;
  virtual void PrintNodeInfo(ostream &OStr) const;
};

/**
  */
class RelNLPlanOp : public RelSPJPlanOp {
 public:
  class NLJoinInfo { public:
    XxxBool Valid;
    int NLPos;
    NLJoinInfo() { NLPos = 0; Valid = XXX_TRUE;}
    virtual ~NLJoinInfo() {}
   };

  virtual NLJoinInfo* createJoinInfo() const {
    return new NLJoinInfo();
   }
  // create a record once all the info is in the handle
  virtual XxxErrCode reclaimCurrent(XxxDerivedRelation::CursorInfo* Handle,
				    int NLPos) const = 0;

  //Begin: Moustafa
  virtual XxxErrCode reclaimCurrent(XxxDerivedRelation::CursorInfo* Handle,
  				    int NLPos,
  				    const XxxRecordSchema* RecSchema) const = 0;
  //End: Moustafa
  virtual XxxErrCode getNextItem(XxxRelation* *RelationArray,
				 XxxDerivedRelation::CursorInfo* Handle,
				 int NLPos, XxxBool& Valid) const = 0;
  virtual XxxErrCode openChildScan(XxxDerivedRelation::CursorInfo* Handle,
				   XxxRelation* *RelationArray,
				   int NLPos) const = 0;
  virtual XxxErrCode closeChildScan(XxxDerivedRelation::CursorInfo* Handle,
				    XxxRelation* *RelationArray,
				    int NLPos) const = 0;
  virtual XxxErrCode checkFinal(XxxDerivedRelation::CursorInfo* Handle,
				XxxCoreADTClass::BoolValue& Satisfied) const 
				 { Satisfied.Value = XXX_TRUE; return XXX_OK;}

 public:
  virtual XxxErrCode GetNextRecord(XxxSetCursor& SCsr,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
				   XxxRecord*& ReturnRec) const;
  virtual XxxErrCode InitHandle(XxxDerivedRelation::CursorInfo* Handle,
				XxxRelation* *RelationArray) const;
  virtual XxxErrCode CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				 XxxRelation* *RelationArray) const;

  RelNLPlanOp(int NumChildren, 
	       RelPlanOp* Children[],
	       RelCatalogInfo* CatInfo,
	       XxxValExprPlanList* ExprPlanList,
	       XxxBooleanExpressionPlan** SelPredPlans,
	       XxxBooleanExpressionPlan** JoinPredPlans,
	       XxxBooleanExpressionPlan** ResidPredPlans,
	       XxxBooleanExpressionPlan* InitPlan) 
    : RelSPJPlanOp(NumChildren, Children, CatInfo, ExprPlanList, 
		   SelPredPlans, JoinPredPlans, ResidPredPlans, 
		   InitPlan) {}
  virtual ~RelNLPlanOp() {}
 };

/**
 // Tuple-at-a-time NL join
 */
class RelTNLPlanOp : public RelNLPlanOp {
 protected:
  // create a record once all the info is in the handle
  virtual XxxErrCode reclaimCurrent(XxxDerivedRelation::CursorInfo* Handle,
				    int NLPos) const;
  //Begin: Moustafa
  virtual XxxErrCode reclaimCurrent(XxxDerivedRelation::CursorInfo* Handle,
  				    int NLPos,
  				    const XxxRecordSchema* RecSchema) const;
  //End: Moustafa

  virtual XxxErrCode getNextItem(XxxRelation* *RelationArray,
				 XxxDerivedRelation::CursorInfo* Handle,
				 int NLPos, XxxBool& Valid) const;
  virtual XxxErrCode openChildScan(XxxDerivedRelation::CursorInfo* Handle,
				   XxxRelation* *RelationArray,
				   int NLPos) const;
  virtual XxxErrCode closeChildScan(XxxDerivedRelation::CursorInfo* Handle,
				    XxxRelation* *RelationArray,
				    int NLPos) const;

 public:
  RelTNLPlanOp(int NumChildren, 
	       RelPlanOp* Children[],
	       RelCatalogInfo* CatInfo,
	       XxxValExprPlanList* ExprPlanList,
	       XxxBooleanExpressionPlan** SelPredPlans,
	       XxxBooleanExpressionPlan** JoinPredPlans,
	       XxxBooleanExpressionPlan** ResidPredPlans,
	       XxxBooleanExpressionPlan* InitPlan) 
    : RelNLPlanOp(NumChildren, Children, CatInfo, ExprPlanList, 
		  SelPredPlans, JoinPredPlans, ResidPredPlans, 
		  InitPlan) {}
  virtual ~RelTNLPlanOp() {}
  virtual const char *PrintString()  const { return "TupleNL SPJ" ;}
  virtual void PrintNodeInfo(ostream &OStr) const;
 };

/**
 // Page-at-a-time NL join
 */
class RelPNLPlanOp : public RelTNLPlanOp {
 protected:
  //// 
  class BlockRelTuples { public:
    // eventually, this should be based on PageSize and TupleSize
    int BlockSize; // max number of records in block
    int NumRecords; // actual number of records in block
    int CurrentPos; // position of current scan thru the records
    XxxRecord** Records;
    BlockRelTuples(int Size);
    ~BlockRelTuples();
   };
  ////
  class PNLJoinInfo : public RelNLPlanOp::NLJoinInfo {
   public:
    // a block of tuples for each relation
    BlockRelTuples** RelBlocks;
    unsigned int BlockCardinality;
    int Arity;
    PNLJoinInfo(int NumChildren, unsigned int BlockCard);
    virtual ~PNLJoinInfo();
   };

  // number of tuples in each "block"/"page" of the NL join
  unsigned int blockCardinality;

  virtual NLJoinInfo* createJoinInfo() const {
    return new PNLJoinInfo(this->Arity(), blockCardinality);
   }

  // create a record once all the info is in the handle
  virtual XxxErrCode reclaimCurrent(XxxDerivedRelation::CursorInfo* Handle,
				    int NLPos) const;
  //Begin: Moustafa
  virtual XxxErrCode reclaimCurrent(XxxDerivedRelation::CursorInfo* Handle,
				    int NLPos,
				    const XxxRecordSchema* RecSchema) const;
  //End: Moustafa
  
  virtual XxxErrCode getNextItem(XxxRelation* *RelationArray,
				 XxxDerivedRelation::CursorInfo* Handle,
				 int NLPos, XxxBool& Valid) const;
  virtual XxxErrCode openChildScan(XxxDerivedRelation::CursorInfo* Handle,
				   XxxRelation* *RelationArray,
				   int NLPos) const;
  virtual XxxErrCode closeChildScan(XxxDerivedRelation::CursorInfo* Handle,
				    XxxRelation* *RelationArray,
				    int NLPos) const;
  virtual XxxErrCode checkFinal(XxxDerivedRelation::CursorInfo* Handle,
				XxxCoreADTClass::BoolValue& Satisfied) const;

 public:
  static const int DEFAULT_PNL_BLOCKCARD; // 100

  RelPNLPlanOp(int NumChildren, 
	       RelPlanOp* Children[],
	       RelCatalogInfo* CatInfo,
	       XxxValExprPlanList* ExprPlanList,
	       XxxBooleanExpressionPlan** SelPredPlans,
	       XxxBooleanExpressionPlan** JoinPredPlans,
	       XxxBooleanExpressionPlan** ResidPredPlans,
	       XxxBooleanExpressionPlan* InitPlan,
	       unsigned int BlockCard) 
    : RelTNLPlanOp(NumChildren, Children, CatInfo, ExprPlanList, 
		   SelPredPlans, JoinPredPlans, ResidPredPlans, 
		   InitPlan) { blockCardinality = BlockCard;}
  virtual ~RelPNLPlanOp() {}
  virtual const char *PrintString()  const { return "PageNL SPJ" ;}
  virtual void PrintNodeInfo(ostream &OStr) const;
 };

/**
 * The Sort-merge Join plan operator.  This operator assumes it has two
 * input streams or records, both sorted by a single attribute.  The join
 * condition must be an equality of these two attributes, e.g., 
 *                        ... WHERE A.id = B.id
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 8/12/97
 */
class RelSMPlanOp : public RelSPJPlanOp {
protected:
  /**
   * Encapsulates the concept of a record group; maintains a linked-
   * list of records and information on which attribute the join
   * is based upon.
   */
  class Group {
  private:
    /// The linked-list node structure
    struct Member {
      XxxRecord *Rec;
      Member *Next;
    };
    /// The head of the linked list
    Member *members;
    /// The cursor in scans
    Member *currentMember;
    /// The attribute of the records in the equality join condition
    int attrIndex;
    /// Whether or not the current scan is valid
    XxxBool scanValid;
  public:
    /**
     * Constructor.
     */
    Group(int attrIndexNum) : members(NULL), currentMember(NULL),
                              attrIndex(attrIndexNum), scanValid(XXX_FALSE) { }
    /**
     * Destructor.
     */
    ~Group() {
      FlushAllRecords(XXX_FALSE);
    }
    /**
     * Returns the index of the critical attribute of the contained records.
     */
    int AttrIndex() { return attrIndex; }
    /**
     * Returns XXX_TRUE if this group has no records in it.
     */
    XxxBool IsEmpty() { return members == NULL; }
    /**
     * Adds a record to this group.
     */
    void AddRecord(XxxRecord *Rec) {
      ASSERT(Rec);
      Member *NewMember = new Member;
      NewMember->Rec = Rec;
      NewMember->Next = members;
      members = NewMember;
    }
    /**
     * Removes all records from this group.
     * @param DestroyRecs If XXX_TRUE, the records within will be explicitly
     *                    destroyed; otherwise, only the linked-list structure
     *                    will be destroyed.
     */
    void FlushAllRecords(XxxBool DestroyRecs) {
      Member *cur = members, *tmp;
      while (cur) {
	if (DestroyRecs)
	  delete cur->Rec;
	tmp = cur->Next;
	delete cur;
	cur = tmp;
      }
      scanValid = XXX_FALSE;
      members = NULL;
    }
    /**
     * Initializes a scan over the group.
     */
    void InitScan() {
      currentMember = members; 
      scanValid = XXX_TRUE;
    }
    /**
     * Returns XXX_TRUE if the current scan is valid.
     */
    XxxBool ScanIsValid() { return (scanValid); }
    /**
     * Gets the next record from the scan.  Returns NULL if the scan is
     * now invalid.
     */
    XxxRecord *GetNext() {
      if (currentMember == NULL) {
	scanValid = XXX_FALSE;
	return NULL;
      } else {
	XxxRecord *Rec = currentMember->Rec;
	currentMember = currentMember->Next;
	scanValid = XXX_TRUE;
	return Rec;
      }
    }
    /**
     * Closes the current scan.
     */
    void CloseScan() { currentMember = NULL; scanValid = XXX_FALSE; }
  };

  /**
   * Contains the relevant information to store in the operator handle
   * between calls to GetNextRecord.
   */
  class SMJoinInfo {
  public:
    /// Is the scan valid?
    XxxBool Valid;
    /// The indices of the left and right relations in the RelationArray
    int LeftRelIndex, RightRelIndex;
    /// The left, right, and output group structures
    Group LeftGroup, 
          RightGroup, 
          OutputGroup;
    /**
     * Constructor.
     */
    SMJoinInfo(int RightRelIndex, int LeftRelIndex,
	       int RightAttrIndex, int LeftAttrIndex) : 
      Valid(XXX_TRUE), LeftRelIndex(LeftRelIndex), RightRelIndex(RightRelIndex),
      LeftGroup(LeftAttrIndex), RightGroup(RightAttrIndex), OutputGroup(0) { }
  };

  /**
   * Creates the join info object for the SM run.
   */
  virtual XxxErrCode createJoinInfo(XxxDerivedRelation::CursorInfo* Handle,
				    SMJoinInfo *&JoinInfo) const;
  /**
   * Gets the next record from the relation in the RelationArray 
   * indexed by the parameter.  Returns Valid == XXX_TRUE if the next
   * record was returned.
   */
  virtual XxxErrCode getNextItem(XxxRelation* *RelationArray,
				 XxxDerivedRelation::CursorInfo* Handle,
				 int Index, XxxBool& Valid) const;
  /**
   * Joins the left and right groups by a cross-product (assuming the groups
   * ought to be joined) and populates the output group.  In the process, it
   * empties the input groups.
   */
  virtual XxxErrCode 
    joinGroups(XxxDerivedRelation::CursorInfo* Handle,
	       XxxRelation* *RelationArray, const XxxRecordSchema* Schema,
	       SMJoinInfo *JoinInfo) const;
  /**
   * Parses the join plan, finding the location of the left and right
   * relations and finding the indices of the critical attributes in
   * each's records.
   */
  virtual 
    XxxErrCode parseJoinPlan(XxxDerivedRelation::CursorInfo* Handle,
			     int &LeftAttrIndex, int &LeftRelIndex,
			     int &RightAttrIndex, int &RightRelIndex) const;
  /**
   * Compares a left and right record and returns XXX_TRUE if they are
   * equal with respect to their critical attributes.
   */
  virtual XxxBool isEqual(XxxRecord *LeftRec, int LeftAttrIndex,
			  XxxRecordSchema *LeftSchema,
			  XxxRecord *RightRec, int RightAttrIndex,
			  XxxRecordSchema *RightSchema) const;
  /**
   * Compares a left and right record and returns XXX_TRUE if the left is
   * less than the right with respect to their critical attributes.
   */
  virtual XxxBool isLessThan(XxxRecord *LeftRec, int LeftAttrIndex,
			     XxxRecordSchema *LeftSchema,
			     XxxRecord *RightRec, int RightAttrIndex,
			     XxxRecordSchema *RightSchema) const;
public:
  /**
   * Returns a string representation of this plan operator.
   */
  virtual const char *PrintString()  const { return "SM SPJ" ;}
  /**
   * Gets the next record from this plan operator.
   */
  virtual XxxErrCode GetNextRecord(XxxSetCursor& SCsr,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
				   XxxRecord*& ReturnRec) const;
  /**
   * Initializes this plan operator.
   */
  virtual XxxErrCode InitHandle(XxxDerivedRelation::CursorInfo* Handle,
				XxxRelation* *RelationArray) const;
  /**
   * Closes the scan this plan operator offers.
   */
  virtual XxxErrCode CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				 XxxRelation* *RelationArray) const;
  /**
   * Constructor.
   */
  RelSMPlanOp(int NumChildren, 
	      RelPlanOp* Children[],
	      RelCatalogInfo* CatInfo,
	      XxxValExprPlanList* ExprPlanList,
	      XxxBooleanExpressionPlan** SelPredPlans,
	      XxxBooleanExpressionPlan** JoinPredPlans,
	      XxxBooleanExpressionPlan** ResidPredPlans,
	      XxxBooleanExpressionPlan* InitPlan);
  /**
   * Destructor.
   */
  virtual ~RelSMPlanOp() {}

  virtual void PrintNodeInfo(ostream &OStr) const;
};


/**
 * The Symmetric Hash Join plan operator.  
 * @author Tok Wee Hyong, tokweehy@comp.nus.edu.sg
 * School of Computing, National University Of Singapore
 * @version 9/19/99
 */
class RelHashJoinPlanOp : public RelSPJPlanOp {
protected:
  /**
   * Encapsulates the hash table
   */
   class HashTable {
	private:
		int hashsize;

		// buckets used for chaining
		struct item {
			XxxRecord *Rec;
			item *Next;
		};
		typedef item *itemptr;
		itemptr *table;
		itemptr curptr;

		int attrIndex;
		XxxBool scanValid;

	public:
	  /**
	   * Constructor .
	   */
		HashTable(const int tablesize, const int attrIndexNum) : 
		  curptr(NULL), attrIndex(attrIndexNum), scanValid(XXX_FALSE) {
		  // allocate enough contiguous storage for hash table
		  int i;
		  
		  hashsize = tablesize;
		  table = new itemptr[tablesize];
		  
		  for (i=0; i < hashsize; i++) 
		    table[i] = NULL;
		}

	  /**
	   * Destructor
	   */
	   ~HashTable() {
		 delete [] table;
	   }

	  /**
	   * Add a XxxRecord to the hashtable
	   */
	   void Insert(XxxRecord *newRec, XxxRecordSchema *RecSchema) {
		// insert an item into the hash table using chaining
		int pos;
		//Bjarke, sep. 13 2001: added const qualifier to variable
		const char *key;
		itemptr newItem;
		
		ASSERT(newRec);

		newRec->GetField(attrIndex, RecSchema, key);

		//cout << "value is " << ((XxxCoreADTClass::IntValue*)key)->Value << endl;
		// pos = Hash(((XxxCoreADTClass::StrValue*)key)->Value);
		// cout << "inserting" << ((XxxCoreADTClass::IntValue*)key)->Value << endl;
		pos = ((XxxCoreADTClass::IntValue*)key)->Value % hashsize;
		//cout << "pos : " << pos << endl;

		newItem = new item;
		newItem->Rec  = newRec;

		if ( table[pos] ) 
		  newItem->Next = table[pos];
		else
		  newItem->Next = NULL;
		 
		table[pos] = newItem; 
	   }

	   /**
	    * Compute the hash key
		*/
	   int Hash(char *key) {
		 unsigned h = 0;
				 
         while (*key) h += *key++;
		 return h % hashsize;
	   }

	   /**
	    * Check whether hash table is empty 
		*/
	   bool IsEmpty() {	return ( table == NULL ); }

	   void printTable() {
		int i;
		itemptr cur;

		for (i=0; i < hashsize; i++ ) {
		  if ( table[i] != NULL ) {
			cur = table[i];
			while ( cur ) {
				printf("[%s] ",cur->Rec);
				cur = cur->Next;
			} // while cur
			printf("\n");
		  } // if
		}  // for
	   }

	   	   
	   /**
	    * Probe the hash table to see whether the hash value 
		* has a valid ptr
		*/
	   XxxBool probe(XxxRecord *probeRec, XxxRecordSchema *RecSchema) {
		 int hashvalue;
		//Bjarke, sep. 13 2001: added const qualifier to variable
		 const char *key;
		 
		 ASSERT(probeRec);
		 probeRec->GetField(attrIndex, RecSchema, key);
		 		 
		 curptr = NULL;
		 //hashvalue = Hash(key);
		 //hashvalue = Hash(((XxxCoreADTClass::StrValue*)key)->Value);
		 hashvalue = ((XxxCoreADTClass::IntValue*)key)->Value % hashsize;

		 if ( hashvalue < hashsize && table[hashvalue] ) {
			curptr = table[hashvalue];
			scanValid = XXX_TRUE;
			return XXX_TRUE;
		 }
		 else  {
			scanValid = XXX_FALSE;
			return XXX_FALSE;
		 }
	   }

	   void InitScan() { scanValid = XXX_TRUE; }
	   XxxBool ScanIsValid() { return ( scanValid); }

	   XxxRecord *GetNext() {
		   if ( curptr == NULL ) {
			 scanValid = XXX_FALSE;
			 return NULL;
		   } else {
			  XxxRecord *Rec = curptr->Rec;
			  curptr = curptr->Next;
			  scanValid = XXX_TRUE;
			  return Rec;
		   }
	   } 
	   
	   void CloseScan() { curptr = NULL; scanValid = XXX_FALSE; }
	   int AttrIndex() { return attrIndex; } 
  }; // HashTable

  /* 
   *	shared buffer implementation to store the XxxRecords object 
   *	@author Wee Hyong, Tok
   *	@@version 06 Jun 2000
   */
   class XxxRecordBuffer {

   public:
		XxxRecord		*curRec;
		XxxRecord		*Table[XXX_MAX_SHAREDBUFFER_RECORDS];
		XxxBool         State[XXX_MAX_SHAREDBUFFER_RECORDS];
		XxxBaseMutex    *Mutex;
		XxxCondVar      *CondFull;
		XxxCondVar      *CondEmpty;
		XxxBool			isValid;

		int itemCount;
   		int bIndex;
		int timeoutCount;	// timeout and return no recrods after specific no of tries
		int First, Last;
   protected:
		XxxBool			moreRecs;
   public:
	   XxxRecordBuffer(int);
	   virtual ~XxxRecordBuffer();
	   
	   virtual XxxBool IsFull();
	   virtual XxxBool IsEmpty();
	   virtual XxxErrCode AddItem(XxxRecord *Item);
	   virtual XxxErrCode Add(XxxRecord *Item);
	   virtual XxxErrCode Pop();
	   virtual XxxErrCode Get();
	   virtual XxxBool IsValid();
	   virtual void signal();
   }; // XxxRecordBuffer

	
   class PopulateThread : public XxxBaseThread {
   public:
	   class PopulateShutDown {
	   private:
			XxxBool Shutdown;
	   public:
		   PopulateShutDown() { Shutdown = XXX_FALSE; }
		   ~PopulateShutDown() {}

		   void set() { Shutdown = XXX_TRUE;}
		   XxxBool isSet() { return Shutdown;}
	   };

   public:
		XxxRecordBuffer *Buffer;
		PopulateShutDown *Shutdown;
		XxxRelation* *populateRelationArray;
		XxxSetCursor *childCursorArray;
		XxxBooleanExpressionPlan *PredPlan;

		int populateIndex;
		int ArrayLen;

		XxxBaseMutex *populateMutex;		
		const RelExecParameters* parameters;
		XxxRecord *CurRec;
		XxxBool populateValid;

   public:
	   PopulateThread(XxxRelation **RelationArray,
			 XxxSetCursor *cursorArray, 
			 int Index, XxxRecordBuffer *buffer,
			 XxxBooleanExpressionPlan *plan, const RelExecParameters* params, int length) {
						
			// parameters needed to initcursor
			parameters				= params;
			PredPlan				= plan;
			ArrayLen				= length;

			childCursorArray		= cursorArray;
			Buffer					= buffer; 
			populateRelationArray	= RelationArray;
			populateIndex			= Index;
			
			this->Shutdown   = new PopulateShutDown();
			W_COERCE(fork()); 
	   }

	   void ShutDown() { Shutdown->set(); }
	   virtual void run();
   };

	
  class Store {
  private:
    /// The linked-list node structure
    struct Member {
      XxxRecord *Rec;
      Member *Next;
    };
    /// The head of the linked list
    Member *members;
    /// The cursor in scans
    Member *currentMember;
    /// The attribute of the records in the equality join condition
    
    /// Whether or not the current scan is valid
    XxxBool scanValid;
  public:
    /**
     * Constructor.
     */
    Store() : members(NULL), currentMember(NULL), scanValid(XXX_FALSE) { }
    /**
     * Destructor.
     */
    ~Store() {
      FlushAllRecords(XXX_FALSE);
    }
    /**
     * Returns XXX_TRUE if this group has no records in it.
     */
    XxxBool IsEmpty() { return members == NULL; }
    /**
     * Adds a record to this group.
     */
    void AddRecord(XxxRecord *Rec) {
      ASSERT(Rec);
      Member *NewMember = new Member;
      NewMember->Rec = Rec;
      NewMember->Next = members;
      members = NewMember;
    }
    /**
     * Removes all records from this group.
     * @param DestroyRecs If XXX_TRUE, the records within will be explicitly
     *                    destroyed; otherwise, only the linked-list structure
     *                    will be destroyed.
     */
    void FlushAllRecords(XxxBool DestroyRecs) {
      Member *cur = members, *tmp;
	  while (cur) {
		if (DestroyRecs)
		  delete cur->Rec;
		
		tmp = cur->Next;
		delete cur;
		cur = tmp;
	  }

	  scanValid = XXX_FALSE;
	  members = NULL;
	}
    /**
     * Initializes a scan over the group.
     */
    void InitScan() {
      currentMember = members; 
      scanValid = XXX_TRUE;
    }
    /**
     * Returns XXX_TRUE if the current scan is valid.
     */
    XxxBool ScanIsValid() { return (scanValid); }
    /**
     * Gets the next record from the scan.  Returns NULL if the scan is
     * now invalid.
     */
    XxxRecord *GetNext() {
      if (currentMember == NULL) {
		scanValid = XXX_FALSE;
		return NULL;
      } 
	  else {
		XxxRecord *Rec = currentMember->Rec;
		currentMember = currentMember->Next;
		scanValid = XXX_TRUE;
		return Rec;
      }
    }
    /**
     * Closes the current scan.
     */
    void CloseScan() { currentMember = NULL; scanValid = XXX_FALSE; }
  }; // store

  class HashJoinInfo {
	public:
	XxxBool Valid;						// is the scan valid
	int LeftRelIndex, RightRelIndex;	// indices of the left and right relations in the RelationArray
	int validIndex, probeIndex;
	HashTable LeftHashT, RightHashT;	// Left hash table, right hash table

	XxxRecord *sbrec[2];
	XxxRecordBuffer *sb[2];				// Shared buffer- Used to hold data populated from the scans
	PopulateThread *PopulThread[2];	// 2 separate populate threads to run to retrieve data
	XxxBaseMutex   *populateMutex;
	XxxBool	firstCall;
	Store OutputStore;
	

	/**
	 * Constructor
	 */
	 HashJoinInfo(int RightRelIndex, int LeftRelIndex,
		          int RightAttrIndex, int LeftAttrIndex) :
	  Valid(XXX_TRUE), 
	   LeftRelIndex(LeftRelIndex), 
	   RightRelIndex(RightRelIndex),
	   validIndex(0), probeIndex(0),
	   LeftHashT(XXX_HASHTABLE_MAX, LeftAttrIndex),
	   RightHashT(XXX_HASHTABLE_MAX, RightAttrIndex),
	   OutputStore() 
	   { 
	     for (int i=0; i < 2; i++ ) {
	       sb[i] = new XxxRecordBuffer(i);
	       sbrec[i] = NULL;
	     }
	     populateMutex = new XxxBaseMutex();
	     firstCall = XXX_TRUE;
	   }
  }; // HashJoinInfo

  /**
   * Create the join object for the Hash Join run
   */
  virtual XxxErrCode createJoinInfo(XxxDerivedRelation::CursorInfo* Handle,
						HashJoinInfo *&JoinInfo) const;
  
 /**
   * Gets the next record from the relation in the RelationArray 
   * indexed by the parameter.  Returns Valid == XXX_TRUE if the next
   * record was returned.
   */
  
/* 
commented on 7 Jun 2000
used to test out the threaded version of SHJ
*/
virtual XxxErrCode getNextItemSb(HashJoinInfo* JoinInfo, int Index, XxxBool& Valid) const;
virtual XxxErrCode getNextItem(XxxRelation* *RelationArray,
				 XxxDerivedRelation::CursorInfo* Handle,
				 int Index, XxxBool& Valid) const;

  /**
   * Parses the join plan, finding the location of the left and right
   * relations and finding the indices of the critical attributes in
   * each's records.
   */
  virtual 
    XxxErrCode parseJoinPlan(XxxDerivedRelation::CursorInfo* Handle,
			     int &LeftAttrIndex, int &LeftRelIndex,
			     int &RightAttrIndex, int &RightRelIndex) const;
  /**
   * Compares a left and right record and returns XXX_TRUE if they are
   * equal with respect to their critical attributes.
   */
  virtual XxxBool isEqual(XxxRecord *LeftRec, int LeftAttrIndex,
			  XxxRecordSchema *LeftSchema,
			  XxxRecord *RightRec, int RightAttrIndex,
			  XxxRecordSchema *RightSchema) const;
  
public:
  /**
   * Returns a string representation of this plan operator.
   */
  virtual const char *PrintString()  const { return "Hash Join SPJ" ;}
  /**
   * Gets the next record from this plan operator.
   */
  virtual XxxErrCode GetNextRecord(XxxSetCursor& SCsr,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
				   XxxRecord*& ReturnRec) const;
  /**
   * Initializes this plan operator.
   */
  virtual XxxErrCode InitHandle(XxxDerivedRelation::CursorInfo* Handle,
				XxxRelation* *RelationArray) const;
  /**
   * Closes the scan this plan operator offers.
   */
  virtual XxxErrCode CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				 XxxRelation* *RelationArray) const;

  virtual XxxErrCode checkFinal(XxxDerivedRelation::CursorInfo* Handle,
				XxxCoreADTClass::BoolValue& Satisfied) const;
				 
  /**
   * Constructor.
   */
  RelHashJoinPlanOp(int NumChildren, 
	      RelPlanOp* Children[],
	      RelCatalogInfo* CatInfo,
	      XxxValExprPlanList* ExprPlanList,
	      XxxBooleanExpressionPlan** SelPredPlans,
	      XxxBooleanExpressionPlan** JoinPredPlans,
	      XxxBooleanExpressionPlan** ResidPredPlans,
	      XxxBooleanExpressionPlan* InitPlan);
  
 /**
   * Destructor.
   */
  virtual ~RelHashJoinPlanOp() {}
 
  virtual void PrintNodeInfo(ostream &OStr) const;
};
/**
 * End of Symmetric Hash Join Operator
 */

/**
 * Sorts a stream of tuples.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version July 10, 1997
 */
class RelSortPlanOp : public RelDerivedPlanOp {
protected:
  /// Information regarding the sort attributes
  XxxSortInfo sortInfo;
  /// Should duplicates be eliminated?
  XxxBool distinct;
  /// Should a temp file be used
  XxxBool useTempFile;
  /// size of runs for 1st pass of external sort
  int maxRunSize;
  /// Holds handle information
  struct SortHandleInfo {
    XxxShorePhysicalId *TempFile;
    scan_file_i *ScanIter;
  };
public:
  virtual void PrintThis(ostream& OStr)  const;
  virtual const char *PrintString()  const { return "SORT" ;}
  virtual XxxErrCode InitHandle(XxxDerivedRelation::CursorInfo* Handle,
				XxxRelation* *RelationArray) const;
  virtual XxxErrCode CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				 XxxRelation* *RelationArray) const;
  virtual XxxErrCode GetNextRecord(XxxSetCursor& SCsr,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
				   XxxRecord*& ReturnRec) const;
  RelSortPlanOp(RelPlanOp* Child,
		const XxxSortInfo &SInfo,
		XxxBool Distinct,
		XxxBool UseTempFile,
		int MaxRunSize,
		RelCatalogInfo* CatInfo);
  virtual ~RelSortPlanOp() ;
  virtual void NodePrint (ostream &OStr, int curBlockId, int Level, 
			  int numChildren, int *childrenIds) const;
 };

/**
 * Relational plan operator to eliminate duplicate tuples.  This
 * implementation relies on the fact that the incoming stream of tuples
 * is sorted on all attributes.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version July 7, 1997
 * @see SortPlanOp
 */
class RelDistinctPlanOp : public RelDerivedPlanOp {
 public:
  virtual void PrintThis(ostream& OStr) const;
  virtual const char *PrintString() const { return "DISTINCT" ;}
  virtual XxxErrCode InitHandle(XxxDerivedRelation::CursorInfo* Handle,
				XxxRelation* *RelationArray) const;
  virtual XxxErrCode CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				 XxxRelation* *RelationArray) const;
  virtual XxxErrCode GetNextRecord(XxxSetCursor& SCsr,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
				   XxxRecord*& ReturnRec) const;
  RelDistinctPlanOp(RelPlanOp* Child,
		    RelCatalogInfo* CatInfo);
  virtual ~RelDistinctPlanOp() ;
  virtual void NodePrint (ostream &OStr, int curBlockId, int Level, 
			  int numChildren, int *childrenIds) const;
 };

/**
 * Relational plan operator to apply a simple select filter.
 */
class RelFilterPlanOp : public RelDerivedPlanOp {
 protected:
  //// Selection to apply
  XxxBooleanExpressionPlan* SelPredPlan;

 public:
  virtual void PrintThis(ostream& OStr) const;
  virtual const char *PrintString() const { return "FILTER" ;}
  virtual XxxErrCode InitHandle(XxxDerivedRelation::CursorInfo* Handle,
				XxxRelation* *RelationArray) const;
  virtual XxxErrCode CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				 XxxRelation* *RelationArray) const;
  virtual XxxErrCode GetNextRecord(XxxSetCursor& SCsr,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
				   XxxRecord*& ReturnRec) const;
  RelFilterPlanOp(RelPlanOp* Child,
		  RelCatalogInfo* CatInfo,
		  XxxBooleanExpressionPlan* SelPredPlan);
  virtual ~RelFilterPlanOp() ;
  virtual void NodePrint (ostream &OStr, int curBlockId, int Level, 
			  int numChildren, int *childrenIds) const;
 };

//begin Ihab


class RelStopAfterPlanOp : public RelDerivedPlanOp {
 protected:
  int K;

 public:
  virtual void PrintThis(ostream& OStr) const;
  virtual const char *PrintString() const { return "STOP AFTER" ;}
  virtual XxxErrCode InitHandle(XxxDerivedRelation::CursorInfo* Handle,
				XxxRelation* *RelationArray) const;
  virtual XxxErrCode CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				 XxxRelation* *RelationArray) const;
  virtual XxxErrCode GetNextRecord(XxxSetCursor& SCsr,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
				   XxxRecord*& ReturnRec) const;
  RelStopAfterPlanOp(RelPlanOp* Child, int k, RelCatalogInfo* CatInfo);
  virtual ~RelStopAfterPlanOp() ;
  virtual void NodePrint (ostream &OStr, int curBlockId, int Level, 
			  int numChildren, int *childrenIds) const;
 };

//------------------------


class RelKRJPlanOp : public RelSPJPlanOp {
  
 protected:
  
  class Tuple{
    public:  
     Tuple( XxxRecord * LeftRec, 
	    XxxRecord* RightRec, 
	    double Grade, 
	    double WG,  
	    double BG, 
	    int Source): 
	LeftRec(LeftRec), RightRec(RightRec), Grade(Grade),WG(WG), BG(BG),Source(Source), 
	Next(NULL), Prev(NULL), Type(0){}
     
     ~Tuple(){}
     
     XxxRecord * LeftRec;
     XxxRecord * RightRec;
     double Grade;
     double WG; // Worst Grade
     double BG; // Best  Grade
     int Source; // final 0 left 1 right 2
     Tuple * Next;
     Tuple * Prev;
     int Type;
  };
  
  
  class Queue{
    private:
     Tuple * Front;
     Tuple * Rear; 
    public:
     int LeftAttrIndex, RightAttrIndex;
     int LeftGradeRel, RightGradeRel;
     XxxRecordSchema *LeftSchema, *RightSchema;
     int Size;
     double MaxB;
     Queue();
     ~Queue();
     void AddTuple(Tuple* tuple, int k);
     Tuple* GetNext ();
     XxxBool isEqual(XxxRecord *Rec, int AttrIndex,
		     XxxRecord *QRec, int QAttrIndex, 
		     XxxRecordSchema *Schema, XxxRecordSchema* QSchema) const;
     Tuple* GetTuple(XxxRecord* Rec, 
		     XxxRecordSchema *Schema, 
		     int RecSource,
		     int AttrIndex, 
		     XxxDerivedRelation::CursorInfo* Handle,
		     XxxBooleanExpressionPlan** ResidPredPlans,
		     int Remove =0);
     Tuple* ElementAt(int k);
     XxxBool IsEmpty();
     void updateMaxB (int k);
     void updateRanks(double,double);
  };
  
  
  
  public: 
  XxxValExprPlanList* OrderPlanList;
  int Top;
  
  class KRJJoinInfo { 
    public:
     XxxBool Valid;
     XxxBool eof;
     int Top;
     int LeftRelIndex, RightRelIndex;
     int LeftAttrIndex, RightAttrIndex;
     int LeftGradeRel, RightGradeRel;
     int LeftWGradeIndex, RightWGradeIndex;
     int LeftBGradeIndex, RightBGradeIndex;
     XxxRecord *DummyLeftRec, *DummyRightRec;
     double LeftTh, RightTh; // thresholds
     int depth;//depth in children
     int K;   // current K
     
     Queue *Buffer; // the operator Buffer
     
     KRJJoinInfo() { 
	Top = 0; 
	Valid = XXX_TRUE;
	eof = XXX_FALSE;
	depth = 0;
	K =-1;
	Buffer = new Queue();
	LeftTh = RightTh = 0.0;
	
	LeftRelIndex  = RightRelIndex=0;
	LeftWGradeIndex= RightWGradeIndex=0;
	LeftBGradeIndex= RightBGradeIndex=0;
	LeftAttrIndex = RightAttrIndex=0;
	LeftGradeRel =  RightGradeRel=0;
     }
     KRJJoinInfo(int RightRelIndex, int LeftRelIndex,
		 int RightAttrIndex, int LeftAttrIndex,
		 int LeftWGradeIndex, int RightWGradeIndex,
		 int LeftBGradeIndex, int RightBGradeIndex,
		 int LeftGradeRel, int RightGradeRel); 
     
     virtual ~KRJJoinInfo() ;
  };
  virtual XxxErrCode createJoinInfo(XxxDerivedRelation::CursorInfo* Handle,
				    KRJJoinInfo *&JoinInfo) const;
  
  // create a record once all the info is in the handle
  virtual XxxErrCode reclaimCurrent(XxxDerivedRelation::CursorInfo* Handle,
				    int NLPos) const;
  virtual XxxErrCode getNextItem(XxxRelation* *RelationArray,
				 XxxDerivedRelation::CursorInfo* Handle,
				 int NLPos, XxxBool& Valid) const;
  /**
   * Parses the join plan, finding the location of the left and right
   * relations and finding the indices of the critical attributes in
   * each's records.
   */
  virtual  XxxErrCode parseJoinPlan(XxxDerivedRelation::CursorInfo* Handle,
				    int &LeftAttrIndex, int &LeftRelIndex,
				    int &RightAttrIndex, int &RightRelIndex,
				    int &LeftWGradeIndex, int &RightWGradeIndex,
				    int &LeftBGradeIndex, int &RightBGradeIndex,
				    int &LeftGradeRel,  int &RightGradeRel ) const;
  
  virtual XxxErrCode openChildScan(XxxDerivedRelation::CursorInfo* Handle,
				   XxxRelation* *RelationArray,
				   int NLPos) const ;
  virtual XxxErrCode closeChildScan(XxxDerivedRelation::CursorInfo* Handle,
				    XxxRelation* *RelationArray,
				    int NLPos) const ;
  virtual XxxErrCode checkFinal(XxxDerivedRelation::CursorInfo* Handle,
				XxxCoreADTClass::BoolValue& Satisfied) const 
				 { Satisfied.Value = XXX_TRUE; return XXX_OK;}

 public:
  
  virtual XxxErrCode createRecord(XxxDerivedRelation::CursorInfo* Handle,
				  XxxRelation* *RelationArray,
				  const XxxRecordSchema* Schema,
				  XxxRecord*& ReturnRec,
				  double WG=0,
				  double BG =0) const;
  virtual XxxErrCode CreateDummyRec(  const XxxRecordSchema* Schema,  
				      XxxRecord*& ReturnRec) const;
  virtual XxxErrCode GetNextRecord(XxxSetCursor& SCsr,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
				   XxxRecord*& ReturnRec) const;
  virtual XxxErrCode InitHandle(XxxDerivedRelation::CursorInfo* Handle,
				XxxRelation* *RelationArray) const;
  virtual XxxErrCode CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				 XxxRelation* *RelationArray) const;
  //virtual double GetGrade  (XxxRecord *Rec, int AttrIndex, XxxRecordSchema *Schema) const;

  RelKRJPlanOp(int NumChildren, 
	      RelPlanOp* Children[],
	      RelCatalogInfo* CatInfo,
	      XxxValExprPlanList* ExprPlanList,
	      XxxValExprPlanList* OrderPlanList,
	      XxxBooleanExpressionPlan** SelPredPlans,
	      XxxBooleanExpressionPlan** JoinPredPlans,
	      XxxBooleanExpressionPlan** ResidPredPlans,
	      XxxBooleanExpressionPlan* InitPlan,
	      int top =0);
  
  virtual ~RelKRJPlanOp() {  }
  
  virtual const char *PrintString()  const { return "KRJOP : KRJ" ;}
  virtual void PrintNodeInfo(ostream &OStr) const;
  virtual XxxBool isEqual( KRJJoinInfo* JoinInfo, 
			   XxxDerivedRelation::CursorInfo* Handle,
			   XxxBooleanExpressionPlan** ResidPredPlans) const;
};



//end Ihab

#endif /* REL_PLANOP_H */


