#include <iostream.h>
    
#include "relerror.h"
#include "relquery.h"
#include "relplanop.h"
#include "relenv.h"
#include "costspjopt.h"

#include "relopt.h"
    
    
//
// This method analyzes the join graph to see if there is a linear chain
// of joins. If not, cross-products will have to be considered in left
// deep trees.
//
void RelSPJCostBasedOpt::
analyzeJoinGraph(XxxBool AvoidExpensivePreds, 
  		 double MaxCheapPredCost) // definition of "expensive"
{
  linearJoinGraph = XXX_FALSE; // by default: false
    
  int RelCount = BlockPlan->ChildArrayLen;
  XxxBitMap CurrentRels(RelCount);
  XxxBitMap TmpRels(RelCount);
    
  // try each rel as a possible starter for a linear chain
  for (unsigned char i = 0; i < RelCount; i++) {
    CurrentRels.ClearAll();
    CurrentRels.SetBit(i);
      
    unsigned char j;
    for (j = 0; j < RelCount; j++) {
      if (CurrentRels.GetBit(j)) continue;
      XxxBool Found = XXX_FALSE;
      // go thru all the preds and find a pred such that its dependencies 
      //    a) overlap with CurrentRels
      //    b) are totally contained in CurrentRels except for rel j
      for (int k=0; k < NumPreds; k++) {
    	CurrentRels.And(PredDependencyInfo[k], TmpRels);
    	if (!TmpRels.NumBitsSet()) continue;
    
    	CurrentRels.Xor(PredDependencyInfo[k], TmpRels);
    	TmpRels.And(PredDependencyInfo[k], TmpRels);
    	if ((TmpRels.NumBitsSet() == 1) && TmpRels.GetBit(j)) {
  	  // check here is this is an expensive predicate
 	  if (AvoidExpensivePreds && 
 	      PredArray[k]->GetPlanUsage().FullCost() > MaxCheapPredCost) {
 	    continue; // cannot use as a join predicate
 	  }
    	  Found = XXX_TRUE;
    	  break;
    	}
      } // for each pred
      // if an appropriate pred wasnt found, no point continuing with this
      if (!Found) break;
    } // for each relation in this linear chain
    
    if (j == RelCount) {
      // this means that the previous loop made it successfully
      // through all the rels (finding a linear chain)
      linearJoinGraph = XXX_TRUE;
      break;
    }
  } // for loop over all possible starting rels for linear chain
    
  if (XXX_THISRELENV()->DebugOpt) {
    cerr << "Linear Join Graph: " << (linearJoinGraph?"YES":"NO") << endl;
  }
}
    
XxxErrCode RelSPJCostBasedOpt::
GeneratePlanOp(RelOptEngine::PlanObject* Plan,
    	       RelPlanOp*& PlanOp) const
{
  PhysicalPlan* SPJPlan = (PhysicalPlan*)(Plan);
    
  // This remap is used to alter the various expressions.
  // Initially, it is the identity mapping
  XxxExprPlan::VarRemapping Remap(BlockPlan->ChildArrayLen);
  Remap.NewArity = 2;
  for (int i=0; i < BlockPlan->ChildArrayLen; i++) {
    Remap.SrcMappingArray[i].NewSrc = -1; // default bad value
    int ChildArity = 
      BlockPlan->GetChild(i)->CatalogInfo->RecordSchema()->Arity();
    Remap.SrcMappingArray[i].NewAttrMapping = new int[ChildArity];
    // initially, leave the attribute mapping alone
    for (int j = 0; j < ChildArity; j++)
      Remap.SrcMappingArray[i].NewAttrMapping[j] = j;
  }
    
  if (SPJPlan->GeneratePlanOp(this, XXX_TRUE, 
			      Remap, 0, 0, PlanOp) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error generating planop");
    return XXX_NOT_OK;
  }
    
  // Take care of a possible DISTINCT clause
  if (Distinct) {
    XxxSortInfo SInfo(BlockPlan->CatalogInfo->RecordSchema()->Arity());
    int i;
    for (i = 0; i < BlockPlan->CatalogInfo->RecordSchema()->Arity(); i++) {
      SInfo[i].AttrIndex = i;
      SInfo[i].Asc = XXX_TRUE;
    }
    // This takes some explaining: if the input is already sorted in
    // the necessary order, of course, just add a Distinct planop.
    // However, if the input is not sorted, add a sort operator.
    // The sort operator takes a "distinct" flag which allows 
    // duplicates to be eliminated early in the sort. So in this
    // case, combine the two planops.
    /**** For some reason, there is a bug in the sort that removes
    *** non-equal values!!!
    **** TODO: find out why and fix it!! likewise in naivespj.C
    if (!PlanOp->CatalogInfo->ImplInfo()->SortInfo.Matches(SInfo)) {
    // Create the sort + distinct planop 
    PlanOp = new RelSortPlanOp(PlanOp, SInfo, XXX_TRUE, XXX_TRUE,
                               XXX_THISENV().DBEnv.BufPoolSize/2,
                               BlockPlan->CatalogInfo);
    }
    else {
    // Create the distinct planop 
    PlanOp = new RelDistinctPlanOp(PlanOp, BlockPlan->CatalogInfo);
    }
    ****/
    if (!PlanOp->CatalogInfo->ImplInfo()->SortInfo.Matches(SInfo)) {
      PlanOp = new RelSortPlanOp(PlanOp, SInfo, XXX_FALSE, XXX_TRUE,
				 XXX_THISENV().DBEnv.BufPoolSize/2,
				 BlockPlan->CatalogInfo);
    }
    PlanOp = new RelDistinctPlanOp(PlanOp, BlockPlan->CatalogInfo);
    
    // make sure to specify the SortInfo in the result
    PlanOp->CatalogInfo->ImplInfo()->SortInfo.Init(1);
    PlanOp->CatalogInfo->ImplInfo()->SortInfo.SortInfoArray[0].Init(SInfo);
  }
    
  return XXX_OK;
}
    
XxxErrCode RelSPJCostBasedOpt::
generateBestRelPlan(const XxxDBEnv& Env,
    		    const RelOptEngine::LogProperties& RelLProps,
    		    unsigned char RelId,
    		    PhysicalPlan*& RelPPlan)
{
  // Step 1: create a PlanTable
  MinPlanTable PlanTab;
    
     // Step 2: if none exists, call generateRelPlans
  if (generateRelPlans(Env, RelLProps, RelId, PlanTab) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error generating phys rel plan");
    return XXX_NOT_OK;
  }
    
  // Step 3: extract PPlan from PlanTable
  // get the plan out of the PlanTab : there should be just one
  if (PlanTab.GetMinPlan(RelPPlan) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error getting plan from plan table");
    return XXX_NOT_OK;   
  }
  PlanTab.Reset();
     
  return XXX_OK;
}
    
XxxErrCode RelSPJCostBasedOpt::
generateRelPlans(const XxxDBEnv& Env,
    		 const RelOptEngine::LogProperties& RelLProps,
    		 unsigned char RelId,
    		 PlanTable& PlanTab)
{
  int NumPlansAdded, NumPlansPruned;
    
  const RelCatalogInfo* CatInfo = ChildCat(RelId);
    
  if (!CatInfo) {
    REL_NOTE_ERROR(REL_BADOPT, "Invalid cat info for rel");
    return XXX_NOT_OK;
  }
  RelImplInfo* ImplInfo = CatInfo->ImplInfo();
  if (!ImplInfo) {
    REL_NOTE_ERROR(REL_BADOPT, "Invalid impl info for rel");
    return XXX_NOT_OK;
  }
  RelPhysPlan* Plan;
  //Begin: Moustafa
  // If relation has stream attribute, then there is only a stream scan
  // to choose from (Mar. 2003)
  if(BlockPlan->HasStreamInput){
    // Add one and only SScanPlan
    Plan = new RelSScanPlan(BlockPlan->HighestPlanId++,
			    BlockPlan->BlockId,
			    RelLProps, RelId,
			    RelLProps.Preds,
			    Env, CatInfo->StatsInfo());
    if (XXX_THISRELENV()->DebugOpt) {
      cerr << "Generated Plan: ";  Plan->Print(cerr); cerr << endl;
    }
    // this may delete (prune) the plan 
    if (PlanTab.AddPlan(Env, Plan, 
			NumPlansAdded, NumPlansPruned) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADOPT, "Error adding plan");
      return XXX_NOT_OK;
    }
    return XXX_OK;    
  }
  //End: Moustafa
  // there is always a FILESCAN plan
  Plan = new RelFScanPlan(BlockPlan->HighestPlanId++,
			  BlockPlan->BlockId,
			  RelLProps, RelId,
			  RelLProps.Preds,
			  Env, CatInfo->StatsInfo());
  if (XXX_THISRELENV()->DebugOpt) {
    cerr << "Generated Plan: ";  Plan->Print(cerr); cerr << endl;
  }
  // this may delete (prune) the plan 
  if (PlanTab.AddPlan(Env, Plan, 
		      NumPlansAdded, NumPlansPruned) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error adding plan");
    return XXX_NOT_OK;
  }
    
  // for each index available in ImplInfo, add an index scan
  XxxBitMap IndxPreds(NumPreds);
  XxxBitMap IndxDeps(BlockPlan->Arity());
  RelImplInfo::IndexList* CurrIndex = ImplInfo->Indexes->Next;
  while (CurrIndex) {
    RelImplInfo::IndexImplInfo* CurrIndexInfo = 
      CurrIndex->IndxInfo;
    // first, add an indexed access method for any predicates 
    // in RelLProps.Preds. This corresponds to simple selection preds.
    for (int j = 0; j < NumPreds; j++) {
      // Look at each predicate in RelLProps.Preds
      Plan = NULL;
      if (RelLProps.Preds.GetBit(j)) {
	XxxMatchInfo* MatchedInfo = NULL;
	XxxBool Success = XXX_FALSE;
	if (CurrIndexInfo->Matches(PredArray[j],
				   BlockPlan->Arity(), RelId, 
				   Success, MatchedInfo) != XXX_OK) {
	  REL_NOTE_ERROR(REL_BADOPT, "Error matching index");
	  return XXX_NOT_OK;	 
    	}
	if (Success) {
	  IndxPreds.ClearAll();
	  IndxPreds.SetBit(j);
	  IndxDeps.ClearAll();
	  Plan = new RelIScanPlan(BlockPlan->HighestPlanId++,
				  BlockPlan->BlockId,
				  RelLProps, RelId,
				  RelLProps.Preds, // this is NewPreds too
				  CurrIndexInfo->IndexName,
				  IndxPreds,
				  IndxDeps,
				  MatchedInfo,
				  Env, CatInfo->StatsInfo());
    	} // if this predicate matches the index 
      } // if the predicate is applicable
      if (Plan) {
	if (XXX_THISRELENV()->DebugOpt) {
	  cerr << "Generated Plan: ";  Plan->Print(cerr); cerr << endl;
	}
	// this may delete/prune the plan
	if (PlanTab.AddPlan(Env, Plan, 
			    NumPlansAdded, NumPlansPruned) != XXX_OK) {
	  REL_NOTE_ERROR(REL_BADOPT, "Error adding plan");
	  return XXX_NOT_OK;
	}
      }
    } // for each predicate
    
    // then add indexed access methods for each index scan needed
    // for ordered access based on interesting orders : TODO:L_OPT
    /****
      // then add indexed access paths for indexes used in joins : i.e.,
      // the predicate used does not appear in RelLProps.Preds, but
      // it is a join predicate, part of which is bound by relations
      // occurring earlier in the join order. This is a dependency which
      // must be specified : TODO:L_OPT
      ***/
    
    CurrIndex = CurrIndex->Next;
  } // for each index
    
  // do not reclaim plans : if any plan is
  // to be reclaimed, this will happen within AddPlan
  return XXX_OK;
}
    
XxxErrCode RelSPJCostBasedOpt::
generateBestJoinPlan(const XxxDBEnv& Env,
    		     const PhysicalPlan* OuterPPlan,
    		     const RelOptEngine::LogProperties& InnerLProps,
    		     const RelCatalogInfo* InnerCatInfo,
    		     const RelOptEngine::LogProperties& JoinLProps,
    		     const XxxBitMap& JoinPreds,
    		     PhysicalPlan*& JoinPPlan)
{
  // Step 1: create a PlanTable
  MinPlanTable PlanTab;
  MinPlanTable DummyPlanTab;
  // add by zhiyuan, PlanTab store Minimal Plan tab
  PlanTab.Joinplannow = 1;
  // Step 2: call generateJoinPlans
  if (generateJoinPlans(Env, OuterPPlan, InnerLProps, InnerCatInfo,
			DummyPlanTab,
			JoinLProps, JoinPreds, PlanTab) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error generating phys rel plan");
    return XXX_NOT_OK;
  }
    
  // Step 3: extract PPlan from PlanTable
  // get the plan out of the PlanTab : there should be just one
  if (PlanTab.GetMinPlan(JoinPPlan) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error getting plan from plan table");
    return XXX_NOT_OK;   
  }
  PlanTab.Reset();
    
  return XXX_OK;
}
    
XxxErrCode RelSPJCostBasedOpt::
generateJoinPlans(const XxxDBEnv& Env,
    		  const PhysicalPlan* OuterPPlan,
    		  const RelOptEngine::LogProperties& InnerLProps,
    		  const RelCatalogInfo* InnerCatInfo,
    		  PlanTable& InnerPlanTab,
    		  const RelOptEngine::LogProperties& JoinLProps,
    		  const XxxBitMap& NewPreds,
    		  PlanTable& PlanTab)
{   
  int NumPlansAdded, NumPlansPruned;
  
  RelImplInfo* InnerImplInfo = NULL;
  RelStatsInfo* InnerStatsInfo = NULL;
  if (InnerCatInfo) {
    InnerImplInfo = InnerCatInfo->ImplInfo();
    InnerStatsInfo = InnerCatInfo->StatsInfo();
  }
  
  // loops through available join methods
  RelOptEngine::JoinMethod JMeth = RelOptEngine::UNKNOWN_JOIN;
  for (JMeth = (RelOptEngine::JoinMethod)(((int)JMeth) + 1); 
       JMeth < RelOptEngine::INVALID_JOIN; 
       JMeth = (RelOptEngine::JoinMethod)(((int)JMeth) + 1)) {
    //Begin: Moustafa
    // Exclude any join plans other than HBEWJOIN and NLBEWJOIN or HFEWJOIN and NLFEWJOINif one
    // of the inputs is a stream
    if(BlockPlan->HasStreamInput){
      if((JMeth != RelOptEngine::HBEWINDOW)&&
	 (JMeth != RelOptEngine::NLBEWINDOW)&&
	 (JMeth != RelOptEngine::HFEWINDOW)&&
	 (JMeth != RelOptEngine::NLFEWINDOW)) continue;
    }
    else{
      if((JMeth == RelOptEngine::HBEWINDOW)||
	 (JMeth == RelOptEngine::NLBEWINDOW)||
	 (JMeth == RelOptEngine::HFEWINDOW)||
	 (JMeth == RelOptEngine::NLFEWINDOW)) continue;
    }

    // No sort_merge plan if this block has_stream_input
    if ((JMeth == RelOptEngine::SORTMERGE)&&
	((!XXX_THISRELENV()->SortMergeEnabled)||(BlockPlan->HasStreamInput))) continue;
    //    if ((JMeth == RelOptEngine::SORTMERGE) &&
    //	(!XXX_THISRELENV()->SortMergeEnabled)) continue;
    //End: Moustafa
    if ((JMeth == RelOptEngine::PNLOOPS) &&
	(XXX_THISRELENV()->NLJoinMethod != XxxRelEnv::PAGE_NL)) continue;
    
    JoinPhysPlan* JPlan;
    XxxBool Success = XXX_FALSE;
    
    
    REL_CALL_FN(createJoinPlan(Env, JoinLProps, OuterPPlan, 
			       InnerLProps, InnerStatsInfo, InnerPlanTab,
			       NewPreds, JMeth, Success, JPlan),
		REL_BADOPT, "Error creating join plan");
    if (Success) {       
      if (XXX_THISRELENV()->DebugOpt) {
	cerr << "Generated Plan: ";
	JPlan->Print(cerr);
	
      }
      
      int NumPlansAdded = 0;
      int NumPlansPruned = 0;
      REL_CALL_FN(PlanTab.AddPlan(Env, JPlan, 
				  NumPlansAdded, NumPlansPruned),
		  REL_BADOPT, "Error adding plan");
      if (XXX_THISRELENV()->DebugOpt) {
	cerr << " :: Added:" << NumPlansAdded << " Pruned: " << NumPlansPruned << endl;
      }
    } // if Success
  } // for each join method
     
  return XXX_OK;
}
    
//
// Create a join between OuterPPlan and one or more PPlans for the inner 
// table. Try to get these plans from InnerPlanTab. If no plans are
// available in InnerPlanTab, try to create them.
//
XxxErrCode RelSPJCostBasedOpt::
createJoinPlan(const XxxDBEnv& Env,
    	       const RelOptEngine::LogProperties& JoinLProps,
    	       const PhysicalPlan* OuterPPlan,
    	       const RelOptEngine::LogProperties& InnerLProps,
    	       const RelStatsInfo* InnerStatsInfo,
    	       PlanTable& InnerPlanTab,
    	       const XxxBitMap& NewPreds,
    	       RelOptEngine::JoinMethod JMeth,
    	       XxxBool& Success,
    	       JoinPhysPlan*& JPlan)
{
  Success = XXX_FALSE;
  if (InnerLProps.Rels.NumBitsSet() != 1) {
    REL_NOTE_ERROR(REL_BADOPT, "Inner plan should have exactly one rel");
    return XXX_NOT_OK;
  }
  RelOptEngine::LogProperties NewInnerLProps(InnerLProps);
  unsigned char SMPredIndex;
    
  // do some pre-processing based on the join method
  switch (JMeth) {
  case RelOptEngine::TNLOOPS: {
    // add the join preds to the InnerLProps
    NewInnerLProps.Preds.Or(NewPreds, NewInnerLProps.Preds);
    Success = XXX_TRUE;
    break;
  }
  case RelOptEngine::PNLOOPS:
    Success = XXX_TRUE;
    break;
  case RelOptEngine::SORTMERGE: {
    // need to check if there is a sort-merge predicate
    for (unsigned char i = 0; i < NewPreds.ArraySize; i++) {
      if (!NewPreds.GetBit(i)) continue;
      XxxBool SMJoinable = XXX_FALSE;
      XxxRelBooleanPlan* SMPlan; // dummy
      XxxBooleanExpressionPlan* ResidPlan; // dummy
      if (isSMJoinable(PredArray[i], SMPlan, ResidPlan,
    		       SMJoinable) != XXX_OK) {
    	REL_NOTE_ERROR(REL_BADOPT, "Error checking sort-merge");
    	return XXX_NOT_OK;
      }
      if (SMJoinable) {
    	SMPredIndex = i;
    	Success = XXX_TRUE;
    	break;
      }
    } // for each NewPred
    break;
  }
  case RelOptEngine::NDXLOOPS:
    // don't create a separate plan for this: it gets tried as part
    // of TNL join (the inner table access may be indexed)
  default:
    Success = XXX_FALSE;
  }
    
  if (Success) {
    // now generate the best plan for the inner
    PhysicalPlan* InnerPPlan = NULL;
    unsigned char InnerRelId;
    for (InnerRelId = 0; InnerRelId < InnerLProps.Rels.ArraySize; InnerRelId++) {
      if (InnerLProps.Rels.GetBit(InnerRelId)) break;
    }
    
    // TODO: step thru all the plans in the InnerPlanTab
    // 1) try to see if there are any plans in InnerPlanTab for this LProp
    // 2) if not, create the plans.
    // 3) then step thru the plans

    // if debug, dump the course of choosing best inner plan
    if (generateBestRelPlan(Env, NewInnerLProps,
			    InnerRelId, InnerPPlan) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADOPT, "Error generating phys rel plan");
      return XXX_NOT_OK;
    }

    // define the SortInfo if any based on the PhysProp for PPlans
	// need to check the PhysProperties of boths PPlans to 
	// see what should the SortInfo be
	int numOrderAttrs  = 0;
	int SInfoType = 0;
	int i;

	if ( OuterPPlan->PhysProps.SortOrder.Matches(InnerPPlan->PhysProps.SortOrder) || 
		OuterPPlan->PhysProps.SortOrder.interestingMatches(InnerPPlan->PhysProps.SortOrder) ) {
		// both SortOrder for OuterPPlan and InnerPPlan match
		// SortInfo is populated with the SortOrder found in either of the Plans
		// or
		// SortInfo for InnerPPlan is a prefix for the SortInfo for OuterPPlan
		// init the SInfo to be the number of ordered attribute in the prefix
			
		if ( JMeth == RelOptEngine::SORTMERGE ) {
			numOrderAttrs = 1;
			SInfoType = 2;
		}
		else
			numOrderAttrs = InnerPPlan->PhysProps.SortOrder.NumOrderedAttrs(); 
	}
	else if ( InnerPPlan->PhysProps.SortOrder.interestingMatches(OuterPPlan->PhysProps.SortOrder) ) {
		// SortInfo for OuterPPlan is a prefix for the SortInfo for InnerPPlan
		// init the SInfo to be the number of ordered attribute in the prefix
		  numOrderAttrs = OuterPPlan->PhysProps.SortOrder.NumOrderedAttrs();
		  SInfoType = 1;
	}
	else {
		// no exact match, no prefix match
		if ( JMeth == RelOptEngine::SORTMERGE ) {
			numOrderAttrs = 1;
			SInfoType = 2;
		}
		else {
			SInfoType = 1;
			numOrderAttrs = OuterPPlan->PhysProps.SortOrder.NumOrderedAttrs();
		}
	}

	XxxSortInfo SInfo(numOrderAttrs);
	
	
	switch ( SInfoType ) {
	case 0: 
		// populate based on SortInfo in Inner Plan
		for (i = 0; i < numOrderAttrs; i++) {
			SInfo[i].AttrIndex = InnerPPlan->PhysProps.SortOrder[i].AttrIndex;
			SInfo[i].Asc = InnerPPlan->PhysProps.SortOrder[i].Asc;
		}
		break;
	case 1:
	  // populate based on SortInfo in Inner Plan
	  for (i = 0; i < numOrderAttrs; i++) {
		  SInfo[i].AttrIndex = OuterPPlan->PhysProps.SortOrder[i].AttrIndex;
		  SInfo[i].Asc = OuterPPlan->PhysProps.SortOrder[i].Asc;
	  }
	  break;
	case 2: 
		if ( JMeth == RelOptEngine::SORTMERGE ) {
			// for sort-merge, the sorting is based on the first attribute referenced in
			// the WHERE clause. To obtain it, we reference the SubPlans array and obtain the SrcAttrIndex from there
			// this correspond to the attribute used in sorting the plans
		  SInfo[0].AttrIndex= ((XxxUnknownValue::UnknownValuePlan *)(this->PredArray[0]->SubPlans[0]))->SrcAttrIndex;
		  SInfo[0].Asc		= XXX_TRUE;
		}
		break;
	}  // switch

	RelOptEngine::PhysProperties JoinPProps(SInfo);

    switch (JMeth) {
    case RelOptEngine::TNLOOPS: {
      // Now generate the TNL plan
      // Note: there are no "join preds" here since
      // they are all pushed into the inner.
      XxxBitMap TNLJoinPreds(NewPreds);
      TNLJoinPreds.ClearAll();
      JPlan = new TNLJoinPlan(BlockPlan->HighestPlanId++,
			      BlockPlan->BlockId,
			      JoinLProps, JoinPProps,
			      OuterPPlan, InnerPPlan, TNLJoinPreds,
			      Env, InnerStatsInfo);
      break;
    }
    case RelOptEngine::PNLOOPS:
      /**  TODO: use this to get a better estimate of BlockCard
	int OuterTuplesPerPage = 
	DBEnv.PageSize/ (OuterLProps.TupleSize + DBEnv.RecordOverhead);
	int NumInvocs = OuterLProps.Card / (OuterTuplesPerPage * PagesPerBlock);
	**/
      JPlan = new PNLJoinPlan(BlockPlan->HighestPlanId++,
			      BlockPlan->BlockId,
			      JoinLProps, JoinPProps,
			      OuterPPlan, InnerPPlan, NewPreds,
			      Env, InnerStatsInfo,
			      RelPNLPlanOp::DEFAULT_PNL_BLOCKCARD);
      break;
    case RelOptEngine::SORTMERGE: {
      // we can ignore SMPlan and ResidPlan because we know
      // that SMPlan == PredArray[i] and ResidPlan == NULL
      XxxBitMap JPreds(NewPreds.ArraySize);
      XxxBitMap RPreds(NewPreds);
      JPreds.SetBit(SMPredIndex);
      RPreds.ClearBit(SMPredIndex);
      
	  // TODO:L_OPT check if sorting needed on inputs
      // Use the PhysProperties of the PPlans  
	  XxxBool toSortOuterPlan = XXX_TRUE;
	  XxxBool toSortInnerPlan = XXX_TRUE;

	  if ( OuterPPlan->PhysProps.SortOrder.NumOrderedAttrs() != 0
		  && InnerPPlan->PhysProps.SortOrder.NumOrderedAttrs() != 0 ) {
		// ordered attributes exists for both plans
		// check whether they match or one is the prefix of the other
		// if they do, then there is no need to performing sorting

		if ( OuterPPlan->PhysProps.SortOrder.interestingMatches(InnerPPlan->PhysProps.SortOrder) ||
			InnerPPlan->PhysProps.SortOrder.interestingMatches(OuterPPlan->PhysProps.SortOrder) ) {
			toSortOuterPlan = XXX_FALSE;
			toSortInnerPlan = XXX_FALSE;
		}
	  }
	  else {
		  int sourceIndex = ((XxxUnknownValue::UnknownValuePlan *)(this->PredArray[0]->SubPlans[0]))->SrcAttrIndex;
		// check whether we need to sort the OuterPPlan by looking at its PhysProperties
		  if ( OuterPPlan->PhysProps.SortOrder.NumOrderedAttrs() != 0 ) {
			if (OuterPPlan->PhysProps.SortOrder[0].AttrIndex == sourceIndex )
				toSortOuterPlan = XXX_FALSE;
		  }
		  if ( InnerPPlan->PhysProps.SortOrder.NumOrderedAttrs() != 0 ) {
			if (InnerPPlan->PhysProps.SortOrder[0].AttrIndex == sourceIndex )
				toSortInnerPlan = XXX_FALSE;
		  }

	  }

	  // added by wee hyong to include in the Physical properties of the plan
	  // @version 19 June 2000
	  // tok19!
	  JPlan = new SMJoinPlan(BlockPlan->HighestPlanId++,
			     BlockPlan->BlockId,
			     JoinLProps,JoinPProps, OuterPPlan, InnerPPlan, 
			     toSortOuterPlan, toSortInnerPlan,
			     NewPreds, JPreds, RPreds,
			     Env, InnerStatsInfo);
	  

      break;
    }
	
    case RelOptEngine::NDXLOOPS:
      // don't create a separate plan for this: it gets tried as part
      // of TNL join (the inner table access may be indexed)
    default:
      // shouldn't come here
      return XXX_NOT_OK;
    } // switch
    
  } // if Success
    
  return XXX_OK;
}
    
//
// Generate all plans that are formed by joining the two logical plans.
// Step thru all the physical plans of the outer. However, do not at
// this stage step thru the physical plans of the inner. Instead,
// just extract the InnerLogProps and extract the logical plan from
// InnerPlanTab in the nested call to generateJoinPlans().
// The reason is that for something like an index NL join, we need
// the join preds to be pushed into the InnerPPlan, so the actual
// LogProps of this plan will differ (have the join preds too) from
// the InnerLPlan.
XxxErrCode RelSPJCostBasedOpt::
generateJoinPlans(const XxxDBEnv& Env,
    		  const LogicalPlan* OuterLPlan,
    		  const LogicalPlan* InnerLPlan,
    		  const RelCatalogInfo* InnerCatInfo,
    		  PlanTable& InnerPlanTab,
    		  const RelOptEngine::LogProperties& JoinLProps,
    		  const XxxBitMap& NewPreds,
    		  PlanTable& PlanTab)
{
  // LOOP 1: loops through available plans for outer
  XxxSetCursor Csr0; PhysicalPlan* PPlan0;
  if (OuterLPlan->OpenPlanScan(Csr0) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "open scan: outer physplans");
    return XXX_NOT_OK;
  }
  while (1) {
    if (OuterLPlan->GetNextPlan(Csr0, PPlan0) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADOPT, "get next: outer physplans");
      return XXX_NOT_OK;
    }
    if (!Csr0.IsValid()) break;
    if (generateJoinPlans(Env, PPlan0, InnerLPlan->LogProps, 
			  InnerCatInfo, InnerPlanTab,
			  JoinLProps, NewPreds, PlanTab) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADOPT, "error generating join plans");
      return XXX_NOT_OK;
    }
  }
  if (OuterLPlan->ClosePlanScan(Csr0) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT,"Error in close scan: outer physplans");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}
    
