#include <iostream.h>
#include <math.h>

#include "relerror.h"
#include "relquery.h"
#include "relplanop.h"
#include "relenv.h"
#include "greedyspj.h"

//
// Methods of RelSPJGreedyOpt
//
/**
 * Overloads this method to determine the join order.
 * The greedy algorithm avoids cross-products and 
 * tries to apply the most selective join first
 */
XxxErrCode RelSPJGreedyOpt::
determineJoinStrategy(int NumChildren, int* JoinOrder,
		      RelOptEngine::JoinMethod* JoinMethods) const 
{
 int i;
 if (XXX_THISRELENV() && XXX_THISRELENV()->DebugOpt) {
   cerr << "Determining greedy join order" << endl;
 }

 int *TempJoinOrder = new int[NumChildren];
 RelOptEngine::JoinMethod *TempJoinMethods = 
   new RelOptEngine::JoinMethod[NumChildren-1];

 RelOptEngine::Cost Cost, TempCost;
 Cost.SetMax();

 // determine greedy join orders trying each child as the first join table
 for (i=0; i<NumChildren; i++) {
   if (determineGreedyJoinOrder(NumChildren, TempJoinOrder, TempJoinMethods,
				TempCost, i) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error determining greedy join order");
     delete[] TempJoinOrder;
     delete[] TempJoinMethods;
     return XXX_NOT_OK;
    }
   if (XXX_THISRELENV() && XXX_THISRELENV()->DebugOpt) {
     printPlan(cerr, NumChildren, TempJoinOrder, TempJoinMethods, TempCost);
   }

   // If the new cost is less than the current cost, choose the new plan.
   if (TempCost.FullCost() < Cost.FullCost()) {
     memcpy(JoinOrder, TempJoinOrder, NumChildren*sizeof(int));
     memcpy(JoinMethods, TempJoinMethods, 
	    (NumChildren-1)*sizeof(RelOptEngine::JoinMethod));
     Cost = TempCost;
     if (XXX_THISRELENV() && XXX_THISRELENV()->DebugOpt) {
       cerr << "Plan is chosen as lowest cost" << endl;
     }
    }
  } // for each child as the first table in the join

 if (XXX_THISRELENV()->DebugOpt) {
   cerr << "The Chosen Plan is :" << endl;
   printPlan(cerr, NumChildren, JoinOrder, JoinMethods, Cost);
 }

 // clean up
 delete[] TempJoinOrder;
 delete[] TempJoinMethods;

 if (Cost.FullCost() == DBL_MAX) {
   REL_NOTE_ERROR(REL_BADOPT, "Error in determining best greedy join order");
   return XXX_NOT_OK;
  }

 return XXX_OK;
}


/**
 * Chooses a greedy join order, starting at relation # First
 */
XxxErrCode RelSPJGreedyOpt::
determineGreedyJoinOrder(int NumChildren, int *JoinOrder, 
			 RelOptEngine::JoinMethod *JoinMethods,
			 RelOptEngine::Cost &Cost, const int First) const 
{
 RelOptEngine::Cost InputCost, CurrCost;
 RelOptEngine::LogProperties 
   InputLogProps(NumChildren, NumPreds), 
   CurrLogProps(NumChildren, NumPreds);
 // ignore phys props for now
 RelOptEngine::PhysProperties InputPhysProps, CurrPhysProps;
 RelOptEngine::JoinMethod JM;

 if (XXX_THISRELENV() && XXX_THISRELENV()->DebugOpt) {
   cerr << "Starting with relation:" << First << " .. ";
 }
 InputCost.SetMin();
 InputLogProps.Card = 1;
 InputLogProps.TupleSize = 0;
 InputLogProps.Preds.ClearAll();
 InputLogProps.Rels.ClearAll();

 // first take care of the predicates that depend on only the "First" child
 if (includeNewRel(First, 0 /* join order position */, 
		   InputCost /* initial cost */, 
		   InputLogProps,
		   InputPhysProps,
		   CurrCost, 
		   CurrLogProps,
		   CurrPhysProps,
		   JM) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error including initial Rel");
   return XXX_NOT_OK;
  }
 JoinOrder[0]=First;
 InputLogProps.Rels.CopyBits(CurrLogProps.Rels);
 InputLogProps.Preds.CopyBits(CurrLogProps.Preds);
 InputPhysProps.SortOrder = CurrPhysProps.SortOrder;

 // now use the greedy algorithm to determine the rest of the join order.
 // we are looping over the spots in the join order
 int i;
 for (i=1; i<NumChildren; i++) {
   RelOptEngine::Cost MinCost;
   RelOptEngine::LogProperties MinLogProps(NumChildren, NumPreds);
   // ignore phys props for now
   RelOptEngine::PhysProperties MinPhysProps;
   MinCost.SetMax();

   int ChosenRel = -1;
   RelOptEngine::JoinMethod ChosenJM;

   // try each of the possible relations and choose the best one
   for (int j=0; j<NumChildren; j++) {
     if (InputLogProps.Rels.GetBit(j)) continue;

     // determine properties if relation j was used in the next spot
     if (includeNewRel(j, i /* position in join order */,
		       InputCost, 
		       InputLogProps,
		       InputPhysProps,
		       CurrCost, 
		       CurrLogProps,
		       CurrPhysProps,
		       JM) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error including j'th Rel");
       return XXX_NOT_OK;
      }

     if (CurrCost.FullCost() < MinCost.FullCost()) {
       MinLogProps = CurrLogProps;
       MinPhysProps = CurrPhysProps;
       MinCost = CurrCost;
       ChosenRel = j;
       ChosenJM = JM;
      }
    } // loop over all the children for a particular spot
   if (MinCost.FullCost() == DBL_MAX) {
     REL_NOTE_ERROR(REL_BADOPT, "Error in finding next join child");
     return XXX_NOT_OK;
    }
   InputCost = MinCost;
   InputLogProps = MinLogProps;
   InputPhysProps = MinPhysProps;
   JoinOrder[i] = ChosenRel;
   JoinMethods[i-1] = ChosenJM;
   if (XXX_THISRELENV() && XXX_THISRELENV()->DebugOpt) {
     cerr << " " << ChosenRel << "(" << ChosenJM << "," << InputCost.FullCost() << ")";
   }
  } // loop over all the spots

 if (XXX_THISRELENV() && XXX_THISRELENV()->DebugOpt) {
   cerr << endl;
 }

 // make sure that all bits are set in RelsUsed and PredsUsed
 if (InputLogProps.Preds.NumBitsSet() != NumPreds) {
   REL_NOTE_ERROR(REL_BADOPT, "All the predicates were not used.");
   return XXX_NOT_OK;
  }      

 if (InputLogProps.Rels.NumBitsSet() != NumChildren) {
   REL_NOTE_ERROR(REL_BADOPT, "All the relations were not used.");
   return XXX_NOT_OK;
  }      

 // specify output cost
 Cost = InputCost;

 return XXX_OK;
}

/**
 * Estimate the cost/cardinality of joining with CurrentRel
 */
XxxErrCode RelSPJGreedyOpt::
includeNewRel(int CurrentRel,
	      int CurrentRelPos, // position in join order
	      const RelOptEngine::Cost& InputCost,
	      const RelOptEngine::LogProperties& InputLProps,
	      const RelOptEngine::PhysProperties& InputPProps,
	      RelOptEngine::Cost& OutputCost,
	      RelOptEngine::LogProperties& OutputLProps,
	      RelOptEngine::PhysProperties& OutputPProps,
	      RelOptEngine::JoinMethod& JM) const
{
 int k;
 OutputLProps = InputLProps;
 OutputPProps = InputPProps; // not being used currently

 OutputLProps.Rels.SetBit(CurrentRel);
     
 XxxBitMap OverlapRels(InputLProps.Rels.ArraySize);
 // Lets find all the predicates (join and selection) that can be
 // applied to the new relation.
 // Also, compute the selectivity of the filter predicates
 double Selectivity = 1.0;
 XxxBitMap NewPreds(InputLProps.Preds.ArraySize);
 for(k=0; k < InputLProps.Preds.ArraySize; k++) {
   if (OutputLProps.Preds.GetBit(k)) {
     // this predicate has already been used
     continue;
    }
   OverlapRels.ClearAll();
   OutputLProps.Rels.Or(PredDependencyInfo[k], OverlapRels);
   if (OverlapRels.Equals(OutputLProps.Rels)) {
     // all the necessary relations for this predicate are present
     OutputLProps.Preds.SetBit(k);
     NewPreds.SetBit(k);
     Selectivity = Selectivity * PredSelectivity[k];
    }
  }

 int TableCard = 0;
 int AvgTupleSize;

 // first get some stats on the current rel
 const RelCatalogInfo *CatInfo = ChildCat(CurrentRel);
 RelStatsInfo *StatsInfo = CatInfo->StatsInfo();
 if (StatsInfo) {
   TableCard = StatsInfo->Cardinality;
   AvgTupleSize = StatsInfo->AvgTupleSize;
  }
 else {
   TableCard = RelStatsInfo::DefaultTableCard;
   AvgTupleSize = RelStatsInfo::DefaultTupleSize;
  }

 // use NewPreds to check if there is an index match ...
 XxxBool IsIndexScan = XXX_FALSE;
 XxxBool IsEqualityPred = XXX_FALSE;
 XxxBool EqualityIndexScan = XXX_FALSE;
 XxxMatchInfo* MatchedInfo = NULL;
 RelImplInfo::IndexImplInfo* IndxInfo;
 int MatchingPredIndex = -1;
 if (CatInfo->ImplInfo() &&
     CatInfo->ImplInfo()->Indexes) {
   for (int j = 0; j < NewPreds.ArraySize; j++) {
     if (NewPreds.GetBit(j)) {
       if (CatInfo->ImplInfo()->Indexes->
	   Matches(PredArray[j], InputLProps.Rels.ArraySize, 
		   CurrentRelPos,
		   IsIndexScan, MatchedInfo, IndxInfo) != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADINDEX, "Error matching index");
	 return XXX_NOT_OK;
	} // if Matches() != XXX_OK
       // find out if it is an equality pred
       if ((PredArray[j]->BoolType == XXX_REL_BOOL_EXPR) &&
	   (((XxxRelBooleanPlan*)(PredArray[j]))->OperType 
	    == XXX_EQ_OP)) {
	 IsEqualityPred = XXX_TRUE;
	 if (IsIndexScan)
	   EqualityIndexScan = XXX_TRUE;
       } // if equality pred
       if (IsIndexScan) {
	 MatchingPredIndex = j;
	 break;
       }
      } // if NewPreds[j]
    } // for each pred
  } // if there is an Indexlist


/***
  ** Modified by Wee Hyong, tokweehy@comp.nus.edu.sg
  ** to add support for Hash Join
  **
  ** Now its time to assign outputcost.
  ** Here is the logic used:
  **   a) if it matches a clustered index, use it (INL join)
  **   b) if it matches an equality pred on an index, use it (INL)
  **   c) if there's no index, but an equality pred, check env variable
  **      If SortMergeEnabled is set, then use SM join
  **   d) If HashJoinEnabled is set, then use HashJoin
  **   e) else use PNL join
  **/
 // PNL join with filescan: this is the default
 int TuplesPerBlock = RelPNLPlanOp::DEFAULT_PNL_BLOCKCARD;
 CostPNLJoin(InputLProps, InputPProps, InputCost,
	     TableCard, AvgTupleSize,
	     OptSpecs()->DBEnv, OutputCost, TuplesPerBlock);
 JM = RelOptEngine::PNLOOPS;

 RelOptEngine::Cost TempCost;
 
 if (IsIndexScan) {
   
   // INL join
   CostINLJoin(InputLProps, InputPProps, InputCost,
	       TableCard, AvgTupleSize,
	       IndxInfo, MatchedInfo,
	       EqualityIndexScan,
	       PredSelectivity[MatchingPredIndex],
	       OptSpecs()->DBEnv, TempCost);
   if (TempCost.FullCost() < OutputCost.FullCost()) {
	 cout << "[Using NDXLoops] " << endl;
     OutputCost = TempCost;
     JM = RelOptEngine::NDXLOOPS;
   }
 }
 else if (IsEqualityPred) {
   // SM join
   cout << "[Equality pred on non-index item]" << endl;
   int RunLengthInput, RunLengthThis;
   XxxRelEnv* REnv =((XxxRelEnv*)(XXX_THISENGENV(RelationEngineId)));

   if ( REnv->SortMergeEnabled ) {
    CostSMJoin(InputLProps, InputPProps, InputCost, XXX_TRUE,
	      TableCard, AvgTupleSize, XXX_TRUE,
	      OptSpecs()->DBEnv, TempCost, RunLengthInput, RunLengthThis);
	cout << "SM Join" << endl;
    if (TempCost.FullCost() < OutputCost.FullCost()) {
     OutputCost = TempCost;
     JM = RelOptEngine::SORTMERGE;
	 
    }
   } else if ( REnv->HashJoinEnabled ) {
	 JM = RelOptEngine::HASHJOIN;
	 cout << "Hash Join" << endl;
   } else {
	 JM = RelOptEngine::PNLOOPS;
   }

   
 }

 if (MatchedInfo) delete MatchedInfo;

 // TODO:LOpt : add cost of any expensive predicates
 OutputLProps.Card = InputLProps.Card * (int)(TableCard * Selectivity);
 // TODO:LOpt : not pushing projections
 OutputLProps.TupleSize = InputLProps.TupleSize + AvgTupleSize;

 return XXX_OK;
}
