//
// Note: these classes are nested within RelSPJCostBasedOpt (see costspjopt.h)
// Do not include this header file directly.
// Classes defined are:
//     PhysicalPlan
//         -- RelPhysPlan
//              -- RelFScanPlan
//              -- RelIScanPlan
//Begin: Moustafa
//              -- RelSScanPlan
//End: Moustafa
//         -- JoinPhysPlan
//              -- TNLJoinPlan
//              -- PNLJoinPlan
//              -- INLJoinPlan
//              -- SMJoinPlan
//
#ifndef PHYSPLAN_H
#define PHYSPLAN_H

  #include "indexmatch.h"
  /**
    * @memo  RelSPJDynOptimizer::PhysicalPlan
    * This is either a plan for a relation or for a join
    */
  class PhysicalPlan : public RelOptEngine::PlanObject { 
   public:
    XxxBitMap NewPreds; // any preds to apply
    virtual void PrintTree(ofstream& OStr) const;
    virtual void TreePrint(ofstream &OStr, int &curPlanId, int level) const; // add by zhiyuan

    // this used to be a reference to shared logical properties
    // common to several physical plans, but it turned out to be
    // a problem to maintain. so instead, keep a copy in each physical
    // plan and swallow the overhead
    RelOptEngine::LogProperties LogProps;

	
    PhysicalPlan(int PlanIdent,
		 unsigned char BlockIdent,
		 const RelOptEngine::LogProperties& LProps,
		 const RelOptEngine::PhysProperties& PProps,
		 const XxxBitMap& NPreds)
      : RelOptEngine::PlanObject(PlanIdent, BlockIdent,PProps),
      NewPreds(NPreds), LogProps(LProps)
      {
      }
    virtual XxxErrCode GeneratePlanOp(const RelSPJOptimizer* Opt,
				      XxxBool TopLevel,
				      XxxExprPlan::VarRemapping& Remap,
				      int InitRemapSrcNum,
				      int FinalRemapSrcNum,
				      RelPlanOp*& PlanOp) const = 0;
    double Cost() const { return Usage.FullUsage.Cost();}
   };
  //// plan for access to stored table (or sub-block)
  class RelPhysPlan : public PhysicalPlan {
   public:
    unsigned char RelId; // index into array of rels for this block
    RelOptEngine::AccessMethod Method;
	
	virtual void TreePrint(ofstream &OStr, int &curPlanId, int level) const; // add by zhiyuan
	RelPhysPlan(int PlanIdent,
		unsigned char BlockIdent,
		const RelOptEngine::LogProperties& LProps,
		unsigned char RelationId,
		const XxxBitMap& NPreds,
		RelOptEngine::AccessMethod AMeth)
      : PhysicalPlan(PlanIdent, BlockIdent, LProps,NULL, NPreds)
	  //Bjarke, sep. 13 2001: the NULL value passed on is for a reference. References should not be NULL. Thus, we get a warning.
       {
	RelId = RelationId;
	Method = AMeth;
	}
	/* June 25 2001: Linking problem here - Back to previous definition
	RelPhysPlan(int PlanIdent,
		    unsigned char BlockIdent,
		    const RelOptEngine::LogProperties& LProps,
		    unsigned char RelationId,
		    const XxxBitMap& NPreds,
		    RelOptEngine::AccessMethod AMeth);
	*/
	virtual void Print(ostream& OStr) const;

   };
  //// plan for access to stored table (or sub-block)
  class RelFScanPlan : public RelPhysPlan {
  public:
    RelFScanPlan(int PlanIdent,
		 unsigned char BlockIdent,
		 const RelOptEngine::LogProperties& LProps,
		 unsigned char RelationId,
		 const XxxBitMap& NPreds,
		 const XxxDBEnv& Env,
		 const RelStatsInfo* StatsInfo);
    virtual void Print(ostream& OStr) const;
	virtual void TreePrint(ofstream &OStr, int &curPlanId, int level) const; // add by zhiyuan
    virtual XxxErrCode GeneratePlanOp(const RelSPJOptimizer* Opt,
				      XxxBool TopLevel,
				      XxxExprPlan::VarRemapping& Remap,
				      int InitRemapSrcNum,
				      int FinalRemapSrcNum,
				      RelPlanOp*& PlanOp) const;
  };

  //Begin: Moustafa
  //// plan for access to relation with stream data type (Mar. 2003)
  class RelSScanPlan : public RelPhysPlan {
   public:
    RelSScanPlan(int PlanIdent,
		 unsigned char BlockIdent,
		 const RelOptEngine::LogProperties& LProps,
		 unsigned char RelationId,
		 const XxxBitMap& NPreds,
		 const XxxDBEnv& Env,
		 const RelStatsInfo* StatsInfo);
    virtual void Print(ostream& OStr) const;
    virtual void TreePrint(ofstream &OStr, int &curPlanId, int level) const;
    virtual XxxErrCode GeneratePlanOp(const RelSPJOptimizer* Opt,
				      XxxBool TopLevel,
				      XxxExprPlan::VarRemapping& Remap,
				      int InitRemapSrcNum,
				      int FinalRemapSrcNum,
				      RelPlanOp*& PlanOp) const;
   };
  //End: Moustafa

  //// plan for indexed access to stored table (or sub-block)
  class RelIScanPlan : public RelPhysPlan {
   public:
    XxxBitMap IndxPreds; // preds applied to index(subset of NewPreds)
    XxxBitMap RelDependencies; // any rels that must appear before
                               // this (to bind search arguments)
    // Index into array of indexes for this rel.
    // This array is maintained by the optimizer block.
    // unsigned char IndexId; 

    // for now, just use the index name
    char IndexName[REL_RELNAME_MAX];
    // any match information needed at run time
    const XxxMatchInfo* MatchedInfo;

    RelIScanPlan(int PlanIdent,
		 unsigned char BlockIdent,
		 const RelOptEngine::LogProperties& LProps,
		 unsigned char RelationId,
		 const XxxBitMap& NPreds,
		 const char* IName,
		 const XxxBitMap& IPreds,
		 const XxxBitMap& RelDeps,
		 const XxxMatchInfo* MInfo,
		 const XxxDBEnv& Env,
		 const RelStatsInfo* StatsInfo);
    virtual ~RelIScanPlan() {
      if (MatchedInfo) delete MatchedInfo;
     }

    virtual void Print(ostream& OStr) const;
	virtual void TreePrint(ofstream &OStr, int &curPlanId, int level) const; // add by zhiyuan
 
    virtual XxxErrCode GeneratePlanOp(const RelSPJOptimizer* Opt,
				      XxxBool TopLevel,
				      XxxExprPlan::VarRemapping& Remap,
				      int InitRemapSrcNum,
				      int FinalRemapSrcNum,
				      RelPlanOp*& PlanOp) const;
   };


  //// plan for join, and several of its subclasses
  class JoinPhysPlan : public PhysicalPlan { 
   protected:
    virtual XxxErrCode getCatInfo(const RelSPJOptimizer* Opt,
				  RelPlanOp** TmpPlanOps,
				  XxxBool TopLevel,
				  RelCatalogInfo*& CatInfo) const;
    virtual XxxErrCode getProjInfo(const RelSPJOptimizer* Opt,
				   RelPlanOp** TmpPlanOps,
				   XxxBool TopLevel,
				   const XxxExprPlan::VarRemapping& Remap,
				   XxxValExprPlanList*& TargetList) const;
    virtual void printThis(ostream& OStr) const = 0;
	virtual void TreePrint(ofstream &OStr, int &curPlanId, int level) const; // add by zhiyuan
	virtual void PrintTree(ofstream& OStr) const;

    virtual void getRemap(int InputSrcNum, 
			  int& InitRemapSrcNum,
			  int& FinalRemapSrcNum) const {
			    InitRemapSrcNum = InputSrcNum;
			    FinalRemapSrcNum = InputSrcNum;
			   }
    virtual RelPlanOp* 
      makePlanOp(int NumChildren, 
		 RelPlanOp* Children[],
		 RelCatalogInfo* CatInfo,
		 XxxValExprPlanList* ExprPlanList,
		 XxxBooleanExpressionPlan** SelPredPlans,
		 XxxBooleanExpressionPlan** JoinPredPlans,
		 XxxBooleanExpressionPlan** ResidPredPlans,
		 XxxBooleanExpressionPlan* InitPlan) const = 0;
   public:
    const PhysicalPlan* OuterPlan;
    const PhysicalPlan* InnerPlan;

    RelOptEngine::JoinMethod  Method;
    XxxBitMap JoinPreds;  // subset of NewPreds used in join
    XxxBitMap ResidPreds; // subset of NewPreds applied after join

    JoinPhysPlan(int PlanIdent,
		 unsigned char BlockIdent,
		 const RelOptEngine::LogProperties& LProps,
		 const RelOptEngine::PhysProperties& PProps,
		 const PhysicalPlan* Plan0,
		 const PhysicalPlan* Plan1,
		 const XxxBitMap& NPreds,
		 RelOptEngine::JoinMethod JMeth);

    virtual void Print(ostream& OStr) const;

    virtual XxxErrCode GeneratePlanOp(const RelSPJOptimizer* Opt,
				      XxxBool TopLevel,
				      XxxExprPlan::VarRemapping& Remap,
				      int InitRemapSrcNum,
				      int FinalRemapSrcNum,
				      RelPlanOp*& PlanOp) const;
   };
  //// tuple-at-a-time nested loops
  class TNLJoinPlan : public JoinPhysPlan { 
   protected:
    void planStats(const XxxDBEnv& Env,
		   const RelStatsInfo* InnerStatsInfo);
    virtual void printThis(ostream& OStr) const {}
    virtual RelPlanOp* 
      makePlanOp(int NumChildren, 
		 RelPlanOp* Children[],
		 RelCatalogInfo* CatInfo,
		 XxxValExprPlanList* ExprPlanList,
		 XxxBooleanExpressionPlan** SelPredPlans,
		 XxxBooleanExpressionPlan** JoinPredPlans,
		 XxxBooleanExpressionPlan** ResidPredPlans,
		 XxxBooleanExpressionPlan* InitPlan) const;
   public:
    TNLJoinPlan(int PlanIdent,
		unsigned char BlockIdent,
		const RelOptEngine::LogProperties& LProps,
		const RelOptEngine::PhysProperties& PProps,
		const PhysicalPlan* Plan0,
		const PhysicalPlan* Plan1,
		const XxxBitMap& NPreds,
		const XxxDBEnv& Env,
		const RelStatsInfo* InnerStatsInfo)
      : JoinPhysPlan(PlanIdent, BlockIdent, LProps, PProps,
		     Plan0, Plan1, NPreds, RelOptEngine::TNLOOPS)
    { planStats(Env, InnerStatsInfo);}
   };
  //// page-at-a-time nested loops
  class PNLJoinPlan : public JoinPhysPlan { 
   protected:
    void planStats(const XxxDBEnv& Env,
		   const RelStatsInfo* InnerStatsInfo);
    virtual void printThis(ostream& OStr) const;
    virtual RelPlanOp* 
      makePlanOp(int NumChildren, 
		 RelPlanOp* Children[],
		 RelCatalogInfo* CatInfo,
		 XxxValExprPlanList* ExprPlanList,
		 XxxBooleanExpressionPlan** SelPredPlans,
		 XxxBooleanExpressionPlan** JoinPredPlans,
		 XxxBooleanExpressionPlan** ResidPredPlans,
		 XxxBooleanExpressionPlan* InitPlan) const;
   public:
    int TuplesPerBlock;
    PNLJoinPlan(int PlanIdent,
		unsigned char BlockIdent,
		const RelOptEngine::LogProperties& LProps,
		const RelOptEngine::PhysProperties& PProps,
		const PhysicalPlan* Plan0,
		const PhysicalPlan* Plan1,
		const XxxBitMap& NPreds,
		const XxxDBEnv& Env,
		const RelStatsInfo* InnerStatsInfo,
		unsigned int BlockCard)
      : JoinPhysPlan(PlanIdent, BlockIdent, LProps, PProps,
		     Plan0, Plan1, NPreds, RelOptEngine::PNLOOPS),
    	TuplesPerBlock(BlockCard) {
	planStats(Env, InnerStatsInfo);
       }
   };
  //// indexed nested loops
  class INLJoinPlan : public JoinPhysPlan { 
   protected:
    void planStats(const XxxDBEnv& Env,
		   const RelStatsInfo* InnerStatsInfo);
    virtual void printThis(ostream& OStr) const;
   public:
    INLJoinPlan(int PlanIdent,
		unsigned char BlockIdent,
		const RelOptEngine::LogProperties& LProps,
		const RelOptEngine::PhysProperties& PProps,
		const PhysicalPlan* Plan0,
		const PhysicalPlan* Plan1,
		const XxxBitMap& NPreds,
		const XxxBitMap& JPreds,
		const XxxBitMap& RPreds,
		const XxxDBEnv& Env,
		const RelStatsInfo* InnerStatsInfo)
      : JoinPhysPlan(PlanIdent, BlockIdent, LProps, PProps,
		     Plan0, Plan1, NPreds, RelOptEngine::NDXLOOPS) {
	JoinPreds.CopyBits(JPreds);
	ResidPreds.CopyBits(RPreds);
	planStats(Env, InnerStatsInfo);
      }
   };
  //// sort-merge join
  class SMJoinPlan : public JoinPhysPlan { 
   protected:
    void planStats(const XxxDBEnv& Env,
		   const RelStatsInfo* InnerStatsInfo);
    virtual void printThis(ostream& OStr) const;
    virtual void getRemap(int InputSrcNum, 
			  int& InitRemapSrcNum,
			  int& FinalRemapSrcNum) const {
			    InitRemapSrcNum = 0;
			    FinalRemapSrcNum = InputSrcNum;
			   }
    virtual RelPlanOp* 
      makePlanOp(int NumChildren, 
		 RelPlanOp* Children[],
		 RelCatalogInfo* CatInfo,
		 XxxValExprPlanList* ExprPlanList,
		 XxxBooleanExpressionPlan** SelPredPlans,
		 XxxBooleanExpressionPlan** JoinPredPlans,
		 XxxBooleanExpressionPlan** ResidPredPlans,
		 XxxBooleanExpressionPlan* InitPlan) const;
   public:
    XxxBool ToSortPlan0;  // should plan 0 be sorted
    XxxBool ToSortPlan1;  // should plan 1 be sorted
    int SortRunLength0;
    int SortRunLength1;
    SMJoinPlan(int PlanIdent,
	       unsigned char BlockIdent,
	       const RelOptEngine::LogProperties& LProps,
		   const RelOptEngine::PhysProperties& PProps,
	       const PhysicalPlan* Plan0,
	       const PhysicalPlan* Plan1,
	       XxxBool SortPlan0,
	       XxxBool SortPlan1,
	       const XxxBitMap& NPreds,
	       const XxxBitMap& JPreds,
	       const XxxBitMap& RPreds,
	       const XxxDBEnv& Env,
	       const RelStatsInfo* InnerStatsInfo)
      : JoinPhysPlan(PlanIdent, BlockIdent, LProps, PProps,
		     Plan0, Plan1, NPreds, RelOptEngine::SORTMERGE){
	JoinPreds.CopyBits(JPreds);
	ResidPreds.CopyBits(RPreds);
	ToSortPlan0 = SortPlan0;
	ToSortPlan1 = SortPlan1;
	planStats(Env, InnerStatsInfo);
       }
   };


#endif
