#include <iostream.h>

#include "relerror.h"
#include "relquery.h"
#include "dynspj.h"
#include "relplanop.h"

static int Choose(int N, int C)
{
 // N-choose-C == N!/(N-C)!C!
 int Numerator = 1;
 for (int i = N; i > (N-C); i--) {
   Numerator *= i;
  }
 int Denominator = 1;
 for (int j = C; j > 0; j--) {
   Denominator *= j;
  }
 return (int)(Numerator/Denominator);
}

//
// Methods of PlanTable
//
RelSPJDynOptimizer::
DynPlanTable::DynPlanTable() 
{
 NumRels = 0;
 PlanArraySize = 0;
 PlanArray = NULL;
 EndArray = 0;
}
RelSPJDynOptimizer::
DynPlanTable::~DynPlanTable()
{
 for (int i = 0; i < PlanArraySize; i++) {
   if (PlanArray[i]) delete PlanArray[i];
  }
 if (PlanArray) delete [] PlanArray;
}

void RelSPJDynOptimizer::
DynPlanTable::Init(int Size, const PlanSpaceClass* PSpace) 
{
 PlanSpace = PSpace;
 NumRels = Size;

 // N choose NumRels
 PlanArraySize = Choose(PlanSpace->RelCount, NumRels);
 cerr << "For " << NumRels << " Rels, use " << PlanArraySize << " combinations" << endl;

 PlanArray = new LogicalPlan*[PlanArraySize];
 for (int i = 0; i < PlanArraySize; i++)
   PlanArray[i] = NULL;
 EndArray = 0;
}

XxxErrCode RelSPJDynOptimizer::
DynPlanTable::Print(ostream& OStr) const
{
 for (int i = 0; i < EndArray; i++) {
   OStr << "\t";
   OStr << " *** Plan " << i << " *** " << endl;
   PlanArray[i]->Print(OStr);
  }
 return XXX_OK;
}

XxxErrCode RelSPJDynOptimizer::
DynPlanTable::PrintStats(ostream& OStr) const
{
 OStr << "\tNumRels = " << NumRels << ", ArraySize = " << PlanArraySize << ", EndArray = " << EndArray << endl;
 OStr << "\tPlans Generated: " << Stats.PlansGenerated << endl;
 OStr << "\tPlans Added: " << Stats.PlansAdded << endl;
 OStr << "\tPlans Pruned: " << Stats.PlansPruned << endl;
 OStr << endl;

 return XXX_OK;
}

XxxErrCode RelSPJDynOptimizer::
DynPlanTable::AddPlan(const XxxDBEnv& Env,
		   PhysicalPlan* NewPlan,
		   int& NumPlansAdded,
		   int& NumPlansPruned)
{
 XxxSetCursor Csr;
 NumPlansAdded = 0;
 NumPlansPruned = 0;

 // check to see if this plan is already there
 if (OpenLPlanScan(NewPlan->LogProps, Csr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, " Error opening plan scan ");
   return XXX_NOT_OK;
  }
 LogicalPlan* LPlan = NULL;
 if (GetNextLPlan(Csr, LPlan) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, " Error in getNext of plan scan ");
   return XXX_NOT_OK;   
  }
 if (Csr.IsValid() && LPlan) {
   // add NewPlan to existing plan options
   if (LPlan->AddPlan(Env, NewPlan, 
		      NumPlansAdded, NumPlansPruned) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error adding plan to plan options");
     return XXX_NOT_OK;
    }
   Stats.PlansPruned += NumPlansPruned;
   Stats.PlansAdded += NumPlansAdded;
  }
 else {
   if (EndArray >= PlanArraySize) {
     REL_NOTE_ERROR(REL_BADOPT, " Plan Array Full!! ");
     return XXX_NOT_OK;
    }
   // create a new LogicalPlan with this plan
   LPlan = new LogicalPlan(NewPlan->LogProps);
   if (LPlan->AddPlan(Env, NewPlan, 
		      NumPlansAdded, NumPlansPruned) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error adding plan to plan options");
     return XXX_NOT_OK;
    }
   PlanArray[EndArray++] = LPlan;
   Stats.PlansPruned += NumPlansPruned;
   Stats.PlansAdded += NumPlansAdded;
  }

 // close the scan
 if (CloseLPlanScan(Csr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, " Error closing plan scan ");
   return XXX_NOT_OK;
  }

 Stats.PlansGenerated++;
 return XXX_OK;
}

// LogicalPlan scan
XxxErrCode RelSPJDynOptimizer::
DynPlanTable::OpenLPlanScan(const RelOptEngine::LogProperties& LogSpec,
			 XxxSetCursor& Csr) const
{
 Csr.NumberRead = 0;
 Csr.AtEnd = false;
 struct LPlanScanHandle* SH = new struct LPlanScanHandle(LogSpec);
 Csr.State = SH;
 return XXX_OK;
}

XxxErrCode RelSPJDynOptimizer::
DynPlanTable::GetNextLPlan(XxxSetCursor& Csr,
			LogicalPlan*& LPlan) const
{
 struct LPlanScanHandle* SH = (struct LPlanScanHandle *)(Csr.State);
 
 while (SH->ScanIndex < EndArray) {
   enum RelOptEngine::MatchQuality Quality =
    RelOptEngine::BAD;
   if (PlanArray[SH->ScanIndex])
     PlanArray[SH->ScanIndex]->LogProps.Matches(SH->LogProps,Quality);
  if (Quality == RelOptEngine::GOOD) {
     LPlan = PlanArray[SH->ScanIndex];
     Csr.NumberRead++;
     SH->ScanIndex++;
     return XXX_OK;
    }
   else {
     // return to loop
     SH->ScanIndex++;
    }
  }
 Csr.AtEnd = true;
 return XXX_OK;
}

XxxErrCode RelSPJDynOptimizer::
DynPlanTable::CloseLPlanScan(XxxSetCursor& Csr) const
{
 struct LPlanScanHandle* SH = (struct LPlanScanHandle *)(Csr.State);
 delete SH;
 return XXX_OK;
}

// Plan scan
XxxErrCode RelSPJDynOptimizer::
DynPlanTable::OpenPPlanScan(const RelOptEngine::LogProperties& LogSpec, 
			 XxxSetCursor& Csr) const
{
 Csr.NumberRead = 0;
 Csr.AtEnd = false;
 struct PPlanScanHandle* SH = new struct PPlanScanHandle();
 if (SH->Init(this, LogSpec) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error initializing LPlan scan");
   return XXX_NOT_OK;
  }
 Csr.State = SH;
 return XXX_OK;
}

XxxErrCode RelSPJDynOptimizer::
DynPlanTable::GetNextPPlan(XxxSetCursor& Csr,
			PhysicalPlan*& Plan) const
{
 struct PPlanScanHandle* SH = (struct PPlanScanHandle *)(Csr.State);

 if (!SH->CurrLPlan) {
   // invoke GetNextLPlan
   if (GetNextLPlan(SH->LPlanCsr, SH->CurrLPlan) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error reading next plan option");
     return XXX_NOT_OK;
    }
   if (!SH->LPlanCsr.IsValid() || !SH->CurrLPlan) {
     Csr.AtEnd = true;
     return XXX_OK;
    }
   SH->CurrPlanList = SH->CurrLPlan->PPlanList.Next;
  }

 Csr.NumberRead++;
 Plan = (PhysicalPlan*)(SH->CurrPlanList->Plan);
 SH->CurrPlanList = SH->CurrPlanList->Next;
 if (!SH->CurrPlanList) SH->CurrLPlan = NULL;
 return XXX_OK;
}

XxxErrCode RelSPJDynOptimizer::
DynPlanTable::ClosePPlanScan(XxxSetCursor& Csr) const
{
 struct PPlanScanHandle* SH = (struct PPlanScanHandle *)(Csr.State);
 if (SH->ShutDown(this) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, " Error closing plan options scan");
   return XXX_NOT_OK;
  }
 delete SH;
 return XXX_OK;
}

XxxErrCode RelSPJDynOptimizer::
DynPlanTable::PPlanScanHandle::Init(const DynPlanTable* PlanTab, 
				 const RelOptEngine::LogProperties& LProps)
{
 // open the LPlanScan
 if (PlanTab->OpenLPlanScan(LProps, LPlanCsr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error opening Plan Options Scan");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelSPJDynOptimizer::
DynPlanTable::PPlanScanHandle::ShutDown(const DynPlanTable* PlanTab)
{
 // close the LPlanScan
 if (PlanTab->CloseLPlanScan(LPlanCsr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error closing Plan Options Scan");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}
