#include <iostream.h>

#include "reladt.h"
#include "relopt.h"
#include "relerror.h"
#include "relquery.h"
#include "relexec.h"
#include "relenv.h"
#include "shoresm.h"

const char *RelOptEngine::JoinMethodNames[7] = {
  "Unknown",
  "SortMerge",
  "PageNLoops",
  "TupleNLoops",
  "NdxLoops",
  "HashJoin",
  "Invalid"
 };

RelSpecParameters::RelSpecParameters(const char* RName, 
				     XxxBool Named,
				     XxxBool Create, 
				     XxxBool Append)
{
 PrintMode = XXX_FALSE;
 ToCreate = Create;
 ToAppend = Append;
 NamedRel = Named;
 if (Named && RName) {
   strncpy(RelName, RName, XXX_NODE_NAME_MAX);
   RelName[XXX_NODE_NAME_MAX-1] = 0;
  }
 initDBEnv();
}

RelSpecParameters::RelSpecParameters(XxxBool ToPrint) 
: PrintMode(ToPrint) 
{
 if (ToPrint)
   ToCreate = XXX_FALSE;
 else
   ToCreate = XXX_TRUE;
 ToAppend = XXX_FALSE;
 NamedRel = XXX_FALSE;
 initDBEnv();
}
 
void RelSpecParameters::initDBEnv()
{
  // get the values from XXX_THISENV()
  DBEnv = XXX_THISENV().DBEnv;
  // Of course, one option is to actually get the current
  // values at this time, but this prevents us from explicitly
  // setting the values thru XXX_THISENV (needed for debugging)
  if (XXX_THISRELENV()->DebugOpt) {
    cerr << "Current DBEnv" << endl;
    DBEnv.Print(cerr);
  }
}

/**
 * Each query node is a "query block" in the Starburst sense.
 * Optimization across query blocks is heuristic. Within each
 * query block, it is cost-based.
 */
XxxErrCode RelOptEngine::OptimizeQuery(RelQueryClass *Query, 
				       const RelSpecParameters* Params,
				       RelQueryPlan*& NewPlan)
{
 // NOTE: Params should always be non-NULL!
 ASSERT((Params));

 // Set up the environment for the planning process.
 // This involves traversing the query, and counting the
 // the number of query blocks, and the number of predicates, 
 // relations, etc involved.
 OptQueryEnv *Env = new OptQueryEnv(Query, Params);
 
 // actually do the planning in a bottom up manner
 if (createQueryPlan(Query, Env, NewPlan) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error creating block-level query plan");
   return XXX_NOT_OK;
  }
 
 // clean up after planning
 delete Env;

 return XXX_OK;
}

//
// Planning of query blocks is done bottom-up. Therefore, this
// function is recursively invoked on the children in a postfix
// depth-first traversal of the query graph.
//
XxxErrCode RelOptEngine::createQueryPlan(RelQueryClass *Query,
					 OptQueryEnv  *Env,
					 RelQueryPlan*& NewPlan)
{
 XxxErrCode rc = XXX_OK;
 // traverse the query in a bottom up manner
 void *Cursor; XxxBool Success; XxxQueryNode* Node;
 int BlockCounter = 0;

 if (Query->InitTraversal(Cursor, XxxQueryDag::DEPTH_FIRST_POSTFIX,
			  Query->Root) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to open traversal of query graph");
   return XXX_NOT_OK;
  }
 while (((rc = Query->TraverseNext(Cursor, Node, Success)) == XXX_OK) &&
	(Success)) {
   if (createBlockPlan((RelQueryNode*)Node, BlockCounter, Env,
		       Env->BlockEnvs[BlockCounter].PlanC) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Unable to optimize query block");
     return XXX_NOT_OK;
    }
   BlockCounter++;

  }
 if (rc != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to get next in query graph traversal");
   return XXX_NOT_OK;
  }
 if (Query->EndTraversal(Cursor) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to close traversal of query graph");
   return XXX_NOT_OK;
  }

 
 ASSERT((BlockCounter >= Env->BlockCount));

 // return the top-level plan
 NewPlan = new RelQueryPlan(Env->BlockEnvs[Env->BlockCount-1].PlanC,
			    Env->OptSpecs);
 return XXX_OK;
}

// 
// Cost-based planning for one query block.
//
XxxErrCode RelOptEngine::createBlockPlan(RelQueryNode* Node,
					 unsigned char BlockIdent,
					 OptQueryEnv  *Env,
					 RelBlockPlan*& PlanC)
{
 XxxErrCode rc = XXX_OK;

 if (!Node->Op) { // base relation
   return planBaseBlock(Node, BlockIdent, Env, PlanC);
  }

 switch (Node->Op->Type()){
  case REL_SPJ:
   // defined in planspj.C
   return planSPJBlock(Node, BlockIdent, Env, PlanC); 
  case REL_GBY: 
   // defined in plangby.C
   return planAggrBlock(Node, BlockIdent, Env, PlanC);  
  case REL_OBY:
   // defined in planoby.C
   return planOrderBlock(Node, BlockIdent, Env, PlanC); 
//begin Ihab
 case REL_KRJ:
   //defined in planKRJ.C
   return planKRJBlock(Node, BlockIdent, Env, PlanC); 
	 return XXX_OK;
 case REL_STOP:
   // defined in planstop.C
   return planStopBlock(Node, BlockIdent, Env, PlanC); 
//end Ihab
  default:
   XXX_NOTE_ERROR(XXX_INCOMPL, "Unknown relation block type");
   return XXX_NOT_OK;
  } 
}

//
// This will return the chosen plan in ChildPlanOp, if found.
// Otherwise, it will return a NULL. Either way, the return
// value of the function is XXX_OK, unless there is some
// error that occurs.
//
XxxErrCode 
RelOptEngine::findChildBlockPlan(RelQueryNode* ChildNode,
				 OptQueryEnv* Env,
				 XxxBool& BlockFound,
				 RelBlockPlan* &ChildBlockPlan)
{
 ChildBlockPlan = 0;
 BlockFound = XXX_FALSE;

 int j;
 for (j = 0; j < Env->BlockCount; j++) {
   if (Env->BlockEnvs[j].Block == ChildNode) {
     BlockFound = XXX_TRUE;
     if (!Env->BlockEnvs[j].PlanC) {
       REL_NOTE_ERROR(REL_BADOPT, "Unable to find plan for child");
       return XXX_NOT_OK;
      } // if no plan or bad plan
     ChildBlockPlan = Env->BlockEnvs[j].PlanC;
     break;
    } // if the right node was found
  } // for all nodes in the env

 if (j == Env->BlockCount) {
   // REL_NOTE_ERROR(REL_BADOPT, "Unable to find plan for child");
   BlockFound = XXX_FALSE;
   return XXX_OK;
  }

 return XXX_OK;
}

//
// Methods of RelOptEngine::OptQueryEnv
//
RelOptEngine::OptQueryEnv::OptQueryEnv(RelQueryClass* Query,
				       const RelSpecParameters* Params)
{
 XxxErrCode rc = XXX_OK;

 OptSpecs = Params;

 BlockCount = 0;
 BlockEnvs = 0;

 // traverse the query and determine the number of query blocks (nodes)
 void *Cursor; XxxBool Success; XxxQueryNode* Node;

 if (Query->InitTraversal(Cursor, XxxQueryDag::DEPTH_FIRST_POSTFIX,
			  Query->Root) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to open traversal of query graph");
   return;
  }
 while (((rc = Query->TraverseNext(Cursor, Node, Success)) == XXX_OK) &&
	(Success)) {
   BlockCount++;
  }
 if (rc != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to get next in query graph traversal");
   return;
  }
 if (Query->EndTraversal(Cursor) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to close traversal of query graph");
   return;
  }

 if (XXX_THISRELENV()->DebugOpt) {
   cerr << " Query Block Count  = " << BlockCount << endl;
 }

 BlockEnvs = new PerBlockEnv[BlockCount];

 // traverse the query again, and fill one PerBlockEnv per node
 int BlockCounter = 0;
 if (Query->InitTraversal(Cursor, XxxQueryDag::DEPTH_FIRST_POSTFIX,
			  Query->Root) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to open traversal of query graph");
   return;
  }
 while (((rc = Query->TraverseNext(Cursor, Node, Success)) == XXX_OK) &&
	(Success)) {
   BlockEnvs[BlockCounter].Block = (RelQueryNode*)Node;
   BlockEnvs[BlockCounter].PlanC = 0;
   BlockCounter++;
  }
 if (rc != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to get next in query graph traversal");
   return;
  }
 if (Query->EndTraversal(Cursor) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to close traversal of query graph");
   return;
  }
 ASSERT((BlockCounter == BlockCount));
}

RelOptEngine::OptQueryEnv::~OptQueryEnv()
{
 if (BlockCount && BlockEnvs)
   delete [] BlockEnvs;
}

//
// Methods of RelOptEngine::OptStats
//
void RelOptEngine::
OptStats::Print(ostream& OStr) const
{
 OStr << "  REL Optimizer Engine:" << endl;
 OStr << "\t OptimizeQuery : " << OptimizeQuery << endl;
 OStr << "\t PlansConsidered : " << PlansConsidered << endl;
}

//
// Methods of RelOptEngine::PlanObjectList
//
void RelOptEngine::PlanObjectList::Destroy()
{
 PlanObjectList* Tmp = Next;
 while (Tmp) {
   // delete the plan
   if (Tmp->Plan) delete Tmp->Plan;
   PlanObjectList* Tmp2 = Tmp->Next;
   // delete the list cell
   delete Tmp;
   Tmp = Tmp2;
  }
}

//
// Methods of RelQueryPlan
//
void RelQueryPlan::Print(ostream& OStr) const
{
 if (TopBlockPlan) TopBlockPlan->Print(OStr);
}

XxxErrCode 
RelQueryPlan::GenerateExecPlan(const RelExecParameters* Params,
			       RelPlanExecutor*& PlanExec) const
{
 if (!IsValid()) {
   REL_NOTE_ERROR(REL_BADEXEC, "Invalid query plan");
   return XXX_NOT_OK;
  }

 RelPlanOp* TopPlanOp;
 // Use PlanId == -1 since this is a top-level query block
 if (TopBlockPlan->GeneratePlanOp(Params, 
				  TopPlanOp) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Error generating RelPlanOp");
   return XXX_NOT_OK;
  }

 if (!OptParams || OptParams->PrintMode) {
   PlanExec = new RelExecEngine::PrintPlanExecutor(TopPlanOp);
  }
 else {
   PlanExec = new RelExecEngine
     ::StorePlanExecutor(TopPlanOp,
			 (OptParams->NamedRel?OptParams->RelName:0),
			 OptParams->ToCreate,
			 OptParams->ToAppend);
  }
 return XXX_OK;
}

XxxErrCode 
RelQueryPlan::GenerateEvalCursor(const RelExecParameters* Params,
				 RelEvalCursor*& EvalCursor) const
{
 if (!IsValid()) {
   REL_NOTE_ERROR(REL_BADEXEC, "Invalid query plan");
   return XXX_NOT_OK;
  }

 RelPlanOp* TopPlanOp;
 // Use PlanId == -1 since this is a top-level query block
 if (TopBlockPlan->GeneratePlanOp(Params, 
				  TopPlanOp) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Error generating RelPlanOp");
   return XXX_NOT_OK;
  }

 EvalCursor = new RelEvalCursor(TopPlanOp, Params);
 return XXX_OK;
}

//
// Methods of RelBlockPlan
//
RelBlockPlan::RelBlockPlan(XxxErrCode& RC,
			   unsigned char BlockIdent,
			   int NumChildren, 
			   RelBlockPlan* Children[],
			   RelCatalogInfo* CatInfo,
			   RelOptEngine::OptQueryEnv* Env)
{
 BlockId = BlockIdent; 
 //Begin: Moustafa
 // To mark plan for nodes that has streaming input (Mar. 2003)
 HasStreamInput = XXX_FALSE;
 //End: Moustafa

 // replicate the catalog info, so that the owner may reclaim it
 CatalogInfo = CatInfo->Replicate1();
 ChildArrayLen = NumChildren;
 ChildBlockPlanArray = new RelBlockPlan*[ChildArrayLen];
 for (int i = 0; i < ChildArrayLen; i++) {
   ChildBlockPlanArray[i] = Children[i];
   //Begin: Moustafa
   // mark the plan as has_streaming_input, if one of the children
   // has_streaming_input (Mar. 2003)
   if (Children[i]->HasStreamInput)
     HasStreamInput = XXX_TRUE;
   //End: Moustafa
 }

 NumPlans = 0;
 // FirstPlan = NULL;
 HighestPlanId = 0;

 RC = XXX_OK;
}
RelBlockPlan::~RelBlockPlan()
{
 for (int i = 0; i < ChildArrayLen; i++) {
   if (ChildBlockPlanArray[i]) delete ChildBlockPlanArray[i];
  }
 delete [] ChildBlockPlanArray;

 if (CatalogInfo) {
   CatalogInfo->Destroy();
   delete CatalogInfo; 
   CatalogInfo = NULL; 
  }

 PlanList.Destroy(); // destroy the list of plans
}

void RelBlockPlan::Print(ostream& OStr) const
{
 OStr << "(";
 for (int i = 0; i < Arity(); i++) {
   ChildBlockPlanArray[i]->Print(OStr);
   if (i < Arity()-1) {
     OStr << ", ";
    }
  }
 OStr << ")";
}

XxxErrCode RelBlockPlan::
GetBestPlan(const RelOptEngine::PhysProperties& DesiredPProps,
	    RelOptEngine::PhysProperties*& PlanProps,
	    int& PlanId,
	    RelOptEngine::Cost& Usage,
	    XxxBool& PlanFound)
{
 // walk thru the plans in the plan list, and find the best plan
 RelOptEngine::PlanObject *BestPlan;
 RelOptEngine::PlanObjectList *CurrentList;
 BestPlan = NULL;
 
 CurrentList = PlanList.Next;

 BestPlan = CurrentList->Plan;

 while (CurrentList) {
   RelOptEngine::MatchQuality Quality = RelOptEngine::BAD;
   CurrentList->Plan->PhysProps.Matches(DesiredPProps, Quality);
   if (Quality == RelOptEngine::PARTIAL) {
     // TODO:L_OPT
     // Step 1: add glue to make the match GOOD
     // Step 2: if this plan already exists, drop it
     // Step 3: otherwise, add the plan _after_ the CurrentPlan
    }
   else if (Quality == RelOptEngine::GOOD) {
     if (!BestPlan ||
	 (CurrentList->Plan->Usage.FullCost() < 
	  BestPlan->Usage.FullCost())) {
       // compare costs of best plan and current plan
       BestPlan = CurrentList->Plan;
      }
     // else ignore current plan
    }
   CurrentList = CurrentList->Next;
  }
 if (BestPlan) {
   PlanFound = XXX_TRUE;
   PlanId = BestPlan->PlanId;
   Usage = BestPlan->Usage;
   PlanProps = &(BestPlan->PhysProps);
  }
 else {
   PlanFound = XXX_FALSE;
  }
 return XXX_OK;
}

XxxErrCode RelBlockPlan::
GetBestPlan(const RelExecParameters* Params, // ignored for now
	    int& PlanId,
	    RelOptEngine::Cost& Usage,
	    XxxBool& PlanFound) const
{
 PlanFound = XXX_FALSE;
 // walk thru the plans in the plan list, and find the best plan

 RelOptEngine::PlanObject *ChosenPlan;
 RelOptEngine::PlanObjectList* CurrentList;
 ChosenPlan = NULL;
 
 CurrentList = PlanList.Next;
 while (CurrentList) {
   if (!ChosenPlan ||
       (CurrentList->Plan->Usage.FullCost() < 
	ChosenPlan->Usage.FullCost())) {
     // compare costs of best plan and current plan
     ChosenPlan = CurrentList->Plan;
    }
   // else ignore current plan
   CurrentList = CurrentList->Next;
  } // while
 if (ChosenPlan) {
   PlanFound = XXX_TRUE;
   Usage = ChosenPlan->Usage;
   PlanId = ChosenPlan->PlanId;
   return XXX_OK;
  }
 else {
   REL_NOTE_ERROR(REL_BADOPT, "No plan in query block");
   return XXX_NOT_OK;
  }
}

XxxErrCode RelBlockPlan::
GetPlan(int PlanId,
	RelOptEngine::PlanObject*& ChosenPlan) const
{
 if (PlanId < 0) {
   REL_NOTE_ERROR(REL_BADOPT, "Given invalid plan id");
   return XXX_NOT_OK;
  } // if PlanId < 0
 else {
   // walk thru the plans in the plan list, and find the desired one
   RelOptEngine::PlanObjectList *CurrentList;
   ChosenPlan = NULL;
 
   CurrentList = PlanList.Next;
   while (CurrentList) {
     if (CurrentList->Plan->PlanId == PlanId) {
       ChosenPlan = CurrentList->Plan;
       return XXX_OK;
      }
     CurrentList = CurrentList->Next;
    } // while
   REL_NOTE_ERROR(REL_BADOPT, "Unable to find desired plan");
   return XXX_NOT_OK;
  }
}

XxxErrCode RelBlockPlan::
GeneratePlanOp(const RelExecParameters* Params,
	       RelPlanOp*& PlanOp) const
{
 XxxGenericPlan::PlanUsage Usage;
 XxxBool PlanFound = XXX_FALSE;
 int PlanId;

 // find an appropriate PlanId
 if (GetBestPlan(Params, PlanId, Usage, PlanFound) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to find plan");
   return XXX_NOT_OK;
  }

 if (GeneratePlanOp(PlanId, PlanOp) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to find plan");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelBlockPlan::
redirectDependencies(const XxxValueExprPlan::VarRemapping& Remap,
		     int CorrelationHt)
{
  for (int i = 0; i < ChildArrayLen; i++) {
    REL_CALL_FN(ChildBlockPlanArray[i]->
		RedirectDependencies(Remap, CorrelationHt+1),
		REL_BADOPT, "Error redirecting child dependencies");
  }
  return XXX_OK;
}

//
// Methods of RelOptEngine::LogProperties
//
RelOptEngine::
LogProperties::LogProperties(const LogProperties& LProps)
: Rels(LProps.Rels), Preds(LProps.Preds)
{
 Valid = LProps.Valid;
 Card = LProps.Card;
 TupleSize = LProps.TupleSize;
}

void RelOptEngine::
LogProperties::Print(ostream& OStr) const
{
 OStr << " Card = " << Card << " ";
 OStr << " TupleSize = " << TupleSize;
 OStr << " Rels: ";
 Rels.Print(OStr);
 OStr << " Preds: ";
 Preds.Print(OStr);
}

void RelOptEngine::
LogProperties::Matches(const LogProperties& LProps,
		       RelOptEngine::MatchQuality& Quality) const
{
 Quality = RelOptEngine::GOOD;

 // if the LogProperties are invalid, the match is trivially true
 if ((!LProps.Valid) || !Valid) return ;

 if (!((Rels.Equals(LProps.Rels)) &&
     (Preds.Equals(LProps.Preds)))) {
   Quality = RelOptEngine::BAD;
  }
}

void RelOptEngine::
PhysProperties::Matches(const PhysProperties& PProps,
			RelOptEngine::MatchQuality& Quality) const
{
 if (SortOrder.Matches(PProps.SortOrder))
   Quality = RelOptEngine::GOOD;
 else {
	 if (SortOrder.interestingMatches(PProps.SortOrder))
		 Quality = RelOptEngine::GOOD;
	 else
		 Quality = RelOptEngine::BAD;
 }
}

