#ifndef XXX_RELOPT_H
#define XXX_RELOPT_H

#include "reladt.h"
#include "generic.h"
#include "relquery.h"
#include "prd-sort.h"
#include "misc.h"  // for BitMaps

XXX_EXTERN_CLASS_DECL(RelPlanOp);
XXX_EXTERN_CLASS_DECL(RelPlanExecutor);
XXX_EXTERN_CLASS_DECL(RelExecParameters);
XXX_EXTERN_CLASS_DECL(RelEvalCursor);
XXX_EXTERN_CLASS_DECL(XxxRelation);

// forward declarations of classes in this file
XXX_EXTERN_CLASS_DECL(RelSpecParameters);
XXX_EXTERN_CLASS_DECL(RelBlockPlan);
XXX_EXTERN_CLASS_DECL(RelQueryPlan);

/**
 * Encapsulates all optimization code.
 */
class RelOptEngine {
 public:
  typedef enum {
    UNKNOWNSCAN,
    FILESCAN,
    NDXSCAN,
    //Begin: Moustafa
    // decalsre SSCAN as one of the eligible operators (Mar. 2003)
    SSCAN     //Stream Scan
    //End: Moustafa
   } AccessMethod;
  typedef enum {
    UNKNOWN_JOIN,
    SORTMERGE,
    PNLOOPS, // page-at-a-time
    TNLOOPS, // tuple-at-a-time
    NDXLOOPS, // indexed nl
    //Begin: Moustafa
    // decalre HWINDOWN and NLWINDOW as eligible operators (Mar. 2003) using the FEW-join and BEW-join 
    //  algorithms
    HBEWINDOW,  // hash-window join (BEW-join Algorithm)
    NLBEWINDOW, // Nested-Loop-window join  (BEW-join Algorithm)
    HFEWINDOW,  // hash-window join (FEW-join Algorithm)
    NLFEWINDOW, // Nested-Loop-window join  (FEW-join Algorithm)
    //End: Moustafa
    HASHJOIN, // symmetric hash join
    // GRACEHASH, // not supported yet
    RJOIN,        //ihab rank join
    INVALID_JOIN
   } JoinMethod;
//Bjarke, aug. 8: const added to type
  static const char* JoinMethodNames[];
  typedef enum {
    UNKNOWN,
    TUPLE,      // tuple-at-a-time
    TUPLECACHE, // tuple-at-a-time with result caching
    SET,        // set-at-a-time
   } ExprEvalMethod;
  enum MatchQuality {
    BAD = 0,
    PARTIAL = 1,
    GOOD = 2
   };

  // Physical props make sense in the context of one particular schema.
  //// Physical properties of a plan.       
  class PhysProperties { public:
    XxxSortInfo SortOrder;
    PhysProperties() : SortOrder(0) {}
    PhysProperties(int NumAttrs) : SortOrder(NumAttrs) {}

	PhysProperties(const PhysProperties& PProps) 
      : SortOrder(PProps.SortOrder) {}
    ~PhysProperties() {}

	// @author: wee hyong
	// @version: 19 June 2000
	// added to initialize SortOrder
	PhysProperties(const XxxSortInfo& SInfo) 
      : SortOrder(SInfo) {}

    void Matches(const PhysProperties& PProps,
		 enum MatchQuality& Quality) const;
   };

  // Logical props only make
  // sense in the context of one particular query block.
  // Therefore, there is no point using logical props across
  // query blocks (one could do this, but it makes all the
  // bitmaps, etc very large).
  //// Logical properties of a plan. 
  class LogProperties { public:
    XxxBool Valid;
    int Card; // cardinality
    int TupleSize; 

    //(these next two bitmaps are in the context of one particular
    // query block)
    XxxBitMap Rels; // which relations does it hold
    XxxBitMap Preds; // which predicates does it hold

    LogProperties(int TotalRels, int TotalPreds) 
      : Rels(TotalRels), Preds(TotalPreds) { Valid = true;}
    LogProperties(const LogProperties& LProps);
    LogProperties() { Valid = false;}
    void Matches(const LogProperties& LProps,
		 enum MatchQuality& Quality) const;
    void Print(ostream& OStr) const;
   };

  typedef XxxGenericPlan::PlanUsage Cost;

  /**
    * This is the superclass that can be subclassed to
    * keep track of the plans in each query block
    */
  class PlanObject { 
   public:
    const int PlanId;
    const unsigned char BlockId;

    // usage === "cost" + "resources"
    Cost Usage;
    // any physical properties
    PhysProperties PhysProps;

    PlanObject(int PlanIdent,
	       unsigned char BlockIdent)
     : PlanId(PlanIdent), BlockId(BlockIdent) {}

	// added to provide constructor to init PhysProps
	// @author Wee Hyong
	// @version 19 June 2000
	PlanObject(int PlanIdent,
	       unsigned char BlockIdent,
		   const PhysProperties& PProps)
     : PlanId(PlanIdent), BlockId(BlockIdent), PhysProps(PProps) {}
	
    virtual ~PlanObject() {}
    virtual void Print(ostream& OStr) const = 0;
   };

  /**
   *  This is a list of plan objects
   */
  class PlanObjectList { public:
    PlanObject* Plan;
    PlanObjectList* Next;
    PlanObjectList(PlanObject* NewPlan = NULL,
		   PlanObjectList* NextPlan = NULL) {
      Plan = NewPlan;
      Next = NextPlan;
     }
    ~PlanObjectList() {}
    void Destroy();// recursively step thru and destroy
    void Add(PlanObject* NewPlan) {
      PlanObjectList* Tmp = new PlanObjectList(NewPlan, Next);
      Next = Tmp;
     }
   };

  /**
   * This structure keeps track of _all_ plans of all
   * query blocks.
   **/
  class OptQueryEnv { public:
    struct PerBlockEnv {
      RelQueryNode* Block;
      RelBlockPlan* PlanC;
     };
    PerBlockEnv* BlockEnvs;
    int BlockCount;

    const RelSpecParameters* OptSpecs;

    // any other data structures: probably need a hash table of
    // all predicates, etc

    OptQueryEnv(RelQueryClass* Query,
		const RelSpecParameters* Params);
    ~OptQueryEnv();
   };

 protected:
  ////
  class OptStats {
   public:
    // need to flesh this out once there is a real optimizer
    int OptimizeQuery;
    int PlansConsidered;
    int Extra[5];
    void Reset() {
      OptimizeQuery = PlansConsidered = 0;
      Extra[0] = Extra[1] = Extra[2] = Extra[3] = Extra[4] = 0;
     }
    void Print(ostream& OStr) const;
    OptStats() { Reset(); }
   };

  OptStats Stats;
  XxxRelationEngine* RelEng;

  // 1) plan an entire query
  XxxErrCode createQueryPlan(RelQueryClass* Query,
			     OptQueryEnv* Env,
			     RelQueryPlan*& NewPlan);
  // 2) plan a specific block (assuming all blocks that
  //      it accesses have already been planned)
  XxxErrCode createBlockPlan(RelQueryNode* Node,
			     unsigned char BlockId,
			     OptQueryEnv* Env,
			     RelBlockPlan*& Plan);

  XxxErrCode planBaseBlock(RelQueryNode* Node,
			   unsigned char BlockIdent,
			   OptQueryEnv* Env,
			   RelBlockPlan*& Plan);

  // routines for planning individual blocks
  XxxErrCode planSPJBlock(RelQueryNode* Node,
			  unsigned char BlockIdent,
			  OptQueryEnv* Env,
			  RelBlockPlan*& Plan);
  XxxErrCode planAggrBlock(RelQueryNode* Node,
			   unsigned char BlockIdent,
			   OptQueryEnv* Env,
			   RelBlockPlan*& Plan);
  XxxErrCode planOrderBlock(RelQueryNode* Node,
			    unsigned char BlockIdent,
			    OptQueryEnv* Env,
			    RelBlockPlan*& Plan);

// begin Ihab
// Rank join query block
// right now it is in-block optimization
 XxxErrCode planKRJBlock(RelQueryNode* Node,
			  unsigned char BlockIdent,
			  OptQueryEnv* Env,
			  RelBlockPlan*& Plan); 
  XxxErrCode planStopBlock(RelQueryNode* Node,
			   unsigned char BlockIdent,
			   OptQueryEnv* Env,
			   RelBlockPlan*& Plan);
//end Ihab  

  // the above routines call findChildBlockPlan to retrieve the
  // already computed plans for a particular child
  XxxErrCode findChildBlockPlan(RelQueryNode* ChildNode,
				OptQueryEnv* Env,
				XxxBool& BlockFound,
				RelBlockPlan* &ChildBlockPlan);

 public:  
  RelOptEngine(XxxRelationEngine* SrcEng) { RelEng = SrcEng; }
  XxxErrCode OptimizeQuery(RelQueryClass *Query, 
			   const RelSpecParameters* Params,
			   RelQueryPlan*& DummyPlan);
  virtual void PrintStats(ostream& OStr) const 
   { Stats.Print(OStr); }
  virtual void ResetStats() 
   { Stats.Reset(); }
 };

/**
 * This class encapsulates the input parameters to the optimizer
 */
class RelSpecParameters {
  void initDBEnv();
 public:
  XxxBool PrintMode; // if this is set, print out the answer, 

  // these matter only if the answer is being written to a relation
  XxxBool NamedRel;
  char RelName[XXX_NODE_NAME_MAX];
  XxxBool ToCreate; // if only this is set, create a new relation
  XxxBool ToAppend; // if only this is set, append to an existing relation
  // if both are set, try to append, if there is an existing relation,
  // otherwise, create one...

  // should it be optimized for all answers? alternative is
  // optimize for first answer tuple
  XxxBool OptForAllAnswers;

  // Desirable physical properties.
  // In addition, the best plan that does not match
  // this will be generated.
  // This should be ignored by all but the top query block
  RelOptEngine::PhysProperties PhysProps;
  
  // Representation of system state for optimization.
  // This specifies several parameters like buffer size, etc
  XxxDBEnv DBEnv;

  // more constructors may need to be added...
  RelSpecParameters(const char* RName, 
		    XxxBool Named,
		    XxxBool Create, 
		    XxxBool Append);
  RelSpecParameters(XxxBool ToPrint = XXX_TRUE);

  ~RelSpecParameters() {}
 };


/**
 * A logical plan set for one query block. It is
 * a collection of possible plans, indexed by 
 * their physical properties. Each plan has a PlanID
 * which is local to the RelBlockPlan. It can be used
 * to identify one particular plan.
 */
class RelBlockPlan { 
 public:
  unsigned char BlockId;
  RelBlockPlan* *ChildBlockPlanArray;
  int ChildArrayLen;
  RelCatalogInfo* CatalogInfo;
  //Begin: Moustafa
  // A flag to indicate if the plan has a streaming input
  XxxBool HasStreamInput;

  //End: Moustafa
  virtual void Print(ostream& OStr) const;

  // The first physical plan for this query block.
  // Other plans are linked up thru the Next field.
  RelOptEngine::PlanObjectList PlanList;
  // RelOptEngine::PlanObject* FirstPlan;
  int NumPlans;

  int HighestPlanId; // ever increasing counter of PlanIds

  RelBlockPlan(XxxErrCode& RC,
	       unsigned char BlockIdent,
	       int NumChildren, 
	       RelBlockPlan* Children[],
	       RelCatalogInfo* CatInfo,
	       RelOptEngine::OptQueryEnv* Env);
  virtual ~RelBlockPlan();

  virtual int Arity() const { return ChildArrayLen;}
  RelBlockPlan* GetChild(int i) const { 
    if ( i >= Arity()) return NULL;
    else return ChildBlockPlanArray[i]; }

  // For the top level query block, the RelExecParameters will
  // specify how to choose the appropriate plan.
  virtual XxxErrCode GeneratePlanOp(const RelExecParameters* Params,
				    RelPlanOp*& PlanOp) const;
  // For sub-blocks, the PlanId specifies the desired plan.
  virtual XxxErrCode GeneratePlanOp(int PlanId,
				    RelPlanOp*& PlanOp) const = 0;

  // This method returns the PlanId and Usage of a plan that
  // comes closest to the desired physical properties.
  // Do not make this a const method, because it could lead
  // to the actual generation of fresh plans, especially
  // in the context of selection predicates.
  virtual XxxErrCode 
    GetBestPlan(const RelOptEngine::PhysProperties& DesiredPProps,
		RelOptEngine::PhysProperties*& PlanProps,
		int& PlanId,
		RelOptEngine::Cost& Usage,
		XxxBool& PlanFound);
  // This is similar to the previous method, but uses
  // RelExecParameters instead
  virtual XxxErrCode 
    GetBestPlan(const RelExecParameters* Params,
		int& PlanId,
		RelOptEngine::Cost& Usage,
		XxxBool& PlanFound) const;
  XxxErrCode AddPlan(RelOptEngine::PlanObject* NewPlan) {
    PlanList.Add(NewPlan);
    NumPlans++;
    return XXX_OK;
   }
  // Get the specified plan
  XxxErrCode GetPlan(int PlanId,
		     RelOptEngine::PlanObject*& ChosenPlan) const;

  // This method replaces the dependencies of all correlation
  // variables (i.e XxxUnknownVars) within this plan.
  virtual XxxErrCode RedirectDependencies(const XxxValueExprPlan::VarRemapping& Remap,
					  int CorrelationHt = 0) = 0;
protected:
  // This method replaces the dependencies of all correlation
  // variables (i.e XxxUnknownVars) within this plan.
  XxxErrCode redirectDependencies(const XxxValueExprPlan::VarRemapping& Remap,
				  int CorrelationHt = 0);

 };

/**
//
// encapsulates entire plan: composed of several block plans
//
*/
class RelQueryPlan : public XxxGenericPlan {
  friend class RelOptEngine;
 private:
  PlanUsage  planUsage;
 public:
  RelSpecParameters* OptParams;
  RelBlockPlan* TopBlockPlan;

  RelQueryPlan() { 
    TopBlockPlan = NULL;
    OptParams = NULL;
   }
  RelQueryPlan(RelBlockPlan* TopPlan,
	       const RelSpecParameters* Params) { 
    TopBlockPlan = TopPlan;
    OptParams = new RelSpecParameters(*Params);
    // should be allocated because input arg may be deleted
   }
  //Bjarke, sep. 13 2001: Explicitly initialize base class XxxGenericPlan
  RelQueryPlan(const RelQueryPlan &Plan) : 
    XxxGenericPlan(), OptParams(NULL), TopBlockPlan(NULL) {
    cerr << "Cannot replicate a query plan" << endl;
  }
  virtual XxxGenericPlan * Replicate() const {
    return new RelQueryPlan(*this);
   }
  virtual ~RelQueryPlan() { 
    if (TopBlockPlan) delete TopBlockPlan;
    if (OptParams) delete OptParams;
   }

  virtual XxxErrCode UsageOfPlan(PlanUsage& Usage) const
   { Usage = planUsage; return XXX_OK;}

  void Print(ostream& OStr) const; 
  XxxBool IsValid() const
   { return (TopBlockPlan ? XXX_TRUE : XXX_FALSE); }

  XxxErrCode GenerateExecPlan(const RelExecParameters* Params,
			      RelPlanExecutor*& PlanExec) const;
  XxxErrCode GenerateEvalCursor(const RelExecParameters* Params,
				RelEvalCursor*& EvalCursor) const;

  // This method replaces the dependencies of all correlation
  // variables (i.e XxxUnknownVars) within this plan.
  virtual XxxErrCode RedirectDependencies(const XxxValueExprPlan::VarRemapping& Remap,
					  int CorrelationHt = 0) {
    if (TopBlockPlan) 
      return TopBlockPlan->RedirectDependencies(Remap, CorrelationHt);
    return XXX_OK;
  }
 };


#endif  /* XXX_RELOPT_H */
