#include <iostream.h>

#include "relerror.h"
#include "relquery.h"
#include "relplanop.h"
#include "relenv.h"
#include "seminaivespj.h"

//
// Methods of RelSPJSemiNaiveOpt
//

XxxErrCode RelSPJSemiNaiveOpt::
GeneratePlanOp(RelOptEngine::PlanObject* Plan,
	       RelPlanOp*& PlanOp) const
{
 SPJNaivePlan* SPJPlan = (SPJNaivePlan*)(Plan);
 int NumChildren = SPJPlan->ArraySize;
 int NumJoins = NumChildren-1;
 if (NumJoins == 0) 
   return RelSPJNaiveOptimizer::GeneratePlanOp(Plan, PlanOp);
 
 // this is entered only if there is at least one join
 int i;
 
 // Step 1: Generate plan ops for all the child query blocks.
 RelPlanOp** ChildPlanOps = new RelPlanOp*[NumChildren];
 // InputArity keeps track of the arity of each input relation
 // before join reordering. This could be needed for Remapping
 int *InputArity = new int[NumChildren];
 if (getChildPlans(SPJPlan, ChildPlanOps, InputArity) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error getting child plans");
   return XXX_NOT_OK;
  }
 
 // Generating a pipeline of several SPJPlanOps,
 // one for each binary join.
 // The painful part is creating a cataloginfo
 // for each intermediate result.
 // However, the good news is that the recordschema
 // is easy to determine if projections are not
 // pushed down (it is the concatenation of the
 // input recordschema).
 
 // Step 2: construct catalog info, targetlist, predplans, residplan 
 // and initplan for each join
 RelSPJOptimizer::PredPlanStruct PredPlans;
 XxxValExprPlanList* PlanList = NULL;
 RelCatalogInfo* CatInfo;
 
 // Step 3: this remap is used to alter the various expressions
 XxxExprPlan::VarRemapping Remap(NumChildren);
 if (constructRemap(SPJPlan, Remap, 2, InputArity,
		    XXX_TRUE) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error constructing remap");
   return XXX_NOT_OK;
  }

 PlanOp = ChildPlanOps[0];
 int CurrentArity = // keeps track of arity of partial result
   ChildPlanOps[0]->CatalogInfo->RecordSchema()->Arity();

 // step thru all the joins (remember that all joins are binary)
 for (i = 0; i < NumJoins; i++) {
   // construct the input planops
   RelPlanOp* TmpPlanOps[2];
   TmpPlanOps[0] = PlanOp; // outer relation is the temporary
   TmpPlanOps[1] = ChildPlanOps[i+1];

   // Step 4: get all the predicates to be applied
   PredPlans.InitPlan = NULL;
   // InitPlan makes sense only for the first join
   if (i == 0) {
     if (getInitPlan(SPJPlan, PredPlans.InitPlan) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error in getInitPlan");
       return XXX_NOT_OK;
      }
    } // if (i == 0)

   // determine the PredPlans to be applied
   if (getJoinPlans(SPJPlan, PredPlans, 2, i, XXX_TRUE) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error in getJoinPlans");
     return XXX_NOT_OK;
    }
   
   // Step 5: modify the Remap to reflect the new temporary planop
   if (changeRemap(Remap, NumChildren, 
		   i+1 /* SourceToChange*/, 1 /* New source*/, 
		   0, InputArity) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error changing remap");
     return XXX_NOT_OK;
    }

   // Step 6: modify all the expressions using the VarRemapping
   if (exprRemap(Remap, PredPlans, 2, NULL) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error applying remap");
     return XXX_NOT_OK;
    }

   // Step 7: determine the targetlist and construct the catalog info
   XxxBool ReclaimCat = XXX_FALSE; // should the catinfo be reclaimed?
   if (makeOutputInfo(i, NumJoins, Remap, TmpPlanOps,
		      PlanList, CatInfo, ReclaimCat) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error making output info");
     return XXX_NOT_OK;
    }

   // Step 8: adjust for sort-merge join (note: this "semi-naive"
   // optimizer will try to apply sort-merge wherever possible)
   if (alwaysTrySM && checkSortMerge(NumChildren,
				     PredPlans, Remap, 
				     SPJPlan->JoinMethods[i]) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error checking sort merge");
     return XXX_NOT_OK;
    }
   // Step 9: generate a plan operator for the SPJ.
   // Make sure "distinct" is handled only for the last op
   XxxBool ApplyDistinct = ((i < NumJoins-1)?XXX_FALSE:Distinct);
   if (makePlanOp(TmpPlanOps, 2, PredPlans, PlanList,
		  CatInfo, SPJPlan->JoinMethods[i], 
		  ApplyDistinct, PlanOp) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error making plan operator");
     return XXX_NOT_OK;
    }
   if (ReclaimCat) {
     CatInfo->Destroy();
     delete CatInfo;
    }
   // Step 10: modify the Remap to reflect the new temporary planop
   if (changeRemap(Remap, NumChildren, 
		   1 /* SourceToChange*/, 0 /* New source*/, 
		   CurrentArity, InputArity) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error changing remap");
     return XXX_NOT_OK;
    }

    // important: update this only at the end of the loop
    // becaue it is used to update the Remap
    CurrentArity += 
      ChildPlanOps[i+1]->CatalogInfo->RecordSchema()->Arity();

   // Do not reclaim the pred arrays, because the plan ops use them.
   // Change this eventually : TODO:L2
   // PredPlans.Reclaim();
  } // for NumJoins (main loop creating SPJPlanOps)

 delete [] InputArity;
 delete [] ChildPlanOps;
 
 return XXX_OK;
}

XxxErrCode RelSPJSemiNaiveOpt::
makeOutputInfo(int CurrentJoinPos,
	       int NumJoins,
	       XxxExprPlan::VarRemapping& Remap,
	       RelPlanOp** TmpPlanOps,
	       XxxValExprPlanList*& TmpPlanList,
	       RelCatalogInfo*& TmpCatInfo,
	       XxxBool& ReclaimCat) const
{
 if (CurrentJoinPos == NumJoins-1) {
   // if this is the last join
   TmpCatInfo = BlockPlan->CatalogInfo;
   ReclaimCat = XXX_FALSE;
   TmpPlanList = ExprPlanList;
   // modify the mapping in the targetlist
   if (TmpPlanList &&
       TmpPlanList->RedirectDependencies(Remap) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error redirecting target");
     return XXX_NOT_OK;
    }
  }
 else {
   // if it is an intermediate join
   TmpCatInfo = new RelCatalogInfo();
   // give it a dummy implinfo: needed because SortOrders are attached to it
   TmpCatInfo->implInfo = new RelImplInfo();
   // make sure to reclaim this
   ReclaimCat = XXX_TRUE;

   // create the schema
   const XxxRecordSchema* TmpRecSchema[2];
   TmpRecSchema[0] = TmpPlanOps[0]->CatalogInfo->RecordSchema();
   TmpRecSchema[1] = TmpPlanOps[1]->CatalogInfo->RecordSchema();
   TmpCatInfo->recSchema = new XxxRecordSchema(2,TmpRecSchema);
   
   //  new targetlist is the concatenation of the record schemas
   TmpPlanList = NULL;
   // Two steps: first create an XxxValueExprList, then 
   // optimize it to get an XxxValExprPlanList
   XxxValueExprList *TmpExprList = NULL;
   for (int ci =0; ci < 2; ci++) {
     const XxxRecordSchema* ChildSchema = TmpRecSchema[ci];
     for (int j = 0; j < ChildSchema->Arity(); j++) {
       XxxValueExpression* NewExpr = 
	 new XxxUnknownValue(j, ChildSchema->GetAttr(j)->Type,
			     ChildSchema->GetAttr(j)->TypeCatInfo);
       ((XxxUnknownValue *)NewExpr)->SrcChild = ci;
       if (!TmpExprList)
	 TmpExprList = new XxxValueExprList(NewExpr);
       else 
	 TmpExprList = TmpExprList->AddExpr(NewExpr);
      } // for each attribute
    } // for both children
   // now optimize TmpExprList to create plan list
   if (TmpExprList && 
       (TmpExprList->Optimize(TmpPlanList) != XXX_OK)) {
     REL_NOTE_ERROR(REL_BADOPT, "Error optimizing tmp expr list");
     return XXX_NOT_OK;
    }
  } // else
 return XXX_OK;
}

XxxErrCode RelSPJSemiNaiveOpt::
checkSortMerge(int NumChildren,
	       RelSPJOptimizer::PredPlanStruct& TmpPredPlan,
	       XxxExprPlan::VarRemapping& Remap,
	       RelOptEngine::JoinMethod& JoinMethod) const
{
 XxxBool SMJoinable = XXX_FALSE;
 XxxRelBooleanPlan* TmpJoinPlan = NULL;
 if (isSMJoinable(TmpPredPlan.JoinPredPlans[1], TmpJoinPlan,
		  TmpPredPlan.ResidPredPlans[1], SMJoinable) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error checking for sort-merge");
   return XXX_NOT_OK;
  }

 if (SMJoinable) {
   JoinMethod = RelOptEngine::SORTMERGE;
   TmpPredPlan.JoinPredPlans[1] = TmpJoinPlan;

   // need to further remap the SelPredPlans, because for 
   // sort-merge, these selections get applied independent of
   // a join context. So the "inner" should also be treated
   // as being input #0 (i.e. it looks a bit "bushy")
   // The original Remap should not be changed for this, because it
   // needs to be preserved to accurately remap later expressions
   // that reference the inner. Instead, use a dummy Remap
   XxxExprPlan::VarRemapping DummyRemap(Remap.OldArity);
   DummyRemap.NewArity = Remap.NewArity;
   for (int i = 0; i < Remap.OldArity; i++) {
     DummyRemap.SrcMappingArray[i].NewSrc = Remap.SrcMappingArray[i].NewSrc;
     // dont bother with AttrMappings, since they are not used here
    }
   if (changeRemap(DummyRemap, NumChildren, 
		   1 /* SourceToChange*/, 0 /* New source*/, 
		   0, NULL) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error changing remap");
     return XXX_NOT_OK;
    }
   // modify the selection expressions using the VarRemapping
   if (TmpPredPlan.SelPredPlans[1] &&
       TmpPredPlan.SelPredPlans[1]->RedirectDependencies(DummyRemap) != XXX_OK){
     REL_NOTE_ERROR(REL_BADOPT, "Error redirecting predicate");
     return XXX_NOT_OK;
    }
  }
 return XXX_OK;
}

/**
 * Modify the Remap to reflect a new temporary planop.
 */
XxxErrCode RelSPJSemiNaiveOpt::
changeRemap(XxxExprPlan::VarRemapping& Remap,
	    int NumChildren,
	    int SourceToChange,
	    int NewSourceValue,
	    int AttrIncrement,
	    int* InputArity) const
{
 for (int k=0; k < NumChildren; k++) {
   int SrcNum = Remap.SrcMappingArray[k].NewSrc;
   if (SrcNum == SourceToChange) {
     Remap.SrcMappingArray[k].NewSrc = NewSourceValue;
     // if the change in the join order also affects
     // the positions of attrs for subsequent joins,
     // then modify the remap appropriately
     if (AttrIncrement > 0) {
       for (int j = 0; j < InputArity[k]; j++)
	 Remap.SrcMappingArray[k].NewAttrMapping[j] += AttrIncrement;
      }
    } // if SrcNum == SourceToChange
  } // for k < NumChildren
 return XXX_OK;
}
