#include "ShoreBTree.h"
#include "shorerel.h"
#include "shoresm.h"
#include "relexec.h"

// Cursor used to scan a SHORE relation with a B+ index.

ShoreBTree::
IScanCursorInfo::IScanCursorInfo(const XxxShoreLogicalOID& RelId,
				 const XxxShoreObjectId& IndexId,
				 XxxRecordSchema* Schema,
				 XxxBooleanExpressionPlan* Expr,
				 XxxBooleanExpressionPlan* RExpr, // residual
				 XxxBool Exact,
				 const BTreeRange* RangeSpec,
				 const RelExecParameters* Parameters,
				 const XxxRecord** SibRecs,
				 const XxxRecordSchema** SibSchema,
				 int CurrJoinPos)
:  CursorInfo(Schema, Expr, RExpr, Exact, Parameters, 
	      SibRecs, SibSchema, CurrJoinPos), 
	      ScanIter(RelId.VolId, IndexId, 
		       RangeSpec->LowerCmp, 
		       (RangeSpec->LowerBoundSize 
			? RangeSpec->LowerBound 
			: cvec_t::neg_inf),
		       RangeSpec->UpperCmp, 
		       (RangeSpec->UpperBoundSize 
			? RangeSpec->UpperBound 
			: cvec_t::pos_inf))
    
{
 VolId = RelId.VolId;
 RSpec = RangeSpec;
}

ShoreBTree::
IScanCursorInfo::~IScanCursorInfo()
{
 // reclaim the RangeSpec
 if (RSpec) delete RSpec;
}

XxxErrCode ShoreBTree::
IScanCursorInfo::NextRecord(XxxShoreObjectId*& ObjId, 
			    XxxRecord*& ReturnRec, XxxBool& Success)
{
 Success = XXX_FALSE;
 XxxErrCode rc = XXX_OK;

 if (!RSpec->Satisfiable) {
   return XXX_OK;
  }

 bool AtEnd;
 w_rc_t RetVal;

 XXX_CALL((XXX_SHORE_RETFN(ScanIter.next(AtEnd), RetVal)),
	  XXX_SHOREERR, "Couldnt scan index", rc);
 XXX_BADRC_RCRETURN(rc);

 if (AtEnd) {
   return XXX_OK;
  }

 // TODO:L2 here, we should know the key type so that we can use a buffer of
 // the max-size of the key's type
 XxxShoreObjectId RecordId;
 XxxShoreKey DataElem(&RecordId, sizeof(XxxShoreObjectId));
 // 1000 is an upper bound that we hope is safe : ugh!
 char KeyBuf[1000];
 XxxShoreKey KeyElem(&KeyBuf, 1000);
 size_type KeyLen, DataLen;

 XXX_CALL((XXX_SHORE_RETFN(ScanIter.curr(&KeyElem, KeyLen, 
					 &DataElem, DataLen), 
			   RetVal)),
	  XXX_SHOREERR, "Couldnt get curr element from index", rc);
 XXX_BADRC_RCRETURN(rc);

 ASSERT((DataLen == sizeof(XxxShoreObjectId)));

 // for now, we are allocating a new record
 if (XxxShoreSM->GetAllocateRecord(VolId, RecordId, 
				   ReturnRec) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_SHOREERR, "Error getting indexed record");
   return XXX_NOT_OK;
  }

 ObjId = new XxxShoreObjectId(RecordId);
 Success = XXX_TRUE;
 return XXX_OK;
}

//
// BTreeRange methods.
//
void ShoreBTree:: 
BTreeRange::CheckSatisfiability()
{
 // there are a couple of strange cases where the RangeSpec
 // can obviously be false (eg. = 2 and = 1).
 // so check for satisfiability of RangeSpec (NP-what?)
 // Only need to check this the first time though:
 // ignore pos_inf and neg_inf
 if (!LowerBoundSize || !UpperBoundSize) {
   Satisfiable = XXX_TRUE;
  }
 // otherwise, both are real bounds
 else {
#if defined(XXX_WIN32)
   // The >,< operators of Shore cvec_t are buggy on NT. so can't do this.
   // TODO:L2 get this fixed in Shore
#else
   //Bjarke: comparison used in the wrong way
   //     if (LowerBound > UpperBound) {
   //       Satisfiable = XXX_FALSE;
   //      }
   //     else if ((LowerCmp == scan_index_i::eq) &&
   //  	    (UpperCmp == scan_index_i::eq)) {
   //       if (UpperBound > LowerBound) 
   //         Satisfiable = XXX_FALSE;
   //      }
   //     else 
#endif
     Satisfiable = XXX_TRUE;
  } // if both are real bounds
}

void ShoreBTree:: 
BTreeRange::Copy(const XxxIDXClass::IndexRange& R)
{
  // TODO: check that this works with the new XxxADTValue regime
 if (this == &R)
   return;

 BTreeRange& BR = (BTreeRange&) R;

 DataLen = BR.DataLen;
 Data = new char[DataLen];
 memcpy(Data, BR.Data, DataLen);

 LowerCmp = BR.LowerCmp;
 LowerBoundOffset = BR.LowerBoundOffset;
 LowerBoundSize = BR.LowerBoundSize;
 LowerBound.set(BR.LowerBound);
 if (!LowerBoundOffset && !LowerBoundSize) // if a constant like pos_inf
   LowerBound.set(BR.LowerBound);
 else 
   LowerBound.set((void *)(Data+LowerBoundOffset), LowerBoundSize);

 UpperCmp = BR.UpperCmp;
 UpperBoundOffset = BR.UpperBoundOffset;
 UpperBoundSize = BR.UpperBoundSize;
 if (!UpperBoundOffset && !UpperBoundSize) // if a constant like pos_inf
   UpperBound.set(BR.UpperBound);
 else 
   UpperBound.set((void *)(Data+UpperBoundOffset), UpperBoundSize);
 Satisfiable = BR.Satisfiable;
}


// ShoreBTree methods.

XxxErrCode
ShoreBTree::Print(const XxxShoreVolumeId& VolumeId,
		  const XxxShoreObjectId& IndexId) 
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->print_index(VolumeId, IndexId),
			   RetVal)),
	  XXX_SHOREERR, "Cannot print B+ index", rc);
 XXX_BADRC_RCRETURN(rc);
 
 return rc;
}

XxxErrCode 
ShoreBTree::Create(const XxxShoreVolumeId& VolumeId,
		   XxxShoreObjectId& Index,
		   XxxValueExpression *Expr,
		   const XxxBool Unique)
{
 XxxBasicType IndexColType;
//Bjarke, aug. 8: IndexString changed from char* to char[5] and strcpy
//used to assign a value to it
 char IndexString[5];
 
 // check if index is built on an expression of valid type
 // Expr MUST be typed checked before!

 IndexColType = Expr->BasicType();
 switch (IndexColType)
  {
  case XXX_INT:
   strcpy(IndexString, "i4");
   break;
  case XXX_DOUBLE:
   strcpy(IndexString, "f8");
   break;
  case XXX_STRING:
   strcpy(IndexString, "b*20");
   break;
  default:
   return XXX_NOT_OK;
  } 

 // SHORE call to create B+ index

 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 
 ss_m::ndx_t IndexType = 
   ((Unique == XXX_TRUE)?ss_m::t_uni_btree : ss_m::t_btree);
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->create_index(VolumeId, 
							    IndexType,
							    ss_m::t_regular,
							    IndexString,
							    0,
							    Index),
			   RetVal)),
	  XXX_SHOREERR, "Unable to create index", rc);
 XXX_BADRC_RCRETURN(rc);
 return rc;

#ifdef XXX_DEBUG
 // Print(VolumeId,Index);
#endif
 
}

XxxErrCode 
ShoreBTree::Create(const XxxShoreVolumeId& VolumeId,
		   XxxShoreObjectId& Index,
		   XxxBasicType IndexColType,
		   const XxxBool Unique)
{
//Bjarke, aug. 8: IndexString changed from char* to char[5] and strcpy
//used to assign a value to it
 char IndexString[5];

 switch (IndexColType) {
  case XXX_INT:
   strcpy(IndexString, "i4");
   break;
  case XXX_DOUBLE:
   strcpy(IndexString, "f8");
   break;
  case XXX_STRING:
   strcpy(IndexString, "b*20");
   break;
  default:
   return XXX_NOT_OK;
  } 
 
 // SHORE call to create B+ index
 
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 
 ss_m::ndx_t IndexType = 
   ((Unique == XXX_TRUE)?ss_m::t_uni_btree : ss_m::t_btree);
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->create_index(VolumeId, 
							    IndexType,
							    ss_m::t_regular,
							    IndexString,
							    0,
							    Index),
			   RetVal)),
	  XXX_SHOREERR, "Unable to create index", rc);
 XXX_BADRC_RCRETURN(rc);
 return rc;
 
#ifdef XXX_DEBUG
 // Print(VolumeId,Index);
#endif

}



XxxErrCode 
ShoreBTree::Delete(const XxxShoreVolumeId& VolumeId,
		   XxxShoreObjectId& Index)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->destroy_index(VolumeId, Index),
			   RetVal)),
	  XXX_SHOREERR, "Couldnt destroy B+ index", rc);
 XXX_BADRC_RCRETURN(rc);
 return rc;
}

// Record has already been evaluated into RecValue

XxxErrCode 
ShoreBTree::InsertEntry(const XxxShoreVolumeId& VolumeId,
			const XxxShoreObjectId& Index,
			const XxxADTValue* RecValue,
			const XxxShoreObjectId* RecId,
			XxxValueExprPlan *ValuePlan)
{
 XxxBasicType IndxColType = ValuePlan->BasicType();

 ASSERT(XxxCoreADTClass::IsCoreADT(IndxColType));
 int DatumSize = 
   XxxCoreADTClass::STypeSize(IndxColType, RecValue, ValuePlan->GenerateCatInfo())
   - XxxCoreADTClass::ValueOffset(IndxColType);
 
 XxxShoreKey Key((void *)(((char*)RecValue) + 
                          XxxCoreADTClass::ValueOffset(IndxColType)), DatumSize);
 
 // SHORE call

 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 XxxShoreKey Elem((void *) RecId, sizeof(XxxShoreObjectId));
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->create_assoc(VolumeId,
							    Index, 
							    Key, 
							    Elem), 
			   RetVal)),
	  XXX_SHOREERR, "Cannot make entry in index", rc);
 XXX_BADRC_RCRETURN(rc);
 
 return rc;
}

XxxErrCode 
ShoreBTree::InsertEntry(const XxxShoreVolumeId& VolumeId,
			const XxxShoreObjectId& Index,
			const XxxADTValue* RecValue,
			int RecSize,
			XxxBasicType IndxColType,
			const XxxShoreObjectId* RecId)
{
 XxxShoreKey Key((void *)(((char*)RecValue)+XxxCoreADTClass::ValueOffset(IndxColType)),
		                  RecSize - XxxCoreADTClass::ValueOffset(IndxColType));
 
 // SHORE call

 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 XxxShoreKey Elem((void *) RecId, sizeof(XxxShoreObjectId));
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->create_assoc(VolumeId,
							    Index, 
							    Key, 
							    Elem), 
			   RetVal)),
	  XXX_SHOREERR, "Cannot enter file into index", rc);
 XXX_BADRC_RCRETURN(rc);
 
 return rc;
}

XxxErrCode 
ShoreBTree::DeleteAllEntries(const XxxShoreVolumeId& VolumeId,
			     const XxxShoreObjectId& Index,	
			     const XxxADTValue* RecValue,
			     int RecSize,
			     XxxBasicType IndxColType,
			     int& NumDeleted)
{
 XxxShoreKey Key((void *)(((char*)RecValue)+XxxCoreADTClass::ValueOffset(IndxColType)),
		 RecSize-XxxCoreADTClass::ValueOffset(IndxColType));
 
 XxxErrCode rc  = XXX_OK;
 w_rc_t RetVal;
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->destroy_all_assoc(VolumeId,
								 Index, 
								 Key,
								 NumDeleted), 
			   RetVal)),
	  XXX_SHOREERR, "Cannot delete entry in index", rc);
 XXX_BADRC_RCRETURN(rc);

 return rc;
}

XxxErrCode 
ShoreBTree::DeleteOneEntry(const XxxShoreVolumeId& VolumeId,
			   const XxxShoreObjectId& Index,
			   const XxxADTValue* RecValue,
			   const XxxShoreObjectId* RecId,
			   XxxValueExprPlan *ValuePlan) 
{
 XxxBasicType IndxColType = ValuePlan->BasicType();

 ASSERT(XxxCoreADTClass::IsCoreADT(IndxColType));
 int DatumSize = 
   XxxCoreADTClass::STypeSize(IndxColType, RecValue,ValuePlan->GenerateCatInfo()) 
   - XxxCoreADTClass::ValueOffset(IndxColType);
 
 XxxShoreKey Key((void *)(((char*)RecValue)+XxxCoreADTClass::ValueOffset(IndxColType)),DatumSize);
 
 XxxErrCode rc  = XXX_OK;
 w_rc_t RetVal;
 XxxShoreKey Elem((void *) RecId, sizeof(XxxShoreObjectId));
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->destroy_assoc(VolumeId,
							     Index, 
							     Key,
							     Elem), 
			   RetVal)),
	  XXX_SHOREERR, "Cannot delete entry in index", rc);
 XXX_BADRC_RCRETURN(rc);
 
 return rc;
}

XxxErrCode 
ShoreBTree::DeleteOneEntry(const XxxShoreVolumeId& VolumeId,
			   const XxxShoreObjectId& Index,
			   const XxxADTValue* RecValue,
			   int RecSize,
			   XxxBasicType IndxColType,
			   const XxxShoreObjectId* RecId)
{
 XxxShoreKey Key((void *)(((char*)RecValue)+XxxCoreADTClass::ValueOffset(IndxColType)),
		 RecSize-XxxCoreADTClass::ValueOffset(IndxColType));
 
 XxxErrCode rc  = XXX_OK;
 w_rc_t RetVal;
 XxxShoreKey Elem((void *) RecId, sizeof(XxxShoreObjectId));
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->destroy_assoc(VolumeId,
							     Index, 
							     Key,
							     Elem), 
			   RetVal)),
	  XXX_SHOREERR, "Cannot delete entry in index", rc);
 XXX_BADRC_RCRETURN(rc);
 
 return rc;
}


XxxErrCode 
ShoreBTree::GetEntry(const XxxShoreVolumeId& VolumeId,
		     const XxxShoreObjectId& Index,
		     const XxxADTValue* RecValue,
		     XxxValueExprPlan *ValuePlan,
		     void *DataBuf,
		     int DataBufSize,
		     bool& Found)
{
 XxxBasicType IndxColType = ValuePlan->BasicType();
 
 ASSERT(XxxCoreADTClass::IsCoreADT(IndxColType));
 int DatumSize = 
   XxxCoreADTClass::STypeSize(IndxColType, RecValue, ValuePlan->GenerateCatInfo())
   - XxxCoreADTClass::ValueOffset(IndxColType);
 
 XxxShoreKey Key((void *)(((char*)RecValue)+XxxCoreADTClass::ValueOffset(IndxColType)),DatumSize);
 
 XxxErrCode rc  = XXX_OK;
 w_rc_t RetVal;
 smsize_t DBufSize = DataBufSize;
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->find_assoc(VolumeId, 
							  Index, 
							  Key,
							  DataBuf, 
							  DBufSize,
							  Found), RetVal)),
	  XXX_SHOREERR, "Cannot find entry in index", rc);
 XXX_BADRC_RCRETURN(rc);
 DataBufSize = DBufSize;
 // TODO:L1 is a memcpy needed here for alignment bugs??
 
 return rc;
}

XxxErrCode 
ShoreBTree::GetEntry(const XxxShoreVolumeId& VolumeId,
		     const XxxShoreObjectId& Index,
		     const XxxADTValue* RecValue,
		     int RecSize,
		     XxxBasicType IndxColType,
		     void *DataBuf,
		     int DataBufSize,
		     bool& Found)
{
 XxxShoreKey Key((void *)(((char*)RecValue)+XxxCoreADTClass::ValueOffset(IndxColType)),
		 RecSize-XxxCoreADTClass::ValueOffset(IndxColType));
 
 XxxErrCode rc  = XXX_OK;
 w_rc_t RetVal;
 smsize_t DBufSize = DataBufSize;
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->find_assoc(VolumeId, 
							  Index, 
							  Key,
							  DataBuf, 
							  DBufSize,
							  Found), RetVal)),
	  XXX_SHOREERR, "Cannot find entry in index", rc);
 XXX_BADRC_RCRETURN(rc);
 DataBufSize = DBufSize;
 // TODO:L1 is a memcpy needed here for alignment bugs??
 
 return rc;
}

//AUTHOR: Fabian Camargo DATE: 3/24/97
//PURPOSE: Obtains the conjunction of two ranges
//NOTES: This is used when then query involves an AND.
XxxErrCode 
ShoreBTree::MergeRanges(const XxxIDXClass::IndexRange* R1, 
			const XxxIDXClass::IndexRange* R2,
			XxxIDXClass::IndexRange* R3) const
{
 // cast Ranges to BTree ranges
 const BTreeRange *Range1 = (const BTreeRange *)R1;
 const BTreeRange *Range2 = (const BTreeRange *)R2;
 BTreeRange *MergedRange = (BTreeRange *)R3;
 
 int R1DataLen = Range1->DataLen;
 int R2DataLen = Range2->DataLen;
 
 char *DataPtr;
 DataPtr = new char[R1DataLen+R2DataLen];
 memcpy(DataPtr,Range1->Data,R1DataLen);
 memcpy(DataPtr+R1DataLen,Range2->Data,R2DataLen);

 // the resulting lower bound is the higher of the two bounds
 int BoundOffset, BoundSize;
 XxxBool ChooseRng1;
 if (!Range1->LowerBoundSize) // neg_inf
   ChooseRng1 = XXX_FALSE;
 else if (!Range2->LowerBoundSize) // neg_inf
   ChooseRng1 = XXX_TRUE;
 // The >,< operators of Shore cvec_t are buggy on NT. i think!
 // TODO:L2 are we sure this works?? see CheckSatisfiability() ...
 else if (Range1->LowerBound < Range2->LowerBound)
   ChooseRng1 = XXX_FALSE;
 else ChooseRng1 = XXX_TRUE;
 
 if (!ChooseRng1) {
   BoundOffset = R1DataLen + Range2->LowerBoundOffset;
   BoundSize = Range2->LowerBoundSize;
   if (!BoundOffset && !BoundSize)// if it is a constant like pos_inf
     MergedRange->LowerBound.set(Range2->LowerBound);
   else 
     MergedRange->LowerBound.set((void *)(DataPtr+BoundOffset), BoundSize);
   MergedRange->LowerCmp=Range2->LowerCmp;
  }
 else {
   BoundOffset = Range1->LowerBoundOffset;
   BoundSize = Range1->LowerBoundSize;
   if (!BoundOffset && !BoundSize) // if it is a constant like pos_inf
     MergedRange->LowerBound.set(Range1->LowerBound);
   else 
     MergedRange->LowerBound.set((void *)(DataPtr+BoundOffset), BoundSize);
   MergedRange->LowerCmp = Range1->LowerCmp;
  }
 MergedRange->LowerBoundOffset = BoundOffset;
 MergedRange->LowerBoundSize = BoundSize;

 // the resulting upper bound is the lower of the two bounds
 if (!Range1->UpperBoundSize) // pos_inf
   ChooseRng1 = XXX_FALSE;
 else if (!Range2->UpperBoundSize) // pos_inf
   ChooseRng1 = XXX_TRUE;
 // The >,< operators of Shore cvec_t are buggy on NT. i think!
 // TODO:L2 are we sure this works?? see CheckSatisfiability ...
 else if (Range1->UpperBound < Range2->UpperBound)
   ChooseRng1 = XXX_TRUE;
 else ChooseRng1 = XXX_FALSE;

 if (ChooseRng1) {
   BoundOffset = Range1->UpperBoundOffset;
   BoundSize = Range1->UpperBoundSize;
   if (!BoundOffset && !BoundSize) // if it is a constant like pos_inf
     MergedRange->UpperBound.set(Range1->UpperBound);
   else 
     MergedRange->UpperBound.set((void *)(DataPtr+BoundOffset), BoundSize);
   MergedRange->UpperCmp=Range1->UpperCmp;
  }
 else {
   BoundOffset = R1DataLen + Range2->UpperBoundOffset;
   BoundSize = Range2->UpperBoundSize;
   if (!BoundOffset && !BoundSize) // if it is a constant like pos_inf
     MergedRange->UpperBound.set(Range2->UpperBound);
   else 
     MergedRange->UpperBound.set((void *)(DataPtr+BoundOffset), BoundSize);
   MergedRange->UpperCmp=Range2->UpperCmp;
  }
 MergedRange->UpperBoundOffset = BoundOffset;
 MergedRange->UpperBoundSize = BoundSize;


 // Do this only at the end : this function could be called as follows:
 // MergeRanges(R1, R2, R1) i.e the result is one the inputs too.
 // So do not modify the result till the very end. This is a little dicey.
 MergedRange->Data=DataPtr;
 MergedRange->DataLen = R1DataLen+R2DataLen;

 // of course, if the two ranges are non-overlapping, we come up
 // with a range where the lower bound is higher than the upper bound.
 // this should be fine, and result in no answers.
 
 MergedRange->CheckSatisfiability();

 return XXX_OK;
}

XxxErrCode
ShoreBTree::getRangeSpec(XxxExprPlan* LeftPlan,
			 XxxExprPlan* RightPlan,
			 XxxRelOpType OpType,
			 const XxxRecord** SibRecs,
			 const XxxRecordSchema** SibSchema,
			 const RelExecParameters* Parameters,
			 XxxIDXClass::IndexRange* Range,
			 const XxxIDXClass::MatchInfo* MatchedInfo) const
{
 // Cast range to b+ tree range.
 BTreeRange* RangeSpec = (BTreeRange *)Range;
 
 if (!MatchedInfo) {
   REL_NOTE_ERROR(REL_BADINDEX, "Empty match info");
   return XXX_NOT_OK;
  }
 if (MatchedInfo->Type != XxxIDXClass::MatchInfo::REL) {
   REL_NOTE_ERROR(REL_BADINDEX, "Bad type for match info");
   return XXX_NOT_OK;
  }
 const XxxIDXClass::RelMatchInfo* RMInfo = 
   (const XxxIDXClass::RelMatchInfo*)MatchedInfo;
 
 // B+ trees get their ranges from the constant part of the expression.
 // i.e expr = x > 5 then we get range from '>' and 5.
 XxxExprPlan* ExprPlan;
 if (RMInfo->LeftConstant)
   ExprPlan = LeftPlan;
 else
   ExprPlan = RightPlan;

 // These variables need to persist beyond this
 // function invocation. therefore, use dynamic allocation.
 // Use the destructor of ShoreIndexRangeSpec to reclaim memory.
 
 XxxValueExprPlan* ConstExprPlan = (XxxValueExprPlan*)(ExprPlan);
 XxxBasicType TypeId = ConstExprPlan->BasicType();
 ASSERT(XxxCoreADTClass::IsCoreADT(TypeId));
 XxxADTClass* TypeClass = XxxADTArray[TypeId];
 RangeSpec->Data = 
   new char[XxxCoreADTClass::
	   SMaxObjectSize(TypeId, ConstExprPlan->GenerateCatInfo())];
 
 // evaluate the expression
 XXX_CALL_RET(ExprPlan->Evaluate((const XxxRecord **)SibRecs, 
				 (const XxxRecordSchema **)SibSchema, 
				 (XxxADTValue*)RangeSpec->Data, 
				 (Parameters?Parameters->Params:0)),
	      XXX_BADEXPR, "Unable to evaluate const expr");
 
 RangeSpec->DataLen = XxxCoreADTClass::
   STypeSize(TypeId, (XxxADTValue *)RangeSpec->Data, 
	     ConstExprPlan->GenerateCatInfo());
 
 if (RMInfo->LeftConstant == XXX_TRUE) {
   // switch it around so that you imagine that the right side has the const
   // so it is of the form A <op> <const>
   switch(OpType) {
    case XXX_GT_OP:
     OpType = XXX_LT_OP;
     break;
    case XXX_LT_OP:
     OpType = XXX_GT_OP;
     break;
    case XXX_LEQ_OP:
     OpType = XXX_GEQ_OP;
     break;
    case XXX_GEQ_OP:
     OpType = XXX_LEQ_OP;
     break;
    case XXX_EQ_OP:
     // leave this untouched
     break;
    default:
     // should not be here
     XXX_NOTE_ERROR(REL_BADINDEX, "bad Op type for Btree index");
     return XXX_NOT_OK;
    }
  }
 // TODO:L2 Get rid of hacks below in next release
 int HdrSize = XxxCoreADTClass::ValueOffset(TypeId);
 switch (OpType) {
  case XXX_GT_OP:
   RangeSpec->LowerCmp = scan_index_i::gt;
   RangeSpec->LowerBound.set((void *)(RangeSpec->Data+HdrSize), RangeSpec->DataLen-HdrSize);
   RangeSpec->LowerBoundOffset = HdrSize;
   RangeSpec->LowerBoundSize = RangeSpec->DataLen-HdrSize;
   RangeSpec->UpperCmp = scan_index_i::le;
   // Note: Setting this is useless : Shore compares the actual pointers!
   RangeSpec->UpperBound.set(vec_t::pos_inf);
   RangeSpec->UpperBoundOffset = 0; // hack to specify that it is a constant
   RangeSpec->UpperBoundSize = 0;
   break;
  case XXX_GEQ_OP:
   RangeSpec->LowerCmp = scan_index_i::ge;
   RangeSpec->LowerBound.set((void *)(RangeSpec->Data+HdrSize), RangeSpec->DataLen-HdrSize);
   RangeSpec->LowerBoundOffset = HdrSize;
   RangeSpec->LowerBoundSize = RangeSpec->DataLen-HdrSize;
   RangeSpec->UpperCmp = scan_index_i::le;
   // Note: Setting this is useless : Shore compares the actual pointers!
   RangeSpec->UpperBound.set(vec_t::pos_inf);
   RangeSpec->UpperBoundOffset = 0; // specify that it is a constant
   RangeSpec->UpperBoundSize = 0;
   break;
  case XXX_LT_OP:
   RangeSpec->LowerCmp = scan_index_i::ge;
   // Note: Setting this is useless : Shore compares the actual pointers!
   RangeSpec->LowerBound.set(vec_t::neg_inf);
   RangeSpec->LowerBoundOffset = 0; // specify that it is a constant
   RangeSpec->LowerBoundSize = 0;
   RangeSpec->UpperCmp = scan_index_i::lt;
   RangeSpec->UpperBound.set((void *)(RangeSpec->Data+HdrSize), RangeSpec->DataLen-HdrSize);
   RangeSpec->UpperBoundOffset = HdrSize;
   RangeSpec->UpperBoundSize = RangeSpec->DataLen-HdrSize;
   break;
  case XXX_LEQ_OP:
   RangeSpec->LowerCmp = scan_index_i::ge;
   // Note: Setting this is useless : Shore compares the actual pointers!
   RangeSpec->LowerBound.set(vec_t::neg_inf);
   RangeSpec->LowerBoundOffset = 0; // specify that it is a constant
   RangeSpec->LowerBoundSize = 0;
   RangeSpec->UpperCmp = scan_index_i::le;
   RangeSpec->UpperBound.set((void *)(RangeSpec->Data+HdrSize), RangeSpec->DataLen-HdrSize);
   RangeSpec->UpperBoundOffset = HdrSize;
   RangeSpec->UpperBoundSize = RangeSpec->DataLen-HdrSize;
   break;
  case XXX_EQ_OP:
  case XXX_LIKE_OP:
   RangeSpec->LowerCmp = scan_index_i::eq;
   RangeSpec->LowerBound.set((void *)(RangeSpec->Data+HdrSize), RangeSpec->DataLen-HdrSize);
   RangeSpec->LowerBoundOffset = HdrSize;
   RangeSpec->LowerBoundSize = RangeSpec->DataLen-HdrSize;
   RangeSpec->UpperCmp = scan_index_i::eq;
   RangeSpec->UpperBound.set((void *)(RangeSpec->Data+HdrSize), RangeSpec->DataLen-HdrSize);
   RangeSpec->UpperBoundOffset = HdrSize;
   RangeSpec->UpperBoundSize = RangeSpec->DataLen-HdrSize;
   break;
  default:
   REL_NOTE_ERROR(REL_BADINDEX, "Not valid Op type for BTree index");
   return XXX_NOT_OK;
  }
 RangeSpec->CheckSatisfiability();
 return XXX_OK;
}

XxxIDXClass::CursorInfo* 
ShoreBTree::GetCursor(const XxxShoreLogicalOID& RelId,
		      const XxxShoreObjectId& IndexId,
		      XxxRecordSchema* Schema,
		      XxxBooleanExpressionPlan* Expr,
		      XxxBooleanExpressionPlan* RExpr, // residual
		      XxxBool Exact,
		      const XxxIDXClass::IndexRange* RangeSpec,
		      const RelExecParameters* Parameters,
		      const XxxRecord** SibRecs,
		      const XxxRecordSchema** SibSchema,
		      int CurrentJoinPos)
{
 return new IScanCursorInfo(RelId,
			    IndexId,
			    Schema,
			    Expr,
			    RExpr,
			    Exact,
			    (BTreeRange *)RangeSpec,
			    Parameters,
			    SibRecs,
			    SibSchema,
			    CurrentJoinPos);
}
