#include "ShoreRTree.h"
#include "shorerel.h"
#include "funcexpr.h"

// cursor used to scan a shore relation with an R* tree

ShoreRTree::
RIScanCursorInfo::RIScanCursorInfo(const XxxShoreLogicalOID& RelId,
				   const XxxShoreObjectId& IndexId, 
				   XxxRecordSchema* Schema,
				   XxxBooleanExpressionPlan* Expr,
				   XxxBooleanExpressionPlan* RExpr, // residual
				   XxxBool Exact,
				   const ShoreRTree::RTreeRange* RangeSpec,
				   const RelExecParameters* Parameters,
				   const XxxRecord** SibRecs,
				   const XxxRecordSchema** SibSchema,
				   int CurrJoinPos)
: CursorInfo(Schema, Expr, RExpr, Exact, Parameters, SibRecs, 
	     SibSchema, CurrJoinPos), 
	     ScanIter(RelId.VolId, IndexId, 
		      RangeSpec->Cmp, 
		      RangeSpec->box.ShoreBox()),
		      box(RangeSpec->box)
{
 VolId = RelId.VolId;
 RSpec = RangeSpec;
}

// close the rtree index scan on the SHORE file
ShoreRTree::
RIScanCursorInfo::~RIScanCursorInfo()
{
 // reclaim the Range Spec
 if (RSpec) delete RSpec;
}

XxxErrCode ShoreRTree::
RIScanCursorInfo::NextRecord(XxxShoreObjectId*& ObjId, 
			     XxxRecord*& ReturnRec, XxxBool& Success)
{
 Success = XXX_FALSE;
 XxxErrCode rc = XXX_OK;

 bool AtEnd;
 w_rc_t RetVal;

 XxxShoreObjectId RecordId;
 nbox_t KeyElem;
 long unsigned int DataLen=sizeof(XxxShoreObjectId);

 XXX_CALL((XXX_SHORE_RETFN(ScanIter.next(KeyElem,&RecordId,DataLen,
					 AtEnd), 
			   RetVal)),
	  XXX_SHOREERR, "Couldnt get next element from index", rc);
 XXX_BADRC_RCRETURN(rc);

 if (AtEnd) {
   return XXX_OK;
  }

 ASSERT((DataLen == sizeof(XxxShoreObjectId)));

 // for now, we are allocating a new record

 if (XxxShoreSM->GetAllocateRecord(VolId, RecordId, 
				   ReturnRec) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_SHOREERR, "Error getting indexed record");
   return XXX_NOT_OK;
  }

 ObjId = new XxxShoreObjectId(RecordId);
 Success = XXX_TRUE;
 return XXX_OK;
}

// RTreeRange methods.

void ShoreRTree:: 
RTreeRange::Copy(const XxxIDXClass::IndexRange& R)
{
 if (this == &R)
   return;

 RTreeRange& RR = (RTreeRange&) R;

 Cmp=RR.Cmp;
 box=RR.box;
}


// ShoreRTree methods.

XxxErrCode 
ShoreRTree::Print(const XxxShoreVolumeId& VolumeId,
		  const XxxShoreObjectId& IndexId)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->print_md_index(VolumeId,IndexId),
			   RetVal)),
	  XXX_SHOREERR, "Cannot print rtree index", rc);
 XXX_BADRC_RCRETURN(rc);
 
 return rc;
}

XxxErrCode 
ShoreRTree::Create(const XxxShoreVolumeId& VolumeId,
		   XxxShoreObjectId& Index,
		   XxxValueExpression *Expr,
		   const XxxBool Unique)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 
 ss_m::ndx_t IndexType = ss_m::t_rtree;

 // Right now the dimension is hard coded to 2. Once SHORE is able
 // to support more dimensions we should be able to extract the number
 // of dimensions from an ADT in the Expr.

 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->create_md_index(VolumeId, 
							       IndexType,
							       ss_m::t_regular,
							       Index,
							       2),
			   RetVal)),
	  XXX_SHOREERR, "Unable to create r-tree index", rc);
 XXX_BADRC_RCRETURN(rc);
 return rc;

#ifdef XXX_DEBUG
 // Print(VolumeId,Index);
#endif
 
}

XxxErrCode 
ShoreRTree::Create(const XxxShoreVolumeId& VolumeId,
		   XxxShoreObjectId& Index,
		   XxxBasicType IndxColType,
		   const XxxBool Unique)
{
 return Create(VolumeId,Index,NULL,XXX_TRUE);
}


XxxErrCode 
ShoreRTree::Delete(const XxxShoreVolumeId& VolumeId,
		   XxxShoreObjectId& Index)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->destroy_md_index(VolumeId,Index),
			   RetVal)),
	  XXX_SHOREERR, "Couldnt destroy R* index", rc);
 XXX_BADRC_RCRETURN(rc);
 return rc;
}

XxxErrCode 
ShoreRTree::InsertEntry(const XxxShoreVolumeId& VolumeId,
			const XxxShoreObjectId& Index,
			const XxxADTValue* RecValue,
			const XxxShoreObjectId* RecId,
			XxxValueExprPlan *ValuePlan)
{
 XxxBasicType IndxColType = ValuePlan->BasicType();
 return InsertEntry(VolumeId, Index, RecValue,
		    0, IndxColType, RecId);
}

XxxErrCode 
ShoreRTree::InsertEntry(const XxxShoreVolumeId& VolumeId,
			const XxxShoreObjectId& Index,
			const XxxADTValue* RecValue,
			int RecSize,
			XxxBasicType IndxColType,
			const XxxShoreObjectId* RecId)
{ 
 XxxBoundingBox Box;
 if (XxxADTArray[IndxColType]->GetBoundingBox(RecValue,Box)==XXX_FALSE) {
   return XXX_NOT_OK;
  }
 
 nbox_t ShoreKey= Box.ShoreBox();
 
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 XxxShoreKey KeyElem((void *) (RecId), sizeof(XxxShoreObjectId));
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->create_md_assoc(VolumeId, Index, 
							       ShoreKey, 
							       KeyElem), 
			   RetVal)),
	  XXX_SHOREERR, "Cannot enter file into rtree index", rc);
 XXX_BADRC_RCRETURN(rc);
 
 return rc;
}


XxxErrCode 
ShoreRTree::DeleteOneEntry(const XxxShoreVolumeId& VolumeId,
			   const XxxShoreObjectId& Index,
			   const XxxADTValue* RecValue,
			   const XxxShoreObjectId* RecId,
			   XxxValueExprPlan *ValuePlan) 
{
 XxxBasicType IndxColType = ValuePlan->BasicType();
 
 XxxBoundingBox Box;
 if (XxxADTArray[IndxColType]->GetBoundingBox(RecValue,Box)==XXX_FALSE) {
   return XXX_NOT_OK;
  }
 
 nbox_t ShoreKey= Box.ShoreBox();
 
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 XxxShoreKey KeyElem((void *) (RecId), sizeof(XxxShoreObjectId));
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->destroy_md_assoc(VolumeId, Index, 
								ShoreKey, 
								KeyElem), 
			   RetVal)),
	  XXX_SHOREERR, "Cannot enter file into rtree index", rc);
 XXX_BADRC_RCRETURN(rc);
 
 return rc;
}

XxxErrCode 
ShoreRTree::DeleteOneEntry(const XxxShoreVolumeId& VolumeId,
			   const XxxShoreObjectId& Index,
			   const XxxADTValue* RecValue,
			   int RecSize,
			   XxxBasicType IndxColType,
			   const XxxShoreObjectId* RecId)
{
 XxxBoundingBox Box;
 if (XxxADTArray[IndxColType]->GetBoundingBox(RecValue,Box)==XXX_FALSE) {
   return XXX_NOT_OK;
  }
 
 nbox_t ShoreKey= Box.ShoreBox();
 
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 XxxShoreKey KeyElem((void *) (RecId), sizeof(XxxShoreObjectId));
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->destroy_md_assoc(VolumeId, Index, 
								ShoreKey, 
								KeyElem), 
			   RetVal)),
	  XXX_SHOREERR, "Cannot enter file into rtree index", rc);
 XXX_BADRC_RCRETURN(rc);
 
 return rc;
}


XxxErrCode 
ShoreRTree::GetEntry(const XxxShoreVolumeId& VolumeId,
		     const XxxShoreObjectId& Index,
		     const XxxADTValue* RecValue,
		     XxxValueExprPlan *ValuePlan,
		     void *DataBuf,
		     int DataBufSize,
		     bool& Found)
{
 XxxBasicType IndxColType = ValuePlan->BasicType();
 
 XxxBoundingBox Box;
 if (XxxADTArray[IndxColType]->GetBoundingBox(RecValue,Box)==XXX_FALSE) {
   return XXX_NOT_OK;
  }
 
 nbox_t ShoreKey = Box.ShoreBox();
 
 XxxErrCode rc  = XXX_OK;
 w_rc_t RetVal;
 smsize_t DBufSize = DataBufSize;
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->find_md_assoc(VolumeId, 
							     Index, 
							     ShoreKey,
							     DataBuf, 
							     DBufSize,
							     Found), RetVal)),
	  XXX_SHOREERR, "Cannot find entry in index", rc);
 XXX_BADRC_RCRETURN(rc);
 DataBufSize = DBufSize;
 // is a memcpy needed here for alignment bugs??
 
 return rc;
}

XxxErrCode 
ShoreRTree::GetEntry(const XxxShoreVolumeId& VolumeId,
		     const XxxShoreObjectId& Index,
		     const XxxADTValue* RecValue,
		     int RecSize,
		     XxxBasicType IndxColType,
		     void *DataBuf,
		     int DataBufSize,
		     bool& Found)
{
 XxxBoundingBox Box;
 if (XxxADTArray[IndxColType]->GetBoundingBox(RecValue,Box)==XXX_FALSE) {
   return XXX_NOT_OK;
  }
 
 nbox_t ShoreKey = Box.ShoreBox();
 
 XxxErrCode rc  = XXX_OK;
 w_rc_t RetVal;
 smsize_t DBufSize = DataBufSize;
 XXX_CALL((XXX_SHORE_RETFN(getSSM(XxxShoreSM)->find_md_assoc(VolumeId, 
							     Index, 
							     ShoreKey,
							     DataBuf, 
							     DBufSize,
							     Found), RetVal)),
	  XXX_SHOREERR, "Cannot find entry in index", rc);
 XXX_BADRC_RCRETURN(rc);
 DataBufSize = DBufSize;
 // is a memcpy needed here for alignment bugs??
 
 return rc;
}

// This is really not sufficient. We need to check that a bounding box
// can be extracted from the first argument of the function.
// So i guess we need to get a plan and evaluate the first argument.
// This is the same work done in getRangeSpec
XxxErrCode
ShoreRTree::Match(const XxxValueExpression* IndexExpr,
		  const XxxValueExpression* Expr,
		  XxxBool& Success,
		  XxxIDXClass::RelMatchInfo*& MatchedInfo) const
{
 Success = XXX_FALSE;
 // Check to see that Expr is a func expression and that Expr->Owner
 // Matches the Index Expr.
 if (Expr->ValType() ==  XxxValueExpression::FUNC_EXPR) {
   XxxFuncExpression* FuncExpr = (XxxFuncExpression *) Expr;

   // See if we can match the index on the owner, or args of the funcexpr
   WhereMatched Where = NOWHERE;
   
   char* Env = new char[XXX_NODE_NAME_MAX];
   strcpy(Env,"");

   if (FuncExpr->Owner->Match(IndexExpr,Env)) {
     // Index expression matches owner
     Where = OWNER;
    }
   else if (FuncExpr->ArgList!= NULL) {
     // Try to match index on args of function
     for (int i=0;i<FuncExpr->ArgList->Count();i++) {
       strcpy(Env,"");
       if ((*(FuncExpr->ArgList))[i]->Match(IndexExpr,Env)) {
	 // Argument matches index.
	 Where = ARG;
	 break;
	}
      }
    }
   
   if (Where != NOWHERE) {
     // Index Matched, get compare op for function
     // and check if we can handle that compare op.
     
     XxxCompareOp CompOp;

     // Praveen: this code is pretty hairy. There may be a bug here 
     // if we are trying to match two different ADTs (for instance,
     // polygon and point). A further complication is that since
     // we dont yet have a boolean data type, there is no notion
     // of a boolean function. Clean this up in the next release.
     // TODO:L2 Should be using FuncExpr->OwnerType 
     // instead of IndexExpr->BasicType(), but there's a bug which makes
     // the OwnerType incorrect at times (so says Fabian)
     XxxBool MatchOccured = XXX_FALSE;
     if (XxxADTArray[IndexExpr->BasicType()]->
	 GetCompOp(FuncExpr->Name,
		   CompOp, MatchOccured) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADINDEX, "Error getting comparison operator");
       return XXX_NOT_OK;
      }
     if (MatchOccured) {
       switch(CompOp) {
	case XXX_IDX_OVERLAPS:
	case XXX_IDX_INSIDE:
	case XXX_IDX_COVERS:
	case XXX_IDX_EQUALS:
	case XXX_IDX_MEETS:
	case XXX_IDX_CONTAINS:
	case XXX_IDX_COVERED_BY:
	 MatchedInfo = new RMatchInfo(CompOp,Where);
	 Success = XXX_TRUE;
	default:
	 // Can't use SHORE R* index for DISJOINT compare op.
	 Success = XXX_FALSE;
	} // switch
      } // if MatchOccured
    } // if Where != NOWHERE
  } // if it is a funcexpr

 // Does not match
 return XXX_OK;
}

//OWNER: Fabian Camargo Date:3/5/97
//PURPOSE: Get range for R* index scan
//NOTE: Changing this so that the ADT returns the range.
XxxErrCode
ShoreRTree::getRangeSpec(XxxExprPlan* LeftPlan,
			 XxxExprPlan* RightPlan,
			 XxxRelOpType OpType,
			 const XxxRecord** SibRecs,
			 const XxxRecordSchema** SibSchema,
			 const RelExecParameters* Parameters,
			 XxxIDXClass::IndexRange* Range,
			 const XxxIDXClass::MatchInfo* MatchedInfo) const
{

 if (!MatchedInfo) {
   REL_NOTE_ERROR(REL_BADINDEX, "Empty match info");
   return XXX_NOT_OK;
  }
 if (MatchedInfo->Type != XxxIDXClass::MatchInfo::REL) {
   REL_NOTE_ERROR(REL_BADINDEX, "Bad type for match info");
   return XXX_NOT_OK;
  }
 RMatchInfo *RMatchedInfo = (RMatchInfo *)MatchedInfo;
 
 // Cast range to a r* tree range.
 RTreeRange* RangeSpec = (RTreeRange *)Range;

 // R* trees get there range from the non-constant part of a
 // func expression i.e expr = f.overlaps(box) >1 then get range from f.overlaps(box).

 XxxExprPlan* ExprPlan;
 if (RMatchedInfo->LeftConstant)
   ExprPlan = RightPlan;
 else
   ExprPlan = LeftPlan;

 // Check to see that ExprPlan is a Func expression plan.

 XxxValueExprPlan* ValPlan;
 
 ValPlan = (XxxValueExprPlan *)ExprPlan;

 XxxFuncExpression::FuncPlan* FuncPlan;

 if (ValPlan->ValType() == XxxValueExpression::FUNC_EXPR) 
   FuncPlan = (XxxFuncExpression::FuncPlan *)ValPlan;
 else {
   REL_NOTE_ERROR(REL_BADINDEX, "Indexing expression must be a funcexpr");
   return XXX_NOT_OK;
  }

 if (FuncPlan->ArgPlans == NULL) {
   REL_NOTE_ERROR(REL_BADINDEX, "Indexing expression must have arguments");
   return XXX_NOT_OK;
  }

 if (RMatchedInfo->Where == OWNER) {
   // Index matched on owner. Get bounding box from arguments

   // Construct an array to hold all the argument values.
   XxxValuesArray Args(FuncPlan->ArgPlans);
   
   // Now evaluate all arguments.
   int i;
   for (i = 0; i < Args.NumValues; i++) {
     XxxValueExprPlan* ArgPlan = (*(FuncPlan->ArgPlans))[i];
     // check that the argument is constant.
     if (!ArgPlan->IsConstant()) {
       REL_NOTE_ERROR(REL_BADINDEX, "Arguments must be constant");
       return XXX_NOT_OK;       
      }
     
     if (ArgPlan->Evaluate((const XxxRecord **)SibRecs, 
			   (const XxxRecordSchema **)SibSchema, 
			   Args.Values[i]) != XXX_OK) {
       XXX_NOTE_ERROR(XXX_BADEXPR, "Unable to evaluate arguments");
       return XXX_NOT_OK;
      }
    } // for each argument
   
   // Get bounding box from first argument of function.
   //Bjarke, sep. 14 2001: Added cast to remove use of temporary
   if (XxxADTArray[Args.Types[0]]->
       GetBoundingBox((const XxxADTValue*&)Args.Values[0],
		      RangeSpec->box) != XXX_TRUE){
     XXX_NOTE_ERROR(XXX_BADEXPR, "Unable to Get bounding box for range.");
     return XXX_NOT_OK;
    }
  }
 else if (RMatchedInfo->Where == ARG) {
   // Matched index on argument. Get bounding box from owner.
   
   if (FuncPlan->OwnerPlan == NULL) {
     REL_NOTE_ERROR(REL_BADINDEX, "Indexing expression must have owner");
     return XXX_NOT_OK;
    }
   XxxBasicType OwnerType = FuncPlan->OwnerPlan->BasicType();
   
   int KeySize = XxxADTArray[OwnerType]->
     MaxObjectSize(FuncPlan->OwnerPlan->GenerateCatInfo());
   XxxADTValue *OwnerVal = (XxxADTValue*)new char[KeySize];
   OwnerVal->Init();
   
   if (FuncPlan->OwnerPlan->Evaluate((const XxxRecord **)SibRecs, 
				     (const XxxRecordSchema **)SibSchema, 
				     OwnerVal) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADEXPR, "Unable to evaluate owner.");
     return XXX_NOT_OK;
    }
   // Get bounding box from owner of function.
   //Bjarke, sep. 14 2001: Added cast to remove use of temporary
   if (XxxADTArray[OwnerType]->
       GetBoundingBox((const XxxADTValue*&)OwnerVal,RangeSpec->box) != XXX_TRUE){
     XXX_NOTE_ERROR(XXX_BADEXPR, "Unable to Get bounding box for range.");
     return XXX_NOT_OK;
    }    
   delete [] ((char*)OwnerVal);
  }

 // Decide which SHORE R* compare op to use from the CompOp
 // obtained when matching info. For certain ops we, need to reverse
 // the SHORE compare op depending on where the index was matched (OWNER or
 // ARG).
 
 switch(RMatchedInfo->Comp) {
  case XXX_IDX_OVERLAPS:
  case XXX_IDX_MEETS:
   RangeSpec->Cmp = nbox_t::t_overlap;
   break;
  case XXX_IDX_EQUALS:
   RangeSpec->Cmp = nbox_t::t_exact;
   break;
  case XXX_IDX_INSIDE:
  case XXX_IDX_COVERED_BY:
   if (RMatchedInfo->Where == OWNER)
     RangeSpec->Cmp = nbox_t::t_cover;
   else if (RMatchedInfo->Where == ARG)
     RangeSpec->Cmp = nbox_t::t_inside;
   break;
  case XXX_IDX_COVERS:
  case XXX_IDX_CONTAINS:
   if (RMatchedInfo->Where == OWNER)
     RangeSpec->Cmp = nbox_t::t_inside;
   else if (RMatchedInfo->Where == ARG)
     RangeSpec->Cmp = nbox_t::t_cover;
   break;
  default:
   XXX_NOTE_ERROR(XXX_BADEXPR, "Unable to compare method for range.");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxIDXClass::CursorInfo* 
ShoreRTree::GetCursor(const XxxShoreLogicalOID& RelId,
		      const XxxShoreObjectId& IndexId,
		      XxxRecordSchema* Schema,
		      XxxBooleanExpressionPlan* Expr,
		      XxxBooleanExpressionPlan* RExpr, // residual
		      XxxBool Exact,
		      const XxxIDXClass::IndexRange* RangeSpec,
		      const RelExecParameters* Parameters,
		      const XxxRecord** SibRecs,
		      const XxxRecordSchema** SibSchema,
		      int CurrentJoinPos)
{
 // R trees are always inexact matches. 

 return new RIScanCursorInfo(RelId,
			     IndexId,
			     Schema,
			     Expr,
			     RExpr, 
			     XXX_FALSE,
			     (RTreeRange *)RangeSpec,
			     Parameters,
			     SibRecs,
			     SibSchema,
			     CurrentJoinPos);
}
