#ifndef XXX_RELCAT_H
#define XXX_RELCAT_H
//
// This file defines the classes that hold meta-info for each relation,
// the class hierarchy RelCatalog, that stores a table of such meta-info
// and the RelCatalogEngine that maintains a table of RelCatalogs
//

#include "relconfig.h"
#include "types.h"
#include "reladt.h"
#include "record.h"
#include "expr.h"
#include "exprplan.h"
#include "indexmatch.h" 
#include "prd-sort.h"

XXX_EXTERN_CLASS_DECL(XxxRelation);
/**
 * Statistics on a relation : each implementation can subclass this
 */
class RelStatsInfo {
 public:
  static const int DefaultTableCard; // defined as 1000 (relopt.C)
  static const int DefaultTupleSize; // defined as 100 (relopt.C)

  int Cardinality;
  int AvgTupleSize;

  // keep info on each attribute 
  int Arity;
  XxxADTClass::TypeStats** AttrStats;

 private:
  void init() {
    Cardinality = DefaultTableCard;
    AvgTupleSize = DefaultTupleSize;
    Arity = 0;
    AttrStats = NULL;
   }
  void reclaim();

 public:
  RelStatsInfo() { init();  }
  RelStatsInfo(int NumFields);
  virtual ~RelStatsInfo() { reclaim(); }

  virtual void Print(ostream& OStr, const XxxRecordSchema* RecSchema = NULL) const;
  virtual RelStatsInfo* Replicate(const XxxRecordSchema* RecSchema)const;

  virtual XxxErrCode ResetStatsInfo(const XxxRecordSchema* RecSchema);
  virtual XxxErrCode AccumulateStats(XxxRecord* Rec,
				     const XxxRecordSchema* RecSchema);
  virtual XxxErrCode TerminateStats(const XxxRecordSchema* RecSchema);
 };

/// Determines if a relation is stored, or a derived relation
enum RelTypeEnum { REL_STORED, REL_DERIVED };
/// A typedef for easy access
typedef enum RelTypeEnum RelationType;

/**
 * impl info on a relation : each implementation can subclass this
 */
class RelImplInfo {
 public:
  /**
   * Implementation information for an index
   */
  class IndexImplInfo { 
   public:
    // store the index pred. string
    char *IndexString;
    char* IndexName;
    XxxIndexType IndexType; 
    // This is an unsatisfactory way to deal with clustered indexes.
    // Firstly, we are assuming that the person creating the index
    // correctly specifies whether it is clustered.
    // Secondly, we only have a binary state -- clustered or not.
    // However, indexes could be mostly clustered.
    XxxBool Clustered;
    XxxBool Unique;
    XxxValueExpression *IndexExpr;
    XxxExprPlan *IndexPlan;

    IndexImplInfo (const char *IndxStr, const char *IndxName, 
		   const XxxIndexType type, 
		   XxxBool IsUnique,
		   XxxBool IsClustered,
		   const XxxRecordSchema *RecordSchema);
    // copy constructor : needed for replicate
    IndexImplInfo(const IndexImplInfo& IInfo);
    virtual ~IndexImplInfo() {
      if (IndexString) delete [] IndexString;
      if (IndexName) delete [] IndexName;
      if (IndexExpr) delete IndexExpr;
      if (IndexPlan) delete IndexPlan;
     };
    virtual void Print(ostream& OStr) const {}
    virtual XxxErrCode Matches(XxxBooleanExpressionPlan* Expr,
			       int SiblingArity,
			       int CurrentRelationPos,
			       XxxBool& Success,
			       XxxMatchInfo*& MatchedInfo) const;
    virtual IndexImplInfo* Replicate() const = 0;
   private:
    //// to match a relboolean expr
    virtual XxxErrCode matches(XxxRelBooleanPlan* Expr,
			       int SiblingArity,
			       int CurrentRelationPos,
			       XxxBool& Success,
			       XxxMatchInfo*& MatchedInfo) const;
    //// to match a boolboolean expr
    virtual XxxErrCode matches(XxxBoolBooleanPlan* Expr,
			       int SiblingArity,
			       int CurrentRelationPos,
			       XxxBool& Success,
			       XxxMatchInfo*& MatchedInfo) const;
    
   };
  /**
   *
   */
  class IndexList { 
   public:
    IndexImplInfo* IndxInfo;
    IndexList* Next;
    IndexList(IndexImplInfo* NewIndex, IndexList* Nxt = NULL) {
      IndxInfo = NewIndex;
      Next = Nxt;
     }
    virtual ~IndexList() {
      // Do not delete recursively here! Recursive deletes are
      // bad for the threads package (puts frames on the stack).
      // Instead, use a while loop where this is being deleted.
    }
    IndexList* Replicate() const;
    void Destroy();
    virtual XxxErrCode Matches(XxxBooleanExpressionPlan* Expr,
			       int SiblingArity,
			       int CurrentRelationPos,
			       XxxBool& Success,
			       XxxMatchInfo*& MatchedInfo,
			       IndexImplInfo*& IndxInfo) const;
   };
  /// a list of sort order properties
  class SortOrders {
   public:
    unsigned int NumSortOrders;
    XxxSortInfo* SortInfoArray;
    SortOrders() { NumSortOrders = 0; SortInfoArray = NULL;}
    ~SortOrders() { if (SortInfoArray) delete [] SortInfoArray;}
    void Init(unsigned int NumSorts) {
      NumSortOrders = NumSorts;
      SortInfoArray = new XxxSortInfo[NumSorts];
     }
    XxxBool Matches(const XxxSortInfo& SortInfo) const;
    XxxSortInfo & operator[](int index) const
     { return SortInfoArray[index];}
   };

  /**
   * Marks a relation as stored on disk, or as a derived relation which is
   * the output of a relational operator.
   */
  RelationType RelType;
  IndexList* Indexes;  // list of available indexes
  int NumIndexes;
  int SourceIndex; // which data src did this come from?
  //// Sort properties of the relation: unsorted by default
  SortOrders SortInfo; 
  
  RelImplInfo() : RelType(REL_STORED), NumIndexes(0), SourceIndex(0) 
  { Indexes = new IndexList(NULL); }

  //// copy constructor : needed for Replicate()
  RelImplInfo(const RelImplInfo&, XxxBool CopyIndexes);
  virtual ~RelImplInfo();
  // All subclasses should redefine this method !!
  virtual RelImplInfo* Replicate(XxxBool CopyIndexes) const {
    return new RelImplInfo(*this, CopyIndexes);
   }
  void AddIndex(IndexImplInfo* IndxInfo) {
    Indexes->Next = new IndexList(IndxInfo, Indexes->Next);
    NumIndexes++;
   }
  virtual void Print(ostream& OStr) const;

 };

/**
 * structure that holds meta-information on a particular relation
 */
class RelCatalogInfo : public XxxADTMetaInfo {
  friend class RelTextFileCatalog;
  friend class RelDataEngine;
  friend class RelQueryNode;
  void init() {
    Named = XXX_TRUE;
    recSchema = NULL;
    statsInfo = NULL;
    implInfo = NULL;
   }
  
 public:
  XxxRecordSchema *recSchema;   // record structure of relation
  RelStatsInfo *statsInfo;      // associated statistics
  RelImplInfo *implInfo;        // associated implementation information
  XxxBool Named;
  char RelName[REL_RELNAME_MAX];
  // generic catalog interface
  const char *RelationName() const { return RelName;}
  XxxRecordSchema *RecordSchema() const { return recSchema;}
  RelStatsInfo *StatsInfo() const { return statsInfo;}
  RelImplInfo *ImplInfo() const { return implInfo;}

  // following are needed by the template Table class
  RelCatalogInfo() { this->init(); }
  RelCatalogInfo(const char *RName) {
    strncpy(RelName, RName, REL_RELNAME_MAX);
    RelName[REL_RELNAME_MAX-1] = 0;
    this->init();
   }
  virtual ~RelCatalogInfo() {}
  virtual XxxErrCode Replicate(XxxADTMetaInfo*& NewMI) const {
    NewMI = Replicate1();
    return XXX_OK;
  }
  // change the name of this to avoid compiler warnings
  RelCatalogInfo* Replicate1(XxxBool CopyIndexes = XXX_TRUE) const;
  /**
   * Returns the number of bytes necessary to hold the serialized version
   * of this meta-info object.
   * @param Size (out) The required number of bytes
   * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
   */
  virtual XxxErrCode SerializeCopySize(int &Size) const;
  /**
   * Creates a binary version of this meta-info object.
   * @param Data An array of bytes of at least the size specified by the
   *             SerializeCopySize() function.
   * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
   */
  virtual XxxErrCode SerializeCopy(char *Data) const;
  
  virtual void Destroy() {
    if (recSchema) delete recSchema;
    if (statsInfo) delete statsInfo;
    if (implInfo) delete implInfo;
   }
  XxxBool Valid() {
    return ((implInfo && statsInfo && recSchema) ?XXX_TRUE:XXX_FALSE);
   }
  virtual int Hash() const { return (RelName[0] + RelName[1]);}
  virtual int operator==(const RelCatalogInfo& RInfo) const
   { return !strcasecmp(RelName, RInfo.RelName);}
  virtual void Print(ostream& OStr) const ;
  virtual void PrintSchema(ostream& OStr) const ;

  // TODO:L2 : check if the schemas are the same
  virtual XxxBool IsCompatible(const XxxADTMetaInfo* OtherMI) const {
    return XXX_FALSE;
   }
 };

/**
// Each RelCatalog is a table of catalog entries (RelCatalogInfo), 
// one for each named relation. There is also one entry for all the
// unnamed relations that occur as values of some particular attribute of 
// of another structure. This entry is given the composite name derived from
// the structure name and the attribute name.
**/



class RelCatalog : virtual public XxxMultiSet<RelCatalogInfo> {
 protected:
 public:
  RelCatalog() {} 
  virtual ~RelCatalog() {}

  // well defined public interface for a catalog
  virtual void PrintCats(ostream& OStr);
  virtual void PrintSchemas(ostream& OStr);

  // subclasses should redefine this method
  virtual XxxErrCode GetCatInfo(const char* RelName, RelCatalogInfo &,
				XxxBool& Found) const = 0;
  int operator==(const RelCatalog& Other) { return 0; }
 };

#endif
