#include <stdio.h>
#include <stdlib.h>
#include <fstream.h>
#include <dirent.h>
#include <sys/stat.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "xxx_thread.h"
#include "env.h"
#include "reladt.h"
#include "reldata.h"
#include "relenv.h"
#include "shorerel.h"
#include "relerror.h"
#include "relation.h"
#include "XxxIndex.h"
#include "funcexpr.h"

// after how many records loaded should we break and commit
const int RelDataEngine::LoadCommitThreshold = 1024;
// prefix for primary key index name
const char* const RelDataEngine::PKEY_INDX_PREFIX = "_PKEY_";

//
// Methods of RelDataEngine
//
RelDataEngine::RelDataEngine(XxxRelationEngine* SrcEng)
{
 RelDataSrc* Tmp;

 RelEng = SrcEng;
 defaultSrc = REL_INVALID_SRC;

 // initialize the array
 for (int i = 0; i < RelNumSrcTypes; i++) {
   sourceArray[i] = NULL;
 }

 /****
   // add a data src for relations as Ascii Unix files
   addDataSrc(Tmp = new RelUnixAsciiDataSrc(this));
 defaultSrc = Tmp->SrcType;
 **/
 // add a SHORE data src
 if (XxxShoreSM) {
   addDataSrc(Tmp = new RelShoreDataSrc(this));
   defaultSrc = Tmp->SrcType;
  }

 //addDataSrc(new RelMemoryDataSrc(this));
}

RelDataEngine::~RelDataEngine()
{
 for (int i = 0; i < RelNumSrcTypes; i++) {
   if (sourceArray[i]) delete sourceArray[i];
  }
}

void RelDataEngine::PrintRels(ostream& OStr)
{
 for (int i = 0; i < RelNumSrcTypes; i++) {
   if (sourceArray[i]) sourceArray[i]->PrintRels(OStr);
  }
}
void RelDataEngine::PrintSchemas(ostream& OStr)
{
 for (int i = 0; i < RelNumSrcTypes; i++) {
   if (sourceArray[i]) sourceArray[i]->PrintSchemas(OStr);
  }
}
void RelDataEngine::PrintCats(ostream& OStr)
{
 for (int i = 0; i < RelNumSrcTypes; i++)
   if (sourceArray[i]) sourceArray[i]->PrintCats(OStr);
}
void RelDataEngine::PrintDevs(ostream& OStr)
{
 for (int i = 0; i < RelNumSrcTypes; i++) {
   if (sourceArray[i]) sourceArray[i]->PrintDevs(OStr);
  }
}
void RelDataEngine::PrintSrcs(ostream& OStr)
{
 for (int i = 0; i < RelNumSrcTypes; i++) {
   if (sourceArray[i]) {
     sourceArray[i]->Print(OStr);
     OStr << endl;
    }
  }
}

XxxErrCode RelDataEngine::DeleteCatInfo(const char* RelName)
{
 // get catalog information to insure it exists
 RelCatalogInfo CatInfo;
 XxxBool Found = XXX_FALSE;

 if (GetCatInfo(RelName, CatInfo, Found) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADSRC, "Error getting catalog info");
   return XXX_NOT_OK;
  }
 if (Found != XXX_TRUE) {
   REL_NOTE_ERROR(REL_BADTABLE, "Attempt to delete catalog info on a non-existent relation");
   return XXX_NOT_OK;
  }

 // get hold of the relation
 RelHandle *Handle;
 if (GetRelation(RelName, Handle) != XXX_TRUE) {
   REL_NOTE_ERROR(REL_BADTABLE, "Cannot find relation handle to load");
   return XXX_NOT_OK;
  }

 // delete the catalog information from the device on which it resides
 if (sourceArray[Handle->SrcType]->DeleteCatInfo(RelName) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADSRC, "Error deleting catalog info on a relation");
   return XXX_NOT_OK;
  }

 return XXX_OK;
}

XxxErrCode RelDataEngine::GetCatInfo(const char* RelName,
				     RelCatalogInfo& AnswerCat,
				     XxxBool& Found)
{
 Found = XXX_FALSE;
 for (int i = 0; i < RelNumSrcTypes; i++) {
   if (sourceArray[i] &&
       sourceArray[i]->GetCatInfo(RelName, AnswerCat, Found) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADSRC, "Error getting catalog info");
     return XXX_NOT_OK;
    }
   if (Found) {
     RelImplInfo* ImplInfo = AnswerCat.implInfo;
     if (ImplInfo) ImplInfo->SourceIndex = i;
     break;
    }
  }
 return XXX_OK;
}

XxxErrCode RelDataEngine::AddCatInfo(const char* RelName,
				     const  XxxBool Named,
				     const RelDataSrc::AttrTypeList* Attrs,
				     const RelDataSrc::KeyList* Keys,
				     const RelSrcTypes SrcType,
				     const char* DevName,
				     const char* RelImplType)
{
 if (sourceArray[SrcType]) {
   return sourceArray[SrcType]->AddCatInfo(RelName, DevName, RelImplType, 
					   Attrs, Keys);
  }
 else {
   REL_NOTE_ERROR(REL_BADSRC, "Invalid source type");
   return XXX_NOT_OK;
  }
}

XxxBool RelDataEngine::RelationIsProtected(const char* RelName) {
  for (int i = 0; i < RelNumSrcTypes; i++) 
    if (sourceArray[i] &&
	sourceArray[i]->RelationIsProtected(RelName))
      return XXX_TRUE;
 return XXX_FALSE;
}

XxxBool RelDataEngine::GetRelation(const char* RelName,
				   XxxRelation*& AnswerRel,
				   const RelImplInfo* ImplInfo)
{
 for (int i = 0; i < RelNumSrcTypes; i++) {
   if (sourceArray[i] &&
       sourceArray[i]->GetRelation(RelName, AnswerRel, ImplInfo) == XXX_TRUE) {
     return XXX_TRUE;
    }
  }
 return XXX_FALSE;
}

XxxBool RelDataEngine::GetRelation(const char* RelName,
				   RelHandle*& Handle)
{
 for (int i = 0; i < RelNumSrcTypes; i++) {
   if (sourceArray[i] &&
       sourceArray[i]->GetRelation(RelName, Handle) == XXX_TRUE) {
     return XXX_TRUE;
    }
  }
 return XXX_FALSE;
}

XxxErrCode RelDataEngine::GetRelation(const RelHandle* Handle,
				      XxxRelation*& AnswerRel,
				      const RelImplInfo* ImplInfo)
{
 return sourceArray[Handle->SrcType]->GetRelation(Handle, AnswerRel, ImplInfo);
}

// generate a suitable unique index for a primary key
XxxErrCode RelDataEngine::makePKeyIndex(const char* TableName, 
					RelDataSrc::AttrTypeList* Attrs,
					RelDataSrc::KeyList* Keys)
{
 if (!Keys) return XXX_OK;

 if (Keys->Next) {
   cerr << "No support yet for multi-attribute keys" << endl;
   return XXX_OK;
  }
 
 XxxValueExpression* IndexExpr = NULL;
 // otherwise, make an index expression corresponding to the key attr
 if (!Attrs) {
   REL_NOTE_ERROR(REL_BADCAT, "No attrs available");
   return XXX_NOT_OK;
  }
 int AttrPos = 0;
 XxxBool AttrFound = XXX_FALSE;
 while (Attrs) {
   if (!strcmp(Keys->AttrName, Attrs->Name)) {
     AttrFound= XXX_TRUE;
     break;
    }
   AttrPos++;
   Attrs = Attrs->Next;
  } // while
 if (!AttrFound) {
   REL_NOTE_ERROR(REL_BADCAT, "Key does not match any attribute");
   return XXX_NOT_OK;
  }

 // now construct an XxxUnknownValue from AttrPos
 // TODO:L1 : who reclaims this?
 IndexExpr = new XxxUnknownValue(Keys->AttrName, TableName);

 char *IndexName = new char[strlen(TableName) + strlen(PKEY_INDX_PREFIX) + 1];
 strcpy(IndexName, PKEY_INDX_PREFIX);
 strcat(IndexName, TableName);
 if (CreateNewIndex(TableName, IndexName, IndexExpr, XXX_TRUE, 
		    XXX_TRUE, // assume clustered on primary key
		    XxxIDXArray[XXX_BTREE]->TypeName) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADCAT, "Error making index for primary key");
   return XXX_NOT_OK;
  }
 delete [] IndexName;

 // delete IndexExpr; // Check if this is ok : TODO:L1
 return XXX_OK;
}

XxxErrCode RelDataEngine::CreateNewRelation(RelHandle*& Handle, 
					    const char* Name, 
					    RelSrcTypes SrcType,
					    const char* DevName,
					    const char* RelImplType,
					    RelDataSrc::AttrTypeList* Attrs,
					    RelDataSrc::KeyList* Keys,
					    XxxBool MakeCatEntry) 
{
 if (sourceArray[SrcType]) {
   if (sourceArray[SrcType]->CreateNewRelation(Handle, Name, DevName, 
					       RelImplType, Attrs, Keys,
					       MakeCatEntry) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADCAT, "Error creating new relation");
     return XXX_NOT_OK;
    }
   // also try to create a unique index on the primary key
   if (makePKeyIndex(Name, Attrs, Keys) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADCAT, "Error making index expr for primary key");
     return XXX_NOT_OK;
    }
  }
 else {
   REL_NOTE_ERROR(REL_BADSRC, "Invalid source type");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelDataEngine::CreateNewRelation(RelHandle*& Handle, 
					    const char* Name, 
					    const RelImplInfo* ImplInfo,
					    RelDataSrc::AttrTypeList* Attrs,
					    RelDataSrc::KeyList* Keys,
					    XxxBool MakeCatEntry) 
{
 // SrcType should be stored in the ImplInfo
 // for now, use default
 RelSrcTypes SrcType;
 SrcType = DefaultSrcType();
 if (sourceArray[SrcType]) {
   if (sourceArray[SrcType]->CreateNewRelation(Handle, Name, ImplInfo,
					       Attrs, Keys, MakeCatEntry) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADCAT, "Error creating new relation");
     return XXX_NOT_OK;
    }
   // also try to create a unique index on the primary key
   if (makePKeyIndex(Name, Attrs, Keys) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADCAT, "Error making index expr for primary key");
     return XXX_NOT_OK;
    }
  }
 else {
   REL_NOTE_ERROR(REL_BADSRC, "Invalid source type");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

// OWNER: Fabian DATE: 1/23/97
// PURPOSE: Creates a new index of a specified type based 
// on the expression passed in as an argument.
// NOTES: Only for Arithmetic Expression and ADT Function.

XxxErrCode 
RelDataEngine::CreateNewIndex(const char* RelName,
			      const char* IndxName,
			      XxxValueExpression *Expr,
			      const XxxBool Unique,
			      const XxxBool Clustered,
			      const char *IndxType)
{
 XxxBool Success;

 if (Expr->IsConstant() || 
     (Expr->ValType() == XxxValueExpression::BOOL_EXPR) ||
     (Expr->ValType() == XxxValueExpression::HOST_VAR) ||
     ((Expr->ValType() == XxxValueExpression::FUNC_EXPR) &&
      (((XxxFuncExpression*)Expr)->IsAggr()))) {
   REL_NOTE_ERROR (REL_BADINDEX,
		   "Can ONLY create index on value expression");
   return XXX_NOT_OK;
 }
 
 XxxIndexType IdxType = XXX_UNKNOWN;
 int i;
 for (i=0;i<XxxNumIDXs;i++) {
   if (strcasecmp(IndxType,XxxIDXArray[i]->TypeName)==0) {
     IdxType = i;
     break;
   }
 }

 if (IdxType == XXX_UNKNOWN) {
   REL_NOTE_ERROR (REL_BADINDEX,
		   "Can't create index of this type.");
   return XXX_NOT_OK;
 }

 for (i = 0; i < RelNumSrcTypes; i++) {
   if (sourceArray[i]) {
     if (sourceArray[i]->CreateNewIndex(RelName,
					IndxName,
					Expr,
					Unique,
					Clustered,
					Success,
					IdxType) != XXX_OK) {
       REL_NOTE_ERROR (REL_BADINDEX, "Error creating index");
       return XXX_NOT_OK;
     }
     if (Success)
       return XXX_OK;
   }
 }
 REL_NOTE_ERROR (REL_BADINDEX, "Couldnt find corresponding relation");
 return XXX_NOT_OK;
}

XxxErrCode RelDataEngine::
CreateNewView(const char* Name, const char* ViewInfo)
{
 RelSrcTypes SrcType;
 SrcType = DefaultSrcType();
 if (sourceArray[SrcType]) {
   REL_CALL_FN(sourceArray[SrcType]->CreateNewView(Name, ViewInfo),
	       REL_BADCAT, "Error creating new view");
  }
 else {
   REL_NOTE_ERROR(REL_BADSRC, "Invalid source type");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelDataEngine::DeleteView (const char *ViewName)
{
 RelSrcTypes SrcType;
 SrcType = DefaultSrcType();
 if (sourceArray[SrcType]) {
   REL_CALL_FN(sourceArray[SrcType]->DeleteView(ViewName),
	       REL_BADCAT, "Error creating new view");
  }
 else {
   REL_NOTE_ERROR(REL_BADSRC, "Invalid source type");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelDataEngine::
RetrieveView(const char* ViewName, 
	     RelViewClass*& View, XxxBool& Found) const
{
  Found = XXX_FALSE;
  RelSrcTypes SrcType;
  SrcType = DefaultSrcType();
  if (sourceArray[SrcType]) {
    REL_CALL_FN(sourceArray[SrcType]->RetrieveView(ViewName, View, Found),
		REL_BADCAT, "Error retrieving new view");
  }
  else {
    REL_NOTE_ERROR(REL_BADSRC, "Invalid source type");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode RelDataEngine::PrintViews(ostream& OStr) const
{
 RelSrcTypes SrcType;
 SrcType = DefaultSrcType();
 if (sourceArray[SrcType]) {
   REL_CALL_FN(sourceArray[SrcType]->PrintViews(OStr),
	       REL_BADCAT, "Error retrieving new view");
  }
 else {
   REL_NOTE_ERROR(REL_BADSRC, "Invalid source type");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelDataEngine::DeleteRelation(const char *RelName) {
  if (RelationIsProtected(RelName)) {
    char Buf[100];
    sprintf (Buf, "%s is a protected relation", RelName);
    REL_NOTE_ERROR(REL_BADTABLE, Buf);
    return XXX_NOT_OK;
  }

  // get the relation's handle
  RelHandle *Handle;
  if (GetRelation(RelName, Handle) != XXX_TRUE) {
    REL_NOTE_ERROR(REL_BADTABLE, "Cannot find relation handle to load");
    return XXX_NOT_OK;
  }
  
  // delete the relation from the data source
  return sourceArray[Handle->SrcType]->DeleteRelation(RelName, Handle);
}

// TODO:L2 Change this to be able to load from a URL
XxxErrCode RelDataEngine::LoadRelation(const char* RelName, 
				       istream& InStream, 
				       const RelCatalogInfo* CatInfo,
				       RelHandle*& Handle,
				       const XxxBool ToCreate)
{
 // note: we break loads to commit the xacts to conserve log space.
 // so, we cannot allow multi-stmt xacts with load commands in them.
 if (XXX_THISENV().MultiStmtXact) {
   XXX_NOTE_ERROR(XXX_BADCMD, "Don't load data within multi-stmt xact");
   return XXX_NOT_OK;   
  }
 // get hold of the relation
 if (ToCreate) {
   // srctype and dev name should come from the catalog info!!
   if (CreateNewRelation(Handle, NULL, CatInfo->ImplInfo(), 
			 NULL, NULL, XXX_FALSE) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADTABLE, "Cannot create relation handle to load");
     return XXX_NOT_OK;
    }
  }
 else {
   if (GetRelation(RelName, Handle) != XXX_TRUE) {
     REL_NOTE_ERROR(REL_BADTABLE, "Cannot find relation handle to load");
     return XXX_NOT_OK;
    }
  }
 XxxRelation* Rel;
 if (GetRelation(Handle, Rel, CatInfo->ImplInfo()) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADTABLE, "Cannot find relation to load");
   return XXX_NOT_OK;
  }
 Rel->CatalogInfo = CatInfo;

 // here's the place to change the store property of the relation
 if (Rel->PrepareForLoad() != XXX_OK) {
   REL_NOTE_ERROR(REL_BADTABLE, "Not a loadable table");
   return XXX_NOT_OK;
  }
 
 XxxRecordSchema* RecSchema = Rel->CatalogInfo->RecordSchema();
 XxxValuesArray ValArray(RecSchema);
 unsigned int RecSize;
 XxxRecord* Rec;
 int RecordCounter = 0;
 XxxBool StartParenFound = XXX_FALSE;
 char FirstChar;
 // try to find the opening bracket
 InStream >> FirstChar;
 if (!InStream.good()) {
   char buf[100];
   sprintf(buf, "Error reading file at position %ld", InStream.tellg());
   REL_NOTE_ERROR(REL_BADTXTFILE, buf);
   return XXX_NOT_OK;
  }
 if (FirstChar == '(') StartParenFound = XXX_TRUE;
 else InStream.putback(FirstChar);

 //--> emc 2/25/97
 // Initialize for character delimited files
 XxxBool Delimited = XXX_FALSE;
 char Delimiter='\t';
 XxxRelEnv* REnv =((XxxRelEnv*)(XXX_THISENGENV(RelationEngineId)));
  if (REnv->Delimited) { // default is false
    Delimited = XXX_TRUE;
    Delimiter = REnv->Delimiter; // default is '\t'
  }

 // read the records and insert them into the file
 while (1) {
   // if there was an opening bracket, try to find the closing bracket 
   if (StartParenFound == XXX_TRUE) {
     InStream >> FirstChar;
     if (!InStream.good()) {
       char buf[100];
       sprintf(buf, "Error reading file at position %ld", InStream.tellg());
       REL_NOTE_ERROR(REL_BADTXTFILE, buf);
       return XXX_NOT_OK;
     }
     if (FirstChar == ')') break;
     else InStream.putback(FirstChar);
   }
   // read a record
   char buf[XXX_TEXTLINE_MAX];
   if (ValArray.GetTextRecordValues(InStream, RecSchema, RecSize, 
				    Delimited, Delimiter) != XXX_OK) {
     // if the file is at eof, break     
     if (InStream.eof()) break;

     if (InStream.bad()) {
       char buf[100];
       sprintf(buf, "Serious error reading record %d at byte %ld: quitting", 
	       RecordCounter, InStream.tellg());
       REL_NOTE_ERROR(REL_BADTXTFILE, buf);
       // Read to the end of the line to further the stream
       InStream.getline(buf, XXX_TEXTLINE_MAX);
       return XXX_NOT_OK;
     } else {
       sprintf(buf, 
	       "Non-serious error reading record %d at byte %ld; skipping...", 
	       RecordCounter, InStream.tellg());
       REL_NOTE_ERROR(REL_BADTXTFILE, buf);
       // clear the error bits in the InStream
       InStream.clear();
       // read upto the next newline
       // This is not really correct. it should skip upto the
       // next synchronization point, whatever that is.
       // For nested data, this may span several lines.
       // What should i do about this?
       InStream.getline(buf, XXX_TEXTLINE_MAX);
       continue;
     }
   }

   Rec = new XxxRecord(ValArray, RecSchema);
   
   // insert the record
   if (Rel->AddItem(Rec) != XXX_OK) {
     char buf[200];
     sprintf(buf, "Unable to insert record %d from byte %ld: skipping", 
        RecordCounter, InStream.tellg());
     REL_NOTE_ERROR(REL_BADTABLE, buf);
     delete Rec;
     continue;
    }
   
   delete Rec;
   
   if ((++RecordCounter % LoadCommitThreshold) == 0) {
     *(XXX_THISENV().OutputStr) << " .. records loaded: " << RecordCounter << endl;
     if (XXX_IS_ERROR()) {
       XXX_PRINT_ERROR(*(XXX_THISENV().ErrorStr), 
		       "Error loading records");
       XXX_CLEAR_ERROR();
     }
     // TODO:L2 check this: is this necessary with t_loadfile?
     if (XxxShoreSM && !XXX_THISENV().MultiStmtXact) {
       // Needed to ensure that the log space does not get exhausted
       XxxShoreSM->CommitTrans();
       XxxShoreSM->BeginTrans();
       // change the store property of the relation again 
       // since the commit may have affected it
       if (Rel->PrepareForLoad() != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADTABLE, "Not a loadable table");
	 return XXX_NOT_OK;
	}
      }
   } // if exceeded LoadCommitThreshold
   if (InStream.eof()) break;
 }
 
 // get rid of the relation object
 delete Rel;
 
 if (RecordCounter) {
   *(XXX_THISENV().OutputStr) << " Rel:: Total records loaded: " << RecordCounter << endl;
   return XXX_OK;
  }
 else return XXX_NOT_OK;
}

XxxErrCode RelDataEngine::LoadRelation(const char* RelName, 
           const char* FileName)
{
 // get hold of the file
 ifstream InFStream;
 if (CheckFileExists(InFStream, FileName) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADTXTFILE, "Cannot find file to load from");
   return XXX_NOT_OK;
  }

 // get hold of the catalog info
 RelCatalogInfo CatInfo;
 XxxBool MIFound = XXX_FALSE;
 if ((GetCatInfo(RelName, CatInfo, MIFound) != XXX_OK) ||
     (MIFound != XXX_TRUE)) {
   REL_NOTE_ERROR(REL_BADTABLE, "Cannot find catalog for relation");
   return XXX_NOT_OK;
  }
 // move into loadRelation

 RelHandle *Handle;
 return LoadRelation(RelName, InFStream, &CatInfo, Handle, XXX_FALSE);

 // delete the catalog info  (automatic)
 // close the file  (automatic)

}

void RelDataEngine::addDataSrc(RelDataSrc* NewSrc)
{
 sourceArray[NewSrc->SrcType] = NewSrc;
}

int RelDataEngine::GetHandleSize(const RelHandle* Handle) const
{
 return sourceArray[Handle->SrcType]->HandleSize();
}


int RelDataEngine::CopyHandle(const RelHandle*SrcHandle, 
         RelHandle* DstHandle,
         const RelCatalogInfo* CatInfo) const
{
 int CopySize = GetHandleSize(SrcHandle);
 memcpy((char *)DstHandle, (char *)SrcHandle, CopySize);
 return CopySize;
}
// this may decrement refcnts
void RelDataEngine::RemoveHandle(const RelHandle* Rel,
     const RelCatalogInfo* CatInfo) const
{
}

RelSrcTypes RelDataEngine::FindSrcType(const char* SrcName) const
{
 if (!SrcName) return REL_INVALID_SRC;

 for (int i = 0; i < RelNumSrcTypes; i++) {
   if (sourceArray[i] && 
       sourceArray[i] && !strcmp(sourceArray[i]->SrcName, SrcName))
     return sourceArray[i]->SrcType;
  }
 return REL_INVALID_SRC;
}

//
// Methods of RelDataSrc
//
static int printRelEntry(RelNamedHandle& DataEntry, void *arg)
{
 DataEntry.Print(*((ostream *)arg));
 *((ostream *)arg) << endl;
 // reclaim the RelHandle
 if (DataEntry.Hndl)  delete DataEntry.Hndl;
 return 1;
}
void RelDataSrc::PrintRels(ostream& OStr)
{
 // should print some identifier for the catalog src here
 ApplyAll(printRelEntry, (void *)(&OStr));
}


//
// Function : RelDataEngine::InsertTuples
// Purpose  : Insert the tuples with ValueList into relation 
//            named RelName.
//
XxxErrCode RelDataEngine::InsertTuples(const char *RelName,
				       XxxValueExprList * ValueList)
{
 // get hold of the catalog info
 RelCatalogInfo CatInfo;
 XxxBool MIFound = XXX_FALSE;

 REL_CALL_FN(GetCatInfo (RelName, CatInfo, MIFound),
	     REL_BADTABLE, "Cannot find catalog for relation");

 if (MIFound != XXX_TRUE) {
   REL_NOTE_ERROR (REL_BADTABLE, "Cannot find catalog for relation");
   return XXX_NOT_OK;
  }

 RelHandle *Handle;
 if (GetRelation (RelName, Handle) != XXX_TRUE) {
   REL_NOTE_ERROR (REL_BADTABLE, "Cannot find relation to load");
   return XXX_NOT_OK;
  }
 XxxRelation *Rel;
 REL_CALL_FN(GetRelation (Handle, Rel, CatInfo.ImplInfo ()),
	     REL_BADTABLE, "Cannot find relation to load");

 XxxRecordSchema *RecSchema;
 Rel->CatalogInfo = &CatInfo;
 RecSchema = Rel->CatalogInfo->RecordSchema ();
 XxxValuesArray ValArray(RecSchema);

 XXX_CALL_RET(ValueList->TypeCheck(RecSchema),
	      XXX_BADVALUE, "Type check failed");

 if (ValueList->IsConstant () != XXX_TRUE) {
   XXX_NOTE_ERROR (XXX_BADVALUE, "Only constant values can be inserted");
   return XXX_NOT_OK;
  }

 XxxValExprPlanList *PlanList;
 XXX_CALL_RET(ValueList->Optimize(PlanList),
	      XXX_BADVALUE, "Error optimizing value list");

 XXX_CALL_RET(PlanList->Evaluate((const XxxRecord **) NULL,
				 (const XxxRecordSchema **) NULL,
				 ValArray.Values),
	      XXX_BADVALUE, "Error evaluating value list");

 XxxRecord *Rec = new XxxRecord(ValArray, RecSchema);
 // insert the record
 if (Rel->AddItem(Rec) != XXX_OK) {
   char buf[200];
   sprintf (buf, "Unable to insert record\n");
   REL_NOTE_ERROR (REL_BADTABLE, buf);
   delete PlanList;
   delete Rec;
   delete Rel;
   return XXX_NOT_OK;
  }

 delete PlanList;
 delete Rec;
 delete Rel;

 return XXX_OK;
}

XxxErrCode
RelDataEngine::DeleteIndex (const char *IndxName) 
{
 // cannot delete a primary index directly from parser
 if (!strncmp(IndxName, PKEY_INDX_PREFIX, strlen(PKEY_INDX_PREFIX))) {
   REL_NOTE_ERROR(REL_BADINDEX, "Cannot delete primary key index");
   return XXX_NOT_OK;
  }

 char *RelName = 0;
 int Src = REL_INVALID_SRC;
 for (int i = 0; i < RelNumSrcTypes; i++)
   if (sourceArray[i] &&
       sourceArray[i]->GetRelationOfIndex(IndxName, RelName) == XXX_TRUE) {
     Src = i;
     break;
    }
 if (Src == REL_INVALID_SRC) {
   REL_NOTE_ERROR (REL_BADINDEX, "Cannot find index in any data source");
   if (RelName) delete [] RelName;
   return XXX_NOT_OK;
  }
 if (sourceArray[Src]->DeleteIndex(IndxName, RelName) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADSRC, "Could not delete index from data source");
   if (RelName) delete [] RelName;
   return XXX_NOT_OK;
  } 
 else {
   if (RelName) delete [] RelName;
   return XXX_OK;
  }
}

XxxErrCode
RelDataEngine::DeleteTuples (const char *RelName,
			     XxxBooleanExpression *BoolExpr) 
{
  if (RelationIsProtected(RelName)) {
    char Buf[100];
    sprintf (Buf, "%s is a protected relation: no deletes allowed", RelName);
    REL_NOTE_ERROR(REL_BADTABLE, Buf);
    return XXX_NOT_OK;
  }
  // get the relation's handle
  RelHandle *Handle;
  if (GetRelation(RelName, Handle) != XXX_TRUE) {
    REL_NOTE_ERROR(REL_BADTABLE, "Cannot find relation handle to load");
    return XXX_NOT_OK;
  }

  // Get the relation's implementation information
  XxxBool Found = XXX_FALSE;
  RelCatalogInfo CatInfo;

  if ((GetCatInfo(RelName, CatInfo, Found) != XXX_OK) ||
      (Found == XXX_FALSE)) {
    REL_NOTE_ERROR(REL_BADSRC, "Error getting catalog info");
    return XXX_NOT_OK;
  }

  XxxExprPlan *FilterPlan = 0;

  if (BoolExpr) { 
    // where clause is specified; if not, SQL-92 requires default of "true"
    // so leave filter plan non-existent; the cursor which uses it will
    // return every tuple since no filter is specified
    REL_CALL_FN(BoolExpr->ResolveVars(CatInfo.RecordSchema()),
		REL_BADQUERY, "Could not resolve variable references");
    REL_CALL_FN(BoolExpr->TypeCheck(),
		REL_BADQUERY, "Could not type-check expression");
    REL_CALL_FN(BoolExpr->Optimize(FilterPlan),
		REL_BADQUERY, "Could not form expression plan");
  }
    
  // delete the tuples from the data source
  RelImplInfo *ImplInfo = CatInfo.ImplInfo();
  REL_CALL_FN(sourceArray[Handle->SrcType]->
	      DeleteTuples(Handle,
			   (XxxBooleanExpressionPlan *)FilterPlan,
			   CatInfo),
	      REL_BADSRC, "Cannot delete tuples from data source");
     
  delete FilterPlan;
  return XXX_OK;
}

//
// Process an update statement
//
XxxErrCode
RelDataEngine::UpdateTuples(const char *RelName,
			    XxxBooleanExpression *BoolExpr,
			    const UpdateItemList* UpdateList) 
{
  // sanity check
  if (!UpdateList) {
    REL_NOTE_ERROR(REL_BADEXEC, "No update list passed to update command");
    return XXX_NOT_OK;
  }
  if (RelationIsProtected(RelName)) {
    char Buf[100];
    sprintf (Buf, "%s is a protected relation: no updates allowed", 
	     RelName);
    REL_NOTE_ERROR(REL_BADTABLE, Buf);
    return XXX_NOT_OK;
  }
  // get the relation's handle
  RelHandle *Handle;
  if (GetRelation(RelName, Handle) != XXX_TRUE) {
    REL_NOTE_ERROR(REL_BADTABLE, "Cannot find relation handle to load");
    return XXX_NOT_OK;
  }

  // Get the relation's implementation information
  XxxBool Found = XXX_FALSE;
  RelCatalogInfo CatInfo;

  if ((GetCatInfo(RelName, CatInfo, Found) != XXX_OK) ||
      (Found == XXX_FALSE)) {
    REL_NOTE_ERROR(REL_BADSRC, "Error getting catalog info");
    return XXX_NOT_OK;
  }

  XxxExprPlan *FilterPlan = 0;
  if (BoolExpr) { 
    // where clause is specified; if not, SQL-92 requires default of "true"
    // so leave filter plan non-existent; the cursor which uses it will
    // return every tuple since no filter is specified
    REL_CALL_FN(BoolExpr->ResolveVars(CatInfo.RecordSchema()),
		REL_BADEXEC, "Could not resolve variable references");
    REL_CALL_FN(BoolExpr->TypeCheck(),
		REL_BADEXEC, "Could not type-check expression");
    REL_CALL_FN(BoolExpr->Optimize(FilterPlan),
		REL_BADEXEC, "Could not form expression plan");
  }

  // now create an ExprPlanList to specify the new record to be generated
  XxxValueExprList* TargetList = NULL;
  XxxRecordSchema* Schema = CatInfo.RecordSchema();
  int ListLength = UpdateList->AttrList->Count();
  for (int AttrIndex = 0; AttrIndex < Schema->Arity(); AttrIndex++) {
    const char* AttrName = ((*Schema)[AttrIndex]).Name;
    int i;
    for (i = 0; i < ListLength; i++) {
      XxxValueExpression* ValExpr;
      REL_CALL_FN(UpdateList->AttrList->GetElement(i, ValExpr),
		  REL_BADEXEC, "Error getting update expr");
      ASSERT((ValExpr->ValType() == XxxValueExpression::COLUMN_VAR));
      XxxUnknownValue* UVal = (XxxUnknownValue*)ValExpr;
      if (UVal->Name && !strcmp(UVal->Name, AttrName)) 
	break;
    } // for loop over the AttrList
    XxxValueExpression* NewExpr;
    if (i == ListLength) {
      // this attribute is not changed by the update statement
      NewExpr = new XxxUnknownValue(AttrIndex, ((*Schema)[AttrIndex]).Type);
    }
    else {
      // get the expr from the replacement list
      XxxValueExpression* ValExpr;
      REL_CALL_FN(UpdateList->ReplList->GetElement(i, ValExpr),
		  REL_BADEXEC, "Error getting replacement expr");
      NewExpr = ValExpr->Replicate();
    }
    if (TargetList)
      TargetList = TargetList->AddExpr(NewExpr);
    else
      TargetList = new XxxValueExprList(NewExpr);
  } // for loop over the schema
  
  // now type check and optimize the update list
  REL_CALL_FN(TargetList->ResolveVars(CatInfo.RecordSchema()),
	      REL_BADQUERY, "Could not resolve update list");

  REL_CALL_FN(TargetList->TypeCheck(),
	      REL_BADQUERY, "Could not type-check update list");

  XxxValExprPlanList *TargetPlanList;
  REL_CALL_FN(TargetList->Optimize(TargetPlanList),
	      REL_BADQUERY, "Could not plan update list");
    
  XxxRelation *Rel;
  REL_CALL_FN(GetRelation(Handle, Rel, CatInfo.ImplInfo()),
	      REL_BADDEV, "Cannot retrieve relation from handle");

  Rel->CatalogInfo = &CatInfo;
  int NumUpdated = 0;
  REL_CALL_FN(Rel->UpdateRecords((XxxBooleanExpressionPlan *)FilterPlan, 
				 TargetPlanList, NumUpdated),
	      REL_BADEXEC, "Cannot update records in relation");
  if (Rel) delete Rel;

  if (FilterPlan) delete FilterPlan;
  if (TargetPlanList) delete TargetPlanList;

  return XXX_OK;
}

//
// Collect stats for a specific relation
//
XxxErrCode
RelDataEngine::RunStats (const char *RelName)
{
 if (RelName)
   return runStats(RelName);
 else {
   // need to run stats on all relations in the collection
   for (int i = 0; i < RelNumSrcTypes; i++) {
     if (sourceArray[i]) {
       // open cursor on sourceArray[i]
       XxxSetCursor Csr;
       if (sourceArray[i]->InitCursor(Csr) != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADSRC, "init cursor failed");
	 return XXX_NOT_OK;
	}

       // for each RelNamedHandle
       while (1) {
	 RelNamedHandle RelName;
	 if (sourceArray[i]->NextItem(Csr, RelName) != XXX_OK) {
	   REL_NOTE_ERROR(REL_BADSRC, "next item failed");
	   return XXX_NOT_OK;
	  }
	 if (!Csr.IsValid()) break;
	 if (runStats(RelName.Name) != XXX_OK) {
	   char buf[100];
	   sprintf(buf, "runstats failed on %s", RelName.Name);
	   REL_NOTE_ERROR(REL_BADTABLE, buf);
	   return XXX_NOT_OK;
	  }
	}
       // close cursor in sourceArray[i]
       if (sourceArray[i]->CloseCursor(Csr) != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADSRC, "init cursor failed");
	 return XXX_NOT_OK;
	} // if close cursor is not OK
      } // while (1)
    } // if (sourceArray[i])
  } // for each source
 return XXX_OK;
}

XxxErrCode RelDataEngine::runStats (const char *RelName)
{
 if (XXX_THISRELENV()->DebugStats) {
   cerr << "Running stats on relation " << RelName << endl;
 }

 // Get Catalog
 RelCatalogInfo* CatalogInfo = new RelCatalogInfo(RelName);
 XxxBool MIFound = XXX_FALSE;
 
 if ((GetCatInfo(RelName, *CatalogInfo, MIFound) != XXX_OK) ||
     (MIFound != XXX_TRUE)) {
   REL_NOTE_ERROR(REL_BADTABLE, "Cannot find catalog for relation");
   return XXX_NOT_OK;
  }
 
 // Get relation 
 XxxRelation* Rel;
 if (GetRelation(RelName, Rel, CatalogInfo->ImplInfo()) != XXX_TRUE) {
   char buf[100];
   sprintf(buf, "Cannot find relation %s", RelName);
   REL_NOTE_ERROR(REL_BADTABLE, buf);
   return XXX_NOT_OK;
  }
 Rel->CatalogInfo = CatalogInfo;

 // Set up a scan over the relation
 XxxSetCursor RelCsr;  XxxRecord* RelRecord;
 XxxRecordSchema* RecSchema = CatalogInfo->RecordSchema();
 int Arity = RecSchema->Arity();
 if (!CatalogInfo->StatsInfo()) {
   // allocate memory for statistics collection
   CatalogInfo->statsInfo = new RelStatsInfo(Arity);
  }
 RelStatsInfo* StatsInfo = CatalogInfo->StatsInfo();

 if (StatsInfo->ResetStatsInfo(RecSchema) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADCAT, "Unable to allocate stats info");
   return XXX_NOT_OK;
  }

 // Initialize Cursor
 if (Rel->InitCursor(RelCsr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Unable to open cursor");
   return XXX_NOT_OK;
  }

  // Iterate thru relation; See one record at a time.  
 while ((Rel->NextItem(RelCsr, RelRecord) == XXX_OK) && 
	(RelCsr.IsValid() == XXX_TRUE)) {
   if (StatsInfo->AccumulateStats(RelRecord, RecSchema) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADCAT, "Error accumulating stats");
     return XXX_NOT_OK;
    }
  }

 // Close Cursor
 if (Rel->CloseCursor(RelCsr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Unable to close cursor");
   return XXX_NOT_OK;
  }

 if (StatsInfo->TerminateStats(RecSchema) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADCAT, "Error terminating stats to disk");
   return XXX_NOT_OK;
  }

 if (XXX_THISRELENV()->DebugStats) {
   cerr << "Stats collected for " << RelName << endl;
 }

 // make sure stats are written to the _STATS and _SASTATS tables on disk
 int SourceIndex = CatalogInfo->ImplInfo()->SourceIndex;
 if (!sourceArray[SourceIndex] ||
     sourceArray[SourceIndex]->ModifyStatsInfo(CatalogInfo, 
					       StatsInfo) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADCAT, "Error propagating stats to disk");
   return XXX_NOT_OK;
  }

 return XXX_OK;
}
