#ifndef XXX_RELDATA_H
#define XXX_RELDATA_H

#include "gensrc.h"
#include "relconfig.h"
#include "reladt.h"
#include "relcat.h"
#include "relquery.h"

XXX_EXTERN_CLASS_DECL(RelDataEngine);
XXX_EXTERN_CLASS_DECL(XxxRelation);

// There is one source type for each implementation of a relation.
// The source type is used in each relation identifier to specify which 
// data source it belongs to. SrcType 0 is the default src type
// All the source types _MUST_ be listed here
enum RelSrcTypes {
  REL_UNIXASCII_SRC = 0, // ascii text
  REL_SHORE_SRC = 1,     // stored in SHORE
  REL_UNIXBIN_SRC = 2,   // binary
  REL_MEM_SRC = 3,       // stored in main memory
  REL_INVALID_SRC = 4    // invalid  
 };
const int RelNumSrcTypes = 2;

/**
// Each data source subclasses the RelHandle. This is basically an
// identifier of the relation. This is the "value" of a relation which
// resides in a record. i.e. it is like an OID. Each data src will
// subclass this to provide its own specific kind of handle.
//
// NOTE: This can have _NO_ virtual methods, since we want it to
// be simply a struct (no virtual table ptrs) that can be copied into
// and out of a record. Further, we want to make sure that this is
// completely aligned on a word boundary to avoid problems inside records.
//
*/
class RelHandle : public XxxADTValue {
 public:
  RelSrcTypes SrcType;

  RelHandle(RelSrcTypes Type = REL_INVALID_SRC) { SrcType = Type;}
  RelSrcTypes SourceType() const { return SrcType;}
  XxxBool Valid() { return (SrcType != REL_INVALID_SRC);}
 };

/**
 *
 */
class RelNamedHandle {
 public:
  char Name[REL_RELNAME_MAX];
  RelHandle* Hndl;
  RelNamedHandle(const char* RelName, RelHandle* Handle) {
    strncpy(Name, RelName, REL_RELNAME_MAX);
    Name[REL_RELNAME_MAX-1] = 0;
    Hndl = Handle;
   }
  // following are needed by the template Table class
  RelNamedHandle() { Hndl = NULL;}
  int Hash() const { return (Name[0] + Name[1]);}
  int operator==(const RelNamedHandle& SInfo) const
   { return !strcmp(Name,SInfo.Name);}
  void Print(ostream& OStr) const { OStr << Name; }
 };

/**
// Each DataSrc maintains a number of devices. Each device contains
// one or more named relations. The interface of a DataSrc permits one to
// create, mount, unmount and destroy devices, and to get handles
// on individual named relations. The DataSrc is therefore a multiset
// of named relations.
*/

class RelDataSrc : public XxxGenDataSrc,
virtual public XxxMultiSet<RelNamedHandle> {
 protected:
  RelDataEngine* dataEngine;
 public:
  RelSrcTypes SrcType;  // this should be assigned by each subclass

  RelDataSrc(RelDataEngine* DataEngine, const char* Name, int defDevSize):
  XxxGenDataSrc(Name, defDevSize) {
   dataEngine = DataEngine;
  }
  virtual ~RelDataSrc() {};
  int operator==(const RelDataSrc& Other) const { return 0; }

  // well known interface of a data src
  virtual void PrintRels(ostream& OStr);

  // methods that retrieve catalog info
  virtual XxxErrCode GetCatInfo(const char* Name, RelCatalogInfo &,
				XxxBool& Found) = 0;
  virtual XxxErrCode AddCatInfo(const char* RelName,
				const char* DevName,
				const char* RelImplType,
				const RelDataSrc::AttrTypeList* Attrs,
				const RelDataSrc::KeyList* Keys) = 0;
  virtual XxxErrCode ModifyStatsInfo(const RelCatalogInfo* CatInfo,
				     const RelStatsInfo* StatsInfo) = 0;

  virtual XxxErrCode DeleteCatInfo(const char *RelName) = 0;

  // used to get named base relations at query execution startup
  virtual XxxBool GetRelation(const char* Name,
			      XxxRelation*& AnswerRel,
			      const RelImplInfo* ImplInfo) = 0;
  // used to get a handle to a named relation
  virtual XxxBool GetRelation(const char* Name,
			      RelHandle*& Handle) = 0;

  // Used to determine if relation is protected
  virtual XxxBool RelationIsProtected (const char *RelName) = 0;
  // used to get a base relation from a handle
  virtual XxxErrCode GetRelation(const RelHandle* Handle,
				 XxxRelation*& AnswerRel,
				 const RelImplInfo* ImplInfo) = 0;
  virtual XxxErrCode CreateNewRelation(RelHandle*& Handle, 
				       const char* Name,
				       const char* DevName,
				       const char* RelImplType,
				       AttrTypeList* Attrs,
				       KeyList* Keys,
				       XxxBool MakeCatEntry) 
   { return XXX_NOT_OK; }
  virtual XxxErrCode CreateNewRelation(RelHandle*& Handle, 
				       const char* Name,
				       const RelImplInfo* ImplInfo,
				       AttrTypeList* Attrs,
				       KeyList* Keys,
				       XxxBool MakeCatEntry) 
   { return XXX_NOT_OK; }

  virtual XxxErrCode CreateNewIndex (const char *RelName,
				     const char *IndxName,
				     XxxValueExpression *Expr,
				     const XxxBool Unique,
				     const XxxBool Clustered,
				     XxxBool & Success,
				     XxxIndexType IndxType) {
    Success = XXX_FALSE;
    return XXX_OK;
  }
  virtual XxxBool GetRelationOfIndex (const char *IndxName,
				      char *&RelName) const
    { return XXX_NOT_OK; /* default implementation */ } 

  virtual XxxErrCode DeleteIndex (const char *IndxName, const char *RelName) 
    { return XXX_NOT_OK; /* default implementation */ } 
  
  virtual XxxErrCode DeleteTuples (const RelHandle *Handle, 
				   XxxBooleanExpressionPlan *ExprPlan,
				   RelCatalogInfo &CatInfo) = 0;

  virtual XxxErrCode DeleteRelation(const char *RelName,
				    const RelHandle* Handle) = 0;
  virtual XxxErrCode CreateNewView(const char* Name, 
				   const char* ViewInfo) 
  { return XXX_NOT_OK;}

  virtual XxxErrCode DeleteView (const char *ViewName) 
  { return XXX_NOT_OK; }
  virtual XxxErrCode RetrieveView(const char* ViewName, 
				  RelViewClass*& View,
				  XxxBool& Found) const 
  { Found = XXX_FALSE; return XXX_OK; }
  virtual XxxErrCode PrintViews(ostream& OStr) const 
  { return XXX_OK; }
 };

/**
 *  Data engine for relational data
 */
class RelDataEngine {
  // prefix for primary key index name
  static const char* const PKEY_INDX_PREFIX;
  // after how many records loaded should we break and commit
  // to conserve log space?
  static const int LoadCommitThreshold;

  XxxRelationEngine* RelEng;
  RelDataSrc* sourceArray[RelNumSrcTypes];
  void addDataSrc(RelDataSrc* NewSrc);  
  RelSrcTypes defaultSrc;
  // method to collect statistics on the relation
  XxxErrCode runStats(const char* RelName);

  XxxErrCode makePKeyIndex(const char* TableName, 
			   RelDataSrc::AttrTypeList* Attrs,
			   RelDataSrc::KeyList* Keys);

 public:  
  // class used to store update items for Update statement
  class UpdateItemList { public:
    // list of attr names
    XxxValueExprList* AttrList;
    // list of replacement expressions
    XxxValueExprList* ReplList;

    UpdateItemList(const char* Name, XxxValueExpression* VExpr) {
      AttrList = new XxxValueExprList(new XxxUnknownValue(Name));
      ReplList = new XxxValueExprList(VExpr);
     }
    ~UpdateItemList() {
      if (AttrList) AttrList->Destroy();
      if (ReplList) ReplList->Destroy();
    }
    UpdateItemList* Add(const char* Name, XxxValueExpression* VExpr) {
      AttrList = AttrList->AddExpr(new XxxUnknownValue(Name));
      ReplList = ReplList->AddExpr(VExpr);
      return this;
     }
   };

  RelDataEngine(XxxRelationEngine* RelEng);
  virtual ~RelDataEngine();
  void PrintRels(ostream& OStr);
  void PrintCats(ostream& OStr);
  void PrintSchemas(ostream& OStr);
  void PrintDevs(ostream& OStr);
  void PrintSrcs(ostream& OStr);

  // methods that retrieve catalog info
  XxxErrCode GetCatInfo(const char* Name, RelCatalogInfo &, XxxBool& Found);
  XxxErrCode DeleteCatInfo (const char* RelName);
  XxxErrCode AddCatInfo(const char* RelName,
			const  XxxBool Named,
			const RelDataSrc::AttrTypeList* Attrs,
			const RelDataSrc::KeyList* Keys,
			const RelSrcTypes SrcType,
			const char* DevName,
			const char* RelImplType);

  // is it one of the catalog relations (that should not ever be deleted)
  XxxBool RelationIsProtected (const char *RelName);

  // methods that retrieve relations
  XxxBool GetRelation(const char* Name, 
		      XxxRelation*& AnswerRel,
		      const RelImplInfo* ImplInfo);
  XxxBool GetRelation(const char* Name, RelHandle*& Handle);
  XxxErrCode GetRelation(const RelHandle* Handle,
			 XxxRelation*& AnswerRel,
			 const RelImplInfo* ImplInfo);

  // methods that create/delete relations
  XxxErrCode CreateNewRelation(RelHandle*& Handle, 
				       const char* Name,
				       RelSrcTypes SrcType, 
				       const char* DevName,
				       const char* RelImplType,
				       XxxGenDataSrc::AttrTypeList* Attrs,
				       XxxGenDataSrc::KeyList* Keys,
				       XxxBool MakeCatEntry);
  XxxErrCode CreateNewRelation(RelHandle*& Handle, 
				       const char* Name,
				       const RelImplInfo* ImplInfo,
				       XxxGenDataSrc::AttrTypeList* Attrs,
				       XxxGenDataSrc::KeyList* Keys,
				       XxxBool MakeCatEntry);

  virtual XxxErrCode CreateNewIndex (const char *RelName,
				     const char *IndxName,
				     XxxValueExpression *Expr,
				     const XxxBool Unique,
				     const XxxBool Clustered,
				     const char *IndxType);

  virtual XxxErrCode DeleteIndex (const char *IndxName);

  virtual XxxErrCode DeleteRelation(const char *RelName);

  virtual XxxErrCode LoadRelation(const char* RelName, const char* FileName);
  virtual XxxErrCode LoadRelation(const char* RelName, istream& InStream, 
				  const RelCatalogInfo* CatInfo, 
				  RelHandle*& Handle, 
				  const XxxBool ToCreate);

  XxxErrCode CreateNewView(const char* Name, const char* ViewInfo);
  virtual XxxErrCode DeleteView (const char *ViewName);
  XxxErrCode RetrieveView(const char* ViewName, 
			  RelViewClass*& View,
			  XxxBool& Found) const;
  XxxErrCode PrintViews(ostream& OStr) const;

  // method to collect statistics on the relation
  XxxErrCode RunStats(const char* RelName);

  // methods that manipulate devices
  virtual XxxErrCode CreateDevice(char*& DeviceName, const int DeviceSize,
				  const RelSrcTypes SrcType) {
    return sourceArray[SrcType]->CreateDevice(DeviceName, DeviceSize);
   }
  virtual XxxErrCode MountDevice(const char* DeviceName,
				 const RelSrcTypes SrcType) {
    return sourceArray[SrcType]->MountDevice(DeviceName);
   }
  virtual XxxErrCode UnMountDevice(const char* DeviceName,
				   const RelSrcTypes SrcType) {
    return sourceArray[SrcType]->UnMountDevice(DeviceName);
   }
  virtual XxxErrCode DestroyDevice(const char* DeviceName,
				   const RelSrcTypes SrcType) {
    return sourceArray[SrcType]->DestroyDevice(DeviceName);
   }

  int GetHandleSize(const RelHandle* Rel) const;
  // this may increment refcnts
  int CopyHandle(const RelHandle*SrcHandle, 
		 RelHandle* DstHandle,
		 const RelCatalogInfo* CatInfo) const;
  // this may decrement refcnts
  void RemoveHandle(const RelHandle* Rel,
		    const RelCatalogInfo* CatInfo) const;

  RelSrcTypes FindSrcType(const char* SrcName) const;
  RelSrcTypes DefaultSrcType() const { return defaultSrc; }
  int DefaultDevSize() const { 
    RelDataSrc* DefSrc = sourceArray[DefaultSrcType()];
    return (DefSrc?DefSrc->DefaultDevSize():-1);
   }
  int DevSize(RelSrcTypes SrcType) const { 
    RelDataSrc* Src = sourceArray[SrcType];
    return (Src?Src->DefaultDevSize():-1);
   }
  XxxBool IsMounted(const char* DevName, RelSrcTypes SrcType) const {
    return sourceArray[SrcType]->IsMounted(DevName);
   }

  XxxErrCode InsertTuples ( const char *RelName, XxxValueExprList * ValueList);
  // XxxErrCode InsertQueryBlock ( const char *RelName, RelViewClass * QueryBlock);

  /// process a delete statement
  XxxErrCode DeleteTuples(const char *RelName,
			  XxxBooleanExpression *BoolExpr);

  /// process an update statement
  XxxErrCode UpdateTuples(const char *RelName,
			  XxxBooleanExpression *BoolExpr,
			  const UpdateItemList* UpdateList);
};

#endif  /* XXX_RELDATA_H */
