#ifndef REL_SHORERELDEV_H
#define REL_SHORERELDEV_H

//
// This is included from within shorerel.h.
// RelShoreDevice is a nested class within RelShoreDataSrc
//

/**
 // A SHORE "device".
 // Each volume is a directory. It maintains a table of named relations.
 // It also maintains an index between the relation id and the relation
 // handle.
 */
class RelShoreDevice : public XxxMultiSet<RelNamedHandle> {
  friend class RelShoreCatalog;

  // these methods define the schemas of each of
  // the basic catalog tables
  static XxxRecordSchema* getInitSchema(const char* CatName);
  static int getSchemaArity(const char* CatName);

  static XxxErrCode createCatalogs(const XxxShoreVolumeId VolId,
				   const XxxShoreObjectId& CatalogIndex);

 public:
  struct tablesInfo {
    const char* TableName;
    int TableArity;
    int ImplType;
   };
  static struct tablesInfo TablesInfo[];
  static const int numTablesInfos;

  struct attrsInfo {
    const char* TableName;
    const char* AttrName;
    int IsKey;
    int TypeId;
    int AttrIndex;
    const XxxADTClass::MetaInfoHandle* AttrMIHndl;
   };
  static struct attrsInfo AttrsInfo[];
  static const int numAttrsInfos;

  struct indxsInfo {
    const char* TableName;
    const char* IndxName;
    const char* IndxString;
    XxxIndexType IndxType;
    int IsUnique;
    int IsClustered;
   };
  static struct indxsInfo IndxsInfo[];
  static const int numIndxsInfos;

  struct statsInfo {
    const char* TableName;
    int Card;
    int TupleSize;
   };
  static struct statsInfo StatsInfo[];
  static const int numStatsInfos;


  struct viewsInfo {
    const char* ViewName;
	const char* ViewInfo;	
  };
  static struct viewsInfo ViewsInfo[];
  static const int numViewsInfos;

  struct astatsInfo {
    const char* TableName;
    int AttrIndex;
    int TypeId;
    const XxxADTMetaInfo* MetaInfo;
    const XxxADTClass::TypeStats* Data;
   };

 private:
  // this is a constructor, moved out of the class to allow
  // it to be viewed as a pure struct
  static tablesInfo *newTablesInfo(const char* Name, 
				   const int Arity, const int IType){
    tablesInfo* TInfo = new tablesInfo();
    TInfo->TableName = Name;
    TInfo->TableArity = Arity;
    TInfo->ImplType = IType;
    return TInfo;
   }
  static XxxErrCode createTablesCat(const XxxShoreVolumeId VolId,
				    const XxxShoreObjectId& CatalogIndex);
  static XxxErrCode createTablesEntry(const XxxShoreVolumeId VolId,
				      XxxShoreObjectId& TablesFile,
				      XxxShoreObjectId& TablesIndex,
				      const XxxRecordSchema* RecSchema,
				      XxxValuesArray& ValsTab,
				      struct tablesInfo& TablesInfo);
  static XxxErrCode makeTablesRec(const XxxRecordSchema* RecSchema,
				  XxxValuesArray& ValsTab,
				  const struct tablesInfo& TablesInfo,
				  XxxRecord*& ReturnRec);

  // this is a constructor, moved out of the class to allow
  // it to be viewed as a pure struct
  static attrsInfo *newAttrsInfo(const char* TableName,
				 const char* AttrName,
				 const int IsKey,
				 const int TypeId,
				 const int AttrIndex,
				 const XxxADTClass::MetaInfoHandle* AttrMIHndl) {
    attrsInfo* AInfo = new attrsInfo();
    AInfo->TableName = TableName;
    AInfo->AttrName = AttrName;
    AInfo->IsKey = IsKey;
    AInfo->TypeId = TypeId;
    AInfo->AttrIndex = AttrIndex;
    AInfo->AttrMIHndl = AttrMIHndl;
    return AInfo;
   }
  
  static XxxErrCode createAttrsCat(const XxxShoreVolumeId VolId,
				   const XxxShoreObjectId& CatalogIndex);
  static XxxErrCode createAttrsEntry(const XxxShoreVolumeId VolId,
				     XxxShoreObjectId& AttrsFile,
				     XxxShoreObjectId& AttrsIndex,
				     const XxxRecordSchema* RecSchema,
				     XxxValuesArray& ValsTab,
				     struct attrsInfo& AttrsInfo);
  static XxxErrCode makeAttrsRec(const XxxRecordSchema* RecSchema,
				 XxxValuesArray& ValsTab,
				 const struct attrsInfo& AttrsInfo,
				 XxxRecord*& ReturnRec);


  static XxxErrCode createIndxsCat(const XxxShoreVolumeId VolId,
				   const XxxShoreObjectId& CatalogIndex);
  static XxxErrCode createIndxsEntry(const XxxShoreVolumeId VolId,
				     XxxShoreObjectId& IndxsFile,
				     XxxShoreObjectId& IndxsTableIndex,
				     XxxShoreObjectId& IndxsNameIndex,
				     const XxxRecordSchema* RecSchema,
				     XxxValuesArray& ValsTab,
				     struct indxsInfo& IndxInfo);

  static statsInfo *newStatsInfo(const char* TableName,
				 int Card,
				 int TupleSize) {
    statsInfo* SInfo = new statsInfo();
    SInfo->TableName = TableName;
    SInfo->Card = Card;
    SInfo->TupleSize = TupleSize;
    return SInfo;
   }
  static XxxErrCode createStatsCat(const XxxShoreVolumeId VolId,
				   const XxxShoreObjectId& CatalogIndex);
  static XxxErrCode createStatsEntry(const XxxShoreVolumeId VolId,
				     XxxShoreObjectId& StatsFile,
				     XxxShoreObjectId& StatsIndex,
				     const XxxRecordSchema* RecSchema,
				     XxxValuesArray& ValsTab,
				     struct statsInfo& StatsInfo);
  static XxxErrCode makeStatsRec(const XxxRecordSchema* RecSchema,
				 XxxValuesArray& ValsTab,
				 const struct statsInfo& StatsInfo,
				 XxxRecord*& ReturnRec);

  static viewsInfo *newViewsInfo(const char* ViewName,
					const char* ViewInfo) {
    viewsInfo* VInfo = new viewsInfo();
    VInfo->ViewName = ViewName;
	VInfo->ViewInfo = ViewInfo;
    return VInfo;
   }
  static XxxErrCode createViewsCat(const XxxShoreVolumeId VolId,
				   const XxxShoreObjectId& CatalogIndex);
  static XxxErrCode createViewsEntry(const XxxShoreVolumeId VolId,
				     XxxShoreObjectId& ViewsFile,
				     XxxShoreObjectId& ViewsIndex,
				     const XxxRecordSchema* RecSchema,
				     XxxValuesArray& ValsTab,
				     struct viewsInfo& ViewsInfo);
  static XxxErrCode deleteViewsEntry(const XxxShoreVolumeId VolId,
				 XxxShoreObjectId& ViewsFile,
				 XxxShoreObjectId& ViewsIndex,
				 const XxxRecordSchema* RecSchema,
				 XxxValuesArray& ValsTab,
				 struct viewsInfo& ViewsInfo);

  static XxxErrCode makeViewsRec(const XxxRecordSchema* RecSchema,
				 XxxValuesArray& ValsTab,
				 const struct viewsInfo& ViewsInfo,
				 XxxRecord*& ReturnRec);

  static astatsInfo *newAStatsInfo(const char* TableName,
				   int AttrIndex,
				   int TypeId,
				   const XxxADTMetaInfo* MetaInfo,
				   const XxxADTClass::TypeStats* DataValue) {
    astatsInfo* ASInfo = new astatsInfo();
    ASInfo->TableName = TableName;
    ASInfo->AttrIndex = AttrIndex;
    ASInfo->TypeId = TypeId;
    ASInfo->MetaInfo = MetaInfo;
    ASInfo->Data = DataValue;
    return ASInfo;
   }
  static XxxErrCode createAttrStatsCat(const XxxShoreVolumeId VolId,
				       const XxxShoreObjectId& CatalogIndex);
  static XxxErrCode makeAStatsRec(const XxxRecordSchema* RecSchema,
				  XxxValuesArray& ValsTab,
				  const astatsInfo& AStatsInfo,
				  XxxRecord*& ReturnRec);


  virtual XxxErrCode addItem(const RelNamedHandle & item);
 public:
  static const char *const RELROOTCAT;
  static const char *const RELTABLESCAT;
  static const char *const RELTABINDXCAT;
  static const char *const RELATTRSCAT;
  static const char *const RELATTRSINDXCAT;
  static const char *const RELSTATSCAT;
  static const char *const RELSTATSINDXCAT;
  static const char *const RELVIEWSCAT;
  static const char *const RELVIEWSINDXCAT;
  static const char *const RELATTRSTATSCAT;
  static const char *const RELATTRSTATSINDXCAT;
  static const char *const RELINDXSCAT;
  static const char *const RELINDXSINDXCAT1;
  static const char *const RELINDXSINDXCAT2;
  static const char *const RELSUPERFILE;

  XxxBool Valid;
  char* DevName;
  XxxShoreVolumeId VolumeId;
  XxxShoreObjectId RootIndex;  // index over the tables and indexes 
  // in the device
  XxxShoreObjectId SuperFile;  // File for all objects that dont have
  // a file to put 'em in.
  RelShoreCatalog* Catalog;
  int operator==(const RelShoreDevice& Other) const 
   { return !strcmp(DevName, Other.DevName); }

  XxxErrCode GetCatInfo(const char* Name, RelCatalogInfo &Cat,
			XxxBool& Found) {
    if (Catalog)
      return Catalog->GetCatInfo(Name, Cat, Found);
    else {
      Found = XXX_FALSE;
      return XXX_OK;
     }
   }
  XxxErrCode AddCatInfo(const char* RelName,
			const RelDataSrc::AttrTypeList* Attrs,
			const RelDataSrc::KeyList* Keys,
			const RelShoreImplType ImplType) {
    return (Catalog?
	    Catalog->AddCatInfo(RelName, Attrs, Keys, ImplType):
	    XXX_NOT_OK);
   }
  XxxErrCode ModifyStatsInfo(const RelCatalogInfo* CatInfo,
			     const RelStatsInfo* StatsInfo) {
    return (Catalog? Catalog->ModifyStatsInfo(CatInfo, StatsInfo): XXX_NOT_OK);
   }

  XxxErrCode DeleteCatInfo(const char* Name) {
    return (Catalog ?
	    Catalog->DeleteCatInfo(Name) : XXX_NOT_OK);
   }

  
  XxxBool GetRelation(const char* Name, 
		      RelHandle*& Handle) const;

  XxxBool GetRelationOfIndex (const char *IndxName,
			      char *&RelName) const;
  XxxErrCode GetRelation(const RelShoreHandle* Handle,
			 XxxRelation*& AnswerRel,
			 const RelImplInfo* ImplInfo);

  // if Name == NULL, it is an unnamed relation otherwise it is named
  XxxErrCode CreateNewRelation(RelShoreHandle*& Handle,
			       RelShoreImplType ImplType,
			       const char* Name,
			       RelDataSrc::AttrTypeList* Attrs,
			       RelDataSrc::KeyList* Keys,
			       XxxBool MakeCatEntry);
  XxxErrCode CreateNewIndex (const char *RelName,
			     const char *IndxName,
			     XxxValueExpression * Expr,
			     const XxxBool Unique,
			     const XxxBool Clustered,
			     XxxBool & Success,
			     XxxIndexType IndxType);
  XxxErrCode DeleteIndex (const char *IndxName, const char *RelName,
			  const RelShoreHandle *Handle);

  XxxErrCode DeleteRelation(const char *RelName,
			    const RelShoreHandle* Handle);
  XxxErrCode DeleteRelation(const char *RelName,
			    const RelShoreHandle* Handle,
			    const RelCatalogInfo* CatInfo);

  XxxErrCode CreateNewView(const char* Name, 
			   const char* ViewInfo);
  XxxErrCode DeleteView(const char *Name);
  XxxErrCode RetrieveView(const char* ViewName, 
			  RelViewClass*& View,
			  XxxBool& Found) const;
  XxxErrCode PrintViews(ostream& OStr) const;

  static XxxErrCode FormatDevice(const XxxShoreVolumeId& VolId);
  RelShoreDevice(const char *Name);
  // actually destroy the underlying device
  XxxErrCode DestroyDev();
  ~RelShoreDevice();

  // to iterate over the RelNamedHandles
  virtual XxxErrCode InitCursor(XxxSetCursor&) const;
  virtual XxxErrCode NextItem(XxxSetCursor&, RelNamedHandle&) const;
  virtual XxxErrCode CloseCursor(XxxSetCursor&) const;
  virtual XxxErrCode RemoveItem(const RelNamedHandle & item);
 };

#endif
