/**
 This file defines the following five classes which are used to
 define methods of ADTs
   XxxFuncParseInfo : result of parsing a function
   XxxFuncPlanInfo  : result of planning a function

   XxxFuncMethodInfo : descriptor of scalar function
   XxxAggrMethodInfo : descriptor of aggregate function
**/

#ifndef XXX_ADTFUNC_H
#define XXX_ADTFUNC_H

#include "types.h"
#include "generic.h"

XXX_EXTERN_CLASS_DECL(XxxAggrOpHandle);

/**
 // Class that can be used for a generic function/method parse 
 // data structure.
 */
class XxxFuncParseInfo : public XxxGenericParse {
 protected:
  XxxBool* preOptimizeArray;
  int NumArgs;
 public:
  XxxBasicType TypeId;     // to which E-ADT does this fn belong?
  int EngineIndex;         // which engine does this fn reference
  // either TypeId or EngineIndex will be meaningful: not both
  int MethodIndex;
  XxxBool IsAggr;
  XxxADTMetaInfo* TypeCatInfo;

  XxxBool IsTopLevel;
  XxxBasicType ReturnType; 
  char FuncName[XXX_FUNCNAME_MAX];

  XxxGenericParse* ExprParse; // any other parse data structure

  /// constructor used by QPEs
  XxxFuncParseInfo(const char* Name,
		   XxxGenericParse* Exp,
		   int Index,
		   XxxBasicType RetType);
  /// constructor used by ADTs
  XxxFuncParseInfo(const char* Name,
		   int MethIndex,
		   XxxGenericParse* ParseData, 
		   XxxBasicType ThisTypeId,
		   XxxBasicType RetType,
		   XxxADTMetaInfo* CatInfo = NULL);
  XxxFuncParseInfo(const XxxFuncParseInfo& src);  
  virtual XxxGenericParse* Replicate() const {
    return new XxxFuncParseInfo(*this);
  }

  virtual ~XxxFuncParseInfo()
   { if (ExprParse) delete ExprParse;}

  virtual XxxADTMetaInfo* GenerateCatInfo() const {
    if (TypeCatInfo) return TypeCatInfo;
    else if (ExprParse) return ExprParse->GenerateCatInfo();
    else return NULL;
   }
  virtual XxxBasicType GenerateType() const {
    if (ExprParse && (ReturnType == XXX_UNKNOWN)) 
      return ExprParse->GenerateType();
    else return ReturnType;
   }
 };

XXX_EXTERN_CLASS_DECL(XxxFuncResultCache);
XXX_EXTERN_CLASS_DECL(XxxFuncScratchPad);

/**
 * class that represents the execution plan for a function call
 */
class XxxFuncPlanInfo : public XxxGenericPlan {
  friend class XxxADTClass;
  friend class XxxAggrExpression; // HACK: TODO: needs to be removed
 protected:
  int NumArgs;
  PlanUsage  planUsage;
  XxxBool    compressOutput;
  XxxBool*   decompressInput;
  XxxBool    decompressOwner;
  XxxBool    suspendComputation;
  
 public:
  XxxBasicType TypeId;   // id of type to which the function belongs
  int MethodIndex;       // index of method within the type

  int EngineIndex;       // in case this represents an engine, not a type

  char FuncName[XXX_FUNCNAME_MAX];

  XxxGenericPlan* ExprPlan; // additional plan info (usually for engines)

  //// scratch area for functions/methods
  XxxFuncScratchPad* ScratchPad;
  //// result cache for functions/methods
  XxxFuncResultCache* ResultCache;

  XxxFuncPlanInfo(const XxxFuncPlanInfo &FPI);
  XxxFuncPlanInfo(const char* Name,
		  XxxBasicType Type,
		  int EngIndex,
		  XxxGenericPlan* Exp);
  XxxFuncPlanInfo(const XxxFuncParseInfo* ParseInfo,
		  int ArgCount,
		  XxxGenericPlan* Plan = NULL);
  virtual ~XxxFuncPlanInfo();
  virtual XxxGenericPlan *Replicate() const {    
    return new XxxFuncPlanInfo(*this);
  }

  virtual XxxErrCode UsageOfPlan(PlanUsage& Usage) const
  { Usage = planUsage; return XXX_OK;}
  virtual const PlanUsage& GetPlanUsage() { return planUsage;}
  // TODO: i tried to keep this protected, but it rather a pain
  virtual PlanUsage& getPlanUsage() { return planUsage;}

  XxxErrCode modifyArgCompression(XxxValueExprPlan* OwnerPlan,
				  XxxValExprPlanList* ArgPlans);


  // member functions for suspending computation
  XxxErrCode SuspendComputation();
  XxxErrCode UnsuspendComputation();
  XxxBool IsSuspended() { return suspendComputation; }

  /// Evaluate this expression
  virtual XxxErrCode Evaluate(XxxValueEnv* Env,
			      XxxFuncMethodInfo* ThisMethInfo,
			      XxxADTValue *ReturnValue) const;
  /**
   * Creates a suspended computation encapsulating the execution of this 
   * plan. The suspended computation can be stored and later executed.
   */
  virtual XxxErrCode SuspendEvaluation(XxxValueEnv* Env, 
				       XxxFuncMethodInfo* ThisMethInfo,
				       XxxADTValue *ReturnValue) const;

  virtual XxxErrCode AggrInit(XxxAggrMethodInfo* MethInfo,
			      XxxAggrOpHandle *Handle) const;
  virtual XxxErrCode AggrTerm(XxxAggrMethodInfo* MethInfo,
			      XxxAggrOpHandle *Handle,
			      XxxADTValue* FinalVal) const;
  virtual XxxErrCode AggrAccum(XxxAggrMethodInfo* MethInfo,
			       XxxValueEnv* Env,
			       XxxAggrOpHandle *Handle) const;
  virtual XxxErrCode AggrDrop(XxxAggrMethodInfo* MethInfo,
			      XxxValueEnv* Env, 
			      XxxAggrOpHandle *Handle) const;
 };


/**
 * Information on each method of an ADT
 */
class XxxMethodInfo { 
 public:
  char           Name[XXX_FUNCNAME_MAX];
  int            NumArgs;
  XxxBasicType   ReturnType;
  XxxADTClass::MetaInfoGen    MIGen;

  // For optimizations, if OptFunc is specified, it takes precedence.
  // Else "Usage" defines the cost (by default, 0)
  XxxADTClass::OptimizeFunc   OptFunc;
  XxxGenericPlan::PlanUsage Usage;

  /// properties of the method (not used so far)
  XxxBool SideEffectFree;

  XxxMethodInfo(const char* FName, int Arity, XxxBasicType RetType,
		XxxADTClass::MetaInfoGen MIG, 
		XxxADTClass::OptimizeFunc OF,
		XxxBool SEFree = XXX_TRUE) 
    : NumArgs(Arity), ReturnType(RetType), MIGen(MIG), 
      OptFunc(OF), SideEffectFree(SEFree) {
      strncpy(Name, FName, XXX_FUNCNAME_MAX);
      Name[XXX_FUNCNAME_MAX-1] = 0;
  }
  XxxMethodInfo(const XxxMethodInfo &MInfo);
  virtual ~XxxMethodInfo() {}
  virtual XxxMethodInfo * Replicate() const = 0;
};

/**
 * Information on each function method of an ADT
 */
class XxxFuncMethodInfo : public XxxMethodInfo { 
public:
  XxxADTClass::ExecuteFunc ExecFunc;

  //// A class that captures a variable-length argument type list
  class ArgInfo {
  public:
    int Count;
    // TODO: also need to keep meta-info for these types
    XxxBasicType ArgTypes[XXX_ARGCNT_MAX];
    ArgInfo(int Arity, ...);
    ArgInfo(const ArgInfo& AInfo);
    ArgInfo() { Count = 0;}
  };
  ArgInfo ArgumentTypes;

  // some common options for generating meta-info for the result
  typedef enum MIGenOption {
    MIGEN_UNKNOWN = 0,
    MIGEN_OWNER = 1,
    MIGEN_ARG0 = 2,
    MIGEN_ARG1 = 3,
    MIGEN_ARG2 = 4,
    MIGEN_ARG3 = 5,
  } MIGenOption;

  /*** a variety of constructors for XxxFuncMethodInfo ***/
  //// maintained for backwards compatibility
  XxxFuncMethodInfo(const char* FName, 
		    XxxBasicType RetType, 
		    XxxMetaInfoGen MIG, 
		    XxxOptimizeFunc OF, 
		    XxxExecuteFunc EF,
		    int Arity, ...);
  //// constructor that allows MIG and OF to be NULL
  XxxFuncMethodInfo(const char* FName, 
		    XxxBasicType RetType, 
		    ArgInfo& TypeSignature,
		    XxxExecuteFunc EF,  
		    XxxMetaInfoGen MIG = 0, 
		    XxxOptimizeFunc OF = 0);
  //// constructor that allows MIG to be NULL
  XxxFuncMethodInfo(const char* FName, 
		    XxxBasicType RetType, 
		    ArgInfo& TypeSignature,
		    XxxExecuteFunc EF,  
		    XxxOptimizeFunc OF,
		    XxxMetaInfoGen MIG = 0);
  //// constructor that specifies the plan usage (i.e. cost).
  XxxFuncMethodInfo(const char* FName, 
		    XxxBasicType RetType, 
		    ArgInfo& TypeSignature,
		    XxxExecuteFunc EF,  
		    const XxxGenericPlan::PlanUsage& Usage,
		    XxxMetaInfoGen MIG = 0);
  //// constructor that specifies the plan usage and MIGenOption
  XxxFuncMethodInfo(const char* FName, 
		    XxxBasicType RetType, 
		    ArgInfo& TypeSignature,
		    XxxExecuteFunc EF,  
		    const XxxGenericPlan::PlanUsage& Usage,
		    MIGenOption MIGenChoice);

  //// copy constructor
  XxxFuncMethodInfo(const XxxFuncMethodInfo &FMInfo);

  virtual XxxMethodInfo *Replicate() const {
    return new XxxFuncMethodInfo(*this);
  }
  // just match the method name, arg types and arity
  virtual XxxErrCode Matches(const char* FuncName,
			     const XxxValueExpression* Owner,
			     const XxxValueExprList *ArgList,
			     XxxBasicType& ReturnType,
			     XxxADTMetaInfo*& ReturnMetaInfo,
			     XxxBool& Success) const;

  // this is used to discern client side UDFs while optimizing (Tobias)
 //mwm1
  virtual XxxBool IsClientSiteExecutable()
  { return XXX_FALSE; }
  /*
  virtual boolean IsClientSiteExecutable()
  { return false; }
  */
  virtual XxxBool IsJavaMethod()
  {return XXX_FALSE;}

  // this is used to identify asynchrounous device UDFs (philippe)
  virtual XxxBool IsDeviceUDF()
    {return XXX_FALSE;}

};

/**
 * Information on each aggregate method of an ADT
 */
class XxxAggrMethodInfo : public XxxFuncMethodInfo { 
public:
  XxxADTClass::AggrInitFunc InitFunc;
  XxxADTClass::AggrAccumFunc AccumFunc;
  XxxADTClass::AggrTermFunc TermFunc;
  XxxADTClass::AggrDropFunc DropFunc;

  XxxAggrMethodInfo(const XxxAggrMethodInfo &AMI);

  XxxAggrMethodInfo(const char* FName, 
		    XxxBasicType RetType,
		    ArgInfo& TypeSignature,
		    XxxADTClass::MetaInfoGen MIG, 
		    XxxADTClass::OptimizeFunc OF,
		    XxxADTClass::AggrInitFunc IFunc,
		    XxxADTClass::AggrAccumFunc AFunc, 
		    XxxADTClass::AggrTermFunc TFunc,
		    XxxADTClass::AggrDropFunc DFunc);

  virtual XxxMethodInfo * Replicate() const {
    return new XxxAggrMethodInfo(*this);
  }
};


#endif
