#include <stdio.h>
#include <string.h>
#include <math.h>
#include "types.h"
#include "basictypes.h"
#include "stringtypes.h"
#include "xxx_thread.h"
#include "strstream.h"

//
// Methods of XxxCoreADTClass
//
const XxxCoreADTClass::IntValue XxxCoreADTClass::XXX_DEFAULT_INT(0);
const XxxCoreADTClass::DblValue XxxCoreADTClass::XXX_DEFAULT_DBL(0.0);
const XxxCoreADTClass::BoolValue XxxCoreADTClass::XXX_DEFAULT_BOOL(XXX_FALSE);

void XxxCoreADTClass::RegisterCoreADTs()
{
  // the core system ADTs
  new XxxIntADT();
  new XxxDoubleADT();
  new XxxStringADTClass();
  new XxxFixedStringADT();
  new XxxBooleanADT();
  new XxxMIBlobADT();
}

XxxErrCode XxxCoreADTClass::DefaultVal(XxxADTValue* Object, 
				       const XxxADTMetaInfo* MetaInfo) const 
{ 
  switch(TypeId) {
  case XXX_INT:
    memcpy((char *)Object, (char *)(&XXX_DEFAULT_INT), sizeof(IntValue));
    break;
  case XXX_DOUBLE:
    memcpy((char*)Object, (char*)(&XXX_DEFAULT_DBL), sizeof(DblValue));
    break;
  case XXX_BOOLEAN:
    memcpy((char*)Object, (char*)(&XXX_DEFAULT_BOOL), sizeof(BoolValue));
    break;
  case XXX_STRING:
    strcpy(((StrValue *)Object)->Value, "");
    break;
  case XXX_FIXEDSTRING:
    strcpy(((StrValue *)Object)->Value, "");
    break;
  default:
    // TODO: why not: why cant each adt define a default
    XXX_NOTE_ERROR(XXX_BADVALUE, "No default for this type");
    return XXX_NOT_OK;
  }
  
  // essentially override this by setting the NULL value bit
  Object->Set(XxxADTValue::NULL_VALUE);
  return XXX_OK; 
}

XxxBool 
XxxCoreADTClass::CanCastFrom(XxxBasicType SourceType, 
			     const XxxADTMetaInfo* SourceMetaInfo,
			     const XxxADTMetaInfo* TargetMetaInfo)
{
  // A string can be produced from any ADT, since they can simply
  // write-text
  if (TypeId == XXX_STRING)
    return XXX_TRUE;
  // can always cast to a fixed length string by truncation
  if ((TypeId == XXX_FIXEDSTRING) && (SourceType == XXX_STRING))
    return XXX_TRUE;

  return XXX_FALSE;
}

XxxErrCode 
XxxCoreADTClass::CastFrom(XxxBasicType SourceType, 
			  const XxxADTMetaInfo* SourceMetaInfo,
			  const XxxADTMetaInfo* TargetMetaInfo,
			  XxxADTValue* OldObject, 
			  XxxADTValue* NewObject)
{
  if (!((TypeId==XXX_STRING) || 
	((TypeId==XXX_FIXEDSTRING) && (SourceType==XXX_STRING)))) {
    XXX_NOTE_ERROR(XXX_BADADT, "Cannot cast from the specified type");
    return XXX_OK;
  }
  ASSERT(NewObject);
  NewObject->Init();

  ostrstream OStr;
  if (XxxADTArray[SourceType]->WriteText(OStr, OldObject, 
					 SourceMetaInfo) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADADT, "Cannot write text");
    return XXX_NOT_OK;
  }
  OStr << '\0';
  OStr.flush();

  if ((TypeId==XXX_FIXEDSTRING) && (SourceType==XXX_STRING)) 
    strncpy(((StrValue *)NewObject)->Value, OStr.str(),
	    ((XxxFixedStringADT::FStringMetaInfo*)TargetMetaInfo)->Length);
  else 
    strcpy(((StrValue *)NewObject)->Value, OStr.str());
  return XXX_OK;
}

XxxBool 
XxxCoreADTClass::CanCastTo(XxxBasicType TargetType, 
			   const XxxADTMetaInfo* SourceMetaInfo,
			   const XxxADTMetaInfo* TargetMetaInfo)
{
 XxxBool ReturnVal = XXX_FALSE;
 // put in the known type conversions for our special-case types
 switch (TypeId) {
  case XXX_INT:
   if ((TargetType == XXX_DOUBLE) || (TargetType == XXX_INT))
     ReturnVal = XXX_TRUE;
   break;
  case XXX_DOUBLE:
   if ((TargetType == XXX_INT) || (TargetType == XXX_DOUBLE))
     ReturnVal = XXX_TRUE;
   break;
  case XXX_STRING:
   // Add a cast to every other data type, so that arbitrary
   // data can be passed in some user-defined format to be
   // read using ReadText
   ReturnVal = XXX_TRUE;
   break;
  }
 return ReturnVal;
}

XxxErrCode 
XxxCoreADTClass::CastTo(XxxBasicType TargetType, 
			const XxxADTMetaInfo* SourceMetaInfo,
			const XxxADTMetaInfo* TargetMetaInfo,
			XxxADTValue* OldObject, 
			XxxADTValue* NewObject)
{
  // dummy case where types are the same
  if (TargetType == TypeId) {
    STypeCopy(TypeId, OldObject, NewObject, SourceMetaInfo);
    return XXX_OK;
  }
  
  // TODO: make sure to copy the header 4 bytes : 
  // put in the known type conversions for our special-case types
  switch (TypeId) {
  case XXX_INT:
    if (TargetType == XXX_DOUBLE)
      ((DblValue *)NewObject)->Value = 
	(double)(((IntValue *)OldObject)->Value);
    return XXX_OK;
  case XXX_DOUBLE:
    // NOTE: VC++ and g++ seem to have different behavior in
    // the result of (int)(92.5) -- VC++ says 93, while g++ says 92.
    // To be consistent, always round down
    if (TargetType == XXX_INT)
      ((IntValue *)NewObject)->Value = 
	(int)(floor(((DblValue *)OldObject)->Value));
    return XXX_OK;
  case XXX_STRING:
    istrstream OrigDataStr(((StrValue *)OldObject)->Value);
    return XxxADTArray[TargetType]->ReadText(OrigDataStr, 
					     NewObject, TargetMetaInfo);
  }
  XXX_NOTE_ERROR(XXX_BADEXEC, "Unknown type conversion executed");
  return XXX_NOT_OK;
}

XxxErrCode XxxCoreADTClass::
CreateStatsInfo(const XxxADTMetaInfo *MetaInfo,
		TypeStats*& StatsInfo) const
{
 // currently defined for int and double
 switch (TypeId) {
  case XXX_INT:
   StatsInfo = new IntStats;
   break;
  case XXX_DOUBLE:
   StatsInfo = new DblStats;
   break;
  case XXX_STRING:
  case XXX_FIXEDSTRING:
  case XXX_BOOLEAN:
  case XXX_MIBLOB:
   StatsInfo = 0;
  default:
   StatsInfo = 0;
  }
 return XXX_OK;
}

XxxErrCode XxxCoreADTClass::
makeStatsInfo(const XxxADTMetaInfo *MetaInfo,
	      const MetaInfoHandle& MIHandle,
	      TypeStats*& StatsInfo) const
{
 switch (TypeId) {
  case XXX_INT: {
    IntStats* Tmp = (IntStats*)(MIHandle.Data);
    IntStats* Stat = (IntStats*)(StatsInfo);
    *Stat = *Tmp;
    break;
   }
  case XXX_DOUBLE: {
    DblStats* Tmp = (DblStats*)(MIHandle.Data);
    DblStats* Stat = (DblStats*)(StatsInfo);
    *Stat = *Tmp;
    break;
   }
  }
 return XXX_OK;
}

// PURPOSE: Reads string field from InStream, using whitespace as delimiters
XxxErrCode 
XxxCoreADTClass::loadTextNoDelimiter(istream& InStream, 
				     StrValue* Object,
				     int MaxStrLen)
{
 char FirstChar=0,CurrentChar=0;
 int j=0;

 // skip whitespace
 InStream >> FirstChar;
 if (InStream.fail()) return XXX_NOT_OK;
 
 if ((FirstChar != '"') &&
     (FirstChar != '\'')) {
   // Text is not enclosed
   CurrentChar=FirstChar;
   while (!isspace(CurrentChar) && j<MaxStrLen-1) {
     *(Object->Value + j++) = CurrentChar;
     InStream.get(CurrentChar);
      if (InStream.fail()) 
	// The string still need to be null-terminated
	break;
   }
   // If CurrentChar is not a whitespace, overflow has occurred
   if (!isspace(CurrentChar))
     skipRecNonWhiteSpace(InStream);
 }
 else {
   // Text is enclosed
   InStream.get(CurrentChar);
   if (InStream.fail()) return XXX_NOT_OK;
   while (CurrentChar != FirstChar && j<MaxStrLen-1) {
     *(Object->Value + j++) = CurrentChar;
     InStream.get(CurrentChar);
     
     if (InStream.fail()) 
       // The string still need to be null-terminated
       break;
   }
   CurrentChar=InStream.peek();
   // If next char is not a whitespace, overflow has occurred
   if (isspace(CurrentChar)==0)
     skipRecNextField(InStream, FirstChar);
 } // else text is enclosed
 // null terminate the string
 *(Object->Value + j++) = 0;
 return XXX_OK;
}

// PURPOSE: Reads string field from InStream, using Delimiter
XxxErrCode 
XxxCoreADTClass::loadTextDelimited(istream& InStream, 
				   StrValue* Object,
				   char Delimiter,
				   int MaxStrLen)
{
  char FirstChar;

  // ENCLOSED text is enclosed by either quote or backslash
  XxxBool Enclosed = XXX_FALSE;
  
  InStream.get(FirstChar);
  if ((FirstChar != '"') &&
      (FirstChar != '\'')) 
    InStream.putback(FirstChar);
  else
    Enclosed = XXX_TRUE;
  
  char CurrentChar = 0;
  int j = 0;
  InStream.get(CurrentChar);
  if (InStream.fail())
    return XXX_NOT_OK;
  
  while (!InStream.eof() && (CurrentChar != Delimiter) && 
	 (CurrentChar != XXX_RECORD_SEPARATOR) &&
	 !(Enclosed && (CurrentChar==FirstChar)) &&
	 (j<MaxStrLen-1)) {
    *(Object->Value + j++) = CurrentChar;
    InStream.get(CurrentChar);
    if (InStream.fail()) 
      return XXX_NOT_OK;
  } // while
  
  *(Object->Value + j++) = 0;
  
  // Check for fields exceeding MaxStrLen
  if (!(CurrentChar==Delimiter || CurrentChar==XXX_RECORD_SEPARATOR)) {
    skipRecNextField(InStream, Delimiter);
  }
  if (InStream.fail()) return XXX_NOT_OK;
  return XXX_OK;
}

const char XxxCoreADTClass::XXX_RECORD_SEPARATOR = '\n';
// PURPOSE: Moves InStream to next field, stopping at whitespace
void XxxCoreADTClass::skipRecNonWhiteSpace(istream& InStream)
{
 char CurrentChar;
 if (!InStream.eof())
   InStream.get(CurrentChar);
 // Stop at first whitespace
 while(! (isspace(CurrentChar) || InStream.eof() || 
	 (CurrentChar==XXX_RECORD_SEPARATOR))) {
   InStream.get(CurrentChar);
  }
}

// PURPOSE: Moves InStream to next field, stopping at StopChar
void XxxCoreADTClass::skipRecNextField(istream& InStream, 
				       char StopChar)
{
 char CurrentChar;
 if (!InStream.eof())
   InStream.get(CurrentChar);
 // Stop at delimiter or text enclosure character (StopChar)
 while(!(InStream.eof() || (CurrentChar==StopChar) || 
	 (CurrentChar==XXX_RECORD_SEPARATOR))) {
   InStream.get(CurrentChar);
  }
}


