#ifndef XXX_BASICTYPES_H
#define XXX_BASICTYPES_H

#include "types.h"

const XxxBasicType XXX_UNKNOWN = -1;
const XxxBasicType XXX_INT = 0;
const XxxBasicType XXX_DOUBLE = 1;
const XxxBasicType XXX_STRING = 2; // variable length string
const XxxBasicType XXX_FIXEDSTRING = 3; // fixed length string
const XxxBasicType XXX_BOOLEAN = 4;
/// BLOBs are used to keep meta info on ADTs
const XxxBasicType XXX_MIBLOB = 5;
const XxxBasicType XXX_GENFUNC = 6;  

// NOTE: XXX_GENFUNC must be the last of the core ADTs.
// The assumption is made that the user-defined ADTs begin after this.

/**
 // The basic types have dummy ADTClass objects just to specify their
 // names, and keep some of the code uniform. however, during execution,
 // these types are special-cased, so all the methods of these types are
 // empty
 */
class XxxCoreADTClass : public XxxADTClass {
 public:
  static const char XXX_RECORD_SEPARATOR; // = '\n'
  // define all the value classes for all CORE ADTs
  class IntValue : public XxxADTValue {
  public:
    IntValue(int i=0) { Value = i;}
    int Value;
  };
  class DblValue : public XxxADTValue {
  public:
    DblValue(double d=0.0) { Value = d;}
    double Value;
  };
  class BoolValue : public XxxADTValue {
  public:
    BoolValue(XxxBool v) { Value = v;}
    XxxBool Value;
  };
  class StrValue : public XxxADTValue {
  public:
    char Value[1];  // this is the start of an array
  };

  // now define all the statistics classes for CORE ADTs
  /**
   * Statistics root class for CORE ADTs
   */
  class BasicStats : public XxxADTClass::TypeStats { 
   public:
    int Count;
    BasicStats() { Count  = 0;}
    virtual XxxErrCode Term(const XxxADTMetaInfo* MetaInfo) {
      return XXX_OK;
     }
   };
  class IntStats : public BasicStats{ public:
    int MinVal;
    int MaxVal;
    double Var;
    double Avg;
    IntStats() 
     { MaxVal = INT_MIN; MinVal = INT_MAX; Var = 0.0; Avg = 0.0; }
    virtual int Size(const XxxADTMetaInfo* MetaInfo) const 
     { return sizeof(IntStats); }
    virtual XxxErrCode Accum(const XxxADTValue*Object, const XxxADTMetaInfo* MetaInfo) {
      int NewVal = ((IntValue*)Object)->Value;
      if (NewVal < MinVal) MinVal = NewVal;
      if (NewVal > MaxVal) MaxVal = NewVal;
      Avg = (Avg * Count/(Count+1.0)) + ((double)NewVal/(Count+1.0));
      Count++;
      return XXX_OK;
     }
    virtual XxxErrCode Replicate(const XxxADTMetaInfo* MetaInfo, TypeStats*& NewSI) const {
      NewSI = new IntStats(*this);
      return XXX_OK;
     }
    virtual XxxErrCode Copy(const XxxADTMetaInfo* MetaInfo, TypeStats* NewSI) const {
      *((IntStats*)NewSI) = *this;
      return XXX_OK;
     }
    virtual void Print(ostream& OStr, const XxxADTMetaInfo* MetaInfo) const{
      OStr << "(Min:" << MinVal << ",Max:" << MaxVal;
      OStr << ",Avg:" << Avg << ",Var:" << Var << ")";
    }
   };
  class DblStats : public BasicStats{ public:
    double MinVal;
    double MaxVal;
    double Avg;
    double Var;
    DblStats() 
     { MaxVal = 0.0; MinVal = DBL_MAX; Var = 0.0; Avg = 0; }
    virtual int Size(const XxxADTMetaInfo* MetaInfo) const 
     { return sizeof(DblStats); }
    virtual XxxErrCode Accum(const XxxADTValue*Object, const XxxADTMetaInfo* MetaInfo) {
      double NewVal = ((DblValue*)Object)->Value;
      if (NewVal < MinVal) MinVal = NewVal;
      if (!Count || (NewVal > MaxVal)) MaxVal = NewVal;
      Avg = (Avg * Count/(Count+1.0)) + ((double)NewVal/(Count+1.0));
      Count++;
      return XXX_OK;
     }
    virtual XxxErrCode Replicate(const XxxADTMetaInfo* MetaInfo, TypeStats*& NewSI) const {
      NewSI = new DblStats(*this);
      return XXX_OK;
     }
    virtual XxxErrCode Copy(const XxxADTMetaInfo* MetaInfo, TypeStats* NewSI) const {
      *((DblStats*)NewSI) = *this;
      return XXX_OK;
     }
    virtual void Print(ostream& OStr, const XxxADTMetaInfo* MetaInfo) const{
      OStr << "(Min:" << MinVal << ",Max:" << MaxVal;
      OStr << ",Avg:" << Avg << ",Var:" << Var << ")";
     }
   };

  static const IntValue XXX_DEFAULT_INT;
  static const DblValue XXX_DEFAULT_DBL;
  static const BoolValue XXX_DEFAULT_BOOL;

 public:
  XxxCoreADTClass(XxxBasicType Id, const char* Name, int ObjSize = 1,
		  XxxBool InField = XXX_TRUE,
		  XxxBool MaintainsMetaInfo = XXX_FALSE) 
    : XxxADTClass(Id,Name, ObjSize, InField, MaintainsMetaInfo) {}

  static void RegisterCoreADTs();

  //// Is TypeId a CoreADT?
  static inline XxxBool IsCoreADT(XxxBasicType TypeId);
  /**
   * This is used to find the offset of the actual value of a core value.
   * It is insufficient to use sizeof, since the compiler might align
   * doubles to 8-byte boundaries.
   */
  static inline int ValueOffset(XxxBasicType TypeId);

  static inline XxxBool SOrderedType(XxxBasicType TypeId,
				     const XxxADTMetaInfo* MInfo);
  //// Are the two values equal?
  static inline XxxBool SEquals(XxxBasicType TypeId,
				const XxxADTValue* Obj1, 
				const XxxADTMetaInfo* MI1,
				const XxxADTValue* Obj2,
				const XxxADTMetaInfo* MI2);
  static inline XxxBool SLessThan(XxxBasicType TypeId,
				  const XxxADTValue* Obj1, 
				  const XxxADTMetaInfo* MI1,
				  const XxxADTValue* Obj2,
				  const XxxADTMetaInfo* MI2);
  static inline XxxBool SGreaterThan(XxxBasicType TypeId,
				     const XxxADTValue* Obj1, 
				     const XxxADTMetaInfo* MI1,
				     const XxxADTValue* Obj2,
				     const XxxADTMetaInfo* MI2);
  static inline int SMaxObjectSize(XxxBasicType TypeId,
				   const XxxADTMetaInfo*);

  static inline int STypeCopy(XxxBasicType TypeId,
			      const XxxADTValue* SrcObject, 
			      XxxADTValue *NewObject,
			      const XxxADTMetaInfo* MetaInfo);
  static inline int STypeSize(XxxBasicType TypeId,
			      const XxxADTValue* Object, 
			      const XxxADTMetaInfo* MetaInfo);

  static inline XxxErrCode SCopyHToN(XxxBasicType TypeId,
				     const XxxADTValue* SrcObject, 
				     XxxADTValue *ResultObject,
				     const XxxADTMetaInfo* MetaInfo);
  static inline XxxErrCode SCopyNToH(XxxBasicType TypeId,
				     const XxxADTValue* SrcObject, 
				     XxxADTValue *ResultObject,
				     const XxxADTMetaInfo* MetaInfo);
  static inline XxxErrCode SWriteText(XxxBasicType TypeId,
				      ostream& OutStream, 
				      const XxxADTValue* Object,
				      const XxxADTMetaInfo* MetaInfo);
  static inline XxxErrCode SReadText(XxxBasicType TypeId,
				     istream& InStream, 
				     XxxADTValue* Object,
				     const XxxADTMetaInfo* MetaInfo,
				     XxxBool Delimited = XXX_FALSE,
				     char Delimiter = '\n');


  virtual XxxBool OrderedType(const XxxADTMetaInfo* MInfo) const 
   { return SOrderedType(TypeId, MInfo);}

  virtual XxxBool  LessThan(const XxxADTValue* Val1, 
			    const XxxADTMetaInfo* MI1,
			    const XxxADTValue* Val2,
			    const XxxADTMetaInfo* MI2) const {
   return SLessThan(TypeId, Val1, MI1, Val2, MI2);
   }
  virtual XxxBool  GreaterThan(const XxxADTValue* Val1, 
			       const XxxADTMetaInfo* MI1,
			       const XxxADTValue* Val2,
			       const XxxADTMetaInfo* MI2) const {
   return SGreaterThan(TypeId, Val1, MI1, Val2, MI2);
   }

  /**
   * Assigns a "default" value, if any exists.
   */
  virtual XxxErrCode DefaultVal(XxxADTValue* Object, 
				const XxxADTMetaInfo* MetaInfo) const;

  // although this should not be called, just in case
  virtual XxxErrCode WriteText(ostream& OutStream, 
			       const XxxADTValue* Object,
			       const XxxADTMetaInfo* MetaInfo) const {
    return SWriteText(TypeId, OutStream, Object, MetaInfo);
  }
  // this should not be called, but just in case
  virtual XxxErrCode ReadText(istream& InStream, XxxADTValue* Object,
			      const XxxADTMetaInfo* MetaInfo) const {
    return SReadText(TypeId, InStream, Object, MetaInfo);
  }

  virtual int MaxObjectSize(const XxxADTMetaInfo* MInfo) const 
   { return SMaxObjectSize(TypeId, MInfo);}
  // this should not be called, but just in case
  virtual int TypeSize(const XxxADTValue* Object, 
		       const XxxADTMetaInfo* MetaInfo) const
   { return STypeSize(TypeId, Object, MetaInfo);}

  static XxxErrCode loadTextNoDelimiter(istream& InStream, StrValue*,
					int MaxStrLen);
  static XxxErrCode loadTextDelimited(istream& InStream, 
				      StrValue*, 
				      char Delimiter,
				      int MaxStrLen);
  static void skipRecNonWhiteSpace(istream& InStream);
  static void skipRecNextField(istream& InStream, char StopChar);

  /**
   * Determines if this ADT can cast an object of another type into
   * its own type.
   * @param SourceType The type to cast from
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   */
  virtual XxxBool CanCastFrom(XxxBasicType SourceType, 
			      const XxxADTMetaInfo* SourceMetaInfo,
			      const XxxADTMetaInfo* TargetMetaInfo);
  /**
   * Executes a cast from another type into this type.
   * @param CastType The type id of the old object
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   * @param OldObject The object to be casted
   * @param NewObject The new object to be created
   */
  virtual XxxErrCode CastFrom(XxxBasicType SourceType, 
			      const XxxADTMetaInfo* SourceMetaInfo,
			      const XxxADTMetaInfo* TargetMetaInfo,
			      XxxADTValue* OldObject, 
			      XxxADTValue* NewObject);
  /**
   * Determines if this ADT can cast an object of its type into
   * another type.
   * @param TargetType The type to cast to
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   */
  virtual XxxBool CanCastTo(XxxBasicType TargetType, 
			    const XxxADTMetaInfo* SourceMetaInfo,
			    const XxxADTMetaInfo* TargetMetaInfo);
  /**
   * Executes a cast from this type into another type.
   * @param CastType The type id of the new object
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   * @param OldObject The object to be casted
   * @param NewObject The new object to be created
   */
  virtual XxxErrCode CastTo(XxxBasicType TargetType, 
			    const XxxADTMetaInfo* SourceMetaInfo,
			    const XxxADTMetaInfo* TargetMetaInfo,
			    XxxADTValue* OldObject, 
			    XxxADTValue* NewObject);

  virtual XxxErrCode CreateStatsInfo(const XxxADTMetaInfo *MetaInfo,
				     TypeStats*& StatsInfo) const;

  /// Never gets called for core types
  virtual XxxErrCode CopyHToN(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const {
    return SCopyHToN(TypeId, SrcObject, ResultObject, MetaInfo);
  }
  /// Never gets called for core types
  virtual XxxErrCode CopyNToH(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const {
    return SCopyNToH(TypeId, SrcObject, ResultObject, MetaInfo);
  }
 protected:
  virtual XxxErrCode makeStatsInfo(const XxxADTMetaInfo *MetaInfo,
				   const MetaInfoHandle& MIHandle,
				   TypeStats*& StatsInfo) const;
 };

/**
 * Integer ADT
 */
class XxxIntADT : public XxxCoreADTClass {
 public:
  XxxIntADT() : XxxCoreADTClass(XXX_INT, "int", sizeof(IntValue)) {}
 };

/**
 * Double ADT
 */
class XxxDoubleADT : public XxxCoreADTClass {
 public:
  XxxDoubleADT() : 
    XxxCoreADTClass(XXX_DOUBLE, "double", sizeof(DblValue)) {}
 };

/**
 * Boolean ADT
 */
class XxxBooleanADT : public XxxCoreADTClass {
 public:
  XxxBooleanADT() : 
    XxxCoreADTClass(XXX_BOOLEAN, "boolean", sizeof(BoolValue)) {}
 };

// invlude the string types
#include "stringtypes.h"

/**
 * The ADT for Binary Large OBjects used to store metainfo handles.
 * @author Praveen Seshadri, praveen@cs.cornell.edu
 * @version June 19, 1997
 */
class XxxMIBlobADT : public XxxCoreADTClass {
public:
  //// Constructor.
  XxxMIBlobADT()
    : XxxCoreADTClass(XXX_MIBLOB, "mi_blob",
		      sizeof(MetaInfoHandle), XXX_FALSE, XXX_FALSE) {}
  //// Informs the system of this type's in-record size.
  static inline int ValueSize(const XxxADTValue* Object, 
			      const XxxADTMetaInfo* MetaInfo) {
    MetaInfoHandle* MIHndl = (MetaInfoHandle*)Object;
    return MIHndl->CopySize();
   }
};

//
// Now for the inline methods
//
inline XxxBool 
XxxCoreADTClass::IsCoreADT(XxxBasicType TypeId) {
  return ((TypeId < XXX_GENFUNC) ? XXX_TRUE : XXX_FALSE);
}

inline int
XxxCoreADTClass::ValueOffset(XxxBasicType TypeId) {
  switch (TypeId) {
  case XXX_DOUBLE:
    return (int)(&((DblValue *)0)->Value);
  case XXX_BOOLEAN:
    return (int)(&((BoolValue *)0)->Value);
  case XXX_STRING:
  case XXX_FIXEDSTRING:
    return (int)(&((StrValue *)0)->Value);
  case XXX_INT:
    return (int)(&((IntValue *)0)->Value);
  case XXX_MIBLOB:
  default:
	ASSERT(1 == 0); // Shouldn't ever get here
  }
  return 0;
}

inline XxxBool 
XxxCoreADTClass::SEquals(XxxBasicType TypeId,
			 const XxxADTValue* Obj1, 
			 const XxxADTMetaInfo* MI1,
			 const XxxADTValue* Obj2,
			 const XxxADTMetaInfo* MI2) {
  XxxBool RetValue = XXX_FALSE;
  switch (TypeId) {
  case XXX_INT:
    if (((IntValue *)Obj1)->Value == ((IntValue *)Obj2)->Value)
      RetValue = XXX_TRUE;
    break;
  case XXX_DOUBLE:
    if (((DblValue *)Obj1)->Value == ((DblValue *)Obj2)->Value)
      RetValue = XXX_TRUE;
    break;
  case XXX_BOOLEAN:
    if (((BoolValue *)Obj1)->Value == ((BoolValue *)Obj2)->Value)
      RetValue = XXX_TRUE;
    break;
  case XXX_FIXEDSTRING:
  case XXX_STRING:
    if (!strcmp(((StrValue*)Obj1)->Value, ((StrValue*)Obj2)->Value))
      RetValue = XXX_TRUE;
    break;
  }
  return RetValue;
}

inline XxxBool 
XxxCoreADTClass::SOrderedType(XxxBasicType TypeId,
			      const XxxADTMetaInfo* MInfo) {
 XxxBool RetValue = XXX_FALSE;
 switch (TypeId) {
  case XXX_INT:
  case XXX_DOUBLE:
  case XXX_FIXEDSTRING:
  case XXX_STRING:
    RetValue = XXX_TRUE;
    break;
  } 
 return RetValue;
}

inline XxxBool 
XxxCoreADTClass::SLessThan(XxxBasicType TypeId,
			   const XxxADTValue* Obj1, 
			   const XxxADTMetaInfo* MI1,
			   const XxxADTValue* Obj2,
			   const XxxADTMetaInfo* MI2) {
 XxxBool RetValue = XXX_FALSE;
 switch (TypeId) {
  case XXX_INT:
    if (((IntValue *)Obj1)->Value < ((IntValue *)Obj2)->Value)
      RetValue = XXX_TRUE;
    break;
  case XXX_DOUBLE:
    if (((DblValue *)Obj1)->Value < ((DblValue *)Obj2)->Value)
      RetValue = XXX_TRUE;
    break;
  case XXX_FIXEDSTRING:
  case XXX_STRING:
    if (strcmp(((StrValue*)Obj1)->Value, ((StrValue*)Obj2)->Value) < 0)
      RetValue = XXX_TRUE;
    break;
  case XXX_BOOLEAN: // always false
    break;
  } 
 return RetValue;
}
inline XxxBool 
XxxCoreADTClass::SGreaterThan(XxxBasicType TypeId,
			 const XxxADTValue* Obj1, 
			 const XxxADTMetaInfo* MI1,
			 const XxxADTValue* Obj2,
			      const XxxADTMetaInfo* MI2) {
 XxxBool RetValue = XXX_FALSE;
 switch (TypeId) {
  case XXX_INT:
    if (((IntValue *)Obj1)->Value > ((IntValue *)Obj2)->Value)
      RetValue = XXX_TRUE;
    break;
  case XXX_DOUBLE:
    if (((DblValue *)Obj1)->Value > ((DblValue *)Obj2)->Value)
      RetValue = XXX_TRUE;
    break;
  case XXX_FIXEDSTRING:
  case XXX_STRING:
    if (strcmp(((StrValue*)Obj1)->Value, ((StrValue*)Obj2)->Value) > 0)
      RetValue = XXX_TRUE;
    break;
  case XXX_BOOLEAN: // always false
    break;
  } 
 return RetValue;
}

inline int 
XxxCoreADTClass::SMaxObjectSize(XxxBasicType TypeId,
				const XxxADTMetaInfo* MetaInfo) {
  if (TypeId == XXX_FIXEDSTRING)
    return XxxFixedStringADT::ValueSize(MetaInfo);
  else
    return ADTArray[TypeId]->XxxADTClass::MaxObjectSize(MetaInfo);
}

inline int 
XxxCoreADTClass::STypeCopy(XxxBasicType TypeId,
			   const XxxADTValue* SrcObject, 
                           XxxADTValue *NewObject,
                           const XxxADTMetaInfo* MetaInfo) {
 int CopySize = 0;
 switch (TypeId) {
 case XXX_INT:
   CopySize = sizeof(IntValue); break;
 case XXX_DOUBLE:
   CopySize = sizeof(DblValue); break;
 case XXX_BOOLEAN:
   CopySize = sizeof(BoolValue); break;
 case XXX_FIXEDSTRING:
   CopySize = XxxFixedStringADT::ValueSize(MetaInfo); break;
 case XXX_STRING:
   CopySize = XxxStringADTClass::ValueSize(SrcObject, MetaInfo); break;
 case XXX_MIBLOB:
   CopySize = XxxMIBlobADT::ValueSize(SrcObject, MetaInfo); break;
 }
 memcpy((char*)NewObject, (char*)SrcObject, CopySize);
 return CopySize;
}

inline int 
XxxCoreADTClass::STypeSize(XxxBasicType TypeId,
			   const XxxADTValue* Object, 
			   const XxxADTMetaInfo* MetaInfo) {
 int CopySize = 0;
 switch (TypeId) {
 case XXX_FIXEDSTRING:
   CopySize = XxxFixedStringADT::ValueSize(MetaInfo); break;
 case XXX_STRING:
   CopySize = XxxStringADTClass::ValueSize(Object, MetaInfo); break;
 case XXX_MIBLOB:
   CopySize = XxxMIBlobADT::ValueSize(Object, MetaInfo); break;
 default:
   CopySize = ADTArray[TypeId]->XxxADTClass::MaxObjectSize(MetaInfo);
 }
 return CopySize;
}

inline XxxErrCode 
XxxCoreADTClass::SCopyHToN(XxxBasicType TypeId,
			   const XxxADTValue* SrcObject, 
			   XxxADTValue *ResultObject,
			   const XxxADTMetaInfo* MetaInfo) {
  // in all cases, copy the initial header bytes
  *ResultObject = *SrcObject;
  switch (TypeId) {
  case XXX_INT: {
    IntValue* SrcVal = (IntValue*)SrcObject;
    IntValue* DestVal = (IntValue*)ResultObject;
    XxxNetworkInt(&(DestVal->Value), SrcVal->Value);
    break;
  }
  case XXX_DOUBLE: {
    DblValue* SrcVal = (DblValue*)SrcObject;
    DblValue* DestVal = (DblValue*)ResultObject;
    XxxNetworkDouble(&(DestVal->Value), SrcVal->Value);
    break;
  }
  case XXX_BOOLEAN: {
    BoolValue* SrcVal = (BoolValue*)SrcObject;
    BoolValue* DestVal = (BoolValue*)ResultObject;
    DestVal->Value = SrcVal->Value;
    break;
  }
  case XXX_MIBLOB:
  case XXX_FIXEDSTRING:
  case XXX_STRING: {
    STypeCopy(TypeId, SrcObject, ResultObject, MetaInfo);
    break;
  }
  } // switch (TypeId)
  return XXX_OK;
}

inline XxxErrCode 
XxxCoreADTClass::SCopyNToH(XxxBasicType TypeId,
			   const XxxADTValue* SrcObject, 
			   XxxADTValue *ResultObject,
			   const XxxADTMetaInfo* MetaInfo) {
  // in all cases, copy the initial header bytes
  *ResultObject = *SrcObject;
  switch (TypeId) {
  case XXX_INT: {
    IntValue* SrcVal = (IntValue*)SrcObject;
    IntValue* DestVal = (IntValue*)ResultObject;
    XxxHostInt(&(DestVal->Value), SrcVal->Value);
    break;
  }
  case XXX_DOUBLE: {
    DblValue* SrcVal = (DblValue*)SrcObject;
    DblValue* DestVal = (DblValue*)ResultObject;
    XxxHostDouble(&(DestVal->Value), SrcVal->Value);
    break;
  }
  case XXX_BOOLEAN: {
    BoolValue* SrcVal = (BoolValue*)SrcObject;
    BoolValue* DestVal = (BoolValue*)ResultObject;
    DestVal->Value = SrcVal->Value;
    break;
  }
 case XXX_MIBLOB:
  case XXX_FIXEDSTRING:
  case XXX_STRING: {
    STypeCopy(TypeId, SrcObject, ResultObject, MetaInfo);
    break;
  }
  } // switch
  return XXX_OK;
}

inline XxxErrCode 
XxxCoreADTClass::SWriteText(XxxBasicType TypeId,
			    ostream& OutStream, 
			    const XxxADTValue* Object,
			    const XxxADTMetaInfo* MetaInfo)
{
 switch(TypeId) {
  case XXX_INT:
   OutStream << ((IntValue *)Object)->Value;
   break;
  case XXX_DOUBLE:
   OutStream << ((DblValue *)Object)->Value;
   break;
  case XXX_BOOLEAN:
    OutStream << (((BoolValue *)Object)->Value ? '1' : '0');
   break;
  case XXX_STRING:
  case XXX_FIXEDSTRING:
   OutStream << "\"" << ((StrValue *)Object)->Value << "\"";
   break;
  case XXX_MIBLOB: {
    MetaInfoHandle* MIHndl = (MetaInfoHandle*)Object;
    OutStream << "MIBlob: size " << MIHndl->Size;
    break;
  }
  default:
   return XXX_NOT_OK;
  }
 if (OutStream.bad()) {
   return XXX_NOT_OK;
  }
 return XXX_OK; 
}

XxxErrCode 
XxxCoreADTClass::SReadText(XxxBasicType TypeId,
			   istream& InStream, 
			   XxxADTValue* Object, 
			   const XxxADTMetaInfo* MetaInfo,
			   XxxBool Delimited,
			   char Delimiter) {
  switch (TypeId) {
  case XXX_INT: {
    InStream >> ((IntValue *)Object)->Value;
    if (InStream.fail() && !InStream.eof())
      return XXX_NOT_OK;
    // Jump to the next field
    if (Delimited)
      skipRecNextField(InStream, Delimiter);
    break;
  }
  case XXX_DOUBLE: {
    InStream >> ((DblValue *)Object)->Value;
    if (InStream.fail() && !InStream.eof())
      return XXX_NOT_OK;
    // Jump to the next field
    if (Delimited)
      skipRecNextField(InStream, Delimiter);
    break;
  }
  case XXX_BOOLEAN: {
    char Buf[10];
    XxxBool Error = XXX_FALSE;
    InStream >> Buf;
    if (!strcasecmp(Buf, "true") || (Buf[0] == '1'))
      ((BoolValue *)Object)->Value = XXX_TRUE;
    else if (!strcasecmp(Buf, "false") || (Buf[0] == '0'))
      ((BoolValue *)Object)->Value = XXX_FALSE;
    else
      Error = XXX_TRUE;
    if ((InStream.fail() && !InStream.eof()) || Error)
      return XXX_NOT_OK;
    
    // Jump to the next field
    if (Delimited)
      skipRecNextField(InStream, Delimiter);
    break;
  }    
  case XXX_FIXEDSTRING:
  case XXX_STRING: {
    XxxErrCode LoadTextStatus = XXX_NOT_OK;
    int StrLength;
    if (TypeId == XXX_STRING)
      StrLength = XXX_TEXTSTRING_MAX;
    else
      StrLength = ((XxxFixedStringADT::FStringMetaInfo*)(MetaInfo))->Length;
    // fall thru to next case
    if (Delimited) 
      LoadTextStatus = 
	loadTextDelimited(InStream, (StrValue*)Object, 
			  Delimiter, StrLength);
    else
      LoadTextStatus = 
	loadTextNoDelimiter(InStream, (StrValue*)Object, StrLength);

    if (LoadTextStatus != XXX_OK) 
      return LoadTextStatus;
    break;
  }

  case XXX_MIBLOB:
  default:
    // XXX_NOTE_ERROR(XXX_BADVALUE, "Bad core ADT type to read");
    return XXX_NOT_OK;
  }
 
  if (InStream.fail() && !InStream.eof()) {
    //   XXX_NOTE_ERROR(XXX_BADADT, "Error reading core ADT type");
    return XXX_NOT_OK;
  }
  return XXX_OK; 
}

#endif  /* XXX_BASICTYPES_H */
