#ifndef XXX_ENGINES_H
#define XXX_ENGINES_H

#include "types.h"
#include "values.h"


XXX_EXTERN_CLASS_DECL(XxxDBEngine);

class XxxDBEngine {
 public:
  /// Latch is used to synchronize access to the engine array
  static smutex_t* Latch;
  //// This is PREDATOR's engine array; each engine inserts itself 
  static XxxDBEngine* DBEngineArray[];
  //// The number of engines defined in the system
  static int NumDBEngines;
  static inline XxxDBEngine* DefaultDBEngine() 
  { return DBEngineArray[0];};

  // to account for old code: eventually remove these #defines
#define XxxNumDBEngines XxxDBEngine::NumDBEngines
#define XxxDBEngineArray XxxDBEngine::DBEngineArray

  typedef char LangName[XXX_LANGNAME_MAX];

  int EngineId;
  char Name[XXX_ENGINE_NAME_MAX];
  int NumLanguages;
  LangName* LanguageNames;

  XxxDBEngine(const int EngId,
	      const char* EngineName, 
	      const int NumLangs) {
    EngineId = EngId;
    strncpy(Name, EngineName, XXX_ENGINE_NAME_MAX);
    Name[XXX_ENGINE_NAME_MAX-1] = 0;
    NumLanguages = NumLangs;
    if (NumLanguages)
      LanguageNames = new LangName[NumLangs];
    XxxDBEngineArray[EngineId] = this;
   }
  // declare destructor virtual so that subclass destructors get called
  virtual ~XxxDBEngine() { if (NumLanguages) delete [] LanguageNames; }

  // print the name of the engine
  void Print(ostream& OStr) const;
  // print any statistics maintained by the engine
  virtual void PrintStats(ostream& OStr) const {
    OStr << Name << ":  No Stats Maintained" << endl;
   }
  virtual void ResetStats() const {}

  // Function that parses parameterized expressions in the ADT language.
  // It returns a data structure that contains a parsed representation
  // of the ADT expression. It also specifies the return type of the
  // expression. The ExprString comes prefixed with the name of the
  // language which should parse it.
  virtual XxxErrCode ExprParse(char* ExprString,
			       const XxxArgEnv* ArgEnv,
			       XxxBasicType& ReturnType,
			       XxxGenericParse*& ParseStructure
			       ) = 0;
  // Function that optimizes a parsed expression in the ADT language.
  // It returns a data structure that contains the optimized execution plan
  // (or plan identifier) for the ADT expression.
  virtual XxxErrCode ExprOptimize(XxxGenericParse* ParseStructure,
				  XxxGenericPlan*& PlanStructure)  = 0;

  // Function that evaluates an optimized plan.
  virtual XxxErrCode ExprEvaluate(XxxGenericPlan* PlanStructure,
				  const XxxValueEnv* ArgEnv,
				  XxxADTValue *ReturnValue)  = 0;

  // Each ADT defines its own error codes. This way, error handling is
  // extensible! The ADT must simply provide this function for printing
  // the message associated with each condition
  virtual const char* ErrorString(XxxErrCond Condition) const
    { return "???"; }

  // each engine may provide certain knobs that can be set by the user.
  // the env encapsulates all such knobs.
  virtual XxxErrCode CreateEnv( XxxEngineEnv* &Env ) const
   { Env = NULL; return XXX_OK;}
  virtual XxxErrCode DeleteEnv( XxxEngineEnv* Env ) const
   { 
    if (Env) return XXX_NOT_OK;
    else return XXX_OK;
   }
 };

#endif
