#ifndef XXX_GENERIC_PLAN_H
#define XXX_GENERIC_PLAN_H

/**
 * This class specifies the database system environment
 */
class XxxDBEnv {
  void computeDerivedVals();
 public:
  /// multi-programming level (tells us how to divide resources)
  int MPL;
  /// number of bytes that make up a database page
  int PageSize;
  /// number of extra bytes per record
  int RecordOverhead; 
  /// number of buffer pages
  int BufPoolSize; 
  /// number of memory pages
  int MemorySize;

  /// number of "avg instructions" per second
  double CPUSpeed; 
  /// latency (seconds) for one random i/o
  double DiskLatency; 
  /// disk bandwidth (bytes per sec)
  double DiskXferRate; 

  // The following are derived values from the basic inputs above.
  // Each is a multiplicatory factor.
  double SeqIOWeight;
  double RndIOWeight;
  double CPUInstrWeight;

  void GetCurrentValues();
  XxxErrCode SetValue(const char* ValName,
		      char NewVal);
  XxxDBEnv();
  void Print(ostream& OStr) const;
 };

/***
 // class representing a generic query plan
 // 
 // Every plan is associated with a "cost".
 // I prefer to call it a "usage", because
 // really, all that can be measured is various
 // resource usages. Sometimes, these can
 // be further lumped together into some derived 
 // metric called "cost". 
 //
 // PlanUsage has a number of sub-components (eg DiskIO, memory, CPU)
 // 
 **/
class XxxGenericPlan { 
 public:
  struct DiskUsage {
    int SeqIO;  // number of sequentially r/w pages
    int RandIO; // number of randomly r/w pages
    void SetMin() {
      SeqIO = 0;
      RandIO = 0;
     }
    void SetMax() {
      SeqIO = INT_MAX;
      RandIO = INT_MAX;
     }
    void Add(const struct DiskUsage& Usage1, double Multiplier=1.0) {
      SeqIO += (int)(Multiplier * Usage1.SeqIO);
      RandIO += (int)(Multiplier * Usage1.RandIO);
     }
   };
  typedef double ProcUsage;  // number of CPU instructions
  // currently not used ...
  struct MemUsage {
    int PagesUsed;    // virtual memory pages used  
    int PagesInUse;   // virtual memory pages in use 
    void SetMax() {
      PagesUsed = PagesInUse = INT_MAX;
    }
    void SetMin() {
      PagesUsed = PagesInUse = 0;
    }

    void Add(const struct MemUsage& Usage1, double Multiplier=1.0) {
      PagesUsed  += (int)(Multiplier * Usage1.PagesUsed);
      PagesInUse += (int)(Multiplier * Usage1.PagesInUse);
     }
   };
  struct BufUsage {
    int PagesUsed;    // buffer pool pages used
    int PagesInUse;   // buffer pool pages in use
    void SetMax() {
      PagesUsed = PagesInUse = INT_MAX;
    }
    void SetMin() {
      PagesUsed = PagesInUse = 0;
    }
    void Add(const struct BufUsage& Usage1, double Multiplier = 1.0) {
      PagesUsed  += (int)(Multiplier * Usage1.PagesUsed);
      PagesInUse += (int)(Multiplier * Usage1.PagesInUse);
     }
   };
  struct UsageComponents {
    DiskUsage  Disk;
    ProcUsage  Proc;
    double DiskProcOverlap; // number between 0.0 and 1.0

    BufUsage   Buf;
    MemUsage   Mem;

    void SetMax() {
      Disk.SetMax();
      Proc = INT_MAX;
      DiskProcOverlap = 0.0;
      Buf.SetMax();
      Mem.SetMax();
    }
    void SetMin() {
      Disk.SetMin();
      Proc = 0;
      DiskProcOverlap = 0.0;
      Buf.SetMin();
      Mem.SetMin();
    }

    void Add(const struct UsageComponents& Usage1,
	     double DiskMultiplier = 1.0,
	     double CPUMultiplier = 1.0,
	     double BufMultiplier = 1.0,
	     double MemMultiplier = 1.0) {
      Disk.Add(Usage1.Disk, DiskMultiplier);
      Proc += (CPUMultiplier*Usage1.Proc);
      DiskProcOverlap = 0.0;
      Buf.Add(Usage1.Buf, BufMultiplier);
      Mem.Add(Usage1.Mem, MemMultiplier);
     }
    // a single metric : need to define this
    double Cost() const;
   };
  struct PlanUsage {
    // to start getting answers
    UsageComponents FirstUsage;
    // to finish entire computation
    UsageComponents FullUsage;
    void SetMin() { 
      FirstUsage.SetMin();
      FullUsage.SetMin();
    }
    void SetMax() { 
      FirstUsage.SetMax();
      FullUsage.SetMax();
    }
    void Reset() { SetMin();}

    PlanUsage() { Reset();}
    void Add(const PlanUsage& Usage2, 
	     double DiskMultiplier = 1.0,
	     double CPUMultiplier = 1.0,
	     double BufMultiplier = 1.0,
	     double MemMultiplier = 1.0) {
      FirstUsage.Add(Usage2.FirstUsage, DiskMultiplier, CPUMultiplier,
		     BufMultiplier, MemMultiplier);
      FullUsage.Add(Usage2.FullUsage, DiskMultiplier, CPUMultiplier,
		    BufMultiplier, MemMultiplier);
     }
    // in case a single value is needed as an aggregate metric
    double FullCost() const {
      return FullUsage.Cost();
     }
    double FirstCost() const {
      return FirstUsage.Cost();
     }
    void Print(ostream& OStr) const;
   };

  XxxGenericPlan() {}
  XxxGenericPlan(const XxxGenericPlan &GP) {}
  virtual XxxGenericPlan * Replicate() const = 0;
  virtual ~XxxGenericPlan() {}
  virtual XxxErrCode UsageOfPlan(PlanUsage& Usage) const = 0;
 };

#endif

