#include "papi_prof.h"
#include <iostream>
#include <iomanip>
#include <cstdio>
#include <cassert>
#include <papi.h>

//initialization of static members
template <int MaxNumEvents>
int PapiProf<MaxNumEvents>::event_set = PAPI_NULL;
template <int MaxNumEvents>
int PapiProf<MaxNumEvents>::ActNumEvents = 0;
//specialise for 2 events
template <>
long long PapiProf<2>::accum_values[2] = {0, 0};
template <>
long long PapiProf<2>::stop_values[2] = {0, 0};
template <>
int PapiProf<2>::events[2] = {0, 0};

//overflow handling
template <int MaxNumEvents>
void PapiProf<MaxNumEvents>::overflow_handler(int EventSet, int EventCode, int EventIndex, long long *values, int *threshold, void *context) {
  //  cout << " ----- PAPI overflow ----- " << endl;
  PAPI_accum(EventSet, accum_values);
}

//output
template <int MaxNumEvents>
void PapiProf<MaxNumEvents>::PrintValues() {
  char ecodestr[PAPI_MAX_STR_LEN];
  char edesc[PAPI_MAX_STR_LEN];

  cout << endl << "-----------------------------------------" << endl
       << "PAPI results:" << endl;
  
  for (int i = 0; i < ActNumEvents; ++i) {
    *ecodestr=' ';
    int res = PAPI_event_code_to_name(events[i], ecodestr);
    cout << setw(12);
    if (res == PAPI_OK)
      cout << ecodestr;
    else 
      cout << events[i];
    cout << ":" << setw(10) << accum_values[i];
    if (PAPI_describe_event(ecodestr, &events[i], edesc) == PAPI_OK) 
      cout << setw(10) << "(" << edesc << ")";
    cout << endl;
  }
  cout << "-----------------------------------------" << endl;
}

template <int MaxNumEvents>
PapiProf<MaxNumEvents>::PapiProf() 
{
  int papi_err = PAPI_OK;
  PAPI_library_init(PAPI_VER_CURRENT);
  assert(MaxNumEvents == PAPI_num_counters());
  //create event set
  if ((papi_err = PAPI_create_eventset(&event_set)) != PAPI_OK) {
    cerr << "Error creating event set (" << PAPI_strerror(papi_err)
	 << ")" << endl;
    return;
  }
}

template <int MaxNumEvents>
void PapiProf<MaxNumEvents>::PapiStart(const char* eventstr) {
  //parse eventstr into vents
  int papi_err = PAPI_OK;
  char token[PAPI_MAX_STR_LEN];
  const char *cur_p, *maxp; 

  int n, i = 0;
  cur_p = eventstr;
  maxp = eventstr + strlen(eventstr);
  //number of events to expect
  sscanf(cur_p, "%d%n", &ActNumEvents, &n);
  cur_p = cur_p + n;
  if (ActNumEvents > MaxNumEvents) {
    cerr << "You have specified too many events (" 
	 << ActNumEvents << ")." << endl
	 << "You can specify at most " 
	 << MaxNumEvents << " events." << endl
	 << "Papi has not been started" << endl;
    return;
  }

  //read the events
  while ((cur_p < maxp) && (i < ActNumEvents)) {
    sscanf(cur_p, "%s%n", token, &n);
    if (PAPI_event_name_to_code(token, &events[i]) != PAPI_OK) {
      //if we fail then assume event is a native event (hex number)
      int native_event;
      sscanf(token, "%x", &native_event);
      //      events[i] = 0 | native_event << 8 | i; //the i in the end tells us what counter to use
      events[i] = native_event << 8; 
    }
    cur_p = cur_p + n;
    ++i;
  }
  if (i != ActNumEvents) {
    cerr << "Wrong number of events specified. You must specify " 
	 << ActNumEvents << " events." << endl 
	 << "Papi has not been started" << endl;
    return;
  }

  if ((papi_err = PAPI_add_events(&event_set, events, ActNumEvents)) != PAPI_OK) {
    cerr << "Error adding to event set ("  << PAPI_strerror(papi_err)
	 << ")" << endl;
    return;
  }

  if ((papi_err = PAPI_overflow(event_set, events[0], 50000, 0, 
				overflow_handler)) != PAPI_OK) {
    cerr << "Error setting up overflow handling ("  << PAPI_strerror(papi_err) 
	 << ")" << endl;
    return;
  } 
  if ((papi_err = PAPI_start(event_set)) != PAPI_OK) {
    cerr << "Error starting counting ("  << PAPI_strerror(papi_err)
	 << ")" << endl;
    return;
  }
}

template <int MaxNumEvents>
void PapiProf<MaxNumEvents>::PapiStop() {
  int papi_err = PAPI_OK;
  PAPI_accum(event_set, accum_values);
  //Stop PAPI
  if ((papi_err = PAPI_stop(event_set, stop_values)) != PAPI_OK) {
    cerr << "Error stopping counting ("  << PAPI_strerror(papi_err)
	 << ")" << endl;
    return;
  }
  //really, this step shouldn't be necessary
  for (int i = 0; i < ActNumEvents; ++i) {
    accum_values[i] += stop_values[i];
  }
  if ((papi_err = PAPI_cleanup_eventset(&event_set)) != PAPI_OK) {
    cerr << "Error cleaning up event set (" << PAPI_strerror(papi_err)
	 << ")" << endl;
    //There seems to be problems with calling cleanup_eventset when
    //using native events so we accept the error for now.
  }
  PrintValues();
}

template <int MaxNumEvents>
PapiProf<MaxNumEvents>::~PapiProf() {
  //call PapiStop here if papi is running.
  int status = 0;
  int papi_err = PAPI_OK;
  if ((papi_err = PAPI_state(event_set, &status)) != PAPI_OK) {
    cerr << "Error getting papi state ("  << PAPI_strerror(papi_err)
	 << ")" << endl;
  } 
 
  if ((status & PAPI_RUNNING) || (status & PAPI_PAUSED)) {
    PapiStop();
  }

  if ((papi_err = PAPI_destroy_eventset(&event_set)) != PAPI_OK) {
    cerr << "Error destroying event set ("  << PAPI_strerror(papi_err)
	 << ")" << endl;
    //There seems to be problems with calling destroy_eventset when
    //using native events so we accept the error for now.
  }
}

#ifdef PRED_EXPLICIT_TEMPLATE
//explicit instantiation
template class PapiProf<2>;
#endif
