//
// This file is included from within types.h. It contains many of the nested
// classes (hence -nc) within XxxADTClass.
// Specifically:
//     ParseInfo: root class with parse information for functions
//     OptPlanInfo: root class with parse information for functions
//     MetaInfoGen
//     OptimizeFunc
//     ExecuteFunc
//     MethodInfo
//     FuncMethodInfo
//     AggrMethodInfo
//

#ifndef XXX_TYPES_NC_H
#define XXX_TYPES_NC_H

  /**
   *   Parse Structure for methods of ADTs
   */
  class ParseInfo : public XxxFuncParseInfo {
  public:
    int MethodIndex;
    XxxADTMetaInfo* TypeCatInfo;
    ParseInfo(const char* Name,
        int MethIndex,
        XxxGenericParse* ParseData,  // any parse tree for the function
        XxxBasicType ThisTypeId,
        XxxBasicType RetType,
        XxxADTMetaInfo* CatInfo = NULL) : 
      XxxFuncParseInfo (Name, ThisTypeId, ParseData, RetType) {
        MethodIndex = MethIndex;
        TypeCatInfo = CatInfo;
    }
    ParseInfo(const ParseInfo& src) : XxxFuncParseInfo(src) {
      MethodIndex=src.MethodIndex;
      TypeCatInfo=NULL;
      if (src.TypeCatInfo)
	src.TypeCatInfo->Replicate(TypeCatInfo);
    }
    virtual XxxGenericParse* Replicate() const {
      return new ParseInfo(*this);
    }
    virtual ~ParseInfo() {}
    virtual XxxADTMetaInfo* GenerateCatInfo() const { 
      if (TypeCatInfo) return TypeCatInfo;
      else return XxxFuncParseInfo::GenerateCatInfo();
    }
  };

  
  /**
   * Plan structure for methods of ADTs
   */
  class OptPlanInfo : public XxxFuncPlanInfo {
  protected:
    PlanUsage  planUsage;
    XxxBool    compressOutput;
    XxxBool*   decompressInput;
    XxxBool    decompressOwner;
    XxxBool    suspendComputation;
  public:
    XxxErrCode SuspendComputation();
    XxxErrCode UnsuspendComputation();
    XxxBool IsSuspended() { return suspendComputation; }
    XxxErrCode modifyArgCompression(XxxValueExprPlan* OwnerPlan,
				    XxxValExprPlanList* ArgPlans);
    int MethodIndex;
    OptPlanInfo(const XxxFuncParseInfo* FuncParse, int ArgCount,
		XxxGenericPlan* Plan = NULL);
    OptPlanInfo(const OptPlanInfo &OPI);
    virtual ~OptPlanInfo() { if (decompressInput) delete [] decompressInput;} 
    virtual XxxErrCode UsageOfPlan(PlanUsage& Usage) const
    { Usage = planUsage; return XXX_OK;}
    virtual XxxErrCode Evaluate(XxxValueEnv* Env, 
                                FuncMethodInfo* ThisMethInfo,
                                XxxADTValue *ReturnValue) const;
    /**
     * Creates a suspended computation encapsulating the execution of this 
     * plan. The suspended computation can be stored and later executed.
     */
    virtual XxxErrCode SuspendEvaluation(XxxValueEnv* Env, 
					 FuncMethodInfo* ThisMethInfo,
					 XxxADTValue *ReturnValue) const;
    virtual XxxGenericPlan * Replicate() const
    {
      return (XxxGenericPlan *)(new OptPlanInfo(*this));
    }
  }; 

  // 4) Method registry information
  typedef XxxADTMetaInfo* (*MetaInfoGen)(XxxValueExpression* Owner,
      XxxValueExprList *ArgList);

  typedef XxxErrCode (*OptimizeFunc)(ParseInfo* FParse,
 				     XxxValueExprPlan* OwnerPlan,
 				     XxxValExprPlanList* ArgPlans,
				     OptPlanInfo*& FuncPlan);

  typedef XxxErrCode (OptPlanInfo::*ExecuteFunc)(XxxValueEnv* Env,
      XxxADTValue *ReturnValue) const;

  /**
   * Information on each method of an ADT
   */
  class MethodInfo { public:
    char           Name[XXX_FUNCNAME_MAX];
    int            NumArgs;
    XxxBasicType   ReturnType;
    MetaInfoGen    MIGen;
    OptimizeFunc   OptFunc;
    MethodInfo(const char* FName, int Arity, XxxBasicType RetType,
        MetaInfoGen MIG, OptimizeFunc OF) 
      : NumArgs(Arity), ReturnType(RetType), MIGen(MIG), OptFunc(OF) {
        strncpy(Name, FName, XXX_FUNCNAME_MAX);
	Name[XXX_FUNCNAME_MAX-1] = 0;
    }
    MethodInfo(const MethodInfo &MInfo);
    virtual MethodInfo * Replicate() const {
      return new MethodInfo(*this);
    }
  };

  /**
   * Information on each function method of an ADT
   */
  class FuncMethodInfo : public MethodInfo { 
  public:
    ExecuteFunc ExecFunc;
    XxxBasicType ArgTypes[XXX_ARGCNT_MAX];
    FuncMethodInfo(const char* FName, 
        XxxBasicType RetType, MetaInfoGen MIG, 
        OptimizeFunc OF, ExecuteFunc EF,  int Arity, ...);
    FuncMethodInfo(const FuncMethodInfo &FMInfo);
    virtual MethodInfo * Replicate() const {
      return (MethodInfo *)(new FuncMethodInfo(*this));
    }
    // just match the method name, arg types and arity
    XxxErrCode Matches(const char* FuncName,
        XxxValueExpression* Owner,
        XxxValueExprList *ArgList,
        int NumArgs,
        XxxBasicType& ReturnType,
        XxxADTMetaInfo*& ReturnMetaInfo,
        XxxBool& Success) const;
  };

  typedef XxxErrCode (*AggrInitFunc)(XxxAggrOpHandle& Hndl);
  typedef XxxErrCode (*AggrAccumFunc)(XxxAggrOpHandle& Hndl, 
				      XxxADTValue *NewVal);
  typedef XxxErrCode (*AggrTermFunc)(XxxAggrOpHandle& Hndl, 
				     XxxADTValue* FinalVal);
  typedef XxxErrCode (*AggrReclaimFunc)(XxxAggrOpHandle& Hndl);
  typedef XxxErrCode (*AggrDropFunc)(XxxAggrOpHandle& Hndl, 
				     XxxADTValue* NewVal);

  /**
   * Information on each aggregate method of an ADT
   */
  class AggrMethodInfo : public MethodInfo { 
   public:
    XxxBasicType ArgType;
    AggrInitFunc InitFunc;
    AggrAccumFunc AccumFunc;
    AggrTermFunc TermFunc;
    AggrReclaimFunc ReclaimFunc;
    AggrDropFunc DropFunc;
    AggrMethodInfo(const char* FName, XxxBasicType AType,
        XxxBasicType RetType, MetaInfoGen MIG, 
        OptimizeFunc OF, AggrInitFunc IFunc,
        AggrAccumFunc AFunc, AggrTermFunc TFunc,
        AggrReclaimFunc RFunc, AggrDropFunc DFunc) :
      MethodInfo(FName, 1, RetType, MIG, OF), ArgType(AType),
      InitFunc(IFunc), AccumFunc(AFunc), TermFunc(TFunc),
      ReclaimFunc(RFunc), DropFunc(DFunc) {}
    AggrMethodInfo(const AggrMethodInfo &AMI) :
      MethodInfo(AMI), ArgType(AMI.ArgType), InitFunc(AMI.InitFunc),
      AccumFunc(AMI.AccumFunc), TermFunc(AMI.TermFunc), 
      ReclaimFunc(AMI.ReclaimFunc), DropFunc(AMI.DropFunc) { }
    virtual MethodInfo * Replicate() const {
      return (MethodInfo *)(new AggrMethodInfo(*this));
    }
    XxxErrCode Matches(const char* FuncName,
        XxxValueExpression* Child,
        XxxBasicType& ReturnType,
        XxxADTMetaInfo*& ReturnMetaInfo,
        XxxBool& Success) const;
  };

#endif
