/*****

****/

#ifndef XXX_AGGREXPR_H
#define XXX_AGGREXPR_H
#include <math.h>
#include <limits.h>
#include "types.h"
#include "expr.h"
#include "exprplan.h"
#include "funcexpr.h"

/**
  * Handle used to maintain state during aggregate computations.
  */
class XxxAggrOpHandle {
 public:
  class AggrState {
  public:
    AggrState() {}
    virtual ~AggrState() {}
  };

  //// in which aggr result will be accumulated
  XxxADTValue *AccumVal; 

  XxxBool Terminate;
  XxxBool Started;
  unsigned int Count;

  AggrState* State;

  void Init() {
    Terminate = XXX_FALSE; Started = XXX_FALSE; Count = 0; State = NULL;
   }
  XxxAggrOpHandle(XxxBasicType OutType, const XxxADTMetaInfo* OutMInfo) { 
    //// this is sufficient for most simple aggregates
    AccumVal = (XxxADTValue *)(new char[XxxADTArray[OutType]->MaxObjectSize(OutMInfo)]);
    AccumVal->Init();
    Init();
   }
  ~XxxAggrOpHandle() {
    if (AccumVal)
      delete [] ((char *)AccumVal);
    if (State) delete State;
   }
 };

/**
 * Generic spec of various aggregate functions as iterative computations
 */
class XxxAggrOpSpec {
 public:
  virtual const char* const AggrName() const = 0;
  /**
   * Abstract functions defining the behavior of the aggregate function
   */
  //@{
  //// initialize
  virtual XxxErrCode Init(const XxxFuncExpression::AggrPlan* APlan,
			  XxxAggrOpHandle& Hndl) const { 
    Hndl.AccumVal->Set(XxxADTValue::NULL_VALUE);
    return XXX_OK;
  }
  //// add a value to the aggregate
  virtual XxxErrCode Accum(const XxxFuncExpression::AggrPlan* APlan,
			   XxxAggrOpHandle& Hndl) const=0;
  //// drop a value from an aggregate
  virtual XxxErrCode Drop(const XxxFuncExpression::AggrPlan* APlan,
			  XxxAggrOpHandle& Hndl) const
   { return XXX_NOT_OK;}
  //// perform final computation, reclaim any state stored in handle
  virtual XxxErrCode Terminate(const XxxFuncExpression::AggrPlan* APlan,
			       XxxAggrOpHandle& Hndl, XxxADTValue *FinalVal) const
   { return XXX_OK;}
  //@}
 };

/**
 * Special class of simple arithmetic aggregates (min, max, count, sum, prod)
 **/
class XxxSimpleOpSpec : public XxxAggrOpSpec {
 public:
  virtual XxxErrCode Init(const XxxFuncExpression::AggrPlan* APlan,
			  XxxAggrOpHandle& Hndl) const;
  virtual XxxErrCode Terminate(const XxxFuncExpression::AggrPlan* APlan,
			       XxxAggrOpHandle& Hndl, XxxADTValue *FinalVal) const;
 };

/**
 *
 */
class XxxMinOpSpec : public XxxSimpleOpSpec {
 public:
  virtual const char* const AggrName() const { return "Min";}
  virtual XxxErrCode Init(const XxxFuncExpression::AggrPlan* APlan,
			  XxxAggrOpHandle& Hndl) const;
  virtual XxxErrCode Accum(const XxxFuncExpression::AggrPlan* APlan,
			   XxxAggrOpHandle& Hndl) const;
 };

/**
 *
 */
class XxxMaxOpSpec : public XxxSimpleOpSpec {
 public:
  virtual const char* const AggrName() const { return "Max";}
  virtual XxxErrCode Init(const XxxFuncExpression::AggrPlan* APlan,
			  XxxAggrOpHandle& Hndl) const;
  virtual XxxErrCode Accum(const XxxFuncExpression::AggrPlan* APlan,
			   XxxAggrOpHandle& Hndl) const;
 };

/**
 *
 */
class XxxCountOpSpec : public XxxSimpleOpSpec {
 public:
  virtual const char* const AggrName() const { return "Count";}
  virtual XxxErrCode Init(const XxxFuncExpression::AggrPlan* APlan,
			  XxxAggrOpHandle& Hndl) const;
  virtual XxxErrCode Accum(const XxxFuncExpression::AggrPlan* APlan,
			   XxxAggrOpHandle& Hndl) const;
  virtual XxxErrCode Drop(const XxxFuncExpression::AggrPlan* APlan,
			  XxxAggrOpHandle& Hndl) const;
 };

/**
 *
 */
class XxxSumOpSpec : public XxxSimpleOpSpec {
 public:
  virtual const char* const AggrName() const { return "Sum";}
  virtual XxxErrCode Init(const XxxFuncExpression::AggrPlan* APlan,
			  XxxAggrOpHandle& Hndl) const;
  virtual XxxErrCode Accum(const XxxFuncExpression::AggrPlan* APlan,
			   XxxAggrOpHandle& Hndl) const;
  virtual XxxErrCode Drop(const XxxFuncExpression::AggrPlan* APlan,
			  XxxAggrOpHandle& Hndl) const;
 };

/**
 *
 */
class XxxProdOpSpec : public XxxSimpleOpSpec {
 public:
  virtual const char* const AggrName() const { return "Prod";}
  virtual XxxErrCode Init(const XxxFuncExpression::AggrPlan* APlan,
			  XxxAggrOpHandle& Hndl) const;
  virtual XxxErrCode Accum(const XxxFuncExpression::AggrPlan* APlan,
			   XxxAggrOpHandle& Hndl) const;
  virtual XxxErrCode Drop(const XxxFuncExpression::AggrPlan* APlan,
			  XxxAggrOpHandle& Hndl) const;
 };

/**
 *
 */
class XxxAvgOpSpec : public XxxAggrOpSpec {
 private:
  class AvgInfo : public XxxAggrOpHandle::AggrState {
  public:
    double Val;
   };
 public:
  virtual const char* const AggrName() const { return "Avg";}
  virtual XxxErrCode Init(const XxxFuncExpression::AggrPlan* APlan,
			  XxxAggrOpHandle& Hndl) const;
  virtual XxxErrCode Terminate(const XxxFuncExpression::AggrPlan* APlan,
			       XxxAggrOpHandle& Hndl, XxxADTValue *FinalVal) const;
  virtual XxxErrCode Accum(const XxxFuncExpression::AggrPlan* APlan,
			  XxxAggrOpHandle& Hndl) const;
  virtual XxxErrCode Drop(const XxxFuncExpression::AggrPlan* APlan,
			  XxxAggrOpHandle& Hndl) const;
 };

/**
 * An aggregate expression is a slight variant of a FuncExpression.
 * For standard aggregates like MIN, MAX, etc, they are represented
 * as methods of XXX_GENFUNC_ADT. For extensible aggregates,
 * although textually, the aggregate looks like Foo(A, B, C),
 * we represent it internally as A.Foo(B, C)
 */
class XxxAggrExpression: public XxxFuncExpression {
 public:
  XxxAggrExpression(const char* AggrName, XxxValueExprList* AggrList) :
    XxxFuncExpression(AggrName, 
		      ((AggrList && !stdAggr(AggrName)) ? AggrList->Next : AggrList),
		      ((AggrList && !stdAggr(AggrName)) ? AggrList->Expr : 0)) 
  { isAggr = XXX_TRUE;}
};

//
// definition of inline methods for AggrOpSpecs
//
inline XxxErrCode XxxSimpleOpSpec::Init(const XxxFuncExpression::AggrPlan* APlan,
					XxxAggrOpHandle& Hndl) const
{
 Hndl.Started = XXX_TRUE;
 Hndl.AccumVal->Set(XxxADTValue::NULL_VALUE);
 return XXX_OK;
}
inline XxxErrCode XxxSimpleOpSpec::Terminate(const XxxFuncExpression::AggrPlan* APlan,
					
					     XxxAggrOpHandle& Hndl, XxxADTValue *FinalVal) const
{
 Hndl.Started = XXX_FALSE;
 XxxBasicType OutputType = APlan->BasicType();

 // special case numeric types for faster execution
 if (OutputType == XXX_INT) {
   *((XxxCoreADTClass::IntValue *)FinalVal) = *((XxxCoreADTClass::IntValue*)(Hndl.AccumVal));
   return XXX_OK;
  }
 else if (OutputType == XXX_DOUBLE) {
   *((XxxCoreADTClass::DblValue *)FinalVal) = *((XxxCoreADTClass::DblValue *)(Hndl.AccumVal));
   return XXX_OK;
  }
 else if (OutputType == XXX_STRING || OutputType == XXX_FIXEDSTRING) {
   *((XxxCoreADTClass::StrValue *)FinalVal) = *((XxxCoreADTClass::StrValue *)(Hndl.AccumVal));
   return XXX_OK;
  }
 else XXX_CALL_RET(XxxADTArray[OutputType]->TypeCopy(Hndl.AccumVal, FinalVal, 
					    APlan->GenerateCatInfo()),
		   XXX_BADTYPE, "Error copying aggr result");
 return XXX_OK;
}

inline XxxErrCode XxxMinOpSpec::Init(const XxxFuncExpression::AggrPlan* APlan,
				     XxxAggrOpHandle& Hndl) const
{
 Hndl.Started = XXX_TRUE;
 XxxBasicType OutputType = APlan->BasicType();

 // make sure to initially set the NULL bit
 Hndl.AccumVal->Set(XxxADTValue::NULL_VALUE);

 if (OutputType == XXX_INT)
   ((XxxCoreADTClass::IntValue *)(Hndl.AccumVal))->Value = INT_MAX;
 else if (OutputType == XXX_DOUBLE)
   ((XxxCoreADTClass::DblValue *)(Hndl.AccumVal))->Value = DBL_MAX;
 else if (OutputType == XXX_STRING || OutputType == XXX_FIXEDSTRING)
   Hndl.Started = XXX_FALSE;
 else {
   XXX_NOTE_ERROR(XXX_BADTYPE, "For MIN aggregate");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

inline XxxErrCode XxxMinOpSpec::Accum(const XxxFuncExpression::AggrPlan* APlan,
				      XxxAggrOpHandle& Hndl) const
{
 XxxBasicType OutputType = APlan->BasicType();
 XxxADTValue* NewVal = APlan->ValueEnv->ArgValsPtrs[0];
 if (NewVal->Is(XxxADTValue::NULL_VALUE)) {
   // nothing to do if it is a null value
   return XXX_OK;
 }
 if (!Hndl.Count) {
   // make sure to clear the null bit
   Hndl.AccumVal->Unset(XxxADTValue::NULL_VALUE);
 }

 if (OutputType == XXX_INT) {
   if (((XxxCoreADTClass::IntValue*)NewVal)->Value < ((XxxCoreADTClass::IntValue*)(Hndl.AccumVal))->Value)
     *((XxxCoreADTClass::IntValue*)(Hndl.AccumVal)) = *((XxxCoreADTClass::IntValue*)NewVal);
  }
 else if (OutputType == XXX_DOUBLE) {
   if (((XxxCoreADTClass::DblValue*)NewVal)->Value < ((XxxCoreADTClass::DblValue*)(Hndl.AccumVal))->Value)
     *((XxxCoreADTClass::DblValue*)(Hndl.AccumVal)) = *((XxxCoreADTClass::DblValue*)NewVal);
  }
 else if (OutputType == XXX_STRING || OutputType == XXX_FIXEDSTRING) {
   // the first string automatically becomes the minimum so far.
   if(Hndl.Started == XXX_FALSE) {
     Hndl.Started = XXX_TRUE;
     *((XxxCoreADTClass::StrValue*)(Hndl.AccumVal)) = *((XxxCoreADTClass::StrValue*)NewVal);
   }
   else if (((XxxCoreADTClass::StrValue*)NewVal)->Value < ((XxxCoreADTClass::StrValue*)(Hndl.AccumVal))->Value)
     *((XxxCoreADTClass::StrValue*)(Hndl.AccumVal)) = *((XxxCoreADTClass::StrValue*)NewVal);
  }
 else {
   XXX_NOTE_ERROR(XXX_BADTYPE, "For MIN aggregate");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

inline XxxErrCode XxxMaxOpSpec::Init(const XxxFuncExpression::AggrPlan* APlan,
				     XxxAggrOpHandle& Hndl) const
{
 XxxBasicType OutputType = APlan->BasicType();
 Hndl.Started = XXX_TRUE;
 // initally, set the NULL bit
 Hndl.AccumVal->Set(XxxADTValue::NULL_VALUE);

 if (OutputType == XXX_INT)
   ((XxxCoreADTClass::IntValue*)(Hndl.AccumVal))->Value = INT_MIN;
 else if (OutputType == XXX_DOUBLE)
   ((XxxCoreADTClass::DblValue*)(Hndl.AccumVal))->Value = DBL_MIN;
 else if (OutputType == XXX_STRING || OutputType == XXX_FIXEDSTRING)
   Hndl.Started = XXX_FALSE;
 else {
   XXX_NOTE_ERROR(XXX_BADTYPE, "For MAX aggregate");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

inline XxxErrCode XxxMaxOpSpec::Accum(const XxxFuncExpression::AggrPlan* APlan,
				      XxxAggrOpHandle& Hndl) const
{
 XxxBasicType OutputType = APlan->BasicType();
 XxxADTValue* NewVal = APlan->ValueEnv->ArgValsPtrs[0];

 if (NewVal->Is(XxxADTValue::NULL_VALUE)) {
   // nothing to do if it is a null value
   return XXX_OK;
 }
 if (!Hndl.Count) {
   // make sure to clear the null bit
   Hndl.AccumVal->Unset(XxxADTValue::NULL_VALUE);
 }

 if (OutputType == XXX_INT) {
   if (((XxxCoreADTClass::IntValue*)NewVal)->Value > ((XxxCoreADTClass::IntValue*)(Hndl.AccumVal))->Value)
     *((XxxCoreADTClass::IntValue*)(Hndl.AccumVal)) = *((XxxCoreADTClass::IntValue*)NewVal);
  }
 else if (OutputType == XXX_DOUBLE) {
   if (((XxxCoreADTClass::DblValue*)NewVal)->Value > ((XxxCoreADTClass::DblValue*)(Hndl.AccumVal))->Value)
     *((XxxCoreADTClass::DblValue*)(Hndl.AccumVal)) = *((XxxCoreADTClass::DblValue*)NewVal);
  }
 else if (OutputType == XXX_STRING || OutputType == XXX_FIXEDSTRING) {
   // the first string automatically becomes the maximum so far.
   if(Hndl.Started == XXX_FALSE) {
     Hndl.Started = XXX_TRUE;
     *((XxxCoreADTClass::StrValue*)(Hndl.AccumVal)) = *((XxxCoreADTClass::StrValue*)NewVal);
   }
   else if (((XxxCoreADTClass::StrValue*)NewVal)->Value > ((XxxCoreADTClass::StrValue*)(Hndl.AccumVal))->Value)
     *((XxxCoreADTClass::StrValue*)(Hndl.AccumVal)) = *((XxxCoreADTClass::StrValue*)NewVal);
  }
 else {
   XXX_NOTE_ERROR(XXX_BADTYPE, "For MAX aggregate");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

inline XxxErrCode XxxCountOpSpec::Init(const XxxFuncExpression::AggrPlan* APlan,
				       XxxAggrOpHandle& Hndl) const
{
 Hndl.Started = XXX_TRUE;
 // for COUNT, the initial valueis 0, not null
 Hndl.AccumVal->Unset(XxxADTValue::NULL_VALUE);
 ((XxxCoreADTClass::IntValue*)(Hndl.AccumVal))->Value = 0;
 return XXX_OK;
}

inline XxxErrCode XxxCountOpSpec::Accum(const XxxFuncExpression::AggrPlan* APlan,
					XxxAggrOpHandle& Hndl)const
{
 ((XxxCoreADTClass::IntValue*)(Hndl.AccumVal))->Value += 1;
 return XXX_OK;
}

inline XxxErrCode XxxCountOpSpec::Drop(const XxxFuncExpression::AggrPlan* APlan,
				       XxxAggrOpHandle& Hndl) const
{
 if (!Hndl.Started) {
   XXX_NOTE_ERROR(XXX_INTERNAL, "Drop on an aggregate that has not started");
   return XXX_NOT_OK;
  }
 else {
   ((XxxCoreADTClass::IntValue*)(Hndl.AccumVal))->Value -= 1;
  }
 return XXX_OK;
}

inline XxxErrCode XxxSumOpSpec::Init(const XxxFuncExpression::AggrPlan* APlan,
				     XxxAggrOpHandle& Hndl) const
{
 XxxBasicType OutputType = APlan->BasicType();
 Hndl.Started = XXX_TRUE;
 Hndl.AccumVal->Set(XxxADTValue::NULL_VALUE);
 if (OutputType == XXX_INT)
   ((XxxCoreADTClass::IntValue*)(Hndl.AccumVal))->Value = 0;
 else if (OutputType == XXX_DOUBLE)
   ((XxxCoreADTClass::DblValue*)(Hndl.AccumVal))->Value = 0.0;
 else {
   XXX_NOTE_ERROR(XXX_BADTYPE, "For SUM aggregate");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

inline XxxErrCode XxxSumOpSpec::Accum(const XxxFuncExpression::AggrPlan* APlan,
				      XxxAggrOpHandle& Hndl) const
{
 XxxBasicType OutputType = APlan->BasicType();
 XxxADTValue* NewVal = APlan->ValueEnv->ArgValsPtrs[0];

 if (NewVal->Is(XxxADTValue::NULL_VALUE)) {
   // nothing to do if it is a null value
   return XXX_OK;
 }
 if (!Hndl.Count) {
   // make sure to clear the null bit
   Hndl.AccumVal->Unset(XxxADTValue::NULL_VALUE);
 }

 if (OutputType == XXX_INT) {
   ((XxxCoreADTClass::IntValue*)(Hndl.AccumVal))->Value += ((XxxCoreADTClass::IntValue*)NewVal)->Value;
  }
 else if (OutputType == XXX_DOUBLE) {
   ((XxxCoreADTClass::DblValue*)(Hndl.AccumVal))->Value += ((XxxCoreADTClass::DblValue*)NewVal)->Value;
  }
 else {
   XXX_NOTE_ERROR(XXX_BADTYPE, "For SUM aggregate");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

inline XxxErrCode XxxSumOpSpec::Drop(const XxxFuncExpression::AggrPlan* APlan,
				     XxxAggrOpHandle& Hndl) const
{
 if (!Hndl.Started) {
   XXX_NOTE_ERROR(XXX_INTERNAL, "Drop on an aggregate that has not started");
   return XXX_NOT_OK;
  }
 else {
   XxxADTValue* NewVal = APlan->ValueEnv->ArgValsPtrs[0];
   XxxBasicType OutputType = APlan->BasicType();
   if (OutputType == XXX_INT) {
     ((XxxCoreADTClass::IntValue*)(Hndl.AccumVal))->Value -= ((XxxCoreADTClass::IntValue*)NewVal)->Value;
    }
   else if (OutputType == XXX_DOUBLE) {
     ((XxxCoreADTClass::DblValue*)(Hndl.AccumVal))->Value -= ((XxxCoreADTClass::DblValue*)NewVal)->Value;
    }
   else {
     XXX_NOTE_ERROR(XXX_BADTYPE, "For SUM aggregate");
     return XXX_NOT_OK;
    }
  }
 return XXX_OK;
}

inline XxxErrCode XxxProdOpSpec::Init(const XxxFuncExpression::AggrPlan* APlan,
				      XxxAggrOpHandle& Hndl) const
{
 Hndl.Started = XXX_TRUE;
 XxxBasicType OutputType = APlan->BasicType();
 Hndl.AccumVal->Set(XxxADTValue::NULL_VALUE);
 if (OutputType == XXX_INT)
   ((XxxCoreADTClass::IntValue*)(Hndl.AccumVal))->Value = 1;
 else if (OutputType == XXX_DOUBLE)
   ((XxxCoreADTClass::DblValue*)(Hndl.AccumVal))->Value = 1.0;
 else {
   XXX_NOTE_ERROR(XXX_BADTYPE, "For PROD aggregate");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}
inline XxxErrCode XxxProdOpSpec::Accum(const XxxFuncExpression::AggrPlan* APlan,
				       XxxAggrOpHandle& Hndl) const
{
 XxxBasicType OutputType = APlan->BasicType();
 XxxADTValue* NewVal = APlan->ValueEnv->ArgValsPtrs[0];

 if (NewVal->Is(XxxADTValue::NULL_VALUE)) {
   // nothing to do if it is a null value
   return XXX_OK;
 }
 if (!Hndl.Count) {
   // make sure to clear the null bit
   Hndl.AccumVal->Unset(XxxADTValue::NULL_VALUE);
 }
 if (OutputType == XXX_INT) {
   ((XxxCoreADTClass::IntValue*)(Hndl.AccumVal))->Value *= ((XxxCoreADTClass::IntValue*)NewVal)->Value;
  }
 else if (OutputType == XXX_DOUBLE) {
   ((XxxCoreADTClass::DblValue*)(Hndl.AccumVal))->Value *= ((XxxCoreADTClass::DblValue*)NewVal)->Value;
  }
 else {
   XXX_NOTE_ERROR(XXX_BADTYPE, "For PROD aggregate");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

inline XxxErrCode XxxProdOpSpec::Drop(const XxxFuncExpression::AggrPlan* APlan,
				      XxxAggrOpHandle& Hndl) const
{
 if (!Hndl.Started) {
   XXX_NOTE_ERROR(XXX_INTERNAL, "Drop on an aggregate that has not started");
   return XXX_NOT_OK;
  }
 else {
   XxxBasicType OutputType = APlan->BasicType();
   XxxADTValue* NewVal = APlan->ValueEnv->ArgValsPtrs[0];
   if (OutputType == XXX_INT) {
     ((XxxCoreADTClass::IntValue*)(Hndl.AccumVal))->Value /= ((XxxCoreADTClass::IntValue*)NewVal)->Value;
    }
   else if (OutputType == XXX_DOUBLE) {
     ((XxxCoreADTClass::DblValue*)(Hndl.AccumVal))->Value /= ((XxxCoreADTClass::DblValue*)NewVal)->Value;
    }
   else {
     XXX_NOTE_ERROR(XXX_BADTYPE, "For PROD aggregate");
     return XXX_NOT_OK;
    }
  }
 return XXX_OK;
}

// Methods of AvgOpSpec
inline XxxErrCode XxxAvgOpSpec::Init(const XxxFuncExpression::AggrPlan* APlan,
				     XxxAggrOpHandle& Hndl) const
{
 Hndl.Started = XXX_TRUE;
 XxxBasicType OutputType = APlan->BasicType();
 AvgInfo* Tmp = new AvgInfo();
 Tmp->Val = 0.0;
 Hndl.State = Tmp;
 Hndl.AccumVal->Set(XxxADTValue::NULL_VALUE);
 return XXX_OK;
}

inline XxxErrCode XxxAvgOpSpec::Terminate(const XxxFuncExpression::AggrPlan* APlan,
					  XxxAggrOpHandle& Hndl, 
					  XxxADTValue *FinalVal) const
{
 Hndl.Started = XXX_FALSE;
 XxxBasicType OutputType = APlan->BasicType();
 AvgInfo* StateInfo = (AvgInfo*) (Hndl.State);
 if (OutputType == XXX_INT) {
   ((XxxCoreADTClass::IntValue*)FinalVal)->Value = (int)(floor(StateInfo->Val));
  }
 else if (OutputType == XXX_DOUBLE) {
   ((XxxCoreADTClass::DblValue*)FinalVal)->Value = StateInfo->Val;
  }
 else {
   XXX_NOTE_ERROR(XXX_BADTYPE, "For AVG aggregate");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

inline XxxErrCode XxxAvgOpSpec::Accum(const XxxFuncExpression::AggrPlan* APlan,
				      XxxAggrOpHandle& Hndl) const
{
 XxxBasicType OutputType = APlan->BasicType();
 XxxBasicType InputType = (*(APlan->ArgPlans))[0]->BasicType();

 AvgInfo* Tmp = (AvgInfo*)(Hndl.State);
 XxxADTValue* NewVal = APlan->ValueEnv->ArgValsPtrs[0];

 if (NewVal->Is(XxxADTValue::NULL_VALUE)) {
   // nothing to do if it is a null value
   // but decrement the Hndl.Count, because it should not be
   // counted towards the average
   Hndl.Count--;
   return XXX_OK;
 }
 if (!Hndl.Count) {
   // make sure to clear the null bit
   Hndl.AccumVal->Unset(XxxADTValue::NULL_VALUE);
 }

 if ((OutputType == XXX_INT) && (InputType == XXX_INT)) {
   Tmp->Val = (Tmp->Val*(Hndl.Count) + 
	    ((XxxCoreADTClass::IntValue*)NewVal)->Value)/ (Hndl.Count+1);
  }
 else if ((OutputType == XXX_DOUBLE) && (InputType == XXX_DOUBLE)) {
   Tmp->Val = 
     (Tmp->Val*(Hndl.Count) + ((XxxCoreADTClass::DblValue*)NewVal)->Value)/(Hndl.Count+1);
  }
 else {
   XXX_NOTE_ERROR(XXX_BADTYPE, "For AVG aggregate");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

inline XxxErrCode XxxAvgOpSpec::Drop(const XxxFuncExpression::AggrPlan* APlan,
				     XxxAggrOpHandle& Hndl) const
{
 if (!Hndl.Started) {
   XXX_NOTE_ERROR(XXX_INTERNAL, "Drop on an aggregate that has not started");
   return XXX_NOT_OK;
  }

 XxxBasicType OutputType = APlan->BasicType();
 XxxBasicType InputType = (*(APlan->ArgPlans))[0]->BasicType();
 AvgInfo* Tmp = (AvgInfo*)(Hndl.State);
 if (Hndl.Count == 1) {
   Tmp->Val = 0.0;
 }
 else {
   XxxADTValue* NewVal = APlan->ValueEnv->ArgValsPtrs[0];
   if ((OutputType == XXX_INT) && (InputType == XXX_INT)) {
     Tmp->Val = (Tmp->Val*(Hndl.Count) - 
		 ((XxxCoreADTClass::IntValue*)NewVal)->Value)/(Hndl.Count-1);
   }
   else if ((OutputType == XXX_DOUBLE) && (InputType == XXX_DOUBLE)) {
     Tmp->Val = (Tmp->Val*(Hndl.Count) - 
		 ((XxxCoreADTClass::DblValue*)NewVal)->Value)/(Hndl.Count-1);
   }
   else {
     XXX_NOTE_ERROR(XXX_BADTYPE, "For AVG aggregate");
     return XXX_NOT_OK;
   }
 }
 return XXX_OK;
}

#endif
