#include "exprplan.h"
#include "error.h"
#include "valbool.h"

//
// XxxBooleanExpressionPlan
//

static XxxBool IsAndExpr(const XxxBooleanExpressionPlan* Expr)
{
 if (Expr->BoolType == XXX_BOOL_BOOL_EXPR) {
   XxxBoolBooleanPlan *BoolExpr = (XxxBoolBooleanPlan *)Expr;
   if (BoolExpr->OpType == XXX_AND_OP) 
     return XXX_TRUE;
  }
 return XXX_FALSE;
}

XxxErrCode XxxBooleanExpressionPlan::
MakeConjunction(XxxBooleanExpressionPlan* OldExpr,
		XxxBooleanExpressionPlan* NewExpr,
		XxxBooleanExpressionPlan*& ResultExpr,
		XxxBool& DeleteOld) 
{
 DeleteOld = XXX_FALSE;
 if (!NewExpr) {
   ResultExpr = OldExpr;
   return XXX_OK;
  }
 if (!OldExpr) {
   ResultExpr = NewExpr;
   return XXX_OK;
  }

 int NumChildren = 0;
 int CurrIndex = 0;
 int i;
 XxxBooleanExpressionPlan **Tmp;
 XxxBoolBooleanPlan *BoolOldExpr, *BoolNewExpr;
 XxxBool AndOldExpr = IsAndExpr(OldExpr);
 XxxBool AndNewExpr = IsAndExpr(NewExpr);
 // Note: when this is called from the optimizer, NewExpr is always
 // an atomic boolean factor. However, when called from the runtime
 // (spjplanop.C) NewExpr can itself be a conjunct.

 if (AndOldExpr) {
   BoolOldExpr = (XxxBoolBooleanPlan *)OldExpr;
   NumChildren += BoolOldExpr->Arity();
  }
 else {
   NumChildren += 1;
  }
 if (AndNewExpr) {
   BoolNewExpr = (XxxBoolBooleanPlan *)NewExpr;
   NumChildren += BoolNewExpr->Arity();
  }
 else {
   NumChildren += 1;
  }
 Tmp = new XxxBooleanExpressionPlan*[NumChildren];

 if (AndOldExpr) {
   for (i = 0; i < BoolOldExpr->Arity(); i++) {
     Tmp[CurrIndex++]=(XxxBooleanExpressionPlan*)(BoolOldExpr->SubPlans[i]);
     // We don't want the subplans being deleted
     BoolOldExpr->SubPlans[i] = NULL;
    }
   // in almost all cases, we will want to reclaim OldExpr, if it
   // is already an AND expr
   DeleteOld = XXX_TRUE;
  }
 else {
   Tmp[CurrIndex++] = OldExpr;
  }
 if (AndNewExpr) {
   for (i = 0; i < BoolNewExpr->Arity(); i++) {
     Tmp[CurrIndex++] = (XxxBooleanExpressionPlan*)(BoolNewExpr->SubPlans[i]);
     // We don't want the subplans being deleted
     BoolNewExpr->SubPlans[i] = NULL;
    }
  }
 else {
   Tmp[CurrIndex++] = NewExpr;
  }

 XxxBoolBooleanPlan *TmpResult 
   = new XxxBoolBooleanPlan(XXX_AND_OP, NumChildren, Tmp);
 
 // cleanup
 delete [] Tmp;

 // order the clauses based on their cost and/or selectivity
 if (TmpResult->ReorderSubPlans() != XXX_OK) { 
   XXX_NOTE_ERROR(XXX_BADOPT, "XxxBooleanExpressionPlan::MakeConjunction:"
		  "Error reordering subplans");
   return XXX_NOT_OK;
 }

 ResultExpr = TmpResult;
 return XXX_OK;
}

XxxErrCode XxxBooleanExpressionPlan::
MakeConjunction(XxxBooleanExpressionPlan*& OldExpr,
		XxxBooleanExpressionPlan* NewExpr)
{
 XxxBooleanExpressionPlan* NewPlan;
 XxxBool DeleteOld = XXX_FALSE;
 if (MakeConjunction(OldExpr, NewExpr, NewPlan, DeleteOld) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADEXPR, "Error making conjunction");
   return XXX_NOT_OK;
  }
 if (DeleteOld && OldExpr) delete OldExpr;
 OldExpr = NewPlan;
 return XXX_OK;
}

//
// XxxBoolBooleanPlan
//

XxxBoolBooleanPlan::XxxBoolBooleanPlan(XxxBoolBooleanExpression* BExpr,
				       XxxBooleanExpressionPlan* Children[])
: XxxBooleanExpressionPlan(BExpr, (XxxValueExprPlan**)Children)
{
 OpType = BExpr->OpType;
 SubPlanOrder = new int[Arity()];
 for (int i = 0; i < Arity(); i++) {
   SubPlanOrder[i] = i;
   if (Children[i])
     getPlanUsage().Add(Children[i]->GetPlanUsage());
 } // for
}

XxxBoolBooleanPlan::XxxBoolBooleanPlan(const XxxBoolBooleanPlan &Plan) :
  XxxBooleanExpressionPlan(Plan), OpType(Plan.OpType)
{
  SubPlanOrder = new int[Arity()];
  int i;
  for (i = 0; i < Arity(); i++) {
    SubPlanOrder[i] = Plan.SubPlanOrder[i];
  }
}

// are all the children constant?
static XxxBool AllConstant(int NumChildren,
			   XxxBooleanExpressionPlan* ChildExpr[])
{
 for (int i = 0; i < NumChildren; i++) {
   if (!ChildExpr[i]->IsConstant()) return XXX_FALSE;
  }
 return XXX_TRUE;
}

// this constructor is called from MakeConjunction()
XxxBoolBooleanPlan::
XxxBoolBooleanPlan(XxxBoolOpType Operation,
		   int NumChildren,
		   XxxBooleanExpressionPlan* ChildExpr[])
: XxxBooleanExpressionPlan(XXX_BOOL_BOOL_EXPR,
			   NumChildren, 
			   AllConstant(NumChildren, ChildExpr),
			   (XxxValueExprPlan**)ChildExpr)
{
 OpType = Operation;
 SubPlanOrder = new int[Arity()];
 for (int i = 0; i < Arity(); i++) {
   SubPlanOrder[i] = i;
  }

 ASSERT((NumChildren > 1));

 // now we need to set the dependency bits & the printBuf
 ostrstream Strm(printBuf, 100);
 switch (OpType) {
  case XXX_AND_OP: {
    // The arity of AND-OP is no longer limited to just two
    ChildExpr[0]->dependency.Or(ChildExpr[1]->dependency, dependency);
    Strm << ChildExpr[0]->printBuf;
    for (int i = 1; i < NumChildren; i++) {
      if (i < NumChildren-1) 
	dependency.Or(ChildExpr[i+1]->dependency, dependency);
      Strm << " AND " << ChildExpr[i]->printBuf;
     } // for
    Strm  << '\0';
    break;
   }
  case XXX_OR_OP: {
    ASSERT((NumChildren == 2));
    ChildExpr[0]->dependency.Or(ChildExpr[1]->dependency,
				dependency);
    Strm << ChildExpr[0]->printBuf << " OR " << ChildExpr[1]->printBuf << '\0';
    break;
   }
  case XXX_NOT_OP: {
    ASSERT((NumChildren == 1));
    dependency.CopyBits(ChildExpr[0]->dependency);
    Strm << " NOT " << ChildExpr[0]->printBuf << '\0';
    break;
   }
  case XXX_TRUE_OP: {
    ASSERT((NumChildren == 1));
    Strm << " TRUE " << '\0';
    break;
   }
  case XXX_FALSE_OP: {
    ASSERT((NumChildren == 1));
    Strm << " FALSE " << '\0';
    break;
   }
  } // switch
}

XxxBoolBooleanPlan::~XxxBoolBooleanPlan()
{
 delete [] SubPlanOrder;
}

XxxErrCode 
XxxBoolBooleanPlan::Evaluate(const XxxRecord* InputRecords[],
			     const XxxRecordSchema* InputSchemas[],
			     XxxADTValue *Result,
			     const XxxValueEnv* Env) const 
{
 int i;

 // interpret result as a ptr to a XxxBool
 XxxCoreADTClass::BoolValue* Answer = (XxxCoreADTClass::BoolValue*)Result;
 Answer->Value = XXX_FALSE;
 Answer->Init();

 switch (OpType) {
  case XXX_AND_OP: {
    for (i = 0; i < Arity(); i++) {
      // reset the null bit for each subexpression
      Answer->Init();
      if (SubPlans[SubPlanOrder[i]]->Evaluate(InputRecords, InputSchemas, 
					      Answer, Env) != XXX_OK) {
	return XXX_NOT_OK;
       }
      if (Answer->Is(XxxADTValue::NULL_VALUE) ||
	  (Answer->Value != XXX_TRUE)) {
	Answer->Value = XXX_FALSE;
	Answer->Unset(XxxADTValue::NULL_VALUE);
	break;
      }
     } // for all sub-clauses
    break;
   }
  case XXX_OR_OP: {
    for (i = 0; i < Arity(); i++) {
      // reset the null bit for each subexpression
      Answer->Init();
      if (SubPlans[SubPlanOrder[i]]->Evaluate(InputRecords, InputSchemas, 
					      Answer, Env) != XXX_OK) {
	return XXX_NOT_OK;
       }
      if (!Answer->Is(XxxADTValue::NULL_VALUE) &&
	  (Answer->Value == XXX_TRUE))
	break;
     } // for all sub-clauses
    break;
   }
  case XXX_NOT_OP: {
    if (SubPlans[0]->Evaluate(InputRecords, InputSchemas, 
			      Answer, Env) != XXX_OK) {
      return XXX_NOT_OK;
     }
    Answer->Value = ((Answer->Value == XXX_TRUE) ? XXX_FALSE : XXX_TRUE);
   break;
   }
  case XXX_TRUE_OP: {
    Answer->Value = XXX_TRUE;
    break;
   }
  case XXX_FALSE_OP: {
    Answer->Value = XXX_FALSE;
    break;
   }

  } // switch

 return XXX_OK;
}

XxxErrCode XxxBoolBooleanPlan::ReorderSubPlans() 
{
 SortItem *SortArray = new SortItem[Arity()];
 
 // TODO: get this from XXX_RELENV() instead of hardcoding it
 if (0) {
   // selectivity-based
   if (XXX_THISENV().DebugFlags.Expression) {
     cerr << "Ordering selections based on SELECTIVITY" << endl;
   }
   for(int i=0; i<Arity(); i++) {
     SortArray[i].Key = ((XxxBooleanExpressionPlan*)SubPlans[i])->Selectivity;
     SortArray[i].Index = i;
    }
  } 
 else if (0) {
   // cost-based
   if (XXX_THISENV().DebugFlags.Expression) {
     cerr << "Ordering selections based on COST" << endl;
   }
   for(int i=0; i<Arity(); i++) {
     double Cost = SubPlans[i]->GetPlanUsage().FullCost();
     SortArray[i].Key = Cost;
     SortArray[i].Index = i;
    }
  } 
 else {
   // rank-based 
   if (XXX_THISENV().DebugFlags.Expression) {
     cerr << "Ordering selections based on RANK" << endl;
   }
   for(int i=0; i<Arity(); i++) {
     // Rank = (Selectivity-1)/Cost
     double Cost = SubPlans[i]->GetPlanUsage().FullCost();
     double Rank = -1.0;
     if (Cost>0)
       Rank = (((XxxBooleanExpressionPlan*)SubPlans[i])->Selectivity - 1.0)/Cost;
     SortArray[i].Key = Rank;
     SortArray[i].Index = i;
    }
  }
 
 // now sort the thing
 qsort(SortArray, Arity(), sizeof(SortItem), SortItem::CompareVoid);
 
 // set the SubPlanOrder and cleanup
 for(int i=0; i<Arity(); i++) {
   SubPlanOrder[i] = SortArray[i].Index;
  }
 
 delete[] SortArray;
 
 return XXX_OK;
}

//
// XxxRelBooleanPlan
//
XxxRelBooleanPlan::XxxRelBooleanPlan(XxxRelBooleanExpression* RExpr,
				     XxxValueExprPlan* Children[])
: XxxBooleanExpressionPlan(RExpr, Children)
{
 ArgL = NULL;
 ArgR = NULL;

 // arity is always 2 !!
 TypeL = SubPlans[0]->BasicType();
 MetaInfoL = SubPlans[0]->GenerateCatInfo();
 TypeR = SubPlans[1]->BasicType();
 MetaInfoR = SubPlans[1]->GenerateCatInfo();

 ArgL = (XxxADTValue *)(new char[XxxADTArray[TypeL]->MaxObjectSize(MetaInfoL)]);
 ArgL->Init();
 ArgR = (XxxADTValue *)(new char[XxxADTArray[TypeL]->MaxObjectSize(MetaInfoR)]);
 ArgR->Init();

 OperType = RExpr->OpType;
 
 InvertOp(OperType, OperType, Inverted);

 // add usages of children
 for (int i = 0; i < Arity(); i++) {
   if (Children[i])  getPlanUsage().Add(Children[i]->GetPlanUsage());
 }
}

XxxRelBooleanPlan::~XxxRelBooleanPlan()
{
 // reclaim space of this plan
 if (ArgL) delete [] ((char *)ArgL);
 if (ArgR) delete [] ((char *)ArgR);
}

XxxErrCode XxxRelBooleanPlan::InvertOp(XxxRelOpType OrigOp,
				       XxxRelOpType& FinalOp,
				       XxxBool& Inverted)
{
 // implement only GT, GEQ, EQ (implement LT,LEQ,NEQ by inverting)
 switch (OrigOp) {
  case XXX_LEQ_OP:
   Inverted = XXX_TRUE;
   FinalOp = XXX_GT_OP;
   break;
  case XXX_LT_OP:
   Inverted = XXX_TRUE;
   FinalOp = XXX_GEQ_OP;
   break;
  case XXX_NEQ_OP:
   Inverted = XXX_TRUE;
   FinalOp = XXX_EQ_OP;
   break;
  default:
    FinalOp = OrigOp;
    Inverted = XXX_FALSE;
  } // switch
 return XXX_OK;
}

XxxErrCode 
XxxRelBooleanPlan::ExprEval(const XxxADTValue* ArgL,
			    XxxBasicType TypeL,
			    const XxxADTMetaInfo* MetaInfoL,
			    const XxxADTValue* ArgR,
			    XxxBasicType TypeR,
			    const XxxADTMetaInfo* MetaInfoR,
			    XxxRelOpType OperType,
			    XxxBool Inverted,
			    XxxBool& ExprResult)
{
 // check if either input is a NULL
 if (ArgL->Is(XxxADTValue::NULL_VALUE) ||
     ArgR->Is(XxxADTValue::NULL_VALUE)) {
   // the result is false:
   ExprResult = XXX_FALSE;
   return XXX_OK;
 }

 XxxCoreADTClass::IntValue *IntArg[2];
 XxxCoreADTClass::DblValue *DoubleArg[2];
 XxxCoreADTClass::StrValue *StringArg[2];
 IntArg[0] = (XxxCoreADTClass::IntValue *)ArgL;
 IntArg[1] = (XxxCoreADTClass::IntValue *)ArgR;
 DoubleArg[0] = (XxxCoreADTClass::DblValue *)ArgL;
 DoubleArg[1] = (XxxCoreADTClass::DblValue *)ArgR;
 StringArg[0] = (XxxCoreADTClass::StrValue *)ArgL;
 StringArg[1] = (XxxCoreADTClass::StrValue *)ArgR;

 // implement only GT, GEQ, EQ (implement LT,LEQ,NEQ by inverting)
 switch (OperType) {
  case XXX_GT_OP: {
    if (XxxCoreADTClass::IsCoreADT(TypeL)) {
      ExprResult = XxxCoreADTClass::
	SGreaterThan(TypeL, ArgL, MetaInfoL, ArgR, MetaInfoR);
    }
    else {
      ExprResult = XxxADTArray[TypeL]->
	GreaterThan(ArgL, MetaInfoL, ArgR, MetaInfoR);
    }
   break;
  }
  case XXX_GEQ_OP: {
    if (XxxCoreADTClass::IsCoreADT(TypeL)) {
      ExprResult = !(XxxCoreADTClass::
	SLessThan(TypeL, ArgL, MetaInfoL, ArgR, MetaInfoR));
    }
    else {
      ExprResult = !(XxxADTArray[TypeL]->
	LessThan(ArgL, MetaInfoL, ArgR, MetaInfoR));
    }
    break;
  }
  case XXX_EQ_OP: {
    if (XxxCoreADTClass::IsCoreADT(TypeL)) {
      ExprResult = XxxCoreADTClass::
	SEquals(TypeL, ArgL, MetaInfoL, ArgR, MetaInfoR);
    }
    else {
      ExprResult = XxxADTArray[TypeL]->
	Equals(ArgL, MetaInfoL, ArgR, MetaInfoR);
    }
    break;
   }
  case XXX_LIKE_OP:
  default: {
    char buf[100];
    sprintf(buf, "Bad operator type %d!!\n", OperType);
    XXX_NOTE_ERROR(XXX_BADEXPR, buf);
    return XXX_NOT_OK;
   }
  } // switch

 if (Inverted)
   ExprResult = ((ExprResult == XXX_TRUE)?XXX_FALSE:XXX_TRUE);
 return XXX_OK;
}

XxxErrCode 
XxxRelBooleanPlan::Evaluate(const XxxRecord* InputRecords[],
			    const XxxRecordSchema* InputSchemas[],
			    XxxADTValue *Result,
			    const XxxValueEnv* Env) const 
{
 // interpret result as a ptr to a XxxBool
 XxxCoreADTClass::BoolValue* Answer = (XxxCoreADTClass::BoolValue *)Result;

 Answer->Value = XXX_FALSE;

 if (SubPlans[0]->Evaluate(InputRecords, InputSchemas, ArgL, Env) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating left subexpression");
   return XXX_NOT_OK;
  }
 if (SubPlans[1]->Evaluate(InputRecords, InputSchemas, ArgR, Env) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating right subexpression");
   return XXX_NOT_OK;
  }

 if (ExprEval(ArgL, TypeL, MetaInfoL, ArgR, TypeR, MetaInfoR, 
	      OperType, Inverted, Answer->Value) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating relational expression");
   return XXX_NOT_OK;
 }

 return XXX_OK;
}

XxxRelBooleanPlan::XxxRelBooleanPlan(const XxxRelBooleanPlan &Plan) :
  XxxBooleanExpressionPlan(Plan), TypeL(Plan.TypeL), 
  TypeR(Plan.TypeR), Inverted(Plan.Inverted), OperType(Plan.OperType)
  
{
  if (!Plan.MetaInfoL ||
      (Plan.MetaInfoL->Replicate((XxxADTMetaInfo *&)MetaInfoL) != XXX_OK)) {
    MetaInfoL = NULL;
  }
  if (!Plan.MetaInfoR ||
      (Plan.MetaInfoR->Replicate((XxxADTMetaInfo *&)MetaInfoR) != XXX_OK)) {
    MetaInfoR = NULL;
  }

  ArgL = (XxxADTValue *)(new char[XxxADTArray[TypeL]->MaxObjectSize(MetaInfoL)]);
  ArgL->Init();
  memcpy((char *)ArgL, (char *)(Plan.ArgL), 
	 XxxADTArray[TypeL]->MaxObjectSize(MetaInfoL));

  ArgR = (XxxADTValue *)(new char[XxxADTArray[TypeR]->MaxObjectSize(MetaInfoR)]);
  ArgR->Init();
  memcpy((char *)ArgR, (char *)(Plan.ArgR), 
	 XxxADTArray[TypeR]->MaxObjectSize(MetaInfoR));
}

//
// Methods of XxxValBooleanPlan
//
XxxErrCode 
XxxValBooleanPlan::Evaluate(const XxxRecord* InputRecords[],
			    const XxxRecordSchema* InputSchemas[],
			    XxxADTValue *Result,
			    const XxxValueEnv* Env) const 
{
  if (!SubPlans[0]) {
    XXX_NOTE_ERROR(XXX_BADOPT, "Missing subplan");
    return XXX_NOT_OK;
  }
  else
    return SubPlans[0]->Evaluate(InputRecords, InputSchemas,
				 Result, Env);
}
