#include "caseexpr.h"

XxxCaseExpression::
XxxCaseExpression(int NChildren,
		  XxxExpression *Children[])
  : XxxValueExpression(NChildren, Children)
{
}

XxxCaseExpression::
XxxCaseExpression(const XxxCaseExpression& src)
  : XxxValueExpression(src)
{
}

void XxxCaseExpression::Print(ostream& OStr) const 
{
  OStr << "CASE";

  for (int i = 0; i < arity-1; i+=2) {
    OStr << " WHEN ";
    ChildExpr(i)->Print(OStr);
    OStr << " THEN ";
    ChildExpr(i+1)->Print(OStr);
  }
  OStr << " ELSE ";
  ChildExpr(arity-1)->Print(OStr);
  OStr << " END ";
}

/**
 * TODO: check for metainfo?
 */
XxxErrCode XxxCaseExpression::TypeCheck()
{
  int i;
  XxxErrCode rc;

  // First type check the child.
  if ((rc = XxxExpression::TypeCheck()) != XXX_OK) {
    return rc;
  }

  // Make sure that there is at least one condition-expression pair
  // and a else expression.
  if (arity < 3) {
      XXX_NOTE_ERROR(XXX_BADTYPE, "Missing cases for Case Op!");
      return XXX_NOT_OK;
  }

  // Check that every condition expression has type boolean.
  for (i = 0; i < arity - 1; i+=2) {
    if (ChildExpr(i)->BasicType() != XXX_BOOLEAN) {
      XXX_NOTE_ERROR(XXX_BADTYPE, "Condition type must be boolean for Case Op!");
      return XXX_NOT_OK;
    }
  }

  // Check that every result expression has the same type.
  BaseType = ChildExpr(1)->BasicType();
  for (i = 3; i <= arity; i+=2) {
    // default case has no condition expression
    if (i == arity)
        i = arity - 1;
    switch (BaseType){
    case XXX_INT:
    case XXX_DOUBLE:
      // if any of result expressions are double, then set return type to double
      if (ChildExpr(i)->BasicType() == XXX_DOUBLE)
        BaseType = XXX_DOUBLE;
      else if (ChildExpr(i)->BasicType() != XXX_INT) {
        XXX_NOTE_ERROR(XXX_BADTYPE, "All return expressions must have the same type!");
        return XXX_NOT_OK;
      }
      break;
    default:
      if (ChildExpr(i)->BasicType() != BaseType) {
        XXX_NOTE_ERROR(XXX_BADTYPE, "All return expressions must have the same type!");
        return XXX_NOT_OK;
      }
      break;
    }
  }
  return XXX_OK;
}

XxxErrCode XxxCaseExpression::Optimize(XxxValueExprPlan*& ExprPlan)
{

 XxxValueExprPlan** TmpPlans = new XxxValueExprPlan*[Arity()];
 for (int i = 0; i < arity; i++)
   XXX_CALL_RET(((XxxValueExpression *)(ChildExpr(i)))->Optimize(TmpPlans[i]),
	      XXX_BADEXEC, "Unable to optimize case expression");

 CasePlan *CEPlan = new CasePlan(this, TmpPlans);
 ExprPlan = CEPlan;
 return XXX_OK;
} 

XxxCaseExpression::
CasePlan::CasePlan(XxxCaseExpression* CExpr,
		   XxxValueExprPlan* ChildrenPlan[])
  : XxxValueExprPlan(CExpr, ChildrenPlan) 
{ 
  int i;
  ConditionVal = new XxxCoreADTClass::BoolValue(XXX_FALSE);
  ChildType = SubPlans[1]->BasicType();

  // if any result expression are double, then set childtype to double
  for (i = 3; i <= arity; i+=2) {
    if (i == arity)
        i = arity - 1;
    if (SubPlans[i]->BasicType() == XXX_DOUBLE)
      ChildType = XXX_DOUBLE;
  }

  for (i = 0; i < arity; i++) {
    getPlanUsage().Add(SubPlans[i]->GetPlanUsage());
  }
}

XxxCaseExpression::
CasePlan::CasePlan(const CasePlan &Plan)
  : XxxValueExprPlan(Plan)
{
  ConditionVal = new XxxCoreADTClass::BoolValue(XXX_FALSE);
  ChildType = Plan.ChildType;
}

XxxErrCode XxxCaseExpression::
CasePlan::Evaluate(const XxxRecord* Records[],
		   const XxxRecordSchema* Schemas[],
		   XxxADTValue *Result,
		   const XxxValueEnv* Env) const 
{
  int i;
  XxxCoreADTClass::BoolValue *Cond = (XxxCoreADTClass::BoolValue*)ConditionVal;
  XxxCoreADTClass::DblValue *DbAnswer = (XxxCoreADTClass::DblValue*)Result;

  if (ChildType == XXX_DOUBLE)
    DbAnswer->Value = 0.0 ;

  // step through conditions until a matching one found
  for (i = 0; i < arity - 1; i+=2) {
    // Reset the null bit for each expression.
    Cond->Init();
    XXX_CALL_RET(SubPlans[i]->Evaluate(Records, Schemas, Cond, Env),
              XXX_BADEXPR, "Error evaluating case subexpression");
 
    // a matching condition.  return corresponding expression
    if (!Cond->Is(XxxADTValue::NULL_VALUE) && Cond->Value == XXX_TRUE) {
      XXX_CALL_RET(SubPlans[i+1]->Evaluate(Records, Schemas, Result, Env),
              XXX_BADEXPR, "Error evaluating case subexpression");

      return XXX_OK;
    }
  }

  // no matching condition.  return default expression
  XXX_CALL_RET(SubPlans[arity-1]->Evaluate(Records, Schemas, Result, Env),
            XXX_BADEXPR, "Error evaluating case subexpression");

  return XXX_OK;
}

XxxBool XxxCaseExpression::
Match(const XxxExpression *OtherExpr, char* Env) const
{
  if (arity != OtherExpr->Arity())
    return XXX_FALSE;
  for (int i = 0; i < arity; i++) {
    if (!ChildExpr(i)->Match(OtherExpr->ChildExpr(i)))
      return XXX_FALSE;
  }
  return XXX_TRUE;
}
