/********
 * This file defines different kinds of E-ADT transformation rules
 *     a) Reorganize rule
 *     b) Merge rule
 *     c) Rule to move a function out of an aggregate
 *     d) Rule to move a function into an aggregate
 *********/
#include <string.h>
#include "optrules.h"
#include "funcexpr.h"
#include "adtfunc.h"
#include "aggrexpr.h"

//
// XxxRule methods
//
XxxRule::XxxRule(RulePrecondFunc PCF, RuleActionFunc RAF, 
		 const char* Name, const XxxOptRuleEngine::TravDir TD)
{ 
 preCond = PCF; 
 act = RAF; 
 TDir=TD;
 if (Name) {
   strncpy(RuleName, Name, XXX_SUNDRYNAME_MAX);
  }
 else {
   strcpy(RuleName, "UNNAMED");
  }
 RuleName[XXX_SUNDRYNAME_MAX-1] = 0;
}

//
// Methods of XxxSimpleCondRule
//

// This constructor will set up this rule to use the default PreCondition
// function.
// The OwnerList is a list of ids of functions allowed to be owners of VExpr
// such that we can do the reorder.  Analogously, ThisList is a list of ids of
// functions that can be operated on (i.e. VExpr must be one of these types).
// ownerType and thisType are used for type compatibility, if we're trying to
// switch an image function whose owner is an audio, for example.
XxxSimpleCondRule::XxxSimpleCondRule(const char OwnerNames[][XXX_FUNCNAME_MAX], 
				     const int NumOwners, 
				     const XxxBasicType OwnerType, 
				     const char ThisNames[][XXX_FUNCNAME_MAX],
				     const int NumThis, 
				     const XxxBasicType ThisType, 
				     RuleActionFunc RAF,
				     const char* Name, 
				     const XxxOptRuleEngine::TravDir TD) 
: XxxRule(NULL, RAF, Name, TD)
{
 ownerList=NULL;
 if (NumOwners > 0) {
   ownerList =new char*[NumOwners];
  }
 int i;
 for (i=0; i < NumOwners; i++) {
   ownerList[i] = new char[strlen(OwnerNames[i])+1];
   strcpy(ownerList[i],OwnerNames[i]);
  }
 ownerListLen = NumOwners;
 ownerType = OwnerType;
 thisList = new char*[NumThis];
 for (i=0; i < NumThis; i++) {
   thisList[i] = new char[strlen(ThisNames[i])+1];
   strcpy(thisList[i], ThisNames[i]);
  }
 thisListLen = NumThis;
 thisType = ThisType;
}

XxxSimpleCondRule::~XxxSimpleCondRule() 
{
 if (ownerList) {
   for (int i=0; i < ownerListLen; i++) {
     delete [] ownerList[i];
    }
   delete [] ownerList;
   ownerList = NULL;
  }
 if (thisList) {
   for (int i=0; i < thisListLen; i++) {
     delete [] thisList[i];
    }
   delete [] thisList;
   thisList = NULL;
  }
}

XxxErrCode
XxxSimpleCondRule::PreCondition(XxxValueExpression *VExpr, XxxBool& Success,
				void*& RuleData) const
{
 Success = XXX_FALSE;
 // this must be a function expression
 if (VExpr->ValType() != XxxValueExpression::FUNC_EXPR) {
   return XXX_OK;
  }
 XxxFuncExpression* FExpr = (XxxFuncExpression*) VExpr;
 XxxFuncParseInfo *ThisFuncParse = FExpr->FuncParse;
 
 // Make sure owner exists
 if (!FExpr->Owner)
   return XXX_OK;
 
 XxxValueExpression* Owner = FExpr->Owner;
 // Owner must also be a function
 if (Owner->ValType() != XxxValueExpression::FUNC_EXPR) {
   return XXX_OK;
  }
 XxxFuncExpression* OwnerExpr = (XxxFuncExpression*) (FExpr->Owner);
 XxxFuncParseInfo *OwnerFuncParse = OwnerExpr->FuncParse;
 
 // check if the ADT types of this function and its owner match the
 // requirement
 if (ThisFuncParse->TypeId!=thisType || OwnerFuncParse->TypeId!=ownerType)
   return XXX_OK;
 
 // check if this function is among those allowed to use this rule.
 int i;
 for (i=0; i < thisListLen; i++) {
   if (!strcmp(thisList[i], ThisFuncParse->FuncName))
     break;
  }
 if (i == thisListLen) {
   // not eligible, so return
   return XXX_OK;
  }
 
 // check if the owner is among those allowed to use this rule
 for (i=0; i < ownerListLen; i++) {
   // owner not eligible, so return
   if (!strcmp(OwnerFuncParse->FuncName, ownerList[i]))
     break;
  }
 if (i == ownerListLen) {
   // not eligible, so return
   return XXX_OK;
  }
 
 // if we've gotten this far, we're done, and we can act on this expression
 RuleData = NULL;     // no extra rule data to pass
 Success = XXX_TRUE;  // condition fired
 return XXX_OK;
}

//
// XxxFuncReorderRule methods
//
XxxErrCode
XxxFuncReorderRule::Action(XxxValueExpression *InVExpr, void* RuleData,
			   XxxValueExpression*& OutVExpr) const 
{
 // Swap all the fields of the functon expression with its owner

 // Since we know we can apply the rule, InVExpr must be a funcexpression
 XxxFuncExpression *InFExpr = (XxxFuncExpression*) InVExpr;
 XxxFuncExpression* OwnerFunc = (XxxFuncExpression*)(InFExpr->Owner);

 char SwapName[XXX_FUNCNAME_MAX];
 
 strncpy(SwapName, InFExpr->Name, XXX_FUNCNAME_MAX);
 SwapName[XXX_FUNCNAME_MAX-1] = 0;
 XxxValueExprList* SwapArgList = InFExpr->ArgList;
 XxxFuncParseInfo *SwapFuncParse = InFExpr->FuncParse;
 
 // Swap the FuncParses
 InFExpr->FuncParse = OwnerFunc->FuncParse;
 OwnerFunc->FuncParse = SwapFuncParse;
 // set the IsTopLevel flag to false, to prevent future FuncReorganize
 // calls from doing the rewrite all over again.
 OwnerFunc->FuncParse->IsTopLevel = XXX_FALSE;
 // TODO:L2: what about setting InFExpr->FuncParse->IsTopLevel to true??

 // Swap the two argument list 
 InFExpr->ArgList = OwnerFunc->ArgList;
 OwnerFunc->ArgList = SwapArgList;
 
 // Swap the text print of the method type of each
 strncpy(InFExpr->Name, OwnerFunc->Name, XXX_FUNCNAME_MAX);
 InFExpr->Name[XXX_FUNCNAME_MAX-1] = 0;
 strncpy(OwnerFunc->Name, SwapName, XXX_FUNCNAME_MAX);
 OwnerFunc->Name[XXX_FUNCNAME_MAX-1] = 0;
 
 OutVExpr = InFExpr;
 if (XXX_THISENV().DebugFlags.ADTRules) {
   cerr << "Commuted " << InFExpr->Name << " and " << OwnerFunc->Name << endl;
   InFExpr->Print(cerr);
   cerr << endl;
 }
 
 return XXX_OK;
}

//
// Methods of XxxMergeRule 
//
XxxMergeRule::XxxMergeRule(const char OwnerID[][XXX_FUNCNAME_MAX],
			   const int NumOwners, 
			   const XxxBasicType OwnerType,
			   const char ThisID[][XXX_FUNCNAME_MAX],
			   const int NumThis, 
			   const XxxBasicType ThisType,
			   const ArgRule* AR[], 
			   int NumArgRules, Dominator Dom,
			   const char* Name,
			   const XxxOptRuleEngine::TravDir TD, 
			   const char* DestString)
: XxxSimpleCondRule(OwnerID, NumOwners, OwnerType, 
		    ThisID, NumThis, ThisType, 0, Name, TD) 
{
 argRuleArray = NULL;
 destName[0] = 0;
 argRuleArrayLen = NumArgRules;
 dominator = Dom;
 
 if (argRuleArrayLen > 0) {
   argRuleArray = new const ArgRule*[argRuleArrayLen];
  }
 for (int i=0; i < argRuleArrayLen; i++) {
   argRuleArray[i] = AR[i];
  }
 if (DestString) {
   strcpy(destName, DestString);
  }
}

XxxMergeRule::~XxxMergeRule() 
{
 if (argRuleArray) {
   // reclaim each rule
   for (int i = 0; i < argRuleArrayLen; i++)
     if (argRuleArray[i]) delete argRuleArray[i];
   // reclaim array
   delete [] argRuleArray;
  }
}

// The merge action always deletes the Owner object. However, the
// merge rule may be used in cases where the desired behavior is
// to retain the owner and delete "this". In these cases, the
// fields of the owner are copied on to "this".
XxxErrCode 
XxxMergeRule::Action(XxxValueExpression *InVExpr, void* RuleData,
		     XxxValueExpression*& OutVExpr) const
{
 // Since we know we can apply the rule, InVExpr must be a funcexpression
 XxxFuncExpression *InFExpr = (XxxFuncExpression*) InVExpr;
 
 XxxValueExprList* ArgList = InFExpr->ArgList;
 XxxFuncExpression* Owner = (XxxFuncExpression*)(InFExpr->Owner);
 XxxFuncParseInfo *ThisFuncParse = InFExpr->FuncParse;

 // at the end we want to say what we merged, but InFExpr and Owner change
 // in many different ways, so we just copy the original names here to make
 // life simple.
 char OwnerName[XXX_FUNCNAME_MAX], ThisName[XXX_FUNCNAME_MAX];
 strncpy(OwnerName, Owner->Name, XXX_FUNCNAME_MAX);
 OwnerName[XXX_FUNCNAME_MAX-1] = 0;
 strncpy(ThisName, InFExpr->Name, XXX_FUNCNAME_MAX);
 ThisName[XXX_FUNCNAME_MAX-1] = 0;

 // if we are bypassing the "dominator" mechanism
 if (destName[0] != 0) {
   strncpy(InFExpr->Name, destName, XXX_FUNCNAME_MAX);
   InFExpr->Name[XXX_FUNCNAME_MAX-1] = 0;
   strncpy(InFExpr->FuncParse->FuncName, destName, XXX_FUNCNAME_MAX);
   InFExpr->FuncParse->FuncName[XXX_FUNCNAME_MAX-1] = 0;
  } 
 else if (dominator == OWNEREXP) {
   // the owner expression dominates, so copy its name into this exp
   strncpy(InFExpr->Name, Owner->Name, XXX_FUNCNAME_MAX);
   InFExpr->Name[XXX_FUNCNAME_MAX-1] = 0;
   // switch FuncParses
   InFExpr->FuncParse=Owner->FuncParse;
   Owner->FuncParse=ThisFuncParse;
  }

 XxxValueExprList* OwnerArgList = Owner->ArgList;

 // start going through the ArgRules and putting results in ResExprList
 // start off by replicating the dominator's arg list (if there is a
 // dominator).
 XxxValueExprList* ResExprList = NULL;
 if (dominator == THISEXP) {
   if (ArgList) {
     ResExprList = ArgList->Replicate();
    }
   else if (argRuleArrayLen > 0) {
     // dominator's arglist is NULL, so this merge rule should have 0
     // arg rules because any arg rule would end up putting something in
     // ResExprList, which later becomes InFExpr's arg list but in this
     // case, it should be null.
     XXX_NOTE_ERROR(XXX_BADVALUE, "dominator function takes no args, but argrules are still defined!");
     return XXX_NOT_OK;
    }
  } 
 else if (dominator == OWNEREXP) {
  if (Owner->ArgList) {
    ResExprList=OwnerArgList->Replicate();
   } 
  else if (argRuleArrayLen > 0) {
    // dominator's arglist is NULL, so this merge rule should have 0
    // arg rules because any arg rule would end up putting something in
    // ResExprList, which later becomes InFExpr's arg list...but in this
    // case, it should be null.
    XXX_NOTE_ERROR(XXX_BADVALUE, "dominator function takes no args, but argrules are still defined!");
    return XXX_NOT_OK;
   }
 } 
 else { // if there is a new merged function
   ResExprList = new XxxValueExprList(NULL);
  }

 // now step through the argument creation rules
 for (int i=0; i < argRuleArrayLen; i++) {
   const ArgRule* CurRule = argRuleArray[i];
   XxxValueExpression* ResExpr;
   if (!CurRule ||
       CurRule->CreateExpr(OwnerArgList, ArgList, ResExpr) != XXX_OK) {
     if (ResExprList) ResExprList->Destroy();
     XXX_NOTE_ERROR(XXX_BADOPT, "Error creating argument expression");
     return XXX_NOT_OK;
    }
   // store this result in ResExprList, and we're done
   if (ResExprList->SetElement(CurRule->DestArgNum, ResExpr) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADOPT, "Error adding argument expression");
     return XXX_NOT_OK;
    }
  } // for each rule

 // ResExprList is fully generated.  Let's get rid of the current ArgList and
 // make InFExpr point to the new arg list.
 if (InFExpr->ArgList)
   InFExpr->ArgList->Destroy();
 // if the resultlist contains nothing, we can delete it and just set the
 // expression's arglist to NULL.
 if (ResExprList && ResExprList->Expr==NULL && ResExprList->Next==NULL) {
   ResExprList->Destroy();
   ResExprList = NULL;
  } 
 InFExpr->ArgList = ResExprList;

 // delete Owner, but we don't want Grandpa to die, so we do a small hack
 XxxValueExpression *Grandpa = Owner->Owner;
 Owner->Owner = NULL;
 delete Owner; // recursive delete stops because of NULL Owner
 InFExpr->Owner = Grandpa;
 
 OutVExpr = InFExpr;
 
 if (XXX_THISENV().DebugFlags.ADTRules) {
   cerr << "Successful merge of " << OwnerName << " and " << ThisName << endl;
   InFExpr->Print(cerr);
   cerr << endl;
 }

 return XXX_OK;
}

//
// Methods of MergeRule::ArgRule and its subclasses
// 
XxxErrCode XxxMergeRule::
SimpleArgRule::CreateExpr(const XxxValueExprList* OwnerArgList,
			  const XxxValueExprList* ThisArgList,
			  XxxValueExpression*& NewExpr) const 
{
 XxxValueExpression *OwnerArg, *ThisArg;
 if (OwnerArgList->GetElement(OwnerArgNum, OwnerArg) != XXX_OK) {
   char buf[100];
   sprintf(buf,"Couldn't get %d'th element of owner arglist",
	   OwnerArgNum);
   XXX_NOTE_ERROR(XXX_BADOPT, buf);
   return XXX_NOT_OK;
  }
 if (ThisArgList->GetElement(ThisArgNum, ThisArg) != XXX_OK) {
   char buf[100];
   sprintf(buf,"Couldn't get %d'th element of this function's arglist",
	   ThisArgNum);
   XXX_NOTE_ERROR(XXX_BADOPT, buf);
   return XXX_NOT_OK;
  }
 if (createExpr(OwnerArg, ThisArg, NewExpr) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADOPT, "Error creating new argument");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode XxxMergeRule::
CopyArgRule::createExpr(const XxxValueExpression* OwnerArg,
			const XxxValueExpression* ThisArg,
			XxxValueExpression*& NewExpr) const 
{
 if (FirstExp==THISEXP) {
   NewExpr = (XxxValueExpression*)(ThisArg->Replicate());
  } 
 else {
  NewExpr = (XxxValueExpression*)(OwnerArg->Replicate());
  }
 return XXX_OK;
}

XxxErrCode XxxMergeRule::
ArithArgRule::createExpr(const XxxValueExpression* OwnerArg,
			 const XxxValueExpression* ThisArg,
			 XxxValueExpression*& NewExpr) const 
{
 XxxExpression *Children[2];
 if (FirstExp == THISEXP) {
   Children[0] = ThisArg->Replicate();
   Children[1] = OwnerArg->Replicate();
  } 
 else {
   Children[0] = OwnerArg->Replicate();
   Children[1] = ThisArg->Replicate();
  }

 if (Op == XXX_UMINUS_OP)
   NewExpr = new XxxArithExpression(Op, 1, Children);
 else
   NewExpr = new XxxArithExpression(Op, 2, Children);

 if (NewExpr->TypeCheck() != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADOPT, "Type check failed for new arith expression");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

// 
// Methods of MoveFuncIntoAggrRule
//
XxxErrCode 
XxxMoveFuncIntoAggrRule::PreCondition(XxxValueExpression *VExpr, 
				      XxxBool &Success,
				      void *&RuleData) const
{
 RuleData = NULL;
 Success = XXX_FALSE;
 if (VExpr->ValType() != XxxValueExpression::FUNC_EXPR) {
   return XXX_OK;
  } 
 // it is indeed a function expression
 XxxFuncExpression *FuncExpr = (XxxFuncExpression *)VExpr;
 XxxFuncParseInfo *FuncParse = (XxxFuncParseInfo *)FuncExpr->FuncParse;
    
 // make sure that the owner exists
 if (!FuncExpr->Owner) 
   return XXX_OK;

 XxxValueExpression* Owner = FuncExpr->Owner;
 if ((Owner->ValType() != XxxValueExpression::FUNC_EXPR) ||
     !(((XxxFuncExpression*)Owner))->IsAggr())
   // the owner is not an aggregate expression
   return XXX_OK;

 XxxAggrExpression *AggrExpr = (XxxAggrExpression *)Owner;
 XxxFuncParseInfo *AggrParse = (XxxFuncParseInfo *)AggrExpr->FuncParse;
    
 // check if the ADT types of this function and its owner match the 
 // requirement.  This check should never fail.
 if ((FuncParse->TypeId != thisType) || (AggrParse->TypeId != ownerType)) 
   return XXX_OK;

 // check if function is among those allowed to use this rule
 int i;
 for (i=0; i < thisListLen; i++) {
   if (!strcmp(thisList[i],FuncParse->FuncName))
     break;
  } // for

 if (i == thisListLen) {
   // nope, this function has not been specified to be interchangable 
   // with an aggregate
   return XXX_OK;
  }

 // see if the owner aggregate has been specified to use this optimization
 for (i=0; i < ownerListLen; i++) {
   // owner not eligible, so return
   if (!strcmp(AggrParse->FuncName, ownerList[i]))
     break;
  }
 if (i == ownerListLen) {
   // nope, this aggregate cannot be interchanged with a function
   return XXX_OK;
  }
 
 Success = XXX_TRUE;
 return XXX_OK;
}

XxxErrCode 
XxxMoveFuncIntoAggrRule::Action(XxxValueExpression *InVExpr, void *RuleData,
				XxxValueExpression *&OutVExpr) const
{
 // since we're at this point, we know that the InVExpr is a 
 // func-expression and its owner is an aggregate expression
 XxxFuncExpression *FuncExpr = (XxxFuncExpression *) InVExpr;
 XxxAggrExpression *AggrExpr = (XxxAggrExpression *) FuncExpr->Owner;

 XxxValueExpression *AggrArg = AggrExpr->Owner;

 // make the argument of the aggregate the owner of the function instead
 FuncExpr->Owner = AggrArg;
 FuncExpr->OwnerType = FuncExpr->Owner->BaseType;
 FuncExpr->OwnerTypeCatInfo = FuncExpr->Owner->GenerateCatInfo();
 FuncExpr->FuncParse->IsTopLevel = XXX_FALSE;

 // make the function expression the argument of the aggregate
 AggrExpr->Owner = FuncExpr;

 if (XXX_THISENV().DebugFlags.ADTRules) {
   cerr << "XxxMoveFuncIntoAggrRule successful!" << endl;
   AggrExpr->Print(cerr);
   cerr << endl;
 }

 // return the aggregate expression
 OutVExpr = AggrExpr;
 return XXX_OK;
}

// The XxxMoveFuncOutOfAggrRule Methods
XxxErrCode 
XxxMoveFuncOutOfAggrRule::PreCondition(XxxValueExpression *VExpr, 
				       XxxBool &Success,
				       void *&RuleData) const
{  
 RuleData = NULL;
 Success = XXX_FALSE;
 if ((VExpr->ValType() != XxxValueExpression::FUNC_EXPR) ||
     (!((XxxFuncExpression*)VExpr)->IsAggr())) {
   return XXX_OK;
  } 
    
 // it is indeed an aggregate expression
 XxxAggrExpression *AggrExpr = (XxxAggrExpression *)VExpr;
 XxxFuncParseInfo *AggrParse = (XxxFuncParseInfo *)AggrExpr->FuncParse;

 // now lets get its expression and see if it is a func-expr
 XxxValueExpression *Expr = AggrExpr->Owner;
 
 if (Expr->ValType() != XxxValueExpression::FUNC_EXPR) 
   // not a func-expr
   return XXX_OK;

 XxxFuncExpression *FuncExpr = (XxxFuncExpression *)Expr;
 XxxFuncParseInfo *FuncParse = (XxxFuncParseInfo *)FuncExpr->FuncParse;
    
 // make sure that the owner exists, although I don't think I need to
 // check for this case
 if (!FuncExpr->Owner)
   return XXX_OK;
    
 // check if the ADT types of this aggrexpr and its expr match the 
 // requirement.  This check should never fail.
 if ((AggrParse->TypeId != thisType) || (FuncParse->TypeId != ownerType)) 
   return XXX_OK;
    
 // check if the aggregate is among those allowed to use this rule
 int i;
 for (i=0; i < thisListLen; i++) {
   if (!strcmp(thisList[i],AggrParse->FuncName))
     break;
  } // for

 if (i == thisListLen) 
   // nope, this aggregate can't be commuted this way
   return XXX_OK;

 // see if the owner function has been specified to use this optimization
 for (i=0; i < ownerListLen; i++) {
   // owner not eligible, so return
   if (!strcmp(FuncParse->FuncName, ownerList[i]))
     break;
  } // for
    
 if (i == ownerListLen) 
   // nope, this aggregate cannot be interchanged with a function
   return XXX_OK;
    
 Success = XXX_TRUE;
 return XXX_OK;
}

XxxErrCode 
XxxMoveFuncOutOfAggrRule::Action(XxxValueExpression *InVExpr, void *RuleData,
				 XxxValueExpression *&OutVExpr) const
{
 // since we're at this point, we know that the InVExpr is an 
 // aggrexpression and its owner is a func-expression
 XxxAggrExpression *AggrExpr = (XxxAggrExpression *) InVExpr;
 XxxFuncExpression *FuncExpr = (XxxFuncExpression *) (AggrExpr->Owner);

 XxxValueExpression *FuncOwner = FuncExpr->Owner;

 // make the owner of the function the argument of the aggregate instead
 AggrExpr->Owner = FuncOwner;
 AggrExpr->FuncParse->IsTopLevel = XXX_FALSE;
 
 // make the aggregate the owner of the func-expr
 FuncExpr->Owner = AggrExpr;
 FuncExpr->OwnerType = FuncExpr->Owner->BaseType;
 FuncExpr->OwnerTypeCatInfo = FuncExpr->Owner->GenerateCatInfo();

 if (XXX_THISENV().DebugFlags.ADTRules) {
   cerr << "XxxMoveFuncOutOfAggrRule successful!" << endl;
   FuncExpr->Print(cerr);
   cerr << endl;
 }

 // return the aggregate expression
 OutVExpr = FuncExpr;
 return XXX_OK;
}
