#ifndef XXX_EXPR_H
#define XXX_EXPR_H
#include <iostream.h>
#ifdef XXX_WIN32
#include <strstrea.h>
#else
#include <strstream.h>
#endif
#include "types.h"
#include "xxx_thread.h" // included solely for error handling!!
#include "misc.h"

XXX_EXTERN_CLASS_DECL(XxxRecord);
XXX_EXTERN_CLASS_DECL(XxxRecordSchema);
XXX_EXTERN_CLASS_DECL(XxxQueryNode);
XXX_EXTERN_CLASS_DECL(XxxValueExprPlan);
XXX_EXTERN_CLASS_DECL(XxxValExprPlanList);
XXX_EXTERN_CLASS_DECL(XxxValueEnv);
XXX_EXTERN_CLASS_DECL(XxxBindingEnv);
XXX_EXTERN_CLASS_DECL(XxxUnknownValue);

XXX_EXTERN_CLASS_DECL(XxxArgEnv);
XXX_EXTERN_CLASS_DECL(XxxAggrOpHandle); // in aggrexpr.h

/**
 * Root class for the expression hierarchy.
 * Here's how the expression "library" works. The XxxExpression class
 * provides a virtual base class for all kinds of expressions. It provides
 * interfaces for type checking, etc. In the XXX system, every expression 
 * needs to be optimized. Therefore, there is a XxxValueExprPlan that is generated 
 * from an XxxExpression by invoking the Optimize() method. It is this 
 * XxxValueExprPlan that is executed in order to evaluate the expression.
 * @author Praveen
 */
class XxxValueExpression{
  friend class XxxValueExprPlan;
 protected:
  /// how many subexpressions?
  int arity;
  /// array of subexpressions
  XxxValueExpression** childExpr;
 public:
  /// Optional name for expression.
  char *ExprName;
  /// Is this a constant expression?
  virtual XxxBool IsConstant() const;
  /// What is the arity?
  virtual int Arity() const { return arity; }
  /// Get the i'th subexpressions
  inline XxxValueExpression* ChildExpr(int i) const;
  /// Print the expression
  virtual void Print(ostream&) const = 0;

  /// Tries to resolve unknown variables using a Schema
  virtual XxxErrCode ResolveVars (const XxxRecordSchema *Schema);
  /// Tries to resolve unknown variables using the child Source 
  virtual XxxErrCode ResolveVars(XxxQueryNode* Node, char Source);
  /// Tries to resolve host variables using the argument environment
  virtual XxxErrCode ResolveVars(const XxxArgEnv* Env);

  /// This should be called after all variables are resolved.
  virtual XxxErrCode TypeCheck();

  /**
   * Most expresions have hard coded execution strategies
   * and so are automatically optimized. Function expressions are not.
   * Every expr must be Optimize()ed during query optimization!!
   */
  virtual XxxErrCode Optimize(XxxValueExprPlan*& ExecPlan) = 0;

  /**
   * Gets input stream dependencies (that were resolved in ResolveVars)
   * Sets flags in the input argument Dependency. The size of
   * the array is 1+ the actual number of input streams.
   * The extra (0th) entry stands for the run-time ArgEnv
   */
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependendency,
				     int CorrelationHt) const;

  /// Specify a particular subexpression (used in query rewrite)
  XxxErrCode SetChild(int Index, XxxValueExpression* NewExp);
  /// Checks to see if the expression matches another expression
  virtual XxxBool Match(const XxxValueExpression* OtherExpr, 
			char *Env = NULL) const;
  /**
   *  Checks if the expression matches another rule input expression 
   *  If match is successful, return true, and assign BindTerm array
   *  inside XxxBindingEnv class.
   */
  virtual XxxBool Match(const XxxValueExpression* RuleExpr, 
			XxxBindingEnv *BindEnv) const;
  /// Provides a convenient way to copy arbitrary expressions
  virtual XxxValueExpression * Replicate() const = 0;
  /**
   * Replaces all references to the target expression with references
   * to the replacement expression.  Uses the match function to determine
   * if the target expression is referenced.  The replacement is replicated
   * for all replacements.
   */
  virtual XxxErrCode Replace (const XxxValueExpression *Target,
			      const XxxValueExpression *Replacement);
   /**
   * Replaces all references to the target expression with references
   * to the replacement expression.  Used in XxxStrRules (Binquan)
   */
  virtual XxxErrCode Replace (const XxxUnknownValue *Target,
			      const XxxValueExpression *Replacement);
  /**
   * Substitutes NewAttrIndex for OldAttrIndex in the SrcAttrIndex field 
   * of XxxUnknownValues subexpressions with SrcChild == ChildNumber
   * Appears should be initialized with XXX_FALSE and is set to true if
   * a column reference was found and updated.
   */
  virtual XxxErrCode UpdateUnknowns(int ChildNumber,
				   int OldAttrIndex,
				   int NewAttrIndex,
				   XxxBool &Appears);
 /**
   * Returns XXX_TRUE if this expression references an aggregate.
   */
  virtual XxxBool ReferencesAggregate() const;
  /**
   * Returns a list of the immediate subexpressions of the expression.
   * The list consists of the children or the owner and arguments of an 
   * expression. Used in the AggregateFunctionPushDownRule. (mayr@cs)
   */
  virtual XxxValueExprList *DirectSubexpressions();
  /**
   * Returns a list of replications of all value expressions referenced in
   * the expression.  It is the caller's responsibility to destroy the
   * list.
   */
  virtual XxxValueExprList *ValExprsReferenced() const;

  /** Returns replications of all subexpressions. Caller has to destroy
   * them. (Tobias) */ 
  virtual XxxValueExprList *SubExprsReferenced() const;

  /***** These are methods from the old XxxValueExpression ****/
  /**
   * A list of all possible value types of value expressions.  These
   * are used to identify subclass instances.
   */
  enum ValExprTypeEnum {
    INVALID,
    ARITH_EXPR,
    COLUMN_VAR,
    HOST_VAR,
    CONST_EXPR,
    FUNC_EXPR,
    CAST_EXPR,
    BOOL_EXPR,
    CASE_EXPR,
	JAVA_EXPR
  };
  /// A typedef for easy reference
  typedef enum ValExprTypeEnum ValExprType;

  XxxBasicType BaseType;    // return value of expression
  XxxADTMetaInfo *MetaInfo;   // return meta-info of expression

  XxxBasicType BasicType() const { return BaseType;}
  virtual ValExprType ValType() const = 0;

  //// needed for expressions that generate complex data types
  virtual XxxADTMetaInfo *GenerateCatInfo() const { return MetaInfo;}

  /// Common constructor
  XxxValueExpression(int NumChildren = 0, 
		     XxxValueExpression* Children[] = NULL);
  /// Copy constructor (used in most query rewrites)
  XxxValueExpression(const XxxValueExpression& src);
  /// Destructor
  virtual ~XxxValueExpression();
 };

/// short-term hack : TODO: remove this
#define XxxExpression XxxValueExpression 

/**
 * Enumeration of possible boolean expression types.
 * @author Praveen
 */
typedef enum {
  /// boolean expressions with AND, OR or NOT
  XXX_BOOL_BOOL_EXPR,    
  /// relational boolean expressions
  XXX_REL_BOOL_EXPR,
  /// values of boolean data type
  XXX_VAL_BOOL_EXPR
 } XxxBoolExprType;

/**
 * Class of boolean expressions.
 */
class XxxBooleanExpression : public XxxValueExpression {
 public:
  /**
   * Represents selectivity statistics for a predicate.
   */
  class SelectivityStatsInfo { 
  public:
    XxxRecordSchema* RecSchema;
    XxxADTClass::TypeStats* *AttrStats;
   };

  static XxxErrCode ConvertToCNF(XxxBooleanExpression* BExpr);
  static XxxErrCode 
    GenerateBooleanFactors(XxxBooleanExpression* Cond,
			   int& NumFactors,
			   XxxBooleanExpression**& Factors,
			   XxxBool ApplyXforms = XXX_TRUE);

  XxxBoolExprType BoolType;
  XxxBooleanExpression(int NumChildren = 0, XxxExpression* Children[] = NULL) :
    XxxExpression(NumChildren, Children) { BaseType = XXX_BOOLEAN;}

  /**
   * Copy constructor.  Makes a complete copy (recursing if necessary) of
   * the supplied expression.
   */
  XxxBooleanExpression(const XxxBooleanExpression& src) : XxxExpression(src)
  { BoolType = src.BoolType; }
  virtual double Selectivity(SelectivityStatsInfo* StatsArray, 
			     int ArraySize) const;
  virtual ValExprType ValType() const { return BOOL_EXPR;}
 };

/**
 * well known boolean connective operators
 * @author Praveen
 */
typedef enum {
  XXX_TRUE_OP,
  XXX_FALSE_OP,
  XXX_AND_OP,
  XXX_OR_OP,
  XXX_NOT_OP
 } XxxBoolOpType;
/**
 * A boolean term (AND, OR, NOT).
 * @author Praveen
 */
class XxxBoolBooleanExpression : public XxxBooleanExpression {
  friend class XxxBoolBooleanPlan;
 protected:
 public:
  XxxBoolOpType OpType;
  XxxBoolBooleanExpression(XxxBoolOpType Type, int NumChildren = 0,
			   XxxExpression* Children[] = NULL);

  /**
   * Copy constructor.  Makes a complete copy (recursing if necessary) of
   * the supplied expression.
   */
  XxxBoolBooleanExpression(const XxxBoolBooleanExpression& src) : 
    XxxBooleanExpression(src)
  { OpType = src.OpType; }

  /**
   * Provides a convenient way to copy arbitrary expressions.  It is unsafe
   * to use copy constructors for expressions, since a derived class's
   * copy constructor may not be called for a parent class.  Thus every
   * expression has a replicate method, which provides a common way to
   * duplicate an expression.
   */
  virtual XxxExpression * Replicate() const
    { return new XxxBoolBooleanExpression(*this); }
  virtual void Print(ostream&) const;
  virtual XxxErrCode Optimize(XxxValueExprPlan*& ExecPlan);
  virtual XxxErrCode TypeCheck();
  virtual XxxBool Match(const XxxExpression* Expr, 
			char *Env = NULL) const;
  virtual double Selectivity(SelectivityStatsInfo* StatsArray, 
			     int ArraySize) const;
 };

/**
 * well known boolean relational operators on basic types
 * @author Praveen
 */
typedef enum {
  XXX_GT_OP,
  XXX_LT_OP,
  XXX_EQ_OP,
  XXX_LEQ_OP,
  XXX_GEQ_OP,
  XXX_NEQ_OP,
  XXX_LIKE_OP,
  XXX_OTHER_OP // this is used by ANY, ALL, EXISTS, IN, NOT IN etc
 } XxxRelOpType;

/**
 * A relational predicate (GT, LT, etc)
 * @author Praveen
 */
class XxxRelBooleanExpression : public XxxBooleanExpression {
  friend class XxxRelBooleanPlan;
 protected:
  static XxxErrCode getStats(SelectivityStatsInfo *StatsArray, int ArraySize, 
		      XxxValueExpression *VExpr, double &High, double &Low,
		      double &Avg, int &Count);
  static double eqSelectivity(SelectivityStatsInfo* StatsArray,
			      int ArraySize,
			      XxxValueExpression* LExpr, 
			      XxxValueExpression* RExpr);
  static double gtSelectivity(SelectivityStatsInfo* StatsArray, 
			      int ArraySize,
			      XxxValueExpression* LExpr, 
			      XxxValueExpression* RExpr);
 public:
  XxxRelOpType OpType;
  XxxRelBooleanExpression(XxxRelOpType Type, int NumChildren = 0,
			   XxxExpression* Children[] = NULL);
  /**
   * Copy constructor.  Makes a complete copy (recursing if necessary) of
   * the supplied expression.
   */
  XxxRelBooleanExpression(const XxxRelBooleanExpression& src) : 
    XxxBooleanExpression(src)
    { OpType = src.OpType; }
  /// Provides a convenient way to copy arbitrary expressions
  virtual XxxExpression * Replicate() const
    { return new XxxRelBooleanExpression(*this); }
  virtual void Print(ostream&) const;
  virtual XxxErrCode Optimize(XxxValueExprPlan*& ExecPlan);
  virtual XxxErrCode TypeCheck();
  virtual XxxBool Match(const XxxExpression* Expr, 
			char *Env = NULL) const;
  virtual double Selectivity(SelectivityStatsInfo* StatsArray, 
			     int ArraySize) const;
 };

/**
 * List of value expressions.
 * @author Praveen
 */
class XxxValueExprList {
 private:
  void printThis(ostream& OStr) const
   { if(Expr) Expr->Print(OStr); }
 public:
  XxxValueExpression* Expr;
  XxxValueExprList* Next;  
  XxxValueExprList(XxxValueExpression *InitExpr, 
		   XxxValueExprList* InitNext = NULL) {
    Expr = InitExpr;
    Next = InitNext;
  }
  XxxValueExprList(const XxxValueExprList& src);
  virtual ~XxxValueExprList() {}

  XxxValueExprList *Replicate() const { return new XxxValueExprList(*this); }
  XxxValueExprList* Append(XxxValueExprList *ToAdd);
  // Destroy deletes the entire list, while the destructor only deletes
  // the particular item. No recursive deletes should be used!
  void Destroy();
  void DestroyListOnly();

  //// functions to allow index access to list
  XxxErrCode SetElement(const int ind, XxxValueExpression* Exp);
  XxxErrCode GetElement(const int ind, XxxValueExpression*& Exp) const;

  //// adds expr to the end of the list
  XxxValueExprList* AddExpr(XxxValueExpression *NextExpr); // -> 2/20/97

  //// adds expr to end, and returns the newly added elt in its second arg
  XxxValueExprList *AddExpr(XxxValueExpression *NextExpr, 
			    XxxValueExprList *&NewElt);  // -> 2/20/97

  /**
   * Replaces all references to the target expression with references
   * to the replacement expression.  Uses the match function to determine
   * if the target expression is referenced.
   */
  virtual XxxErrCode Replace (const XxxExpression *Target,
			      const XxxExpression *Replacement);
  /**
   * Replaces all references to the target expression with references
   * to the replacement expression. Used in XxxStrRule (Binquan)
   */
  virtual XxxErrCode Replace (const XxxUnknownValue *Target,
			      const XxxExpression *Replacement);
  /**
   * Substitutes NewAttrIndex for OldAttrIndex in the SrcAttrIndex field 
   * of XxxUnknownValues subexpressions with SrcChild == ChildNumber
   */
  virtual XxxErrCode UpdateUnknowns(int ChildNumber,
				    int OldAttrIndex,
				    int NewAttrIndex,
				    XxxBool &Appears);

  virtual void Print(ostream&) const;
  /**
   * Removes all expressions in the list except those of the specified type, 
   * possibly making the list empty.  Returns the new head of the list, or 
   * NULL if the list is now empty.
   */
  virtual XxxValueExprList *
    TypeFilter(const XxxValueExpression::ValExprType Type);
  virtual XxxBool Contains(XxxValueExpression *Expr);
  virtual XxxBool IsSubsetOf(XxxValueExprList *SuperSet);
  /// Is this expression based in any way on an aggregate expression?
  virtual XxxBool ReferencesAggregate() const;
  virtual XxxBool IsConstant() const;
  XxxErrCode ResolveVars (const XxxRecordSchema *Schema);
  XxxErrCode ResolveVars(XxxQueryNode* Node, char Source);
  XxxErrCode ResolveVars(const XxxArgEnv* Env);

  virtual XxxErrCode TypeCheck();
  virtual XxxErrCode TypeCheck(const XxxRecordSchema * RecSchema);

  // Gets input stream dependencies (that were resolved in ResolveVars)
  // Sets flags in the input argument Dependency. The
  // array must be initialized to be all false. The size of
  // the array is 1+ the actual number of input streams.
  // The extra (0th) entry stands for the run-time Env
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependendency,
				     int CorrelationHt) const;
  virtual XxxErrCode Optimize(XxxValExprPlanList*& PlanList);
  int Count() const;
  XxxValueExpression* operator[](int index) const;
};

// Now, time for the inline methods
inline XxxExpression* XxxExpression::ChildExpr(int i) const {
  if ((i < arity) && (i >= 0))
    return childExpr[i];
  else
    return NULL;
 }	

/**********
//// This represents a list of expressions.
XXX_EXTERN_CLASS_DECL(XxxExprList);
class XxxExprList {
private:
  void printThis(ostream& OStr) const
    { if(Expr) Expr->Print(OStr); }
public:
  XxxExpression* Expr;
  XxxExprList* Next;
  XxxExprList(XxxExpression *InitExpr, 
	      XxxExprList* InitNext = NULL) {
    Expr = InitExpr;
    Next = InitNext;
  }
  XxxExprList(const XxxExprList& src);

  XxxExprList* Append(XxxExprList *ToAdd); //--> sandman 02/13/97

  //// adds expr to the end of the list
  XxxExprList* AddExpr(XxxExpression *NextExpr);
  virtual void Print(ostream&) const;
  //// Destroy deletes the entire list and all the internal expressions
  void Destroy(); //--> sandman 02/13/97
  void DestroyListOnly();
  ~XxxExprList(){}
  int Count() const;
  XxxExpression* operator[](int index) const;
};
**********/

#endif

