#ifndef XXX_EXPRPLAN_H
#define XXX_EXPRPLAN_H
#include "expr.h"
#include "generic.h"

/**
 * An expression plan is used to evaluate an expression (i.e ValueExpression)
 * All plans must be subclasses of XxxGenericPlan, which 
 * defines a cost for the expression.
 * @see XxxGenericPlan
 * @author Praveen
 */
class XxxValueExprPlan : public XxxGenericPlan {
  friend class XxxExpression;
  XxxBasicType basicType;
  XxxValueExpression::ValExprType valType;
  XxxADTMetaInfo *catInfo;         
  //// a copy of the source expression (needed for index expressions)
  XxxValueExpression* sourceExpr;

 public:
  /**
   * A VarRemapping helps modify an expression plan
   * through join order changes.
   * Every variable (XxxUnknownValue) in an expression
   * gets resolved with a SrcChild and SrcAttrIndex during
   * type checking. These persist into the ExprPlan.
   * However, during join optimization, a query block
   * can be split into several different join operators
   * with a different join order. The ExprPlans need to be
   * suitably modified to reflect this. The VarReMapping
   * provides the necessary information to do so.
   */
  class VarRemapping { public:
    struct SrcMapping {
      int NewSrc;
      /// mappings for child attrs
      int* NewAttrMapping; 
     };
    int OldArity;
    int NewArity;
    //// array that specifies the new mapping 
    struct SrcMapping* SrcMappingArray; 
    VarRemapping(int Size);
    ~VarRemapping();
   };
 protected:
  PlanUsage planUsage;
  XxxBool isConstant;
  int arity;
  char printBuf[100];
  XxxBitMap dependency;
  /// this method will be called from RedirectDependencies
  XxxErrCode changeDependency(const VarRemapping& Remap);

  virtual PlanUsage& getPlanUsage() { return planUsage;}
 public:
  XxxValueExprPlan** SubPlans;

  /// this is the most commonly used constructor
  XxxValueExprPlan (XxxExpression * Exp, XxxValueExprPlan* Children[]);
  /// Copy constructor
  XxxValueExprPlan (const XxxValueExprPlan &Plan);
  /**
  // This constructor is used when building an exprplan
  // from existing exprplans. The printBuf, dependency BitMap
  // and sourceExpr need to be initialized outside the constructor.
  // This is typically called when making a conjunction of two exprplans.
  */
  XxxValueExprPlan(XxxValueExpression::ValExprType ExprType, int Arity, XxxBool Constant,
		   XxxValueExprPlan* Children[]);
  virtual ~XxxValueExprPlan();
  /// Replication
  virtual XxxGenericPlan * Replicate() const = 0;
  virtual void Print(ostream& OStr) const 
  { OStr << printBuf; }
  virtual XxxBool IsConstant()  const { return isConstant;}
  virtual int Arity() const { return arity; }
  char *PrintBuf() { return printBuf; }

  virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
			      const XxxRecordSchema* Schema[],
			      XxxADTValue *Result,
			      const XxxValueEnv* Env = NULL) 
    const = 0;
  //// evaluate, when there is a single input
  XxxErrCode Evaluate(const XxxRecord* Rec,
		      const XxxRecordSchema* Schema,
		      XxxADTValue *Result,
		      const XxxValueEnv* Env = NULL) 
    const { return Evaluate(&Rec, &Schema, Result, Env); }

  //// 4 methods to evaluate an aggregate expression
  //// Make a handle to hold ongoing aggregation
  virtual XxxErrCode MakeOpHandle(XxxAggrOpHandle*& Handle) const;
  //// Initialize handle
  virtual XxxErrCode Init(XxxAggrOpHandle *Handle) const;
  //// Accumulate into handle 
  virtual XxxErrCode Accumulate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxAggrOpHandle *Handle,
				const XxxValueEnv* Env = NULL) const;
  //// Terminates the aggregation and performs any postprocessing
  virtual XxxErrCode Terminate(XxxAggrOpHandle *Handle,
			       const XxxRecord* Rec[],
			       const XxxRecordSchema* Schema[],
			       XxxADTValue *FinalVal,
			       const XxxValueEnv* Env = NULL) const;

  //// member of XxxGenericPlan
  virtual XxxErrCode UsageOfPlan(PlanUsage& Usage) const {
    Usage = planUsage;
    return XXX_OK;
   }
  virtual const PlanUsage& GetPlanUsage() { return planUsage;}

  //// Gets input stream dependencies.
  //// Similar method for XxxExpression as well
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependendency) const;

  /**
  // This method replaces the dependencies of all variables
  // (i.e XxxUnknownVars) within this expression plan.
  // For example, an XxxUnknownValue with SrcChild = 1
  // will be replaced with SrcChild = Remap[1]
  // This is necessary if join reordering, etc is going to
  // happen. A similar method for changing the SrcAttrIndex
  // fields will be needed for projection pushdown.
  */
  virtual XxxErrCode RedirectDependencies(const VarRemapping& Remap,
					  int CorrelationHeight = 0);

  /**** former methods of XxxValueExprPlan ****/
  XxxBasicType BasicType() const { return basicType;}
  XxxValueExpression::ValExprType ValType() const { return valType;}
  virtual XxxADTMetaInfo *GenerateCatInfo() const { return catInfo; }
  const XxxValueExpression* SourceExpr() const { return sourceExpr; }
 };

/// short-term hack : TODO: remove this
#define XxxExprPlan XxxValueExprPlan

/**
 * List of value expression plans
 */
class XxxValExprPlanList {
 private:
  void printThis(ostream& OStr) const
   { Plan->Print(OStr); }
 public:
  XxxValueExprPlan* Plan;
  XxxValExprPlanList* Next;
  XxxValExprPlanList(XxxValueExprPlan *InitPlan, 
		     XxxValExprPlanList* InitNext = NULL) {
    Plan = InitPlan;
    Next = InitNext;
   }
  // adds expr to the end of the list
  XxxValExprPlanList* AddExprPlan(XxxValueExprPlan *NextExprPlan);
  virtual void Print(ostream&) const;
  XxxValExprPlanList * Replicate() const;
  // Destroy deletes the entire list, while the destructor only deletes
  // the particular item. No recursive deletes should be used in the code!!
  virtual void Destroy();  // --> Anil Sachdeva  2/20/97
  // DestroyListOnly only gets rid of the outer shell, leaving the exprplan
  // intact
  virtual void DestroyListOnly(); // --> Anil Sachdeva  2/20/97
  //Bjarke, sep. 11 2001: added virtual decl.
  virtual ~XxxValExprPlanList();
  int Count();
  XxxValueExprPlan* operator[](int index) const;

  // Used to evaluate en entire target list of value expressions
  XxxErrCode Evaluate(const XxxRecord * Rec[],
		      const XxxRecordSchema * Schema[],
		      XxxADTValue *Result[],
		      const XxxValueEnv * Env = NULL) const;

  virtual XxxErrCode RedirectDependencies(const XxxValueExprPlan::VarRemapping& Remap,
					  int CorrelationHeight = 0);
};

/**
 * Root plan class for all boolean expressions
 */
class XxxBooleanExpressionPlan : public XxxValueExprPlan { 
  friend class XxxBooleanExpression;
  friend class XxxBoolBooleanPlan;
 public:
  double Selectivity;
  /**
   * This method makes a conjunction of two expression plans.
   * If the OldExpr == NULL, the NewExpr is returned.
   * Otherwise, a fresh expression is allocated.
   * @param  OldExpr : one conjunct to be added
   * @param  NewExpr : the other conjunct to be added
   * @param  ResultExpr : the resulting conjunction.
   * @param  DeleteOld : was the old expr an AND (to be deleted)?
   */
  static XxxErrCode MakeConjunction(XxxBooleanExpressionPlan* OldExpr,
				    XxxBooleanExpressionPlan* NewExpr,
				    XxxBooleanExpressionPlan*& ResultExpr,
				    XxxBool& DeleteOld);
  /**
   * This method makes a conjunction of two expression plans.
   * It calls the previous method, and also deallocates the
   * original contents of the OldExpr.
   * @param  OldExpr : the input expression plan
   * @param  NewExpr : the other conjunct to be added
   */
  static XxxErrCode MakeConjunction(XxxBooleanExpressionPlan*& OldExpr,
				    XxxBooleanExpressionPlan* NewExpr);

  XxxBoolExprType BoolType;
  XxxBooleanExpressionPlan(XxxBooleanExpression* BExpr,
			   XxxValueExprPlan* Children[])
    : XxxValueExprPlan(BExpr, Children) 
  { BoolType = BExpr->BoolType; Selectivity = 0.1; }
  XxxBooleanExpressionPlan(XxxBoolExprType Type, int Arity,
			   XxxBool Constant, XxxValueExprPlan* Children[])
			   : XxxValueExprPlan(XxxValueExpression::BOOL_EXPR, Arity, Constant, Children) 
     {BoolType = Type;}
  XxxBooleanExpressionPlan(const XxxBooleanExpressionPlan &Plan) :
    XxxValueExprPlan(Plan), Selectivity(Plan.Selectivity),
    BoolType(Plan.BoolType)  {}
  virtual XxxGenericPlan * Replicate() const = 0;
  virtual ~XxxBooleanExpressionPlan() {}
 };

/**
 * Plan class for all bool boolean expressions
 */
class XxxBoolBooleanPlan : public XxxBooleanExpressionPlan { 
  friend class XxxBoolBooleanExpression;
 protected:
  ////
  class SortItem {
  public:
    double Key; // the cost or selectivity of a predicate
    int Index;  // of SubPlanOrder Array
    SortItem() { Key = 0.0; Index = 0; }
    static int Compare(const SortItem *A, const SortItem *B) {
      if(A->Key < B->Key) return -1;
      else if(A->Key > B->Key) return 1;
      else return 0;
    }
    static int CompareVoid(const void *A, const void *B) {
      return Compare((const SortItem*)A,
                     (const SortItem*)B);
    }
  };
 public:
  int *SubPlanOrder;  // indirection array to order the SubPlans
  XxxBoolOpType OpType;

  XxxBoolBooleanPlan(XxxBoolBooleanExpression* BExpr,
		     XxxBooleanExpressionPlan* Children[]);
  XxxBoolBooleanPlan(XxxBoolOpType Operation,
		     int NumChildren,
		     XxxBooleanExpressionPlan* ChildExpr[]);
  virtual ~XxxBoolBooleanPlan();
  XxxErrCode ReorderSubPlans();
  virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
			      const XxxRecordSchema* Schema[],
			      XxxADTValue *Result,
			      const XxxValueEnv* Env = NULL) const;
  XxxBoolBooleanPlan(const XxxBoolBooleanPlan &Plan);
  virtual XxxGenericPlan * Replicate() const {
    return (XxxGenericPlan *)(new XxxBoolBooleanPlan(*this));
  }
 };

/**
 * Plan class for all relational boolean expressions
 */
class XxxRelBooleanPlan : public XxxBooleanExpressionPlan { 
  friend class XxxRelBooleanExpression;
 public:
  XxxADTValue *ArgL, *ArgR;
  XxxBasicType TypeL, TypeR;
  const XxxADTMetaInfo *MetaInfoL, *MetaInfoR;

  // special case the following
  XxxBool Inverted;
  XxxRelOpType OperType;
  XxxRelBooleanPlan(XxxRelBooleanExpression* BExpr,
		    XxxValueExprPlan* Children[]);
  virtual ~XxxRelBooleanPlan();
  virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
			      const XxxRecordSchema* Schema[],
			      XxxADTValue *Result,
			      const XxxValueEnv* Env = NULL) const;

  // Inverts LT, LEQ, and NEQ into GEQ, GT and EQ respectively
  static XxxErrCode InvertOp(XxxRelOpType OrigOp,
			     XxxRelOpType& FinalOp,
			     XxxBool& Inverted);
  // generic code to evaluate a relational expression
  // (assumes that LT, LEQ, and NEQ have already been inverted)
  static XxxErrCode ExprEval(const XxxADTValue* ArgL,
			     XxxBasicType TypeL,
			     const XxxADTMetaInfo* MetaInfoL,
			     const XxxADTValue* ArgR,
			     XxxBasicType TypeR,
			     const XxxADTMetaInfo* MetaInfoR,
			     XxxRelOpType OperType,
			     XxxBool Inverted,
			     XxxBool& ExprResult);

  XxxRelBooleanPlan(const XxxRelBooleanPlan &Plan);
  virtual XxxGenericPlan * Replicate() const {
    return (XxxGenericPlan *)(new XxxRelBooleanPlan(*this));
  }
 };

#endif
