#include "optrules.h"
#include "funcexpr.h"
#include "adtfunc.h"

//
// Some local functions used to provide graphical query processing information
// 
// cycle thru eight possible colors
static int NodeColor(int LevelsBelow)
{
 int Val = LevelsBelow % 8;
 int RetVal = 0xffffff;
 if (Val & 1) {
   RetVal -= 128;
  }
 if (Val & 2) {
   RetVal -= 128*256;
  }
 if (Val & 4) {
   RetVal -= 128*256*256;
  }
 return RetVal;
}

// do a bottom-up print of each node
static void PrintNode(const XxxValueExpression* Expr,
		      ostream& OStr, int Level, int NodeId,
		      int& MaxChildNodeId,
		      int& NumLevelsBelow) 
{
 if (Expr->ValType() == XxxValueExpression::FUNC_EXPR) {
   XxxFuncExpression* FExpr = (XxxFuncExpression*)Expr;
   int ChildArrayLen = (FExpr->ArgList?FExpr->ArgList->Count():0);
   if (FExpr->Owner) ChildArrayLen++;
   int* ChildNodeIds = new int[ChildArrayLen];
   int CurrentChild = 0;
   MaxChildNodeId = NodeId;
   NumLevelsBelow = 0;
   int LevelsBelow;
   if (FExpr->Owner) {
     ChildNodeIds[CurrentChild++] = MaxChildNodeId+1;
     PrintNode(FExpr->Owner, OStr, Level+1, 
	       MaxChildNodeId+1, MaxChildNodeId, LevelsBelow);
     if (LevelsBelow > NumLevelsBelow)
       NumLevelsBelow = LevelsBelow;
    }
   if (FExpr->ArgList) {
     XxxValueExprList* ArgList = FExpr->ArgList;
     while (ArgList) {
       ChildNodeIds[CurrentChild++] = MaxChildNodeId+1;
       PrintNode(ArgList->Expr, OStr, Level+1, 
		 MaxChildNodeId+1, MaxChildNodeId, LevelsBelow);
       if (LevelsBelow > NumLevelsBelow)
	 NumLevelsBelow = LevelsBelow;
       ArgList = ArgList->Next;
      }
    }
   // finally, print this node
   OStr << "NODE " << FExpr->Name << endl;
   OStr << "ID " << NodeId << endl;
   OStr << "LEVEL " << Level << endl;
   if (ChildArrayLen) {
     OStr << "CHILDREN " ;
     for (int j = 0; j < ChildArrayLen; j++)
       OStr << ChildNodeIds[j] << " ";
     OStr << endl;
    }
   OStr << "COLOR " << NodeColor(NumLevelsBelow) << endl;
   if (FExpr->IsAggr())
     OStr << "INFO AggregateExpression : ";
   else
     OStr << "INFO FunctionExpression : ";
   FExpr->Print(OStr);
   OStr << endl << "ENDNODE" << endl;
   NumLevelsBelow++;
   if (ChildNodeIds) delete [] ChildNodeIds;
  }
 else {
   NumLevelsBelow = 0;
   MaxChildNodeId = NodeId;
   // print this node
   OStr << "NODE " ;
   Expr->Print(OStr);
   OStr << endl;
   OStr << "ID " << NodeId << endl;
   OStr << "LEVEL " << Level << endl;
   OStr << "COLOR " << NodeColor(NumLevelsBelow) << endl;
   OStr << "INFO ValueExpression : ";
   Expr->Print(OStr);
   OStr << endl << "ENDNODE" << endl;
   NumLevelsBelow++;
  }
}

static void XxxPrintFuncExprTree(ostream& OStr,
				 const XxxValueExpression* Expr, 
				 const char* Label)
{
 OStr << "TREE " << Label << endl;
 // Note: it is crucial that these not be the same variable,
 // since there are several recursive calls within PrintNode that
 // use these variables to pass different kinds of information
 // to each other.
 int Dummy1, Dummy2; 
 PrintNode(Expr, OStr, 0, 0, Dummy1, Dummy2);
 OStr << "ENDTREE" << endl;
}

//
// This function is used to specify whether a function is "top-level"
// or not (needed for ADT optimizations). This is used in funcexpr.C
//
void XxxAssignTopLevel(XxxValueExpression* ChildExpr,
		       int ParentTypeId)
{

  // PRD: CHG2
 XxxBasicType ChildTypeId;
 XxxBool* TopLevelRef;
 if (ChildExpr->ValType()==XxxValueExpression::FUNC_EXPR) {
   TopLevelRef = &(((XxxFuncExpression*)ChildExpr)->FuncParse->IsTopLevel);
   ChildTypeId = ((XxxFuncExpression*)ChildExpr)->FuncParse->TypeId;
  }
 else 
   return;

 // TODO:L2 In the next release, with a better version of inter-EADT
 // optimizations, this condition can be relaxed.
 if (ChildTypeId == ParentTypeId) 
   *TopLevelRef = XXX_FALSE;
 // otherwise, leave TopLevel the way it is (probably XXX_TRUE)
}

// returns true is the child is a "top-level" expression
static XxxBool IsTopLevel(XxxValueExpression* VExpr)
{
 if (VExpr->ValType() == XxxValueExpression::FUNC_EXPR)
   return ((XxxFuncExpression*)VExpr)->FuncParse->IsTopLevel;
 // otherwise, return true, because it is implicitly opaque
 return XXX_TRUE;
}

//
// Methods of XxxOptRuleEngine
//

// Constructor
XxxOptRuleEngine::XxxOptRuleEngine(int MaxFirings, int MaxRules, int MaxTravs) 
: MAXRULES(MaxRules), MAXFIRINGS(MaxFirings), MAXTRAVERSALS(MaxTravs)
{
 ruleArray = new XxxRule*[MAXRULES];
 for (int i = 0; i < MAXRULES; i++)
   ruleArray[i] = NULL;
 numRules = 0;
}

// Destructor
XxxOptRuleEngine::~XxxOptRuleEngine()
{
 for (int i = 0; i < numRules; i++)
   if (ruleArray[i]) delete ruleArray[i];
 delete [] ruleArray;
}

// Adding a rule to the engine
XxxErrCode XxxOptRuleEngine::AddRule(XxxRule* NewRule)
{
 if (numRules == MAXRULES) {
   XXX_NOTE_ERROR(XXX_OVFLO, "No more rules please!");
   return XXX_NOT_OK;
  }
 ruleArray[numRules++] = NewRule;
 return XXX_OK;
}

// Fire all the rules in the engine.
// To do this, perform traversals over the expression,
// trying to apply rules at each stage.
XxxErrCode XxxOptRuleEngine::FireAllRules(XxxValueExpression*& VExpr,
					  XxxBool ApplyTopDown,
					  XxxBool ApplyBottomUp) const
{
 if (NumRules() <= 0)
   return XXX_OK;

 ostream *QPStream = NULL;
 // try to get the output stream from the DiagnosticStream stack
 if (!XXX_THISENV().DiagnosticStream.IsEmpty())
   QPStream = (ostream *)(XXX_THISENV().DiagnosticStream.TopStack());
/*
 if (QPStream) {
   *(QPStream) << "SEQ ";
   VExpr->Print(*QPStream);
   *(QPStream) << endl; 
   XxxPrintFuncExprTree(*QPStream, VExpr, "Initial Expression");
  }
*/
 
 XxxBool ProgressDown = XXX_FALSE; // was there progress on DOWN traversal?
 XxxBool ProgressUp = XXX_FALSE;   // was there progress on UP traversal?
 int NumTraversals = 0;
 do {
   if (ApplyTopDown) {
     if (XXX_THISENV().DebugFlags.ADTRules) {
       // Debug information for the expression
       cerr << "Begin firing top-down rules!" << endl;
       cerr << " The Expression before applying the rule is " << endl;
       VExpr->Print(cerr);
       cerr << endl;
     }

     if (topDownTrav(VExpr, ProgressDown, QPStream) != XXX_OK) {
       XXX_NOTE_ERROR(XXX_BADEXEC, "Error in TopDown Traversal");
       return XXX_NOT_OK;
      }

     if (XXX_THISENV().DebugFlags.ADTRules) {
       cerr << " The Expression after applying the rule is " << endl;
       VExpr->Print(cerr);
       cerr << endl;
     }
   }
   if (ApplyBottomUp) {
     if (XXX_THISENV().DebugFlags.ADTRules) {
       // Debug information for the expression
       cerr << "Begin firing bottom-up rules!" << endl;
       cerr << " The Expression before applying the rule is " << endl;
       VExpr->Print(cerr);
       cerr << endl;
     }
     XXX_CALL_RET(bottomUpTrav(VExpr, ProgressUp, QPStream),
		  XXX_BADEXEC, "Error in BottomUp Traversal");
     if (XXX_THISENV().DebugFlags.ADTRules) {
       cerr << " The Expression after applying the rule is " << endl;
       VExpr->Print(cerr);
       cerr << endl;
     }
   } // if (ApplyBottomUp)
 } while ((ProgressDown || ProgressUp) && 
	  ++NumTraversals < MAXTRAVERSALS);

/*
 if (QPStream) {
   XxxPrintFuncExprTree(*QPStream, VExpr, "Final Expression");
   *(QPStream) << "ENDSEQ " << endl;
  }
*/
 return XXX_OK;
}

//
// This is a recursive traversal of the function expression tree.
// It is not usually a good idea to have recursive calls in a multi-threaded
// program, but in this case, it should be okay, since we do not expect
// much depth of recursion (it was also easier to code this way)
// 
XxxErrCode
XxxOptRuleEngine::topDownTrav(XxxValueExpression*& VExpr, 
			      XxxBool& DidSomething,
			      ostream *QPStream) const
{
 DidSomething = XXX_FALSE;

 int ActualFirings;
 XxxValueExpression* NewVExpr;
 if (fireRules(VExpr, NewVExpr, ActualFirings, 
	       TOP_DOWN, QPStream) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADOPT, "Error firing rule engine");
   return XXX_NOT_OK;
  }
 if (ActualFirings > 0) { // diagnostic
   if (XXX_THISENV().DebugFlags.ADTRules) {
     cerr << " Engine fired " << ActualFirings << " opt rules" << endl;
   }
   VExpr = NewVExpr;
   DidSomething = XXX_TRUE;
  }

 // At this point, VExpr may have changed as a result of rule firings!
 // Now apply rules on the owner if feasible.
 XxxValueExpression *Owner;
 if (VExpr->ValType() == XxxValueExpression::FUNC_EXPR) {
    Owner=((XxxFuncExpression*)VExpr)->Owner;
  } 
 else {
   return XXX_OK;
  }

 // ChildrenDidSomething represents whether or not something changed in 
 // the children after applying rules.
 XxxBool ChildrenDidSomething = XXX_FALSE;
 if (Owner && !IsTopLevel(Owner))
   if (topDownTrav(Owner, ChildrenDidSomething, QPStream) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADOPT, "Error in top down traversal");
     return XXX_NOT_OK;
    }
 
 // the Owner might have changed because of rewrites
 if (VExpr->ValType() == XxxValueExpression::FUNC_EXPR) {
   ((XxxFuncExpression *)VExpr)->Owner = Owner;
  } 

 // update DidSomething (if it was false before, and some rule was applied
 // in the owner or in the owner's children, DidSomething should be true).
 DidSomething = DidSomething || ChildrenDidSomething;

 // we're ready to reorganize arguments IF this is a funcexpression
 if (VExpr->ValType() != XxxValueExpression::FUNC_EXPR) {
   return XXX_OK;
  }

 XxxFuncExpression *FExpr=(XxxFuncExpression*)VExpr;
 // now apply rules on the arglist of this expression
 ChildrenDidSomething = XXX_FALSE;
 XxxValueExprList *Iter = FExpr->ArgList;
 while (Iter) {
   if (!IsTopLevel(Iter->Expr))
     if (topDownTrav(Iter->Expr, ChildrenDidSomething, QPStream) != XXX_OK) {
       XXX_NOTE_ERROR(XXX_BADOPT, "Error in top down traversal");
       return XXX_NOT_OK;
      }
   DidSomething= DidSomething || ChildrenDidSomething;
   Iter = Iter->Next;
  }
 return XXX_OK;
}

XxxErrCode
XxxOptRuleEngine::bottomUpTrav(XxxValueExpression*& VExpr, 
			       XxxBool& DidSomething,
			       ostream *QPStream) const
{
 DidSomething=XXX_FALSE;
 
 // the traversal only works on funcexpressions
 // NO. changed it to work with other expressions too
 XxxValueExpression *Owner = NULL;
 XxxValueExprList *Iter = NULL;
 if (VExpr->ValType() == XxxValueExpression::FUNC_EXPR) {
   XxxFuncExpression *FExpr = (XxxFuncExpression *) VExpr;
   Owner = FExpr->Owner;
   Iter = FExpr->ArgList;
  } 

 // ChildrenDidSomething represents whether or not something changed in 
 // the children after applying rules.
 XxxBool ChildrenDidSomething = XXX_FALSE;
 if (Owner) {
   if (!IsTopLevel(Owner)) {
     XXX_CALL_RET(bottomUpTrav(Owner, ChildrenDidSomething, QPStream),
		  XXX_BADOPT, "Error in bottom up traversal");
     // the Owner might have changed because of rewrites
     if (VExpr->ValType() == XxxValueExpression::FUNC_EXPR) {
       ((XxxFuncExpression *)VExpr)->Owner = Owner;
     } 
     // update DidSomething (if it was false before, and some rule was applied
     // in the owner or in the owner's children, DidSomething should be true).
     DidSomething= DidSomething || ChildrenDidSomething;
   } // if (!TopLevel)
 } // if Owner

 // now apply rules on the arglist of this expression.
 ChildrenDidSomething = XXX_FALSE;
 while (Iter) {
   if (!IsTopLevel(Iter->Expr)) {
     XXX_CALL_RET(bottomUpTrav(Iter->Expr, ChildrenDidSomething, QPStream),
		  XXX_BADOPT, "Error in top down traversal");
     DidSomething= DidSomething || ChildrenDidSomething;
   } // if !TopLevel
   Iter = Iter->Next;
 } // while (Iter)

 int ActualFirings;
 XxxValueExpression* NewVExpr;
 XXX_CALL_RET(fireRules(VExpr, NewVExpr, ActualFirings, 
	       BOTTOM_UP, QPStream),
	      XXX_BADOPT, "Error firing rule engine");
 if (ActualFirings > 0) { 
   if (XXX_THISENV().DebugFlags.ADTRules) { // diagnostic
     cerr << " Engine fired " << ActualFirings << " opt rules" << endl;
   }
   VExpr = NewVExpr;
   DidSomething = XXX_TRUE;
  }
 return XXX_OK;
}

XxxErrCode XxxOptRuleEngine::fireRules(XxxValueExpression* InVExpr,
				       XxxValueExpression*& OutVExpr,
				       int& ActualNumFirings,
				       TravDir TD,
				       ostream *QPStream) const
{
 ActualNumFirings = 0;

 // Rules are fired on the expression until there is a fixpoint,
 // or the limit on firings is exceeded.
 while (1) {
   XxxBool FiredRule = XXX_FALSE;
   
   for (int i = 0; i < numRules; i++) {
     XxxBool Success = XXX_FALSE;
     void* RuleData;
     // only execute rules that match the desired traversal direction
     if (TD != ruleArray[i]->TDir)
       continue;
     if (ruleArray[i]->PreCondition(InVExpr, Success, RuleData) != XXX_OK) {
       XXX_NOTE_ERROR(XXX_BADOPT, "Error checking rule precond");
       return XXX_NOT_OK;
      }
     if (Success) {
       ActualNumFirings++;
       // print out the graphical version
       if (QPStream) 
	 XxxPrintFuncExprTree(*QPStream, InVExpr, "Before Rule firing");
       if (ruleArray[i]->Action(InVExpr, RuleData, OutVExpr) != XXX_OK) {
	 XXX_NOTE_ERROR(XXX_BADOPT, "Error applying rule action");
	 return XXX_NOT_OK;
	}
       // print out the graphical version
       if (QPStream) 
	 XxxPrintFuncExprTree(*QPStream, OutVExpr, "After Rule firing");
       InVExpr = OutVExpr;
       FiredRule = XXX_TRUE;
       break; // for now, break out of loop: perhaps this is unnecessary
      } // if success
    } // for all rules

   if (!FiredRule) break;
   if (ActualNumFirings >= MAXFIRINGS) break;

  } // while (1)

 return XXX_OK;
}

