#ifndef XXX_OPTRULES_H
#define XXX_OPTRULES_H

#include "funcexpr.h"
#include "types.h"
#include "error.h"

XXX_EXTERN_CLASS_DECL(XxxRule);

/**
 * The Rule Engine allows simple rules to be registered.
 * These rules are used to perform E-ADT optimizations.
 * This rule engine has the following policy:
 *      --- repeatedly traverse the E-ADT expression top-down
 *            followed by bottom-up
 *      --- in each traversal, apply any rules that apply
 *
 * This is probably the most primitive rule engine possible.
 * More fancy engines could be added instead (future additions).
 */
class XxxOptRuleEngine {
 public:
  typedef enum { TOP_DOWN, BOTTOM_UP } TravDir;

 protected:
  const int MAXRULES;       // max number of rules in engine
  const int MAXFIRINGS;     // max number of firings at one expression
  const int MAXTRAVERSALS;  // maximum number of travrsals (down + up)

  int numRules;
  XxxRule** ruleArray;  

  XxxErrCode topDownTrav(XxxValueExpression*& VExp, XxxBool& DidSomething,
			 ostream *QPStream) const;

  XxxErrCode bottomUpTrav(XxxValueExpression*& VExp, XxxBool& DidSomething,
			  ostream *QPStream) const;

  //// Fire the rules against a single expression node
  XxxErrCode fireRules(XxxValueExpression* InVExpr,
		       XxxValueExpression*& OutVExpr,
		       int& ActualNumFirings,
		       TravDir TDir,
		       ostream *QPStream) const;
 public:
  XxxOptRuleEngine(int MaxFirings = 10,
		   int MaxRules = 20,
		   int MaxTraversals = 5);
  virtual ~XxxOptRuleEngine();

  //// Add a rule to the set of rules
  XxxErrCode AddRule(XxxRule* NewRule);

  /**
   * Fire all rules to transform VExpr, producing the resulting expression.
   * This could be subclassed to provide a better control strategy.
   */
  virtual XxxErrCode FireAllRules(XxxValueExpression*& VExpr,
				  XxxBool ApplyTopDown,
				  XxxBool ApplyBottomUp) const;

  int NumRules() const { return numRules;}
 };

/**
 * XxxRule is the root of all rule classes for E-ADT rules.
 * Each rule has a precondition and an action.
 */
class XxxRule {
 public:
  friend class XxxOptRuleEngine;
  typedef XxxErrCode (*RulePrecondFunc)(XxxValueExpression* VExpr,
					XxxBool& Success,
					void*& RuleData);
  typedef XxxErrCode (*RuleActionFunc)(XxxValueExpression* InVExpr,
				       void* RuleData,
				       XxxValueExpression*& OutVExpr);
 protected:
  RulePrecondFunc preCond;
  RuleActionFunc act;
 public:
  XxxOptRuleEngine::TravDir TDir;
  char RuleName[XXX_SUNDRYNAME_MAX];

  XxxRule(RulePrecondFunc PCF, 
	  RuleActionFunc RAF,
	  const char* Name = NULL, 
	  const XxxOptRuleEngine::TravDir TD = XxxOptRuleEngine::TOP_DOWN);
  virtual ~XxxRule() {} 

  virtual XxxErrCode PreCondition(XxxValueExpression* VExpr,
				  XxxBool& Success,
				  void*& RuleData) const {
    if (preCond) return preCond(VExpr, Success, RuleData);
    else {
      XXX_NOTE_ERROR(XXX_BADEXEC, "No rule condition specified");
      return XXX_NOT_OK;
     }
   }
  virtual XxxErrCode Action(XxxValueExpression* InVExpr,
			    void* RuleData, XxxValueExpression*& OutVExpr) const{
    if (act) return act(InVExpr, RuleData, OutVExpr);
    else {
      XXX_NOTE_ERROR(XXX_BADEXEC, "No rule action specified");
      return XXX_NOT_OK;
     }
   }
 };

/**
 * XxxSimpleCondRule provides a simple form of a rule condition.
 * It matches against all function expressions whose name is in 
 * "thisList", and whose owner's name is in "ownerList".
 * Also, the OwnerType and the type of this function must be specified.
 */
class XxxSimpleCondRule : public XxxRule {
 protected:
  char** ownerList;
  int ownerListLen;
  XxxBasicType ownerType;
  char** thisList;
  int thisListLen;
  XxxBasicType thisType;
 public:
  XxxSimpleCondRule(const char OwnerID[][XXX_FUNCNAME_MAX],
		    const int NumOwners,
		    const XxxBasicType OwnerType,
		    const char ThisID[][XXX_FUNCNAME_MAX], 
		    const int NumThis,
		    const XxxBasicType ThisType,
		    RuleActionFunc RAF,
		    const char* Name = NULL, 
		    const XxxOptRuleEngine::TravDir TD = XxxOptRuleEngine::TOP_DOWN);
  virtual ~XxxSimpleCondRule();

  virtual XxxErrCode PreCondition(XxxValueExpression* VExpr,
				  XxxBool& Success,void*& RuleData) const;
 };

/**
 * This rule defines an action that switches two functions.
 * For example, X.f().g()  --> X.g().f()
 * Here, "g" is the current function and "f" is its owner.
 */
class XxxFuncReorderRule: public XxxSimpleCondRule {
 public:
  XxxFuncReorderRule(const char OwnerID[][XXX_FUNCNAME_MAX],
		     const int NumOwners, 
		     const XxxBasicType OwnerType, 
		     const char ThisID[][XXX_FUNCNAME_MAX], 
		     const int NumThis, const XxxBasicType ThisType,
		     const char* Name = NULL,
		     const XxxOptRuleEngine::TravDir TD = XxxOptRuleEngine::TOP_DOWN)
    : XxxSimpleCondRule(OwnerID, NumOwners, OwnerType, ThisID, NumThis, 
			ThisType, 0, Name, TD) {}
  
  virtual XxxErrCode Action(XxxValueExpression *InVExpr, void* RuleData,
			    XxxValueExpression*& OutVExpr) const;
 };

/**
 * This rule defines an action that merges two functions into a single function.
 * For example, X.f().g()  --> X.fg().
 * This is implemented in a pretty complicated fashion, to allow for general
 * kinds of merging.
 * The strategy followed is :
 *      1) the Owner (i.e. "f") structure is always removed 
 *      2) either the Owner or This is identified as the "dominator"
 *          (i.e., the function which is to be retained). If a totally
 *          new merged function is being created, that can be indicated too.
 *      3) in order to construct the arguments of the merged function,
 *         individual "ArgRule"s are specified. some simple kinds of
 *         ArgRules have been provided. More can be written.
 */
class XxxMergeRule: public XxxSimpleCondRule {
 public:
  /// The "Dominator": which of the two expressions "dominates" the result
  typedef enum {THISEXP, OWNEREXP, NONE} Dominator;
  XXX_EXTERN_CLASS_DECL(ArgRule);

 protected:
  const ArgRule** argRuleArray;
  int argRuleArrayLen;
  char destName[XXX_FUNCNAME_MAX];
  Dominator dominator;
  
 public:
  
  // the Dominator type tells the Action rule which function (InFExpr or
  // InFExpr->Owner) will result from the merge and copies the function's
  // name and arguments into the destination expression before applying
  // arg rules.  This way, the user doesn't have to spend a lot of effort
  // defining the rule when the merged expression is similar to one of the
  // sources.  If the merge is going to result in something completely new,
  // we set Dominator to NONE.
  XxxMergeRule(const char OwnerID[][XXX_FUNCNAME_MAX],
	       const int NumOwners, const XxxBasicType OwnerType,
	       const char ThisID[][XXX_FUNCNAME_MAX],
	       const int NumThis, const XxxBasicType ThisType,
	       const ArgRule* AR[],
	       const int NumArgRules, Dominator Dom,
	       const char* Name = NULL, 
	       const XxxOptRuleEngine::TravDir TD = XxxOptRuleEngine::TOP_DOWN,
	       const char* DestString=NULL);
  virtual ~XxxMergeRule();
  virtual XxxErrCode Action(XxxValueExpression *InVExpr, void* RuleData,
			    XxxValueExpression*& OutVExpr) const;
 public:
  /// Class that specifies how each resulting attribute should be formed
  class ArgRule {
   public:
    int DestArgNum;
    ArgRule(int DestArg = -1) { DestArgNum = DestArg;}
    virtual ~ArgRule() {}
    //// this method creates a new expression for the specific position
    virtual XxxErrCode CreateExpr(const XxxValueExprList* OwnerArgList,
				  const XxxValueExprList* ThisArgList,
				  XxxValueExpression*& NewExpr) const {
				    XXX_NOTE_ERROR(XXX_INTERNAL, "Pure method");
				    return XXX_NOT_OK;
				   }
   };
  //// This argrule makes use of no more than one arg from each list
  class SimpleArgRule : public ArgRule {
   public:
    int OwnerArgNum;
    int ThisArgNum;
    // for operations like -, /, the Dominator tells Action which operand goes
    // first.  THISEXP means "ThisExp Op OwnerExp". 
    // For Copy, THISEXP means ThisExp should be copied to Dest.  OWNEREXP
    // works the other way.  case NONE is equivalent to THISEXP.
    Dominator FirstExp;
    SimpleArgRule(int OwnerArgNum, 
		  int ThisArgNum,
		  int DestArgNum, 
		  Dominator FirstExp) 
      : ArgRule(DestArgNum) {
	this->OwnerArgNum=OwnerArgNum;
	this->ThisArgNum=ThisArgNum;
	this->DestArgNum=DestArgNum;
	this->FirstExp=FirstExp;
       }
    virtual XxxErrCode CreateExpr(const XxxValueExprList* OwnerArgList,
				  const XxxValueExprList* ThisArgList,
				  XxxValueExpression*& NewExpr) const;
   protected:
    virtual XxxErrCode createExpr(const XxxValueExpression* OwnerArg,
				  const XxxValueExpression* ThisArg,
				  XxxValueExpression*& NewExpr) const = 0;
   }; // class SimpleArgRule

  /// ArgRule to copy one of the args
  class CopyArgRule : public SimpleArgRule {
   public:
    CopyArgRule(int OwnerArgNum, 
		int ThisArgNum,
		int DestArgNum, 
		Dominator FirstExp) 
      : SimpleArgRule(OwnerArgNum, ThisArgNum, DestArgNum, FirstExp) {}
   protected:
    virtual XxxErrCode createExpr(const XxxValueExpression* OwnerArg,
				  const XxxValueExpression* ThisArg,
				  XxxValueExpression*& NewExpr) const;
   }; // CopyArgRule

  /// ArgRule to copy one of the args
  class ArithArgRule : public SimpleArgRule {
   public:
    XxxArithOpType Op;
    ArithArgRule(int OwnerArgNum, 
		 int ThisArgNum,
		 int DestArgNum, 
		 Dominator FirstExp,
		 XxxArithOpType ArithOp)
      : SimpleArgRule(OwnerArgNum, ThisArgNum, DestArgNum, FirstExp) 
       { Op = ArithOp;}
   protected:
    virtual XxxErrCode createExpr(const XxxValueExpression* OwnerArg,
				  const XxxValueExpression* ThisArg,
				  XxxValueExpression*& NewExpr) const;
   }; // CopyArgRule
  
 };

/**
 * If a funcexpression has an aggregate as an owner, and the two can be 
 * interchanged to make the evaluation more efficient, this rule does so.
 */
class XxxMoveFuncIntoAggrRule : public XxxSimpleCondRule {
 public:
  XxxMoveFuncIntoAggrRule(XxxBasicType ADTType,
			  const char Aggrs[][XXX_FUNCNAME_MAX], 
			  const int NumAggrs, 
			  const char Funcs[][XXX_FUNCNAME_MAX], 
			  const int NumFuncs,
			  const char* Name = NULL,
			  const XxxOptRuleEngine::TravDir TD = XxxOptRuleEngine::BOTTOM_UP)
    : XxxSimpleCondRule(Aggrs, NumAggrs, ADTType, Funcs, NumFuncs, ADTType, 
	      0, Name, TD) {}
  
  virtual XxxErrCode PreCondition(XxxValueExpression *VExpr, XxxBool &Success,
				  void *&RuleData) const;
  
  virtual XxxErrCode Action(XxxValueExpression *InVExpr, void *RuleData,
			    XxxValueExpression *&OutVExpr) const;
 };

/**
 * If an aggregate has a funcexpression as its expr, and if evaluating the
 * aggregate first and then the funcexpression will be more efficient, 
 * this rule does so.
 */
class XxxMoveFuncOutOfAggrRule : public XxxSimpleCondRule {
 public:
  XxxMoveFuncOutOfAggrRule(XxxBasicType ADTType,
			   const char Aggrs[][XXX_FUNCNAME_MAX], 
			   const int NumAggrs,			   
			   const char Funcs[][XXX_FUNCNAME_MAX], 
			   const int NumFuncs,
			   const char* Name = NULL, 
			   const XxxOptRuleEngine::TravDir TD = XxxOptRuleEngine::TOP_DOWN)
    : XxxSimpleCondRule(Funcs, NumFuncs, ADTType, Aggrs, NumAggrs, ADTType, 
	      0, Name, TD) {}
  
  virtual XxxErrCode PreCondition(XxxValueExpression *VExpr, XxxBool &Success,
				  void *&RuleData) const;
  
  virtual XxxErrCode Action(XxxValueExpression *InVExpr, void *RuleData,
			    XxxValueExpression *&OutVExpr) const;
 };

#endif
