#ifndef XXX_SORT_H
#define XXX_SORT_H

#include "types.h"
#include "record.h"

XXX_EXTERN_CLASS_DECL(XxxValExprPlanList);

/**
 * Encapsulates sort specifications, such as which attributes, which order, 
 * etc.  This is PREDATOR's general sort criterion object, and is
 * used in all sort calls.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version May 23, 1997
 */
class XxxSortInfo { 
public:
  /// Specifications for a sort column
  struct OrderAttrInfo {
    /// The index of the attribute in the schema
    int AttrIndex;
    /// XXX_TRUE if sort is in ascending order
    XxxBool Asc;
  };
  /// A typedef for simplicity
  typedef struct OrderAttrInfo OrderedAttrInfo;
protected:
  /// The number of attributed on which to sort
  unsigned int numOrderedAttrs;
  /// An array of OrderedAttrInfo objects describing those attribute's specs
  OrderedAttrInfo* orderedAttrs;
public:
  /**
   * Constructor.
   * @param NumAttrs The number of OrderAttrInfo objects to allocate
   */
  XxxSortInfo(unsigned int NumAttrs) { Init(NumAttrs);}
  /**
   * Copy constructor.
   * @param SInfo the object to copy
   */
  XxxSortInfo(const XxxSortInfo& SInfo) { Init(SInfo);}
  /**
   * Default Constructor.
   */
  XxxSortInfo() { numOrderedAttrs = 0; orderedAttrs = NULL;}

  void Init(unsigned int NumAttrs);
  void Init(const XxxSortInfo& SInfo);
  /**
   * Destructor.
   */
  ~XxxSortInfo() { if (orderedAttrs) delete [] orderedAttrs;}
  /**
   * Determines if two XxxSortInfo objects are identical
   * @param SInfo The XxxSortInfo object to compare with
   * @return XXX_TRUE if the match is exact, XXX_FALSE otherwise
   */
  XxxBool Matches(const XxxSortInfo& SInfo) const;
  /**
   * @author: wee hyong
   * @version 20 June 2000
   * @desc    Added method interesting match to see whether the match correspond to
   *          an interesting order    
   * @param SInfo The XxxSortInfo object to compare with
   * @return XXX_TRUE if the match is interesting, XXX_FALSE otherwise
   */
  XxxBool interestingMatches(const XxxSortInfo& SInfo) const;
  /**
   * Returns the number of ordered attributes.
   */
  unsigned int NumOrderedAttrs() const { return numOrderedAttrs; }
  /**
   * A method of accessing the attribute ordering information.
   */
  OrderedAttrInfo & operator[](int index) const;
  /**
   * Returns a count of the number of attributes in the prefix of the
   * XxxSortInfo that are also found as XxxUnknownValues in the TargetList.
   * This is used to detemine sort orders for relational operations.
   */
  int CommonAttrs(const XxxValExprPlanList* TargetList,
		  int InputSrc = 0) const;
  /**
   * Fills an OutSInfo with the sort information about any attributes of
   * TargetList that are ordered by the spec in SInfo.
   * This is used by the relational operaors in conjunction with CommonAttrs.
   */
  XxxErrCode FillTarget(const XxxValExprPlanList* TargetList,
			XxxSortInfo& OutSInfo,
			int InputSrc = 0) const;
};

/**
 * This is the additional data required by the comparison function
 * below.
 */
struct XxxCompData {
  const XxxRecordSchema *Schema;
  const XxxSortInfo *SortInfo;
};

	
/// A typedef for simplicity
typedef struct XxxCompData XxxComparisonData;


/**
 * This is a typedef for a SHORE defined comparison function which
 * accepts two objects and their sizes along with some extra data
 * and determines a relative ordering of the two objects.
 */
typedef int (*XxxComparisonFunction)(size_type size1,
				     const void *item1,
				     size_type size2,
				     const void *item2,
				     const void *ExtraData);
#ifdef XXX_SHORE_V2
/**
*	This is the data which is associated to each key and permits
*   to extract the key in the extraction function below.
*/
class XxxKeyCompData {
public:
	XxxComparisonData keyInfo;
	int keyIndex;

	XxxKeyCompData(const XxxRecordSchema* Schema, const XxxSortInfo* SortInfo, int keyIndex)
	{
		this->keyInfo.Schema = Schema;
		this->keyInfo.SortInfo = SortInfo;
		this->keyIndex = keyIndex;
	}
	~XxxKeyCompData()
	{}
};

/**
* This is the callback key extraction function (CSKF) in Shore V2.0
*/
XxxShoreRetVal XxxKeyExtractionFunction(const XxxShoreRecId& rid,
	const XxxShoreSortObject &	in_obj,
	XxxShoreSortKeyData 	cookie,  // type info
	XxxShoreFactory&	factory,
	XxxShoreSortKey*	skey);
	
/**
* This is the callback key comparison function (CF) in Shore V2.0
*/
typedef int (*XxxKeyComparisonFunction)(size_type size1,
				     const void *item1,
				     size_type size2,
				     const void *item2);
#endif

/**
 * This is PREDATOR's relative ordering function for records.  It
 * accepts two records and their sizes, along with a pointer to 
 * a valid XxxComparisonData struct and determines the ordering of
 * the records by checking most-significant attributes first.
 * All ADT attributes are checked based upon their ordering if they
 * define one, or they are checked to see if they are equal and then
 * one is chosen arbitrarily to be "less than" the other.
 */

int XxxCompareFunc(size_type size1,
		   const void *Rec1,
		   size_type size2,
		   const void *Rec2, 
		   const void *ExtraData);

#ifdef XXX_SHORE_V2
int XxxCompareFuncADT(size_type size1,
		   const void *Rec1,
		   size_type size2,
		   const void *Rec2);
#endif

#endif
