#ifndef XXX_RECORD_H
#define XXX_RECORD_H

#include "error.h"
#include "xxx_thread.h"
#include "prd-config.h"
#include "types.h"
#include "templates.h"

XXX_EXTERN_CLASS_DECL(XxxUnknownValue);
XXX_EXTERN_CLASS_DECL(XxxQueryNode);
XXX_EXTERN_CLASS_DECL(XxxValueExprList);
XXX_EXTERN_CLASS_DECL(XxxValuesArray);

/**
  * Represents a single named and typed field.
  * @see XxxRecordSchema
  */
class XxxAttribute {
 public:
  /// name of the attribute
  char           Name[XXX_ATTR_NAME_MAX];
  /// type of attribute
  XxxBasicType   Type;       
  /// further type info if available
  XxxADTMetaInfo *TypeCatInfo;

  // Couple of attributes used in type checking only.
  // TODO: Will eventually be moved out of here.
  /// which input relation does this come from
  char AttrSource;
  /// index of src attribute in AttrSource
  int  AttrIndex;                         

  // constructors and destructor
  XxxAttribute() { TypeCatInfo = NULL; }
  XxxAttribute(char *XxxName, XxxBasicType XxxType, 
	       XxxADTMetaInfo* TypeInfo = NULL) 
    { this->Assign(XxxName, XxxType, TypeInfo); }
  ~XxxAttribute();

  /// Initialize this attribute
  inline void Assign(const char *XxxName, XxxBasicType XxxType,
		     const XxxADTMetaInfo* TypeInfo);
  /// Initialize this attribute
  inline void Assign(const XxxAttribute& Attr);

  /// Print the value of this attribute
  void Print(ostream& OStr) const;
  /**
    * Checks if this attribute corresponds directly or transitively to
    * the specified var value
    */
  XxxBool FindValueAttr(const XxxQueryNode*, 
			XxxUnknownValue*, 
			XxxBool) const;
};

/**
  * Defines the schema of a record.
  * @see XxxAttribute, XxxRecord
  */
class XxxRecordSchema {
 public:
  int                          RecArity;
  int                          RecordSize;
  int                          *RecordOffsets;
  int                          ExtraAreaOffset;
  XxxAttribute*                Attrs;   // list of typed attributes
  char*                        KeyVector; // bitvector indicating key attrs

  /// Computes offsets. Must call before the schema can be used
  void ComputeStats();
  int Arity() const { return RecArity;}
  /// Size of attrs alone (excl extra space for strings)
  int Size() const { return RecordSize;} 
  void Print(ostream&) const;

  XxxRecordSchema(const XxxRecordSchema& Schema);
  /// Construct the concatenation of many schemas
  XxxRecordSchema(int NumSchemas,
		  const XxxRecordSchema* Schemas[]);
  XxxRecordSchema(int Arity);
  ~XxxRecordSchema();

  XxxAttribute& operator [] (int index) const
    { return Attrs[index]; }

  unsigned int GetRecordSize(const XxxValuesArray& ValArray) const;
  const char *GetKeyVec() const { return KeyVector; }
  XxxAttribute *GetAttr(int index) const
    { return &(Attrs[index]); }
  XxxAttribute *GetAttr(char *AttrName) const ;
  XxxErrCode    GetAttrCol(char *AttrName, int &col) const;
  XxxErrCode    GetAttrType(char *AttrName, XxxBasicType &Type) const;

  int GetOffset(int AttrNum) const {
    return ((AttrNum < RecArity)?RecordOffsets[AttrNum]:-1);
   }
  XxxErrCode Matches(const XxxRecordSchema& Schema, XxxBool& Match) const;
 };

/**
 * The definition of a record in memory.
 */

//SMostafa


// The time stamps vector
class TheTimeStampVector{

 public:
  struct timeval minTS; // used to calculate the period (end period = minTS + w)
  struct timeval maxTS; // used to calculate the period (start period = maxTS + w)
  int VecSize;
  struct timeval *TheVec;

  TheTimeStampVector(TheTimeStampVector* v);
  TheTimeStampVector(struct timeval V);
  TheTimeStampVector(TheTimeStampVector* v, struct timeval ts);
  TheTimeStampVector(int size);
  ~TheTimeStampVector();
  void SetMaximum(TheTimeStampVector* v);
  XxxBool IsToLeftByWindowOf(TheTimeStampVector* v, int w);
  XxxBool TimeStampsGreater(struct timeval TS1, struct timeval TS2){
    double diff = ((TS1.tv_sec-TS2.tv_sec)+
		   (TS1.tv_usec-TS2.tv_usec)/1000000.0);
    return ((diff>0?XXX_TRUE:XXX_FALSE));
  };
  double TimeStampsDiff(struct timeval TS1, struct timeval TS2){
    return ((TS1.tv_sec-TS2.tv_sec)+
	    (TS1.tv_usec-TS2.tv_usec)/1000000.0);
  };
  
};
//EMostafa
class XxxRecord {
  unsigned int size;
  XxxBool borrowedData;
  void printBinary(ostream& OStr, const XxxRecordSchema* Schema) const;
  void printAscii(ostream& OStr, const XxxRecordSchema* Schema) const;

 public:
  char*  DataArea;
  //SMostafa
  TheTimeStampVector* TSVec;  // vector of time stamps in the record
  //EMostafa
  XxxRecord() {
    borrowedData = XXX_FALSE;
    size = 0;
    DataArea = NULL;
    //SMostafa
    TSVec = NULL;
    //EMostafa
   }
  XxxRecord(unsigned int Size) {
    borrowedData = XXX_FALSE;
    size = Size;
    DataArea = new char[size];
    //SMostafa
    TSVec = NULL;
    //EMostafa
   }
  XxxRecord(unsigned int Size, char *Data) {
    borrowedData = XXX_TRUE;
    size = Size;
    DataArea = Data;
    //SMostafa
    TSVec = NULL;
    //EMostafa
  }
  XxxRecord(const XxxRecord& Record) {
    borrowedData = XXX_FALSE;
    size  = Record.size;
    DataArea = new char[size];
    memcpy(DataArea, Record.DataArea, size);
    //SMostafa
    TSVec = NULL;
    //EMostafa
   }
  // constructor to take the place of XxxMakeNewRecord
  XxxRecord(const XxxValuesArray& ValArray,
	    const XxxRecordSchema* RecSchema);
  ~XxxRecord() {
    if (!borrowedData && DataArea) delete [] DataArea;
    //SMostafa
    if(TSVec!=NULL)
      delete TSVec;
    //EMostafa
   }

  XxxBool Equals(const XxxRecord *Rec, const XxxRecordSchema* Schema) const;
  void Print(ostream& OStr, const XxxRecordSchema* Schema,
    XxxBool Binary = false) const {
    if (Binary)
      printBinary(OStr, Schema);
    else
      printAscii(OStr, Schema);
  }
  // FieldPtr will be pointed to the appropriate range of bytes
  // within the record. If it is a variable sized record, ExtraOffset
  // will be set to the offset of the extra data within the record.
  // It is upto the caller to interpret these bytes in whatever fashion.
  XxxErrCode GetField(int i, const XxxRecordSchema* Schema,
		      const char*& FieldPtr) const;
  unsigned int Size() const { return size;}
  void Reclaim(const XxxRecordSchema* Schema);

  //// A list of schemas : should use a list template instead!
  class RecSchemaList { 
    struct recSchemaEntry {
      const XxxRecord* Record;
      const XxxRecordSchema* Schema;
      recSchemaEntry* Next;
      recSchemaEntry(const XxxRecord* Rec = NULL, 
		     const XxxRecordSchema* Sch = NULL,
		     recSchemaEntry* Nxt = NULL) :
      Record(Rec), Schema(Sch) { Next = Nxt; }
      ~recSchemaEntry() {
	if (Next) delete Next; // recursive delete
       }
     }; // struct recSchemaEntry
   public:
    int Count;
    recSchemaEntry* ListHead;
    
    RecSchemaList() { 
      Count = 0;
      ListHead = new recSchemaEntry();
     }
    ~RecSchemaList() { delete ListHead; }
    void Add(const XxxRecord* Rec, const XxxRecordSchema* Sch) {
      recSchemaEntry* Tmp = new recSchemaEntry(Rec, Sch, ListHead->Next);
      ListHead->Next = Tmp;
      Count++;
     }

    void StartScan(void *&Cursor) const
     { Cursor = (void *)(ListHead->Next); }
    XxxBool GetNext(void*& Cursor, const XxxRecord* &Rec, 
		    const XxxRecordSchema*& Sch) const
		     {
		      recSchemaEntry* Tmp = (recSchemaEntry*) Cursor;
		      if (!Tmp) return XXX_FALSE;
		      Rec = Tmp->Record;
		      Sch = Tmp->Schema;
		      Cursor = (void *)(Tmp->Next);
		      return XXX_TRUE;
		     }
    void EndScan(void *Cursor) const { }
   }; // RecSchemaList
  // Static method that concatenates a list of records.
  // The second argument saves an extra traversal of the list.
  static XxxRecord* Concat(const RecSchemaList& RSList, int TotalSize);
 }; // XxxRecord

/**
 * A set of records
 */
class XxxRecordSet : public XxxMultiSet<XxxRecord*> {
 protected:
  virtual XxxErrCode addItem(XxxRecord* const& AddRecord) = 0;
 public:
  XxxRecordSet() {}
  virtual ~XxxRecordSet() {}
  virtual XxxErrCode InitCursor(XxxSetCursor&) const = 0;
  virtual XxxErrCode NextItem(XxxSetCursor&, XxxRecord*& ) const = 0;
  virtual XxxErrCode CloseCursor(XxxSetCursor&) const = 0;
 };

/**
 * Used for textual records, to extract the record values 
 */
class XxxValuesArray {
 public:
  static const char XXX_RECORD_SEPARATOR; // = '\n'

  XxxADTValue**   Values;
  int NumValues;
  int *ValSize;
  XxxBasicType* Types;

  XxxValuesArray(const XxxValuesArray &ValArray);
  ~XxxValuesArray();
  XxxValuesArray(const XxxRecordSchema* RecSchema);
  XxxValuesArray(XxxValueExprList* ExprList);
  XxxValuesArray(XxxValExprPlanList* PlanList);
  // computes the size of the record stored in this XxxValuesArray
  unsigned int RecordSize(const XxxRecordSchema* RecSchema) const {
    return RecSchema->GetRecordSize(*this);
   }
  XxxErrCode GetTextRecordValues(istream& InStream, 
				 XxxRecordSchema* RecSchema, 
				 unsigned int& RecSize,
				 XxxBool Delimited,
				 char Delimiter);
 private:
  XxxErrCode handleMissingField(istream& InStream,
				const XxxAttribute& Attr,
				int i,
				XxxBool Delimited,
				char Delimiter);
 };

//
// Now for the inline methods
//
inline XxxAttribute::~XxxAttribute() {
  if (TypeCatInfo) 
    delete TypeCatInfo;
}
inline void XxxAttribute::
Assign(const char *XxxName, XxxBasicType XxxType, 
       const XxxADTMetaInfo* TypeInfo) {
  strncpy(Name, XxxName, XXX_ATTR_NAME_MAX);
  Name[XXX_ATTR_NAME_MAX-1] = 0;
  Type = XxxType;
  TypeCatInfo = (XxxADTMetaInfo *)TypeInfo;
  AttrSource = -1;
}

inline void XxxAttribute::
Assign(const XxxAttribute& Attr) {
  XxxADTMetaInfo *TypeInfo;
  if (Attr.TypeCatInfo) {
    Attr.TypeCatInfo->Replicate(TypeInfo);
  }
  else
    TypeInfo = NULL;
  this->Assign(Attr.Name, Attr.Type, TypeInfo) ;
}

inline XxxRecordSchema::
XxxRecordSchema(int InputArity) {
  RecArity = InputArity;
  Attrs = new XxxAttribute[RecArity];
  RecordSize = -1;
  KeyVector = new char[RecArity];
}

inline XxxRecordSchema::
~XxxRecordSchema() {
  if (Attrs) delete [] Attrs;
  if (RecordOffsets) delete [] RecordOffsets;
  if (KeyVector) delete [] KeyVector;
}

#endif  /* XXX_RECORD_H */
