#ifndef XXX_SHORESM_H
#define XXX_SHORESM_H

#include "error.h"
#include "shoresm-defs.h"
#include "record.h"
#include "expr.h"
#include "exprplan.h"
#include "prd-sort.h"

// Some error-handling macros
#define XXX_SHORE_FN(X) (X ? XXX_NOT_OK : XXX_OK)

#define XXX_SHORE_RETFN(X, R) \
  ((R=X)?(XXX_NOTE_ERROR(XXX_SHOREERR, w_error_t::error_string(R.err_num())), \
	    XXX_NOT_OK) : XXX_OK)

#define XXX_SM_CHECK \
 { if (!XxxShoreSM || !(XxxShoreSM->Valid())) { \
     XXX_NOTE_ERROR(XXX_SHOREERR, "Shore SM not up"); \
     return XXX_NOT_OK; \
   } \
 }
#define XXX_NULL_SM_CHECK \
 { if (!XxxShoreSM || !(XxxShoreSM->Valid())) { \
     XXX_NOTE_ERROR(XXX_SHOREERR, "Shore SM not up"); \
     return; \
   } \
 }

// Forward declarations of index classes which need direct access to 
// the Storage Manager
XXX_EXTERN_CLASS_DECL(ShoreBTree);
XXX_EXTERN_CLASS_DECL(ShoreRTree);

/**
 * A queue of SHORE file identifiers which are deleted lazily.
 */
class ShoreFileDeletionQueue {
public:
  class QueueEntry {
  public:
    /**
     * The time after which the entry may be deleted (in milliseconds 
     * since Jan 1, 1970)
     */
    long DeleteTime;
    /// The volume ID to be deleted
    XxxShoreVolumeId VolId;
    /// The file ID to be deleted
    XxxShoreObjectId FileId;
    /// The next queue entry (NULL if none)
    QueueEntry *NextEntry;
    /// The previous queue entry (NULL if none)
    QueueEntry *PrevEntry;
    /**
     * Default Constructor
     */
    QueueEntry (XxxShoreVolumeId VolumeId, 
		XxxShoreObjectId LocalId,
		long DeletionTime,
		QueueEntry *Next,
		QueueEntry *Prev);
  };

  /**
   * Constructor.
   */
  ShoreFileDeletionQueue();
  /**
   * Gets the time from now, adding the specified number of milliseconds.
   */
  static long GetTimeFromNow(long Delay);
  /**
   * Enqueues a file for deletion.
   */
  XxxErrCode EnqueueFile(XxxShoreVolumeId VolId, 
			 XxxShoreObjectId FileId,
			 long DeletionTime);
  /**
   * Deletes all files in the queue which have expired.
   */
  XxxErrCode GarbageCollect();
  /**
   * Deletes all files in the queue, regardless of when they should be deleted.
   */
  XxxErrCode ForceFlush();
protected:
  /// The queue of items
  QueueEntry *Entries;
  /// A latch to protect access to the queue
  XxxBaseMutex Latch;
};

/**
 * The wrapper for the SHORE storage manager.
 * @author Praveen Seshadri, praveen@cs.cornell.edu
 * @version June 25, 1997
 */
class XxxStorageManager {
  friend class XxxIDXClass;
 protected:
  //// Lower bound on number of buffers
  static const int MIN_NUM_BUFFERS;

  //// How long is each sort run? (set to 100)
  static const int XxxStorageManager::SORT_RUN_SIZE;

  //// A handle on the SHORE storage manager
  ss_m* storageManager;

  /// Options used in configuring the SHORE storage manager
  option_group_t *smOptions;

  //// number of buffer pages
  int numBuffers;

public:
  /**
   * A handle on a "default" volume.  As a convenience, the Storage
   * Manager maintains a volume accessible to all of the ADTs so they
   * may store objects persistently without managing an entire volume.
   */
  XxxShoreVolumeId defaultVolId; 
  /**
   * A handle on a "default" persistent file.  ADTs which choose not
   * to manage their own persistent storage may use this file to store
   * objects.  It is located on the default volume.
   */
  XxxShoreObjectId persistentFileId;
  /**
   * A handle on the "default" persistent file which is used only for data types information.  
   * Right now, it contains records for previous created instances of composite data types, such as alias,
   * struct, and enum. Predator server could use these records to recreate the ADT instance when restarted.
   * It is located on the default volume.
   */
  XxxShoreObjectId typeFileId;
  /**
   * A queue used to delete files after a period of time.
   */
  ShoreFileDeletionQueue DeletionQueue;

  typedef enum {
    XXX_SHORE_REGULAR, // all pages logged
    XXX_SHORE_LOAD,   // temporary (unlogged) till commit
    XXX_SHORE_APPEND, // only existing pages logged
    XXX_SHORE_TEMP  // removed when volume is dismounted (no logging)
  } XxxStoreProperty;

  /**
   * This class provides an iterator over a large object stored in the SM.
   * This iterator is to be used __ONLY__ to read an object, not to
   * update it.
   * TODO:L2: move all uses of pin_i into this object
   **/
  class ObjectIterator {
    //// is this the first time the iterator is being accessed?
    XxxBool firstAccess;
   public:
    //// did the constructor work okay?
    XxxBool Valid;
    //// handle to the object 
    pin_i* Pin;

    ObjectIterator(const XxxShoreVolumeId& VolumeId,
		   const XxxShoreObjectId& RecordId);
    ~ObjectIterator();
    XxxErrCode GetNextBytes(const char*& NextBytes, unsigned int& Length);
   };

  /**
   * Converts a SHORE Logical Id into a Physical Id.
   */
  XxxErrCode LogicalToPhysical(const XxxShoreLogicalOID &OID,
			       XxxShorePhysicalId &PhysId);

  /**
   * Creates a new SHORE file.
   * @param VolumeId The volume on which the file should be created
   * @param FileId (out) The file ID of the newly created file
   */
  XxxErrCode CreateFile(const XxxShoreVolumeId& VolumeId,
			XxxShoreObjectId& File,
			XxxStoreProperty Property = XXX_SHORE_REGULAR);
  /**
   * Creates a temporary (unlogged) SHORE file that will be removed
   * when the volume is unmounted.
   * @param VolumeId The file's volume
   * @param FileId (out) The file ID of the newly created file
   */
  XxxErrCode CreateTempFile(const XxxShoreVolumeId& VolumeId,
			    XxxShorePhysicalId& FileId);

  /**
   * Inserts a record into a temporary file.
   * @param VolumeId The volume's logical ID
   * @param FileId The file's ID
   * @param Record The record to insert
   */
  XxxErrCode InsertTempRecord(const XxxShorePhysicalId& FileId,
			      const XxxRecord* Record);
  /**
   * Sorts a SHORE temporary file, destroying it and creating a new 
   * SHORE file if the input file id is the same as the output file id.
   * @param InFileId The file id of the input file
   * @param VolumeId The output file's volume
   * @param OutFileId The file id of the output file
   * @param Func The comparison function which determines the 
   *             relative ordering of tuples
   * @param Data A pointer to the data the comparison function uses
   * @param Distinct Should duplicates be eliminated?
   * @param RunSize The number of buffer pages used for each sorted run
   */
  XxxErrCode SortPhysicalFile (const XxxShorePhysicalId &InFileId,
			       const XxxShoreVolumeId &VolumeId,
			       XxxShorePhysicalId &OutFileId,
			       XxxComparisonFunction Func,
			       const XxxComparisonData *Data,
			       XxxBool Distinct,
			       int RunSize);
  /**
   * Deletes a temporary (unlogged) SHORE file 
   * @param VolumeId The file's volume
   * @param FileId The file ID of the file
   */
  XxxErrCode DeleteTempFile(const XxxShorePhysicalId& FileId);

  /**
   * Changes the store property of the file.
   * @param Volumeid The volume for the file
   * @param FileId The file identifier
   * @param Property The desired file property
   */
  XxxErrCode SetStoreProperty(const XxxShoreVolumeId& VolumeId,
			      const XxxShoreObjectId& FileId,
			      XxxStoreProperty Property);
  /**
   * Gets the store property of the file.
   * @param Volumeid The volume for the file
   * @param FileId The file identifier
   * @param Property The desired file property
   */
  XxxErrCode GetStoreProperty(const XxxShoreVolumeId& VolumeId,
			      const XxxShoreObjectId& FileId,
			      XxxStoreProperty& Property);
protected:
  //Bjarke, aug. 16 01: replacements for storageManager->get_vid() method
  //don't know if they should be made virtual, especially GetTempVolume
  XxxShorePhysVolumeId GetPhysVolumeId(const XxxShoreVolumeId& VolumeID);
  XxxShorePhysVolumeId GetTempVolume(const XxxShoreVolumeId& VolumeID);

  /**
   * Creates sort key information for a sort predicate.
   * @param Comp The information regarding the sort
   * @param ShoreKey (out) The SHORE key information
   */
  XxxErrCode constructSortKeyInfo (XxxComparisonFunction Func,
				   const XxxComparisonData *Data,
				   XxxShoreKeyInfo &ShoreKey);

  /**
  * Creates sort key for Shore V2.0
  */
#ifdef XXX_SHORE_V2
  XxxErrCode constructSortKey(const XxxComparisonData *Data,
				   XxxShoreSortKeyList *&ShoreSortKeyList);
#endif
public:
  /**
   * Sorts a SHORE file, creating a new SHORE file.
   * @param InputFile The OID of the input file
   * @param DestroyInputFile Should the input file be destroyed upon
   *                         completion of the sort?
   * @param OutputFile The OID of the output file (created within)
   * @param NewFileProp The file property of the newly-created result file
   * @param Func The comparison function which determines the 
   *             relative ordering of tuples
   * @param Data A pointer to the data the comparison function uses
   * @param Distinct Should duplicates be eliminated?
   * @param RunSize The number of buffer pages used for each sorted run
   */
  XxxErrCode SortFile (const XxxShoreLogicalOID &InputFile,
		       XxxBool DestroyInputFile,
		       XxxShoreLogicalOID &OutputFile,
		       XxxStoreProperty NewFileProp,
		       XxxComparisonFunction Func,
		       const XxxComparisonData *Data,
		       XxxBool Distinct,
		       int RunSize)
  {
    return SortFile(InputFile.VolId, InputFile.LocalId,
		    DestroyInputFile, OutputFile.VolId, 
		    OutputFile.LocalId, NewFileProp, Func, 
		    Data, Distinct, RunSize);
  }
  /**
   * Sorts a SHORE file, creating a new SHORE file.
   * @param InVolumeId The volume id of the input file
   * @param InFileId The file id of the input file
   * @param DestroyInputFile Should the input file be destroyed upon
   *                         completion of the sort?
   * @param OutVolumeId The volume id of the output file
   * @param OutFileId The file id of the output file
   * @param NewFileProp The file property of the newly-created result file
   * @param Func The comparison function which determines the 
   *             relative ordering of tuples
   * @param Data A pointer to the data the comparison function uses
   * @param Distinct Should duplicates be eliminated?
   * @param RunSize The number of buffer pages used for each sorted run
   */
  XxxErrCode SortFile (const XxxShoreVolumeId &InVolumeId,
		       const XxxShoreObjectId &InFileId,
		       XxxBool DestroyInputFile,
		       XxxShoreVolumeId &OutVolumeId,
		       XxxShoreObjectId &OutFileId,
		       XxxStoreProperty NewFileProp,
		       XxxComparisonFunction Func,
		       const XxxComparisonData *Data,
		       XxxBool Distinct,
		       int RunSize);
  /**
   * Deletes an existing SHORE file.
   * @param VolumeId The file's volume
   * @param FileId The file's ID
   */
  XxxErrCode DeleteFile(const XxxShoreVolumeId& VolumeId,
			const XxxShoreObjectId& File);
  /**
   * Deletes an existing SHORE file.
   * @param LogicalOID The OID of the file
   */
  XxxErrCode DeleteFile(const XxxShoreLogicalOID& LogicalOID) {
    return DeleteFile(LogicalOID.VolId, LogicalOID.LocalId);
  }
  /**
   * Inserts a <Key, Datum> pair into the Root Index of a volume of the
   * SHORE storage manager.
   * @param VolumeId The volume in which an insertion should be made
   * @param Key The key for the pair
   * @param Datum The data for the pair
   * @param DatumSize The size of the data
   */
  XxxErrCode InsertInRootIndex(const XxxShoreVolumeId& VolumeId,
			       const XxxShoreKey& Key, 
			       void *Datum, 
			       int DatumSize);
  /**
   * Retrieves a <Key, Datum> pair from the Root Index of a volume of the
   * SHORE storage manager.
   * @param VolumeId The volume in which the pair resides
   * @param Key The key for the pair
   * @param DataBuf A pre-allocated buffer for the pair's data
   * @param DataBufSize The size of the data buffer
   * @param Found true if the pair was found
   */
  XxxErrCode GetFromRootIndex(const XxxShoreVolumeId& VolumeId,
			      const XxxShoreKey& Key,
			      void *DataBuf,
			      int DataBufSize,
			      bool& Found);
  /**
   * Deletes <Key, Datum> pairs from the Root Index of a volume of the
   * SHORE storage manager matching a given key
   * @param VolumeId The volume in which the pair(s) resides
   * @param Key The key for the pair
   * @param NumRemoved (out) The number of entries removed
   */
  XxxErrCode DeleteFromRootIndex(const XxxShoreVolumeId& VolumeId,
				 const XxxShoreKey& Key, 
				 int& NumRemoved);

  /**
   * Inserts a record into a file.
   * @param VolumeId The file's volume
   * @param FileId The file's ID
   * @param Record The record to insert
   * @param RecordId (out) The newly assigned OID of the record
   */
  XxxErrCode InsertRecord(const XxxShoreVolumeId& VolumeId,
			  const XxxShoreObjectId& FileId,
			  const XxxRecord* Record,
			  XxxShoreObjectId& RecordId);
  /**
   * Gets the total size (in bytes) of a SHORE object.
   * @param VolumeId The object's volume
   * @param FileId The object's ID
   * @param Size (out) The size of the object
   */
  XxxErrCode GetObjectSize(const XxxShoreVolumeId& VolumeId,
			   const XxxShoreObjectId& ObjectId,
			   int &Size);
  /**
   * Inserts a record into a file.
   * @param VolumeId The file's volume
   * @param FileId The file's ID
   * @param RecordData The data to insert
   * @param RecordSize The size of that data
   * @param RecordId (out) The newly assigned OID of the record
   */
  XxxErrCode InsertRecord(const XxxShoreVolumeId& VolumeId,
			  const XxxShoreObjectId& FileId,
			  const void* RecordData,
			  size_t RecordSize,
			  XxxShoreObjectId& RecordId);
  /**
   * Inserts a record into a file, allowing the user to specify an OID
   * for the record.
   * @param VolumeId The file's volume
   * @param FileId The file's ID
   * @param RecordData The data to insert
   * @param RecordSize The size of that data
   * @param RecordId (in) A specified OID to give the record
   */
  XxxErrCode InsertRecordWithId(const XxxShoreVolumeId& VolumeId,
				const XxxShoreObjectId& FileId,
				const void* RecordData,
				size_t RecordSize,
				const XxxShoreObjectId& RecordId);

  /**
   * Deletes a record from a file.
   * @param VolumeId The file's volume
   * @param RecordId The record's OID
   */
  XxxErrCode DeleteRecord(const XxxShoreVolumeId& VolumeId,
			  const XxxShoreObjectId& RecordId);
  /**
   * Gets an object from a file into a preallocated buffer.
   * @param VolumeId The file's volume
   * @param RecordId The record's OID
   * @param MemoryBuffer A preallocated buffer of BufferSize bytes
   * @param BufferSize The size of the supplied buffer
   * @param StartByte The starting byte number of the object to fetch
   * @param Length The extent of the object, starting with the start byte
   *               to fetch
   */
  XxxErrCode GetObjectPreAlloc(const XxxShoreVolumeId& VolumeId,
			       const XxxShoreObjectId& RecordId,
			       char* MemoryBuffer,
			       unsigned int BufferSize,
			       unsigned int StartByte,
			       unsigned int Length);
  /*
   * Gets a pinned object from a file into a preallocated buffer.
   * @param Pin A valid pin on the object
   * @param MemoryBuffer A preallocated buffer of BufferSize bytes
   * @param BufferSize The size of the supplied buffer
   * @param StartByte The starting byte number of the object to fetch
   * @param Length The extent of the object, starting with the start byte
   *               to fetch
   */
  XxxErrCode GetObjectPreAlloc(pin_i* Pin,
			       char* MemoryBuffer,
			       unsigned int BufferSize,
			       unsigned int StartByte,
			       unsigned int Length);
  /**
   * Gets an object from a file into an internally allocated buffer.
   * @param VolumeId The file's volume
   * @param RecordId The record's OID
   * @param MemoryBuffer (out) A buffer containing the data
   * @param BufferSize (out) The size of the supplied buffer
   * @param StartByte The starting byte number of the object to fetch
   * @param Length The extent of the object, starting with the start byte
   *               to fetch
   */
  XxxErrCode GetAllocateObject(const XxxShoreVolumeId& VolumeId,
			       const XxxShoreObjectId& RecordId,
			       char*& MemoryBuffer,
			       unsigned int& BufferSize,
			       unsigned int StartByte=0,
			       unsigned int Length=0);
  /**
   * Gets a pinned object from a file into an internally allocated buffer.
   * @param Pin A valid pin on the object
   * @param MemoryBuffer (out) A buffer containing the data
   * @param BufferSize (out) The size of the supplied buffer
   * @param StartByte The starting byte number of the object to fetch
   * @param Length The extent of the object, starting with the start byte
   *               to fetch
   */
  XxxErrCode GetAllocateObject(pin_i* Pin,
			       char*& MemoryBuffer,
			       unsigned int& BufferSize,
			       unsigned int StartByte=0,
			       unsigned int Length=0);
  /**
   * Gets (and allocates) a record from a file.
   * @param VolumeId The file's volume
   * @param RecordId The record's OID
   * @param Record (out) The newly allocated record.
   */
  XxxErrCode GetAllocateRecord(const XxxShoreVolumeId& VolumeId,
			       const XxxShoreObjectId& RecordId,
			       XxxRecord*& Record);
  /**
   * Gets (and allocates) a pinned record from a file.
   * @param Pin A valid pin on the record
   * @param Record (out) The newly allocated record.
   */
  XxxErrCode GetAllocateRecord(pin_i* Pin,
			       XxxRecord*& Record);
  /**
   * Updates a record by replacing a byte region with a new byte
   * region of the same size.
   * @param VolumeId The file's volume
   * @param RecordId The record's OID
   * @param startByte The starting byte in the record to replace
   * @param length The length of the replacement
   * @param newData The new data to place in the range
   */
  XxxErrCode UpdateRecord(const XxxShoreVolumeId& VolumeId,
			  const XxxShoreObjectId& RecordId,
			  const int StartByte, 
			  const int Length,
			  const char *NewData);
  /**
   * Replaces the contents of a record while retaining the record ID.
   * @param VolumeId The file's volume
   * @param RecordId The record's OID
   * @param NewRec The new record to store
   */
  XxxErrCode ReplaceRecord(const XxxShoreVolumeId& VolumeId,
			   const XxxShoreObjectId& RecordId,
			   const XxxRecord* NewRec);
  /**
   * Writes the contents of an object to an operating system file.
   * @param VolumeId The file's volume
   * @param RecordId The record's OID
   * @param FileName The name of the file to write to
   */
  XxxErrCode WriteToOSFile(const XxxShoreVolumeId& VolumeId,
			   const XxxShoreObjectId& RecordId,
			   const char* FileName);
  /**
   * Writes the contents of an object to an output stream.
   * @param VolumeId The file's volume
   * @param RecordId The record's OID
   * @param OStr The stream to write to
   */
  XxxErrCode WriteToStream(const XxxShoreVolumeId& VolumeId,
			   const XxxShoreObjectId& RecordId,
			   ostream &OStr);
  /**
   * Gets an object/record iterator object (used for large objects)
   * @param VolumeId The object's volume
   * @param RecordId The object's OID
   */
  XxxErrCode GetObjectIterator(const XxxShoreVolumeId& VolumeId,
			       const XxxShoreObjectId& RecordId,
			       ObjectIterator*& Iterator) {
    Iterator = new ObjectIterator(VolumeId, RecordId);
    if (!Iterator->Valid) {
      delete Iterator;
      return XXX_NOT_OK;
     }
    return XXX_OK;
   }
  /**
   * Constructor.
   * @param BufSize The size of the buffer pool, in KB
   * @param RC (out) The error code result
   */
  XxxStorageManager(int BufSize);
  /**
   * Destructor.
   */
  virtual ~XxxStorageManager();
  /**
   * Determines if the Storage Manager is up and running.
   */
  XxxBool Valid() const { return (storageManager?XXX_TRUE:XXX_FALSE);}
  /**
   * Number of buffers allocated
   */
  int NumBuffers() const { return numBuffers;}
  /**
   * Returns the location of the default volume.
   */
  XxxShoreVolumeId DefaultVol() const { return defaultVolId; }
  /**
   * Returns the location of the default persistent file.  ADTs which 
   * choose not
   * to manage their own persistent storage may use this file to store
   * objects.  It is located on the default volume.
   */
  XxxShoreObjectId PersistentFileId() const { return persistentFileId; }
  /**
   * Begins a transaction.
   */
  virtual XxxErrCode BeginTrans();
  /**
   * Commits a transaction.
   */
  virtual XxxErrCode CommitTrans();
  /**
   * Aborts a transaction.
   */
  virtual XxxErrCode AbortTrans();
  /**
   * Creates a device at the specified location.
   * @param FilePath The location of the device
   * @param DeviceSize The size of the device in KB
   * @param VolId (out) The newly assigned volume ID
   */
  virtual XxxErrCode CreateDevice(const char* FilePath, const int DeviceSize,
				  XxxShoreVolumeId& VolId);
  /**
   * Creates the default device at the specified location.
   * @param FilePath The location of the device
   * @param DeviceSize The size of the device in KB
   */
  virtual XxxErrCode CreateDefDevice(const char* FilePath, 
				     const int DeviceSize);
  /**
   * Mounts a device at the specified location.
   * @param FilePath The location of the device
   * @param VolId (out) The newly assigned volume ID
   */
  virtual XxxErrCode MountDevice(const char* FilePath, 
				 XxxShoreVolumeId& VolId);
  /**
   * Mounts the default device.
   * @param FilePath The location of the device
   */
  virtual XxxErrCode MountDefDevice(const char* FilePath);
  /**
   * Dismounts a device.
   * @param FilePath The location of the device
   */
  virtual XxxErrCode UnMountDevice(const char* FilePath);
  /**
   * Dismounts a device, if mounted; then destroys it.
   * @param FilePath The location of the device
   */
  virtual XxxErrCode DestroyDevice(const char* FilePath);
  /**
   * Determines if a device exists.
   * @param FilePath The location of the device
   * @param Found (out) XXX_TRUE if the device exists
   */
  virtual XxxErrCode DeviceExists(const char* FilePath, XxxBool& Found) const;
  /**
   * Prints disk usage statistics to the output stream.
   */
  XxxErrCode DiskUsage(ostream& OStr);
  /**
   * Prints SHORE statistics to the output stream.
   */
  XxxErrCode PrintStats(ostream& OStr);
  /**
   * Resets SHORE statistics.
   */
  XxxErrCode ResetStats();
  /**
   * Determines if the supplied pin is valid.
   */
  static XxxBool PinValid(pin_i* Pin, lock_mode_t LM)
   { 
     if (Pin->pinned()) {
       // TODO:L2 check if this is necessary for all store properties
       if (!Pin->up_to_date()) {
	 // need to repin the record
	 Pin->repin(LM);
	 /** Sometimes, this still isnt up to date. I dont know why.
	   if (!Pin->up_to_date())
	   cerr << "Bother!!: Pin->up_to_date still bad" << endl;
	   **/
       }
       return XXX_TRUE; 
     }
     else {
       return XXX_FALSE;
     }
   }

  static void PinPrint(const pin_i* Pin, char* Buf) {
#if (SHORE_INTERIM < 2)    
    const char* format = "Pin() failed to pin!: %d %d %d %d %ld %ld";
#else
    const char* format = "Pin() failed to pin!: %d %d %d %d %d %d";
#endif
   sprintf(Buf, format,
	   Pin->pinned(), Pin->up_to_date(), 
	   Pin->pinned_all(), Pin->is_small(), 
	   Pin->start_byte(), Pin->length());
  }
  /**
   * Returns the OID of a pinned object.
   */
  static XxxShoreObjectId PinObjectId(const pin_i* Pin)
   { return Pin->serial_no(); }
  /**
   * Translates a transaction to its ID.
   * @param Xact A pointer to the transaction
   * @param XactId The transaction ID of that transaction
   */
  XxxErrCode XactToId (const XxxShoreXact *Xact,
		       XxxShoreXactId &XactId) const;
  /**
   * Translates a transaction ID to a pointer to the transaction.
   * @param Xact A pointer to the transaction
   * @param XactId The transaction ID of that transaction
   */
  XxxErrCode IdToXact (const XxxShoreXactId &XactId,
		       XxxShoreXact *&Xact) const;
};

/**
 * A global handle on the Storage Manager
 */
extern XxxStorageManager* XxxShoreSM;
/**
 * Begins a transaction for XXX_THISTHREAD.
 */
extern XxxErrCode XxxBeginTrans();
/**
 * Commits a transaction for XXX_THISTHREAD.
 */
extern XxxErrCode XxxCommitTrans();
/**
 * Aborts a transaction for XXX_THISTHREAD.
 */
extern XxxErrCode XxxAbortTrans();

#endif
