#ifndef XXX_STRRULES_H
#define XXX_STRRULES_H

#include "optrules.h"

/* Binquan, Jan 11, 1998
The string rule engine is used to construct rules that utilizes the semantic of the data types.
Example can be found in image/image.C, such as:

RuleEngine->AddRule(new XxxStrRule("X.pic.blur(v1).scale(v2)", "X.pic.scale(v2).blur(v1)", 
ImageTypeId, "reorder1", XxxOptRuleEngine::TOP_DOWN, "1.0 > v2"));

where the parameter lists are
1. "X.pic.blur(v1).scale(v2)", is the pattern to be matched against any query expression,
    we call it precondtion string,
2. "X.pic.scale(v2).blur(v1)", is the pattern which the query expression will be transformed into.
3. ImageTypeId, the type of the owner of the rule
4. "reorder1", the name of the rule
5. XxxOptRuleEngine::TOP_DOWN, the direction to apply the rule
6. "1.0 > v2", the condition to apply the rule, we call it parameter condition string.

The idea is  to apply the rule whenever the query expression matchs the precondition
string, and the parameters inside the expression statisfy the parameter condition string.

Notice, if one variable occurs more than once in the input expression,
any expression matching this rule should provide the same value for all
instances of that variable in the input string.

For example, 
RuleEngine->AddRule(new XxxStrRule("X.pic.scale(v1).scale(v1)", "X.pic.scale(v1*v1)", 
ImageTypeId, "merge1", XxxOptRuleEngine::TOP_DOWN));

the rule will match "X.pic.scale(0.1).scale(0.1)", but not for "X.pic.scale(0.1).scale(0.2)"

*/

/**
 * The Binding environment variables for the string rule, it contains all the 
 * unknown variables in the precondition string, and the corresponding values in the 
 * matched query expression.
 */
class XxxBindingEnv {
  friend class XxxStrRule;
protected:
  /**
   * The structure that contains a pair of the unknown variable and its value
   */
  class BindTerm {
  public:
    XxxUnknownValue*  UnknownVar;
    XxxExpression* BindValue;
  };

  BindTerm *BindArray;
  int Len;
  
public:
  XxxBindingEnv(XxxValueExprList* ExprList);
  XxxBindingEnv(XxxBindingEnv& Old);
  virtual ~XxxBindingEnv();
  
  // Set the value of the unknown variable
  virtual XxxErrCode SetElement(XxxUnknownValue* Var, XxxExpression* Value);
  // Get the value of the unknown variable
  virtual XxxExpression* GetElement(XxxUnknownValue* Var);
  /**
   * This function will be used in a rule which matches a collection of function.
   * such as both "blur", "negate" method of the image ADT can be reordered with method
   * "scale", so if in the rule precondition "X.pic.foo(v1).scale(v2)", the const string
   * "foo" should match both "blur" and "negate" when doing typecheck
   */
  virtual XxxBool Match(const char* foo, const char* funcName);
};

/**
 * XxxStrRule provides a form of a rule condition.
 * It takes three strings, InStr, OutStr, CondStr
 * Any function expressions containing the Instr, and satisfies the
 * parameter condition string, would match this
 * rule, and change the matching part into outStr.
 */
class XxxStrRule :  public XxxRule {
  friend class XxxBindingEnv;
 protected:
  char *InStr;
  char *OutStr;
  char *CondStr;
  XxxBasicType Type;

  class RuleDataStruct {
  public:
    XxxBindingEnv * BindEnv;
    XxxValueExpression *OutExpr;
    RuleDataStruct(XxxBindingEnv *bindEnv, XxxValueExpression *outExpr) {
      BindEnv = bindEnv;
      OutExpr = outExpr;
    }
  };

public:
  XxxStrRule(const char* InString, const char* OutString,
             XxxBasicType type, const char* Name=NULL, 
             const XxxOptRuleEngine::TravDir TD = XxxOptRuleEngine::TOP_DOWN,
			 const char* CondString=NULL);
   
  virtual ~XxxStrRule(); 

  virtual XxxBasicType BasicType() { return Type; }
  virtual XxxErrCode PreCondition(XxxValueExpression* VExpr,
				  XxxBool& Success,
				  void*& RuleData) const ;
   

  virtual XxxErrCode Action(XxxValueExpression* InVExpr, void* RuleData, 
                            XxxValueExpression*& OutVExpr) const;
 };


#endif



