#include "suspcomp.h"
#include "adtfunc.h"
//Bjarke, feb 1 2002: added time.h for win32
#ifdef WIN32
#include <time.h>
#endif

const int XxxSuspCompLib::DEFAULT_CAPACITY = 20;
const XxxSuspCompLib::XxxSuspCompId XxxSuspCompLib::INVALID_ID = 0;

XxxSuspendedComp::
XxxSuspendedComp(const XxxFuncPlanInfo *ExecPlan,
                 const XxxValueEnv *Environment, 
                 const XxxADTClass::FuncMethodInfo *MethInfo) :
  Plan(NULL), Env(NULL), MethodInfo(NULL) 
{ 
  if (ExecPlan) {
    Plan = (XxxFuncPlanInfo*)(ExecPlan->Replicate());
    // Important: unmark this plan as suspended
    Plan->UnsuspendComputation();
  }
  if (Environment)
    Env = Environment->Replicate();
  if (MethInfo)
    MethodInfo = (XxxADTClass::FuncMethodInfo *)MethInfo->Replicate();
}

XxxSuspendedComp::~XxxSuspendedComp()
{
  if (Plan != NULL) delete Plan;
  if (Env != NULL) delete Env;
  if (MethodInfo != NULL) delete MethodInfo;
}

XxxErrCode XxxSuspendedComp::Evaluate (XxxADTValue *ReturnValue)
{ 
#ifdef XXX_WIN32
 clock_t Time0 = clock();
#else
  struct timeval Time[2];
  struct timezone TZ[2];
  if (XXX_THISENV().DebugFlags.SuspComp)
    gettimeofday(&Time[0], &TZ[0]);
#endif

  // Important: Unmark this value as suspended
  ReturnValue->Unset(XxxADTValue::SUSPENDED_COMPUTATION);
  // Actually compute the value
  XxxErrCode RC = Plan->Evaluate(Env, MethodInfo, ReturnValue);

  if (XXX_THISENV().DebugFlags.Connection) {
	  double ElapsedTime;
#ifdef XXX_WIN32
	clock_t Time1 = clock();
	ElapsedTime = (Time1 - Time0) /1000000.0;
#else
    gettimeofday(&Time[1], &TZ[1]);
    ElapsedTime = (Time[1].tv_sec - Time[0].tv_sec) +
      (Time[1].tv_usec - Time[0].tv_usec)/1000000.0;
#endif
    cerr << "Suspended computation took " 
	 << ElapsedTime << " seconds." << endl;
  }

  return RC;
}

XxxSuspCompLib::Entry::Entry(XxxSuspCompId IDNumber, 
			     XxxSuspendedComp *SuspCompPtr) :
  SuspComp(SuspCompPtr), ID(IDNumber) { }

XxxSuspCompLib::Entry::~Entry()
{
  if ((SuspComp != NULL) && (ID != INVALID_ID))
    delete SuspComp;
}

XxxSuspCompLib::XxxSuspCompLib(int InitialCapacity) : 
  entries(new Entry[InitialCapacity]), currentCapacity(InitialCapacity),
  size(0), nextID(INVALID_ID + 1)
{
}

XxxSuspCompLib::~XxxSuspCompLib()
{
  Flush();
}

XxxErrCode XxxSuspCompLib::GetSuspComp(const XxxSuspCompId IdNumber, 
				       XxxSuspendedComp *& SuspComp)
{
  if (IdNumber == INVALID_ID) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Invalid suspended computation ID provided");
    return XXX_NOT_OK;
  }
  XxxUseMutex XxxAutoMutex(Latch);
  for (int i = 0; i < currentCapacity; i++)
    if (entries[i].ID == IdNumber) {
      SuspComp = entries[i].SuspComp;
      return XXX_OK;
    }
  XXX_NOTE_ERROR(XXX_BADEXEC, "Unable to find suspended computation");
  return XXX_NOT_OK;
}

int XxxSuspCompLib::GetSize() const
{
  return size;
}

int XxxSuspCompLib::GetCapacity() const
{
  return currentCapacity;
}

XxxErrCode XxxSuspCompLib::UpdateCapacity(int NewCapacity)
{
  XxxUseMutex XxxAutoMutex(Latch);
  if (NewCapacity < size) {
    XXX_NOTE_ERROR(XXX_BADEXEC, 
		   "New capacity too low; cannot delete indescriminately");
    return XXX_NOT_OK;
  }
  if (NewCapacity == currentCapacity)
    return XXX_OK;
  else if (NewCapacity > currentCapacity) {
    Entry *newEntries = new Entry[NewCapacity];
    memcpy(newEntries, entries, sizeof(Entry) * currentCapacity);
    delete [] entries;
    entries = newEntries;
    currentCapacity = NewCapacity;
    return XXX_OK;
  } else {
    Entry *newEntries = new Entry[NewCapacity];
    int newPos = 0;
    for (int i = 0; i < currentCapacity; i++)
      if (entries[i].ID != INVALID_ID) {
	newEntries[newPos].ID = entries[i].ID;	
	newEntries[newPos].SuspComp = entries[i].SuspComp;
	newPos++;
      }
    currentCapacity = NewCapacity;
    delete [] entries;
    entries = newEntries;
    return XXX_OK;
  }
}

XxxErrCode XxxSuspCompLib::RemoveSuspComp(const XxxSuspCompId IdNumber)
{
  if (IdNumber == INVALID_ID) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Invalid suspended computation ID provided");
    return XXX_NOT_OK;
  }
  XxxUseMutex XxxAutoMutex(Latch);
  for (int i = 0; i < currentCapacity; i++)
    if (entries[i].ID == IdNumber) {
      delete entries[i].SuspComp;
      entries[i].SuspComp = NULL;
      entries[i].ID = INVALID_ID;
      return XXX_OK;
    }
  XXX_NOTE_ERROR(XXX_BADEXEC, "Unable to find suspended computation");
  return XXX_NOT_OK;
}

XxxErrCode XxxSuspCompLib::InsertSuspComp(XxxSuspendedComp *SuspComp, 
					  XxxSuspCompId &IdNumber)
{
  XxxUseMutex XxxAutoMutex(Latch);
  if (size >= currentCapacity) {
    if (currentCapacity == 0) {
      entries = new Entry[currentCapacity = DEFAULT_CAPACITY];
    } else {
      // Double the number of entries
      Entry *newEntries = new Entry[currentCapacity * 2];
      memcpy(newEntries, entries, sizeof(Entry) * currentCapacity);
      delete [] entries;
      entries = newEntries;
      currentCapacity *= 2;
    }
  }
  if (nextID == INVALID_ID) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "ID wrap-around occured!");
    return XXX_NOT_OK;
  }
  // Search for an empty slot
  for (int i = 0; i < currentCapacity; i++)
    if (entries[i].ID == INVALID_ID) {
      entries[i].ID = IdNumber = nextID++;
      entries[i].SuspComp = SuspComp;
      return XXX_OK;
    }

  // Should never get here
  ASSERT(1 == 0);
  return XXX_NOT_OK;
}

XxxErrCode XxxSuspCompLib::Flush()
{
  XxxUseMutex XxxAutoMutex(Latch);
  if (entries != NULL) {
    delete [] entries;
    entries = NULL;
    size = currentCapacity = 0;
  }
  return XXX_OK;
}
