#ifndef XXX_SUSPCOMP_H
#define XXX_SUSPCOMP_H

#include "prd-config.h"
#include "types.h"
#include "values.h"
#include "xxx_thread.h"


/**
 * Objects of this class represent suspended computations--specifically,
 * method calls whose execution has been postponed.
 */
class XxxSuspendedComp {
protected:
  XxxFuncPlanInfo *Plan;
  XxxValueEnv *Env;
  XxxFuncMethodInfo *MethodInfo;
public:
  XxxSuspendedComp(const XxxFuncPlanInfo *ExecPlan,
                   const XxxValueEnv *Environment, 
                   const XxxFuncMethodInfo *MethInfo);
  virtual ~XxxSuspendedComp();
  virtual XxxErrCode Evaluate(XxxADTValue *ReturnValue);
};

/**
 * This class represents a persistent collection of suspended computations.
 */
class XxxSuspCompLib {
public:
  /**
   * The identifiers used for stored suspended computation objects.
   */
  typedef unsigned short XxxSuspCompId;
  /**
   * An invalid suspended computation ID.
   */
  static const XxxSuspCompId INVALID_ID;
  /**
   * The number of suspended computations this library can hold.
   */
  static const int DEFAULT_CAPACITY;
  /**
   * Constructor.
   */
  XxxSuspCompLib(int InitialCapacity = DEFAULT_CAPACITY);
  /**
   * Destructor.  Calls Flush() to remove all entries.
   */
  ~XxxSuspCompLib();
  /**
   * Updates the capacity of this collection of suspended computations.
   * If the new capacity is less than the current number of computatations
   * stored, an error is generated.
   */
  XxxErrCode UpdateCapacity(int NewCapacity);
  /**
   * Gets the number of stored suspended computations.
   */
  int GetSize() const;
  /**
   * Gets the current capacity of this object.
   */
  int GetCapacity() const;
  /**
   * Gets a suspended computation object, given its ID number.
   */
  XxxErrCode GetSuspComp(const XxxSuspCompId, 
			 XxxSuspendedComp *& SuspComp);
  /**
   * Inserts a new suspended computation object, assigning it an ID number.
   */
  XxxErrCode InsertSuspComp(XxxSuspendedComp *SuspComp, 
			    XxxSuspCompId &IdNumber);
  /**
   * Deletes the suspended computation associated with the given ID.
   */
  XxxErrCode RemoveSuspComp(const XxxSuspCompId);
  /**
   * Removes all suspended computation objects from the library, 
   * deleting them.
   */
  XxxErrCode Flush();
protected:
  /**
   * A linked list entry.
   */
  class Entry {
  public:
    XxxSuspendedComp *SuspComp;
    XxxSuspCompId ID;

    Entry(XxxSuspCompId IDNumber = INVALID_ID, 
	  XxxSuspendedComp *SuspCompPtr = NULL);
    ~Entry();
  };
  /**
   * An array of entries.
   */
  Entry *entries;
  /**
   * The current capacity of this object.
   */
  int currentCapacity;
  /**
   * The current number of stored computations.
   */
  int size;
  /**
   * The next ID to assign.
   */
  XxxSuspCompId nextID;
  /**
   * A mutex to guard access to this object.
   */
  XxxBaseMutex Latch;
};

#endif
