#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "values.h"
#include "record.h"
#include "absquery.h"
#include "exprplan.h"
#include "strrules.h"

XxxArithExpression::XxxArithExpression(XxxArithOpType OperatorType,
				       int NumChildren,
				       XxxExpression* Children[]) :
  XxxValueExpression(NumChildren, Children)
{
  OpType = OperatorType;
  if (arity == 1) {
    if (OperatorType != XXX_UMINUS_OP) {
      XXX_NOTE_ERROR(XXX_BADEXPR, " Bad unary arithmetic operator ");
    }
  }
  else if (arity == 2) {
    if ((OperatorType != XXX_PLUS_OP) &&
	(OperatorType != XXX_MINUS_OP) &&
	(OperatorType != XXX_TIMES_OP) &&
	(OperatorType != XXX_DIV_OP))
      {
	XXX_NOTE_ERROR(XXX_BADEXPR, " Bad binary arithmetic operator ");
      }
  }
  else {
    XXX_NOTE_ERROR(XXX_BADEXPR, " Bad arithmetic operator arity");
  }
}

XxxArithExpression::XxxArithExpression (const XxxArithExpression &src) :
  XxxValueExpression(src)
{
  OpType = src.OpType;
}

void XxxArithExpression::Print(ostream& OStr) const 
{
  if (Arity() == 1) {
    if (OpType == XXX_UMINUS_OP) {
      OStr << " - (";
      ChildExpr(0)->Print(OStr);
	  OStr << ")";
    }
    else {
      XXX_NOTE_ERROR(XXX_BADVALUE, "Bad Arith Op!");
    }     
  }
  else if (Arity() == 2) {
	OStr << "(";
    ChildExpr(0)->Print(OStr);
    switch (OpType) {
    case XXX_PLUS_OP:
      OStr << " + ";
      break;
    case XXX_MINUS_OP:
      OStr << " - ";
      break;
    case XXX_TIMES_OP:
      OStr << " * ";
      break;
    case XXX_DIV_OP:
      OStr << " / ";
      break;
    default:
      XXX_NOTE_ERROR(XXX_BADVALUE, "Bad Arith Op!");
    };
    ChildExpr(1)->Print(OStr); 
	OStr << ")";
  }
}

XxxErrCode XxxArithExpression::TypeCheck()
{
  XxxErrCode rc;
	
  if ((rc = XxxExpression::TypeCheck()) != XXX_OK) {
    return rc;
  }
	
  if ((Arity()>0) && ChildExpr(0)->ValType() == BOOL_EXPR) {
    XXX_NOTE_ERROR(XXX_BADTYPE, "Bad left operand of Arith Op!");
    return XXX_NOT_OK;
  }
  if ((Arity()>1) && ChildExpr(1)->ValType() == BOOL_EXPR) {
    XXX_NOTE_ERROR(XXX_BADTYPE, "Bad right operand of Arith Op!");
    return XXX_NOT_OK;
  }
  XxxBasicType LType, RType;
  LType = (Arity()>0)?
    ((XxxValueExpression*)ChildExpr(0))->BasicType(): XXX_UNKNOWN;
  RType = (Arity()>1)?
    ((XxxValueExpression*)ChildExpr(1))->BasicType(): XXX_UNKNOWN;
	
  switch (OpType) {
  case XXX_PLUS_OP:
  case XXX_MINUS_OP:
  case XXX_TIMES_OP:
  case XXX_DIV_OP:
    if ((RType == XXX_DOUBLE) || (RType == XXX_INT))
      if ((LType == XXX_DOUBLE) || (LType == XXX_INT)) {
	BaseType = ((RType == XXX_INT)&&(LType==XXX_INT))?XXX_INT:XXX_DOUBLE;
	break;
      }
  case XXX_UMINUS_OP:
    if ((LType == XXX_DOUBLE) || (LType == XXX_INT)) {
      BaseType = (LType==XXX_INT) ? XXX_INT:XXX_DOUBLE;
      break;
    }
  default:
    BaseType = XXX_UNKNOWN;
    XXX_NOTE_ERROR(XXX_BADTYPE, "Bad operands in arith expression");
    return XXX_NOT_OK;
  }
	
  return XXX_OK;
}

XxxErrCode XxxArithExpression::Optimize(XxxValueExprPlan*& ExecPlan)
{
  XxxValueExprPlan** TmpPlans = new XxxValueExprPlan*[Arity()];
  for (int i = 0; i < Arity(); i++) {
    XXX_CALL_RET(ChildExpr(i)->Optimize(TmpPlans[i]),
		 XXX_BADOPT, "Error optimizing subexp");
  }
  switch (OpType) {
  case XXX_PLUS_OP:
    ExecPlan = new PlusPlan(this, TmpPlans); break;
  case XXX_MINUS_OP:
    ExecPlan = new MinusPlan(this, TmpPlans); break;
  case XXX_TIMES_OP:
    ExecPlan = new TimesPlan(this, TmpPlans); break;
  case XXX_DIV_OP:
    ExecPlan = new DivPlan(this, TmpPlans); break;
  case XXX_UMINUS_OP:
    ExecPlan = new UMinusPlan(this, TmpPlans); break;
  };

  delete [] TmpPlans;
  return XXX_OK;
}

//AUTHOR: Fabian Camargo DATE:2/10/97
//PURPOSE: Checks to see if expression matches index expr
//NOTE: If expression is 'a + b', it will NOT match with 'b + a'
XxxBool
XxxArithExpression::Match(const XxxExpression *OtherExpr,
			  char* Env) const
{
  if (!(XxxValueExpression::Match(OtherExpr, Env)))
    return XXX_FALSE;
  XxxArithExpression *CastExpr = (XxxArithExpression *)OtherExpr;
  if (OpType != CastExpr->OpType)
    return XXX_FALSE;
  return XXX_TRUE;
}

XxxBool
XxxArithExpression::Match(const XxxExpression *RuleExpr,
			  XxxBindingEnv* BindEnv) const
{
  // For arithmetic expression, we don't need to check the following thing,
  //  because "0.1 * 0.2" will not match "v2", since the arity of them are different
  if(((XxxValueExpression *)RuleExpr)->ValType() == COLUMN_VAR) {
    if((BindEnv->SetElement((XxxUnknownValue *)RuleExpr, (XxxExpression *)this)) != XXX_OK){
 	  cerr << "Value conflict, rule doesn't match" << endl;
      return XXX_FALSE;
    }
    return XXX_TRUE;
  }
  if (!(XxxValueExpression::Match(RuleExpr, BindEnv))) {
    return XXX_FALSE;
  }
	
  XxxArithExpression *CastExpr = (XxxArithExpression *)RuleExpr;
  if (OpType != CastExpr->OpType)
    return XXX_FALSE;
  return XXX_TRUE;
}

XxxArithExpression::ArithPlan::ArithPlan(XxxArithExpression* AExpr,
					 XxxValueExprPlan* Children[])
  : XxxValueExprPlan(AExpr, Children)
{
  OpType = AExpr->OpType;
  init();
}

void XxxArithExpression::ArithPlan::init()
{
  TypeL = SubPlans[0]->BasicType();
  switch (TypeL) {
  case XXX_INT:
    ArgL = &(IntArg[0]);
    break;
  case XXX_DOUBLE:
    ArgL = &(DoubleArg[0]);
    break;
  default:
    XXX_NOTE_ERROR(XXX_BADTYPE, "");
    return;
  }
  if (Arity() == 2) {
    TypeR = SubPlans[1]->BasicType();
    switch (TypeR) {
    case XXX_INT:
      ArgR = &(IntArg[1]);
      break;
    case XXX_DOUBLE:
      ArgR = &(DoubleArg[1]);
      break;
    default:
      XXX_NOTE_ERROR(XXX_BADTYPE, "");
      return;
    }
  } // arity == 2
}

XxxErrCode XxxArithExpression::
ArithPlan::Evaluate(const XxxRecord* InputRecords[],
		    const XxxRecordSchema* InputSchemas[],
		    XxxADTValue *Result,
		    const XxxValueEnv* Env) const 
{	
  // this should never get called : delete it TODO
  return XXX_NOT_OK;
}

XxxErrCode XxxArithExpression::
UMinusPlan::Evaluate(const XxxRecord* InputRecords[],
		    const XxxRecordSchema* InputSchemas[],
		    XxxADTValue *Result,
		    const XxxValueEnv* Env) const 
{	
  XXX_CALL_RET(PreEvaluate(InputRecords, InputSchemas, Result, Env),
	       XXX_BADEXPR, "");

  XxxCoreADTClass::DblValue *DoubleAns = 
    (XxxCoreADTClass::DblValue *)Result;
  XxxCoreADTClass::IntValue *IntAns = 
    (XxxCoreADTClass::IntValue *)Result;
	
  if (BasicType() == XXX_INT) {
    if (TypeL == XXX_INT) 
      IntAns->Value = -IntArg[0].Value;
    else if (TypeL == XXX_DOUBLE) 
      IntAns->Value = (int)(-DoubleArg[0].Value);
  }
  else if (BasicType() == XXX_DOUBLE) {
    if (TypeL == XXX_INT)
      DoubleAns->Value = (double)(-IntArg[0].Value);
    else if (TypeL == XXX_DOUBLE)
      DoubleAns->Value = -DoubleArg[0].Value;
  }
  return XXX_OK;
}

XxxErrCode XxxArithExpression::
PlusPlan::Evaluate(const XxxRecord* InputRecords[],
		   const XxxRecordSchema* InputSchemas[],
		   XxxADTValue *Result,
		   const XxxValueEnv* Env) const 
{
  XXX_CALL_RET(PreEvaluate(InputRecords, InputSchemas, Result, Env),
	       XXX_BADEXPR, "");

  XxxCoreADTClass::DblValue *DoubleAns = 
    (XxxCoreADTClass::DblValue *)Result;
  XxxCoreADTClass::IntValue *IntAns = 
    (XxxCoreADTClass::IntValue *)Result;

  ASSERT ((OpType == XXX_PLUS_OP));

  if (BasicType() == XXX_INT) {
    if (TypeL == XXX_INT) { 
      if (TypeR == XXX_INT) 
	IntAns->Value = IntArg[0].Value+IntArg[1].Value;
      else 
	IntAns->Value = (int)(IntArg[0].Value+DoubleArg[1].Value); }
    else if (TypeL == XXX_DOUBLE) {
      if (TypeR == XXX_INT) 
	IntAns->Value = (int)(DoubleArg[0].Value+IntArg[1].Value);
      else 
	IntAns->Value = (int)(DoubleArg[0].Value+DoubleArg[1].Value); }
  }
  else if (BasicType() == XXX_DOUBLE) {
    if (TypeL == XXX_INT) { 
      if (TypeR == XXX_INT)
	DoubleAns->Value = (double)(IntArg[0].Value+IntArg[1].Value);
      else 
	DoubleAns->Value = (double)IntArg[0].Value+DoubleArg[1].Value; 
    }
    else if (TypeL == XXX_DOUBLE) {
      if (TypeR == XXX_INT) 
	DoubleAns->Value = DoubleArg[0].Value+(double)(IntArg[1].Value);
      else 
	DoubleAns->Value = DoubleArg[0].Value+DoubleArg[1].Value; 
    }
  }
  return XXX_OK;
}

XxxErrCode XxxArithExpression::
MinusPlan::Evaluate(const XxxRecord* InputRecords[],
		    const XxxRecordSchema* InputSchemas[],
		    XxxADTValue *Result,
		    const XxxValueEnv* Env) const 
{
  XXX_CALL_RET(PreEvaluate(InputRecords, InputSchemas, Result, Env),
	       XXX_BADEXPR, "");

  XxxCoreADTClass::DblValue *DoubleAns = 
    (XxxCoreADTClass::DblValue *)Result;
  XxxCoreADTClass::IntValue *IntAns = 
    (XxxCoreADTClass::IntValue *)Result;

  ASSERT ((OpType == XXX_MINUS_OP));

  if (BasicType() == XXX_INT) {
    if (TypeL == XXX_INT) { 
      if (TypeR == XXX_INT) 
	IntAns->Value = IntArg[0].Value-IntArg[1].Value;
      else 
	IntAns->Value = (int)(IntArg[0].Value-DoubleArg[1].Value); 
    }
    else if (TypeL == XXX_DOUBLE) {
      if (TypeR == XXX_INT) 
	IntAns->Value = (int)(DoubleArg[0].Value-IntArg[1].Value);
      else 
	IntAns->Value = (int)(DoubleArg[0].Value-DoubleArg[1].Value); 
    }
  }
  else if (BasicType() == XXX_DOUBLE) {
    if (TypeL == XXX_INT) { 
      if (TypeR == XXX_INT)
	DoubleAns->Value = (double)(IntArg[0].Value-IntArg[1].Value);
      else 
	DoubleAns->Value = (double)IntArg[0].Value-DoubleArg[1].Value; 
    }
    else if (TypeL == XXX_DOUBLE) {
      if (TypeR == XXX_INT) 
	DoubleAns->Value = DoubleArg[0].Value-(double)(IntArg[1].Value);
      else 
	DoubleAns->Value = DoubleArg[0].Value-DoubleArg[1].Value; 
    }
  }
  return XXX_OK;
}

XxxErrCode XxxArithExpression::
TimesPlan::Evaluate(const XxxRecord* InputRecords[],
		    const XxxRecordSchema* InputSchemas[],
		    XxxADTValue *Result,
		    const XxxValueEnv* Env) const 
{
  XXX_CALL_RET(PreEvaluate(InputRecords, InputSchemas, Result, Env),
	       XXX_BADEXPR, "");

  XxxCoreADTClass::DblValue *DoubleAns = 
    (XxxCoreADTClass::DblValue *)Result;
  XxxCoreADTClass::IntValue *IntAns = 
    (XxxCoreADTClass::IntValue *)Result;

  ASSERT ((OpType == XXX_TIMES_OP));

  if (BasicType() == XXX_INT) {
    if (TypeL == XXX_INT) { 
      if (TypeR == XXX_INT) 
	IntAns->Value = IntArg[0].Value*IntArg[1].Value;
      else 
	IntAns->Value = (int)(IntArg[0].Value*DoubleArg[1].Value); 
    }
    else if (TypeL == XXX_DOUBLE) {
      if (TypeR == XXX_INT) 
	IntAns->Value = (int)(DoubleArg[0].Value*IntArg[1].Value);
      else 
	IntAns->Value = (int)(DoubleArg[0].Value*DoubleArg[1].Value); 
    }
  }
  else if (BasicType() == XXX_DOUBLE) {
    if (TypeL == XXX_INT) { 
      if (TypeR == XXX_INT)
	DoubleAns->Value = (double)(IntArg[0].Value*IntArg[1].Value);
      else 
	DoubleAns->Value = (double)IntArg[0].Value*DoubleArg[1].Value; 
    }
    else if (TypeL == XXX_DOUBLE) {
      if (TypeR == XXX_INT) 
	DoubleAns->Value = DoubleArg[0].Value*(double)(IntArg[1].Value);
      else 
	DoubleAns->Value = DoubleArg[0].Value*DoubleArg[1].Value; 
    }
  }
  return XXX_OK;
}

XxxErrCode XxxArithExpression::
DivPlan::Evaluate(const XxxRecord* InputRecords[],
		  const XxxRecordSchema* InputSchemas[],
		  XxxADTValue *Result,
		  const XxxValueEnv* Env) const 
{
  XXX_CALL_RET(PreEvaluate(InputRecords, InputSchemas, Result, Env),
	       XXX_BADEXPR, "");

  XxxCoreADTClass::DblValue *DoubleAns = 
    (XxxCoreADTClass::DblValue *)Result;
  XxxCoreADTClass::IntValue *IntAns = 
    (XxxCoreADTClass::IntValue *)Result;

  ASSERT ((OpType == XXX_DIV_OP));

  if(((TypeR == XXX_INT) && IntArg[1].Value == 0) ||
     ((TypeR == XXX_DOUBLE) && DoubleArg[1].Value == 0)){
    XXX_NOTE_ERROR(XXX_BADVALUE, "Divide by zero");
    return XXX_NOT_OK;
  }

  if (BasicType() == XXX_INT) {
    if (TypeL == XXX_INT) { 
      if (TypeR == XXX_INT)
	IntAns->Value = IntArg[0].Value/IntArg[1].Value;
      else 
	IntAns->Value = (int)(IntArg[0].Value/DoubleArg[1].Value); 
    }
    else if (TypeL == XXX_DOUBLE) {
      if (TypeR == XXX_INT) 
	IntAns->Value = (int)(DoubleArg[0].Value/IntArg[1].Value);
      else 
	IntAns->Value = (int)(DoubleArg[0].Value/DoubleArg[1].Value); }
  }
  else if (BasicType() == XXX_DOUBLE) {
    if (TypeL == XXX_INT) { 
      if (TypeR == XXX_INT)
	DoubleAns->Value = (double)(IntArg[0].Value/IntArg[1].Value);
      else 
	DoubleAns->Value = (double)IntArg[0].Value/DoubleArg[1].Value; }
    else if (TypeL == XXX_DOUBLE) {
      if (TypeR == XXX_INT) 
	DoubleAns->Value = DoubleArg[0].Value/(double)(IntArg[1].Value);
      else 
	DoubleAns->Value = DoubleArg[0].Value/DoubleArg[1].Value; }
  }
  return XXX_OK;
}

//
// Methods of XxxUnknownValue
//

XxxUnknownValue::XxxUnknownValue (const char *VarName, 
				  const char *NodeName,
				  const XxxBasicType Type, 
				  const int AttrIndex, 
				  const int Child) : XxxValueExpression()
{
  BaseType = Type;
  typeCatInfo = NULL;
  strncpy (Name, VarName, XXX_ATTR_NAME_MAX);
  Name[XXX_ATTR_NAME_MAX-1] = 0;
  if (NodeName) {
    strncpy (XxxName, NodeName, XXX_NODE_NAME_MAX);
    XxxName[XXX_NODE_NAME_MAX-1] = 0;
  }
  else
    XxxName[0] = '\0';
  SrcChild = Child;
  SrcAttrIndex = AttrIndex;
  CorrelationHeight = 0;
}

XxxUnknownValue::XxxUnknownValue(int Index, 
				 const char *NodeName) : XxxValueExpression()
{
  BaseType = XXX_UNKNOWN;
  typeCatInfo = NULL;
  Name[0] = 0;
  if (NodeName) {
    strncpy(XxxName, NodeName, XXX_NODE_NAME_MAX);
    XxxName[XXX_NODE_NAME_MAX-1] = 0;
  }
  else
    XxxName[0] = '\0';
  SrcChild = -1;
  SrcAttrIndex = Index;
  CorrelationHeight = 0;
}

XxxUnknownValue::XxxUnknownValue(int Index, XxxBasicType Type, 
				 XxxADTMetaInfo *CatInfo)  
  : XxxValueExpression()
{
  sprintf(Name, "$%d", Index);
  XxxName[0] = 0;
  SrcChild = 0;
  SrcAttrIndex = Index;
  BaseType = Type;
  typeCatInfo = CatInfo;
  CorrelationHeight = 0;
}

XxxUnknownValue::XxxUnknownValue(const XxxUnknownValue& src):
  XxxValueExpression(src) 
{
  typeCatInfo=NULL;
  if (src.typeCatInfo)
    src.typeCatInfo->Replicate(typeCatInfo);
  strncpy(Name, src.Name, XXX_ATTR_NAME_MAX);
  Name[XXX_ATTR_NAME_MAX-1] = 0;
  strncpy(XxxName, src.XxxName, XXX_NODE_NAME_MAX);
  XxxName[XXX_NODE_NAME_MAX-1] = 0;
  SrcChild = src.SrcChild;
  SrcAttrIndex = src.SrcAttrIndex;
  CorrelationHeight = src.CorrelationHeight;
}

void XxxUnknownValue::Print(ostream& OStr) const 
{
  if (XxxName[0] && Name[0])
    OStr << XxxName << "." << Name;
  else if (XxxName[0])
    OStr << XxxName << ".#" << SrcAttrIndex;
  else if (Name[0])
    OStr <<  Name;
}

XxxBool 
XxxUnknownValue::Equals(const XxxUnknownValue *Var)
{
  if(strcmp(Name, Var->Name))
    return XXX_FALSE;
  if(strcmp(XxxName, Var->XxxName))
    return XXX_FALSE;
  if(SrcChild != Var->SrcChild)
    return XXX_FALSE;
  if(SrcAttrIndex != Var->SrcAttrIndex)
    return XXX_FALSE;
  return XXX_TRUE;
}

XxxErrCode 
XxxUnknownValue::GetDependencies(XxxBitMap& Dependency,
				 int CorrelationHt) const
{
  if (SrcChild < 0) {
    XXX_NOTE_ERROR(XXX_BADEXPR, "Unresolved source of Unknown value");
    return XXX_NOT_OK;
  }
  if (CorrelationHt == CorrelationHeight) {
    // increment by one, because bit 0 represents env args
    Dependency.SetBit(SrcChild+1);
  }
  return XXX_OK;
}

XxxErrCode XxxUnknownValue::Optimize(XxxExprPlan*& ExecPlan)
{
  UnknownValuePlan* UPlan = new UnknownValuePlan(this);
  ExecPlan = UPlan;
  return XXX_OK;
}

XxxErrCode XxxUnknownValue::
UnknownValuePlan::RedirectDependencies(const VarRemapping& Remap,
				       int CorrelationHt)
{
  // be careful and check for bad case
  if (SrcChild < 0) {
    XXX_NOTE_ERROR(XXX_BADOPT, "DANGER!!: Bad SrcChild: shouldnt ever happen");
    return XXX_NOT_OK;
  }

  // no redirection for correlated attributes arising from higher
  // (or lower) query blocks than this source (i.e. the
  // correlation values should match)
  if (this->CorrelationHeight != CorrelationHt) return XXX_OK;

  // first change SrcAttrIndex, then SrcChild (not other way!!)
  if (Remap.SrcMappingArray[SrcChild].NewAttrMapping) {
    // This check makes sure that an attr mapping has been provided.
    // Sometimes, there is no change to the attribute mapping,
    // but only to the SrcChild mapping
    SrcAttrIndex = 
      Remap.SrcMappingArray[SrcChild].NewAttrMapping[SrcAttrIndex];
  }
  SrcChild = Remap.SrcMappingArray[SrcChild].NewSrc;
	
  if (changeDependency(Remap) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADOPT, "Error changing depends mapping");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode XxxUnknownValue::
UnknownValuePlan::Evaluate(const XxxRecord* Rec[],
			   const XxxRecordSchema* Schema[],
			   XxxADTValue *Result,
			   const XxxValueEnv* Env) const 
{
  if (CorrelationHeight > 0) {
    //if (!Env || !Env->NEEnv.Valid || !Env->NEEnv.Env) {
    if (!Env || !Env->NEEnv.Valid) {
      XXX_NOTE_ERROR(XXX_BADEXPR, "No valid environment passed to correlated value");
      return XXX_NOT_OK;
    }
    // need this convoluted mechanism since this is a const method
    ((XxxUnknownValue::UnknownValuePlan*)this)->CorrelationHeight--;
    // recursively invoke the evaluate method
    if (Evaluate(Env->NEEnv.InputRecords, Env->NEEnv.InputSchemas,
		 Result, Env->NEEnv.Env) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXPR, "Error getting correlated value");
      ((XxxUnknownValue::UnknownValuePlan*)this)->CorrelationHeight++;
      return XXX_NOT_OK;
    }
    ((XxxUnknownValue::UnknownValuePlan*)this)->CorrelationHeight++;
    return XXX_OK;
  } // if CorrelationHeight > 0

  const char* FieldPtr;
  Rec[SrcChild]->GetField(SrcAttrIndex, Schema[SrcChild], FieldPtr);
  if (XxxCoreADTClass::IsCoreADT(BasicType())) {
    XxxCoreADTClass::
      STypeCopy(BasicType(), (XxxADTValue*)FieldPtr, Result,
		Schema[SrcChild]->GetAttr(SrcAttrIndex)->TypeCatInfo);
  }
  else {
    // reaches here if it is an ADT
    XxxADTClass* AttrClass = XxxADTArray[BasicType()];
    AttrClass->TypeCopy((XxxADTValue*)FieldPtr, Result, 
			Schema[SrcChild]->GetAttr(SrcAttrIndex)->TypeCatInfo);
    // This is one PRIMARY place where TypeCopy gets called.
    // (The other is in record.C where values are copied into a new record).
    // TODO:L1 There should be matching calls to TypeDestroy when the XxxValuesArray 
    // entry gets overwritten and when the record gets deleted
  } // if it is not a basic type
  return XXX_OK;  
}

XxxErrCode
XxxUnknownValue::ResolveVars(const XxxRecordSchema *Schema)
{
  XxxBasicType Type;
	
  // once the value is resolved, do not try resolving it again!
  if (BaseType != XXX_UNKNOWN)
    return XXX_OK;

  if (Schema->GetAttrType(Name, Type) != XXX_OK)
    {
      XXX_NOTE_ERROR (XXX_BADEXPR, "Bad attribute.");
      return XXX_NOT_OK;
    }
	
  BaseType = Type;
	
  SrcChild = 0;
  Schema->GetAttrCol(Name, SrcAttrIndex);
	
  if ((*Schema)[SrcAttrIndex].TypeCatInfo) {
    (*Schema)[SrcAttrIndex].TypeCatInfo->Replicate(typeCatInfo);
  }
	
  return XXX_OK;
}

//AUTHOR: Fabian Camargo DATE: 2/7/97
//PURPOSE: Check to see if expression matches an index expression 
//NOTES: The Env char* parameter is used to carry a relation name binding
// across multiple invocations of the Match() method.
// Consider the case that f(R.x, R.y) is the indexing expression, which
// has to be matched with f(A.x, A.y). First, A.x::Match(R.x, Env) is
// called. At this stage, Env is the empty string. The match succeeds and 
// Env == "A". Now A.y::Match(R.y, Env) gets called. At this point, 
// since Env and A match, the attrnames are checked and the match 
// succeeds. Try this matching f(A.x, B.y) instead -- this fails.
// This is a special case of a general deep matching with a binding
// environment. In our case, the binding environment is guaranteed to
// have only one variable, since indexes can only be built on a single
// table.
XxxBool 
XxxUnknownValue::Match(const XxxExpression* OtherExpr,
		       char *Env) const
{
  if (!(XxxValueExpression::Match(OtherExpr, Env)))
    return XXX_FALSE;
  XxxUnknownValue *Other = (XxxUnknownValue *)OtherExpr;
  if (Env) {
    // This invocation is being used with the environment to match
    // multiple expressions for indexes.
    if (!strcmp(Env, ""))
      // Env not defined. Assign it the relation's name for later comparison
      strcpy(Env, XxxName);
    else {
      // Env previously defined. See if it matches relation name
      if (strcmp(Env, XxxName))
	return XXX_FALSE;
    }
    if (!strcmp(Name, Other->Name))
      return XXX_TRUE;
  } else {
    /*
     * This is a standard invocation of the function. If either expression 
     * is named and the other references that name, they match.
     */
    if (ExprName && Other->ExprName && (!strcmp(ExprName, Other->ExprName)))
      return XXX_TRUE;
    if (ExprName && (!strcmp(ExprName, Other->Name)))
      return XXX_TRUE;
    if (Other->ExprName && (!strcmp(Other->ExprName, Name)))
      return XXX_TRUE;
    /*
     * If the either expression's XxxName is unresolved or they match,
     * check the VarNames.
     */
    else if (!strcmp(Other->XxxName, "") || !strcmp(Other->XxxName, "") ||
	     !strcmp(XxxName, Other->XxxName)) {
      if (!strcmp(Name, Other->Name))
	return XXX_TRUE;
    }
  }
  return XXX_FALSE;
}

void XxxArgValue::Print(ostream& OStr) const 
{
  OStr << "$" << ArgIndex;
}

XxxErrCode XxxArgValue::Optimize(XxxExprPlan*& ExecPlan)
{
  XxxValueExprPlan *TmpPlan;
  XXX_CALL_RET(Expr->Optimize(TmpPlan),
	       XXX_BADOPT, "Error optimizing subexp");
  ExecPlan = new ArgValuePlan(this, TmpPlan);
  return XXX_OK;
}

XxxBool 
XxxArgValue::Match(const XxxExpression* OtherExpr,
		   char* Env) const
{
  if (!(XxxValueExpression::Match(OtherExpr, Env)))
    return XXX_FALSE;
  XxxArgValue *CastExpr = (XxxArgValue *)OtherExpr;
	
  return ((ArgIndex == CastExpr->ArgIndex) && 
	  Expr->Match(CastExpr->Expr,Env));
} 

XxxErrCode XxxArgValue::
ArgValuePlan::Evaluate(const XxxRecord* InputRecords[],
		       const XxxRecordSchema* InputSchemas[],
		       XxxADTValue *Result,
		       const XxxValueEnv* Env) const 
{
  if (ArgIndex < Env->NumArgs) {
    if (XxxCoreADTClass::IsCoreADT(BasicType())) {
      XxxCoreADTClass::
	STypeCopy(BasicType(), 
		  Env->ArgValsPtrs[ArgIndex],
		  Result,
		  ArgCatInfo);
    }
    else {
      XxxADTArray[BasicType()]->
	TypeCopy(Env->ArgValsPtrs[ArgIndex],
		 Result, ArgCatInfo);
    }
    return XXX_OK;
  }
  else {
    XXX_NOTE_ERROR(XXX_BADEXPR, "Invalid Host Variable");
    return XXX_NOT_OK;
  }
}

void XxxConstIntValue::Print(ostream& OStr) const 
{
  OStr << i;
}

void XxxConstDoubleValue::Print(ostream& OStr) const 
{
  OStr << d;
}

void XxxConstStringValue::Print(ostream& OStr) const 
{
  OStr << '\"' << c << '\"';
}

XxxBool
XxxConstValue::Match(const XxxExpression *OtherExpr,
		     char* Env) const 
{
  if (!(XxxValueExpression::Match(OtherExpr, Env)))
    return XXX_FALSE;
  return (BasicType() == ((XxxValueExpression *)OtherExpr)->BasicType());
}

XxxBool
XxxConstValue::Match(const XxxExpression *RuleExpr,
		     XxxBindingEnv *BindEnv) const 
{
  // this match will do any binding that needs to happen
  return XxxValueExpression::Match(RuleExpr, BindEnv);
}

XxxBool 
XxxConstIntValue::Match(const XxxExpression *OtherExpr,
			char* Env) const 
{
  if (!(XxxConstValue::Match(OtherExpr, Env)))
    return XXX_FALSE;
  XxxConstIntValue *CastExpr = (XxxConstIntValue *)OtherExpr;
  if (i != CastExpr->i)
    return XXX_FALSE;
  return XXX_TRUE;
}


XxxBool 
XxxConstDoubleValue::Match(const XxxExpression *OtherExpr,
			   char* Env) const 
{
  if (!(XxxConstValue::Match(OtherExpr, Env)))
    return XXX_FALSE;
  XxxConstDoubleValue *CastExpr = (XxxConstDoubleValue *)OtherExpr;
  if (d != CastExpr->d)
    return XXX_FALSE;
  return XXX_TRUE;
}

XxxBool 
XxxConstStringValue::Match(const XxxExpression *OtherExpr, 
			   char* Env) const 
{
  if (!(XxxConstValue::Match(OtherExpr, Env)))
    return XXX_FALSE;
  XxxConstStringValue *CastExpr = (XxxConstStringValue *)OtherExpr;
  if (strcmp(c, CastExpr->c))
    return XXX_FALSE;
  return XXX_TRUE;
}

// try to resolve this var using the schema of Node or its children
XxxErrCode XxxUnknownValue::ResolveVars(XxxQueryNode* Node, 
					char Source)
{
  // once the value is resolved, do not try resolving it again!
  if (BaseType != XXX_UNKNOWN)
    return XXX_OK;
	
  int Tmp;
  if ((Tmp = Node->FindValueAttrIndex(this)) >= 0) {
    SrcChild = Source;
    // Either of the following was previously undefined, so copy both.
    SrcAttrIndex = Tmp;
    strncpy(Name, (*(Node->Schema()))[SrcAttrIndex].Name, XXX_ATTR_NAME_MAX);
    Name[XXX_ATTR_NAME_MAX-1] = 0;
		
    BaseType = (*(Node->Schema()))[SrcAttrIndex].Type;
		
    if ((*(Node->Schema()))[SrcAttrIndex].TypeCatInfo) {
      (*(Node->Schema()))[SrcAttrIndex].TypeCatInfo->Replicate(typeCatInfo);
    }
    else 
      typeCatInfo = NULL;
  }
  return XXX_OK;
}

// used for correlation variables
XxxErrCode XxxUnknownValue::ResolveVars(const XxxArgEnv* Env)
{
  // once the value is resolved, do not try resolving it again!
  if (BaseType != XXX_UNKNOWN)
    return XXX_OK;
  if (!Env) return XXX_OK;
	
  // Go thru the correlation environment and check if it provides the
  // necessary correlation
  // Process Env->NEEnv: first the children, then its env:
  CorrelationHeight++;
  for (int i = 0; i < Env->NPEnv.NumChildren; i++) {
    if (ResolveVars(Env->NPEnv.Children[i], i) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXPR, "Error resolving correlation value");
      return XXX_NOT_OK;
    }
    if (BaseType != XXX_UNKNOWN)
      break;
  } // for
	
  if (BaseType == XXX_UNKNOWN) {
    // try to use the environment of Env->NEEnv (recursion to move up
    // the query block nesting)
    if (ResolveVars(Env->NPEnv.Env) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXPR, "Error resolving correlation value");
      return XXX_NOT_OK;
    }
  }
  // if the value was not resolved from the correlation environment, reset it
  if (BaseType == XXX_UNKNOWN)
    CorrelationHeight--;
	
  return XXX_OK;
}

XxxErrCode XxxUnknownValue::UpdateUnknowns(int ChildNumber,
					   int OldAttrIndex,
					   int NewAttrIndex,
					   XxxBool &Appears)
{
  if(SrcChild == ChildNumber && SrcAttrIndex == OldAttrIndex) {
    SrcAttrIndex = NewAttrIndex;
    Appears = XXX_TRUE;
  }
  return XXX_OK;
}

//
// Methods for XxxArgValue
//

XxxArgValue::XxxArgValue(const XxxArgValue& src):XxxValueExpression(src) {
  ArgIndex=src.ArgIndex;
  Expr=NULL;
  if (src.Expr)
    Expr=(XxxValueExpression*) src.Expr->Replicate();
}

// try to resolve this host var using the arg environment
XxxErrCode XxxArgValue::ResolveVars(const XxxArgEnv* Env)
{
  if (Env && (ArgIndex <= Env->ExprCount)) {
    Expr = (*Env)[ArgIndex-1];
    BaseType = Expr->BasicType();
  }
  return XXX_OK;
}

/////////////////////////////////////////////////////

XxxValueEnv::XxxValueEnv(XxxValueExprList* HostExprs, int SizeOfOwner,
			 XxxADTMetaInfo *OwnerTypeCatInfo)
  : ArgVals(HostExprs)
{
  OwnerSize = SizeOfOwner;
  int i = 0;
  NumArgs = (HostExprs?HostExprs->Count():0);
	
  if (OwnerSize > 0) {
    OwnerValPtr = (XxxADTValue *)(new char[OwnerSize]);
  }
  else {
    OwnerValPtr = NULL;
  }
	
  OwnerCatInfo = OwnerTypeCatInfo;
	
  // Make pointers to the storage in ArgVals
  if (NumArgs) {
    ArgValsPtrs = new XxxADTValue*[NumArgs];
    for (i =0; i < NumArgs; i++) {
      ArgValsPtrs[i] = (XxxADTValue *)(ArgVals.Values[i]);
    }
  }
  else
    ArgValsPtrs = NULL;
}

XxxValueEnv::XxxValueEnv(const XxxValueEnv &Env) :
  ArgVals(Env.ArgVals), OwnerSize(Env.OwnerSize), OwnerValPtr(NULL),
  NumArgs(Env.NumArgs), OwnerCatInfo(NULL), NEEnv(Env.NEEnv)
{
  int i;
  // Make pointers to the storage in ArgVals
  ArgValsPtrs = new XxxADTValue*[NumArgs];
  for (i = 0; i < NumArgs; i++) {
    ArgValsPtrs[i] = (XxxADTValue *)(ArgVals.Values[i]);
  }
  // Copy the owner
  if (OwnerSize > 0) {
    OwnerValPtr = (XxxADTValue *)(new char[OwnerSize]);
    memcpy(OwnerValPtr, Env.OwnerValPtr, OwnerSize);
  }
  // Copy its catalog info
  if (Env.OwnerCatInfo != NULL) {
    if (Env.OwnerCatInfo->Replicate(OwnerCatInfo) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADADT, "Cannot replicate catalog info");
      OwnerCatInfo = NULL;
    }
  }
}

XxxValueEnv::~XxxValueEnv()
{
  if (ArgValsPtrs) delete [] ArgValsPtrs;
  if (OwnerSize>0 && OwnerValPtr) delete [] ((char *)OwnerValPtr);
}
